/* --------------------------------------------------------------------------
 *
 * class BSplineBasis
 *   B-spline basis computations handler
 *
 * required headers
 *   see include header file BSplineBasis
 *
 * acknowledgments
 *   many discussions with Chris Genovese
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __BSplineBasis_H__
#define __BSplineBasis_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class BSplineBasis
 {
  public:

   // --------------------------------------------------- construct/reinitialize
   // --------------------------------------------------------------------------

   BSplineBasis( const fpVector& knots,
                       int       bsOrder,
                       int       leftEndPoint,
                       int       rightEndPoint,
                       bool      rightEndLeftContinuous = true,
                       bool      dropLastBasisFn        = false );
                            //
                            //      knot sequence is adjusted to make a basis of
                            //  order bsOrder possible over the interval defined
                            //    by knots[leftEndPoint] and knots[rightEndoint]
                            //           by adding multiple end knots, if needed
                            //
                            //      This "centering" of the interval of interest
                            //   within the adjuested knot sequence ensures that
                            //         the basis forms a partition of unity over
                            //                                     this interval
                            //
                            //    the flag rightEndLeftContinuous applies to the
                            //   extreme right knot value (after adjustment): if
                            // true, spline values and derivatives at this point
                            //                    are calculated left-continuous
                            //
                            //   dropLastBasisFn == true: the coefficient of the
                            //   last B-spline is taken to be be zero, and basis
                            //        size is reduced by 1 (similar in spirit to
                            //                      noIntercept = true in Splus)
                            //
                            //                                       ASSUMPTIONS
                            //
                            //       knots[0] <= ... <= knots[knots.size( ) - 1]
                            //                                 knots.size( ) > 1
                            // 0 <= leftEndPoint < rightEndPoint < knots.size( )

   BSplineBasis( const BSplineBasis& another );

   BSplineBasis( );
                            //
                            //                                    default values
                            //
                            //                      knots: { 0, 1 }, bsOrder: 1,
                            //                leftEndPoint: 0, rightEndPoint: 1,
                            //                     rightEndLeftContinuous: true,
                            //                            dropLastBasisFn: false

   void
   reinitialize( const fpVector& knots,
                       int       bsOrder,
                       int       leftEndPoint,
                       int       rightEndPoint,
                       bool      rightEndLeftContinuous = true,
                       bool      dropLastBasisFn        = false );

   // --------------------------------------------------------- basis attributes
   // --------------------------------------------------------------------------

   inline
   int
   order( ) const;                                // order of the B-spline basis

   inline
   int
   numKnots( ) const;                                        // knots( ).size( )

   inline
   const fpVector&
   knots( ) const;                                     // complete knot sequence

   inline
   int
   leftEndPoint( ) const;

   inline
   int
   rightEndPoint( ) const;
                                     //
                                     //   end points of the interval of interest
                                     // returned as knot indices within knots( )
                                     //
                                     //    because of the adjustment of the knot
                                     //         sequence made by the constructor
                                     //             partition of unity holds for
                                     //   knots( )[leftEndPoint( )] <= abscissa
                                     //            <= knots( )[rightEndPoint( )]

   inline
   bool
   isRightEndLeftContinuous( ) const;                   // rightmost knot status

   inline
   bool
   isDropLastBasisFn( ) const;      // has the last basis function been dropped?

   inline
   int
   basisSize( ) const;
                                       //
                                       // Total # of B-spline basis functions ==
                                       //                 numKnots( ) - order( )
                                       //
                                       // individual B-splines are labelled from
                                       //                  0 to basisSize( ) - 1

   // ----------------------------------------- spline and derivative evaluation
   // --------------------------------------------------------------------------

                                                //                   ASSUMPTIONS
                                                //
                                                //   0 <= bsIndex < basisSize( )
                                                // coeff.size( ) == basisSize( )

   inline
   fpType
   eval( const fpType* coeff, fpType abscissa, int dOrder = 0 ) const;
                                    //
                                    // evaluate dOrder-th derivative of a spline
                                    // defined by coefficients coeff at abscissa

   inline
   fpType
   eval( int bsIndex, fpType abscissa, int dOrder = 0 ) const;
                                //
                                // same as above, except for one single B-spline

   // --------------------------------------------------- single spline integral
   // --------------------------------------------------------------------------

   inline
   fpType
   evalIntegral( const fpType* coeff,
                 fpType lowerLimit, fpType upperLimit ) const;
                                    //
                                    // integral of a spline between given limits

   inline
   fpType
   evalIntegral( int bsIndex, fpType lowerLimit, fpType upperLimit ) const;
                                  //
                                  // integral of a B-spline between given limits

   inline
   fpType
   evalIntegral( int bsIndex ) const;
                                  //
                                  // B-spline integrated over the entire support

   // ---------------------------------------------- two-spline product integral
   // --------------------------------------------------------------------------

   inline
   fpType
   evalIntegral( const fpType* coeff1,
                 const fpType* coeff2,
                       fpType  lowerLimit,
                       fpType  upperLimit,
                       int     dOrder = 0 ) const;
                                  //
                                  // integral of a product of derivatives of two
                                  //                splines between given limits
                                  //
                                  //                  0 <= dOrder < order( ) - 1

   inline
   fpType
   evalIntegral( int    bsIndex1,
                 int    bsIndex2,
                 fpType lowerLimit,
                 fpType upperLimit,
                 int    dOrder = 0 ) const;
                                  //
                                  // integral of a product of derivatives of two
                                  //              B-splines between given limits
                                  //
                                  //                  0 <= dOrder < order( ) - 1

   inline
   void
   evalIntegral( fpType    lowerLimit,
                 fpType    upperLimit,
                 fpMatrix& productIntegrals,
                 int       dOrder = 0 ) const;
                                      //
                                      // all integrals of B-spline pair products
                                      //
                                      //              0 <= dOrder < order( ) - 1

   inline
   int
   maxHalfBandWidth( ) const;     // maximum number of nonzero product integrals
                                  //    for any B-spline on one side of diagonal
                                  //
                                  // notice that the matrix of product integrals
                                  // is symmetric, and banded because of compact
                                  //                        support of B-splines

   inline
   int
   bandMin( int bsIndex ) const;  //    minimum index for a B-spline to have any
                                  //     nonzero product integral with the given
                                  //                                    B-spline
                                  //
                                  //                returns 0 if bsIndex invalid

   inline
   int
   bandMax( int bsIndex ) const;  //    maximum index for a B-spline to have any
                                  //     nonzero product integral with the given
                                  //                                    B-spline
                                  //
                                  // returns basisSize( ) - 1 if bsIndex invalid

   // -------------------------------------------------- miscellaneous utilities
   // --------------------------------------------------------------------------

   inline
   bool
   hasSupport( int bsIndex ) const;
                                        //
                                        //    true if the given B-spline has any
                                        //  support over the interval defined by
                                        //  leftEndPoint( ) and rightEndPoint( )

   inline
   int
   lowerSupport( int bsIndex ) const;

   inline
   int
   upperSupport( int bsIndex ) const;
                          //
                          // support of a specific B-spline of the default order
                          //         returned as an index into the knot sequence
                          //
                          //                      lowerSupport( ... ) := bsIndex
                          //           upperSupport( ... ) := bsIndex + order( )
                          //
                          //     the values returned by these two functions make
                          //         sense only if hasSupport( bsIndex ) == true
                          //
                          //                         0 <= bsIndex < basisSize( )

   inline
   bool
   isInSupport( fpType abscissa, int bsIndex ) const;
                            //
                            //   true if abscissa is in the support of the given
                            //                                          B-spline
                            //
                            //    The right end point is handled consistent with
                            //                       isRightEndLeftContinuous( )

   inline
   bool
   isInSupport( fpType abscissa ) const;
                            //
                            // true if  knots( )[0] <= abscissa
                            //      && abscissa     <  knots( )[numKnots( ) - 1]
                            //       (for isRightEndLeftContinuous( ) == false )
                            //      && abscissa     <= knots( )[numKnots( ) - 1]
                            //       (for isRightEndLeftContinuous( ) == true  )

   inline
   pair<int, int>
   bracket( fpType abscissa ) const;
               //
               // abscissa                                               returns
               //
               // abscissa < knots( )[0]                                 (0, -1)
               //
               // knots( )[0] <= abscissa &&
               // absicssa <  knots( )[numKnots( ) - 1]
               //             (n : knots( )[n] <= abscissa < knots( )[n + 1], 0)
               //
               // abscissa == knots( )[numKnots( ) - 1]
               // && rightEndLeftContinuous == true
               //             (n : knots( )[n] < abscissa <= knots( )[n + 1], 0)
               //
               // abscissa == knots( )[numKnots( ) - 1]
               // && rightEndLeftContinuous == false       (numKnots( ) - 1, +1)
               //
               // abscissa > knots( )[numKnots( ) - 1]     (numKnots( ) - 1, +1)
               //
               //    first of the returned pair is the left bracket for abscissa
               //        second is its out-of-range status: +/-1 if out of range

   inline
   bool
   isPartitionOfUnity( fpType abscissa ) const;
                                    //
                                    // true if partition of unity property holds
                                    //                     at the given abscissa
                                    //
                                    //                                     NOTES
                                    //
                                    //                             see NOTES for
                                    // isPartitionOfUnity( intervalIndex ) below

                                       //      utility pertaining to an interval
                                       // defined by a pair of consecutive knots

   inline
   bool
   isPartitionOfUnity( int intervalIndex ) const;
                                    //
                                    // true if partition of unity property holds
                                    //               for the interval defined as
                                    //
                                    //               knots( )[intervalIndex]
                                    //            <= abscissa
                                    //            <  knots( )[intervalIndex + 1]
                                    //
                                    //                               ASSUMPTIONS
                                    //
                                    //      0 <= intervalIndex < numKnots( ) - 1
                                    //
                                    //                                     NOTES
                                    //
                                    // due to adjustment of the knot sequence by
                                    // the constructor, PoU holds over intervals
                                    //  leftEndPoint( ) ... rightEndPoint( ) - 1

   // ------------------------------------------------------------------ private
   // --------------------------------------------------------------------------

  private:
   fpVector my_knots;
   int      my_bsOrder;
   bool     my_rightEndLeftContinuous;
   bool     my_dropLastBasisFn;

   inline
   void
   initKnots( const fpVector& knots,
                    int       leftEndPoint,
                    int       rightEndPoint );
                                    //
                                    // adjust knot sequence so that partition of
                                    //  unity holds between the given end points

   inline
   void
   getProductKnots(       int       currentOrder,
                    const fpVector& currentKnots,
                          int&      newOrder,
                          fpVector& newKnots,
                          int       dOrder = 0 ) const;
                                  //
                                  // knot sequence adjusted to represent product
                                  //   of two splines as a spline of appropriate
                                  //  order, preserving the number of smoothness
                                  //                                  conditions
                                  //
                                  // used by evalIntegral( coeff1, coeff2, ... )

   inline
   void
   getProductCoeff( const fpType*   coeff1,
                    const fpType*   coeff2,
                          int       pOrder,
                    const fpVector& productKnots,
                          fpVector& productCoeff,
                          int       dOrder = 0 ) const;
                                  //
                                  //    product of two splines coeff1 and coeff2
                                  //      represented as a spline of appropriate
                                  //     order over an appropriate knot sequence
                                  //
                                  //                                 ASSUMPTIONS
                                  //
                                  //           productKnots is the knot sequence
                                  //             generated by getProductKnots( )
                                  //
                                  // used by evalIntegral( coeff1, coeff2, ... )
 };

// ------------------------------------------------------------ basis attributes
// -----------------------------------------------------------------------------

int
BSplineBasis::order( ) const
 {
  return my_bsOrder;
 }

int
BSplineBasis::numKnots( ) const
 {
  return my_knots.size( );
 }

const fpVector&
BSplineBasis::knots( ) const
 {
  return my_knots;
 }

int
BSplineBasis::leftEndPoint( ) const
 {
  return ( order( ) - 1 );
 }

int
BSplineBasis::rightEndPoint( ) const
 {
  return ( numKnots( ) - order( ) );
 }

bool
BSplineBasis::isRightEndLeftContinuous( ) const
 {
  return my_rightEndLeftContinuous;
 }

bool
BSplineBasis::isDropLastBasisFn( ) const
 {
  return my_dropLastBasisFn;
 }

int
BSplineBasis::basisSize( ) const
 {
  return
   (
       BSplineTools::basisSize( order( ), numKnots( ) )
     - isDropLastBasisFn( )
   );
 }

// -------------------------------------------- spline and derivative evaluation
// -----------------------------------------------------------------------------

fpType
BSplineBasis::eval( const fpType* coeff,
                          fpType  abscissa,
                          int     dOrder ) const
 {
  if ( !isInSupport( abscissa ) ) return 0.0;

  return BSplineTools::eval( coeff, abscissa, bracket( abscissa ).first,
                             order( ), knots( ), dOrder );
 }

fpType
BSplineBasis::eval( int    bsIndex,
                    fpType abscissa,
                    int    dOrder ) const
 {
  if ( !isInSupport( abscissa, bsIndex ) ) return 0.0;

  fpVector coeff( 0.0, basisSize( ) + isDropLastBasisFn( ) );
  coeff[bsIndex] = 1.0;

  return BSplineTools::eval( coeff( 0 ), abscissa, bracket( abscissa ).first,
                             order( ), knots( ), dOrder );
 }

// ------------------------------------------------------ single spline integral
// -----------------------------------------------------------------------------

fpType
BSplineBasis::evalIntegral( const fpType* coeff,
                                  fpType  lowerLimit,
                                  fpType  upperLimit ) const
 {
  return BSplineTools::evalIntegral( coeff, lowerLimit, upperLimit,
                                     bracket( lowerLimit ).first,
                                     bracket( upperLimit ).first,
                                     order( ), knots( ) );
 }

fpType
BSplineBasis::evalIntegral( int    bsIndex,
                            fpType lowerLimit,
                            fpType upperLimit ) const
 {
  fpVector coeff( 0.0, basisSize( ) + isDropLastBasisFn( ) );
  coeff[bsIndex] = 1.0;

  return BSplineTools::evalIntegral( coeff( 0 ), lowerLimit, upperLimit,
                                     bracket( lowerLimit ).first,
                                     bracket( upperLimit ).first,
                                     order( ), knots( ) );
 }

fpType
BSplineBasis::evalIntegral( int bsIndex ) const
 {
  return BSplineTools::evalIntegral( bsIndex, order( ), knots( ) );
 }

// ------------------------------------------------- two-spline product integral
// -----------------------------------------------------------------------------

fpType
BSplineBasis::evalIntegral( const fpType* coeff1,
                            const fpType* coeff2,
                                  fpType  lowerLimit,
                                  fpType  upperLimit,
                                  int     dOrder ) const
 {
  int pOrder;
  fpVector pKnots;
  getProductKnots( order( ), knots( ), pOrder, pKnots, dOrder );
                              //
                              // adjusted knot sequence for representing product
                              //  of two splines of order order( ) - dOrder as a
                              //   spline of order 2 * ( order( ) - dOrder ) - 1
                              //   preserving smoothness conditions at the knots

  fpVector pCoeff;
  getProductCoeff( coeff1, coeff2, pOrder, pKnots, pCoeff, dOrder );
                                   //
                                   //    spline product expressed as a spline of
                                   // order pOrder over the knot sequence pKnots

                                     //   Finally, evaluate the product integral
                                     //
  return BSplineTools::evalIntegral( pCoeff( 0 ), lowerLimit, upperLimit,
                                     BSplineTools::bracket( lowerLimit, pKnots,
                                            isRightEndLeftContinuous( ) ).first,
                                     BSplineTools::bracket( upperLimit, pKnots,
                                            isRightEndLeftContinuous( ) ).first,
                                     pOrder, pKnots );
 }

fpType
BSplineBasis::evalIntegral( int    bsIndex1,
                            int    bsIndex2,
                            fpType lowerLimit,
                            fpType upperLimit,
                            int    dOrder ) const
 {
  if (    (   max( bsIndex1, bsIndex2 )
            - min( bsIndex1, bsIndex2 ) >= order( ) )
       || (   dOrder                    >= order( ) ) ) return 0.0;

  fpVector a1( 0.0, basisSize( ) );
  a1[bsIndex1] = 1.0;

  fpVector a2( 0.0, basisSize( ) );
  a2[bsIndex2] = 1.0;

  return
   evalIntegral( a1( 0 ), a2( 0 ), lowerLimit, upperLimit, dOrder );
 }

void
BSplineBasis::evalIntegral( fpType    lowerLimit,
                            fpType    upperLimit,
                            fpMatrix& productIntegrals,
                            int       dOrder ) const
 {
  productIntegrals.resize( basisSize( ), basisSize( ), false, 0.0 );

  if ( ( dOrder >= 0 ) && ( dOrder < order( ) ) )
   {
    int pOrder;
    fpVector pKnots;
    getProductKnots( order( ), knots( ), pOrder, pKnots, dOrder );

    int lowerBracket = BSplineTools::bracket( lowerLimit, pKnots,
                                              isRightEndLeftContinuous( ) ).first;
    int upperBracket = BSplineTools::bracket( upperLimit, pKnots,
                                              isRightEndLeftContinuous( ) ).first;

    fpVector integrals;
    BSplineTools::evalIntegral( lowerLimit, upperLimit,
                                lowerBracket, upperBracket,
                                pOrder, pKnots, integrals );

    fpVector a1( 0.0, basisSize( ) );
    fpVector a2( 0.0, basisSize( ) );

    fpVector pCoeff;

    for ( int i = 0; i < basisSize( ); i++ )
     {
      if ( i > 0 ) a1[i-1] = 0.0; a1[i] = 1.0;

      getProductCoeff( a1( 0 ), a1( 0 ), pOrder, pKnots, pCoeff, dOrder );

      productIntegrals( i, i ) = BlasDrivers::dot( integrals, pCoeff );

      for ( int k = 0; k < i; k++ ) a2[k] = 0.0;

      for ( int j = bandMin( i ); j < i; j++ )
       {
        if ( j > 0 ) a2[j-1] = 0.0; a2[j] = 1.0;

        getProductCoeff( a1( 0 ), a2( 0 ), pOrder, pKnots, pCoeff, dOrder );

        productIntegrals( i, j ) = productIntegrals( j, i ) = BlasDrivers::dot( integrals, pCoeff );
       }
     }
   }
 }

int
BSplineBasis::maxHalfBandWidth( ) const
 {
  return ( order( ) - 1 );
 }

int
BSplineBasis::bandMin( int bsIndex ) const
 {
  return max( 0, bsIndex - maxHalfBandWidth( ) );
 }

int
BSplineBasis::bandMax( int bsIndex ) const
 {
  return min( basisSize( ) - 1, bsIndex + maxHalfBandWidth( ) );
 }

// ----------------------------------------------------- miscellaneous utilities
// -----------------------------------------------------------------------------

bool
BSplineBasis::hasSupport( int bsIndex ) const
 {
  return
   (
       ( 0       <= bsIndex      )
    && ( bsIndex <  basisSize( ) )
   );
 }

int
BSplineBasis::lowerSupport( int bsIndex ) const
 {
  return BSplineTools::lowerSupport( bsIndex );
 }

int
BSplineBasis::upperSupport( int bsIndex ) const
 {
  return BSplineTools::upperSupport( bsIndex, order( ) );
 }

bool
BSplineBasis::isInSupport( fpType abscissa,
                           int    bsIndex ) const
 {
  return
   (
       hasSupport( bsIndex )
    && ( knots( )[lowerSupport( bsIndex )] <= abscissa )
    && (
           (    isRightEndLeftContinuous( )
             && ( abscissa == knots( )[numKnots( ) - 1] )  )
         ? ( abscissa <= knots( )[upperSupport( bsIndex )] )
         : ( abscissa <  knots( )[upperSupport( bsIndex )] )
       )
   );
 }

bool
BSplineBasis::isInSupport( fpType abscissa ) const
 {
  return
   (
       ( knots( )[0] <= abscissa )
    && (
           (    isRightEndLeftContinuous( )
             && ( abscissa == knots( )[numKnots( ) - 1] ) )
         ? ( true                                         )
         : ( abscissa < knots( )[numKnots( ) - 1]         )
       )
   );
 }

pair<int, int>
BSplineBasis::bracket( fpType abscissa ) const
 {
  return
   BSplineTools::bracket( abscissa, knots( ), isRightEndLeftContinuous( ) );
 }

bool
BSplineBasis::isPartitionOfUnity( fpType abscissa ) const
 {
  return isPartitionOfUnity( bracket( abscissa ).first );
 }

bool
BSplineBasis::isPartitionOfUnity( int intervalIndex ) const
 {
  return
   BSplineTools::isPartitionOfUnity( intervalIndex, order( ), numKnots( ) );
 }

// --------------------------------------------------------------------- private
// -----------------------------------------------------------------------------

void
BSplineBasis::initKnots( const fpVector& knots,
                               int       leftEndPoint,
                               int       rightEndPoint )
 {
  int lep = max( 0, leftEndPoint );
  int rep = min( ( (int) knots.size( ) ) - 1, rightEndPoint );

  if ( lep >= rep )
   {
    lep = 0;
    rep = ( (int) knots.size( ) ) - 1;         // ASSUMPTION : knots.size( ) > 1
   }

  BSplineTools::adjustEndKnots( lep, rep, my_bsOrder, knots, my_knots );
 }

void
BSplineBasis::getProductCoeff( const fpType*   coeff1,
                               const fpType*   coeff2,
                                     int       pOrder,
                               const fpVector& pKnots,
                                     fpVector& pCoeff,
                                     int       dOrder ) const
 {
  pCoeff.resize( BSplineTools::basisSize( pOrder, pKnots.size( ) ) );

  fpType factor = 1.0;
  for ( int n = 2; n < pOrder; n++ ) factor /= ( (fpType) n );

  fpType tau;
  int leftBracket;

  for ( int i = 0;
            i < BSplineTools::basisSize( pOrder, pKnots.size( ) );
            i++ )
   {
    tau = pKnots[BSplineTools::lowerSupport( i )];
                                //
                                //  choice of tau is supposed to be arbitrary so
                                // long as it is in the support of i-th B-spline

    leftBracket = bracket( tau ).first;

    for ( int n = 0; n < pOrder; n++ )
     pCoeff[i] += (
                     BSplineTools::evalProduct( coeff1, coeff2,
                                                tau, leftBracket,
                                                order( ), knots( ),
                                                pOrder - 1 - n,
                                                dOrder, dOrder )
                   * BSplineTools::DnPsi( tau, i, pOrder, pKnots, n )
                 );

    pCoeff[i] *= factor;
   }
 }

void
BSplineBasis::getProductKnots(       int       currentOrder,
                               const fpVector& currentKnots,
                                     int&      newOrder,
                                     fpVector& newKnots,
                                     int       dOrder ) const
 {
  currentOrder -= dOrder;
  newOrder      = 2 * currentOrder - 1;

                               // compute the size of the required knot sequence
                               //
  int numNewKnots = 0;
  for ( int knotIndex = 0, m = 0; knotIndex < currentKnots.size( ); )
   {
    m = BSplineTools::knotMultiplicity( knotIndex, currentKnots );

    numNewKnots += min( newOrder, m + currentOrder - 1 );
    knotIndex   += m;
   }

  newKnots.resize( numNewKnots );

                        // copy knots with proper multiplicity so as to preserve
                        // the correct number of smoothness conditions at a knot
                        //
  int newKnotIndex = 0;
  for ( int knotIndex = 0, m = 0, mNew = 0; knotIndex < currentKnots.size( ); )
   {
    m    = BSplineTools::knotMultiplicity( knotIndex, currentKnots );
    mNew = min( newOrder, m + currentOrder - 1 );

    for ( int i = 0; i < mNew; i++ )
     newKnots[newKnotIndex + i] = currentKnots[knotIndex];

    newKnotIndex += mNew;
    knotIndex    += m;
   }
 }

#endif /* __BSplineBasis_H__ */
