/* --------------------------------------------------------------------------
 *
 * namespace BSplineTools
 *   general tools for B-spline computations
 *
 * required headers
 *   see include header file BSplineTools
 *
 * acknowledgments
 *   many discussions with Chris Genovese
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __BSplineTools_H__
#define __BSplineTools_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

cpp_namespace_begin( BSplineTools )

   // --------------------------------------------------------------------------
   //                          a toolbox containing some B-splines-related tools
   //
   //                                                          useful references
   //
   //           Carl de Boor, A Practical Guide to Splines, Springer-Verlag 1978
   //
   //                                      Carl de Boor, "B(asic)-Spline Basics"
   //         in: Fundamental Developments of Computer-Aided Geometric Modelling
   //                                       Les Piegl (ed.), Academic Press 1993
   //
   //                              http://www.cs.wisc.edu/~deboor/ftpreadme.html
   // --------------------------------------------------------------------------

   // -------------------------------------------------------------- ASSUMPTIONS
   // --------------------------------------------------------------------------

                                  // knots[0] <= ... <= knots[knots.size( ) - 1]
                                  //
                                  //               numKnots := knots.size( ) > 1
                                  //
                                  //   bsOrderMin( ) <= bsOrder <= bsOrderMax( )
                                  //   bsIndexMin( ) <= bsIndex <= bsIndexMax( )

   // ------------------------------------- utilities related to number of knots
   // --------------------------------------------------------------------------

   cpp_namespace_decl
   inline
   int
   bsOrderMin( int numKnots );
                                           //
                                           // minimum order of B-splines allowed
                                           //       by the given number of knots
                                           //
                                           //         returns 0 if numKnots <= 1
                                           //                returns 1 otherwise

   cpp_namespace_decl
   inline
   int
   bsOrderMax( int numKnots );
                                           //
                                           // maximum order of B-splines allowed
                                           //       by the given number of knots
                                           //
                                           //         returns 0 if numKnots <= 1
                                           //     returns numKnots - 1 otherwise

   cpp_namespace_decl
   inline
   bool
   hasSupport( int bsOrder, int numKnots );
                                   //
                                   // true if the given number of knots supports
                                   //            any B-spline of the given order

   // --------------------------- utilities related to number of knots and order
   // --------------------------------------------------------------------------

   cpp_namespace_decl
   inline
   int
   basisSize( int bsOrder, int numKnots );
                                            //
                                            // Total # of B-splines of the given
                                            //    order for the given # of knots
                                            //
                                            //                           returns
                                            //      max( 0, numKnots - bsOrder )

   cpp_namespace_decl
   inline
   int
   bsIndexMin( );
                                            //
                                            //    B-spline index range lower end
                                            //    (for any order and # of knots)
                                            //                              := 0

   cpp_namespace_decl
   inline
   int
   bsIndexMax( int bsOrder, int numKnots );
                                            //
                                            //    B-spline index range upper end
                                            // (for the given order, # of knots)
                                            //     := max( 0, basisSize( ) - 1 )

   cpp_namespace_decl
   inline
   bool
   hasSupport( int bsIndex, int bsOrder, int numKnots );
                                   //
                                   // true if the given number of knots supports
                                   //      the given B-spline of the given order

   cpp_namespace_decl
   inline
   int
   lowerSupport( int bsIndex );
                                   //
                                   //                            returns bsIndex
                                   //
                                   //   the value returned makes sense only when
                                   //         hasSupport( bsIndex, ... ) == true

   cpp_namespace_decl
   inline
   int
   upperSupport( int bsIndex, int bsOrder );
                                   //
                                   //                  returns bsIndex + bsOrder
                                   //
                                   //   the value returned makes sense only when
                                   //         hasSupport( bsIndex, ... ) == true

   // --------------------------------------------- utilities involving abscissa
   // --------------------------------------------------------------------------

   cpp_namespace_decl
   inline
   bool
   isInSupport(       fpType    abscissa,
                      int       bsIndex,
                      int       bsOrder,
                const fpVector& knots );
                        //
                        // true if the given B-spline is supprted by the knots &
                        // the abscissa lies in it's support (right-continuous):
                        //
                        //         hasSupport( bsIndex, bsOrder, knots.size( ) )
                        // &&         knots[lowerSupport( bsIndex )] <= abscissa
                        // && abscissa < knots[upperSupport( bsIndex, bsOrder )]

   cpp_namespace_decl
   inline
   bool
   isInSupport( fpType abscissa, const fpVector& knots );
                              //
                              //       true if the abscissa lies in the interval
                              //        defined by the knots (right-continuous):
                              //
                              // knots[0] <= abscissa < knots[knots.size( ) - 1]

   cpp_namespace_decl
   inline
   pair<int, int>
   bracket(       fpType    abscissa,
            const fpVector& knots,
            const bool      rightEndLeftContinuous = true );
               //
               // abscissa                                               returns
               //
               // abscissa < knots[0]                                    (0, -1)
               //
               // knots[0] <= abscissa &&
               // absicssa <  knots[knots.size( ) - 1]
               //                   (n : knots[n] <= abscissa < knots[n + 1], 0)
               //
               // abscissa == knots[knots.size( ) - 1]
               // && rightEndLeftContinuous == true
               //                               (max n : knots[n] < abscissa, 0)
               //
               // abscissa == knots[knots.size( ) - 1]
               // && rightEndLeftContinuous == false     (knots.size( ) - 1, +1)
               //
               // abscissa > knots[knots.size( ) - 1]    (knots.size( ) - 1, +1)
               //
               //    first of the returned pair is the left bracket for abscissa
               //        second is the out-of-range status: +/-1 if out of range

   // -------------------------------------------------- miscellaneous utilities
   // --------------------------------------------------------------------------

   cpp_namespace_decl
   inline
   bool
   isPartitionOfUnity( int intervalIndex, int bsOrder, int numKnots );
                             //
                             //      true if B-splines of the given order form a
                             //        partition of unity for the given interval
                             //            defined by a pair of successive knots
                             //
                             //       intervalIndex is the index of the left-end
                             //                             knot of the interval
                             // knots[intervalIndex] .. knots[intervalIndex + 1]
                             //
                             //                                       ASSUMPTION
                             //
                             //                0 <= intervalIndex < numKnots - 1

   cpp_namespace_decl
   inline
   int
   knotMultiplicity( int knotIndex, const fpVector& knots );
                                     //
                                     // returns multiplicity of knots[knotIndex]
                                     //    i. e., # of knots == knots[knotIndex]

   cpp_namespace_decl
   inline
   void
   adjustEndKnots(       int       leftEndPoint,
                         int       rightEndPoint,
                         int       bsOrder,
                   const fpVector& knots,
                         fpVector& newKnots );
                     //
                     //    returns an adjusted knot sequence consistent with the
                     //                      given order and interval end points
                     //
                     //    rightEndPoint and leftEndPoint are indices of the two
                     //   knots serving as endpoints of the interval of interest
                     //
                     //    if there is an insufficient number of knots on either
                     //    side of the end points, multiple knots are added. Any
                     //     knots unnecessary for the given order and end points
                     //                 are absent in the adjusted knot sequence
                     //
                     //    end point positions in the adjusted knot sequence are
                     //                                              bsOrder - 1
                     //                                                      and
                     //       ( rightEndPoint - leftEndPoint + 1 ) + bsOrder - 1
                     //                                             respectively
                     //
                     //                    size of the adjusted knot sequence is
                     // ( rightEndPoint - leftEndPoint + 1 ) + 2 ( bsOrder - 1 )
                     //
                     //                                               ASSUMPTION
                     //        0 <= leftEndPoint < rightEndPoint < knots.size( )
                     //                           knots.size( ) > 1, bsOrder > 0
                     //
                     //                                                 DEFAULTS
                     //                              if second assumption fails:
                     //                                       return fpVector( )
                     //                if first assumption fails but not second:
                     //      leftEndPoint = 0, rightEndPoint = knots.size( ) - 1

   cpp_namespace_decl
   inline
   fpType
   DnPsi(       fpType    abscissa,
                int       bsIndex,
                int       bsOrder,
          const fpVector& knots,
                int       dOrder = 0 );
                             //
                             // evaluates dOrder-th derivative (-D)^n Psi( ... )
                             //      at abscissa of a special degree bsOrder - 1
                             //                            polynomial defined as
                             //
                             //                Psi( bsOrder, bsIndex, abscissa )
                             //            := product from j = 1 .. bsOrder-1 of
                             //                ( knots[bsIndex + j] - abscissa )
                             //
                             //                               equation (4.1b) in
                             //   Carl de Boor's article "B(asic)-Spline Basics"
                             //    http://www.cs.wisc.edu/~deboor/ftpreadme.html
                             //
                             // this function appears in Marseden's identity and
                             // is useful in getting the B-spline representation
                             //  of any arbitray polynomial: see equations (4.5)

   // ------------------------------------------------------ B-spline evaluation
   // --------------------------------------------------------------------------

   cpp_namespace_decl
   inline
   fpType
   eval( const fpType*   coeff,
               fpType    abscissa,
               int       leftBracket,
               int       bsOrder,
         const fpVector& knots,
               int       dOrder = 0 );
                             //
                             //        evaluate dOrder-th derivative of a spline
                             //        defined by coefficients coeff at abscissa
                             //
                             //                                      ASSUMPTIONS
                             //
                             //                                      leftBracket
                             //         == bracket( abscissa, knots, ... ).first
                             //
                             //                             coeff must have size
                             //           >= basisSize( bsOrder, knots.size( ) )

   cpp_namespace_decl
   inline
   fpType
   evalIntegral( const fpType*   coeff,
                       fpType    lowerLimit,
                       fpType    upperLimit,
                       int       bracketLower,
                       int       bracketUpper,
                       int       bsOrder,
                 const fpVector& knots );
                             //
                             //     evaluate integral of a spline between limits
                             //
                             //         evaluate integral of a spline defined by
                             //          coefficients coeff between given limits
                             //
                             //                                      ASSUMPTIONS
                             //
                             //                                     bracketLower
                             //       == bracket( lowerLimit, knots, ... ).first
                             //                                     bracketUpper
                             //       == bracket( upperLimit, knots, ... ).first
                             //
                             //                             coeff must have size
                             //           == basisSize( bsOrder, knots.size( ) )

   cpp_namespace_decl
   inline
   void
   evalIntegral(       fpType    lowerLimit,
                       fpType    upperLimit,
                       int       bracketLower,
                       int       bracketUpper,
                       int       bsOrder,
                 const fpVector& knots,
                       fpVector& integrals );
                                //
                                //  integrals of all the B-spline between limits

   cpp_namespace_decl
   inline
   fpType
   evalIntegral(       int       bsIndex,
                       int       bsOrder,
                 const fpVector& knots );
                                //
                                //       integral of the given B-spline over its
                                //                               entire support:
                                //
                                // hasSupport( bsIndex, bsOrder, knots.size( ) )
                                // * (   knots[upperSupport( bsIndex, bsOrder )]
                                //     - knots[lowerSupport( bsIndex )] )
                                // / bsOrder

   cpp_namespace_decl
   inline
   fpType
   evalProduct( const fpType*   coeff1,
                const fpType*   coeff2,
                      fpType    abscissa,
                      int       leftBracket,
                      int       bsOrder,
                const fpVector& knots,
                      int       dOrder  = 0,
                      int       dOrder1 = 0,
                      int       dOrder2 = 0 );
                                  //
                                  //   evaluate dOrder-th derivative at abscissa
                                  //  of the product of dOrder1-th derivative of
                                  //   first spline and dOrder2-th derivative of
                                  //                                  the second
                                  //
                                  //            coeff1 and coeff2 must have size
                                  //      == basisSize( bsOrder, knots.size( ) )

   // ------------------------------- core functions used in B-spline evaluation
   // --------------------------------------------------------------------------

   cpp_namespace_decl
   inline
   fpType
   weight(       fpType    abscissa,
                 int       bsIndex,
                 int       bsOrder,
           const fpVector& knots );
                          //
                          //            weights required for B-spline recurrence
                          //
                          //   ( abscissa                     - knots[bsIndex] )
                          // / ( knots[bsIndex + bsOrder - 1] - knots[bsIndex] )
                          //
                          //            if denominator != 0.0, and 0.0 otherwise

   cpp_namespace_decl
   inline
   fpType
   dWeight(       int       bsIndex,
                  int       bsOrder,
            const fpVector& knots );
                          //
                          //                    derivatives of the weights above
                          //
                          //   1.0
                          // / ( knots[bsIndex + bsOrder - 1] - knots[bsIndex] )
                          //
                          //            if denominator != 0.0, and 0.0 otherwise

   cpp_namespace_decl
   inline
   fpType
   invWeight(       fpType    abscissa,
                    int       bsIndex,
                    int       bsOrder,
              const fpVector& knots );
                          //
                          //              inverse weights required by sCoeffPlus
                          //
                          //   ( knots[bsIndex + bsOrder - 1] - knots[bsIndex] )
                          // / ( abscissa                     - knots[bsIndex] )
                          //
                          //            if denominator != 0.0, and 0.0 otherwise

   cpp_namespace_decl
   inline
   fpType
   spline( const fpType*   coeff,
                 fpType    abscissa,
                 int       leftBracket,
                 int       bsOrder,
           const fpVector& knots );
                             //
                             //      value of a spline (:= linear combination of
                             //         B-splines with given coefficients) using
                             //            recurrence on the spline coefficients
                             //
                             //                                      ASSUMPTIONS
                             //
                             //                                      leftBracket
                             //         == bracket( abscissa, knots, ... ).first
                             //
                             //                             coeff must have size
                             //           == basisSize( bsOrder, knots.size( ) )

   cpp_namespace_decl
   inline
   void
   Dspline( const fpType*   coeff,
                  int       bsOrder,
            const fpVector& knots,
                  fpVector& dCoeff,
                  int       dOrder = 1 );
                             //
                             //    returns the dOrder-th derivative of the given
                             //  spline as a lower-order spline on the same knot
                             //                                         sequence
                             //
                             //                            if dOrder >= bsOrder:
                             // returns order-1 spline with all coefficients = 0
                             //          and dCoeff.size( ) == knots.size( ) - 1
                             //
                             //                             coeff must have size
                             //           == basisSize( bsOrder, knots.size( ) )

   cpp_namespace_decl
   inline
   void
   Ispline( const fpType*   coeff,
                  int       bsOrder,
            const fpVector& knots,
                  fpVector& iCoeff,
                  int       iOrder = 1 );
                             //
                             // returns the iOrder-th indefinite integral of the
                             //     given spline as a higher-order spline on the
                             //                               same knot sequence
                             //
                             //  the value of the integral calculated at a point
                             //      will be accurate only if partition of unity
                             //                              holds at that point
                             //
                             //                             coeff must have size
                             //           == basisSize( bsOrder, knots.size( ) )

cpp_namespace_end

// ---------------------------------------- utilities related to number of knots
// -----------------------------------------------------------------------------

int
BSplineTools::bsOrderMin( int numKnots )
 {
  return ( ( numKnots > 1 ) ? 1 : 0 );
 }

int
BSplineTools::bsOrderMax( int numKnots )
 {
  return ( ( numKnots > 1 ) ? numKnots - 1 : 0 );
 }

bool
BSplineTools::hasSupport( int bsOrder, int numKnots )
 {
  return
   (
       ( numKnots               >  1                      )
    && ( bsOrderMin( numKnots ) <= bsOrder                )
    && ( bsOrder                <= bsOrderMax( numKnots ) )
   );
 }

// ------------------------------ utilities related to number of knots and order
// -----------------------------------------------------------------------------

int
BSplineTools::basisSize( int bsOrder, int numKnots )
 {
  return ( ( numKnots > bsOrder ) ? numKnots - bsOrder : 0 );
 }

int
BSplineTools::bsIndexMin( )
 {
  return 0;
 }

int
BSplineTools::bsIndexMax( int bsOrder, int numKnots )
 {
  int sz = basisSize( bsOrder, numKnots );
  return ( sz ? sz - 1 : 0 );
 }

bool
BSplineTools::hasSupport( int bsIndex, int bsOrder, int numKnots )
 {
  return
   (
       hasSupport( bsOrder, numKnots )
    && ( bsIndexMin( ) <= bsIndex                         )
    && ( bsIndex       <= bsIndexMax( bsOrder, numKnots ) )
   );
 }

int
BSplineTools::lowerSupport( int bsIndex )
 {
  return bsIndex;
 }

int
BSplineTools::upperSupport( int bsIndex, int bsOrder )
 {
  return ( bsIndex + bsOrder );
 }

// ------------------------------------------------ utilities involving abscissa
// -----------------------------------------------------------------------------

bool
BSplineTools::isInSupport(       fpType    abscissa,
                                 int       bsIndex,
                                 int       bsOrder,
                           const fpVector& knots )
 {
  return
   (
       hasSupport( bsIndex, bsOrder, knots.size( ) )
    && ( knots[lowerSupport( bsIndex )] <= abscissa         )
    && ( abscissa < knots[upperSupport( bsIndex, bsOrder )] )
   );
 }

bool
BSplineTools::isInSupport(       fpType    abscissa,
                           const fpVector& knots )
 {
  return
   (
       ( knots.size( ) >  1                        )
    && ( knots[0]      <= abscissa                 )
    && ( abscissa      <  knots[knots.size( ) - 1] )
   );
 }

pair<int, int>
BSplineTools::bracket(       fpType    abscissa,
                       const fpVector& knots,
                       const bool      rightEndLeftContinuous )
 {
  if ( ( knots.size( ) == 0 ) || ( abscissa < knots[0] ) )
   return ( pair<int, int> (0, -1) );

  if ( abscissa > knots[knots.size( ) - 1] )
   return ( pair<int, int> (knots.size( ) - 1, +1) );

  if ( abscissa == knots[knots.size( ) - 1] )
   {
    int leftBracket = knots.size( ) - 1;
    int outOfRange  = +1;

    if ( rightEndLeftContinuous )
     {
      while (    ( leftBracket >  0                  )
              && ( abscissa    == knots[leftBracket] ) ) leftBracket--;
      outOfRange = 0;
     }

    return ( pair<int, int> (leftBracket, outOfRange) );
   }

  int  leftBracket = 0;
  bool found       = false;

                 // might use binary search/standard lib functions for this part
                 //
  while ( !found && ( leftBracket < ( knots.size( ) - 1 ) ) )
   {
    found =    ( knots[leftBracket] <= abscissa               )
            && ( abscissa           <  knots[leftBracket + 1] );

    leftBracket++;
   }

  return ( pair<int, int> (leftBracket - 1, 0) );
 }

// ----------------------------------------------------- miscellaneous utilities
// -----------------------------------------------------------------------------

bool
BSplineTools::isPartitionOfUnity( int intervalIndex, int bsOrder, int numKnots )
 {
  return
   (
       hasSupport( bsOrder, numKnots )
    && ( ( bsOrder - 1 ) <= intervalIndex          )
    && ( intervalIndex   <  ( numKnots - bsOrder ) )
   );
 }

int
BSplineTools::knotMultiplicity( int knotIndex, const fpVector& knots )
 {
  if ( ( knotIndex < 0 ) || ( knotIndex >= knots.size( ) ) ) return 0;

  int left, right;

  for ( left  = knotIndex;    ( left  >= 0 )
                           && ( knots[left]  == knots[knotIndex] ); left--  ) ;
  for ( right = knotIndex;    ( right <  knots.size( ) )
                           && ( knots[right] == knots[knotIndex] ); right++ ) ;

  return ( right - left - 1 );
 }

void
BSplineTools::adjustEndKnots(       int       leftEndPoint,
                                    int       rightEndPoint,
                                    int       bsOrder,
                              const fpVector& knots,
                                    fpVector& newKnots )
 {
  if ( ( knots.size( ) <= 1 ) || ( bsOrder < 1 ) )
   newKnots.resize( 0 );
  else
   {
    if ( rightEndPoint <= leftEndPoint )
     {
      leftEndPoint  = 0;
      rightEndPoint = (int) knots.size( ) - 1;
     }

    int interiorKnots = rightEndPoint - leftEndPoint + 1;

    newKnots.resize( interiorKnots + 2 * ( bsOrder - 1 ) );

    int shift = ( bsOrder - 1 ) - leftEndPoint;

    for ( int i = 0; i < bsOrder; i++ )
     newKnots[i] = knots[max( 0, i - shift )];
                                                        //
                                                        //  adjust/copy left end

    for ( int i = bsOrder; i < interiorKnots + bsOrder - 2; i++ )
     newKnots[i] = knots[i - shift];
                                                        //
                                                        //   copy middle portion

    for ( int i = interiorKnots + bsOrder - 2; i < newKnots.size( ); i++ )
     newKnots[i] = knots[min( (int) (knots.size( ) - 1), i - shift )];
                                                        //
                                                        // adjust/copy right end
   }
 }

fpType
BSplineTools::DnPsi(       fpType    abscissa,
                           int       bsIndex,
                           int       bsOrder,
                     const fpVector& knots,
                           int       dOrder )
 {
  if ( dOrder < 0 ) dOrder = 0;

  if ( !hasSupport( bsIndex, bsOrder, knots.size( ) ) || ( dOrder >= bsOrder ) )
   return 0.0;
  else if ( bsOrder == 1 )
   return 1.0;
  else
   {
    int numTerms = bsOrder - 1;

    vector<bool> bitKey( numTerms, false );
    for ( int i = dOrder; i < numTerms; i++ ) bitKey[i] = true;
                                            //
                                            // first key with n 0's left aligned

    fpType value = 0.0;
    fpType productTerm;

    do
     {
      productTerm = 1.0;
      for ( int i = 0; i < numTerms; i++ )
       if ( bitKey[i] ) productTerm *= ( knots[bsIndex + i + 1] - abscissa );

      value += productTerm;
     }
    while ( next_permutation( bitKey.begin( ), bitKey.end( ) ) );

    productTerm = 1.0;
    for ( int i = 2; i <= dOrder; i++ ) productTerm *= ( (fpType) i );

    return value * productTerm;
   }
 }

// --------------------------------------------------------- B-spline evaluation
// -----------------------------------------------------------------------------

fpType
BSplineTools::eval( const fpType*   coeff,
                          fpType    abscissa,
                          int       leftBracket,
                          int       bsOrder,
                    const fpVector& knots,
                          int       dOrder )
 {
  if ( dOrder >= bsOrder )
   return 0.0;
  else if ( dOrder <= 0 )
   return spline( coeff, abscissa, leftBracket, bsOrder, knots );
  else
   {
    fpVector dCoeff;
    Dspline( coeff, bsOrder, knots, dCoeff, dOrder );

    return
     spline( dCoeff( 0 ), abscissa, leftBracket, bsOrder - dOrder, knots );
   }
 }

fpType
BSplineTools::evalIntegral( const fpType*   coeff,
                                  fpType    lowerLimit,
                                  fpType    upperLimit,
                                  int       bracketLower,
                                  int       bracketUpper,
                                  int       bsOrder,
                            const fpVector& knots )
 {
  int numKnots = knots.size( );

  if ( lowerLimit < knots[0] )
   {
    lowerLimit   = knots[0];
    bracketLower = 0;
   }
  else if ( lowerLimit > knots[numKnots - 1] )
   {
    lowerLimit   = knots[numKnots - 1];
    bracketLower = numKnots - 2; // ATTN: right end left continuous
   }

  if ( upperLimit < knots[0] )
   {
    upperLimit   = knots[0];
    bracketUpper = 0;
   }
  else if ( upperLimit > knots[numKnots - 1] )
   {
    upperLimit   = knots[numKnots - 1];
    bracketUpper = numKnots - 2; // ATTN: right end left continuous
   }

  if ( lowerLimit == upperLimit ) return 0.0;

  fpVector iKnots;
  adjustEndKnots( 0, numKnots - 1, bsOrder + 1, knots, iKnots );

  fpVector a( 0.0, basisSize( bsOrder + 1, iKnots.size( ) ) );
  for ( int i = 0, j = bsOrder;
            i < basisSize( bsOrder, numKnots ); i++, j++ ) a[j] = coeff[i];

  fpVector iCoeff;
  Ispline( a( 0 ), bsOrder, iKnots, iCoeff );

  fpType lower = spline( iCoeff( 0 ), lowerLimit,
                         bracketLower + bsOrder, bsOrder + 1, iKnots );
  fpType upper = spline( iCoeff( 0 ), upperLimit,
                         bracketUpper + bsOrder, bsOrder + 1, iKnots );

  return ( upper - lower );
 }

void
BSplineTools::evalIntegral(       fpType    lowerLimit,
                                  fpType    upperLimit,
                                  int       bracketLower,
                                  int       bracketUpper,
                                  int       bsOrder,
                            const fpVector& knots,
                                  fpVector& integrals )
 {
  int numKnots = knots.size( );

  if ( lowerLimit < knots[0] )
   {
    lowerLimit   = knots[0];
    bracketLower = 0;
   }
  else if ( lowerLimit > knots[numKnots - 1] )
   {
    lowerLimit   = knots[numKnots - 1];
    bracketLower = numKnots - 2; // ATTN: right end left continuous
   }

  if ( upperLimit < knots[0] )
   {
    upperLimit   = knots[0];
    bracketUpper = 0;
   }
  else if ( upperLimit > knots[numKnots - 1] )
   {
    upperLimit   = knots[numKnots - 1];
    bracketUpper = numKnots - 2; // ATTN: right end left continuous
   }

  integrals.resize( basisSize( bsOrder, numKnots ) );

  if ( lowerLimit != upperLimit )
   {
    fpVector iKnots;
    adjustEndKnots( 0, numKnots - 1, bsOrder + 1, knots, iKnots );
  
    fpVector a( 0.0, basisSize( bsOrder + 1, iKnots.size( ) ) );
    a[bsOrder] = 1.0;

    fpType lower, upper;
    for ( int i = 0, j = bsOrder; i < integrals.size( ); i++, j++ )
     {
      if ( i > 0 ) a[j - 1] = 0.0; a[j] = 1.0;

      fpVector iCoeff;
      Ispline( a( 0 ), bsOrder, iKnots, iCoeff );
    
      lower = spline( iCoeff( 0 ), lowerLimit,
                      bracketLower + bsOrder, bsOrder + 1, iKnots );
      upper = spline( iCoeff( 0 ), upperLimit,
                      bracketUpper + bsOrder, bsOrder + 1, iKnots );
    
      integrals[i] = upper - lower;
     }
   }
 }

fpType
BSplineTools::evalIntegral(       int       bsIndex,
                                  int       bsOrder,
                            const fpVector& knots )
 {
  if ( hasSupport( bsIndex, bsOrder, knots.size( ) ) )
   return ( knots[bsIndex + bsOrder] - knots[bsIndex] ) / bsOrder;
  else
   return 0.0;
 }

fpType
BSplineTools::evalProduct( const fpType*   coeff1,
                           const fpType*   coeff2,
                                 fpType    abscissa,
                                 int       leftBracket,
                                 int       bsOrder,
                           const fpVector& knots,
                                 int       dOrder,
                                 int       dOrder1,
                                 int       dOrder2 )
 {
  int n  = max( 0, dOrder  );
  int m1 = max( 0, dOrder1 );
  int m2 = max( 0, dOrder2 );

  int mMin = max( 0, n - ( bsOrder - 1 ) + m2 );
  int mMax = min( n, ( bsOrder - 1 ) - m1 );

  fpType nCr = MiscNumericFns::nCr( n, mMin );

  fpType d = 0.0;
  for ( int m = mMin; m <= mMax; m++ )
   {
    d +=   nCr
         * eval( coeff1, abscissa, leftBracket, bsOrder, knots, m1 + m     )
         * eval( coeff2, abscissa, leftBracket, bsOrder, knots, m2 + n - m );

    nCr *= ( (fpType) ( n - m ) ) / ( (fpType) ( m + 1 ) );
   }

  return d;
 }

// ---------------------------------- core functions used in B-spline evaluation
// -----------------------------------------------------------------------------

fpType
BSplineTools::weight(       fpType    abscissa,
                            int       bsIndex,
                            int       bsOrder,
                      const fpVector& knots )
 {
   return
    (
     ( abscissa - knots[bsIndex] ) * dWeight( bsIndex, bsOrder, knots )
    );
                                                 //
                                                 //                  ASSUMPTIONS
                                                 //
                                                 // same as those for dWeight( )
 }

fpType
BSplineTools::dWeight(       int       bsIndex,
                             int       bsOrder,
                       const fpVector& knots )
 {
  if ( knots[bsIndex] != knots[bsIndex + bsOrder - 1] )
   return 1.0 / ( knots[bsIndex + bsOrder - 1] - knots[bsIndex] );
  else
   return 0.0;
                                       //
                                       //                            ASSUMPTIONS
                                       //
                                       // 0 <= bsIndex < knots.size( ) - bsOrder
                                       //           1 <= bsOrder < knots.size( )
 }

fpType
BSplineTools::invWeight(       fpType    abscissa,
                               int       bsIndex,
                               int       bsOrder,
                         const fpVector& knots )
 {
  fpType iw = weight( abscissa, bsIndex, bsOrder, knots );
  if ( iw != 0.0 ) iw = 1.0 / iw;
  return iw;
                                                  //
                                                  //                 ASSUMPTIONS
                                                  //
                                                  // same as those for weight( )
 }

/*
fpType
BSplineTools::spline( const fpType*   coeff,
                            fpType    abscissa,
                            int       leftBracket,
                            int       bsOrder,
                      const fpVector& knots )
 {
  fpVector a( coeff, basisSize( 1, knots.size( ) ) );

  fpType current, previous, t, w;
  int    imax;

  for ( int m = 1; m < bsOrder; m++ )
   {
    imax = basisSize( bsOrder - m, knots.size( ) );

    previous = 0.0;
    current  = a[0];

    for ( int index = 0; index < imax; index++ )
     {
      w        = weight( abscissa, index, ( bsOrder - m ) + 1, knots );
      t        = w * current + ( 1.0 - w ) * previous;

      previous = a[index];
      current  = ( index < ( imax - 2 ) ) ? a[index + 1] : 0.0;

      a[index] = t;
     }
   }

  return a[leftBracket];
 }
*/

fpType
BSplineTools::spline( const fpType*   coeff,
                            fpType    abscissa,
                            int       leftBracket,
                            int       bsOrder,
                      const fpVector& knots )
 {
  fpVector a( coeff, leftBracket + 1 );

  fpType current, previous, t, w;
  int    imin, imax;

  for ( int m = 1; m < bsOrder; m++ )
   {
    imin = max( leftBracket - ( bsOrder - m ) + 1, 0 );
    imax = leftBracket;

    current  = a[imin];
    previous = ( ( imin > 0 ) ? a[imin - 1] : 0.0 );

    for ( int index = imin; index <= imax; index++ )
     {
      w = weight( abscissa, index, ( bsOrder - m ) + 1, knots );
      t = w * current + ( 1.0 - w ) * previous;

      previous = a[index];
      current  = ( ( index < ( a.size( ) - 1 ) ) ? a[index + 1] : 0.0 );

      a[index] = t;
     }
   }

  return a[leftBracket];
 }

void
BSplineTools::Dspline( const fpType*   coeff,
                             int       bsOrder,
                       const fpVector& knots,
                             fpVector& dCoeff,
                             int       dOrder /* = 1 */ )
 {
  dCoeff.resize( basisSize( bsOrder - dOrder, knots.size( ) ), 0.0 );
  for ( int i = 0; i < dCoeff.size( ); i++ ) dCoeff[i] = coeff[i];

  fpType current, previous, t;
  int    imax;

  for ( int d = 1; d <= dOrder; d++ )
   {
    imax = basisSize( bsOrder - d, knots.size( ) );

    previous = 0.0;
    current  = dCoeff[0];

    for ( int index = 0; index < imax; index++ )
     {
      t =   ( bsOrder - d ) * ( current - previous )
          * dWeight( index, bsOrder - ( d - 1 ), knots );

      previous = dCoeff[index];
      current  = ( index < ( imax - 2 ) ) ? dCoeff[index + 1] : 0.0;

      dCoeff[index] = t;
     }
   }
 }

void
BSplineTools::Ispline( const fpType*   coeff,
                             int       bsOrder,
                       const fpVector& knots,
                             fpVector& iCoeff,
                             int       iOrder /* = 1 */ )
 {
  fpVector ic( coeff, basisSize( bsOrder, knots.size( ) ) );

  fpType scale;
  int    iiOrder;

  for ( int i = 1; i <= iOrder; i++ )
   {
    iiOrder = bsOrder + ( i - 1 );
    scale   = 1.0 / ( (fpType) iiOrder );

    ic[0] = ( knots[iiOrder] - knots[0] ) * scale * ic[0];

    for ( int index = 1;
              index < basisSize( bsOrder + i, knots.size( ) );
              index++ )
     ic[index] =   ic[index - 1]
                 + ( knots[index + iiOrder] - knots[index] )
                 * scale * ic[index];
   }

  iCoeff.resize( basisSize( bsOrder + iOrder, knots.size( ) ) );
  for ( int i = 0; i < iCoeff.size( ); i++ ) iCoeff[i] = ic[i];
 }

#endif /* __BSplineTools_H__ */
