/* --------------------------------------------------------------------------  
 *
 * class ArrayIndexMap
 *   mapping between indices of an N-dimensional array and
 *   offsets into a one-dimensional array of the same size
 *
 * required headers
 *   see include header file ArrayIndexMap
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __ArrayIndexMap_H__
#define __ArrayIndexMap_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class ArrayIndexMap
 {
  public:
   typedef uintVector LocationType;

   // --------------------------------------------------- construct/reinitialize
   // --------------------------------------------------------------------------

   ArrayIndexMap( );
                                 //
                                 // trivial scalar: numDims = 1, sizes( )[0] = 1

   ArrayIndexMap( const ArrayIndexMap& another );

   ArrayIndexMap&
   operator=( const ArrayIndexMap& rhs );

   ArrayIndexMap( const uintVector&   sizes,
                  const uintVector&   indexOrder       = uintVector( ),
                  const LocationType& indexLowerBounds = LocationType( ) );
                        //
                        //                 number of dimensions == sizes.size( )
                        //                      sizes[d]: size along dimension d
                        //
                        //  indexOrder: order-of-variation  table, a permutation
                        //    of integers 0 <= i < sizes.size( ), specifying the
                        //            order in which the indices vary: the index
                        //    indexOrder[0] varies fastest, indexOrder[1] varies
                        // second fastest, and so on; a generalization of the 2D
                        //       notions of column-major and row-major orderings
                        //
                        //       indexLowerBounds: dimensionwise lower bounds on
                        //    indices; e. g., 0-based, 1-based, etc. In general,
                        //                          n-based with integer n >= 0.
                        //
                        //                              ASSUMPTIONS AND DEFAULTS
                        //
                        //                  # of dimensions := sizes.size( ) > 0
                        //                  if not : dimensions := 1 =: sizes[0]
                        //
                        //                                sizes[d] > 0 for all d
                        //                          if not : it is taken to be 1
                        //
                        //                 if indexOrder.size( ) < sizes.size( )
                        //   or if indexOrder is an invalid sequence of numbers,
                        //             indexOrder is taken to be the permutation
                        //                      { 0, 1, ..., sizes.size( ) - 1 }
                        //
                        //   All unspecified lower bounds for indices assumed 0.

   void
   reinitialize( const uintVector&   sizes,
                 const uintVector&   indexOrder       = uintVector( ),
                 const LocationType& indexLowerBounds = LocationType( ) );

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   size_t
   numDims( ) const;                                          // # of dimensions

   inline
   size_t
   size( ) const;
                                              //
                                              // total volume = product of sizes

   inline
   const uintVector&
   sizes( ) const;
                                                          //
                                                          // sizes dimensionwise

   inline
   const LocationType&
   indexLower( ) const;
                                             //
                                             // index lower bounds dimensionwise

   inline
   const LocationType&
   indexUpper( ) const;
                                             //
                                             // index upper bounds dimensionwise

   inline
   const uintVector&
   order( ) const;
                                       //
                                       // order-of-variation table dimensionwise

   // ------------------------------------------------------------------ mapping
   // --------------------------------------------------------------------------

   inline
   long
   offset( const LocationType& indices ) const;
               //
               //                                                N --> 1 mapping
               //
               //             returns a unique integer between 0 and size( ) - 1
               //             as if it were computed by the multiple loops below
               //
               // size_t offset = 0;
               // for ( size_t iN = 0; iN <= indices[indexOrder[N]]; iN++ )
               //  ...
               //   for ( size_t i2 = 0; i2 <= indices[indexOrder[1]]; i2++ )
               //    for ( size_t i1 = 0; i1 <= indices[indexOrder[0]]; i1++ )
               //     offset++;
               //
               //   indices.size( ) >= numDims( ): first numDims( ) entries used
               //   if indices.size( ) < numDims( ), the unspecified indices are
               //                                      taken to be equal to zero
               //
               //              indexLower( d ) <= indices[d] <= indexUpper( d ):
               //                                      indices[d] is taken to be
               //   max( min( indices[d], indexUpper( d ) ), indexLower( d ) )

   inline
   LocationType
   indices( long offset ) const;
                          //
                          //                                     1 --> N mapping
                          //
                          // given an integer between 0 and size( ) - 1, returns
                          //      multiple indices corresponding to this integer
                          //
                          //                               0 <= offset < size( )
                          //                  if offset < 0: it is taken to be 0
                          //      if offset > size( ): it is taken to be size( )

   inline
   void
   getIndices( long offset, LocationType& indices ) const;

  private:
   uintVector   my_sizes, my_indexOrder;
   LocationType my_lowerBounds, my_upperBounds;
   varray<long> my_volume;
 };

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

size_t
ArrayIndexMap::numDims( ) const
 {
  return my_sizes.size( );
 }

size_t
ArrayIndexMap::size( ) const
 {
  return my_volume[numDims( )];
 }

const uintVector&
ArrayIndexMap::sizes( ) const
 {
  return my_sizes;
 }

const ArrayIndexMap::LocationType&
ArrayIndexMap::indexLower( ) const
 {
  return my_lowerBounds;
 }

const ArrayIndexMap::LocationType&
ArrayIndexMap::indexUpper( ) const
 {
  return my_upperBounds;
 }

const uintVector&
ArrayIndexMap::order( ) const
 {
  return my_indexOrder;
 }

// --------------------------------------------------------------------- mapping
// -----------------------------------------------------------------------------

long
ArrayIndexMap::offset( const LocationType& indices ) const
 {
  long offset = 0;
  size_t dMax = min( numDims( ), indices.size( ) );
  for ( size_t d = 0; d < dMax; d++ )
   offset +=   my_volume[d]
             * (   max(min(indices[d],my_upperBounds[d]),my_lowerBounds[d])
                 - my_lowerBounds[d] );

  // 0-based indices: offset += min( indices[d], my_sizes[d] - 1 ) * my_volume[d];

  return offset;
 }

ArrayIndexMap::LocationType
ArrayIndexMap::indices( long offset ) const
 {
  LocationType idx;
  getIndices( offset, idx );
  return idx;
 }

void
ArrayIndexMap::getIndices( long offset, LocationType& indices ) const
 {
  offset = max( 0L, min( static_cast<long>( size( ) ), offset ) );

  indices.resize( numDims( ) );

  size_t od;
  long sum = 0;
  for ( size_t d = numDims( ) - 1; d > 0; d-- )
   {
    od           = my_indexOrder[d];
    indices[od]  = ( offset - sum ) / my_volume[od];
    sum         += indices[od] * my_volume[od];
   }

  od          = my_indexOrder[0];
  indices[od] = ( offset - sum ) / my_volume[od];

  indices += my_lowerBounds;
 }

#endif /* __ArrayIndexMap_H__ */
