/* --------------------------------------------------------------------------  
 *
 * class ArrayNavigator
 *   arbitrary-dimensional array navigation tool
 *
 * required headers
 *   see include header file ArrayNavigator
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __ArrayNavigator_H__
#define __ArrayNavigator_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class ArrayNavigator : public ArrayIndexMap
 {
  public:
   ArrayNavigator( );

   ArrayNavigator( const ArrayNavigator& another );

   ArrayNavigator&
   operator=( const ArrayNavigator& rhs );

   ArrayNavigator( const uintVector&   sizes,
                   const uintVector&   indexOrder       = LocationType( ),
                   const LocationType& indexLowerBounds = LocationType( ) );

   void
   reinitialize( const uintVector&   sizes,
                 const uintVector&   indexOrder       = LocationType( ),
                 const LocationType& indexLowerBounds = LocationType( ) );

   // -------------------------------------------------------- navigation ranges
   // --------------------------------------------------------------------------

                 // navigation range is defined as a subrange of the index range
                 // (default is indexLower() and indexUpper() ). The naviagation
                 // tools below (next, prev, moveTo, etc.) adhere to this range.

   void
   setRange( size_t dimension, size_t lower = 0, size_t upper = -1 );
                                    //
                                    //          0 <= lower < sizes( )[dimension]
                                    // 0 <= lower <= upper < sizes( )[dimension]
                                    //
                                    // No change if  dimension  is out-of-range.
                                    // out-of-range  lower and upper  bounds are
                                    // replaced by appropriate defaults.
                                    //
                                    // ArrayNavigator is repositioned if current
                                    // location is out-of-range with  respect to
                                    // new ranges.

   inline
   const LocationType&
   rangeLower( ) const;
                                  //
                                  // navigation range dimensionwise lower bounds

   inline
   const LocationType&
   rangeUpper( ) const;
                                  //
                                  // navigation range dimensionwise upper bounds

   // ----------------------------------------------------------------- location
   // --------------------------------------------------------------------------

   inline
   const LocationType&
   first( ) const;

   inline
   const LocationType&
   last( ) const;

   inline
   const LocationType&
   whereAmI( ) const;

   inline
   bool
   atFirst( ) const;

   inline
   bool
   atLast( ) const;

   // --------------------------------------------------------- navigation tools
   // --------------------------------------------------------------------------

   inline
   bool
   toNext( ) const;
                                      //
                                      // return false if whereAmI( ) == last( )

   inline
   bool
   toPrev( ) const;
                                     //
                                     // return false if whereAmI( ) == first( )

   inline
   bool
   toFirst( ) const;

   inline
   bool
   toLast( ) const;

   inline
   bool
   moveTo( const LocationType& location ) const;
                                      //
                                      // move to location; out-of-range location
                                      // indices are ignored  (reports  success)

   inline
   bool
   moveTo( size_t dimension, size_t index ) const;
                                 //
                                 // out-of-range dimension and index are ignored

   // ---------------------------------------- array offset for current location
   // --------------------------------------------------------------------------

   inline
   long
   offset( ) const;
                                 //
                                 // same as ArrayIndexMap::offset( whereAmI( ) )

   // -------------------------------------------------------------- move status
   // --------------------------------------------------------------------------

   inline
   bool
   anyMove( ) const;
                                   //
                                   // any move after the last call to offset( )?

   inline
   void
   setMove( bool move = true ) const;
         //
         // manually set/reset move status reported by anyMove( ), at your risk.

  private:
   LocationType my_rangeLower, my_rangeUpper;

   mutable LocationType my_location;
   mutable long         my_offset;
   mutable bool         my_anyMove;
 };

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

const ArrayNavigator::LocationType&
ArrayNavigator::rangeLower( ) const
 {
  return my_rangeLower;
 }

const ArrayNavigator::LocationType&
ArrayNavigator::rangeUpper( ) const
 {
  return my_rangeUpper;
 }

// -------------------------------------------------------------------- location
// -----------------------------------------------------------------------------

const ArrayNavigator::LocationType&
ArrayNavigator::first( ) const
 {
  return my_rangeLower;
 }

const ArrayNavigator::LocationType&
ArrayNavigator::last( ) const
 {
  return my_rangeUpper;
 }

const ArrayNavigator::LocationType&
ArrayNavigator::whereAmI( ) const
 {
  return my_location;
 }

bool
ArrayNavigator::atFirst( ) const
 {
  bool af = ( my_location[0] == my_rangeLower[0] );
  for ( size_t d = 1; af && ( d < numDims( ) ); d++ )
   af = af && ( my_location[d] == my_rangeLower[d] );
  return af;
 }

bool
ArrayNavigator::atLast( ) const
 {
  bool al = ( my_location[0] == my_rangeUpper[0] );
  for ( size_t d = 1; al && ( d < numDims( ) ); d++ )
   al = al && ( my_location[d] == my_rangeUpper[d] );
  return al;
 }

// ------------------------------------------------------------- navigator tools
// -----------------------------------------------------------------------------

bool
ArrayNavigator::toNext( ) const
 {
  if ( atLast( ) ) return false;

  size_t od;
  bool done = false;
  for ( size_t d = 0; ( d < numDims( ) ) && ( !done ); d++ )
   {
    od = order( )[d];

    if ( my_location[od] < my_rangeUpper[od] )
     {
      my_location[od] += 1;
      done            = true;
     }
    else
     my_location[od] = my_rangeLower[od];
   }

  my_anyMove = true;

  return true;
 }

bool
ArrayNavigator::toPrev( ) const
 {
  if ( atFirst( ) ) return false;

  size_t od;
  bool done = false;
  for ( size_t d = 0; ( d < numDims( ) ) && ( !done ); d++ )
   {
    od = order( )[d];

    if ( my_location[od] > my_rangeLower[od] )
     {
      my_location[od] -= 1;
      done            = true;
     }
    else
     my_location[od] = my_rangeUpper[od];
   }

  my_anyMove = true;

  return true;
 }

bool
ArrayNavigator::moveTo( const LocationType& location ) const
 {
  size_t lsize = min( location.size( ), numDims( ) );

  bool allInRange = true;
  bool inRange    = true;
  bool anyMove    = false;
  for ( size_t d = 0; inRange && ( d < lsize ); d++ )
   {
    inRange    = ( location[d] >= my_rangeLower[d] ) && ( location[d] <= my_rangeUpper[d] );
    allInRange = allInRange && inRange;
    anyMove    = anyMove || inRange;

    if ( inRange ) my_location[d] = location[d];
   }

  my_anyMove = my_anyMove || anyMove;

  return allInRange;
 }

bool
ArrayNavigator::moveTo( size_t d, size_t index ) const
 {
  bool inRange =    ( d     <  numDims( ) )
                 && ( index >= my_rangeLower[d]          )
                 && ( index <= my_rangeUpper[d]          );

  if ( inRange ) my_location[d] = index;

  my_anyMove = my_anyMove || inRange;

  return inRange;
 }

bool
ArrayNavigator::toFirst( ) const
 {
  my_location = my_rangeLower;
  my_anyMove  = true;

  return true;
 }

bool
ArrayNavigator::toLast( ) const
 {
  my_location = my_rangeUpper;
  my_anyMove  = true;

  return true;
 }

// ------------------------------------------- array offset for current location
// -----------------------------------------------------------------------------

long
ArrayNavigator::offset( ) const
 {
  if ( my_anyMove )
   {
    my_offset  = ArrayIndexMap::offset( my_location );
    my_anyMove = false;
   }

  return my_offset;
 }

// ----------------------------------------------------------------- move status
// -----------------------------------------------------------------------------

bool
ArrayNavigator::anyMove( ) const
 {
  return my_anyMove;
 }

void
ArrayNavigator::setMove( bool move ) const
 {
  my_anyMove = move;
 }

#endif /* __ArrayNavigator_H__ */
