/* --------------------------------------------------------------------------
 *
 * namespace MiscFns
 *   miscellaneous non-numeric functions
 *
 * required headers
 *   see include header file MiscFns
 *
 * notes
 *   contains function templates; may not be supported by every compiler
 *   (e. g., SUNWspro 5.0).
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __MiscFns_H__
#define __MiscFns_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

cpp_namespace_begin( MiscFns )

  // ---------------------------------------------------------- case conversions
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  inline
  string
  upperCase( const string& s );

  cpp_namespace_decl
  inline
  string
  lowerCase( const string& s );

  // ------------------------------------ strip leading and trailing whitespaces
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  inline
  string
  stripped( const string& s, const string& whitespace = " \t" );

  // --------------------------------- find/replace all occurrences of a pattern
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  inline
  string
  replace( const string& inThisString,
           const string& thisPattern = " \t",
           const string& with        = "" );

  // ----------------------------------------------- separate string into tokens
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  inline
  void
  getTokens( const string&         tokenStr,
                   vector<string>& tokens,
             const string&         separators = " \t" );

  cpp_namespace_decl
  inline
  vector<string>
  tokens( const string& tokenStr, const string& separators = " \t" );

  // ------------------------------ range options from strings of the form "n:m"
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  inline
  void
  getRange( const string& rangeStr, const pair<int,int>& defaults, pair<int,int>& range );
                          //
                          //         split strings of the form "", "lower", ":",
                          //    "lower:", ":upper", "lower:upper", "upper:lower"
                          //                into integers lower, upper such that
                          // defaults.first <= lower <= upper <= defaults.second

  cpp_namespace_decl
  inline
  pair<int,int>
  range( const string& rangeStr, const pair<int,int>& defaults );

  // -------------------------------- number of significant digits of an integer
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  inline
  size_t
  nDigits( int n );
                                      //
                                      // # of significant digits of an integer n
                                      //                   (for output purposes)
                                      //
                                      // n    return value
                                      //
                                      // > 0  # of significant digits of n
                                      // = 0  1
                                      // < 0  1 + # of significant digits of |n|

  // ------------------------------- keys for iterating over binary permutations
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  inline
  void
  getFirstKey( size_t size, size_t numOnes, MaskType& key );

  cpp_namespace_decl
  inline
  MaskType
  firstKey( size_t size, size_t numOnes );
                                   //
                                   //        returns bit sequence of length size
                                   //         with numOnes 1's all right-aligned
                                   //           (lexicographically the smallest)
                                   //
                                   //            useful with next_permutation( )
                                   //
                                   //                                ASSUMPTIONS
                                   //
                                   //             size > 0, 0 <= numOnes <= size

  cpp_namespace_decl
  inline
  void
  getLastKey( size_t size, size_t numOnes, MaskType& key );

  cpp_namespace_decl
  inline
  MaskType
  lastKey( size_t size, size_t numOnes );
                                   //
                                   //        returns bit sequence of length size
                                   //          with numOnes 1's all left-aligned
                                   //            (lexicographically the largest)
                                   //
                                   //            useful with next_permutation( )
                                   //
                                   //                                ASSUMPTIONS
                                   //
                                   //             size > 0, 0 <= numOnes <= size

#ifndef cpp_missing_template_methods

  // ------------------------------------------------------- write array to cout
  // ---------------------------------------------------------------------------

  template<template<class T> class C, class T>
  cpp_namespace_decl
  inline
  void
  writeArray( const C<T>&    v,
                    ostream& output         = cout,
              const string&  linePrefix     = "",
                    bool     scientific     = true,
                    int      fieldWidth     = 22,
                    size_t   precision      = 15,
                    size_t   numbersPerLine = 4,
              const string&  separator      = " " );

  // ----------------------------------------------------------- array -> string
  // ---------------------------------------------------------------------------

  template<template<class T> class C, class T>
  cpp_namespace_decl
  inline
  string
  asString( const C<T>&   ct,
            const string& singleNumberFormat = "%g",
            const string& separator          = " " );

  // ----------------------------- buffer copy with static_cast type conversions
  // ---------------------------------------------------------------------------

  template< class T1, class T2 >
  cpp_namespace_decl
  inline
  void
  castAndCopy( size_t size, const T1* const from, T2* const into );

#endif /* cpp_missing_template_methods */

cpp_namespace_end

// ------------------------------------------------------------ case conversions
// -----------------------------------------------------------------------------

string
MiscFns::upperCase( const string& s )
 {
  string t( s );
  for ( string::iterator i = t.begin( ); i != t.end( ); ++i ) (*i) = toupper( (*i) );
  return t;
 }

string
MiscFns::lowerCase( const string& s )
 {
  string t( s );
  for ( string::iterator i = t.begin( ); i != t.end( ); ++i ) (*i) = tolower( (*i) );
  return t;
 }

// ------------------------------------------- strip leading and trailing blanks
// -----------------------------------------------------------------------------

string
MiscFns::stripped( const string& s, const string& whitespace )
 {
  string::size_type i = s.find_first_not_of( whitespace );
  return s.substr( i, s.find_last_not_of( whitespace ) - i + 1 );
 }

// ----------------------------------- find/replace all occurrences of a pattern
// -----------------------------------------------------------------------------

string
MiscFns::replace( const string& inThisString, const string& thisPattern, const string& withThis )
 {
  string s( inThisString );

  string::size_type n = max( (string::size_type) 1, withThis.size( ) );
  string::size_type i = s.find( thisPattern, 0 );
  while ( i != string::npos )
   {
    s.replace( i, thisPattern.size( ), withThis );
    i = s.find( thisPattern, i + n );
   }

  return s;
 }

// ------------------------------------------------- separate string into tokens
// -----------------------------------------------------------------------------

void
MiscFns::getTokens( const string&         tokenStr,
                          vector<string>& tokens,
                    const string&         separators )
 {
  tokens.resize( 0 );

  string::size_type i0 = tokenStr.find_first_not_of( separators, 0 );
  string::size_type i1;

  while ( i0 != string::npos )
   {
    i1 = tokenStr.find_first_of( separators, i0 );
    tokens.push_back( tokenStr.substr( i0, i1 - i0 ) );
    i0 = tokenStr.find_first_not_of( separators, i1 );
   }
 }

vector<string>
MiscFns::tokens( const string& tokenStr, const string& separators )
 {
  vector<string> tokens;
  getTokens( tokenStr, tokens, separators );
  return tokens;
 }

// -------------------------------- range options from strings of the form "n:m"
// -----------------------------------------------------------------------------

void
MiscFns::getRange( const string& rangeStr, const pair<int,int>& defaults,
                                                 pair<int,int>& range )
 {
  int i1 = defaults.first;
  int i2 = defaults.second;

  string::size_type colonPos = rangeStr.find( ":" );

  if ( colonPos == string::npos )
   i1 = atoi( rangeStr.c_str( ) );
  else
   {
    string s1 = rangeStr.substr( 0, colonPos  );
    string s2 = rangeStr.substr( colonPos + 1 );

    if ( s1 != "" ) i1 = atoi( s1.c_str( ) );
    if ( s2 != "" ) i2 = atoi( s2.c_str( ) );
   }

  range.first  = max( defaults.first,  min( i1, i2 ) );
  range.second = min( defaults.second, max( i1, i2 ) );
 }

pair<int,int>
MiscFns::range( const string& rangeStr, const pair<int,int>& defaults )
 {
  pair<int,int> r;
  getRange( rangeStr, defaults, r );
  return r;
 }

// ---------------------------------- number of significant digits of an integer
// -----------------------------------------------------------------------------

size_t
MiscFns::nDigits( int n )
 {
  size_t digits = 1;

  if ( n < 0 )
   {
    n = -n;
    ++digits;
   }

  while ( n /= 10 ) ++digits;

  return digits;
 }

// --------------------------------- keys for iterating over binary permutations
// -----------------------------------------------------------------------------

void
MiscFns::getFirstKey( size_t size, size_t numOnes, MaskType& key )
 {
  key.resize( size, true );
  if ( size > numOnes )
   for ( size_t f = 0; f < size - numOnes; f++ ) key[f] = false;
                                                       //
                                                       // all ones right-aligned
 }

MaskType
MiscFns::firstKey( size_t size, size_t numOnes )
 {
  MaskType key;
  getFirstKey( size, numOnes, key );
  return key;
 }

void
MiscFns::getLastKey( size_t size, size_t numOnes, MaskType& key )
 {
  if ( size > numOnes )
   {
    key.resize( size, false );
    for ( size_t f = 0; f < numOnes; f++ ) key[f] = true;
   }
  else
   key.resize( size, true );
                                                        //
                                                        // all ones left-aligned
 }

MaskType
MiscFns::lastKey( size_t size, size_t numOnes )
 {
  MaskType key;
  getLastKey( size, numOnes, key );
  return key;
 }

#ifndef cpp_missing_template_methods

// --------------------------------------------------------- write array to cout
// -----------------------------------------------------------------------------

template<template<class T> class C, class T>
void
MiscFns::writeArray( const C<T>&    v,
                           ostream& output,
                     const string&  linePrefix,
                           bool     scientific,
                           int      width,
                           size_t   precision,
                           size_t   numbersPerLine,
                     const string&  separator )
 {
  ios_base::fmtflags   f = cout.flags( );
  streamsize p = cout.precision( );
  streamsize w = cout.width( );

  if ( scientific ) cout.setf( ios_base::scientific, ios_base::floatfield );

  size_t i = 0;
  while ( i < v.size( ) )
   {
    cout << linePrefix;
    for ( size_t j = 0; ( i < v.size( ) ) && ( j < numbersPerLine ); j++, i++ )
     cout << separator << setw( width ) << setprecision( precision ) << v[i];
    cout << endl;
   }

  cout.flags( f ); cout.precision( p ); cout.width( w );
 }

// ------------------------------------------------------------- array -> string
// -----------------------------------------------------------------------------

template<template<class T> class C, class T>
string
MiscFns::asString( const C<T>&   ct,
                   const string& singleNumberFormat = "%g",
                   const string& separator          = " " )
 {
  string fmt = singleNumberFormat + "%n";
  char s[128]; // ATTN arbitrary size; should suffice except when T is string
  int n;

  string ctstr = "";
  for ( size_t i = 0; i < ct.size( ); i++ )
   {
    sprintf( s, fmt.c_str( ), ct[i], &n );
    char ns[n]; sprintf( ns, singleNumberFormat.c_str( ), ct[i] );
    if ( i != 0 ) ctstr.append( separator );
    ctstr.append( ns, n );
   }

  return ctstr;
 }

// -------------------------------- buffer copy with static_cast type conversion
// -----------------------------------------------------------------------------

template< class T1, class T2 >
void
MiscFns::castAndCopy( size_t size, const T1* const from, T2* const into )
 {
  for ( size_t i = 0; i < size; i++ ) into[i] = static_cast<T2>( from[i] );
 }

#endif /* cpp_missing_template_methods */

#endif /* __MiscFns_H__ */
