/* --------------------------------------------------------------------------
 *
 * namespace MiscNumericFns
 *   miscellaneous numeric functions
 *
 * required headers
 *   see include header file MiscNumericFns
 *
 * notes
 *   contains function templates; may not be supported
 *   by all compilers (e. g., SUNWspro 5.0).
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __MiscNumericFns_H__
#define __MiscNumericFns_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

cpp_namespace_begin( MiscNumericFns )

  // ------------------------------------------------------------- miscellaneous
  // ---------------------------------------------------------------------------

   cpp_namespace_decl
   inline
   fpType
   nCr( int n, int r );
                                               //
                                               //    n! / r! (n-r)! evaluated as
                                               //
                                               // product i = 1 to min( r, n-r )
                                               //      ( 1 + max( r, n-r ) / i )
                                               //
                                               //                    returns 0.0
                                               //     if r > n || r < 0 || n < 0

   cpp_namespace_decl
   inline
   fpType
   adjustIntegralLimits( fpType& lower, fpType& upper );
                                              //
                                              // if lower >  upper : return -1.0
                                              // and interchange lower and upper
                                              //
                                              //
                                              // if lower == upper : return  0.0
                                              // if lower <  upper : return +1.0

  // ------------------------------------------------------- roots of a function
  // ---------------------------------------------------------------------------

#ifndef cpp_missing_template_methods

   cpp_namespace_decl
   template< class Function1D >
   fpType
   BisectionSearch( fpType     lowerBound,
                    fpType     upperBound,
                    int&       iterations,
                    fpType&    closeness,
                    Function1D function1D,
                    fpType     threshold     = 0.0,
                    int        maxIterations = 100,
                    fpType     tolerance     = 1.0e+2 * DBL_EPSILON );

#endif /* cpp_missing_template_methods */

cpp_namespace_end

// --------------------------------------------------------------- miscellaneous
// -----------------------------------------------------------------------------

fpType
MiscNumericFns::nCr( int n, int r )
 {
  fpType p = 0.0;

  if ( ( n >= 0 ) && ( r >= 0 ) && ( r <= n ) )
   {
    fpType k = (fpType) max( r, n - r );

    p = 1.0;
    for ( int i = 1; i <= min( r, n - r ); i++ )
     p *= ( 1.0 + k / ( (fpType) i ) );
   }

  return p;
 }

fpType
MiscNumericFns::adjustIntegralLimits( fpType& lower, fpType& upper )
 {
  fpType sign = +1.0;

  if ( lower > upper )
   {
    fpType t = lower;
    lower    = upper;
    upper    = t;

    sign     = -1.0;
   }
  else if ( lower == upper )
    sign     =  0.0;

  return sign;
 }

// --------------------------------------------------------- roots of a function
// -----------------------------------------------------------------------------

#ifndef cpp_missing_template_methods

template< class Function1D >
fpType
MiscNumericFns::BisectionSearch( fpType     lowerBound,
                                 fpType     upperBound,
                                 int&       iterations,
                                 fpType&    closeness,
                                 Function1D function1D,
                                 fpType     threshold,
                                 int        maxIterations,
                                 fpType     tolerance )
 {
  fpType root, f, sign;

  sign      =   ( function1D( upperBound ) - function1D( lowerBound ) > 0.0 )
              ? +1.0
              : -1.0;

  root      = 0.5 * ( lowerBound + upperBound );
  f         = sign * ( function1D( root ) - threshold );
  closeness = fabs( f );

  for ( iterations = 0;    ( iterations < maxIterations )
                        && ( closeness  > tolerance     ); iterations++ )
   {
    if ( f > 0.0 )
     upperBound = root;
    else if ( f == 0.0 )
     lowerBound = upperBound = root;
    else
     lowerBound = root;

    root      = 0.5 * ( lowerBound + upperBound );
    f         = sign * ( function1D( root ) - threshold );
    closeness = fabs( f );
   }

  return root;
 }

#endif /* cpp_missing_template_methods */

#endif /* __MiscNumericFns_H__ */
