/* --------------------------------------------------------------------------
 *
 * class ParameterManager
 *   parameter block organizer
 *
 * required headers
 *   see include header file ParameterManager
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __ParameterManager_H__
#define __ParameterManager_H__

cpp_using_namespace( std );

template<class T>
class ParameterManager
 {
  public:
   typedef varray< T >                   ArrayType;
   typedef map< string, pair<int, int> > LookupType;
                                   //
                                   // pair.first  == offset of a parameter block
                                   // pair.second == length of a parameter block

   typedef LookupType::const_iterator    const_iterator;
   typedef LookupType::iterator          iterator;

   // -------------------------------------------------------- constructors etc.
   // --------------------------------------------------------------------------

   ParameterManager( int offsetShift = 0 );
                                             //
                                             // any overall shift in the offsets

   ParameterManager( const ParameterManager<T>& another );

   ParameterManager<T>&
   operator=( const ParameterManager<T>& rhs );

   void
   setOffsetShift( int offsetShift = 0 );
                                             //
                                             // any overall shift in the offsets

   // ---------------------------------------------- add/delete parameter blocks
   // --------------------------------------------------------------------------

   void
   addBlock( const string& name, int length );

   void
   delBlock( const string& name );

   // -------------------------------------------- block names, lengths, offsets
   // --------------------------------------------------------------------------

   inline
   int
   numBlocks( ) const;

   inline
   bool
   hasBlock( const string& name ) const;

                                               // 0 <= blockIndex < numBlocks( )

   inline
   string
   blockName( int blockIndex ) const;        // returns "" if blockIndex invalid

   inline
   int
   blockLength( const string& name ) const;     // returns  0 if name is invalid

   inline
   int
   blockLength( int blockIndex ) const;      // returns  0 if blockIndex invalid

   inline
   int
   blockOffset( const string& name ) const;
                                   //
                                   // includes any overall shift in the offsets;
                                   //   - offsetShift( ) - 1 for an invalid name

   inline
   int
   blockOffset( int blockIndex ) const;
                                   //
                                   // includes any overall shift in the offsets;
                                   //  - offsetShift( ) - 1 for an invalid index

   inline
   int
   numParams( ) const;                       // sum total of all blockLength( )s

   inline
   int
   offsetShift( ) const;
                                             //
                                             // any overall shift in the offsets

   inline
   const LookupType&
   map( ) const;                            // for more elaborate map operations

   // ------------------------------------------------------ extract value/block
   // --------------------------------------------------------------------------

   T
   value( const string& name,
                int     index,
          const T*      paramBuf ) const;
                                 //
                                 //   returns T( ) if name &/or index is invalid
                                 // ATTN: may be misleading if not used properly

   ArrayType
   block( const string& name,
          const T*      paramBuf ) const;
                                      //
                                      // returns ArrayType( ) if name &/or index
                                      //                              is invalid

   void
   getValue( const string& name,
                   int     index,
             const T*      paramBuf,
                   T&      here2extract ) const;
                                 //
                                 //   returns T( ) if name &/or index is invalid
                                 // ATTN: may be misleading if not used properly

   void
   getBlock( const string& name,
             const T*      paramBuf,
                   T*      here2extract ) const;
                                      //
                                      // ATTN: here2extract must have right size

   // ---------------------------------------------------------- set value/block
   // --------------------------------------------------------------------------

   void
   setValue( const string& name,
                   int     index,
                   T*      paramBuf,
             const T&      newValue ) const;

   void
   setBlock( const string& name,
                   T*      paramBuf,
             const T*      here2copyFrom ) const;
                               //
                               // ATTN: here2copyFrom must have the correct size

  private:
   LookupType my_m;
   int        my_next_index, my_offset_shift;

   inline
   bool
   check( const string& name, int index, const_iterator& q ) const;

   inline
   bool
   check( const string& name, const_iterator& q ) const;

   inline
   bool
   check( const string& name, int index, iterator& q );

   inline
   bool
   check( const string& name, iterator& q );

   inline
   void
   del( iterator& q );
 };

// ---------------------------------------------------------- constructors, etc.
// -----------------------------------------------------------------------------

template<class T>
ParameterManager<T>::ParameterManager( int offsetShift )
 : my_m( LookupType( ) ),
   my_next_index( 0 ),
   my_offset_shift( offsetShift )
 {
 }

template<class T>
ParameterManager<T>::ParameterManager( const ParameterManager<T>& another )
 : my_m( another.my_m ),
   my_next_index( another.my_next_index ),
   my_offset_shift( another.my_offset_shift )
 {
 }

template<class T>
ParameterManager<T>&
ParameterManager<T>::operator=( const ParameterManager<T>& rhs )
 {
  if ( &rhs != this )
   {
    my_m            = rhs.my_m;
    my_next_index   = rhs.my_next_index;
    my_offset_shift = rhs.my_offset_shift;
   }

  return *this;
 }

template<class T>
void
ParameterManager<T>::setOffsetShift( int offsetShift )
 {
  my_offset_shift = offsetShift;
 }

// ------------------------------------------------- add/delete parameter blocks
// -----------------------------------------------------------------------------

template<class T>
void
ParameterManager<T>::addBlock( const string& name, int length )
 {
  iterator q;

  if ( length > 0 )
   {
    if ( check( name, q ) ) del( q );

    my_m.insert( pair<string, pair<int, int> >
                 ( name, pair<int, int>( my_next_index, length ) ) );

    my_next_index += length;
   }
 }

template<class T>
void
ParameterManager<T>::delBlock( const string& name )
 {
  iterator q;

  if( check( name, q ) ) del( q );
 }

// ---------------------------------------------------- block length and offsets
// -----------------------------------------------------------------------------

template<class T>
int
ParameterManager<T>::numBlocks( ) const
 {
  return my_m.size( );
 }

template<class T>
bool
ParameterManager<T>::hasBlock( const string& name ) const
 {
  const_iterator q;
  return check( name, q );
 }

template<class T>
string
ParameterManager<T>::blockName( int blockIndex ) const
 {
  if ( ( 0 <= blockIndex ) && ( blockIndex < numBlocks( ) ) )
   {
    const_iterator mi = my_m.begin( );
    for ( int i = 0; i < blockIndex; i++, ++mi ) { }

    return mi->first;
   }
  else
   return string( );
 }

template<class T>
int
ParameterManager<T>::blockLength( const string& name ) const
 {
  const_iterator q;
  return ( check( name, q ) ? (*q).second.second : 0 );
 }

template<class T>
int
ParameterManager<T>::blockLength( int blockIndex ) const
 {
  if ( ( 0 <= blockIndex ) && ( blockIndex < numBlocks( ) ) )
   {
    const_iterator mi = my_m.begin( );
    for ( int i = 0; i < blockIndex; i++, ++mi ) { }

    return mi->second.second;
   }
  else
   return 0;
 }

template<class T>
int
ParameterManager<T>::blockOffset( const string& name ) const
 {
  const_iterator q;
  return
     check( name, q )
   ? (*q).second.first - my_offset_shift
   : - ( 1 + my_offset_shift );
 }

template<class T>
int
ParameterManager<T>::blockOffset( int blockIndex ) const
 {
  if ( ( 0 <= blockIndex ) && ( blockIndex < numBlocks( ) ) )
   {
    const_iterator mi = my_m.begin( );
    for ( int i = 0; i < blockIndex; i++, ++mi ) { }

    return mi->second.first - my_offset_shift;
   }
  else
   return - ( 1 + my_offset_shift );
 }

template<class T>
int
ParameterManager<T>::numParams( ) const
 {
  return my_next_index;
 }

template<class T>
int
ParameterManager<T>::offsetShift( ) const
 {
  return my_offset_shift;
 }

template<class T>
const ParameterManager<T>::LookupType&
ParameterManager<T>::map( ) const
 {
  return my_m;
 }

// --------------------------------------------------------- extract value/block
// -----------------------------------------------------------------------------

template<class T>
T
ParameterManager<T>::value( const string& name,
                                  int     index,
                            const T*      paramBuf ) const
 {
  const_iterator q;

  return
   (
    check( name, index, q ) ? paramBuf[(*q).second.first + index] : T( )
   );
 }

template<class T>
ParameterManager<T>::ArrayType
ParameterManager<T>::block( const string& name,
                            const T*      paramBuf ) const
 {
  const_iterator q;

  if ( check( name, q ) )
   {
    ArrayType here2extract( (*q).second.second );
    for ( int i = (*q).second.first, j = 0;
          i < ( (*q).second.first + (*q).second.second ); i++, j++ )
     here2extract[j] = paramBuf[i];
    return here2extract;
   }
  else
   return ArrayType( );
 }

template<class T>
void
ParameterManager<T>::getValue( const string& name,
                                     int     index,
                               const T*      paramBuf,
                                     T&      here2extract ) const
 {
  const_iterator q;

  here2extract =   check( name, index, q )
                 ? paramBuf[(*q).second.first + index]
                 : T( );
 }

template<class T>
void
ParameterManager<T>::getBlock( const string& name,
                               const T*      paramBuf,
                                     T*      here2extract ) const
 {
  const_iterator q;

  if ( check( name, q ) )
   {
    // older version had: here2extract.resize( (*q).second.second );

    for ( int i = (*q).second.first, j = 0;
          i < ( (*q).second.first + (*q).second.second ); i++, j++ )
     here2extract[j] = paramBuf[i];
   }
 }

// ------------------------------------------------------------- set value/block
// -----------------------------------------------------------------------------

template<class T>
void
ParameterManager<T>::setValue( const string& name,
                                     int     index,
                                     T*      paramBuf,
                               const T&      newValue ) const
 {
  const_iterator q;

  if ( check( name, index, q ) ) paramBuf[(*q).second.first + index] = newValue;
 }

template<class T>
void
ParameterManager<T>::setBlock( const string& name,
                                     T*      paramBuf,
                               const T*      here2copyFrom ) const
 {
  const_iterator q;

  if ( check( name, q ) )
   for ( int i = (*q).second.first, j = 0;
         i < ( (*q).second.first + (*q).second.second ); i++, j++ )
    paramBuf[i] = here2copyFrom[j];
 }

// -------------------------------------------------------------------- check( )
// -----------------------------------------------------------------------------

template<class T>
bool
ParameterManager<T>::check( const string& name, int index, const_iterator& q ) const
 {
  return
   (
    check( name, q ) && ( index >= 0 ) && ( index < (*q).second.second )
   );
 }

template<class T>
bool
ParameterManager<T>::check( const string& name, const_iterator& q ) const
 {
  q = my_m.find( name );
  return ( q != my_m.end( ) );
 }

template<class T>
bool
ParameterManager<T>::check( const string& name, int index, iterator& q )
 {
  return
   (
    check( name, q ) && ( index >= 0 ) && ( index < (*q).second.second )
   );
 }

template<class T>
bool
ParameterManager<T>::check( const string& name, iterator& q )
 {
  q = my_m.find( name );
  return ( q != my_m.end( ) );
 }

template<class T>
void
ParameterManager<T>::del( iterator& q )
 {
  int a = (*q).second.first;
  int b = (*q).second.second;

  my_m.erase( q );

  iterator mp;

  for ( mp = my_m.begin( ); mp != my_m.end( ); ++mp )
   if ( (*mp).second.first > a ) (*mp).second.first -= b;

  my_next_index -= b;
 }

#endif /* __ParameterManager_H__ */
