/* --------------------------------------------------------------------------
 *
 * class SplineRampC2
 *   spline-based symmetric up ramp; C^2 at 0 and 1
 *
 * required headers
 *   see include header file SplineRampC2
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __SplineRampC2_H__
#define __SplineRampC2_H__

cpp_using_namespace( BasicTypes );

class SplineRampC2
 {
  public:

                   //                                  spline-based up ramp: C^2
                   //
                   //                                    cubic spline with knots
                   // { -3/3, -2/3, -1/3, 0, 1/3, 2/3, 3/3, 4/3, 5/3, 6/3, 7/3 }
                   //                                  and b-spline coefficients
                   //                                    { 0, 0, 0, 1, 1, 1, 1 }

   static
   inline
   string
   name( );

   static
   inline
   string
   description( );

   inline
   fpType
   operator()( fpType t ) const;

   inline
   fpType
   Dt( fpType t, int dOrder = 1 ) const;

   inline
   fpType
   integral( fpType upperLimit = 1.0 ) const;
                                                 //
                                                 // lower limit taken to be zero

  private:
   static const fpType our_oneThird;
   static const fpType our_twoThird;
 };

string
SplineRampC2::name( )
 {
  return "SplineRampC2";
 }

string
SplineRampC2::description( )
 {
  return
   "cubic spline: knots i/3, i=-3,...,7; coefficients 0, 0, 0, 1, 1, 1, 1";
 }

fpType
SplineRampC2::operator()( fpType t ) const
 {
  if ( t <= 0.0 )
   return 0.0;
  if ( t >= 1.0 )
   return 1.0;
  else if ( t <= our_oneThird )
   return 4.5 * t * t * t;
  else if ( t <= our_twoThird )
   return 0.5 * ( 1.0 + t * ( -9.0 + t * ( 27.0 - 18.0 * t ) ) );
  else
   return 0.5 * ( -7.0 + t * ( 27.0 + t * ( -27.0 + 9.0 * t ) ) );
 }

fpType
SplineRampC2::Dt( fpType t, int dOrder ) const
 {
  if ( dOrder <= 0 )
   return (*this)( t );
  else if ( ( dOrder > 3 ) || ( t <= 0.0 ) || ( t >= 1.0 ) )
   return 0.0;
  else if ( dOrder == 1 )
   {
    if ( t <= our_oneThird )
     return 13.5 * t * t;
    else if ( t <= our_twoThird )
     return 4.5 * ( -1.0 + 6.0 * t * ( 1.0 - t ) );
    else
     return 13.5 * ( 1.0 + t * ( t - 2.0 ) );
   }
  else if ( dOrder == 2 )
   {
    if ( t <= our_oneThird )
     return 27.0 * t;
    else if ( t <= our_twoThird )
     return 27.0 * ( 1.0 - 2.0 * t );
    else
     return 27.0 * ( t - 1.0 );
   }
  else /* if ( dOrder == 3 ) */
   {
    if ( t <= our_oneThird )
     return +27.0;
    else if ( t <= our_twoThird )
     return -54.0;
    else
     return +27.0;
   }
 }

fpType
SplineRampC2::integral( fpType upperLimit ) const
 {
  if ( upperLimit <= 0.0 )
   return 0.0;
  else if ( upperLimit >= 1.0 )
   return 0.5;
  else
   {
    fpType sum = 0.0;
    fpType t;

    if ( upperLimit > our_twoThird )
     {
      t = upperLimit - our_twoThird;
      sum += 0.5 * t * ( -7.0 + t * ( 13.5 + t * ( -9.0 + 2.25 * t ) ) );
     }

    if ( upperLimit > our_oneThird )
     {
      t = min( upperLimit, our_twoThird ) - our_oneThird;
      sum += 0.5 * t * ( 1.0 + t * ( -4.5 + t * ( 9.0 - 4.5 * t ) ) );
     }

    return sum + 1.125 * pow( min( upperLimit, our_oneThird ), 4.0 );
   }
 }

#endif /* __SplineRampC2_H__ */
