/* --------------------------------------------------------------------------
 *
 * class polynomial
 *   a simple floating-point polynomial class
 *
 * required headers
 *   see include header file polynomial
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __polynomial_H__
#define __polynomial_H__

cpp_using_namespace( BasicTypes );

class polynomial
 {
  public:
   // --------------------------------------------------------------------------
   // ------------------------------------------------------------- constructors

   polynomial( );

   explicit polynomial( size_t degree );

   polynomial( size_t degree, const fpType* coeff );

   polynomial( const fpVector& coeff );

   polynomial( const polynomial& another );

   // --------------------------------------------------------------------------
   // ------------------------------------------------------------- reinitialize

   void
   reinitialize( size_t degree = 0 );

   void
   reinitialize( size_t degree, const fpType* coeff );

   void
   reinitialize( const fpVector& coeff );

   void
   reinitialize( const polynomial& another );

   // --------------------------------------------------------------------------
   // --------------------------------------- polynomial degree and coefficients

   inline
   size_t
   degree( ) const;

   inline
   fpType
   operator[]( size_t index ) const;

   inline
   fpType&
   operator[]( size_t index );

   inline
   fpVector
   coeff( ) const;

   inline
   fpVector&
   coeff( );

   // --------------------------------------------------------------------------
   // -------------------------------------- evaluation, integral and derivative

   inline
   fpType
   operator()( fpType abscissa ) const;

   inline
   fpType
   integral( fpType abscissa, int order = 1 ) const;

   inline
   fpType
   derivative( fpType abscissa, int order = 1 ) const;

   inline
   polynomial
   integral( int order = 1 ) const;

   inline
   polynomial
   derivative( int order = 1 ) const;

   // --------------------------------------------------------------------------
   // ----------------------------------------------------- assignment operators

   inline
   polynomial&
   operator=( const polynomial& rhs );

   inline
   polynomial&
   operator=( fpType rhs );

   inline
   polynomial&
   operator+=( fpType rhs );

   inline
   polynomial&
   operator-=( fpType rhs );

   inline
   polynomial&
   operator*=( fpType rhs );

   inline
   polynomial&
   operator/=( fpType rhs );

   inline
   polynomial&
   operator+=( const polynomial& rhs );

   inline
   polynomial&
   operator-=( const polynomial& rhs );

   inline
   polynomial&
   operator*=( const polynomial& rhs );

   // --------------------------------------------------------------------------
   // ---------------------------------------------------------- unary operators

   inline
   polynomial
   operator-( ) const;

   inline
   polynomial
   operator+( ) const;

  private:
   fpVector my_coeff;
 };

// -----------------------------------------------------------------------------
// ------------------------------------------ polynomial degree and coefficients

size_t
polynomial::degree( ) const
 {
  return my_coeff.size( ) - 1;
 }

fpType
polynomial::operator[]( size_t index ) const
 {
  return my_coeff[index];
 }

fpType&
polynomial::operator[]( size_t index )
 {
  return my_coeff[index];
 }

fpVector
polynomial::coeff( ) const
 {
  return my_coeff;
 }

fpVector&
polynomial::coeff( )
 {
  return my_coeff;
 }

// -----------------------------------------------------------------------------
// ----------------------------------------- evaluation, integral and derivative

fpType
polynomial::operator()( fpType abscissa ) const
 {
  size_t myDegree = degree( );
  fpType sum = my_coeff[myDegree];
  for ( int i = myDegree - 1; i >= 0; i-- )
   sum = abscissa * sum + my_coeff[i];
  return sum;
 }

fpType
polynomial::integral( fpType abscissa, int order ) const
 {
  if ( order < 0 )
   return derivative( abscissa, -order );
  else if ( order == 0 )
   return (*this)( abscissa );
  else
   {
    size_t myDegree = degree( );

    fpType factor = 1.0;
    for ( int i = 1; i <= order; i++ ) factor /= (fpType) ( myDegree + i );

    fpType sum = factor * my_coeff[myDegree];
    for ( int i = myDegree - 1; i >= 0; i-- )
     {
      factor *= ( i + 1 + order );
      factor /= ( i + 1 );

      sum = abscissa * sum + factor * my_coeff[i];
     }

    return sum * pow( abscissa, order );
   }
 }

fpType
polynomial::derivative( fpType abscissa, int order ) const
 {
  size_t myDegree = degree( );

  if ( order < 0 )
   return integral( abscissa, -order );
  else if ( order == 0 )
   return (*this)( abscissa );
  else if ( order > static_cast<int>( myDegree ) )
   return 0.0;
  else
   {
    fpType factor = 1.0;
    for ( int i = 0; i < order; i++ ) factor *= ( myDegree - i );

    fpType sum = factor * my_coeff[myDegree];
    for ( int i = myDegree - 1; i >= order; i-- )
     {
      factor /= ( i + 1 );
      factor *= ( i + 1 - order );

      sum = abscissa * sum + factor * my_coeff[i];
     }

    return sum;
   }
 }

polynomial
polynomial::integral( int order ) const
 {
  if ( order < 0 )
   return derivative( -order );
  else if ( order == 0 )
   return (*this);
  else
   {
    size_t myDegree = degree( );
    polynomial p( myDegree + order );

    fpType factor = 1.0;
    for ( int i = 1; i <= order; i++ ) factor /= (fpType) ( myDegree + i );

    p[myDegree + order] = factor * my_coeff[myDegree];
    for ( int i = myDegree - 1; i >= 0; i-- )
     {
      factor *= ( i + 1 + order );
      factor /= ( i + 1 );

      p[order + i] = factor * my_coeff[i];
     }

    return p;
   }
 }

polynomial
polynomial::derivative( int order ) const
 {
  size_t myDegree = degree( );

  if ( order < 0 )
   return integral( -order );
  else if ( order == 0 )
   return (*this);
  else if ( order > static_cast<int>( myDegree ) )
   return polynomial( );
  else
   {
    polynomial p( myDegree - order );

    fpType factor = 1.0;
    for ( int i = 0; i < order; i++ ) factor *= ( myDegree - i );

    p[myDegree - order] = factor * my_coeff[myDegree];
    for ( int i = myDegree - 1; i >= order; i-- )
     {
      factor /= ( i + 1 );
      factor *= ( i + 1 - order );

      p[i - order] = factor * my_coeff[i];
     }

    return p;
   }
 }

// -----------------------------------------------------------------------------
// -------------------------------------------------------- assignment operators

polynomial&
polynomial::operator=( const polynomial& rhs )
 {
  my_coeff.resize( rhs.my_coeff.size( ) );
  my_coeff = rhs.my_coeff;
  return *this;
 }

polynomial&
polynomial::operator=( fpType rhs )
 {
  my_coeff = rhs;
  return *this;
 }

polynomial&
polynomial::operator+=( fpType rhs )
 {
  my_coeff[0] += rhs;
  return *this;
 }

polynomial&
polynomial::operator-=( fpType rhs )
 {
  my_coeff[0] -= rhs;
  return *this;
 }

polynomial&
polynomial::operator*=( fpType rhs )
 {
  my_coeff *= rhs;
  return *this;
 }

polynomial&
polynomial::operator/=( fpType rhs )
 {
  my_coeff /= rhs;
  return *this;
 }

polynomial&
polynomial::operator+=( const polynomial& rhs )
 {
  size_t lhsSize = my_coeff.size( );
  size_t rhsSize = rhs.my_coeff.size( );

  if ( rhsSize <= lhsSize )
   for ( size_t i = 0; i < rhsSize; i++ ) my_coeff[i] += rhs.my_coeff[i];
  else
   {
    fpVector t( rhs.my_coeff );
    for ( size_t i = 0; i < lhsSize; i++ ) t[i] += my_coeff[i];

    my_coeff.resize( rhsSize );
    my_coeff = t;
   }

  return *this;
 }

polynomial&
polynomial::operator-=( const polynomial& rhs )
 {
  size_t lhsSize = my_coeff.size( );
  size_t rhsSize = rhs.my_coeff.size( );

  if ( rhsSize <= lhsSize )
   for ( size_t i = 0; i < rhsSize; i++ ) my_coeff[i] -= rhs.my_coeff[i];
  else
   {
    fpVector t( rhs.my_coeff );
    for ( size_t i = 0; i < lhsSize; i++ ) t[i] -= my_coeff[i];

    my_coeff.resize( rhsSize );
    my_coeff = -t;
   }

  return *this;
 }

polynomial&
polynomial::operator*=( const polynomial& rhs )
 {
  size_t lhsSize = my_coeff.size( );
  size_t rhsSize = rhs.my_coeff.size( );

  fpVector t( 0.0, rhsSize + lhsSize - 1 );

  for ( size_t i = 0; i < lhsSize; i++ )
   for ( size_t j = 0; j < rhsSize; j++ )
    t[i+j] += my_coeff[i] * rhs.my_coeff[j];

  my_coeff.resize( t.size( ) );
  my_coeff = t;

  return *this;
 }

// -----------------------------------------------------------------------------
// ------------------------------------------------------------- unary operators

polynomial
polynomial::operator-( ) const
 {
  fpVector c( my_coeff ); c *= -1.0; // hack for g++ 2.96
  return polynomial( my_coeff );
 }

polynomial
polynomial::operator+( ) const
 {
  return polynomial( my_coeff );
 }

// -----------------------------------------------------------------------------
// --------------------------------------------------------- nonmember functions

inline
polynomial
operator+( const polynomial& rhs1, const polynomial& rhs2 )
 {
  polynomial p( rhs1 );
  p += rhs2;
  return p;
 }

inline
polynomial
operator-( const polynomial& rhs1, const polynomial& rhs2 )
 {
  polynomial p( rhs1 );
  p -= rhs2;
  return p;
 }

inline
polynomial
operator*( const polynomial& rhs1, const polynomial& rhs2 )
 {
  polynomial p( rhs1 );
  p *= rhs2;
  return p;
 }

inline
polynomial
operator+( const polynomial& rhs1, fpType rhs2 )
 {
  polynomial p( rhs1 );
  p += rhs2;
  return p;
 }

inline
polynomial
operator+( fpType rhs1, const polynomial& rhs2 )
 {
  polynomial p( rhs2 );
  p += rhs1;
  return p;
 }

inline
polynomial
operator-( const polynomial& rhs1, fpType rhs2 )
 {
  polynomial p( rhs1 );
  p -= rhs2;
  return p;
 }

inline
polynomial
operator-( fpType rhs1, const polynomial& rhs2 )
 {
  polynomial p( rhs2 );
  p -= rhs1;
  return -p;
 }

inline
polynomial
operator*( const polynomial& rhs1, fpType rhs2 )
 {
  polynomial p( rhs1 );
  p *= rhs2;
  return p;
 }

inline
polynomial
operator*( fpType rhs1, const polynomial& rhs2 )
 {
  polynomial p( rhs2 );
  p *= rhs1;
  return p;
 }

inline
polynomial
operator/( const polynomial& rhs1, fpType rhs2 )
 {
  polynomial p( rhs1 );
  p /= rhs2;
  return p;
 }

#endif /* __polynomial_H__ */
