#include <cmath>
#include <iomanip>
#include <iostream>
#include <ios>
#include <string>

#include <sprng.h>

#include <cppconfig/cppconfig>

#include <b-splines/BSplineBasis>
#include <basic/BasicTypes>
#include <basic/ParameterManager>
#include <brain.model/Bell>
#include <brain.model/ParamBlkNames>
#include <brain.model/SubmodelTable>
#include <brain.model/acProfile>
#include <brain.model/dbProfile>
#include <pghmri/PghMRIhandler>

cpp_using_namespace( BasicTypes );

fpType
normal( fpType ur1, fpType ur2, fpType sd = 1.0, fpType mean = 0.0 )
 {
  return
   mean + sd * sqrt( -2.0 * log( ur1 ) ) * cos( 6.2831853071795864769252867663 * ur2 );
 }

void
makeParamBlkDescriptions( const ParameterManager<fpType>& parManager,
                          const acProfile&                activation,
                                string&                   pBlkDscr,
                                string&                   pBlkDscrResponse,
                                string&                   pBlkDscrShape )
 {
  pBlkDscr = pBlkDscrResponse = pBlkDscrShape = "";

  char s[parManager.numParams( )];

  for ( int b = 0; b < parManager.numBlocks( ); b++ )
   {
    if ( b > 0 ) pBlkDscr.append( " " );
    pBlkDscr.append( parManager.blockName( b ) );

    sprintf( s, "%d", parManager.blockOffset( b ) );
    pBlkDscr.append( "," ); pBlkDscr.append( s );

    sprintf( s, "%d", parManager.blockLength( b ) );
    pBlkDscr.append( "," ); pBlkDscr.append( s );
   }

  for ( size_t a = 0; a < activation.numActive( ); a++ )
   {
    const acProfile::ConditionBellSet& cbs
     = activation.bellSet( activation.condition( a ) );

    if ( a > 0 ) pBlkDscrResponse.append( " " );
    pBlkDscrResponse.append( cbs.name( ) );
    pBlkDscrResponse.append( "." );
    pBlkDscrResponse.append( cbs.tag( ) );
    pBlkDscrResponse.append( "." );
    pBlkDscrResponse.append( cbs.bell( )->name( ) );

    sprintf( s, "%d", a );
    pBlkDscrResponse.append( "," ); pBlkDscrResponse.append( s );

    pBlkDscrResponse.append( ",1" );

    bool printBellInfo = true;
    for ( size_t b = 0; b < a; b++ )
     printBellInfo =    printBellInfo
                     && (    cbs.tag( )
                          != activation.bellSet( activation.condition( b ) ).tag( ) );

    if ( printBellInfo )
     {
      for ( size_t j = 0; j < cbs.numShapeParams( ); j++ )
       {
        if ( ( a > 0 ) || ( j > 0 ) ) pBlkDscrShape.append( " " );
        pBlkDscrShape.append( cbs.bell( )->name( j ) );
        pBlkDscrShape.append( "." );
        pBlkDscrShape.append( cbs.tag( ) );
        pBlkDscrShape.append( "." );
        pBlkDscrShape.append( cbs.bell( )->name( ) );

        sprintf( s, "%d", cbs.shapeOffset( ) + j );
        pBlkDscrShape.append( "," ); pBlkDscrShape.append( s );

        pBlkDscrShape.append( ",1" );
       }
     }
   }
 }

int main ( )
 {
  /* x, y, z, t grid sizes */

  size_t numX, numY, numZ, numT;
  cin >> numX >> numY >> numZ >> numT;

  fpVector timeGrid( numT );
  for ( size_t i = 0; i < numT; i++ ) timeGrid[i] = static_cast<fpType>( i );

  /* drift profile */

  BSplineBasis bs;
  dbProfile    db;

  cpp_scope_begin
   size_t numKnots, bsOrder;
   cin >> bsOrder >> numKnots;

   fpType delta = ( timeGrid[timeGrid.size( ) - 1] - timeGrid[0] ) / ( numKnots - 1 );

   fpVector knots( numKnots );

   for ( size_t i = 0; i < numKnots; i++ ) knots[i] = i * delta;

   int leftEP  = 0;
   int rightEP = numKnots - 1;

   bs.reinitialize( knots, bsOrder, leftEP, rightEP );
   db.reinitialize( timeGrid[0], timeGrid[timeGrid.size( ) - 1], &timeGrid, &bs );
  cpp_scope_end

  /* drift + baseline param specs */

  fpType bMin, bMax, bMean, bSD;
  cin >> bMean >> bSD >> bMin >> bMax;

  fpType dScale, dSD;
  cin >> dScale >> dSD;

  /* noise param specs */

  fpType noiseSD;
  size_t numAR;

  cin >> noiseSD >> numAR;

  fpVector ar( numAR );
  for ( size_t i = 0; i < numAR; i++ ) cin >> ar[i];

  /* activation */

  SubmodelTable smt;
  string        smKey;
  acProfile     ac;

  cpp_scope_begin
   size_t numConditions;
   cin >> numConditions;

   size_t numEpochs;
   string name, tag, bell;
   fpVector epochBegin, epochLength;

   for ( size_t c = 0; c < numConditions; c++ )
    {
     cin >> name >> tag >> bell >> numEpochs;

     epochBegin.resize( numEpochs );
     epochLength.resize( numEpochs );

     for ( size_t e = 0; e < numEpochs; e++ )
      cin >> epochBegin[e] >> epochLength[e];

     ac.addCondition( name, epochBegin, epochLength, tag, BellRegistry::name2bell( bell ) );
    }

   smt.setNumConditions( ac.numConditions( ) );

   cin >> smKey;

   smt.addByKey( smKey );

   ac.setSubmodel( &( *( smt.begin( ) ) ) );
   ac.setDBprofile( &db );
   ac.setTimeGrid( &timeGrid );
   ac.initialize( );
  cpp_scope_end

  /* response and shape param specs for active conditions */

  fpVector rMin( ac.numActive( ) );
  fpVector rMax( ac.numActive( ) );
  fpVector rMean( ac.numActive( ) );
  fpVector rSD( ac.numActive( ) );

  fpVector sMin( ac.numShapeParams( ) );
  fpVector sMax( ac.numShapeParams( ) );
  fpVector sMean( ac.numShapeParams( ) );
  fpVector sSD( ac.numShapeParams( ) );

  for ( size_t a = 0; a < ac.numActive( ); a++ )
   {
    const acProfile::ConditionBellSet& cbs = ac.bellSet( ac.condition( a ) );

    cin >> rMean[a] >> rSD[a] >> rMin[a] >> rMax[a];

    for ( size_t i = 0, j = cbs.shapeOffset( ); i < cbs.numShapeParams( ); i++, j++ )
     cin >> sMean[j] >> sSD[j] >> sMin[j] >> sMax[j];
   }

  /* parameter manager */

  ParameterManager<fpType> pm;

  pm.addBlock( ParamBlkNames::drift,    db.basisSize( )      );
  pm.addBlock( ParamBlkNames::noise,    1                    );
  pm.addBlock( ParamBlkNames::ar,       numAR                );
  pm.addBlock( ParamBlkNames::response, ac.numActive( )      );
  pm.addBlock( ParamBlkNames::shape,    ac.numShapeParams( ) );

  /* initialize .mri files: data, signal, noise, estimates */

  PghMRIhandler mriD, mriS, mriN, mriE;

  cpp_scope_begin
   string smDscr = smt.description( smt.begin( ) );

   string pBlkDscr, pBlkDscrR, pBlkDscrS;
   makeParamBlkDescriptions( pm, ac, pBlkDscr, pBlkDscrR, pBlkDscrS );

   /* data */

   string prefix, dataDscr;
   cin >> dataDscr >> prefix;
 
   mriD.open( prefix + "data", PghMRIhandler::mriWrite );

   mriD.setString( "imagedata.description",  dataDscr );
   mriD.setString( "submodel.key",           smKey    );
   mriD.setString( "submodel.description",   smDscr   );
   mriD.setString( "paramblock.description", pBlkDscr  );

   cpp_scope_begin
    int extents[5] = { 1, numT, numX, numY, numZ };
    mriD.createChunk( "images", "vtxyz", "float32", ".dat", extents );
   cpp_scope_end

   cpp_scope_begin
    int extents[2] = { numZ, numT };
    mriD.createChunk( "missing", "zt", "uint8", "", extents );
    varray<unsigned char> missing( numZ * numT );
    mriD.setChunk( "missing", missing.size( ), 0, PghMRIhandler::mriUchar, &( missing[0] ) );
   cpp_scope_end

   /* estimates == true parameters */

   mriE.open( prefix + smKey + ".estimates", PghMRIhandler::mriWrite );

   mriE.setString( "imagedata.description",  dataDscr  );
   mriE.setString( "submodel.key",           smKey     );
   mriE.setString( "submodel.description",   smDscr    );
   mriE.setString( "paramblock.description", pBlkDscr  );

   cpp_scope_begin
    int extents[5] = { pm.numParams( ), numX, numY, numZ };
    mriE.createChunk( "estimates", "wxyz", "float32", ".dat", extents );
   cpp_scope_end
 
   /* fit == true signal, residuals == true noise */

   mriS.open( prefix + smKey + ".fit", PghMRIhandler::mriWrite );

   mriS.setString( "imagedata.description",  dataDscr  );
   mriS.setString( "submodel.key",           smKey     );
   mriS.setString( "submodel.description",   smDscr    );
   mriS.setString( "paramblock.description", pBlkDscr  );

   mriN.open( prefix + smKey + ".residuals", PghMRIhandler::mriWrite );

   mriN.setString( "imagedata.description",  dataDscr  );
   mriN.setString( "submodel.key",           smKey     );
   mriN.setString( "submodel.description",   smDscr    );
   mriN.setString( "paramblock.description", pBlkDscr  );

   cpp_scope_begin
    int extents[5] = { numT, numX, numY, numZ };
    mriS.createChunk( "fit", "txyz", "float32", ".dat", extents );
    mriN.createChunk( "residuals", "txyz", "float32", ".dat", extents );
   cpp_scope_end

   if ( ac.numActive( ) != 0 )
    {
     mriD.setString( "paramblock.description.response", pBlkDscrR );
     mriD.setString( "paramblock.description.shape",    pBlkDscrS );
     mriE.setString( "paramblock.description.response", pBlkDscrR );
     mriE.setString( "paramblock.description.shape",    pBlkDscrS );
     mriS.setString( "paramblock.description.response", pBlkDscrR );
     mriS.setString( "paramblock.description.shape",    pBlkDscrS );
     mriN.setString( "paramblock.description.response", pBlkDscrR );
     mriN.setString( "paramblock.description.shape",    pBlkDscrS );
    }
  cpp_scope_end

  /* random number generator */

  int numRNS = 2; /* one stream for seeds, another for U(0,1) */
  int* rns_seed;

  cpp_scope_begin
   long int seed;
   cin >> seed;

   rns_seed = init_sprng( 0, numRNS, seed, SPRNG_DEFAULT );
  cpp_scope_end

  /* loop over voxels */

  long numVoxels = numX * numY * numZ;

  int*   rns;
  fpType baseline, dbMin, dbSlope, f, sum, tau, nSD;

  fpVector noise( numT + numAR );
  fpVector signal;

  fpVector params( pm.numParams( ) );

  // copy ar params to parameter buffer

  for ( size_t i = 0, j = pm.blockOffset( ParamBlkNames::ar ); i < numAR; i++, j++ )
   params[j] = ar[i];

  for ( long v = 0, ofsD = 0, ofsE = 0; v < numVoxels; v++, ofsD += numT, ofsE += pm.numParams( ) )
   {
    rns = init_sprng( 1, numRNS, isprng( rns_seed ), SPRNG_DEFAULT );

    // baseline

    do
     baseline = normal( sprng( rns ), sprng( rns ), bSD, bMean );
    while ( ( baseline < bMin ) || ( baseline > bMax ) );

    // linear drift

    dbMin   = baseline * ( 1.0 - dScale );
    dbSlope = ( 2.0 * baseline * dScale ) / ( timeGrid[timeGrid.size( ) - 1] - timeGrid[0] );
    f       = dbSlope / ( bs.order( ) - 1 );

    for ( int i = 0, j = pm.blockOffset( ParamBlkNames::drift );
                     i < pm.blockLength( ParamBlkNames::drift ); i++, j++ )
     {
      tau = bs.knots( )[bs.lowerSupport( i )];

      sum = 0.0;
      for ( int k = i + 1; k < i + bs.order( ); k++ )
       sum += ( bs.knots( )[k] - tau );

      params[j]
       =   dbMin + dbSlope * ( tau - timeGrid[0] ) + f * sum
         + normal( sprng( rns ), sprng( rns ), dSD, 0.0 );
     }

    // response and shape params

    for ( size_t a    = 0,
                 rofs = pm.blockOffset( ParamBlkNames::response ),
                 sofs = pm.blockOffset( ParamBlkNames::shape );
                 a    < ac.numActive( );
                 a++, rofs++ )
     {
      const acProfile::ConditionBellSet& cbs = ac.bellSet( ac.condition( a ) );

      do
       f = normal( sprng( rns ), sprng( rns ), rSD[a], rMean[a] );
      while ( ( f < rMin[a] ) || ( f > rMax[a] ) );

      params[rofs] = f;

      for ( size_t i = 0, j = cbs.shapeOffset( ); i < cbs.numShapeParams( ); i++, j++ )
       {
        do
         f = normal( sprng( rns ), sprng( rns ), sSD[j], sMean[j] );
        while ( ( f < sMin[j] ) || ( f > sMax[j] ) );

        params[sofs + j] = f;
       }
     }

    // noise

    nSD = noiseSD * baseline;
    params[pm.blockOffset( ParamBlkNames::noise )] = 1.0 / ( nSD * nSD );

    for ( size_t i = 0; i < noise.size( ); i++ )
     noise[i] = normal( sprng( rns ), sprng( rns ), nSD, 0.0 );

        // use noise[numAR] .. noise[noise.size( ) - 1]
        //
    if ( numAR > 0 )
     {
      for ( int i = noise.size( ) - 1; i >= numAR; i-- )
       for ( int j = 0; j < numAR; j++ )
        noise[i] -= ar[j] * noise[i-j-1];

      for ( size_t j = 0; j < numAR; j++ ) noise[j] = 0.0;
     }

    mriE.setChunk( "estimates", pm.numParams( ), ofsE, PghMRIhandler::mriDouble, params( 0 ) );

    mriN.setChunk( "residuals", numT, ofsD, PghMRIhandler::mriDouble, noise( numAR ) );

    // signal + noise

    db.getProfile( params( pm.blockOffset( ParamBlkNames::drift ) ), signal );

    signal += ac.profile( params( pm.blockOffset( ParamBlkNames::drift    ) ),
                          params( pm.blockOffset( ParamBlkNames::response ) ),
                          params( pm.blockOffset( ParamBlkNames::shape    ) ) );

    mriS.setChunk( "fit", numT, ofsD, PghMRIhandler::mriDouble, signal( 0 ) );

    for ( size_t i = 0, j = numAR; i < numT; i++, j++ ) signal[i] += noise[j];

    // output

    mriD.setChunk( "images", numT, ofsD, PghMRIhandler::mriDouble, signal( 0 ) );

    free_sprng( rns );
   }

  free_sprng( rns_seed );

  return 0;
 }
