/* --------------------------------------------------------------------------
 *
 * class BasicBell
 *   abstract interface for bell functions
 *
 * required headers
 *   see include header file BasicBell
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __BasicBell_H__
#define __BasicBell_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class BasicBell
 {
  public:

   // --------------------------------------------------------------------------
   // --------------------------------------------------------------------------

   BasicBell( const BasicBell& another );

   BasicBell&
   operator=( const BasicBell& rhs );

   virtual
   ~BasicBell( ) = 0;

   // -------------------------------------- bell function name and descriptions
   // --------------------------------------------------------------------------

   inline
   const string&
   name( ) const;

   inline
   const string&
   description( ) const;

   // --------------------------------------------- shape parameter descriptions
   // --------------------------------------------------------------------------

   inline
   size_t
   numParams( ) const;

   inline
   const string&
   name( size_t paramIndex ) const;

   inline
   const string&
   description( size_t paramIndex ) const;

   inline
   int
   paramIndex( const string& paramName ) const;
                              //
                              // -1 if no such parameter name (paramName) exists

   // ------------------------------------- suggested bounds on shape parameters
   // --------------------------------------------------------------------------

   inline
   const string&
   boundType( size_t paramIndex ) const;
                 //
                 //  "lower", "upper", "both", "none"; "" for invalid paramIndex

   inline
   fpType
   lowerBound( size_t paramIndex ) const;
                 //
                 // -DBL_MAX if boundType( paramIndex ) is "upper", "none" or ""

   inline
   fpType
   upperBound( size_t paramIndex ) const;
                 //
                 // +DBL_MAX if boundType( paramIndex ) is "lower", "none" or ""

   // ------------------------------------------------------------------ support
   // --------------------------------------------------------------------------

   virtual
   fpType
   lowerSupport(       fpType  stimulusBegin,
                       fpType  stimulusLength,
                 const fpType* shapeParams ) const = 0;

   virtual
   fpType
   upperSupport(       fpType  stimulusBegin,
                       fpType  stimulusLength,
                 const fpType* shapeParams ) const = 0;

   // -------------------------------------------- bell extrema over the support
   // --------------------------------------------------------------------------

   virtual
   inline
   fpType
   maxval( fpType stimulusLength, const fpType* shapeParams ) const;
                                       //
                                       // maximum value assumed by the bell over
                                       // its support for given shape parameters
                                       //
                                       //  numerical 1-dimensional maximization;
                                       // works unambiguously for unimodal bells
                                       //
                                       //   ATTN: not implemented; exits as stub

   virtual
   inline
   fpType
   minval( fpType stimulusLength, const fpType* shapeParams ) const;
                                       //
                                       // minimum value assumed by the bell over
                                       // its support for given shape parameters
                                       //
                                       //  numerical 1-dimensional minimization;
                                       // works unambiguously for unimodal bells
                                       //
                                       //   ATTN: not implemented; exits as stub

   // --------------------------------------------------------------- bell value
   // --------------------------------------------------------------------------

   virtual
   fpType
   operator()(       fpType  time,
                     fpType  stimulusBegin,
                     fpType  stimulusLength,
               const fpType* shapeParams ) const = 0;

   // -------------------------------------------------- derivatives w.r.t. time
   // --------------------------------------------------------------------------

   virtual
   fpType
   D1t(       fpType  time,
              fpType  stimulusBegin,
              fpType  stimulusLength,
        const fpType* shapeParams ) const = 0;

   virtual
   fpType
   Dt(       fpType  time,
             fpType  stimulusBegin,
             fpType  stimulusLength,
       const fpType* shapeParams,
             size_t  dOrder = 1 ) const = 0;

   // -------------------------------------- derivatives w.r.t. shape parameters
   // --------------------------------------------------------------------------

   virtual
   fpType
   gradient(       fpType  time,
                   fpType  stimulusBegin,
                   fpType  stimulusLength,
             const fpType* shapeParams,
                   size_t  paramIndex ) const = 0;

   virtual
   fpType
   hessian(       fpType  time,
                  fpType  stimulusBegin,
                  fpType  stimulusLength,
            const fpType* shapeParams,
                  size_t  paramIndex1,
                  size_t  paramIndex2 ) const = 0;

  protected:
   BasicBell( size_t numParams, const string& name, const string& description );

   void
   setParamInfo(       size_t  paramIndex,
                 const string& pname,
                 const string& pdescription = "none available",
                 const string& boundType    = "none",
                       fpType  lowerBound   = -DBL_MAX,
                       fpType  upperBound   = +DBL_MAX );

  private:
   string         my_name, my_description;
   vector<string> my_pnames, my_pdescriptions, my_boundTypes;
   vector<fpType> my_lowerBounds, my_upperBounds;
   static string  my_null;
 };

// ----------------------------------------- bell function name and descriptions
// -----------------------------------------------------------------------------

const string&
BasicBell::name( ) const
 {
  return my_name;
 }

const string&
BasicBell::description( ) const
 {
  return my_description;
 }

// ------------------------------------------------ shape parameter descriptions
// -----------------------------------------------------------------------------

size_t
BasicBell::numParams( ) const
 {
  return my_pnames.size( );
 }

const string&
BasicBell::name( size_t paramIndex ) const
 {
  return
     ( paramIndex >= 0 ) && ( paramIndex < my_pnames.size( ) )
   ? my_pnames[paramIndex]
   : my_null;
 }

const string&
BasicBell::description( size_t paramIndex ) const
 {
  return
     ( paramIndex >= 0 ) && ( paramIndex < my_pnames.size( ) )
   ? my_pdescriptions[paramIndex]
   : my_null;
 }

int
BasicBell::paramIndex( const string& paramName ) const
 {
  vector<string>::const_iterator
   i = find( my_pnames.begin( ), my_pnames.end( ), paramName );

  return
     ( i == my_pnames.end( ) )
   ? -1
   : distance( my_pnames.begin( ), i );
 }

// ---------------------------------------- suggested bounds on shape parameters
// -----------------------------------------------------------------------------

const string&
BasicBell::boundType( size_t paramIndex ) const
 {
  return
     ( paramIndex >= 0 ) && ( paramIndex < my_pnames.size( ) )
   ? my_boundTypes[paramIndex]
   : my_null;
 }

fpType
BasicBell::lowerBound( size_t paramIndex ) const
 {
  return
     ( paramIndex >= 0 ) && ( paramIndex < my_pnames.size( ) )
   ? my_lowerBounds[paramIndex]
   : -DBL_MAX;
 }

fpType
BasicBell::upperBound( size_t paramIndex ) const
 {
  return
     ( paramIndex >= 0 ) && ( paramIndex < my_pnames.size( ) )
   ? my_upperBounds[paramIndex]
   : +DBL_MAX;
 }

// ----------------------------------------------- bell extrema over the support
// -----------------------------------------------------------------------------

fpType
BasicBell::maxval( fpType stimulusLength, const fpType* shapeParams ) const
 {
  return 1.0;                          //   ATTN: not implemented; exits as stub
 }

fpType
BasicBell::minval( fpType stimulusLength, const fpType* shapeParams ) const
 {
  return 0.0;                          //   ATTN: not implemented; exits as stub
 }

#endif /* __BasicBell_H__ */
