/* --------------------------------------------------------------------------
 *
 * template class Bell_ST3
 *   3-parameter bell without dip, without any stimulus dependence
 *   (assuming short enough stimuli, as for single-trial data)
 *
 * required headers
 *   see include header file Bell
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __Bell_ST3_H__
#define __Bell_ST3_H__

template< class Ramp = SplineRampC2 >
class Bell_ST3 : public BasicBell
 {
  public:
   // ------------------------------------------------------- construct/destruct
   // --------------------------------------------------------------------------

   Bell_ST3( );

   ~Bell_ST3( );

   // ------------------------------------------------------------------ support
   // --------------------------------------------------------------------------

   inline
   fpType
   lowerSupport(       fpType  stimulusBegin,
                       fpType  stimulusLength,
                 const fpType* shapeParams ) const;

   inline
   fpType
   upperSupport(       fpType  stimulusBegin,
                       fpType  stimulusLength,
                 const fpType* shapeParams ) const;

   // -------------------------------------------- bell extrema over the support
   // --------------------------------------------------------------------------

   inline
   fpType
   maxval( fpType stimulusLength, const fpType* shapeParams ) const;

   inline
   fpType
   minval( fpType stimulusLength, const fpType* shapeParams ) const;

   // --------------------------------------------------------------- bell value
   // --------------------------------------------------------------------------

   inline
   fpType
   operator()(       fpType  time,
                     fpType  stimulusBegin,
                     fpType  stimulusLength,
               const fpType* shapeParams ) const;

   // -------------------------------------------------- derivatives w.r.t. time
   // --------------------------------------------------------------------------

   inline
   fpType
   D1t(       fpType  time,
              fpType  stimulusBegin,
              fpType  stimulusLength,
        const fpType* shapeParams ) const;
                //
                //  first derivative w. r. t. time: likely to be the common case

   inline
   fpType
   Dt(       fpType  time,
             fpType  stimulusBegin,
             fpType  stimulusLength,
       const fpType* shapeParams,
             size_t  dOrder = 1 ) const;
                //
                //             any general derivative w. r. t. time; dOrder >= 0

   // -------------------------------------- derivatives w.r.t. shape parameters
   // --------------------------------------------------------------------------

   inline
   fpType
   gradient(       fpType  time,
                   fpType  stimulusBegin,
                   fpType  stimulusLength,
             const fpType* shapeParams,
                   size_t  paramIndex ) const;

   inline
   fpType
   hessian(       fpType  time,
                  fpType  stimulusBegin,
                  fpType  stimulusLength,
            const fpType* shapeParams,
                  size_t  paramIndex1,
                  size_t  paramIndex2 ) const;

  private:
   Ramp my_ramp;

   enum params { p_lagon = 0, p_attack, p_decay };
 };

// ---------------------------------------------------------- construct/destruct
// -----------------------------------------------------------------------------

template< class Ramp >
Bell_ST3< Ramp >::Bell_ST3( )
 : BasicBell( 3, "st3",
                   string( "3-parameter bell without dip for single-trial data; ramp: " )
                 + Ramp::name( ) )
 {
  BasicBell::setParamInfo( p_lagon,  "lagon",  "lagon duration",  "lower", 0.0 );
  BasicBell::setParamInfo( p_attack, "attack", "attack duration", "lower", 0.0 );
  BasicBell::setParamInfo( p_decay,  "decay",  "decay duration",  "lower", 0.0 );
 }

template< class Ramp >
Bell_ST3< Ramp >::~Bell_ST3( )
 {
 }

// --------------------------------------------------------------------- support
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_ST3< Ramp >::lowerSupport
 (       fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return stimulusBegin;
 }

template< class Ramp >
fpType
Bell_ST3< Ramp >::upperSupport
 (       fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return
   stimulusBegin + shapeParams[p_lagon] + shapeParams[p_attack] + shapeParams[p_decay];
 }

// ----------------------------------------------- bell extrema over the support
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_ST3< Ramp >::maxval
 (       fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return 1.0;
 }

template< class Ramp >
fpType
Bell_ST3< Ramp >::minval
 (       fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return 0.0;
 }

// ------------------------------------------------------------------ bell value
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_ST3< Ramp >::operator()
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  fpType t0 = ( time - ( stimulusBegin + shapeParams[p_lagon]  ) ) / shapeParams[p_attack];
  fpType t1 = ( time - ( stimulusBegin + shapeParams[p_lagon]
                                       + shapeParams[p_attack] ) ) / shapeParams[p_decay];

  return my_ramp( t0 ) * ( 1.0 - my_ramp( t1 ) );
 }

// ----------------------------------------------------- derivatives w.r.t. time
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_ST3< Ramp >::D1t
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  fpType t0 = ( time - ( stimulusBegin + shapeParams[p_lagon]  ) ) / shapeParams[p_attack];
  fpType t1 = ( time - ( stimulusBegin + shapeParams[p_lagon]
                                       + shapeParams[p_attack] ) ) / shapeParams[p_decay];

  return
     my_ramp.Dt( t0, 1 ) * ( 1.0 - my_ramp( t1 ) ) / shapeParams[p_attack]
   - my_ramp( t0 )       * my_ramp.Dt( t1, 1 )     / shapeParams[p_decay];
 }

template< class Ramp >
fpType
Bell_ST3< Ramp >::Dt
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams,
         size_t  dOrder ) const
 {
  if ( dOrder <= 0 )
   return (*this)( time, stimulusBegin, stimulusLength, shapeParams );
  else if ( dOrder == 1 )
   return (*this).D1t( time, stimulusBegin, stimulusLength, shapeParams );
  else
   {
    fpType t0 = ( time - ( stimulusBegin + shapeParams[p_lagon]  ) ) / shapeParams[p_attack];
    fpType t1 = ( time - ( stimulusBegin + shapeParams[p_lagon]
                                         + shapeParams[p_attack] ) ) / shapeParams[p_decay];

    fpType dt = my_ramp.Dt( t0, dOrder );
  
    fpType f0  = 1.0;
    fpType f1  = 1.0 / pow( shapeParams[p_attack], static_cast<fpType>( dOrder ) );
    fpType nCr = 1.0;
  
    for ( size_t i = 0; i <= dOrder; i++ )
     {
      dt -=   nCr
            * f0 * my_ramp.Dt( t0, i )
            * f1 * my_ramp.Dt( t1, dOrder - i );
  
      f0  /= shapeParams[p_attack];
      f1  *= shapeParams[p_decay];
      nCr *= (fpType) ( dOrder - i ) / (fpType) ( i + 1 );
     }
  
    return dt;
   }
 }

// ----------------------------------------- derivatives w.r.t. shape parameters
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_ST3< Ramp >::gradient
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams,
         size_t  paramIndex ) const
 {
  fpType t0 = ( time - ( stimulusBegin + shapeParams[p_lagon]  ) ) / shapeParams[p_attack];
  fpType t1 = ( time - ( stimulusBegin + shapeParams[p_lagon]
                                       + shapeParams[p_attack] ) ) / shapeParams[p_decay];

  switch ( paramIndex )
   {
    case p_lagon:
     return
      - my_ramp.Dt( t0, 1 ) * ( 1.0 - my_ramp( t1 ) ) / shapeParams[p_attack]
      + my_ramp( t0 )       * my_ramp.Dt( t1, 1 )     / shapeParams[p_decay];

    case p_attack:
     return
      - t0 * my_ramp.Dt( t0, 1 ) * ( 1.0 - my_ramp( t1 ) ) / shapeParams[p_attack]
      +      my_ramp( t0 )       * my_ramp.Dt( t1, 1 )     / shapeParams[p_decay];

    case p_decay:
     return
      t1 * my_ramp( t0 ) * my_ramp.Dt( t1, 1 ) / shapeParams[p_decay];

    default:                                                 // should not occur
     cerr << "Bell_ST3::gradient( ) i " << paramIndex << " should not occur." << endl;
     return 0.0;
   }
 }

template< class Ramp >
fpType
Bell_ST3< Ramp >::hessian
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams,
         size_t  paramIndex1,
         size_t  paramIndex2 ) const
 {
  size_t pSum = paramIndex1 + paramIndex2;

  fpType t0 = ( time - stimulusBegin - shapeParams[p_lagon]  ) / shapeParams[p_attack];
  fpType t1 = ( time - stimulusBegin - shapeParams[p_lagon]
                                     - shapeParams[p_attack] ) / shapeParams[p_decay];

  switch ( pSum )
   {
    case 0:                                             // d( p_lagon, p_lagon )
     return
        my_ramp.Dt( t0, 2 ) * ( 1.0 - my_ramp( t1 ) )
      / ( shapeParams[p_attack] * shapeParams[p_attack] )
      - 2.0 * my_ramp.Dt( t0, 1 ) * my_ramp.Dt( t1, 1 )
      / ( shapeParams[p_attack] * shapeParams[p_decay] )
      - my_ramp( t0 ) * my_ramp.Dt( t1, 2 )
      / ( shapeParams[p_decay] * shapeParams[p_decay] );

    case 1:                                            // d( p_lagon, p_attack )
     {
      fpType d1t0 = my_ramp.Dt( t0, 1 );
      return
         ( 1.0 - my_ramp( t1 ) ) * ( d1t0 + t0 * my_ramp.Dt( t0, 2 ) )
       / ( shapeParams[p_attack] * shapeParams[p_attack] )
       - ( 1.0 + t0 ) * d1t0 * my_ramp.Dt( t1, 1 )
       / ( shapeParams[p_attack] * shapeParams[p_decay] )
       - my_ramp( t0 ) * my_ramp.Dt( t1, 2 )
       / ( shapeParams[p_decay] * shapeParams[p_decay] );
     }

    case 2:
     {
      size_t pDiff = abs( static_cast<int>( paramIndex1 ) - static_cast<int>( paramIndex2 ) );

      switch ( pDiff )
       {
        case 0:                                       // d( p_attack, p_attack )
         {
          fpType d1t0 = my_ramp.Dt( t0, 1 );
          return
             t0 * ( 1.0 - my_ramp( t1 ) ) * ( 2.0 * d1t0 + t0 * my_ramp.Dt( t0, 2 ) )
           / ( shapeParams[p_attack] * shapeParams[p_attack] )
           - 2.0 * t0 * d1t0 * my_ramp.Dt( t1, 1 )
           / ( shapeParams[p_attack] * shapeParams[p_decay] )
           - my_ramp( t0 ) * my_ramp.Dt( t1, 2 )
           / ( shapeParams[p_decay] * shapeParams[p_decay] );
         }

        case 2:                                         // d( p_lagon, p_decay )
         {
          fpType d1t1 = my_ramp.Dt( t1, 1 );
          return
           - my_ramp( t0 ) * ( d1t1 + t1 * my_ramp.Dt( t1, 2 ) )
           / ( shapeParams[p_decay] * shapeParams[p_decay] )
           - t1 * my_ramp.Dt( t0, 1 ) * d1t1
           / ( shapeParams[p_attack] * shapeParams[p_decay] );
         }

        default:                                            // should not occur
         {
          cerr << "Bell_ST3::hessian( ) pSum " << pSum << " pDiff " << pDiff
               << " i1 " << paramIndex1 << " i2 " << paramIndex2 << " should not occur." << endl;
          return 0.0;
         }
       }
     }

    case 3:                                            // d( p_attack, p_decay )
     {
      fpType d1t1 = my_ramp.Dt( t1, 1 );
      return
       - my_ramp( t0 )
       * ( d1t1 + t1 * my_ramp.Dt( t1, 2 ) )
       / ( shapeParams[p_decay] * shapeParams[p_decay] )
       - t0 * t1 * my_ramp.Dt( t0, 1 ) * d1t1
       / ( shapeParams[p_attack] * shapeParams[p_decay] );
     }

    case 4:                                             // d( p_decay, p_decay )
     return
      - t1 * my_ramp( t0 )
      * ( 2.0 * my_ramp.Dt( t1, 1 ) + t1 * my_ramp.Dt( t1, 2 ) )
      / ( shapeParams[p_decay] * shapeParams[p_decay] );

    default:                                                 // should not occur
     {
      cerr << "Bell_ST3::hessian( ) pSum " << pSum
           << " i1 " << paramIndex1 << " i2 " << paramIndex2 << " should not occur." << endl;
      return 0.0;
     }
   }
 }

#endif /* __Bell_ST3_H__ */
