/* --------------------------------------------------------------------------
 *
 * template class Bell, partial specialization
 *   bell function with unit modulator, without dip
 *
 * required headers
 *   see include header file Bell
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __Bell_UN_H__
#define __Bell_UN_H__

template< class Ramp >
class Bell< Ramp, UnitModulator, false > : public BasicBell
 {
  public:
   // ------------------------------------------------------- construct/destruct
   // --------------------------------------------------------------------------

   Bell( );

   ~Bell( );

   // ------------------------------------------------------------------ support
   // --------------------------------------------------------------------------

   inline
   fpType
   lowerSupport(       fpType  stimulusBegin,
                       fpType  stimulusLength,
                 const fpType* shapeParams ) const;
                              //
                              //                                   stimulusBegin

   inline
   fpType
   upperSupport(       fpType  stimulusBegin,
                       fpType  stimulusLength,
                 const fpType* shapeParams ) const;
                              //
                              // stimulusBegin + stimulusLength + shapeParams[p_decay]

   // -------------------------------------------- bell extrema over the support
   // --------------------------------------------------------------------------

   inline
   fpType
   maxval( fpType stimulusLength, const fpType* shapeParams ) const;

   inline
   fpType
   minval( fpType stimulusLength, const fpType* shapeParams ) const;

   // --------------------------------------------------------------- bell value
   // --------------------------------------------------------------------------

   inline
   fpType
   operator()(       fpType  time,
                     fpType  stimulusBegin,
                     fpType  stimulusLength,
               const fpType* shapeParams ) const;

   // -------------------------------------------------- derivatives w.r.t. time
   // --------------------------------------------------------------------------

   inline
   fpType
   D1t(       fpType  time,
              fpType  stimulusBegin,
              fpType  stimulusLength,
        const fpType* shapeParams ) const;
                //
                //  first derivative w. r. t. time: likely to be the common case

   inline
   fpType
   Dt(       fpType  time,
             fpType  stimulusBegin,
             fpType  stimulusLength,
       const fpType* shapeParams,
             size_t  dOrder = 1 ) const;
                //
                //             any general derivative w. r. t. time; dOrder >= 0

   // -------------------------------------- derivatives w.r.t. shape parameters
   // --------------------------------------------------------------------------

   inline
   fpType
   gradient(       fpType  time,
                   fpType  stimulusBegin,
                   fpType  stimulusLength,
             const fpType* shapeParams,
                   size_t  paramIndex ) const;

   inline
   fpType
   hessian(       fpType  time,
                  fpType  stimulusBegin,
                  fpType  stimulusLength,
            const fpType* shapeParams,
                  size_t  paramIndex1,
                  size_t  paramIndex2 ) const;

  private:
   Ramp my_ramp;

   enum params { p_attack = 0, p_decay };
 };

// ---------------------------------------------------------- construct/destruct
// -----------------------------------------------------------------------------

template< class Ramp >
Bell< Ramp, UnitModulator, false >::Bell( )
 : BasicBell( 2, "un2",
                   string( "2-parameter bell; no dip; ramp: " ) + Ramp::name( )
                 + string( ", modulator: " ) + UnitModulator::name( ) )
 {
  BasicBell::setParamInfo( p_attack, "attack", "attack duration", "lower", 0.0 );
  BasicBell::setParamInfo( p_decay,  "decay",  "decay duration",  "lower", 0.0 );
 }

template< class Ramp >
Bell< Ramp, UnitModulator, false >::~Bell( )
 {
 }

// --------------------------------------------------------------------- support
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell< Ramp, UnitModulator, false >::lowerSupport
 (       fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return stimulusBegin;
 }

template< class Ramp >
fpType
Bell< Ramp, UnitModulator, false >::upperSupport
 (       fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return ( stimulusBegin + stimulusLength + shapeParams[p_decay] );
 }

// ----------------------------------------------- bell extrema over the support
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell< Ramp, UnitModulator, false >::maxval
 (       fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  if ( stimulusLength >= shapeParams[p_attack] )
   return 1.0;
  else
   return BasicBell::maxval( stimulusLength, shapeParams );
                               //
                               // ATTN: BasicBell::maxval( ) not implemented yet
 }

template< class Ramp >
fpType
Bell< Ramp, UnitModulator, false >::minval
 (       fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return 0.0;
 }

// ------------------------------------------------------------------ bell value
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell< Ramp, UnitModulator, false >::operator()
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  fpType t0 = ( time - stimulusBegin                  ) / shapeParams[p_attack];
  fpType t1 = ( time - stimulusBegin - stimulusLength ) / shapeParams[p_decay];

  return
   my_ramp( t0 ) * ( 1.0 - my_ramp( t1 ) );
 }

// ----------------------------------------------------- derivatives w.r.t. time
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell< Ramp, UnitModulator, false >::D1t
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  fpType t0 = ( time - stimulusBegin                  ) / shapeParams[p_attack];
  fpType t1 = ( time - stimulusBegin - stimulusLength ) / shapeParams[p_decay];

  return
     my_ramp.Dt( t0, 1 ) * ( 1.0 - my_ramp( t1 ) ) / shapeParams[p_attack]
   - my_ramp( t0 ) * my_ramp.Dt( t1, 1 ) / shapeParams[p_decay];
 }

template< class Ramp >
fpType
Bell< Ramp, UnitModulator, false >::Dt
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams,
         size_t  dOrder ) const
 {
  if ( dOrder <= 0 )
   return (*this)( time, stimulusBegin, stimulusLength, shapeParams );
  else if ( dOrder == 1 )
   return (*this).D1t( time, stimulusBegin, stimulusLength, shapeParams );
  else
   {
    fpType t0 = ( time - stimulusBegin                  ) / shapeParams[p_attack];
    fpType t1 = ( time - stimulusBegin - stimulusLength ) / shapeParams[p_decay];
  
    fpType dt = my_ramp.Dt( t0, dOrder );
  
    fpType f0  = 1.0;
    fpType f1  = 1.0 / pow( shapeParams[p_decay], static_cast<fpType>( dOrder ) );
    fpType nCr = 1.0;
  
    for ( size_t i = 0; i <= dOrder; i++ )
     {
      dt -=   nCr
            * f0 * my_ramp.Dt( t0, i )
            * f1 * my_ramp.Dt( t1, dOrder - i );
  
      f0  /= shapeParams[p_attack];
      f1  *= shapeParams[p_decay];
      nCr *= (fpType) ( dOrder - i ) / (fpType) ( i + 1 );
     }
  
    return dt;
   }
 }

// ----------------------------------------- derivatives w.r.t. shape parameters
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell< Ramp, UnitModulator, false >::gradient
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams,
         size_t  paramIndex ) const
 {
  switch ( paramIndex )
   {
    case p_attack:
     {
      fpType t0 = ( time - stimulusBegin                  ) / shapeParams[p_attack];
      fpType t1 = ( time - stimulusBegin - stimulusLength ) / shapeParams[p_decay];

      return
       - my_ramp.Dt( t0, 1 ) * ( 1.0 - my_ramp( t1 ) ) * ( t0 / shapeParams[p_attack] );
     }

    case p_decay:
     {
      fpType t0 = ( time - stimulusBegin                  ) / shapeParams[p_attack];
      fpType t1 = ( time - stimulusBegin - stimulusLength ) / shapeParams[p_decay];

      return
       my_ramp( t0 ) * my_ramp.Dt( t1, 1 ) * ( t1 / shapeParams[p_decay] );
     }

    default:
     cerr << "Bell_UN::gradient( ) i " << paramIndex << " should not occur." << endl;
     return 0.0;
   }
 }

template< class Ramp >
fpType
Bell< Ramp, UnitModulator, false >::hessian
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams,
         size_t  paramIndex1,
         size_t  paramIndex2 ) const
 {
  if (    ( paramIndex1 < p_attack ) || ( paramIndex1 > p_decay )
       || ( paramIndex2 < p_attack ) || ( paramIndex2 > p_decay ) ) return 0.0;

  fpType t0 = ( time - stimulusBegin                  ) / shapeParams[p_attack];
  fpType t1 = ( time - stimulusBegin - stimulusLength ) / shapeParams[p_decay];

  size_t pSum = paramIndex1 + paramIndex2;

  switch ( pSum )
   {
    case 0:                                           // d( p_attack, p_attack )
     {
      fpType t00 = t0 / shapeParams[p_attack];
      return
         t00 * ( 1.0 - my_ramp( t1 ) )
       * (   2.0 * my_ramp.Dt( t0, 1 ) / shapeParams[p_attack]
           + t00 * my_ramp.Dt( t0, 2 ) );
     }

    case 1:                                            // d( p_attack, p_decay )
     return
      - ( t0 * my_ramp.Dt( t0, 1 ) / shapeParams[p_attack] )
      * ( t1 * my_ramp.Dt( t1, 1 ) / shapeParams[p_decay]  );

    case 2:                                             // d( p_decay, p_decay )
     {
      fpType t11 = t1 / shapeParams[p_decay];
      return
       - t11 * my_ramp( t0 )
       * (   2.0 * my_ramp.Dt( t1, 1 ) / shapeParams[p_decay]
           + t11 * my_ramp.Dt( t1, 2 ) );
     }

    default:
     {
      cerr << "Bell_UN::hessian( ) pSum " << pSum
           << " i1 " << paramIndex1 << " i2 " << paramIndex2 << " should not occur." << endl;
      return 0.0;
     }
   }
 }

#endif /* __Bell_UN_H__ */
