/* --------------------------------------------------------------------------
 *
 * template class Bell_V4
 *   4-parameter vmpfx bell without dip
 *
 * required headers
 *   see include header file Bell
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __Bell_V4_H__
#define __Bell_V4_H__

template< class Ramp = SplineRampC2 >
class Bell_V4 : public BasicBell
 {
  public:
   // ------------------------------------------------------- construct/destruct
   // --------------------------------------------------------------------------

   Bell_V4( );

   ~Bell_V4( );

   // ------------------------------------------------------------------ support
   // --------------------------------------------------------------------------

   inline
   fpType
   lowerSupport(       fpType  stimulusBegin,
                       fpType  stimulusLength,
                 const fpType* shapeParams ) const;

   inline
   fpType
   upperSupport(       fpType  stimulusBegin,
                       fpType  stimulusLength,
                 const fpType* shapeParams ) const;

   // -------------------------------------------- bell extrema over the support
   // --------------------------------------------------------------------------

   inline
   fpType
   maxval( fpType stimulusLength, const fpType* shapeParams ) const;

   inline
   fpType
   minval( fpType stimulusLength, const fpType* shapeParams ) const;

   // --------------------------------------------------------------- bell value
   // --------------------------------------------------------------------------

   inline
   fpType
   operator()(       fpType  time,
                     fpType  stimulusBegin,
                     fpType  stimulusLength,
               const fpType* shapeParams ) const;

   // -------------------------------------------------- derivatives w.r.t. time
   // --------------------------------------------------------------------------

   inline
   fpType
   D1t(       fpType  time,
              fpType  stimulusBegin,
              fpType  stimulusLength,
        const fpType* shapeParams ) const;
                //
                //  first derivative w. r. t. time: likely to be the common case

   inline
   fpType
   Dt(       fpType  time,
             fpType  stimulusBegin,
             fpType  stimulusLength,
       const fpType* shapeParams,
             size_t  dOrder = 1 ) const;
                //
                //             any general derivative w. r. t. time; dOrder >= 0

   // -------------------------------------- derivatives w.r.t. shape parameters
   // --------------------------------------------------------------------------

   inline
   fpType
   gradient(       fpType  time,
                   fpType  stimulusBegin,
                   fpType  stimulusLength,
             const fpType* shapeParams,
                   size_t  paramIndex ) const;

   inline
   fpType
   hessian(       fpType  time,
                  fpType  stimulusBegin,
                  fpType  stimulusLength,
            const fpType* shapeParams,
                  size_t  paramIndex1,
                  size_t  paramIndex2 ) const;

  private:
   Ramp my_ramp;

   enum params { p_lagon = 0, p_attack, p_lagoff, p_decay };
 };

// ---------------------------------------------------------- construct/destruct
// -----------------------------------------------------------------------------

template< class Ramp >
Bell_V4< Ramp >::Bell_V4( )
 : BasicBell( 4, "v4",
                   string( "4-parameter vmpfx bell without dip; ramp: " )
                 + Ramp::name( ) )
 {
  BasicBell::setParamInfo( p_lagon,  "lagon",  "lagon duration",  "lower", 0.0 );
  BasicBell::setParamInfo( p_attack, "attack", "attack duration", "lower", 0.0 );
  BasicBell::setParamInfo( p_lagoff, "lagoff", "lagoff duration", "lower", 0.0 );
  BasicBell::setParamInfo( p_decay,  "decay",  "decay duration",  "lower", 0.0 );
 }

template< class Ramp >
Bell_V4< Ramp >::~Bell_V4( )
 {
 }

// --------------------------------------------------------------------- support
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_V4< Ramp >::lowerSupport
 (       fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return stimulusBegin;
 }

template< class Ramp >
fpType
Bell_V4< Ramp >::upperSupport
 (       fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return
   stimulusBegin + stimulusLength + shapeParams[p_lagoff] + shapeParams[p_decay];
 }

// ----------------------------------------------- bell extrema over the support
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_V4< Ramp >::maxval
 (       fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  if ( stimulusLength >= ( shapeParams[p_lagon] + shapeParams[p_attack] ) )
   return 1.0;
  else
   return BasicBell::maxval( stimulusLength, shapeParams );
                               //
                               // ATTN: BasicBell::maxval( ) not implemented yet
 }

template< class Ramp >
fpType
Bell_V4< Ramp >::minval
 (       fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  return 0.0;
 }

// ------------------------------------------------------------------ bell value
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_V4< Ramp >::operator()
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  fpType t0 = ( time - ( stimulusBegin + shapeParams[p_lagon]                   ) ) / shapeParams[p_attack];
  fpType t1 = ( time - ( stimulusBegin + stimulusLength + shapeParams[p_lagoff] ) ) / shapeParams[p_decay];

  return my_ramp( t0 ) * ( 1.0 - my_ramp( t1 ) );
 }

// ----------------------------------------------------- derivatives w.r.t. time
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_V4< Ramp >::D1t
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams ) const
 {
  fpType t0 = ( time - ( stimulusBegin + shapeParams[p_lagon]                   ) ) / shapeParams[p_attack];
  fpType t1 = ( time - ( stimulusBegin + stimulusLength + shapeParams[p_lagoff] ) ) / shapeParams[p_decay];

  return
     my_ramp.Dt( t0, 1 ) * ( 1.0 - my_ramp( t1 ) ) / shapeParams[p_attack]
   - my_ramp( t0 ) * my_ramp.Dt( t1, 1 ) / shapeParams[p_decay];
 }

template< class Ramp >
fpType
Bell_V4< Ramp >::Dt
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams,
         size_t  dOrder ) const
 {
  if ( dOrder <= 0 )
   return (*this)( time, stimulusBegin, stimulusLength, shapeParams );
  else if ( dOrder == 1 )
   return (*this).D1t( time, stimulusBegin, stimulusLength, shapeParams );
  else
   {
    fpType t0 = ( time - ( stimulusBegin + shapeParams[p_lagon]                   ) ) / shapeParams[p_attack];
    fpType t1 = ( time - ( stimulusBegin + stimulusLength + shapeParams[p_lagoff] ) ) / shapeParams[p_decay];
  
    fpType dt = my_ramp.Dt( t0, dOrder );
  
    fpType f0  = 1.0;
    fpType f1  = 1.0 / pow( shapeParams[p_attack], static_cast<fpType>( dOrder ) );
    fpType nCr = 1.0;
  
    for ( size_t i = 0; i <= dOrder; i++ )
     {
      dt -=   nCr
            * f0 * my_ramp.Dt( t0, i )
            * f1 * my_ramp.Dt( t1, dOrder - i );
  
      f0  /= shapeParams[p_attack];
      f1  *= shapeParams[p_decay];
      nCr *= (fpType) ( dOrder - i ) / (fpType) ( i + 1 );
     }
  
    return dt;
   }
 }

// ----------------------------------------- derivatives w.r.t. shape parameters
// -----------------------------------------------------------------------------

template< class Ramp >
fpType
Bell_V4< Ramp >::gradient
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams,
         size_t  paramIndex ) const
 {
  fpType t0 = ( time - ( stimulusBegin + shapeParams[p_lagon]                   ) ) / shapeParams[p_attack];
  fpType t1 = ( time - ( stimulusBegin + stimulusLength + shapeParams[p_lagoff] ) ) / shapeParams[p_decay];
  
  switch ( paramIndex )
   {
    case p_lagon:
     return
      - my_ramp.Dt( t0, 1 ) * ( 1.0 - my_ramp( t1 ) ) / shapeParams[p_attack];

    case p_attack:
     return
      - t0 * my_ramp.Dt( t0, 1 ) * ( 1.0 - my_ramp( t1 ) ) / shapeParams[p_attack];

    case p_lagoff:
     return
      my_ramp( t0 ) * my_ramp.Dt( t1, 1 ) / shapeParams[p_decay];

    case p_decay:
     return
      t1 * my_ramp( t0 ) * my_ramp.Dt( t1, 1 ) / shapeParams[p_decay];

    default:
     cerr << "Bell_V4::gradient( ) i " << paramIndex << " should not occur." << endl;
     return 0.0;
   }
 }

template< class Ramp >
fpType
Bell_V4< Ramp >::hessian
 (       fpType  time,
         fpType  stimulusBegin,
         fpType  stimulusLength,
   const fpType* shapeParams,
         size_t  paramIndex1,
         size_t  paramIndex2 ) const
 {
  size_t pSum  = paramIndex1 + paramIndex2;
  size_t pDiff = abs( static_cast<int>( paramIndex1 ) - static_cast<int>( paramIndex2 ) );

  fpType t0 = ( time - ( stimulusBegin + shapeParams[p_lagon]                   ) ) / shapeParams[p_attack];
  fpType t1 = ( time - ( stimulusBegin + stimulusLength + shapeParams[p_lagoff] ) ) / shapeParams[p_decay];

  switch ( pSum )
   {
    case 0:                                             // d( p_lagon, p_lagon )
     return
        my_ramp.Dt( t0, 2 ) * ( 1.0 - my_ramp( t1 ) )
      / ( shapeParams[p_attack] * shapeParams[p_attack] );

    case 1:                                            // d( p_lagon, p_attack )
     return
        ( 1.0 - my_ramp( t1 ) ) * ( my_ramp.Dt( t0, 1 ) + t0 * my_ramp.Dt( t0, 2 ) )
      / ( shapeParams[p_attack] * shapeParams[p_attack] );

    case 2:
     switch ( pDiff )
      {
       case 0:                                        // d( p_attack, p_attack )
        return
           t0 * ( 1.0 - my_ramp( t1 ) )
         * ( 2.0 * my_ramp.Dt( t0, 1 ) + t0 * my_ramp.Dt( t0, 2 ) )
         / ( shapeParams[p_attack] * shapeParams[p_attack] );

       case 2:                                         // d( p_lagon, p_lagoff )
        return
         - my_ramp.Dt( t0, 1 ) * my_ramp.Dt( t1, 1 )
         / ( shapeParams[p_attack] * shapeParams[p_decay] );

       default:
        {
         cerr << "Bell_V4::hessian( ) pSum " << pSum << " pDiff " << pDiff
              << " i1 " << paramIndex1 << " i2 " << paramIndex2 << " should not occur." << endl;
         return 0.0;
        }
      }

    case 3:
     switch ( pDiff )
      {
       case 1:                                        // d( p_attack, p_lagoff )
        return
         - t0 * my_ramp.Dt( t0, 1 ) * my_ramp.Dt( t1, 1 )
         / ( shapeParams[p_attack] * shapeParams[p_decay] );

       case 3:                                          // d( p_lagon, p_decay )
        return
         - t1 * my_ramp.Dt( t0, 1 ) * my_ramp.Dt( t1, 1 )
         / ( shapeParams[p_attack] * shapeParams[p_decay] );

       default:
        {
         cerr << "Bell_V4::hessian( ) pSum " << pSum << " pDiff " << pDiff
              << " i1 " << paramIndex1 << " i2 " << paramIndex2 << " should not occur." << endl;
         return 0.0;
        }
      }

    case 4:
     switch ( pDiff )
      {
       case 0:                                        // d( p_lagoff, p_lagoff )
        return
         - my_ramp( t0 ) * my_ramp.Dt( t1, 2 )
         / ( shapeParams[p_decay] * shapeParams[p_decay] );

       case 2:                                         // d( p_attack, p_decay )
        return
         - t0 * t1 * my_ramp.Dt( t0, 1 ) * my_ramp.Dt( t1, 1 )
         / ( shapeParams[p_attack] * shapeParams[p_decay] );

       default:
        {
         cerr << "Bell_V4::hessian( ) pSum " << pSum << " pDiff " << pDiff
              << " i1 " << paramIndex1 << " i2 " << paramIndex2 << " should not occur." << endl;
         return 0.0;
        }
      }

    case 5:                                            // d( p_lagoff, p_decay )
     return
      - t1 * my_ramp.Dt( t0, 1 ) * my_ramp.Dt( t1, 1 )
      / ( shapeParams[p_attack] * shapeParams[p_decay] );

    case 6:                                             // d( p_decay, p_decay )
     return
      - t1 * my_ramp( t0 )
      * ( 2.0 * my_ramp.Dt( t1, 1 ) + t1 * my_ramp.Dt( t1, 2 ) )
      / ( shapeParams[p_decay] * shapeParams[p_decay] );

    default:
     {
      cerr << "Bell_V4::hessian( ) pSum " << pSum
           << " i1 " << paramIndex1 << " i2 " << paramIndex2 << " should not occur." << endl;
      return 0.0;
     }
   }
 }

#endif /* __Bell_V4_H__ */
