/* --------------------------------------------------------------------------
 *
 * class ExpModulator
 *   1-parameter exponential-based modulator to
 *   vary rise characteristics of symmetric up ramps
 *
 * required headers
 *   see include header file ExpModulator
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __ExpModulator_H__
#define __ExpModulator_H__

cpp_using_namespace( BasicTypes );

class ExpModulator
 {
  public:
                          //
                          //    basic 1-parameter modulator m( t, a ) over [0,1]
                          //
                          // m( t; a ) = ( 1 - exp( -a t ) ) / ( 1 - exp( -a ) )
                          //                                 -infty < a < +infty
                          //                                         0 <= t <= 1
                          //

   static
   inline
   string
   name( );

   static
   inline
   string
   description( );

   inline
   fpType
   operator()( fpType t, fpType a ) const;

   inline
   fpType
   Dt( fpType t, fpType a, size_t dOrder = 1 ) const;

   inline
   fpType
   Da( fpType t, fpType a, size_t dOrder = 1 ) const;

   inline
   fpType
   Dta( fpType t, fpType a, size_t dOrderT = 1, size_t dOrderA = 1 ) const;
 };

string
ExpModulator::name( )
 {
  return "ExpModulator";
 }

string
ExpModulator::description( )
 {
  return "( 1 - exp(-at) ) / ( 1 - exp(-a) )";
 }

fpType
ExpModulator::operator()( fpType t, fpType a ) const
 {
  return
   ( a == 0.0 ) ? t : ( 1.0 - exp( - a * t ) ) / ( 1.0 - exp( - a ) );
 }

fpType
ExpModulator::Dt( fpType t, fpType a, size_t dOrder ) const
 {
  if ( dOrder == 0 )
   return (*this)( t, a );
  else if ( a == 0.0 )
   return ( ( dOrder == 1 ) ? 1.0 : 0.0 );
  else
   {
    a *= -1.0;
    return ( - pow( static_cast<double>( a ), static_cast<int>( dOrder ) ) * exp( a * t ) / ( 1.0 - exp( a ) ) );
   }
 }

fpType
ExpModulator::Da( fpType t, fpType a, size_t dOrder ) const
 {
  if ( dOrder == 0 )
   return (*this)( t, a );
  else if ( ( a == 0.0 ) || ( t == 0.0 ) || ( t == 1.0 ) )
   return 0.0;
  else
   {
    fpType Ea      = exp( -a );
    fpType omEaInv = 1.0 / ( 1.0 - Ea );
    fpType Eat     = exp( -a * t );

    fpVector da( dOrder + 1 );
    da[0] = (*this)( t, a );

    fpType dSum, sgn, nCi, asgn = 1.0, tk = t;
    for ( size_t k = 1; k <= dOrder; k++ )
     {
      dSum = 0.0; sgn = -1.0; nCi = k;
      for ( size_t i = 1; i <= k; i++ )
       {
        dSum += sgn * nCi * da[k - i];

        sgn *= -1.0;
        nCi *= ( k - i );
        nCi /= ( i + 1 );
       }

      da[k] = omEaInv * ( Ea * dSum + Eat * tk * asgn );

      asgn *= -1.0;
      tk   *= t;
     }

    return da[dOrder];
   }
 }

fpType
ExpModulator::Dta( fpType t, fpType a, size_t dOrderT, size_t dOrderA ) const
 {
  if ( dOrderA == 0 )
   return Dt( t, a, dOrderT );
  else if ( dOrderT == 0 )
   return Da( t, a, dOrderA );
  else
   {
    fpType Ea         = exp( -a );
    fpType omEaInv    = 1.0 / ( 1.0 - Ea );
    fpType Eat        = exp( -a * t );

    fpType an         = pow( static_cast<double>( a ), static_cast<int>( dOrderT ) );
    fpType atInv      = 1.0 / ( a * t );
    fpType tsgn       = ( dOrderT % 2 == 0 ) ? 1.0: -1.0;

    fpVector da( dOrderA + 1 );
    da[0] = Dt( t, a, dOrderT );

    fpType dSum, rSum, sgn, nCi, Pni, asgn = -1.0, tk = t;
    for ( size_t k = 1; k <= dOrderA; k++ )
     {
      dSum = 0.0; sgn = -1.0; nCi = k;
      for ( size_t i = 1; i <= k; i++ )
       {
        dSum += sgn * nCi * da[k - i];

        sgn *= -1.0;
        nCi *= ( k - i );
        nCi /= ( i + 1 );
       }

      fpVector coeff( 0.0, min( dOrderT, k ) + 1 );
      sgn = 1.0; nCi = 1.0; Pni = 1.0;
      for ( size_t i = 0; i < coeff.size( ); i++ )
       {
        coeff[i] = sgn * nCi * Pni;

        sgn *= -1.0;
        nCi *= ( k - i );
        nCi /= ( i + 1 );
        Pni *= ( dOrderT - i );
       }

      rSum = 0.0; 
      for ( size_t j = 0, i = coeff.size( ) - 1; j < coeff.size( ); j++, i-- )
       rSum = rSum * atInv + coeff[i];

      da[k] = omEaInv * ( Ea * dSum - Eat * an * tk * tsgn * asgn * rSum );

      asgn *= -1.0;
      tk   *= t;
     }

    return da[dOrderA];
   }
 }

#endif /* __ExpModulator_H__ */
