/* ----------------------------------------------------------------------------
 *
 * class LLnormal<true,1,0>
 *   template specialization: log normal/ARMA(1,0) likelihood/with activation
 *
 * required headers
 *   see include header file LLnormal
 *
 * notes
 *   0. follows convention for out-of-range parameter offsets; see BasicObjFn.h
 *   1. the single AR parameter has bounds |phi| <= 1 corresponding to the usual
 *      causality and stationarity conditions; this is to be imposed externally.
 *   2. This particular implementation conditions the likelihood on the first
 *      data point.
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __LLn_a_1_0_H__
#define __LLn_a_1_0_H__

cpp_using_namespace( BasicTypes );

class LLnormal<true,1,0> : public BasicObjFn
 {
  public:
   // --------------------- construct/reinitialize/incremental initialize/assign
   // --------------------------------------------------------------------------

   LLnormal( );

   LLnormal( const LLnormal& another );

   LLnormal&
   operator=( const LLnormal& rhs );

   LLnormal( const fpVector*                 signal,
             const dbProfile*                db,
             const acProfile*                ac,
             const ParameterManager<fpType>& pm,
                   int                       paramBufSize,
                   fpType                    sign = 1.0,
             const string&                   name = "LL normal/activation/arma(1,0)" );

   void
   setSignal( const fpVector* signal );

   void
   setDBprofile( const dbProfile* db );

   void
   setACprofile( const acProfile* ac );

   void
   setParamOffset( const ParameterManager<fpType>& pm );

   void
   reinitialize( const fpVector*                 signal,
                 const dbProfile*                db,
                 const acProfile*                ac,
                 const ParameterManager<fpType>& pm,
                       int                       paramBufSize,
                       fpType                    sign = 1.0,
                 const string&                   name = "LL normal/activation/arma(1,0)" );

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   fpType
   additiveConst( ) const;
                                //
                                // additive constants arising from normalization
                                //   that are not included in value computation.
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   // -------------------------------------------------- likelihood computations
   // ----------------- without initialization of output arguments ll, Dll, DDll
   // --------------------------------------------------------------------------

            // the implementations of gradient and hessian makes the assumptions
            //         (1) drift+baseline is linear in drift+baseline parameters
            //         (and use functions in dbProfile based on this assumption)
            //             (2) activation is linear in responsiveness parameters

   inline
   void
   value_ni( const fpType* const paramBuf, fpType* ll ) const;

   inline
   void
   gradient_ni( const fpType* const paramBuf, fpType* Dll ) const;

   inline
   void
   hessian_ni( const fpType* const parameters, fpMatrix& DDlp ) const;

  private:
   const fpVector*  my_signal;
   const dbProfile* my_db;
   const acProfile* my_ac;
         int        my_dbOffset, my_noiseOffset, my_arOffset,
                    my_respOffset, my_shapeOffset;
         int        my_dbLength, my_noiseLength,
                    my_respLength, my_shapeLength, my_arLength;
 };

// ------------------------ construct/reinitialize/incremental initialize/assign
// -----------------------------------------------------------------------------

LLnormal<true,1,0>::LLnormal( )
 : BasicObjFn( ),
   my_signal( 0 ),
   my_db( 0 ),
   my_ac( 0 ),
   my_dbOffset( 0 ),
   my_noiseOffset( 0 ),
   my_arOffset( 0 ),
   my_respOffset( 0 ),
   my_shapeOffset( 0 ),
   my_dbLength( 0 ),
   my_noiseLength( 0 ),
   my_respLength( 0 ),
   my_shapeLength( 0 ),
   my_arLength( 1 )
 {
  BasicObjFn::setName( "LL normal/activation/arma(1,0)" );
 }

LLnormal<true,1,0>::LLnormal( const LLnormal<true,1,0>& another )
 : BasicObjFn( another.paramBufSize( ), another.sign( ), another.name( ) ),
   my_signal( another.my_signal ),
   my_db( another.my_db ),
   my_ac( another.my_ac ),
   my_dbOffset( another.my_dbOffset ),
   my_noiseOffset( another.my_noiseOffset ),
   my_arOffset( another.my_arOffset ),
   my_respOffset( another.my_respOffset ),
   my_shapeOffset( another.my_shapeOffset ),
   my_dbLength( another.my_dbLength ),
   my_noiseLength( another.my_noiseLength ),
   my_respLength( another.my_respLength ),
   my_shapeLength( another.my_shapeLength )
 {
 }

LLnormal<true,1,0>&
LLnormal<true,1,0>::operator=( const LLnormal<true,1,0>& rhs )
 {
  if ( this == &rhs ) return *this;

  BasicObjFn::setParamBufSize( rhs.paramBufSize( ) );
  BasicObjFn::setSign( rhs.sign( ) );
  BasicObjFn::setName( rhs.name( ) );

  my_signal = rhs.my_signal;

  my_db = rhs.my_db;
  my_ac = rhs.my_ac;

  my_dbOffset    = rhs.my_dbOffset;
  my_noiseOffset = rhs.my_noiseOffset;
  my_arOffset    = rhs.my_arOffset;
  my_respOffset  = rhs.my_respOffset;
  my_shapeOffset = rhs.my_shapeOffset;

  my_dbLength    = rhs.my_dbLength;
  my_noiseLength = rhs.my_noiseLength;
  my_respLength  = rhs.my_respLength;
  my_shapeLength = rhs.my_shapeLength;

  return *this;
 }

LLnormal<true,1,0>::LLnormal( const fpVector*                 signal,
                              const dbProfile*                db,
                              const acProfile*                ac,
                              const ParameterManager<fpType>& pm,
                                    int                       paramBufSize,
                                    fpType                    sign,
                              const string&                   name )
 : BasicObjFn( paramBufSize, sign, name ),
   my_signal( signal ),
   my_db( db ),
   my_ac( ac ),
   my_dbOffset( pm.blockOffset( ParamBlkNames::drift ) ),
   my_noiseOffset( pm.blockOffset( ParamBlkNames::noise ) ),
   my_arOffset( pm.blockOffset( ParamBlkNames::ar ) ),
   my_respOffset( pm.blockOffset( ParamBlkNames::response ) ),
   my_shapeOffset( pm.blockOffset( ParamBlkNames::shape ) ),
   my_dbLength( pm.blockLength( ParamBlkNames::drift ) ),
   my_noiseLength( pm.blockLength( ParamBlkNames::noise ) ),
   my_respLength( pm.blockLength( ParamBlkNames::response ) ),
   my_shapeLength( pm.blockLength( ParamBlkNames::shape ) ),
   my_arLength( 1 )
 {
 }

void
LLnormal<true,1,0>::setSignal( const fpVector* signal )
 {
  my_signal = signal;
 }

void
LLnormal<true,1,0>::setDBprofile( const dbProfile* db )
 {
  my_db = db;
 }

void
LLnormal<true,1,0>::setACprofile( const acProfile* ac )
 {
  my_ac = ac;
 }

void
LLnormal<true,1,0>::setParamOffset( const ParameterManager<fpType>& pm )
 {
  my_dbOffset    = pm.blockOffset( ParamBlkNames::drift );
  my_noiseOffset = pm.blockOffset( ParamBlkNames::noise );
  my_arOffset    = pm.blockOffset( ParamBlkNames::ar );
  my_respOffset  = pm.blockOffset( ParamBlkNames::response );
  my_shapeOffset = pm.blockOffset( ParamBlkNames::shape );

  my_dbLength    = pm.blockLength( ParamBlkNames::drift );
  my_noiseLength = pm.blockLength( ParamBlkNames::noise );
  my_respLength  = pm.blockLength( ParamBlkNames::response );
  my_shapeLength = pm.blockLength( ParamBlkNames::shape );
 }

void
LLnormal<true,1,0>::reinitialize( const fpVector*                 signal,
                                  const dbProfile*                db,
                                  const acProfile*                ac,
                                  const ParameterManager<fpType>& pm,
                                        int                       paramBufSize,
                                        fpType                    sign,
                                  const string&                   name )
 {
  BasicObjFn::reinitialize( paramBufSize, sign, name );

  setSignal( signal );
  setDBprofile( db );
  setACprofile( ac );
  setParamOffset( pm );
 }

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

fpType
LLnormal<true,1,0>::additiveConst( ) const
 {
  return - 0.5 * ( my_signal->size( ) - 1 ) * log( 2.0 * M_PI );
 }

// ----------------------------------------------------- likelihood computations
// -------------------- without initialization of output arguments ll, Dll, DDll
// -----------------------------------------------------------------------------

void
LLnormal<true,1,0>::value_ni( const fpType* const paramBuf, fpType* ll ) const
 {
  const fpType* dbParams = paramBuf + my_dbOffset;

  fpType noiseParam = paramBuf[my_noiseOffset];

  fpVector residual( (*my_signal) );
  residual -= my_db->profile( dbParams );
  residual -= my_ac->profile( dbParams, paramBuf + my_respOffset,
                                        paramBuf + my_shapeOffset );

  ARutils::getARvec( 1, paramBuf + my_arOffset, residual );

  (*ll) += 0.5 * ( ( my_signal->size( ) - 1 ) * log( noiseParam ) - noiseParam * BlasDrivers::dot( residual, residual ) );
 }

void
LLnormal<true,1,0>::gradient_ni( const fpType* const paramBuf, fpType* Dll ) const
 {
  const fpType* dbParams    = paramBuf + my_dbOffset;
  const fpType* respParams  = paramBuf + my_respOffset;
  const fpType* shapeParams = paramBuf + my_shapeOffset;
  const fpType* arParam     = paramBuf + my_arOffset;

  fpType noiseParam = paramBuf[my_noiseOffset];

  int signalSize = my_signal->size( ) - 1;

  fpVector residual( (*my_signal) );
  residual -= my_db->profile( dbParams );
  residual -= my_ac->profile( dbParams, respParams, shapeParams );

  fpVector arResid( residual );
  ARutils::getARvec( 1, arParam, arResid );

                                                  //                  D( noise )
                                                  //
  if ( ( my_noiseOffset >= 0 ) && ( my_noiseOffset < paramBufSize( ) ) )
   Dll[my_noiseOffset] += 0.5 * ( signalSize / noiseParam - BlasDrivers::dot( arResid, arResid ) );

                                                  //                     D( ar )
                                                  //
  if ( ( my_arOffset >= 0 ) && ( my_arOffset < paramBufSize( ) ) )
   Dll[my_arOffset] += noiseParam * BlasDrivers::dot( signalSize, arResid( 1 ), residual( 0 ), 1, 1 );

  int ofsMax;

                                                  //         D( drift+baseline )
                                                  //
  ofsMax = min( my_dbOffset + my_dbLength, paramBufSize( ) );
  for ( int iOffset = max( 0, my_dbOffset ),
            i       = iOffset - my_dbOffset;
            iOffset < ofsMax; iOffset++, i++ )
   Dll[iOffset] += noiseParam * ( BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_db->Dprofile( dbParams, i ) ) ) + BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_ac->DprofileD( dbParams, respParams, shapeParams, i ) ) ) );

                                                  //         D( responsiveness )
                                                  //
  ofsMax = min( my_respOffset + my_respLength, paramBufSize( ) );
  for ( int iOffset = max( 0, my_respOffset ),
            i       = iOffset - my_respOffset;
            iOffset < ofsMax; iOffset++, i++ )
   Dll[iOffset] += noiseParam * BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_ac->DprofileR( dbParams, respParams, shapeParams, i ) ) );

                                                  //                  D( shape )
                                                  //
  ofsMax = min( my_shapeOffset + my_shapeLength, paramBufSize( ) );
  for ( int iOffset = max( 0, my_shapeOffset ),
            i       = iOffset - my_shapeOffset;
            iOffset < ofsMax; iOffset++, i++ )
   Dll[iOffset] += noiseParam * BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_ac->DprofileS( dbParams, respParams, shapeParams, i ) ) );
 }

void
LLnormal<true,1,0>::hessian_ni( const fpType* const paramBuf, fpMatrix& DDll ) const
 {
  const fpType* dbParams    = paramBuf + my_dbOffset;
  const fpType* respParams  = paramBuf + my_respOffset;
  const fpType* shapeParams = paramBuf + my_shapeOffset;
  const fpType* arParam     = paramBuf + my_arOffset;

  fpType noiseParam = paramBuf[my_noiseOffset];

  int signalSize = my_signal->size( ) - 1;

  fpVector residual( (*my_signal) );
  residual -= my_db->profile( dbParams );
  residual -= my_ac->profile( dbParams, respParams, shapeParams );

  fpVector arResid( residual );
  ARutils::getARvec( 1, arParam, arResid );

  int ofsMinDB    = max( 0, my_dbOffset    );
  int ofsMinResp  = max( 0, my_respOffset  );
  int ofsMinShape = max( 0, my_shapeOffset );

  int ofsMaxDB    = min( my_dbOffset    + my_dbLength,    paramBufSize( ) );
  int ofsMaxResp  = min( my_respOffset  + my_respLength,  paramBufSize( ) );
  int ofsMaxShape = min( my_shapeOffset + my_shapeLength, paramBufSize( ) );

  bool doNoise = ( my_noiseOffset >= 0 ) && ( my_noiseOffset < paramBufSize( ) );
  bool doAR    = ( my_arOffset    >= 0 ) && ( my_arOffset    < paramBufSize( ) );

  fpType ddValue;
  fpVector Di, Dj, arDi;

                                        //                    D2( noise, noise )
                                        //
  if ( doNoise )
   {
    DDll( my_noiseOffset, my_noiseOffset )
     -= 0.5 * signalSize / ( noiseParam * noiseParam );

                                                  //             D2( noise, ar )
                                                  //
    if ( doAR )
     {
      ddValue = BlasDrivers::dot( signalSize, arResid( 1 ), residual( 0 ), 1, 1 );

      DDll( my_arOffset, my_noiseOffset ) += ddValue;
      DDll( my_noiseOffset, my_arOffset ) += ddValue;
     }
   }

                                                  //                D2( ar, ar )
                                                  //
  if ( doAR ) DDll( my_arOffset, my_arOffset ) -= noiseParam * BlasDrivers::dot( signalSize, residual( ), residual( ), 1, 1 );

  for ( int iOffset = ofsMinDB,
            i       = iOffset - my_dbOffset;
            iOffset < ofsMaxDB; iOffset++, i++ )
   {
    my_ac->getDprofileD( dbParams, respParams, shapeParams, i, Di );
    Di += my_db->Dprofile( dbParams, i );
    ARutils::getARvec( 1, arParam, Di, arDi );

                                        //           D2( noise, drift+baseline )
                                        //
    if ( doNoise )
     {
      ddValue = BlasDrivers::dot( arResid, arDi );

      DDll( iOffset, my_noiseOffset ) += ddValue;
      DDll( my_noiseOffset, iOffset ) += ddValue;
     }

                                        //              D2( ar, drift+baseline )
                                        //
    if ( doAR )
     {
      ddValue =   noiseParam * ( BlasDrivers::dot( signalSize, arResid( 1 ), Di( 0 ), 1, 1 ) + BlasDrivers::dot( signalSize, arDi( 1 ), residual( 0 ), 1, 1 ) );

      DDll( iOffset, my_arOffset ) -= ddValue;
      DDll( my_arOffset, iOffset ) -= ddValue;
     }

                                        //  D2( drift+baseline, drift+baseline )
                                        //                              diagonal
                                        //
    DDll( iOffset, iOffset ) -= noiseParam * BlasDrivers::dot( arDi,  arDi );

                                        //  D2( drift+baseline, drift+baseline )
                                        //                          off-diagonal
                                        //
    for ( int jOffset = ofsMinDB,
              j       = jOffset - my_dbOffset;
              jOffset < iOffset; jOffset++, j++ )
     {
      my_ac->getDprofileD( dbParams, respParams, shapeParams, j, Dj );
      Dj += my_db->Dprofile( dbParams, j );

      ddValue = noiseParam * BlasDrivers::dot( arDi, ARutils::ARvec( 1, arParam, Dj ) );

      DDll( iOffset, jOffset ) -= ddValue;
      DDll( jOffset, iOffset ) -= ddValue;
     }

                                        //  D2( drift+baseline, responsiveness )
                                        //
    for ( int jOffset = ofsMinResp,
              j       = jOffset - my_respOffset;
              jOffset < ofsMaxResp; jOffset++, j++ )
     {
      ddValue = noiseParam * ( BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_ac->DDprofileDR( dbParams, respParams, shapeParams, i, j ) ) ) - BlasDrivers::dot( arDi, ARutils::ARvec( 1, arParam, my_ac->DprofileR( dbParams, respParams, shapeParams, j ) ) ) );

      DDll( iOffset, jOffset ) += ddValue;
      DDll( jOffset, iOffset ) += ddValue;
     }

                                        //           D2( drift+baseline, shape )
                                        //
    for ( int jOffset = ofsMinShape,
              j       = jOffset - my_shapeOffset;
              jOffset < ofsMaxShape; jOffset++, j++ )
     {
      ddValue = noiseParam * ( BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_ac->DDprofileDS( dbParams, respParams, shapeParams, i, j ) ) ) - BlasDrivers::dot( arDi, ARutils::ARvec( 1, arParam, my_ac->DprofileS( dbParams, respParams, shapeParams, j ) ) ) );

      DDll( iOffset, jOffset ) += ddValue;
      DDll( jOffset, iOffset ) += ddValue;
     }
   }

  for ( int iOffset = ofsMinResp,
            i       = iOffset - my_respOffset;
            iOffset < ofsMaxResp; iOffset++, i++ )
   {
    my_ac->getDprofileR( dbParams, respParams, shapeParams, i, Di );
    ARutils::getARvec( 1, arParam, Di, arDi );

                                        //           D2( noise, responsiveness )
                                        //
    if ( doNoise )
     {
      ddValue = BlasDrivers::dot( arResid, arDi );

      DDll( iOffset, my_noiseOffset ) += ddValue;
      DDll( my_noiseOffset, iOffset ) += ddValue;
     }

                                        //              D2( ar, responsiveness )
                                        //
    if ( doAR )
     {
      ddValue = noiseParam * ( BlasDrivers::dot( signalSize, arResid( 1 ), Di( 0 ), 1, 1 ) + BlasDrivers::dot( signalSize, arDi( 1 ), residual( 0 ), 1, 1 ) );

      DDll( iOffset, my_arOffset ) -= ddValue;
      DDll( my_arOffset, iOffset ) -= ddValue;
     }

                                        //  D2( responsiveness, responsiveness )
                                        //                              diagonal
                                        //
    DDll( iOffset, iOffset ) -= noiseParam * BlasDrivers::dot( arDi,  arDi );

                                        //  D2( responsiveness, responsiveness )
                                        //                          off-diagonal
                                        //
    for ( int jOffset = ofsMinResp,
              j       = jOffset - my_respOffset;
              jOffset < iOffset; jOffset++, j++ )
     {
      ddValue = noiseParam * BlasDrivers::dot( arDi, ARutils::ARvec( 1, arParam, my_ac->DprofileR( dbParams, respParams, shapeParams, j ) ) );

      DDll( iOffset, jOffset ) -= ddValue;
      DDll( jOffset, iOffset ) -= ddValue;
     }

                                        //           D2( responsiveness, shape )
                                        //
    for ( int jOffset = ofsMinShape,
              j       = jOffset - my_shapeOffset;
              jOffset < ofsMaxShape; jOffset++, j++ )
     {
      ddValue = noiseParam * ( BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_ac->DDprofileRS( dbParams, respParams, shapeParams, i, j ) ) ) - BlasDrivers::dot( arDi, ARutils::ARvec( 1, arParam, my_ac->DprofileS( dbParams, respParams, shapeParams, j ) ) ) );

      DDll( iOffset, jOffset ) += ddValue;
      DDll( jOffset, iOffset ) += ddValue;
     }
   }

  for ( int iOffset = ofsMinShape,
            i       = iOffset - my_shapeOffset;
            iOffset < ofsMaxShape; iOffset++, i++ )
   {
    my_ac->getDprofileS( dbParams, respParams, shapeParams, i, Di );
    ARutils::getARvec( 1, arParam, Di, arDi );

                                        //                    D2( noise, shape )
                                        //
    if ( doNoise )
     {
      ddValue = BlasDrivers::dot( arResid, arDi );

      DDll( iOffset, my_noiseOffset ) += ddValue;
      DDll( my_noiseOffset, iOffset ) += ddValue;
     }

                                        //                      D2( ar, shape )
                                        //
    if ( doAR )
     {
      ddValue = noiseParam * ( BlasDrivers::dot( signalSize, arResid( 1 ), Di( 0 ), 1, 1 ) + BlasDrivers::dot( signalSize, arDi( 1 ), residual( 0 ), 1, 1 ) );

      DDll( iOffset, my_arOffset ) -= ddValue;
      DDll( my_arOffset, iOffset ) -= ddValue;
     }

                                        //                    D2( shape, shape )
                                        //                              diagonal
                                        //
    DDll( iOffset, iOffset ) += noiseParam * ( BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_ac->DDprofileSS( dbParams, respParams, shapeParams, i, i ) ) ) - BlasDrivers::dot( arDi,  arDi ) );

                                        //                    D2( shape, shape )
                                        //                          off-diagonal
                                        //
    for ( int jOffset = ofsMinShape,
              j       = jOffset - my_shapeOffset;
              jOffset < iOffset; jOffset++, j++ )
     {
      ddValue = noiseParam * ( BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_ac->DDprofileSS( dbParams, respParams, shapeParams, i, j ) ) ) - BlasDrivers::dot( arDi, ARutils::ARvec( 1, arParam, my_ac->DprofileS( dbParams, respParams, shapeParams, j ) ) ) );

      DDll( iOffset, jOffset ) += ddValue;
      DDll( jOffset, iOffset ) += ddValue;
     }
   }
 }

#endif /* __LLn_a_1_0_H__ */
