/* ----------------------------------------------------------------------------
 *
 * class LLnormal<false,1,0>
 *   template specialization: log normal/ARMA(1,0) likelihood without activation
 *
 * required headers
 *   see include header file LLnormal
 *
 * notes
 *   0. follows convention for out-of-range parameter offsets; see BasicObjFn.h
 *   1. acProfile needed only for the sake of a common interface.
 *   2. the single AR parameter has bounds |phi| <= 1 corresponding to the usual
 *      causality and stationarity conditions; this is to be imposed externally.
 *   3. This particular implementation conditions the likelihood on the first
 *      data point.
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __LLn_n_1_0_H__
#define __LLn_n_1_0_H__

cpp_using_namespace( BasicTypes );

class LLnormal<false,1,0> : public BasicObjFn
 {
  public:
   // --------------------- construct/reinitialize/incremental initialize/assign
   // --------------------------------------------------------------------------

   LLnormal( );

   LLnormal( const LLnormal& another );

   LLnormal&
   operator=( const LLnormal& rhs );

   LLnormal( const fpVector*                 signal,
             const dbProfile*                db,
             const acProfile*                ac,
             const ParameterManager<fpType>& pm,
                   int                       paramBufSize,
                   fpType                    sign = 1.0,
             const string&                   name = "LL normal/no activation/arma(1,0)" );

   void
   setSignal( const fpVector* signal );

   void
   setDBprofile( const dbProfile* db );

   void
   setACprofile( const acProfile* ac );

   void
   setParamOffset( const ParameterManager<fpType>& pm );

   void
   reinitialize( const fpVector*                 signal,
                 const dbProfile*                db,
                 const acProfile*                ac,
                 const ParameterManager<fpType>& pm,
                       int                       paramBufSize,
                       fpType                    sign = 1.0,
                 const string&                   name = "LL normal/no activation/arma(1,0)" );

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   fpType
   additiveConst( ) const;
                                //
                                // additive constants arising from normalization
                                //   that are not included in value computation.
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   // -------------------------------------------------- likelihood computations
   // ----------------- without initialization of output arguments ll, Dll, DDll
   // --------------------------------------------------------------------------

                             //              the implementations of gradient and
                             //  hessian make the assumption that drift+baseline
                             //        is linear in parameters and use functions
                             // in dbProfile which are based on this assumption.

   inline
   void
   value_ni( const fpType* const paramBuf, fpType* ll ) const;

   inline
   void
   gradient_ni( const fpType* const paramBuf, fpType* Dll ) const;

         // the parameters outside the parameter buffer, i. e., with offset < 0,
         // offset >= paramBufSize( ), are treated as constants in the objective
         // function for minimization purposes.  Consequently, the corresponding
         //                   gradients are not calculated for these parameters.
         //
         // the same convention can be used in the hessian calculation, but that
         // not  been  done  in  this  implementation, assuming that most of the
         // minimization methods  of  practical utility are gradient-based,  and
         //                               do not require the hessian explicitly.

   inline
   void
   hessian_ni( const fpType* const parameters, fpMatrix& DDll ) const;

  private:
   const fpVector*  my_signal;
   const dbProfile* my_db;
         int        my_dbOffset, my_noiseOffset, my_arOffset;
         int        my_dbLength, my_noiseLength, my_arLength;
 };

// ------------------------ construct/reinitialize/incremental initialize/assign
// -----------------------------------------------------------------------------

LLnormal<false,1,0>::LLnormal( )
 : BasicObjFn( ),
   my_signal( 0 ),
   my_db( 0 ),
   my_dbOffset( 0 ),
   my_noiseOffset( 0 ),
   my_arOffset( 0 ),
   my_dbLength( 0 ),
   my_noiseLength( 0 ),
   my_arLength( 1 )
 {
  BasicObjFn::setName( "LL normal/no activation/arma(1,0)" );
 }

LLnormal<false,1,0>::LLnormal( const LLnormal& another )
 : BasicObjFn( another.paramBufSize( ), another.sign( ), another.name( ) ),
   my_signal( another.my_signal ),
   my_db( another.my_db ),
   my_dbOffset( another.my_dbOffset ),
   my_noiseOffset( another.my_noiseOffset ),
   my_arOffset( another.my_arOffset ),
   my_dbLength( another.my_dbLength ),
   my_noiseLength( another.my_noiseLength )
 {
 }

LLnormal<false,1,0>&
LLnormal<false,1,0>::operator=( const LLnormal<false,1,0>& rhs )
 {
  BasicObjFn::setParamBufSize( rhs.paramBufSize( ) );
  BasicObjFn::setSign( rhs.sign( ) );
  BasicObjFn::setName( rhs.name( ) );

  my_signal = rhs.my_signal;

  my_db = rhs.my_db;

  my_dbOffset    = rhs.my_dbOffset;
  my_noiseOffset = rhs.my_noiseOffset;
  my_arOffset    = rhs.my_arOffset;

  my_dbLength    = rhs.my_dbLength;
  my_noiseLength = rhs.my_noiseLength;

  return *this;
 }

LLnormal<false,1,0>::LLnormal( const fpVector*                 signal,
                               const dbProfile*                db,
                               const acProfile*                ac,
                               const ParameterManager<fpType>& pm,
                                     int                       paramBufSize,
                                     fpType                    sign,
                               const string&                   name )
 : BasicObjFn( paramBufSize, sign, name ),
   my_signal( signal ),
   my_db( db ),
   my_dbOffset( pm.blockOffset( ParamBlkNames::drift ) ),
   my_noiseOffset( pm.blockOffset( ParamBlkNames::noise ) ),
   my_arOffset( pm.blockOffset( ParamBlkNames::ar ) ),
   my_dbLength( pm.blockLength( ParamBlkNames::drift ) ),
   my_noiseLength( pm.blockLength( ParamBlkNames::noise ) ),
   my_arLength( 1 )
 {
 }

void
LLnormal<false,1,0>::setSignal( const fpVector* signal )
 {
  my_signal = signal;
 }

void
LLnormal<false,1,0>::setDBprofile( const dbProfile* db )
 {
  my_db = db;
 }

void
LLnormal<false,1,0>::setACprofile( const acProfile* ac )
 {
 }

void
LLnormal<false,1,0>::setParamOffset( const ParameterManager<fpType>& pm )
 {
  my_dbOffset    = pm.blockOffset( ParamBlkNames::drift );
  my_noiseOffset = pm.blockOffset( ParamBlkNames::noise );
  my_arOffset    = pm.blockOffset( ParamBlkNames::ar );

  my_dbLength    = pm.blockLength( ParamBlkNames::drift );
  my_noiseLength = pm.blockLength( ParamBlkNames::noise );
 }

void
LLnormal<false,1,0>::reinitialize( const fpVector*                 signal,
                                   const dbProfile*                db,
                                   const acProfile*                ac,
                                   const ParameterManager<fpType>& pm,
                                         int                       paramBufSize,
                                         fpType                    sign,
                                   const string&                   name )
 {
  BasicObjFn::reinitialize( paramBufSize, sign, name );

  setSignal( signal );
  setDBprofile( db );
  setACprofile( ac );
  setParamOffset( pm );
 }

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

fpType
LLnormal<false,1,0>::additiveConst( ) const
 {
  return - 0.5 * ( my_signal->size( ) - 1 ) * log( 2.0 * M_PI );
 }

// ----------------------------------------------------- likelihood computations
// -------------------- without initialization of output arguments ll, Dll, DDll
// -----------------------------------------------------------------------------

void
LLnormal<false,1,0>::value_ni( const fpType* const paramBuf, fpType* ll ) const
 {
  fpType noiseParam = paramBuf[my_noiseOffset];

  fpVector residual( (*my_signal) );
  residual -= my_db->profile( paramBuf + my_dbOffset );

  ARutils::getARvec( 1, paramBuf + my_arOffset, residual );

  (*ll) += 0.5 * ( ( my_signal->size( ) - 1 ) * log( noiseParam ) - noiseParam * BlasDrivers::dot( residual, residual ) );
 }

void
LLnormal<false,1,0>::gradient_ni( const fpType* const paramBuf,
                                        fpType*       Dll ) const
 {
  const fpType* dbParams = paramBuf + my_dbOffset;
  const fpType* arParam  = paramBuf + my_arOffset;

  fpType noiseParam = paramBuf[my_noiseOffset];

  int signalSize = my_signal->size( ) - 1;

  fpVector residual( (*my_signal) );
  residual -= my_db->profile( dbParams );

  fpVector arResid( residual );
  ARutils::getARvec( 1, arParam, arResid );

                                                  //                  D( noise )
                                                  //
  if ( ( my_noiseOffset >= 0 ) && ( my_noiseOffset < paramBufSize( ) ) )
   Dll[my_noiseOffset] += 0.5 * ( signalSize / noiseParam - BlasDrivers::dot( arResid, arResid ) );

                                                  //                     D( ar )
                                                  //
  if ( ( my_arOffset >= 0 ) && ( my_arOffset < paramBufSize( ) ) )
   Dll[my_arOffset] += noiseParam * BlasDrivers::dot( signalSize, arResid( 1 ), residual( 0 ), 1, 1 );

                                                  //         D( drift+baseline )
                                                  //
  int ofsMax = min( my_dbOffset + my_dbLength, paramBufSize( ) );
  for ( int iOffset = max( 0, my_dbOffset ),
            i       = iOffset - my_dbOffset;
            iOffset < ofsMax; iOffset++, i++ )
   {
    Dll[iOffset] += noiseParam * BlasDrivers::dot( arResid, ARutils::ARvec( 1, arParam, my_db->Dprofile( dbParams, i ) ) );
   }
 }

void
LLnormal<false,1,0>::hessian_ni( const fpType* const paramBuf,
                                       fpMatrix&     DDll ) const
 {
  const fpType* dbParams = paramBuf + my_dbOffset;
  const fpType* arParam  = paramBuf + my_arOffset;

  fpType noiseParam = paramBuf[my_noiseOffset];

  int signalSize = my_signal->size( ) - 1;

  fpVector residual( (*my_signal) );
  residual -= my_db->profile( dbParams );

  fpVector arResid( residual );
  ARutils::getARvec( 1, arParam, arResid );

  int ofsMinDB = max( 0, my_dbOffset );
  int ofsMaxDB = min( my_dbOffset + my_dbLength, paramBufSize( ) );

  bool doNoise = ( my_noiseOffset >= 0 ) && ( my_noiseOffset < paramBufSize( ));
  bool doAR    = ( my_arOffset    >= 0 ) && ( my_arOffset    < paramBufSize( ));

  fpType ddValue;
  fpVector df, ardf;

                                        //                    D2( noise, noise )
                                        //
  if ( doNoise )
   {
    DDll( my_noiseOffset, my_noiseOffset )
     -= 0.5 * signalSize / ( noiseParam * noiseParam );

                                                  //             D2( noise, ar )
                                                  //
    if ( doAR )
     {
      ddValue = BlasDrivers::dot( signalSize, arResid( 1 ), residual( 0 ), 1, 1 );

      DDll( my_arOffset, my_noiseOffset ) += ddValue;
      DDll( my_noiseOffset, my_arOffset ) += ddValue;
     }
   }

                                                  //                D2( ar, ar )
                                                  //
  if ( doAR )
   DDll( my_arOffset, my_arOffset ) -= noiseParam * BlasDrivers::dot( residual, residual );

  for ( int iOffset = ofsMinDB,
            i       = iOffset - my_dbOffset;
            iOffset < ofsMaxDB; iOffset++, i++ )
   {
    my_db->getDprofile( dbParams, i, df );
    ARutils::getARvec( 1, arParam, df, ardf );

                                        //           D2( noise, drift+baseline )
                                        //
    if ( doNoise )
     {
      ddValue = BlasDrivers::dot( arResid, ardf );

      DDll( iOffset, my_noiseOffset ) += ddValue;
      DDll( my_noiseOffset, iOffset ) += ddValue;
     }

                                        //              D2( ar, drift+baseline )
                                        //
    if ( doAR )
     {
      ddValue = noiseParam * ( BlasDrivers::dot( signalSize, arResid( 1 ), df( 0 ), 1, 1 ) + BlasDrivers::dot( signalSize, ardf( 1 ), residual( 0 ), 1, 1 ) );

      DDll( iOffset, my_arOffset ) -= ddValue;
      DDll( my_arOffset, iOffset ) -= ddValue;
     }

                                        //  D2( drift+baseline, drift+baseline )
                                        //                              diagonal
                                        //
    DDll( iOffset, iOffset ) -= noiseParam * BlasDrivers::dot( ardf, ardf );

                                        //  D2( drift+baseline, drift+baseline )
                                        //                          off-diagonal
                                        //
    for ( int jOffset = ofsMinDB,
              j       = jOffset - my_dbOffset;
              jOffset < iOffset; jOffset++, j++ )
     {
      ddValue =   noiseParam * BlasDrivers::dot( ardf, ARutils::ARvec( 1, arParam, my_db->Dprofile( dbParams, j ) ) );

      DDll( iOffset, jOffset ) -= ddValue;
      DDll( jOffset, iOffset ) -= ddValue;
     }
   }
 }

#endif /* __LLn_n_1_0_H__ */
