/* ----------------------------------------------------------------------------
 *
 * class LPbaseline
 *   baseline log prior computations
 *
 * required headers
 *   see include header file LPbaseline
 *
 * notes
 *   follows convention for out-of-range parameter offsets; see BasicObjFn.h
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __LPbaseline_H__
#define __LPbaseline_H__

cpp_using_namespace( BasicTypes );

class LPbaseline : public BasicObjFn
 {
  public:
   // ---------------------------- construct/reinitialize/incremental initialize
   // --------------------------------------------------------------------------

   LPbaseline( );

   LPbaseline( const LPbaseline& another );

   LPbaseline&
   operator=( const LPbaseline& rhs );

   LPbaseline( const dbProfile*                db,
               const ParameterManager<fpType>& pm,
                     fpType                    scale,
                     fpType                    shift,
                     int                       paramBufSize,
                     fpType                    sign = 1.0,
               const string&                   name = "LP Baseline" );

   LPbaseline( const dbProfile* db,
                     int        dbOffset,
                     fpType     scale,
                     fpType     shift,
                     int        paramBufSize,
                     fpType     sign = 1.0,
               const string&    name = "LP Baseline" );

   void
   setDBprofile( const dbProfile* db );

   void
   setParamOffset( const ParameterManager<fpType>& pm );

   void
   setParamOffset( int dbOffset );

   void
   setConstants( fpType scale, fpType shift );

   void
   reinitialize( const dbProfile*                db,
                 const ParameterManager<fpType>& pm,
                       fpType                    scale,
                       fpType                    shift,
                       int                       paramBufSize,
                       fpType                    sign = 1.0,
                 const string&                   name = "LP Baseline" );

   void
   reinitialize( const dbProfile* db,
                       int        dbOffset,
                       fpType     scale,
                       fpType     shift,
                       int        paramBufSize,
                       fpType     sign = 1.0,
                 const string&    name = "LP Baseline" );

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   fpType
   scale( ) const;

   inline
   fpType
   shift( ) const;

   inline
   fpType
   additiveConst( ) const;
                                //
                                // additive constants arising from normalization
                                //   that are not included in value computation.
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   // ------------------------------------------------------- prior computations
   // ----------------- without initialization of output arguments lp, Dlp, DDlp
   // --------------------------------------------------------------------------

   inline
   void
   value_ni( const fpType* const paramBuf, fpType* lp ) const;

   inline
   void
   gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const;

         // the parameters outside the parameter buffer, i. e., with offset < 0,
         // offset >= paramBufSize( ), are treated as constants in the objective
         // function for minimization purposes.  Consequently, the corresponding
         //                   gradients are not calculated for these parameters.
         //
         // the same convention can be used in the hessian calculation, but that
         // not  been  done  in  this  implementation, assuming that most of the
         // minimization methods  of  practical utility are gradient-based,  and
         //                               do not require the hessian explicitly.

   inline
   void
   hessian_ni( const fpType* const paramBuf, fpMatrix& DDlp ) const;

  private:
   const dbProfile* my_db;
         int        my_dbOffset;
         fpType     my_scale;
         fpType     my_shift;
 };

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

fpType
LPbaseline::scale( ) const
 {
  return my_scale;
 }

fpType
LPbaseline::shift( ) const
 {
  return my_shift;
 }

fpType
LPbaseline::additiveConst( ) const
 {
  return 0.5 * log( my_scale ) - log( M_PI );
 }

// ---------------------------------------------------------- prior computations
// -------------------- without initialization of output arguments lp, Dlp, DDlp
// -----------------------------------------------------------------------------

void
LPbaseline::value_ni( const fpType* const paramBuf, fpType* lp ) const
 {
  fpType p = my_db->baseline( paramBuf + my_dbOffset ) - my_shift;

  (*lp) -= log( 1.0 + my_scale * p * p );
 }

void
LPbaseline::gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const
 {
  int dbLength = my_db->basisSize( );

  const fpType* dbParams = paramBuf + my_dbOffset;

  fpType p = my_db->baseline( dbParams ) - my_shift;
         p = 2.0 * my_scale * p / ( 1.0 + my_scale * p * p );

  const fpVector& Db = my_db->Dbaseline( dbParams );

  int ofsMax = min( my_dbOffset + dbLength, paramBufSize( ) );
  for ( int iOffset = max( 0, my_dbOffset ),
            i       = iOffset - my_dbOffset;
            iOffset < ofsMax; iOffset++, i++ )
   Dlp[iOffset] -= p * Db[i];
 }

void
LPbaseline::hessian_ni( const fpType* const paramBuf, fpMatrix& DDlp ) const
 {
  int dbLength = my_db->basisSize( );

  const fpType* dbParams = paramBuf + my_dbOffset;

  fpType p1  = my_db->baseline( dbParams ) - my_shift;
         p1 *= p1;
  fpType p2  = 2.0 * my_scale / ( 1.0 + my_scale * p1 );
         p2  = p2 * ( p1 * p2 - 1.0 );
  fpType p3, p4;

  const fpVector& Db = my_db->Dbaseline( dbParams );

  int ofsMinDB = max( 0, my_dbOffset );
  int ofsMaxDB = min( my_dbOffset + dbLength, paramBufSize( ) );

  for ( int iOffset = ofsMinDB,
            i       = iOffset - my_dbOffset;
            iOffset < ofsMaxDB; iOffset++, i++ )
   {
    p3 = p2 * Db[i];
    DDlp( iOffset, iOffset ) += p3 * Db[i];

    for ( int jOffset = ofsMinDB,
              j       = jOffset - my_dbOffset;
              jOffset < iOffset; jOffset++, j++ )
     {
      p4 = p3 * Db[j];

      DDlp( iOffset, jOffset ) += p4;
      DDlp( jOffset, iOffset ) += p4;
     }
   }
 }

#endif /* __LPbaseline_H__ */
