/* ----------------------------------------------------------------------------
 *
 * class LPdrift
 *   drift log prior computations
 *
 * required headers
 *   see include header file LPdrift
 *
 * notes
 *   follows convention for out-of-range parameter offsets; see BasicObjFn.h
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __LPdrift_H__
#define __LPdrift_H__

cpp_using_namespace( BasicTypes );

class LPdrift : public BasicObjFn
 {
  public:
   // ---------------------------- construct/reinitialize/incremental initialize
   // --------------------------------------------------------------------------

   LPdrift( );

   LPdrift( const LPdrift& another );

   LPdrift&
   operator=( const LPdrift& rhs );

   LPdrift( const dbProfile*                db,
            const ParameterManager<fpType>& pm,
                  fpType                    smoothingPar,
                  fpType                    relWeight,
                  int                       paramBufSize,
                  fpType                    sign = 1.0,
            const string&                   name = "LP Drift" );

   LPdrift( const dbProfile* db,
                  int        dbOffset,
                  int        noiseOffset,
                  fpType     smoothingPar,
                  fpType     relWeight,
                  int        paramBufSize,
                  fpType     sign = 1.0,
            const string&    name = "LP Drift" );

   void
   setDBprofile( const dbProfile* db );

   void
   setParamOffset( const ParameterManager<fpType>& pm );

   void
   setParamOffset( int dbOffset, int noiseOffset );

   void
   setConstants( fpType smoothingPar, fpType relWeight );

   void
   reinitialize( const dbProfile*                db,
                 const ParameterManager<fpType>& pm,
                       fpType                    smoothingPar,
                       fpType                    relWeight,
                       int                       paramBufSize,
                       fpType                    sign = 1.0,
                 const string&                   name = "LP Drift" );

   void
   reinitialize( const dbProfile* db,
                       int        dbOffset,
                       int        noiseOffset,
                       fpType     smoothingPar,
                       fpType     relWeight,
                       int        paramBufSize,
                       fpType     sign = 1.0,
                 const string&    name = "LP Drift" );

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   fpType
   relWeight( ) const;

   inline
   fpType
   smoothingPar( ) const;

   inline
   fpType
   additiveConst( ) const;
                                //
                                // additive constants arising from normalization
                                //   that are not included in value computation.
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).
                                //
                                //         ATTN not figured out yet. returns 0.0

   // ------------------------------------------------------- prior computations
   // ----------------- without initialization of output arguments lp, Dlp, DDlp
   // --------------------------------------------------------------------------

   inline
   void
   value_ni( const fpType* const paramBuf, fpType* lp ) const;

   inline
   void
   gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const;

         // the parameters outside the parameter buffer, i. e., with offset < 0,
         // offset >= paramBufSize( ), are treated as constants in the objective
         // function for minimization purposes.  Consequently, the corresponding
         //                   gradients are not calculated for these parameters.
         //
         // the same convention can be used in the hessian calculation, but that
         // not  been  done  in  this  implementation, assuming that most of the
         // minimization methods  of  practical utility are gradient-based,  and
         //                               do not require the hessian explicitly.

   inline
   void
   hessian_ni( const fpType* const paramBuf, fpMatrix& DDlp ) const;

  private:
   const dbProfile* my_db;
         int        my_dbOffset, my_noiseOffset;
         fpType     my_smoothingPar, my_relWeight,
                    my_halfNonNullSize, my_logNormConst;

   void
   setLogNormConst( );
 };

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

fpType
LPdrift::relWeight( ) const
 {
  return my_relWeight;
 }

fpType
LPdrift::smoothingPar( ) const
 {
  return 0.5 / my_smoothingPar;
 }

fpType
LPdrift::additiveConst( ) const
 {
  return my_logNormConst;
 }

// ---------------------------------------------------------- prior computations
// -------------------- without initialization of output arguments lp, Dlp, DDlp
// -----------------------------------------------------------------------------

void
LPdrift::value_ni( const fpType* const paramBuf, fpType* lp ) const
 {
  fpType noiseParam = paramBuf[my_noiseOffset];

  const fpType* dbParams = paramBuf + my_dbOffset;

  (*lp) += my_halfNonNullSize * log( noiseParam );

  noiseParam *= my_smoothingPar;

  (*lp) -= noiseParam * my_db->normSqrDDDrift( dbParams );

  if ( my_relWeight != 0.0 )
   (*lp) -= my_relWeight * noiseParam * my_db->normSqrDrift( dbParams );
                     //
                     // ATTN assuming that my_relWeight == 0.0 most of the times
 }

void
LPdrift::gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const
 {
  fpType noiseParam = paramBuf[my_noiseOffset];

  int dbLength = my_db->basisSize( );

  const fpType* dbParams = paramBuf + my_dbOffset;

  fpVector DnsDDD;
  my_db->getDnormSqrDDDrift( dbParams, DnsDDD );

  fpType c = my_halfNonNullSize / noiseParam;

  if ( ( my_noiseOffset >= 0 ) && ( my_noiseOffset < paramBufSize( ) ) )
   Dlp[my_noiseOffset] += c - my_smoothingPar * my_db->normSqrDDDrift( dbParams );

  c = noiseParam * my_smoothingPar;

  int ofsMax = min( my_dbOffset + dbLength, paramBufSize( ) );
  for ( int iOffset = max( 0, my_dbOffset ),
            i       = iOffset - my_dbOffset;
            iOffset < ofsMax; iOffset++, i++ ) Dlp[iOffset] -= c * DnsDDD[i];

  if ( my_relWeight != 0.0 )
   {
    fpVector DnsD;
    my_db->getDnormSqrDrift( dbParams, DnsD );

    c = my_relWeight * my_smoothingPar;

    if ( ( my_noiseOffset >= 0 ) && ( my_noiseOffset < paramBufSize( ) ) )
     Dlp[my_noiseOffset] -= c * my_db->normSqrDrift( dbParams );

    c *= noiseParam;

    for ( int iOffset = max( 0, my_dbOffset ),
              i       = iOffset - my_dbOffset;
              iOffset < ofsMax; iOffset++, i++ ) Dlp[iOffset] -= c * DnsD[i];
   }
 }

void
LPdrift::hessian_ni( const fpType* const paramBuf, fpMatrix& DDlp ) const
 {
  fpType noiseParam = paramBuf[my_noiseOffset];

  int dbLength = my_db->basisSize( );

  const fpType* dbParams = paramBuf + my_dbOffset;

  fpVector g( 0.0, paramBufSize( ) );
  gradient_ni( paramBuf, g( 0 ) );

  fpType p;

  int ofsMinDB = max( 0, my_dbOffset );
  int ofsMaxDB = min( my_dbOffset + dbLength, paramBufSize( ) );

  if ( ( my_noiseOffset >= 0 ) && ( my_noiseOffset < paramBufSize( ) ) )
   {
    DDlp( my_noiseOffset, my_noiseOffset )
     -= my_halfNonNullSize / ( noiseParam * noiseParam );

    for ( int iOffset = ofsMinDB,
              i       = iOffset - my_dbOffset;
              iOffset < ofsMaxDB; iOffset++, i++ )
     {
      p = g[iOffset] / noiseParam;

      DDlp( iOffset, my_noiseOffset ) += p;
      DDlp( my_noiseOffset, iOffset ) += p;
     }
   }

  fpType c = noiseParam * my_smoothingPar;

  for ( int iOffset = ofsMinDB,
            i       = iOffset - my_dbOffset;
            iOffset < ofsMaxDB; iOffset++, i++ )
   {
    DDlp( iOffset, iOffset ) -= c * my_db->DDnormSqrDDDrift( dbParams, i, i );

    for ( int jOffset = ofsMinDB,
              j       = jOffset - my_dbOffset;
              jOffset < iOffset; jOffset++, j++ )
     {
      p = - c * my_db->DDnormSqrDDDrift( dbParams, i, j );

      DDlp( iOffset, jOffset ) += p;
      DDlp( jOffset, iOffset ) += p;
     }
   }

  if ( my_relWeight != 0.0 )
   {
    c *= my_relWeight;

    for ( int iOffset = ofsMinDB,
              i       = iOffset - my_dbOffset;
              iOffset < ofsMaxDB; iOffset++, i++ )
     {
      DDlp( iOffset, iOffset ) -= c * my_db->DDnormSqrDrift( dbParams, i, i );

      for ( int jOffset = ofsMinDB,
                j       = jOffset - my_dbOffset;
                jOffset < iOffset; jOffset++, j++ )
       {
        p = - c * my_db->DDnormSqrDrift( dbParams, i, j );

        DDlp( iOffset, jOffset ) += p;
        DDlp( jOffset, iOffset ) += p;
       }
     }
   }
 }

#endif /* __LPdrift_H__ */
