/* ----------------------------------------------------------------------------
 *
 * class SubmodelTable
 *   implementation
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#include <brain.model/SubmodelTable>

// ---------------------------------------------------------------------- static
// -----------------------------------------------------------------------------

size_t
SubmodelTable::numActive( const string& key )
 {
  size_t numActive = 0;
  for ( size_t i = 0; i < key.size( ); i++ ) numActive += ( key[i] == '1' );
  return numActive;
 }

SubmodelTable::Submodel
SubmodelTable::submodel( const string& key )
 {
  SubmodelTable::Submodel m( numActive( key ) );

  for ( size_t i = 0, j = 0; i < key.size( ); i++ )
   if ( key[i] == '1' )
    {
     m[j] = i;                          // original: m[j] = key.size( ) - i - 1;
     j++;
    }

  return m;
 }

// --------------------------------------------------------- incremental buildup
// -----------------------------------------------------------------------------

SubmodelTable::SubmodelTable( )
 : my_numConditions( 0 ),
   my_conditionNames( ),
   my_smTable( )
 {
 }

SubmodelTable::SubmodelTable( const SubmodelTable& another )
 : my_numConditions( another.my_numConditions ),
   my_conditionNames( another.my_conditionNames ),
   my_smTable( another.my_smTable )
 {
 }

SubmodelTable::SubmodelTable(       size_t          numConditions,
                              const string&         specification,
                              const string&         separators,
                              const vector<string>& conditionNames )
 : my_numConditions( 0 ),
   my_conditionNames( ),
   my_smTable( )
 {
  setNumConditions( numConditions, conditionNames );
  updateWith( specification, separators );
 }

void
SubmodelTable::setNumConditions(       size_t          numConditions,
                                 const vector<string>& conditionNames )
 {
  if ( my_numConditions == numConditions ) return;

  my_numConditions = numConditions;

  size_t nd = MiscFns::nDigits( my_numConditions );

  string fmt = "%";
  cpp_scope_begin
   char ndstr[nd];
   sprintf( ndstr, "%d", nd );

   fmt.append( ndstr );
   fmt.append( "d" );
  cpp_scope_end

  char s[nd];

  my_conditionNames.resize( my_numConditions );

  size_t cmax = min( my_numConditions, conditionNames.size( ) );

  for ( size_t c = 0; c < cmax; c++ )
   {
    if ( conditionNames[c] != "" )
     my_conditionNames[c] = conditionNames[c];
    else
     {
      my_conditionNames[c] = "C";

      for ( size_t d = 0; d < nd; d++ ) s[d] = '0';
      sprintf( s, fmt.c_str( ), c );

      my_conditionNames[c].append( s );
     }
   }

  for ( size_t c = cmax; c < my_numConditions; c++ )
   {
    my_conditionNames[c] = "C";

    for ( size_t d = 0; d < nd; d++ ) s[d] = '0';
    sprintf( s, fmt.c_str( ), c );

    my_conditionNames[c].append( s );
   }

  delAll( );
 }

void
SubmodelTable::addSubmodel( const Submodel& sm )
 {
  if (    ( my_numConditions == 0 )
       || ( sm.size( )       >  my_numConditions )
       || ( find( my_smTable.begin( ), my_smTable.end( ), sm ) != my_smTable.end( ) ) ) return;

  if ( sm.size( ) < my_smTable.front( ).size( ) )
   my_smTable.push_front( sm );
  else if ( sm.size( ) > my_smTable.back( ).size( ) )
   my_smTable.push_back( sm );
  else
   {
    iterator i = my_smTable.begin( );
    for ( ; ( i != my_smTable.end( ) ) && ( (*i).size( ) < sm.size( ) ); ++i ) ;
    my_smTable.insert( i, sm );
   }
 }

void
SubmodelTable::delSubmodel( const Submodel& sm )
 {
  my_smTable.remove( sm );
 }

void
SubmodelTable::addSubmodel( size_t numActive )
 {
  if (    ( my_numConditions == 0 )
       || ( numActive        >  my_numConditions ) ) return;

  string key( my_numConditions, '1' );
  if ( numActive <= my_numConditions )
   for ( size_t a = 0; a < my_numConditions - numActive; a++ ) key[a] = '0';

  do
   addSubmodel( submodel( key ) );
  while ( next_permutation( key.begin( ), key.end( ) ) );
 }

void
SubmodelTable::delSubmodel( size_t numActive )
 {
  if (    ( my_numConditions == 0 )
       || ( numActive        >  my_numConditions ) ) return;

  my_smTable.remove_if( hasActiveEq( numActive ) );
 }

void
SubmodelTable::addByKey( const string& key )
 {
  if (    ( my_numConditions == 0 )
       || ( key.size( )      != my_numConditions ) ) return;

  addSubmodel( submodel( key ) );
 }

void
SubmodelTable::delByKey( const string& key )
 {
  if (    ( my_numConditions == 0 )
       || ( key.size( )      != my_numConditions ) ) return;

  my_smTable.remove( submodel( key ) );
 }

void
SubmodelTable::addByName( const string& modelName )
 {
  if ( my_numConditions == 0 ) return;

  if ( modelName == "null" )
   addSubmodel( 0 );
  else if ( modelName == "full" )
   addSubmodel( my_numConditions - 1 );
  else if ( modelName == "saturated" )
   addSubmodel( my_numConditions );
  else if ( modelName == "all" )
   for ( size_t c = 0; c < my_numConditions; c++ ) addSubmodel( c );
 }

void
SubmodelTable::delByName( const string& modelName )
 {
  if ( modelName == "null" )
   delSubmodel( 0 );
  else if ( modelName == "full" )
   delSubmodel( my_numConditions - 1 );
  else if ( modelName == "saturated" )
   delSubmodel( my_numConditions );
  else if ( modelName == "all" )
   for ( size_t c = 0; c < my_numConditions; c++ ) delSubmodel( c );
 }

void
SubmodelTable::delAll( )
 {
  my_smTable.resize( 0 );
 }

void
SubmodelTable::updateWith( const string& specification, const string& separators )
 {
  if ( my_numConditions == 0 ) return;

  int numActive;
  Submodel submodel;

  vector<string> tokens;
  MiscFns::getTokens( specification, tokens, separators );

  vector<string>::const_iterator t      = tokens.begin( );
  vector<string>::const_iterator op     = t;
  vector<string>::const_iterator format = ++t;

  while ( !(    ( t  == tokens.end( ) )
             || ( (*op) == "end" )
             || ( ( (*op) != "+" ) && ( (*op) != "-" ) )
             || (    ( (*format) != "name"     ) && ( (*format) != "active" )
                  && ( (*format) != "submodel" ) && ( (*format) != "key"    ) ) ) )
   {
    if ( (*format) == "name" )
     {
      ++t;

      if ( (*op) == "+" ) addByName( (*t) ); else delByName( (*t) );
     }
    else if ( (*format) == "active" )
     {
      ++t;
      numActive = atoi( (*t).c_str( ) );

      if ( numActive > 0 )
       if ( (*op) == "+" ) addSubmodel( numActive ); else delSubmodel( numActive );
     }
    else if ( (*format) == "submodel" )
     {
      ++t;
      numActive = atoi( (*t).c_str( ) );

      if ( numActive > 0 )
       {
        submodel.resize( numActive );

        bool specOK = true;
        for ( int a = 0; ( a < numActive ) && specOK; a++ )
         {
          ++t;
          vector<string>::iterator
           c = find( my_conditionNames.begin( ), my_conditionNames.end( ), (*t) );
          specOK = ( c != my_conditionNames.end( ) );
          if ( specOK ) submodel[a] = distance( my_conditionNames.begin( ), c );
         }

        if ( specOK )
         if ( (*op) == "+" ) addSubmodel( submodel ); else delSubmodel( submodel );
       }
     }
    else if ( (*format) == "key" )
     {
      ++t;

      if ( (*op) == "+" ) addByKey( (*t) ); else delByKey( (*t) );
     }

    op = ++t; format = ++t;
   }
 }
