/* ----------------------------------------------------------------------------
 *
 * class SubmodelTable
 *   submodels list
 *
 * required headers
 *   see include header file SubmodelTable
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __SubmodelTable_H__
#define __SubmodelTable_H__

cpp_using_namespace( std );

class SubmodelTable
 {
  public:
   typedef short int             ConditionType;
   typedef vector<ConditionType> Submodel;
                 //
                 //                         Submodel must support the following:
                 //
                 //                                        a default constructor
                 //                           size_t operator []( size_t ) const
                 //                                         size_t size( ) const
                 //         bool operator ==( const Submodel&, const Submodel& )
                 //
                 //        any representation consistent with the above is fine.
                 //
                 //         conditions are labelled 0, 1, ..., numConditions - 1
                 //  0 <= Submodel[i] < numConditions; 0 <= i < Submodel.size( )
                 //          Submodel.size( ) == # active conditions in Submodel

   typedef list<Submodel>::const_iterator const_iterator;
   typedef list<Submodel>::iterator       iterator;

   // ----------------------------------------------- incremental initialization
   // --------------------------------------------------------------------------

   SubmodelTable(       size_t  numConditions,
                  const string& specification          = "+ name all end",
                  const string& separator              = " \t",
                  const vector<string>& conditionNames = vector<string>( ) );
     //
     // separator consists of characters used to delimit tokens in specification
     // specification: "op format value op format value ... op format value end"
     //
     //                                              op: '+' (add), '-' (delete)
     //                              format: "name", "active", "submodel", "key"
     //                    value: submodel specification in the specified format
     //
     //                                     admissible values for format "name":
     //
     //                                     name          # of active conditions
     //
     //                                     "null"                             0
     //                                     "all"        0 ... numConditions - 1
     //                                     "full"             numConditions - 1
     //                                     "saturated"            numConditions
     //
     //                                   admissible values for format "active":
     //                        0 <= number of active conditions <= numConditions
     //
     //                                 admissible values for format "submodel":
     //  the number of active conditions in the submodel followed by an explicit
     // list of condition names.  Condition names must appear in ConditionNames,
     //           and must use different separators than those for specification
     //
     //                                       admissible value for format "key":
     //        string of 0/1, one for each condition, indicating inactive/active
     //                             lenght of this 0/1 string == # of conditions

     // The resulting submodels list, constructed using any of the functions, is
     //             always in an increasing order of number of active conditions

   SubmodelTable( const SubmodelTable& another );

   SubmodelTable( );

   void
   setNumConditions(       size_t          numConditions,
                     const vector<string>& conditionNames = vector<string>( ) );
                                   //
                                   // existing internal submodel table is purged
                                   //       if numConditions != numConditions( )

   void
   addSubmodel( const Submodel& sm );

   void
   delSubmodel( const Submodel& sm );
                                          //
                                          // 0 <= sm.size( ) <= numConditions( )
                                          //       0 <= sm[i] < numConditions( )

   void
   addSubmodel( size_t numActive );

   void
   delSubmodel( size_t numActive );
                                           //
                                           // 0 <= numActive <= numConditions( )

   void
   addByKey( const string& key );

   void
   delByKey( const string& key );
                      //
                      //                         key.size( ) == numConditions( )
                      // key[i] == '1' if condition i is active, '0' if inactive

   void
   addByName( const string& modelName );

   void
   delByName( const string& modelName );
                         //
                         // see constructor description for details of modelName

   void
   delAll( );
                                         //
                                         //                 delete all submodels

   void
   updateWith( const string& specification = "+ name all end",
               const string& separator     = " \t" );
                                      //
                                      // see constructor description for details

   // ------------------------------------------------- submodel list attributes
   // --------------------------------------------------------------------------

   inline
   size_t
   numConditions( ) const;

   inline
   const string&
   conditionName( size_t condition ) const;

   inline
   size_t
   size( ) const;

   // ----------------------------------------------------- submodeltable access
   // --------------------------------------------------------------------------

   inline
   const_iterator
   begin( ) const;

   inline
   const_iterator
   end( ) const;

   // ------------------------------------------------------ miscellaneous utils
   // --------------------------------------------------------------------------

   inline
   static
   Submodel
   submodel( const string& key );

   inline
   static
   size_t
   numActive( const string& key );

   inline
   static
   string
   key( const Submodel& sm, size_t numConditions );

   inline
   string
   key( const const_iterator& sm ) const;
                      //
                      // a binary key uniquely characterizing the given submodel
                      //
                      // assume that sm "belongs to" this SubmodelTable instance

   inline
   string
   description( const const_iterator& sm, char separator = '/' ) const;
                      //
                      //    submodel description is same as condition names list
                      //
                      // assume that sm "belongs to" this SubmodelTable instance

   // ------------------------------------------------------------------ private
   // --------------------------------------------------------------------------

  private:
   size_t         my_numConditions;
   vector<string> my_conditionNames;
   list<Submodel> my_smTable;

                                    // unary predicate for delSubmodel( size_t )
                                    //
   class hasActiveEq : public unary_function<Submodel,bool>
    {
     public:
      explicit
      hasActiveEq( size_t numActive )
       : my_numActive( numActive )
       {
       }

      bool
      operator()( const Submodel& sm ) const
       {
        return sm.size( ) == my_numActive;
       }

     private:
      size_t my_numActive;
    };

  SubmodelTable& operator=( const SubmodelTable& rhs );
 };

// ---------------------------------------------------- submodel list attributes
// -----------------------------------------------------------------------------

size_t
SubmodelTable::numConditions( ) const
 {
  return my_numConditions;
 }

const string&
SubmodelTable::conditionName( size_t condition ) const
 {
  return my_conditionNames[condition];
 }

size_t
SubmodelTable::size( ) const
 {
  return my_smTable.size( );
 }

// -------------------------------------------------------- submodeltable access
// -----------------------------------------------------------------------------

SubmodelTable::const_iterator
SubmodelTable::begin( ) const
 {
  return my_smTable.begin( );
 }

SubmodelTable::const_iterator
SubmodelTable::end( ) const
 {
  return my_smTable.end( );
 }

// --------------------------------------------------------- miscellaneous utils
// -----------------------------------------------------------------------------

string
SubmodelTable::key( const Submodel& sm, size_t numConditions )
 {
  string id( numConditions, '0' );
  for ( size_t a = 0; a < sm.size( ); a++ ) id[sm[a]] = '1';
  return id;
 }

string
SubmodelTable::key( const const_iterator& sm ) const
 {
  // ATTN any way of checking if sm "belongs to" my_smTable?

  return key( (*sm), my_numConditions );
 }

string
SubmodelTable::description( const const_iterator& sm, char separator ) const
 {
  // ATTN any way of checking if sm "belongs to" my_smTable?

  string d;

  if ( (*sm).size( ) == 0 )
   d = "(null)";
  else
   {
    string s( 1, separator );
    d = my_conditionNames[(*sm)[0]];
    for ( size_t a = 1; a < (*sm).size( ); a++ )
     {
      d.append( s ); d.append( my_conditionNames[(*sm)[a]] );
     }
   }

  return d;
 }

#endif /* __SubmodelTable_H__ */
