/* ----------------------------------------------------------------------------
 *
 * class acProfile
 *   BOLD activation profile
 *
 * required headers
 *   see include header file acProfile
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __acProfile_H__
#define __acProfile_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class acProfile
 {
  public:
   typedef SubmodelTable::Submodel Submodel;

   class ConditionBellSet
    {
     public:

      ConditionBellSet( const ConditionBellSet& another );

      ConditionBellSet&
      operator=( const ConditionBellSet& rhs );

      // ------------------------------------------------------------ attributes
      // -----------------------------------------------------------------------

      inline
      const string&
      name( ) const;

      inline
      const string&
      tag( ) const;

      inline
      const BasicBell*
      bell( ) const;

      inline
      const fpVector*
      timeGrid( ) const;

      inline
      size_t
      timeGridSize( ) const;

      inline
      size_t
      numShapeParams( ) const;

      inline
      int
      shapeOffset( ) const;                       // offset in the shape buffer;
                                                  //    -1 if condition inactive

      // ----------------------------------------------------------- epochs info
      // -----------------------------------------------------------------------

      inline
      size_t
      numEpochs( ) const;

      inline
      const fpVector&
      epochBegin( ) const;

      inline
      const fpVector&
      epochLength( ) const;

      inline
      fpType
      epochBegin( size_t epoch ) const;

      inline
      fpType
      epochLength( size_t epoch ) const;

      inline
      size_t
      beginIndex( size_t epoch, const fpType* shapeParams ) const;
                       //
                       // minimum i such that epochBegin( epoch ) <= timeGrid[i]
                       //
                       // these are  precomputed assuming that the lower support
                       // of a bell corresponding to an epoch is the epoch begin
                       // time itself,  and does not depend on  shape parameters

      inline
      size_t
      endIndex( size_t epoch, const fpType* shapeParams ) const;
                       //
                       // maximum i  such that timeGrid[i]  <=  upper support of
                       //                   the bell given epoch and shapeParams

     protected:

      // -------------------------------------------- over the time grid: form 1
      // -----------------------------------------------------------------------

                                       //        result.size == timeGridSize( );
                                       // calculated values are added to result;
                                       //    shape param indices with respect to
                                       //  the complete shape buffer shapeParams

      inline
      void
      value( const fpType*   shapeParams,
                   fpVector& result,
                   fpType    multiplier = 1.0 ) const;

      inline
      void
      Dt( const fpType*   shapeParams,
                fpVector& result,
                fpType    multiplier = 1.0,
                size_t    dOrder     = 1 ) const;

      inline
      void
      gradient( const fpType*   shapeParams,
                      size_t    paramIndex,
                      fpVector& result,
                      fpType    multiplier = 1.0 ) const;

      inline
      void
      hessian( const fpType*   shapeParams,
                     size_t    paramIndex1,
                     size_t    paramIndex2,
                     fpVector& result,
                     fpType    multiplier = 1.0 ) const;

      // -------------------------------------------- over the time grid: form 2
      // -----------------------------------------------------------------------

                                        // calculated values are added to result

      inline
      void
      value( const fpType* shapeParams,
                   fpType* result,
                   fpType  multiplier = 1.0 ) const;

      inline
      void
      Dt( const fpType* shapeParams,
                fpType* result,
                fpType  multiplier = 1.0,
                size_t  dOrder     = 1 ) const;

      inline
      void
      gradient( const fpType* shapeParams,
                      size_t  paramIndex,
                      fpType* result,
                      fpType  multiplier = 1.0 ) const;

      inline
      void
      hessian( const fpType* shapeParams,
                     size_t  paramIndex1,
                     size_t  paramIndex2,
                     fpType* result,
                     fpType  multiplier = 1.0 ) const;

      // ---------------------------------------------------- at arbitrary times
      // -----------------------------------------------------------------------

      inline
      fpType
      value( const fpType* shapeParams, fpType time ) const;

      inline
      fpType
      Dt( const fpType* shapeParams, fpType time, size_t dOrder = 1 ) const;

      inline
      fpType
      gradient( const fpType* shapeParams, size_t paramIndex, fpType time ) const;

      inline
      fpType
      hessian( const fpType* shapeParams,
                     size_t  paramIndex1,
                     size_t  paramIndex2,
                     fpType  time         ) const;

     private:
            string     my_name, my_tag;
            int        my_shapeOffset;

      const BasicBell* my_bell;
      const fpVector*  my_timeGrid;

            fpVector   my_epochBegin;
            fpVector   my_epochLength;

            uintVector  my_beginIndex;
                                        //
                                        // timeGrid indices for epoch beginnings
                                        //  must be recalculated if the external
                                        //                      timeGrid changes
                                        //
                                        //   this assumes that the lower support
                                        //   of a bell corresponding to an epoch
                                        //   is the epoch begin time itself, and
                                        //   does not depend on shape parameters

      friend class acProfile;

      // ------------------------- construct/reinitialize/incremental initialize
      // -----------------------------------------------------------------------

      ConditionBellSet( const string& name, const string& tag );

      ConditionBellSet( );

      void
      setShapeOffset( int shapeOffset );

      void
      setName( const string& name );

      void
      setTag( const string& tag );

      void
      setBell( const BasicBell* const bell );

      void
      setEpochData( const fpVector& epochBegin, const fpVector& epochLength );
                                   //
                                   // epochBegin.size( ) ==  epochLenght.size( )
                                   //     == number of epochs for this condition

      void
      setTimeGrid( const fpVector* timeGrid, bool init = false );
                                   //
                                   //   time grid values are assumed to be in an
                                   //                           increasing order

      void
      initialize( bool init = true );
    };

   // ---------------------------- construct/reinitialize/incremental initialize
   // --------------------------------------------------------------------------

   acProfile( const acProfile& another );

   acProfile&
   operator=( const acProfile& rhs );

   acProfile( );

   void
   setDBprofile( const dbProfile* const db );

   void
   setSubmodel( const Submodel* const sm );

   void
   setTimeGrid( const fpVector* const timeGrid, bool init = false );

   void
   addCondition( const string&          name,
                 const fpVector&        epochBegin,
                 const fpVector&        epochLength,
                 const string&          tag,
                 const BasicBell* const bell,
                       bool             init = false );

   // void
   // delByName( const string& name );

   // void
   // delByTag( const string& tag );

   void
   setBell( const string&          tag,
            const BasicBell* const bell,
                  bool             init = false );
                               //
                               //  set bell function for all conditions with the
                               //  given tag.  All conditions  with the same tag
                               // share the same set of shape parameters and the
                               //                                  bell function

   void
   initialize( bool init = true );
                               //
                               // performs  all  initializations  assuming  that
                               // time grid, conditions, epochs, tags,  submodel
                               // have all been set appropriately.

   // -------------------------------------------- time grid and parameter sizes
   // --------------------------------------------------------------------------

   inline
   size_t
   timeGridSize( ) const;

   inline
   size_t
   numDBparams( ) const;

   inline
   size_t
   numConditions( ) const;                         // total number of conditions

   inline
   bool
   isActive( size_t condition ) const;      // 0 <= condition < numConditions( )

   inline
   const ConditionBellSet&
   bellSet( size_t condition ) const;       // 0 <= condition < numConditions( )

   inline
   const Submodel&
   submodel( ) const;

   inline
   size_t
   numActive( ) const;        // number of active conditions == size of submodel

   inline
   size_t
   numRespParams( ) const;                               // same as numActive( )

   inline
   size_t
   condition( size_t activeIndex ) const;
                                   //
                                   //       condition corresponding to the given
                                   // active condition index in current submodel
                                   //
                                   //            0 <= activeIndex < numActive( )

   inline
   size_t
   numShapeParams( ) const;
                              //
                              // total number of shape parameters for all active
                              //     conditions together in the current submodel

   inline
   size_t
   numShapeParams( const Submodel& sm ) const;
                              //
                              // total number of shape parameters for all active
                              //       conditions together in the given submodel

   // --------------------------------------------------- external drift profile
   // --------------------------------------------------------------------------

   inline
   const dbProfile&
   drift( ) const;

   // ---------------------------- profile+derivatives: entire time grid; form 1
   // --------------------------------------------------------------------------

   inline
   fpVector
   profile( const fpType* dbParams,
            const fpType* respParams,
            const fpType* shapeParams ) const;
                                                   //
                                                   // profile over the time grid

   inline
   fpVector
   DprofileT( const fpType* dbParams,
              const fpType* respParams,
              const fpType* shapeParams,
                    size_t  dOrder = 1 ) const;
                                    //
                                    // derivative w.r.t. time over the time grid

   inline
   fpVector
   DprofileD( const fpType* dbParams,
              const fpType* respParams,
              const fpType* shapeParams,
                    size_t  paramIndex ) const;
                           //
                           // gradient w.r.t. the given drift+baseline parameter
                           //                                 over the time grid

   inline
   fpVector
   DprofileR( const fpType* dbParams,
              const fpType* respParams,
              const fpType* shapeParams,
                    size_t  paramIndex ) const;
                           //
                           // gradient w.r.t. the given responsiveness parameter
                           //                                 over the time grid
                           //
                           // 0 < paramIndexR <= numActive( ); corresponds to an
                           // active condition (i. e., in the current submodel).

   inline
   fpVector
   DprofileS( const fpType* dbParams,
              const fpType* respParams,
              const fpType* shapeParams,
                    size_t  paramIndex ) const;
                                    //
                                    // gradient w.r.t. the given shape parameter
                                    //                        over the time grid

   inline
   fpVector
   DDprofileDD( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      size_t  paramIndex1,
                      size_t  paramIndex2 ) const;
                       //
                       //     hessian w.r.t. the given drift+baseline parameters
                       //                                     over the time grid
                       //
                       //  note: this is zero if drift+baseline is linear in the
                       //   parameters, which is assumed in this implementation.

   inline
   fpVector
   DDprofileRR( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      size_t  paramIndex1,
                      size_t  paramIndex2 ) const;
                         //
                         //   hessian w.r.t. the given responsiveness parameters
                         //                                   over the time grid
                         //
                         // note: in the present formulation and implementation,
                         //   activation is linear in responsiveness parameters.
                         //  Hence zero second derivatives w.r.t. responsiveness
                         //                                           parameters

   inline
   fpVector
   DDprofileSS( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      size_t  paramIndex1,
                      size_t  paramIndex2 ) const;
                                  //
                                  //   hessian w.r.t. the given shape parameters
                                  //                          over the time grid

   inline
   fpVector
   DDprofileDR( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      size_t  paramIndexD,
                      size_t  paramIndexR ) const;
                           //
                           //         mixed second partial derivative w.r.t. the
                           // given drift+baseline and responsiveness parameters
                           //                                 over the time grid

   inline
   fpVector
   DDprofileDS( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      size_t  paramIndexD,
                      size_t  paramIndexS ) const;
                           //
                           //         mixed second partial derivative w.r.t. the
                           //          given drift+baseline and shape parameters
                           //                                 over the time grid

   inline
   fpVector
   DDprofileRS( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      size_t  paramIndexR,
                      size_t  paramIndexS ) const;
                           //
                           //         mixed second partial derivative w.r.t. the
                           //          given responsiveness and shape parameters
                           //                                 over the time grid

   // ---------------------------- profile+derivatives: entire time grid; form 2
   // --------------------------------------------------------------------------

   inline
   void
   getProfile( const fpType*   dbParams,
               const fpType*   respParams,
               const fpType*   shapeParams,
                     fpVector& p ) const;
                                                   //
                                                   // profile over the time grid

   inline
   void
   getDprofileT( const fpType*   dbParams,
                 const fpType*   respParams,
                 const fpType*   shapeParams,
                       fpVector& Dp,
                       size_t    dOrder = 1 ) const;
                                    //
                                    // derivative w.r.t. time over the time grid

   inline
   void
   getDprofileD( const fpType*   dbParams,
                 const fpType*   respParams,
                 const fpType*   shapeParams,
                       size_t    paramIndex,
                       fpVector& Dp ) const;
                           //
                           // gradient w.r.t. the given drift+baseline parameter
                           //                                 over the time grid

   inline
   void
   getDprofileR( const fpType*   dbParams,
                 const fpType*   respParams,
                 const fpType*   shapeParams,
                       size_t    paramIndex,
                       fpVector& Dp ) const;
                           //
                           // gradient w.r.t. the given responsiveness parameter
                           //                                 over the time grid

   inline
   void
   getDprofileS( const fpType*   dbParams,
                 const fpType*   respParams,
                 const fpType*   shapeParams,
                       size_t    paramIndex,
                       fpVector& Dp ) const;
                                    //
                                    // gradient w.r.t. the given shape parameter
                                    //                        over the time grid

   inline
   void
   getDDprofileDD( const fpType*   dbParams,
                   const fpType*   respParams,
                   const fpType*   shapeParams,
                         size_t    paramIndex1,
                         size_t    paramIndex2,
                         fpVector& DDp ) const;
                        //
                        //    hessian w.r.t. the given drift+baseline parameters
                        //                                    over the time grid
                        //
                        // note: this is zero if drift+baseline is linear in the
                        //  parameters, which is assumed in this implementation.

   inline
   void
   getDDprofileRR( const fpType*   dbParams,
                   const fpType*   respParams,
                   const fpType*   shapeParams,
                         size_t    paramIndex1,
                         size_t    paramIndex2,
                         fpVector& DDp ) const;
                         //
                         //   hessian w.r.t. the given responsiveness parameters
                         //                                   over the time grid
                         //
                         // note: in the present formulation and implementation,
                         //   activation is linear in responsiveness parameters.
                         //  Hence zero second derivatives w.r.t. responsiveness
                         //                                           parameters

   inline
   void
   getDDprofileSS( const fpType*   dbParams,
                   const fpType*   respParams,
                   const fpType*   shapeParams,
                         size_t    paramIndex1,
                         size_t    paramIndex2,
                         fpVector& DDp ) const;
                                    //
                                    // hessian w.r.t. the given shape parameters
                                    //                        over the time grid

   inline
   void
   getDDprofileDR( const fpType*   dbParams,
                   const fpType*   respParams,
                   const fpType*   shapeParams,
                         size_t    paramIndexD,
                         size_t    paramIndexR,
                         fpVector& DDp ) const;
                           //
                           //         mixed second partial derivative w.r.t. the
                           // given drift+baseline and responsiveness parameters
                           //                                 over the time grid

   inline
   void
   getDDprofileDS( const fpType*   dbParams,
                   const fpType*   respParams,
                   const fpType*   shapeParams,
                         size_t    paramIndexD,
                         size_t    paramIndexS,
                         fpVector& DDp ) const;
                                   //
                                   // mixed second partial derivative w.r.t. the
                                   //  given drift+baseline and shape parameters
                                   //                         over the time grid

   inline
   void
   getDDprofileRS( const fpType*   dbParams,
                   const fpType*   respParams,
                   const fpType*   shapeParams,
                         size_t    paramIndexR,
                         size_t    paramIndexS,
                         fpVector& DDprofileRS ) const;
                                   //
                                   // mixed second partial derivative w.r.t. the
                                   //  given responsiveness and shape parameters
                                   //                         over the time grid

   // ------------------------------------- profile+derivatives: arbitrary times
   // --------------------------------------------------------------------------

   inline
   fpType
   profile( const fpType* dbParams,
            const fpType* respParams,
            const fpType* shapeParams,
                  fpType  time  ) const;
                                                  //
                                                  // profile value at given time

   inline
   fpType
   DprofileT( const fpType* dbParams,
              const fpType* respParams,
              const fpType* shapeParams,
                    fpType  time,
                    size_t  dOrder = 1 ) const;
                                     //
                                     // derivative w.r.t. time at the given time

   inline
   fpType
   DprofileD( const fpType* dbParams,
              const fpType* respParams,
              const fpType* shapeParams,
                    fpType  time,
                    size_t  paramIndex ) const;
                           //
                           // gradient w.r.t. the given drift+baseline parameter
                           //                                  at the given time

   inline
   fpType
   DprofileR( const fpType* dbParams,
              const fpType* respParams,
              const fpType* shapeParams,
                    fpType  time,
                    size_t  paramIndex ) const;
                           //
                           // gradient w.r.t. the given responsiveness parameter
                           //                                  at the given time

   inline
   fpType
   DprofileS( const fpType* dbParams,
              const fpType* respParams,
              const fpType* shapeParams,
                    fpType  time,
                    size_t  paramIndex ) const;
                                    //
                                    // gradient w.r.t. the given shape parameter
                                    //                         at the given time

   inline
   fpType
   DDprofileDD( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      fpType  time,
                      size_t  paramIndex1,
                      size_t  paramIndex2 ) const;
                       //
                       //     hessian w.r.t. the given drift+baseline parameters
                       //                                      at the given time
                       //
                       //  note: this is zero if drift+baseline is linear in the
                       //   parameters, which is assumed in this implementation.
   inline
   fpType
   DDprofileRR( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      fpType  time,
                      size_t  paramIndex1,
                      size_t  paramIndex2 ) const;
                         //
                         //   hessian w.r.t. the given responsiveness parameters
                         //                                    at the given time
                         //
                         // note: in the present formulation and implementation,
                         //   activation is linear in responsiveness parameters.
                         //  Hence zero second derivatives w.r.t. responsiveness
                         //                                           parameters

   inline
   fpType
   DDprofileSS( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      fpType  time,
                      size_t  paramIndex1,
                      size_t  paramIndex2 ) const;
                                  //
                                  //   hessian w.r.t. the given shape parameters
                                  //                           at the given time

   inline
   fpType
   DDprofileDR( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      fpType  time,
                      size_t  paramIndexD,
                      size_t  paramIndexR ) const;
                           //
                           //         mixed second partial derivative w.r.t. the
                           // given drift+baseline and responsiveness parameters
                           //                                  at the given time

   inline
   fpType
   DDprofileDS( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      fpType  time,
                      size_t  paramIndexD,
                      size_t  paramIndexS ) const;
                           //
                           //         mixed second partial derivative w.r.t. the
                           //          given drift+baseline and shape parameters
                           //                                  at the given time

   inline
   fpType
   DDprofileRS( const fpType* dbParams,
                const fpType* respParams,
                const fpType* shapeParams,
                      fpType  time,
                      size_t  paramIndexR,
                      size_t  paramIndexS ) const;
                           //
                           //         mixed second partial derivative w.r.t. the
                           //          given responsiveness and shape parameters
                           //                                  at the given time

   // ------------------------------------------------ activation "basis matrix"
   // --------------------------------------------------------------------------

                     // bell values for all active conditions over the time grid

   inline
   fpMatrix
   basisMatrix( const fpType* shapeParams ) const;

   inline
   void
   getBasisMatrix( const fpType* shapeParams, fpMatrix& basisMatrix ) const;

   // ------------------------------------------------------------------ private
   // --------------------------------------------------------------------------

  private:
   const dbProfile*               my_db;
   const Submodel*                my_sm;
         vector<ConditionBellSet> my_cl;
         // vector<string>           my_sParam2cTag;
                                        //
                                        // shape parameters -> condition tag map
 };

// ------------------------------------------------ ConditionBellSet: attributes
// -----------------------------------------------------------------------------

const string&
acProfile::ConditionBellSet::name( ) const
 {
  return my_name;
 }

const string&
acProfile::ConditionBellSet::tag( ) const
 {
  return my_tag;
 }

const BasicBell*
acProfile::ConditionBellSet::bell( ) const
 {
  return my_bell;
 }

const fpVector*
acProfile::ConditionBellSet::timeGrid( ) const
 {
  return my_timeGrid;
 }

size_t
acProfile::ConditionBellSet::timeGridSize( ) const
 {
  return my_timeGrid->size( );
 }

size_t
acProfile::ConditionBellSet::numShapeParams( ) const
 {
  return my_bell->numParams( );
 }

int
acProfile::ConditionBellSet::shapeOffset( ) const
 {
  return my_shapeOffset;
 }

const dbProfile&
acProfile::drift( ) const
 {
  return *my_db;
 }

// ----------------------------------------------- ConditionBellSet: epochs info
// -----------------------------------------------------------------------------

size_t
acProfile::ConditionBellSet::numEpochs( ) const
 {
  return my_epochBegin.size( );
 }

const fpVector&
acProfile::ConditionBellSet::epochBegin( ) const
 {
  return my_epochBegin;
 }

const fpVector&
acProfile::ConditionBellSet::epochLength( ) const
 {
  return my_epochLength;
 }

fpType
acProfile::ConditionBellSet::epochBegin( size_t epoch ) const
 {
  return my_epochBegin[epoch];
 }

fpType
acProfile::ConditionBellSet::epochLength( size_t epoch ) const
 {
  return my_epochLength[epoch];
 }

size_t
acProfile::ConditionBellSet::beginIndex(       size_t  epoch,
                                         const fpType* shapeParams ) const
 {
  return my_beginIndex[epoch];
 }

size_t
acProfile::ConditionBellSet::endIndex(       size_t  epoch,
                                       const fpType* shapeParams ) const
 {
  fpType upperSupport
   = my_bell->upperSupport( my_epochBegin[epoch],
                            my_epochLength[epoch], shapeParams + my_shapeOffset );

  size_t iMax = timeGridSize( ) - 1, i = my_beginIndex[epoch];
  while ( ( i < iMax ) && ( (*my_timeGrid)[i] < upperSupport ) ) i++;

  return i;
 }

// -------------------------------- ConditionBellSet: over the time grid: form 1
// -----------------------------------------------------------------------------

void
acProfile::ConditionBellSet::value( const fpType*   shapeParams,
                                          fpVector& result,
                                          fpType    multiplier ) const
 {
  value( shapeParams, result( 0 ), multiplier );
 }

void
acProfile::ConditionBellSet::Dt( const fpType*   shapeParams,
                                       fpVector& result,
                                       fpType    multiplier,
                                       size_t    dOrder ) const
 {
  Dt( shapeParams, result( 0 ), multiplier, dOrder );
 }

void
acProfile::ConditionBellSet::gradient( const fpType*   shapeParams,
                                             size_t    paramIndex,
                                             fpVector& result,
                                             fpType    multiplier ) const
 {
  gradient( shapeParams, paramIndex, result( 0 ), multiplier );
 }

void
acProfile::ConditionBellSet::hessian( const fpType*   shapeParams,
                                            size_t    paramIndex1,
                                            size_t    paramIndex2,
                                            fpVector& result,
                                            fpType    multiplier ) const
 {
  hessian( shapeParams, paramIndex1, paramIndex2, result( 0 ), multiplier );
 }

// -------------------------------- ConditionBellSet: over the time grid: form 2
// -----------------------------------------------------------------------------

void
acProfile::ConditionBellSet::value( const fpType* shapeParams,
                                          fpType* result,
                                          fpType  multiplier ) const
 {
  if ( multiplier == 0.0 ) return;

  const fpType* s = shapeParams + my_shapeOffset;

  size_t iMax;
  for ( size_t e = 0; e < numEpochs( ); e++ )
   {
    iMax = endIndex( e, shapeParams );
    for ( size_t i = my_beginIndex[e]; i <= iMax; i++ )
     result[i]
      +=   multiplier
         * (*my_bell)( (*my_timeGrid)[i],
                       my_epochBegin[e], my_epochLength[e], s );
   }
 }

void
acProfile::ConditionBellSet::Dt( const fpType* shapeParams,
                                       fpType* result,
                                       fpType  multiplier,
                                       size_t  dOrder ) const
 {
  if ( dOrder == 0 )
   (*this).value( shapeParams, result, multiplier );
  else
   {
    if ( multiplier == 0.0 ) return;

    const fpType* s = shapeParams + my_shapeOffset;

    size_t iMax;
    for ( size_t e = 0; e < numEpochs( ); e++ )
     {
      iMax = endIndex( e, shapeParams );
      for ( size_t i = my_beginIndex[e]; i <= iMax; i++ )
       result[i]
        +=   multiplier
           * my_bell->Dt( (*my_timeGrid)[i],
                          my_epochBegin[e], my_epochLength[e], s, dOrder );
     }
   }
 }

void
acProfile::ConditionBellSet::gradient( const fpType* shapeParams,
                                             size_t  paramIndex,
                                             fpType* result,
                                             fpType  multiplier ) const
 {
  if ( multiplier == 0.0 ) return;

  int pi = paramIndex - my_shapeOffset;
  if ( ( pi < 0 ) || ( pi >= numShapeParams( ) ) ) return;

  const fpType* s = shapeParams + my_shapeOffset;

  size_t iMax;
  for ( size_t e = 0; e < numEpochs( ); e++ )
   {
    iMax = endIndex( e, shapeParams );
    for ( size_t i = my_beginIndex[e]; i <= iMax; i++ )
     result[i]
      +=   multiplier
         * my_bell->gradient( (*my_timeGrid)[i], my_epochBegin[e], my_epochLength[e], s, pi );
   }
 }

void
acProfile::ConditionBellSet::hessian( const fpType* shapeParams,
                                            size_t  paramIndex1,
                                            size_t  paramIndex2,
                                            fpType* result,
                                            fpType  multiplier ) const
 {
  if ( multiplier == 0.0 ) return;

  int pi1 = paramIndex1 - my_shapeOffset;
  if ( ( pi1 < 0 ) || ( pi1 >= numShapeParams( ) ) ) return;

  int pi2 = paramIndex2 - my_shapeOffset;
  if ( ( pi2 < 0 ) || ( pi2 >= numShapeParams( ) ) ) return;

  const fpType* s = shapeParams + my_shapeOffset;

  size_t iMax;
  for ( size_t e = 0; e < numEpochs( ); e++ )
   {
    iMax = endIndex( e, shapeParams );
    for ( size_t i = my_beginIndex[e]; i <= iMax; i++ )
     result[i]
      +=   multiplier
         * my_bell->hessian( (*my_timeGrid)[i], my_epochBegin[e], my_epochLength[e], s, pi1, pi2 );
   }
 }

// ---------------------------------------- ConditionBellSet: at arbitrary times
// -----------------------------------------------------------------------------

fpType
acProfile::ConditionBellSet::value( const fpType* shapeParams,
                                          fpType  time ) const
 {
  const fpType* s = shapeParams + my_shapeOffset;

  fpType result = 0.0;
  for ( size_t e = 0; e < numEpochs( ); e++ )
   result += (*my_bell)( time, my_epochBegin[e], my_epochLength[e], s );

  return result;
 }

fpType
acProfile::ConditionBellSet::Dt( const fpType* shapeParams,
                                       fpType  time,
                                       size_t  dOrder ) const
 {
  if ( dOrder == 0 )
   return (*this).value( shapeParams, time );
  else
   {
    const fpType* s = shapeParams + my_shapeOffset;

    fpType result = 0.0;
    for ( size_t e = 0; e < numEpochs( ); e++ )
     result += my_bell->Dt( time, my_epochBegin[e], my_epochLength[e], s, dOrder );

    return result;
   }
 }

fpType
acProfile::ConditionBellSet::gradient( const fpType* shapeParams,
                                             size_t  paramIndex,
                                             fpType  time ) const
 {
  int pi = paramIndex - my_shapeOffset;
  if ( ( pi < 0 ) || ( pi >= numShapeParams( ) ) ) return 0.0;

  const fpType* s = shapeParams + my_shapeOffset;

  fpType result = 0.0;
  for ( size_t e = 0; e < numEpochs( ); e++ )
   result += my_bell->gradient( time, my_epochBegin[e], my_epochLength[e], s, pi );

  return result;
 }

fpType
acProfile::ConditionBellSet::hessian( const fpType* shapeParams,
                                            size_t  paramIndex1,
                                            size_t  paramIndex2,
                                            fpType  time ) const
 {
  int pi1 = paramIndex1 - my_shapeOffset;
  if ( ( pi1 < 0 ) || ( pi1 >= numShapeParams( ) ) ) return 0.0;

  int pi2 = paramIndex2 - my_shapeOffset;
  if ( ( pi2 < 0 ) || ( pi2 >= numShapeParams( ) ) ) return 0.0;

  const fpType* s = shapeParams + my_shapeOffset;

  fpType result = 0.0;
  for ( size_t e = 0; e < numEpochs( ); e++ )
   result += my_bell->hessian( time, my_epochBegin[e], my_epochLength[e], s, pi1, pi2 );

  return result;
 }

// ----------------------------------------------- time grid and parameter sizes
// -----------------------------------------------------------------------------

size_t
acProfile::timeGridSize( ) const
 {
  return ( my_cl.size( ) > 0 ) ? my_cl[0].timeGridSize( ) : 0;
 }

size_t
acProfile::numDBparams( ) const
 {
  return ( my_db != 0 ) ? my_db->basisSize( ) : 0;
 }

size_t
acProfile::numConditions( ) const
 {
  return my_cl.size( );
 }

bool
acProfile::isActive( size_t condition ) const
 {
  bool found = false;
  for ( size_t a = 0; ( a < numActive( ) ) && !found; a++ )
   found = found || ( condition == (*this).condition( a ) );

  return found;
 }

const acProfile::ConditionBellSet&
acProfile::bellSet( size_t condition ) const
 {
  return my_cl[condition];
 }

const acProfile::Submodel&
acProfile::submodel( ) const
 {
  return *my_sm;
 }

size_t
acProfile::numRespParams( ) const
 {
  return numActive( );
 }

size_t
acProfile::numActive( ) const
 {
  return ( my_sm != 0 ) ? my_sm->size( ) : 0;
 }

size_t
acProfile::condition( size_t activeIndex ) const
 {
  return (*my_sm)[activeIndex];
 }

size_t
acProfile::numShapeParams( ) const
 {
  return ( ( my_sm == 0 ) ? 0 : numShapeParams( *my_sm ) );
 }

size_t
acProfile::numShapeParams( const Submodel& submodel ) const
 {
  size_t n = 0;
  bool found;
  for ( size_t active = 0; active < submodel.size( ); active++ )
   {
    size_t c = submodel[active];
    const string& t = my_cl[c].tag( );

    found = false;
    for ( size_t a = 0; ( a < active ) && !found; a++ )
     found = found || ( t == my_cl[submodel[a]].tag( ) );

    if ( !found ) n += my_cl[c].numShapeParams( );
   }

  return n;
 }

// ------------------------------- profile+derivatives: entire time grid; form 1
// -----------------------------------------------------------------------------

fpVector
acProfile::profile( const fpType* dbParams,
                    const fpType* respParams,
                    const fpType* shapeParams ) const
 {
  fpVector p;
  getProfile( dbParams, respParams, shapeParams, p );

  return p;
 }

fpVector
acProfile::DprofileT( const fpType* dbParams,
                      const fpType* respParams,
                      const fpType* shapeParams,
                            size_t  dOrder ) const
 {
  fpVector p;
  getDprofileT( dbParams, respParams, shapeParams, p, dOrder );

  return p;
 }

fpVector
acProfile::DprofileD( const fpType* dbParams,
                      const fpType* respParams,
                      const fpType* shapeParams,
                            size_t  paramIndex ) const
 {
  fpVector p;
  getDprofileD( dbParams, respParams, shapeParams, paramIndex, p );

  return p;
 }

fpVector
acProfile::DprofileR( const fpType* dbParams,
                      const fpType* respParams,
                      const fpType* shapeParams,
                            size_t  paramIndex ) const
 {
  fpVector p;
  getDprofileR( dbParams, respParams, shapeParams, paramIndex, p );

  return p;
 }

fpVector
acProfile::DprofileS( const fpType* dbParams,
                      const fpType* respParams,
                      const fpType* shapeParams,
                            size_t  paramIndex ) const
 {
  fpVector p;
  getDprofileS( dbParams, respParams, shapeParams, paramIndex, p );

  return p;
 }

fpVector
acProfile::DDprofileDD( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              size_t  paramIndex1,
                              size_t  paramIndex2 ) const
 {
  fpVector p;
  getDDprofileDD( dbParams, respParams, shapeParams,
                  paramIndex1, paramIndex2, p );

  return p;
 }

fpVector
acProfile::DDprofileRR( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              size_t  paramIndex1,
                              size_t  paramIndex2 ) const
 {
  fpVector p;
  getDDprofileRR( dbParams, respParams, shapeParams,
                  paramIndex1, paramIndex2, p );

  return p;
 }

fpVector
acProfile::DDprofileSS( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              size_t  paramIndex1,
                              size_t  paramIndex2 ) const
 {
  fpVector p;
  getDDprofileSS( dbParams, respParams, shapeParams,
                  paramIndex1, paramIndex2, p );

  return p;
 }

fpVector
acProfile::DDprofileDR( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              size_t  paramIndexD,
                              size_t  paramIndexR ) const
 {
  fpVector p;
  getDDprofileDR( dbParams, respParams, shapeParams,
                  paramIndexD, paramIndexR, p );

  return p;
 }

fpVector
acProfile::DDprofileDS( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              size_t  paramIndexD,
                              size_t  paramIndexS ) const
 {
  fpVector p;
  getDDprofileDS( dbParams, respParams, shapeParams,
                  paramIndexD, paramIndexS, p );

  return p;
 }

fpVector
acProfile::DDprofileRS( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              size_t  paramIndexR,
                              size_t  paramIndexS ) const
 {
  fpVector p;
  getDDprofileRS( dbParams, respParams, shapeParams,
                  paramIndexR, paramIndexS, p );

  return p;
 }

// ------------------------------- profile+derivatives: entire time grid; form 2
// -----------------------------------------------------------------------------

void
acProfile::getProfile( const fpType*   dbParams,
                       const fpType*   respParams,
                       const fpType*   shapeParams,
                             fpVector& p  ) const
 {
  p.resize( timeGridSize( ), 0.0 );

  fpType b = my_db->baseline( dbParams );

  for ( size_t a = 0; a < numActive( ); a++ )
   my_cl[condition( a )].value( shapeParams, p, b * respParams[a] );
 }

void
acProfile::getDprofileT( const fpType*   dbParams,
                         const fpType*   respParams,
                         const fpType*   shapeParams,
                               fpVector& Dp,
                               size_t    dOrder ) const
 {
  if ( dOrder <= 0 )
   (*this).getProfile( dbParams, respParams, shapeParams, Dp );
  else
   {
    Dp.resize( timeGridSize( ), 0.0 );

    fpType b = my_db->baseline( dbParams );

    for ( size_t a = 0; a < numActive( ); a++ )
     my_cl[condition( a )].Dt( shapeParams, Dp, b * respParams[a], dOrder );
   }
 }

void
acProfile::getDprofileD( const fpType*   dbParams,
                         const fpType*   respParams,
                         const fpType*   shapeParams,
                               size_t    paramIndex,
                               fpVector& Dp ) const
 {
  Dp.resize( timeGridSize( ), 0.0 );

  if ( paramIndex >= my_db->basisSize( ) ) return;

  fpType Db = my_db->Dbaseline( dbParams, paramIndex );

  for ( size_t a = 0; a < numActive( ); a++ )
   my_cl[condition( a )].value( shapeParams, Dp, Db * respParams[a] );
 }

void
acProfile::getDprofileR( const fpType*   dbParams,
                         const fpType*   respParams,
                         const fpType*   shapeParams,
                               size_t    paramIndex,
                               fpVector& Dp ) const
 {
  Dp.resize( timeGridSize( ), 0.0 );

  if ( paramIndex >= numActive( ) ) return;

  my_cl[condition( paramIndex )].value( shapeParams, Dp,
                                        my_db->baseline( dbParams ) );
 }

void
acProfile::getDprofileS( const fpType*   dbParams,
                         const fpType*   respParams,
                         const fpType*   shapeParams,
                               size_t    paramIndex,
                               fpVector& Dp ) const
 {
  Dp.resize( timeGridSize( ), 0.0 );

  if ( paramIndex >= numShapeParams( ) ) return;

  fpType b = my_db->baseline( dbParams );

  for ( size_t a = 0; a < numActive( ); a++ )
   my_cl[condition( a )].gradient( shapeParams, paramIndex, Dp, b * respParams[a] );
 }

void
acProfile::getDDprofileDD( const fpType*   dbParams,
                           const fpType*   respParams,
                           const fpType*   shapeParams,
                                 size_t    paramIndex1,
                                 size_t    paramIndex2,
                                 fpVector& DDp ) const
 {
  DDp.resize( timeGridSize( ), 0.0 );
 }

void
acProfile::getDDprofileRR( const fpType*   dbParams,
                           const fpType*   respParams,
                           const fpType*   shapeParams,
                                 size_t    paramIndex1,
                                 size_t    paramIndex2,
                                 fpVector& DDp ) const
 {
  DDp.resize( timeGridSize( ), 0.0 );
 }

void
acProfile::getDDprofileSS( const fpType*   dbParams,
                           const fpType*   respParams,
                           const fpType*   shapeParams,
                                 size_t    paramIndex1,
                                 size_t    paramIndex2,
                                 fpVector& DDp ) const
 {
  DDp.resize( timeGridSize( ), 0.0 );

  if (    ( paramIndex1 >= numShapeParams( ) )
       || ( paramIndex2 >= numShapeParams( ) ) ) return;

  fpType b = my_db->baseline( dbParams );

  for ( size_t a = 0; a < numActive( ); a++ )
   my_cl[condition( a )].hessian( shapeParams, paramIndex1, paramIndex2, DDp, b * respParams[a] );
 }

void
acProfile::getDDprofileDR( const fpType*   dbParams,
                           const fpType*   respParams,
                           const fpType*   shapeParams,
                                 size_t    paramIndexD,
                                 size_t    paramIndexR,
                                 fpVector& DDp ) const
 {
  DDp.resize( timeGridSize( ), 0.0 );

  if (    ( paramIndexR >= numActive( )        )
       || ( paramIndexD >= my_db->basisSize( ) ) ) return;

  my_cl[condition( paramIndexR )].value( shapeParams, DDp,
                                         my_db->Dbaseline( dbParams, paramIndexD ) );
 }

void
acProfile::getDDprofileDS( const fpType*   dbParams,
                           const fpType*   respParams,
                           const fpType*   shapeParams,
                                 size_t    paramIndexD,
                                 size_t    paramIndexS,
                                 fpVector& DDp ) const
 {
  DDp.resize( timeGridSize( ), 0.0 );

  if (    ( paramIndexS >= numShapeParams( )   )
       || ( paramIndexD >= my_db->basisSize( ) ) ) return;

  fpType Db = my_db->Dbaseline( dbParams, paramIndexD );

  for ( size_t a = 0; a < numActive( ); a++ )
   my_cl[condition( a )].gradient( shapeParams, paramIndexS, DDp, Db * respParams[a] );
 }

void
acProfile::getDDprofileRS( const fpType*   dbParams,
                           const fpType*   respParams,
                           const fpType*   shapeParams,
                                 size_t    paramIndexR,
                                 size_t    paramIndexS,
                                 fpVector& DDp ) const
 {
  DDp.resize( timeGridSize( ), 0.0 );

  if (    ( paramIndexR >= numActive( )        )
       || ( paramIndexS >= numShapeParams( ) ) ) return;

  my_cl[condition( paramIndexR )].gradient( shapeParams, paramIndexS, DDp,
                                              my_db->baseline( dbParams )
                                            * respParams[paramIndexR] );
 }

// ---------------------------------------- profile+derivatives: arbitrary times
// -----------------------------------------------------------------------------

fpType
acProfile::profile( const fpType* dbParams,
                    const fpType* respParams,
                    const fpType* shapeParams,
                          fpType  time ) const
 {
  fpType result = 0.0;
  for ( size_t a = 0; a < numActive( ); a++ )
   result +=   respParams[a]
             * my_cl[condition( a )].value( shapeParams, time );

  return my_db->baseline( dbParams ) * result;
 }

fpType
acProfile::DprofileT( const fpType* dbParams,
                      const fpType* respParams,
                      const fpType* shapeParams,
                            fpType  time,
                            size_t  dOrder ) const
 {
  fpType result = 0.0;
  for ( size_t a = 0; a < numActive( ); a++ )
   result +=   respParams[a]
             * my_cl[condition( a )].Dt( shapeParams, time, dOrder );

  return my_db->baseline( dbParams ) * result;
 }

fpType
acProfile::DprofileD( const fpType* dbParams,
                      const fpType* respParams,
                      const fpType* shapeParams,
                            fpType  time,
                            size_t  paramIndex ) const
 {
  if ( paramIndex >= my_db->basisSize( ) ) return 0.0;

  fpType result = 0.0;
  for ( size_t a = 0; a < numActive( ); a++ )
   result +=   respParams[a]
             * my_cl[condition( a )].value( shapeParams, time );

  return my_db->Dbaseline( dbParams, paramIndex ) * result;
 }

fpType
acProfile::DprofileR( const fpType* dbParams,
                      const fpType* respParams,
                      const fpType* shapeParams,
                            fpType  time,
                            size_t  paramIndex ) const
 {
  if ( paramIndex >= numActive( ) ) return 0.0;

  return
   (
       my_db->baseline( dbParams )
     * my_cl[condition( paramIndex )].value( shapeParams, time )
   );
 }

fpType
acProfile::DprofileS( const fpType* dbParams,
                      const fpType* respParams,
                      const fpType* shapeParams,
                            fpType  time,
                            size_t  paramIndex ) const
 {
  if ( paramIndex >= numShapeParams( ) ) return 0.0;

  fpType result = 0.0;
  for ( size_t a = 0; a < numActive( ); a++ )
   result +=   respParams[a]
             * my_cl[condition( a )].gradient( shapeParams, paramIndex, time );

  return my_db->baseline( dbParams ) * result;
 }

fpType
acProfile::DDprofileDD( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              fpType  time,
                              size_t  paramIndex1,
                              size_t  paramIndex2 ) const
 {
  return 0.0;
 }

fpType
acProfile::DDprofileRR( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              fpType  time,
                              size_t  paramIndex1,
                              size_t  paramIndex2 ) const
 {
  return 0.0;
 }

fpType
acProfile::DDprofileSS( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              fpType  time,
                              size_t  paramIndex1,
                              size_t  paramIndex2 ) const
 {
  if (    ( paramIndex1 >= numShapeParams( ) )
       || ( paramIndex2 >= numShapeParams( ) ) ) return 0.0;

  fpType result = 0.0;
  for ( size_t a = 0; a < numActive( ); a++ )
   result +=   respParams[a]
             * my_cl[condition( a )].hessian( shapeParams, paramIndex1, paramIndex2, time );

  return my_db->baseline( dbParams ) * result;
 }

fpType
acProfile::DDprofileDR( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              fpType  time,
                              size_t  paramIndexD,
                              size_t  paramIndexR ) const
 {
  if (    ( paramIndexR >= numActive( )        )
       || ( paramIndexD >= my_db->basisSize( ) ) ) return 0.0;

  return
   (
       my_db->Dbaseline( dbParams, paramIndexD )
     * my_cl[condition( paramIndexR )].value( shapeParams, time )
   );
 }

fpType
acProfile::DDprofileDS( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              fpType  time,
                              size_t  paramIndexD,
                              size_t  paramIndexS ) const
 {
  if (    ( paramIndexS >= numShapeParams( )   )
       || ( paramIndexD >= my_db->basisSize( ) ) ) return 0.0;

  fpType result = 0.0;
  for ( size_t a = 0; a < numActive( ); a++ )
   result +=   respParams[a]
             * my_cl[condition( a )].gradient( shapeParams, paramIndexS, time );

  return my_db->Dbaseline( dbParams, paramIndexD ) * result;
 }

fpType
acProfile::DDprofileRS( const fpType* dbParams,
                        const fpType* respParams,
                        const fpType* shapeParams,
                              fpType  time,
                              size_t  paramIndexR,
                              size_t  paramIndexS ) const
 {
  if (    ( paramIndexR >= numActive( )        )
       || ( paramIndexS >= numShapeParams( ) ) ) return 0.0;

  return
     my_db->baseline( dbParams )
   * my_cl[condition( paramIndexR )].gradient( shapeParams, paramIndexS, time );
 }

// --------------------------------------------------- activation "basis matrix"
// -----------------------------------------------------------------------------

fpMatrix
acProfile::basisMatrix( const fpType* shapeParams ) const
 {
  fpMatrix bm;
  getBasisMatrix( shapeParams, bm );
  return bm;
 }

void
acProfile::getBasisMatrix( const fpType* shapeParams, fpMatrix& bm ) const
 {
  bm.resize( timeGridSize( ), numActive( ), false, 0.0 );

  for ( size_t a = 0; a < numActive( ); a++ )
   my_cl[condition( a )].value( shapeParams, bm[a], 1.0 );
 }

// const string&
// acProfile::sParam2cTag( size_t sParamIndex ) const
//  {
//   return my_sParam2cTag[sParamIndex];
//  }

#endif /* __acProfile_H__ */
