/* ----------------------------------------------------------------------------
 *
 * class dbProfile
 *   implementation
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#include <brain.model/dbProfile>

// ------------------------------- construct/reinitialize/incremental initialize
// -----------------------------------------------------------------------------

dbProfile::dbProfile(       fpType        initialTime,
                            fpType        finalTime,
                      const fpVector*     timeGrid,
                      const BSplineBasis* dbBasis )
 : my_initialT( initialTime ),
   my_finalT( finalTime ),
   my_timeRange( finalTime - initialTime ),
   my_timeGrid( timeGrid ),
   my_basis( dbBasis )
 {
  initialize( );
 }

void
dbProfile::reinitialize(       fpType        initialTime,
                               fpType        finalTime,
                         const fpVector*     timeGrid,
                         const BSplineBasis* dbBasis )
 {
  my_initialT  = initialTime;
  my_finalT    = finalTime;
  my_timeRange = finalTime - initialTime;
  my_timeGrid  = timeGrid;
  my_basis     = dbBasis;

  initialize( );
 }

dbProfile::dbProfile( const dbProfile& another )
 : my_initialT( another.my_initialT ),
   my_finalT( another.my_finalT ),
   my_timeRange( another.my_timeRange ),
   my_timeGrid( another.my_timeGrid ),
   my_basis( another.my_basis ),
   my_basisMatrix( another.my_basisMatrix ),
   my_dpDOTdp( another.my_dpDOTdp ),
   my_Dbaseline( another.my_Dbaseline ),
   my_DDnormSqrDrift( another.my_DDnormSqrDrift ),
   my_DDnormSqrDDDrift( another.my_DDnormSqrDDDrift )
 {
 }

dbProfile&
dbProfile::operator=( const dbProfile& rhs )
 {
  if ( &rhs == this ) return *this;

  my_initialT    = rhs.my_initialT;
  my_finalT      = rhs.my_finalT;
  my_timeRange   = rhs.my_timeRange;

  my_timeGrid    = rhs.my_timeGrid;
  my_basis       = rhs.my_basis;

  my_basisMatrix = rhs.my_basisMatrix;
  my_dpDOTdp     = rhs.my_dpDOTdp;

  my_Dbaseline.resize( rhs.my_Dbaseline.size( ) );
  my_Dbaseline   = rhs.my_Dbaseline;

  my_DDnormSqrDrift   = rhs.my_DDnormSqrDrift;
  my_DDnormSqrDDDrift = rhs.my_DDnormSqrDDDrift;

  return *this;
 }

dbProfile::dbProfile( )
 : my_initialT( 0.0 ),
   my_finalT( 0.0 ),
   my_timeRange( 0.0 ),
   my_timeGrid( 0 ),
   my_basis( 0 ),
   my_basisMatrix( fpMatrix( ) ),
   my_dpDOTdp( fpMatrix( ) ),
   my_Dbaseline( fpVector( ) ),
   my_DDnormSqrDrift( fpMatrix( ) ),
   my_DDnormSqrDDDrift( fpMatrix( ) )
 {
 }

void
dbProfile::setTimeGrid(       fpType    initialTime,
                              fpType    finalTime,
                        const fpVector* timeGrid,
                              bool      init )
 {
  my_initialT  = initialTime;
  my_finalT    = finalTime;
  my_timeRange = finalTime - initialTime;
  my_timeGrid  = timeGrid;

  initialize( init );
 }

void
dbProfile::setBasis( const BSplineBasis* dbBasis, bool init )
 {
  my_basis = dbBasis;

  initialize( init );
 }

void
dbProfile::initialize( bool init )
 {
  if ( !init ) return;

  // cout << endl << "(debug) begin dbProfile::initialize( )" << endl;
  // cout << "(debug) my_basis " << my_basis << endl;
  // cout << endl << "(debug) basis attributes: " << endl;
  // cout << "(debug) numKnots( )                 " << my_basis->numKnots( ) << endl;
  // cout << "(debug) knots( )                    ";
  // for ( int i = 0; i < my_basis->numKnots( ); i++ )
  //  cout << my_basis->knots( )[i] << ( ( i == my_basis->numKnots( ) - 1 ) ? '\n' : ' ' );
  // cout << "(debug) leftEndPoint( )             " << my_basis->leftEndPoint( ) << endl;
  // cout << "(debug) rightEndPoint( )            " << my_basis->rightEndPoint( ) << endl;
  // cout << "(debug) isRightEndLeftContinuous( ) "
  //      << my_basis->isRightEndLeftContinuous( ) << endl;
  // cout << "(debug) isDropLastBasisFn( )        "
  //      << my_basis->isDropLastBasisFn( ) << endl;
  // cout << "(debug) order( )                    " << my_basis->order( ) << endl;
  // cout << "(debug) basisSize( )                " << my_basis->basisSize( ) << endl;
  // cout << endl;
  // cout << "(debug) my_timeGrid " << my_timeGrid << endl;
  // cout << "(debug) my_timeGrid->size( ) " << my_timeGrid->size( ) << endl;

  int basisSize = my_basis->basisSize( );
  int gridSize  = my_timeGrid->size( );

  my_basisMatrix.resize( gridSize, basisSize, false );
  // cout << "(debug) ok 0" << endl;
  my_dpDOTdp.resize( basisSize, basisSize, false );
  // cout << "(debug) ok 1" << endl;
  my_Dbaseline.resize( basisSize );
  // cout << "(debug) ok 2" << endl;

  fpType f = 1.0 / my_timeRange;

  for ( int i  = 0; i < basisSize; i++ )
   {
    for ( int j = 0; j < gridSize; j++ ) my_basisMatrix( j, i ) = my_basis->eval( i, (*my_timeGrid)[j] );
    my_Dbaseline[i] = f * my_basis->evalIntegral( i, my_initialT, my_finalT );
    for ( int j = 0; j <= i; j++ )
     my_dpDOTdp( i, j ) = my_dpDOTdp( j, i ) = BlasDrivers::dot( gridSize, my_basisMatrix[i], my_basisMatrix[j], 1, 1 );
   }

  // cout << "(debug) ok 3" << endl;

  my_basis->evalIntegral( my_initialT, my_finalT, my_DDnormSqrDrift );
  my_basis->evalIntegral( my_initialT, my_finalT, my_DDnormSqrDDDrift, 2 );

  // cout << "(debug) ok 4" << endl;

  for ( int i = 0; i < basisSize; i++ )
   {
    my_DDnormSqrDrift( i, i )
     -= my_timeRange * my_Dbaseline[i] * my_Dbaseline[i];

    for ( int j = 0; j < i; j++ )
     {
      f = my_timeRange * my_Dbaseline[i] * my_Dbaseline[j];

      my_DDnormSqrDrift( i, j ) -= f;
      my_DDnormSqrDrift( j, i ) -= f;
     }
   }

  // cout << "(debug) end dbProfile::initialize( )" << endl;
 }

// --------------------------------------------------------- smoothing parameter
// -----------------------------------------------------------------------------

dbProfile::smoothingFn::smoothingFn( int fnChoice, const fpVector& ev )
 : my_fnChoice( fnChoice ),
   my_ev( ev )
 {
 }
