/* ----------------------------------------------------------------------------
 *
 * class dbProfile
 *   drift+baseline computations handler
 *
 * required headers
 *   see include header file dbProfile
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __dbProfile_H__
#define __dbProfile_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class dbProfile
 {
  public:

   // ---------------------------- construct/reinitialize/incremental initialize
   // --------------------------------------------------------------------------

   dbProfile(       fpType        initialTime,
                    fpType        finalTime,
              const fpVector*     timeGrid,
              const BSplineBasis* dbBasis );
                         //
                         //                             single-shot construction
                         //
                         //  dbBasis and timeGrid are external, existing objects
                         // that remain in existence throughout the life of this
                         //                                     dbProfile object

   void
   reinitialize(       fpType        initialTime,
                       fpType        finalTime,
                 const fpVector*     timeGrid,
                 const BSplineBasis* dbBasis );
                                       //
                                       //           single-shot reinitialization

   dbProfile( const dbProfile& another );

   dbProfile&
   operator=( const dbProfile& rhs );

   dbProfile( );
                                       //
                                       //            incremental initialization:
                                       //
                                       //                         dumb defaults;
                                       //
                                       //                         initialTime: 0
                                       //                           finalTime: 0
                                       //                            timeGrid: 0
                                       //                             dbBasis: 0

   void
   setTimeGrid(       fpType    initialTime,
                      fpType    finalTime,
                const fpVector* timeGrid,
                      bool      init = false );
                                       //
                                       //            incremental initialization:
                                       //
                                       //     change internal time grid pointer;
                                       // initialize this object if init == true

   void
   setBasis( const BSplineBasis* dbBasis,
                   bool          init = false );
                                       //
                                       //            incremental initialization:
                                       //
                                       //         change internal basis pointer;
                                       // initialize this object if init == true

   void
   initialize( bool init = true );
                                       //
                                       //            incremental initialization:
                                       //
                                       //           perform all initializations;
                                       //         for use with the set functions

   // ------------------------------------------------- basis and timegrid sizes
   // --------------------------------------------------------------------------

   inline
   int
   basisSize( ) const;

   inline
   int
   timeGridSize( ) const;

   // ------------------------------- profile+derivatives over time grid: form 1
   // --------------------------------------------------------------------------

   inline
   fpVector
   profile( const fpType* dbParams ) const;
                                                   //
                                                   // profile over the time grid

   inline
   fpVector
   DprofileT( const fpType* dbParams, int dOrder = 1 ) const;
                                    //
                                    // derivative w.r.t. time over the time grid

   inline
   fpVector
   Dprofile( const fpType* dbParams, int paramIndex ) const;
                       //
                       // gradient w.r.t. the given parameter over the time grid

   inline
   fpVector
   DDprofile( const fpType* dbParams, int paramIndex1, int paramIndex2 ) const;
                       //
                       // hessian w.r.t. the given parameters over the time grid
                       //
                       //   note: in the present formulation and implementation,
                       //                drift+baseline is linear in parameters.
                       //        Hence zero second derivatives w.r.t. parameters

   // ------------------------------- profile+derivatives over time grid: form 2
   // --------------------------------------------------------------------------

   inline
   void
   getProfile( const fpType* dbParams, fpVector& p ) const;
                                       //
                                       // profile w.r.t. time over the time grid

   inline
   void
   getDprofileT( const fpType* dbParams, fpVector& Dp, int dOrder = 1 ) const;
                                    //
                                    // derivative w.r.t. time over the time grid

   inline
   void
   getDprofile( const fpType* dbParams, int paramIndex, fpVector& Dp ) const;
                       //
                       // gradient w.r.t. the given parameter over the time grid

   inline
   void
   getDDprofile( const fpType*   dbParams,
                       int       paramIndex1,
                       int       paramIndex2,
                       fpVector& DDp ) const;
                       //
                       // hessian w.r.t. the given parameters over the time grid
                       //
                       //   note: in the present formulation and implementation,
                       //                drift+baseline is linear in parameters.
                       //        Hence zero second derivatives w.r.t. parameters

   // ----------------------------------- profile+derivatives at arbitrary times
   // --------------------------------------------------------------------------

   inline
   fpType
   profile( const fpType* dbParams, fpType time ) const;

   inline
   fpType
   DprofileT( const fpType* dbParams, fpType time, int dOrder = 1 ) const;
                                                       //
                                                       // derivative w.r.t. time

   inline
   fpType
   Dprofile( const fpType* dbParams, int paramIndex, fpType time ) const;
                                        //
                                        // derivative w.r.t. the given parameter

   inline
   fpType
   DDprofile( const fpType* dbParams,
                    int     paramIndex1,
                    int     paramIndex2,
                    fpType  time ) const;
                       //
                       //   note: in the present formulation and implementation,
                       //                drift+baseline is linear in parameters.
                       //        Hence zero second derivatives w.r.t. parameters

   // --------------------------------- special functions based on the linearity
   // --------------------------------------------------------------------------

                 //    the following four functions are provided for the sake of
                 // efficiency, and are based on the linearity of drift+baseline
                 //   formulation. Although the same functions can be provide in
                 // the more general case, they may not serve any purpose there.
   inline
   fpType
   vDotDp( int paramIndex, const fpVector& theVector ) const;

   inline
   fpType
   vDotDp( int paramIndex, const fpType*   theVector ) const;
                                         //
                                         // Dprofile( paramIndex ) dot theVector

   inline
   fpType
   DpDotDp( int paramIndex1, int paramIndex2 ) const;
                          //
                          // Dprofile( paramIndex1 ) dot Dprofile( paramIndex2 )

   inline
   const fpMatrix&
   basisMatrix( ) const;                       // timeGridSize( ) X basisSize( )

   // ----------------------------------------------------------------- baseline
   // --------------------------------------------------------------------------

   inline
   fpType
   baseline( const fpType* dbParams ) const;
                                //
                                // baseline value from drift+baseline parameters

   inline
   fpType
   Dbaseline( const fpType* dbParams, int paramIndex ) const;
                        //
                        // baseline gradient with respect to the given parameter

   inline
   fpType
   DDbaseline( const fpType* dbParams, int paramIndex1, int paramIndex2 ) const;
                       //
                       //  baseline hessian with respect to the given parameters
                       //
                       //   note: in the present formulation and implementation,
                       //                drift+baseline is linear in parameters.
                       //        Hence zero second derivatives w.r.t. parameters

   inline
   const fpVector&
   Dbaseline( const fpType* dbParams ) const;
                                                     //
                                                     // entire baseline gradient

   inline
   fpMatrix
   DDbaseline( const fpType* dbParams ) const;
                       //
                       //                                entire baseline hessian
                       //
                       //   note: in the present formulation and implementation,
                       //                drift+baseline is linear in parameters.
                       //        Hence zero second derivatives w.r.t. parameters

   // ---------------------------------------------------------- 1/2 ||drift||^2
   // --------------------------------------------------------------------------

                                             //   ||f||^2 := integral_a^b f^2(t)
                                             // a := initialTime, b := finalTime

   inline
   fpType
   normSqrDrift( const fpType* dbParams ) const;
                                                        //
                                                        // 1/2 ||drift||^2 value

   inline
   fpType
   DnormSqrDrift( const fpType* dbParams, int paramIndex ) const;
                 //
                 // 1/2 ||drift||^2 gradient with respect to the given parameter

   inline
   fpType
   DDnormSqrDrift( const fpType* dbParams,
                         int     paramIndex1,
                         int     paramIndex2 ) const;
                 //
                 // 1/2 ||drift||^2 hessian with respect to the given parameters

   inline
   void
   getDnormSqrDrift( const fpType* dbParams, fpVector& DnormSqrDrift ) const;
                                        //
                                        //       entire 1/2 ||drift||^2 gradient
                                        // DnormSqrDrift.size( ) == basisSize( )

   inline
   const fpMatrix&
   DDnormSqrDrift( const fpType* dbParams ) const;
                                               //
                                               // entire 1/2 ||drift||^2 hessian

   // -------------------------------------------------------- 1/2 ||drift''||^2
   // --------------------------------------------------------------------------

   inline
   fpType
   normSqrDDDrift( const fpType* dbParams ) const;
                                                      //
                                                      // 1/2 ||drift''||^2 value

   inline
   fpType
   DnormSqrDDDrift( const fpType* dbParams, int paramIndex ) const;
               //
               // 1/2 ||drift''||^2 gradient with respect to the given parameter

   inline
   fpType
   DDnormSqrDDDrift( const fpType* dbParams,
                           int     paramIndex1,
                           int     paramIndex2 ) const;
               //
               // 1/2 ||drift''||^2 hessian with respect to the given parameters

   inline
   void
   getDnormSqrDDDrift( const fpType*   dbParams,
                             fpVector& DnormSqrDDDrift ) const;
                                      //
                                      //       entire 1/2 ||drift''||^2 gradient
                                      // DnormSqrDDDrift.size( ) == basisSize( )

   inline
   const fpMatrix&
   DDnormSqrDDDrift( const fpType* dbParams ) const;
                                             //
                                             // entire 1/2 ||drift''||^2 hessian

   // ------------------------------------------------------ smoothing parameter
   // --------------------------------------------------------------------------

   inline
   fpType
   smoothingPar( fpType  relWeight,
                 fpType& DoF,
                 int&    lapackInfo,
                 int&    iterations,
                 fpType& closeness,
                 int     fnChoice      = 0,
                 fpType  tolerance     = 1.0e+2 * DBL_EPSILON,
                 int     maxIterations = 50 ) const;
                                     //
                                     // smoothing parameter lambda corresponding
                                     //  to the given degrees of freedom and the
                                     //  given weight of the ||drift||^2 penalty
                                     //  (relWeight < 0 implies relWeight == 0).
                                     //
                                     //  out-of-range values of DoF are adjusted
                                     //               to an appropriate extreme.
                                     //
                                     //  any lapack errors reported by dsygv are
                                     //  returned in lapackInfo along with value
                                     //         -1.0 for the smoothing paramter.
                                     //
                                     //     tolerance and maxIterations serve as
                                     //             termination criteria for the
                                     //                     root-finding method.
                                     //
                                     //   value of |DoF - smoothing function| is
                                     //                   returned in closeness.
                                     //     iterations taken by the root-finding
                                     //        method is returned in iterations.
                                     //
                                     //   fnChoice            smoothing function
                                     //
                                     //         2              2 Tr S - Tr S S^T
                                     //         1                       Tr S S^T
                                     // otherwise                           Tr S

   // ------------------------------------------------- least-squares smooth fit
   // --------------------------------------------------------------------------

   inline
   void
   smoothFitLS( const fpVector& signal,
                      fpType    relWeight,
                      fpType    smoothingPar,
                      fpType*   parameters,
                      int&      lapackInfo ) const;
                                    //
                                    // least-square fit to the given signal with
                                    //                        desired smoothness
                                    //
                                    //  lapackInfo contains information returned
                                    //                                  by dposv
                                    //
                                    //         signal.size( ) == timeGridSize( )
                                    //
                                    //   parameters: buffer of size basisSize( )

   inline
   void
   getLSFitMatrix( fpType    relWeight,
                   fpType    smoothingPar,
                   fpMatrix& lsFitMatrix ) const;

   // ------------------------------------------------------------------ private
   // --------------------------------------------------------------------------

  private:
         fpType        my_initialT, my_finalT, my_timeRange;

   const fpVector*     my_timeGrid;           //              external time grid
   const BSplineBasis* my_basis;              // external B-spline basis handler

         fpMatrix      my_basisMatrix;   // Basis function values over time grid
                                         // = profile derivatives over time grid

         fpMatrix      my_dpDOTdp;            // Dprofile( i ) DOT Dprofile( j )

         fpVector      my_Dbaseline;                     // gradient of baseline

         fpMatrix      my_DDnormSqrDrift;        //   Hessian of 1/2 ||drift||^2
         fpMatrix      my_DDnormSqrDDDrift;      // Hessian of 1/2 ||drift''||^2

   class smoothingFn
    {
     public:
      smoothingFn( int fnChoice, const fpVector& ev );

      inline
      fpType
      operator()( fpType sParam ) const;

     private:
            int       my_fnChoice;
      const fpVector& my_ev;
    };

   inline
   void
   getBounds(       fpType    DoF,
                    fpType&   sLower,
                    fpType&   sUpper,
              const fpVector& ev,
                    int       numZeroEV,
                    int       fnChoice ) const;
 };

// ---------------------------------------------------- basis and timegrid sizes
// -----------------------------------------------------------------------------

int
dbProfile::basisSize( ) const
 {
  return my_basisMatrix.numCols( );
                      //
                      // same as my_basis.basisSize( ) when properly initialized
 }

int
dbProfile::timeGridSize( ) const
 {
  return my_basisMatrix.numRows( );
 }

// ---------------------------------- profile+derivatives over time grid: form 1
// -----------------------------------------------------------------------------

fpVector
dbProfile::profile( const fpType* dbParams ) const
 {
  fpVector p; getProfile( dbParams, p ); return p;
 }

fpVector
dbProfile::DprofileT( const fpType* dbParams, int dOrder ) const
 {
  fpVector Dp( timeGridSize( ) );
  getDprofileT( dbParams, Dp, dOrder );

  return Dp;
 }

fpVector
dbProfile::Dprofile( const fpType* dbParams, int paramIndex ) const
 {
  return fpVector( my_basisMatrix[paramIndex], timeGridSize( ) );
 }

fpVector
dbProfile::DDprofile( const fpType* dbParams,
                            int     paramIndex1,
                            int     paramIndex2 ) const
 {
  return fpVector( timeGridSize( ) );
 }

// ---------------------------------- profile+derivatives over time grid: form 2
// -----------------------------------------------------------------------------

void
dbProfile::getProfile( const fpType* dbParams, fpVector& p ) const
 {
  p.resize( timeGridSize( ) );
  BlasDrivers::gemv( timeGridSize( ), basisSize( ), my_basisMatrix[0], dbParams, p( ),
                     1.0, 0.0, 1, 1, cppblas::NoTranspose );
 }

void
dbProfile::getDprofileT( const fpType*   dbParams,
                               fpVector& Dp,
                               int       dOrder ) const
 {
  Dp.resize( timeGridSize( ) );

  for ( int i = 0; i < timeGridSize( ); i++ )
   Dp[i] += my_basis->eval( dbParams, (*my_timeGrid)[i], dOrder );
 }

void
dbProfile::getDprofile( const fpType*   dbParams,
                              int       paramIndex,
                              fpVector& Dp ) const
 {
  Dp.resize( timeGridSize( ) );

  const fpType* p = my_basisMatrix[paramIndex];

  for ( int i = 0; i < timeGridSize( ); i++ ) Dp[i] += p[i];
 }

void
dbProfile::getDDprofile( const fpType*   dbParams,
                               int       paramIndex1,
                               int       paramIndex2,
                               fpVector& DDp ) const
 {
  DDp.resize( timeGridSize( ) );
 }

// -------------------------------------- profile+derivatives at arbitrary times
// -----------------------------------------------------------------------------

fpType
dbProfile::profile( const fpType* dbParams, fpType time ) const
 {
  return my_basis->eval( dbParams, time );
 }

fpType
dbProfile::DprofileT( const fpType* dbParams, fpType time, int dOrder ) const
 {
  return my_basis->eval( dbParams, time, dOrder );
 }

fpType
dbProfile::Dprofile( const fpType* dbParams, int paramIndex, fpType time ) const
 {
  return my_basis->eval( paramIndex, time );
 }

fpType
dbProfile::DDprofile( const fpType* dbParams,
                            int     paramIndex1,
                            int     paramIndex2,
                            fpType  time ) const
 {
  return 0.0;
 }

// ------------------------------------ special functions based on the linearity
// -----------------------------------------------------------------------------

fpType
dbProfile::vDotDp( int paramIndex, const fpVector& theVector ) const
 {
  return
   BlasDrivers::dot( min( theVector.size( ), my_basisMatrix.numRows( ) ),
                       theVector( ), my_basisMatrix[paramIndex], 1, 1 );
 }

fpType
dbProfile::vDotDp( int paramIndex, const fpType* theVector ) const
 {
  return
   BlasDrivers::dot( timeGridSize( ), theVector, my_basisMatrix[paramIndex], 1, 1 );
 }

fpType
dbProfile::DpDotDp( int paramIndex1, int paramIndex2 ) const
 {
  return my_dpDOTdp( paramIndex1, paramIndex2 );
 }

const fpMatrix&
dbProfile::basisMatrix( ) const
 {
  return my_basisMatrix;
 }

// -------------------------------------------------------------------- baseline
// -----------------------------------------------------------------------------

fpType
dbProfile::baseline( const fpType* dbParams ) const
 {
  return
   BlasDrivers::dot( basisSize( ), dbParams, my_Dbaseline( ), 1, 1 );
 }

fpType
dbProfile::Dbaseline( const fpType* dbParams, int paramIndex ) const
 {
  return my_Dbaseline[paramIndex];
 }

fpType
dbProfile::DDbaseline( const fpType* dbParams,
                             int     paramIndex1,
                             int     paramIndex2 ) const
 {
  return 0.0;
 }

const fpVector&
dbProfile::Dbaseline( const fpType* dbParams ) const
 {
  return my_Dbaseline;
 }

fpMatrix
dbProfile::DDbaseline( const fpType* dbParams ) const
 {
  return fpMatrix( basisSize( ), basisSize( ) );
 }

// ------------------------------------------------------------- 1/2 ||drift||^2
// -----------------------------------------------------------------------------

fpType
dbProfile::normSqrDrift( const fpType* dbParams ) const
 {
  fpVector p( dbParams, basisSize( ) );
  return 0.5 * BlasDrivers::dot( p, BlasDrivers::symv( my_DDnormSqrDrift, p ) );
 }

fpType
dbProfile::DnormSqrDrift( const fpType* dbParams, int paramIndex ) const
 {
  fpVector d; getDnormSqrDrift( dbParams, d ); return d[paramIndex];
 }

fpType
dbProfile::DDnormSqrDrift( const fpType* dbParams,
                                 int     paramIndex1,
                                 int     paramIndex2 ) const
 {
  return my_DDnormSqrDrift( paramIndex1, paramIndex2 );
 }

void
dbProfile::getDnormSqrDrift( const fpType*   dbParams,
                                   fpVector& DnormSqrDrift ) const
 {
  DnormSqrDrift.resize( basisSize( ) );
  BlasDrivers::symv( basisSize( ), my_DDnormSqrDrift[0], dbParams, DnormSqrDrift( ),
                       1.0, 0.0, 1, 1, cppblas::UpperTriangle );
 }

const fpMatrix&
dbProfile::DDnormSqrDrift( const fpType* dbParams ) const
 {
  return my_DDnormSqrDrift;
 }

// ----------------------------------------------------------- 1/2 ||drift''||^2
// -----------------------------------------------------------------------------

fpType
dbProfile::normSqrDDDrift( const fpType* dbParams ) const
 {
  fpVector p( dbParams, basisSize( ) );
  return 0.5 * BlasDrivers::dot( p, BlasDrivers::symv( my_DDnormSqrDDDrift, p ) );
 }

fpType
dbProfile::DnormSqrDDDrift( const fpType* dbParams, int paramIndex ) const
 {
  fpVector d; getDnormSqrDDDrift( dbParams, d ); return d[paramIndex];
 }

fpType
dbProfile::DDnormSqrDDDrift( const fpType* dbParams,
                                   int     paramIndex1,
                                   int     paramIndex2 ) const
 {
  return my_DDnormSqrDDDrift( paramIndex1, paramIndex2 );
 }

void
dbProfile::getDnormSqrDDDrift( const fpType*   dbParams,
                                     fpVector& DnormSqrDDDrift ) const
 {
  DnormSqrDDDrift.resize( basisSize( ) );
  BlasDrivers::symv( basisSize( ), my_DDnormSqrDDDrift[0], dbParams, DnormSqrDDDrift( ),
                       1.0, 0.0, 1, 1, cppblas::UpperTriangle );
 }

const fpMatrix&
dbProfile::DDnormSqrDDDrift( const fpType* dbParams ) const
 {
  return my_DDnormSqrDDDrift;
 }

// --------------------------------------------------------- smoothing parameter
// -----------------------------------------------------------------------------

fpType
dbProfile::smoothingPar( fpType  relWeight,
                         fpType& DoF,
                         int&    lapackInfo,
                         int&    iterations,
                         fpType& closeness,
                         int     fnChoice      /* = 0      */,
                         fpType  tolerance     /* = 1.0e+2 * DBL_EPSILON */,
                         int     maxIterations /* = 50     */ ) const
 {
  fpMatrix a( my_dpDOTdp );
  fpMatrix omega( my_DDnormSqrDDDrift );

  if ( relWeight > 0.0 )
   {
    fpType p;
    for ( int i = 0; i < basisSize( ); i++ )
     {
      omega( i, i ) += ( relWeight * my_DDnormSqrDrift( i, i ) );
      for ( int j = 0; j < i; j++ )
       {
        p = relWeight * my_DDnormSqrDrift( i, j );
        omega( i, j ) += p;
        omega( j, i ) += p;
       }
     }
   }

  fpVector ev( basisSize( ) );
  fpVector work( 3 * basisSize( ) - 1 );

  cpplapack::dsygv( 1, cpplapack::jobz_0, cppblas::UpperTriangle, basisSize( ),
                    omega[0], basisSize( ), a[0], basisSize( ),
                    ev( ), work( ), work.size( ), lapackInfo );

  // cout << endl << endl;
  // cout << "(smoothingPar) lapackInfo " << lapackInfo << ", eigenvalues" << endl;
  // for ( int i = 0; i < basisSize( ); i++ ) cout << " " << ev[i];
  // cout << endl << endl;

  if ( lapackInfo ) return -1.0;

                 //                                     analyze the eigenvalues:
                 // eigenvalues must be nonnegative, sorted in increasing order.
                 //
  int numZeroEV = 0;
  for ( int i = 0; ( i < basisSize( ) ) && ( fabs( ev[i] ) <= tolerance ); i++ )
    {
     ev[i] = 0.0;
     numZeroEV++;
    }

                  //               the smoothing function is a bounded monotonic
                  // increasing function of the smoothing parameter, which takes
                  //        value numZeroEV at 0.0 and basisSize( ) at infinity.
                  //  Values of DoF outside this range have no inverse under the
                  //                                         smoothing function.
                  //
  if ( DoF <= (fpType) numZeroEV )
   {
    DoF       = (fpType) numZeroEV;
    closeness = 0.0;
    return 0.0;
   }

  if ( DoF >= basisSize( ) ) DoF = (fpType) basisSize( ) - tolerance;
                                   //
                                   // need a better algorithm here to reduce DoF

  fpType sLower, sUpper;
  getBounds( DoF, sLower, sUpper, ev, numZeroEV, fnChoice );

  return
   MiscNumericFns::BisectionSearch( sLower, sUpper,
                                    iterations, closeness,
                                    smoothingFn( fnChoice, ev ),
                                    DoF, maxIterations, tolerance );
 }

fpType
dbProfile::smoothingFn::operator()( fpType sParam ) const
 {
  fpType smFn = 0.0;
  fpType p;

  if ( my_fnChoice == 2 )
   {
    for ( int i = 0; i < my_ev.size( ); i++ )
     {
      p     = 1.0 / ( my_ev[i] + sParam );
      smFn += p * ( 2.0 - sParam * p );
     }
   }
  else if ( my_fnChoice == 1 )
   {
    for ( int i = 0; i < my_ev.size( ); i++ )
     {
      p     = 1.0 / ( my_ev[i] + sParam );
      smFn += p * p;
     }
    smFn *= sParam;
   }
  else
   {
    for ( int i = 0; i < my_ev.size( ); i++ )
     {
      p     = 1.0 / ( my_ev[i] + sParam );
      smFn += p;
     }
   }

  return sParam * smFn;
 }

void
dbProfile::getBounds(       fpType    DoF,
                            fpType&   sLower,
                            fpType&   sUpper,
                      const fpVector& ev,
                            int       numZeroEV,
                            int       fnChoice ) const
 {
  fpType a = ( DoF - (fpType) numZeroEV ) / (fpType) ( ev.size( ) - numZeroEV );

  if ( fnChoice == 2 )
   {
    a  = sqrt( a );
    a /= ( 1.0 - a );
    a *= ( 2.0 * a - 1.0 );
   }
  else if ( fnChoice == 1 )
   {
    a  = sqrt( a );
    a /= ( 1.0 - a );
   }
  else
   a /= ( 1.0 - a );

  sLower = a * ev[numZeroEV];
  sUpper = a * ev[ev.size( ) - 1];
 }

// ------------------------------------------------------------------- drift fit
// -----------------------------------------------------------------------------

void
dbProfile::getLSFitMatrix( fpType    relWeight,
                           fpType    smoothingPar,
                           fpMatrix& lsFitMatrix ) const
 {
  if (    ( lsFitMatrix.numRows( ) < basisSize( ) )
       || ( lsFitMatrix.numCols( ) < basisSize( ) ) )
   lsFitMatrix.resize( basisSize( ), basisSize( ), false, 0.0 );

  fpType s = 1.0 / smoothingPar;

  fpType p;
  for ( int i = 0; i < basisSize( ); i++ )
   {
    lsFitMatrix( i, i ) =   my_dpDOTdp( i, i )
                          + s * (   relWeight * my_DDnormSqrDrift( i, i )
                                  +             my_DDnormSqrDDDrift( i, i ) );

    for ( int j = 0; j < i; j++ )
     {
      p =   my_dpDOTdp( i, j )
          + s * (   relWeight * my_DDnormSqrDrift( i, j )
                  +             my_DDnormSqrDDDrift( i, j ) );

      lsFitMatrix( i, j ) = p;
      lsFitMatrix( j, i ) = p;
     }
   }
 }

void
dbProfile::smoothFitLS( const fpVector& signal,
                              fpType    relWeight,
                              fpType    smoothingPar,
                              fpType*   parameters,
                              int&      lapackInfo ) const
 {
  fpMatrix a; getLSFitMatrix( relWeight, smoothingPar, a );

  for ( int i = 0; i < basisSize( ); i++ )
   parameters[i] = BlasDrivers::dot( min( signal.size( ), my_basisMatrix.numRows( ) ),
                                       signal( ), my_basisMatrix[i], 1, 1 );

  cpplapack::dposv( cppblas::UpperTriangle, basisSize( ), 1,
                    a[0], basisSize( ), parameters, basisSize( ), lapackInfo );
 }

#endif /* __dbProfile_H__ */
