/* ----------------------------------------------------------------------------
 *
 * header initialize
 *   initialization schemes for minimization guess
 *
 * required headers
 *   see include header file initialize
 *
 * notes
 *   current implementation initializes any AR parameters to zero.
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __initialize_H__
#define __initialize_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );
cpp_using_namespace( cpplapack );

             //    The implementation below is for a normal arma(0,0) likelihood
             // with/without activation. other likelihoods need specializations.

template< class LL >
void
initialize( const                 Settings&                 settings,
            const                 fpVector&                 timeCourse,
            const                 ParameterManager<fpType>& parManager,
            const                 acProfile&                activation,
            const                 dbProfile&                drift,
            const                 ObjFnManager&             objectiveFn,
            const                 LL&                       ll,
            const                 LPdrift&                  lpDrift,
            const                 LPgamma&                  lpNoise,
            const                 vector<LPgamma>&          lpResponse,
         /* in_principle_const */ map<string,LPgamma>&      lpShape,
            const                 MinQNB&                   minimizer,
                                  fpVector&                 parameters,
                                  bool                      priorMeanForShape = true )
 {
  // cerr << endl;
  // cerr << "begin initialize" << endl;

  int ofs_d = parManager.blockOffset( ParamBlkNames::drift    );
  int ofs_n = parManager.blockOffset( ParamBlkNames::noise    );
  // int ofs_a = parManager.blockOffset( ParamBlkNames::ar       );
  int ofs_r = parManager.blockOffset( ParamBlkNames::response );
  int ofs_s = parManager.blockOffset( ParamBlkNames::shape    );
                                                           //
                                                           // external ordering

  size_t len_d = parManager.blockLength( ParamBlkNames::drift    );
  size_t len_n = parManager.blockLength( ParamBlkNames::noise    ); // ATTN assume len_n == 1
  size_t len_a = parManager.blockLength( ParamBlkNames::ar       );
  size_t len_r = parManager.blockLength( ParamBlkNames::response );
  size_t len_s = parManager.blockLength( ParamBlkNames::shape    );

  // cerr << endl;
  // cerr << "ofs_d " << ofs_d << " len_d " << len_d << endl;
  // cerr << "ofs_n " << ofs_n << " len_n " << len_n << endl;
  // cerr << "ofs_a " << ofs_a << " len_a " << len_a << endl;
  // cerr << "ofs_r " << ofs_r << " len_r " << len_r << endl;
  // cerr << "ofs_s " << ofs_s << " len_s " << len_s << endl;

  ParameterManager<fpType> initPM;

  initPM.addBlock( ParamBlkNames::drift,    len_d );
  initPM.addBlock( ParamBlkNames::response, len_r );
  initPM.addBlock( ParamBlkNames::ar,       len_a );
  initPM.addBlock( ParamBlkNames::noise,    len_n );
  initPM.addBlock( ParamBlkNames::shape,    len_s );
                            //
                            // internal ordering: drift-response-ar-noise-shape
                            //        all AR parameters are initialized to zero

  int init_ofs_d = initPM.blockOffset( ParamBlkNames::drift    );
  int init_ofs_r = initPM.blockOffset( ParamBlkNames::response );
  int init_ofs_n = initPM.blockOffset( ParamBlkNames::noise    );
  // int init_ofs_a = initPM.blockOffset( ParamBlkNames::ar       );
  int init_ofs_s = initPM.blockOffset( ParamBlkNames::shape    );

  // cerr << endl;
  // cerr << "init_ofs_d " << init_ofs_d << endl;
  // cerr << "init_ofs_a " << init_ofs_a << endl;
  // cerr << "init_ofs_n " << init_ofs_n << endl;
  // cerr << "init_ofs_r " << init_ofs_r << endl;
  // cerr << "init_ofs_s " << init_ofs_s << endl;

  bool varyNoise = (bool) settings.initializer.scheme;

  int    ofsZero = init_ofs_s;    // by default, vary only the shape parameters
  size_t numVars = len_s;

  if ( varyNoise )
   {
    ofsZero  = init_ofs_n;
    numVars += len_n;                                 // ATTN assume len_n == 1
   }

  initPM.setOffsetShift( ofsZero );
     //
     // initPM is used only for setting correct offsets in likelihood functions

  fpVector params( 0.0, parManager.numParams( ) );            // scratch buffer

  parameters.resize( params.size( ), 0.0 );

  ObjFnManager   initOF;
  LL             initLL( ll );
  LPdrift        initLPd( lpDrift );
  LPgamma        initLPn( lpNoise );
  list<LPgamma>  initLPs;
  MinQNB         initMin;

  fpType priorRespMin = 0.0;

  if ( len_r > 0 )
   {
    initMin.setSizes( numVars, settings.initializer.minimizer.maxCorrections );

    initOF.setParamBufSize( numVars );
    initOF.setSign( -1.0 );

    initLL.setParamBufSize( numVars );
    initLL.setParamOffset( initPM );

    initOF.add( &initLL );

    for ( size_t a = 0; a < activation.numActive( ); a++ )
     {
      const acProfile::ConditionBellSet& cbs = activation.bellSet( activation.condition( a ) );

      initLPs.push_back( lpShape[cbs.tag( )] );

      initLPs.back( ).setParamBufSize( numVars );
      initLPs.back( ).setParamOffset( initLPs.back( ).paramOffset( ) - ofsZero );

      initOF.add( &( initLPs.back( ) ) );
     }

    if ( varyNoise )
     {
      initLPd.setParamBufSize( numVars );
      initLPd.setParamOffset( init_ofs_d - ofsZero, init_ofs_n - ofsZero );

      initOF.add( &initLPd );

      initLPn.setParamBufSize( numVars );
      initLPn.setParamOffset( init_ofs_n - ofsZero );

      initOF.add( &initLPn );

      initMin.setBounds( 0, minimizer.boundType( ofs_n ),
                            minimizer.lowerBound( ofs_n ),
                            minimizer.upperBound( ofs_n ) );

      initMin.setScale( 0, minimizer.scale( ofs_n ) );
     }

    for ( int i = 0, iOfs = ofs_s; i < len_s; i++, iOfs++ )
     {
      initMin.setBounds( i + varyNoise, minimizer.boundType( iOfs ),
                                        minimizer.lowerBound( iOfs ),
                                        minimizer.upperBound( iOfs ) );

      initMin.setScale( i + varyNoise, minimizer.scale( iOfs ) );
     }

    initMin.setStoppingCriteria( settings.initializer.minimizer.minRelReduceF,
                                 settings.initializer.minimizer.maxPrGradNorm,
                                 settings.initializer.minimizer.maxEvaluations,
                                 settings.initializer.minimizer.maxIterations );
    initMin.setRestartOpts( settings.initializer.minimizer.maxRestarts,
                            settings.initializer.minimizer.checkHdet,
                            settings.initializer.minimizer.stopAtPlusDefH,
                            settings.initializer.minimizer.mrrfReduction );
    initMin.setVerbosity( settings.initializer.minimizer.verbosity );
    initMin.setObjectiveFn( &initOF );

                                             // shape guess: (gamma) prior means
                                             //
    // cerr << endl;
    for ( list<LPgamma>::const_iterator lp = initLPs.begin( ); lp != initLPs.end( ); ++lp )
     for ( int i = 0, j = init_ofs_s + lp->paramOffset( ) - varyNoise; i < lp->numParams( ); i++, j++ )
      {
       // cerr << "i " << i << " lp->paramOffset( ) " << lp->paramOffset( ) << " j " << j << " shape[j] " << params[j] << endl;
       params[j] =   priorMeanForShape
	            ? lp->constA( i ) / lp->constB( i )                // mean
                   : ( ( lp->constA( i ) - 1.0 ) / lp->constB( i ) ); // mode
      }

                     // minimum prior-expected response across active conditions
                     //
    priorRespMin =   lpResponse[activation.condition( 0 )].constA( 0 )
                   / lpResponse[activation.condition( 0 )].constB( 0 );

    for ( int a = 1; a < len_r; a++ )
     priorRespMin = min( priorRespMin,
                           lpResponse[activation.condition( a )].constA( 0 )
                         / lpResponse[activation.condition( a )].constB( 0 ) );
   }

  // cerr << endl;
  // cerr << "priorRespMin " << priorRespMin << endl;

                                                        // noise prior constants
                                                        //
  fpType noiseConstA = lpNoise.constA( 0 );
  fpType noiseConstB = lpNoise.constB( 0 );

          // matrices for least-squares smooth fit for drift and response params

  fpMatrix bTb;
  drift.getLSFitMatrix( settings.drift.relWeight, settings.drift.smoothPar, bTb );

  const fpMatrix& b = drift.basisMatrix( );

  int lsFitSize = len_d + len_r;
  fpMatrix lsFitMatrix( lsFitSize, lsFitSize );

  fpMatrix a;

  fpVector acp( 0.0, timeCourse.size( ) );
  fpVector dbp( 0.0, timeCourse.size( ) );

  fpType p, ss, respMin;
  int lapackInfo;

                                                    // begin initialization loop

  size_t it   = 0;
  bool   over = false;
  fpType f    = 0.0;
  fpType fOld = +DBL_MAX;

  do
   {
    // cerr << endl;
    // cerr << "it " << it << endl;

    // cerr << endl;
    // cerr << "params ";
    // for ( size_t i = 0; i < params.size( ); i++ ) cerr << " " << params[i];
    // cerr << endl;

                             // activation basis matrix for current shape params
                             //
    activation.getBasisMatrix( params( init_ofs_s ), a );

                                    // form the full drift+response basis matrix
                                    //
    for ( int i = 0, iOfs = init_ofs_d; i < len_d; i++, iOfs++ )
     for ( int j = 0, jOfs = init_ofs_d; j <= i; j++, jOfs++ )
      lsFitMatrix( iOfs, jOfs ) = lsFitMatrix( jOfs, iOfs ) = bTb( i, j );

    for ( int i = 0, iOfs = init_ofs_r; i < len_r; i++, iOfs++ )
     for ( int j = 0, jOfs = init_ofs_r; j <= i; j++, jOfs++ )
      {
       p = BlasDrivers::dot( timeCourse.size( ), a[i], a[j], 1, 1 );
       lsFitMatrix( iOfs, jOfs ) = lsFitMatrix( jOfs, iOfs ) = p;
      }

    for ( int i = 0, iOfs = init_ofs_d; i < len_d; i++, iOfs++ )
     for ( int j = 0, jOfs = init_ofs_r; j < len_r; j++, jOfs++ )
      {
       p = BlasDrivers::dot( timeCourse.size( ), b[i], a[j], 1, 1 );
       lsFitMatrix( iOfs, jOfs ) = lsFitMatrix( jOfs, iOfs ) = p;
      }

    for ( int i = 0, iOfs = init_ofs_d; i < len_d; i++, iOfs++ )
     params[iOfs] = BlasDrivers::dot( timeCourse.size( ), timeCourse( ), b[i], 1, 1 );

    for ( int i = 0, iOfs = init_ofs_r; i < len_r; i++, iOfs++ )
     params[iOfs] = BlasDrivers::dot( timeCourse.size( ), timeCourse( ), a[i], 1, 1 );

                           // solve the drift+response least-squares fit problem
                           //
    dposv( cppblas::UpperTriangle, lsFitSize, 1, lsFitMatrix[0], lsFitSize,
           params( init_ofs_d ), lsFitSize, lapackInfo );

    assert( lapackInfo == 0 );

    if ( len_r > 0 )
     {
                      //    if any of the reponse params are negative, shift all
                      // response params up by that amount. To compensate, shift
                      //         all drift coefficients down by the same amount.
                      //
      respMin = 0.0;
      for ( int i = 0, j = init_ofs_r; i < len_r; i++, j++ )
       if ( params[j] < respMin ) respMin = params[j];

      if ( respMin < 0.0 ) respMin -= priorRespMin;
                   //
                   //    this ensures that the guess response for all conditions
                   // consistent with the response priors, as the drift/response
                   //  guess procedure yields information only about differences
                   //                                              in responses.

      // cerr << endl;
      // cerr << "respMin " << respMin << endl;

      for ( int i = 0, iOfs = init_ofs_d; i < len_d; i++, iOfs++ )
       params[iOfs] += respMin;

      fpType invBaseline = 1.0 / drift.baseline( params( init_ofs_d ) );

      for ( int i = 0, iOfs = init_ofs_r; i < len_r; i++, iOfs++ )
       params[iOfs] = ( params[iOfs] - respMin ) * invBaseline;

      if ( ( it == 0 ) || !varyNoise )
       activation.getProfile( params( init_ofs_d ), params( init_ofs_r ),
                              params( init_ofs_s ), acp );

                         // copy response parameters in the external block order
                         //
      for ( int i = 0, j = init_ofs_r, k = ofs_r; i < len_r; i++, j++, k++ )
       parameters[k] = params[j];
     }

                                               // estimate noise using residuals
                                               //
    if ( ( it == 0 ) || !varyNoise )
     {
      drift.getProfile( params( init_ofs_d ), dbp );

      ss = 0.0;
      for ( size_t i = 0; i < timeCourse.size( ); i++ )
       {
        p   = timeCourse[i] - dbp[i] - acp[i];
        ss += ( p * p );
       }

      params[init_ofs_n]
       = ( 0.5 * timeCourse.size( ) + noiseConstA - 1.0 ) / ( 0.5 * ss + noiseConstB );
     }

                  // copy drift and noise parameters in the external block order
                  //
    for ( int i = 0, j = init_ofs_d, k = ofs_d; i < len_d; i++, j++, k++ )
     parameters[k] = params[j];

    parameters[ofs_n] = params[init_ofs_n];

    if ( len_r > 0 )
     {
      if ( it == 0 )
       {
                            // copy shape parameters in the external block order
                            //
        for ( int i = 0, j = init_ofs_s, k = ofs_s; i < len_s; i++, j++, k++ )
         parameters[k] = params[j];

        // cerr << endl;
        // cerr << "(it=0) parameters ";
        // for ( size_t i = 0; i < parameters.size( ); i++ ) cerr << " " << parameters[i];
        // cerr << endl;

        // cerr << endl;
        // cerr << "(it=0) params ";
        // for ( size_t i = 0; i < params.size( ); i++ ) cerr << " " << params[i];
        // cerr << endl;

        objectiveFn.value( parameters( 0 ), &( f ) );

        // cerr << endl;
        // cerr << "(it=0) f " << f << endl;
       }

      // cerr << endl;
      // cerr << "params before minimization ";
      // for ( size_t i = 0; i < params.size( ); i++ ) cerr << " " << params[i];
      // cerr << endl;

                        // minimize w. r. t. shape (and optionally noise) params
                        //
      initMin.minimize( params( ofsZero ) );

      // cerr << endl;
      // cerr << "initMin.statusCode( )  " << initMin.statusCode( ) << endl;
      // cerr << "initMin.iterations( )  " << initMin.numIterations( ) << endl;
      // cerr << "initMin.evaluations( ) " << initMin.numEvals( ) << endl;
      // cerr << "initMin.restarts( )    " << initMin.numRestarts( ) << endl;

      // cerr << endl;
      // cerr << "params after minimization ";
      // for ( size_t i = 0; i < params.size( ); i++ ) cerr << " " << params[i];
      // cerr << endl;

                      // copy shape/noise parameters in the external block order
                      //
      parameters[ofs_n] = params[init_ofs_n];

      for ( int i = 0, j = init_ofs_s, k = ofs_s; i < len_s; i++, j++, k++ )
       parameters[k] = params[j];

                               // full objective function value: has it reduced?
                               //
      fOld = f;
      objectiveFn.value( parameters( 0 ), &( f ) );
     }

    it++;
    over = ( fOld <= f ) || ( it >= settings.initializer.maxIt ) || ( len_r == 0 );

    // cerr << endl;
    // cerr << "it " << it << " f " << f << " fOld " << fOld << " over " << over << endl;

    // cerr << endl;
    // cerr << "parameters";
    // for ( size_t i = 0; i < parameters.size( ); i++ ) cerr << " " << parameters[i];
    // cerr << endl;
   }
  while ( !over );

  // ATTN parameter bounds haven't been imposed explicitly: it is assumed that
  // ATTN the minimizer will project any out-of-bounds values onto the bounds.

  // cerr << endl;
  // cerr << "end initialize" << endl;
 }

#endif /* __initialize_H__ */
