/* ----------------------------------------------------------------------------
 *
 * class ModelProbability
 *   implementation
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#include <brain.vmp/model.probability>

ModelProbability::ModelProbability( const string& mpChunkSuffix,
                                    const string& mpChunkType,
                                          bool    mpLittleEndian,
                                    const string& bfChunkSuffix,
                                    const string& bfChunkType,
                                          bool    bfLittleEndian,
                                          bool    pghmriMP,
                                          bool    stdoutMP,
                                          bool    pghmriBF,
                                          bool    stdoutBF,
                                          bool    computeBFMP )
 {
  configure( mpChunkSuffix, mpChunkType, mpLittleEndian,
             bfChunkSuffix, bfChunkType, bfLittleEndian );
  setOutput( pghmriMP, stdoutMP, pghmriBF, stdoutBF, computeBFMP );
  my_initialized = false;
 }

void
ModelProbability::configure( const string& mpChunkSuffix,
                             const string& mpChunkType,
                                   bool    mpLittleEndian,
                             const string& bfChunkSuffix,
                             const string& bfChunkType,
                                   bool    bfLittleEndian )
 {
  my_mpChunkSuffix = mpChunkSuffix; my_mpChunkType = mpChunkType; my_mpLittleEndian = mpLittleEndian;
  my_bfChunkSuffix = bfChunkSuffix; my_bfChunkType = bfChunkType; my_bfLittleEndian = bfLittleEndian;
 }

void
ModelProbability::setOutput( bool pghmriMP, bool stdoutMP, bool pghmriBF, bool stdoutBF, bool computeBFMP )
 {
  my_pghmriMP = pghmriMP; my_pghmriBF = pghmriBF; my_pghmri = pghmriMP || pghmriBF;
  my_stdoutMP = stdoutMP; my_stdoutBF = stdoutBF; my_stdout = stdoutMP || stdoutBF;
  my_computeBFMP = ( my_pghmri || my_stdout || computeBFMP );
 }

void
ModelProbability::initialize(       PghMRIhandler*      mriHandler,
                              const TCreader*           dataNavigator,
                              const SubmodelTable*      smTable,
                              const vector<ProductSet>& smProducts,
		                    size_t              nullIndex,
                                    fpVector*           priorProbability )
 {
  // cerr << "(debug) ModelProbability::initialize( ) anyOutput( ) " << anyOutput( ) << " my_computeBFMP " << my_computeBFMP << endl;

  if ( !( anyOutput( ) || my_computeBFMP ) ) return;

  my_smt = smTable; assert( my_smt != 0);

  my_tcr = dataNavigator; assert( my_tcr != 0);
  for ( size_t i = 0; i < my_tcr->numDims( ); i++ )
   assert( my_tcr->names( )[i].size( ) == 1 );

  assert( smTable->size( ) == smProducts.size( ) );

  bool has_logint = true; my_li.resize( smProducts.size( ) );
  for ( size_t i = 0; has_logint && ( i < smProducts.size( ) ); i++ )
  {
   my_li[i]   = cpp_static_cast( LogIntegral*, smProducts[i]( "log.integral" ) );
   has_logint = has_logint && ( my_li[i] != 0 ) && ( my_li[i]->isInitialized( ) );
  }

  assert( has_logint );

  assert( priorProbability != 0 ); my_prior = priorProbability;

  if ( my_prior->size( ) == 0 )
   {
    my_prior->resize( smProducts.size( ) );
    my_prior->operator=( 1.0 / smProducts.size( ) );
   }
  else
   {
    assert( my_prior->size( ) == smProducts.size( ) );

    fpType prior_norm = 0.0;
    for ( size_t i = 0; i < my_prior->size( ); i++ )
     {
      assert( my_prior->operator[]( i ) >= 0.0 );
      prior_norm += my_prior->operator[]( i );
     }
    assert( prior_norm > 0.0 );

    if ( prior_norm != 1.0 )
     {
      prior_norm = 1.0 / prior_norm;
      for ( size_t i = 0; i < my_prior->size( ); i++ ) my_prior->operator[]( i ) *= prior_norm;
     }
   }

  // cerr << "(debug) ModelProbability::initialize( ) my_nullIndex " << my_nullIndex << endl;

  my_nullIndex = nullIndex; assert( my_nullIndex < my_li.size( ) );

  // cerr << "(debug) ModelProbability::initialize( ) my_pghmri " << my_pghmri << endl;

  if ( !my_pghmri ) { my_initialized = true; return; }

  string dimensions = my_tcr->nameStr( "", my_tcr->dimsSansT( ) );
  assert( dimensions.size( ) == my_tcr->dimsSansT( ).size( ) );

  uintVector s( my_tcr->sizesSansT( ) );
  uintVector l( my_tcr->lowerSansT( ) );

  int sizes[s.size( )]; int lower[l.size( )];
  for ( size_t i = 0; i < s.size( ); i++ ) { lower[i] = l[i]; sizes[i] = s[i]; }

  /*
  cerr << "(debug) ModelProbability::initialize( ) sizes";
  for ( size_t i = 0; i < s.size( ); i++ ) cerr << " " << sizes[i];
  cerr << endl;

  cerr << "(debug) ModelProbability::initialize( ) lower";
  for ( size_t i = 0; i < l.size( ); i++ ) cerr << " " << lower[i];
  cerr << endl;
  */

  my_aim.reinitialize( s, uintVector( ), l );

  /*
  cerr << "(debug) ModelProbability::initialize( ) my_aim.numDims( ) " << my_aim.numDims( ) << " my_aim.size( ) " << my_aim.size( ) << endl;
  for ( size_t d = 0; d < my_aim.numDims( ); d++ )
   cerr << "(debug) ModelProbability::initialize( ) dimension " << d
        << " size "             << my_aim.sizes( )[d]
        << " order "            << my_aim.order( )[d]
        << " index range "      << my_aim.indexLower( )[d] << ":" << my_aim.indexUpper( )[d]
        << endl;
  */

  my_mrh = mriHandler; assert( my_mrh->isOpen( ) && ( my_mrh->openMode( ) != PghMRIhandler::mriRead ) );

  string key, name;
  for ( SubmodelTable::const_iterator si = my_smt->begin( ); si != my_smt->end( ); ++si )
   {
    key = my_smt->key( si );

    if ( my_pghmriMP )
     {
      name = key + "." + my_mpChunkSuffix;
      if ( my_mrh->openMode( ) == PghMRIhandler::mriModifyData )
       assert( PghMRIhandler::matchChunk( *my_mrh, name, my_mpChunkType, dimensions, sizes, lower ) );
      else
       {
        my_mrh->createChunk( name, dimensions, my_mpChunkType, name + ".dat", sizes, lower, my_mpLittleEndian );
        my_mrh->setString( name + ".with.respect.to", key );
       }
     }

    if ( my_pghmriBF )
     {
      name = key + "." + my_bfChunkSuffix;
      if ( my_mrh->openMode( ) == PghMRIhandler::mriModifyData )
       assert( PghMRIhandler::matchChunk( *my_mrh, name, my_bfChunkType, dimensions, sizes, lower ) );
      else
       {
        my_mrh->createChunk( name, dimensions, my_bfChunkType, name + ".dat", sizes, lower, my_bfLittleEndian );
        my_mrh->setString( name + ".with.respect.to", key );
       }
     }
   }

  my_initialized = true;
 }

const double ModelProbability::our_lnDBLmax = log( DBL_MAX );
