/* ----------------------------------------------------------------------------
 *
 * class ModelProbability
 *
 * required headers
 *   see include header file products
 *
 * notes
 *  submodelwise log bayes factors and model probabilities
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __ModelProbability_H__
#define __ModelProbability_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

struct ModelProbability
 {
  ModelProbability( const string& mpChunkSuffix  = "model.probability",
                    const string& mpChunkType    = PghMRIhandler::type2name( PghMRIhandler::mriFloat ),
                          bool    mpLittleEndian = PghMRIhandler::endianness( ),
                    const string& bfChunkSuffix  = "log.bayes.factor",
                    const string& bfChunkType    = PghMRIhandler::type2name( PghMRIhandler::mriFloat ),
                          bool    bfLittleEndian = PghMRIhandler::endianness( ),
                          bool    pghmriMP       = true,
                          bool    stdoutMP       = false,
                          bool    pghmriBF       = true,
                          bool    stdoutBF       = false,
                          bool    computeBFMP    = true );
          //
          // pghmri?? and stdout?? specify the output format, compute?? indicate
          // whether  to compute  the quantities  even when  there is no output.

  void
  configure( const string& mpChunkSuffix  = "model.probability",
             const string& mpChunkType    = PghMRIhandler::type2name( PghMRIhandler::mriFloat ),
                   bool    mpLittleEndian = PghMRIhandler::endianness( ),
             const string& bfChunkSuffix  = "log.bayes.factor",
             const string& bfChunkType    = PghMRIhandler::type2name( PghMRIhandler::mriFloat ),
                   bool    bfLittleEndian = PghMRIhandler::endianness( ) );

  void
  setOutput( bool pghmriMP = true, bool stdoutMP = false, bool pghmriBF = true, bool stdoutBF = false, bool computeBFMP = true );
          //
          // pghmri?? and stdout?? specify the output format, compute?? indicate
          // whether  to compute  the quantities  even when  there is no output.

  void
  initialize(       PghMRIhandler*      mriHandler,
              const TCreader*           dataNavigator,
              const SubmodelTable*      smTable,
              const vector<ProductSet>& smProducts,
                    size_t              nullIndex,
                    fpVector*           priorProbability );
                            //
                            // priorProbabilities->size( ) == 0 is taken to  imply
                            // equal a priori submodel probabilities. If this size
                            // is nonzero, it must match the number of  submodels.
                            // supplied probabilities need not be normalized.

  inline const fpVector& prior( ) const;
                                      //
                                      // normalized prior submodel probabilities

  inline const fpVector& bayesFactors( bool tcIsConst ) const;
                                      //
                                      //          submodelwise log bayes factors

  inline bool anyErrorsBF( ) const;
  inline bool anyOutputBF( ) const;

  inline const fpVector& posterior( bool tcIsConst ) const;
                                      //
                                      //        posterior submodel probabilities

  inline bool anyErrorsMP( ) const;
  inline bool anyOutputMP( ) const;

  inline bool anyErrors( ) const;
  inline bool anyOutput( ) const;

  inline void write( bool tcIsConst, const string& stdoutPrefix ) const;

  inline const string& mpChunkSuffix( ) const;
  inline const string& mpChunkType( ) const;
  inline       bool    mpLittleEndian( ) const;

  inline const string& bfChunkSuffix( ) const;
  inline const string& bfChunkType( ) const;
  inline       bool    bfLittleEndian( ) const;

  protected:
   inline bool nothingToDo( ) const;
   inline void updateLocation( ) const;

  private:
   string               my_mpChunkSuffix, my_mpChunkType, my_bfChunkSuffix, my_bfChunkType;
   bool                 my_mpLittleEndian, my_bfLittleEndian;

   bool                 my_pghmriMP, my_stdoutMP, my_pghmriBF, my_stdoutBF, my_computeBFMP;
   bool                 my_pghmri, my_stdout, my_initialized;

   PghMRIhandler*       my_mrh;
   const TCreader*      my_tcr;
   ArrayIndexMap        my_aim;
   const SubmodelTable* my_smt;
   vector<LogIntegral*> my_li;
   size_t               my_nullIndex;
   fpVector*            my_prior;

   mutable fpVector               my_mp, my_bf;
   mutable bool                   my_errorMP, my_errorBF, my_mpUptoDate, my_bfUptoDate;
   mutable TCreader::LocationType my_dataLocation;

   static const fpType  our_lnDBLmax;
 };

const string&
ModelProbability::mpChunkSuffix( ) const
 {
  return my_mpChunkSuffix;
 }

const string&
ModelProbability::bfChunkSuffix( ) const
 {
  return my_bfChunkSuffix;
 }

const string&
ModelProbability::mpChunkType( ) const
 {
  return my_mpChunkType;
 }

const string&
ModelProbability::bfChunkType( ) const
 {
  return my_bfChunkType;
 }

bool
ModelProbability::mpLittleEndian( ) const
 {
  return my_mpLittleEndian;
 }

bool
ModelProbability::bfLittleEndian( ) const
 {
  return my_bfLittleEndian;
 }

bool
ModelProbability::anyOutputMP( ) const
 {
  return my_pghmriMP || my_stdoutMP;
 }

bool
ModelProbability::anyOutputBF( ) const
 {
  return my_pghmriBF || my_stdoutBF;
 }

bool
ModelProbability::anyOutput( ) const
 {
  return my_pghmriBF || my_stdoutBF || my_pghmriMP || my_stdoutMP;
 }

bool
ModelProbability::anyErrorsMP( ) const
 {
  return my_errorMP;
 }

bool
ModelProbability::anyErrorsBF( ) const
 {
  return my_errorBF;
 }

bool
ModelProbability::anyErrors( ) const
 {
  return my_errorBF || my_errorMP;
 }

bool
ModelProbability::nothingToDo( ) const
 {
  return !( my_initialized && ( anyOutput( ) || my_computeBFMP ) );
 }

void
ModelProbability::updateLocation( ) const
 {
  my_mpUptoDate = false;
  my_bfUptoDate = false;

  if ( nothingToDo( ) ) return;

  if ( my_dataLocation != my_tcr->locSansT( ) )
   {
    my_dataLocation.resize( my_tcr->dimsSansT( ).size( ) );
    my_dataLocation = my_tcr->locSansT( );
    my_mpUptoDate = true;
    my_bfUptoDate = true;
   }
 }

const fpVector&
ModelProbability::prior( ) const
 {
  return *my_prior;
 }

const fpVector&
ModelProbability::bayesFactors( bool tcIsConst ) const
 {
  updateLocation( ); if ( my_bfUptoDate ) return my_bf;

  my_bf.resize( my_li.size( ) );

  const LogIntegral::quantityType& liNull = my_li[my_nullIndex]->operator()( tcIsConst );

  // cerr << "(debug) ModelProbability::bayesFactors( ) my_nullIndex " << my_nullIndex << " my_li[my_nullIndex] " << my_li[my_nullIndex] << " liNull " << liNull[0] << " anyErrors( ) " << my_li[my_nullIndex]->anyErrors( ) << endl;

  if ( my_li[my_nullIndex]->anyErrors( ) )
   {
    for ( size_t i = 0; i < my_li.size( ); i++ ) my_bf[i] = -DBL_MAX;
    my_errorBF = true;
   }
  else
   {
    my_bf[my_nullIndex] = 0.0;
    for ( size_t i = 0; i < my_li.size( ); i++ )
     if ( i != my_nullIndex )
      {
       // cerr << "(debug) ModelProbability::bayesFactors( ) i " << i << " my_li[i] " << my_li[i];
       const LogIntegral::quantityType& li = my_li[i]->operator()( tcIsConst );
       // cerr << " li " << li[0];
       my_bf[i] = my_li[i]->anyErrors( ) ? ( -DBL_MAX ) : min( our_lnDBLmax, li[0] - liNull[0] );
       // cerr << " my_bf[i] " << my_bf[i] << " anyErrors( ) " << my_li[i]->anyErrors( ) << endl;
       my_errorBF = my_errorBF || my_li[i]->anyErrors( );
      }
   }

  my_bfUptoDate = true; return my_bf;
 }

const fpVector&
ModelProbability::posterior( bool tcIsConst ) const
 {
  updateLocation( ); if ( my_mpUptoDate ) return my_mp;

  bayesFactors( tcIsConst );

  my_mp.resize( my_li.size( ) );

  if ( my_errorBF )
   {
    for ( size_t i = 0; i < my_li.size( ); i++ ) my_mp[i] = -1;
    my_errorMP = true;
   }
  else
   {
    fpType norm = 0.0;
    for ( size_t i = 0; i < my_li.size( ); i++ )
     {
      my_mp[i]  = my_prior->operator[]( i ) * exp( my_bf[i] );
      norm     += my_mp[i];
     }

    if ( norm <= 0.0 )                             // ATTN use numeric threshold
     {
      for ( size_t i = 0; i < my_li.size( ); i++ ) my_mp[i] = -1;
      my_errorMP = true;
     }
    else if ( norm != 1.0 )
     {
      norm = 1.0 / norm;
      for ( size_t i = 0; i < my_li.size( ); i++ ) my_mp[i] *= norm;
     }
   }

  my_mpUptoDate = true; return my_mp;
 }

void
ModelProbability::write( bool tcIsConst, const string& stdoutPrefix ) const
 {
  // cerr << "(debug) ModelProbability::write( ) anyOutput( ) " << anyOutput( ) << " my_computeBFMP " << my_computeBFMP << endl;

  if ( nothingToDo( ) ) return;

  bayesFactors( tcIsConst ); posterior( tcIsConst );

  /*
  cerr << "(debug) ModelProbability::write( ) mp";
  for ( size_t i = 0; i < my_mp.size( ); i++ ) cerr << " " << my_mp[i];
  cerr << endl;

  cerr << "(debug) ModelProbability::write( ) bf";
  for ( size_t i = 0; i < my_bf.size( ); i++ ) cerr << " " << my_bf[i];
  cerr << endl;
  */

  int offset = 0; if ( my_pghmri ) offset = my_aim.offset( my_tcr->locSansT( ) );

  // cerr << "(debug) ModelProbability::write( ) offset " << offset << endl;

  string key; fpType bf, mp; size_t i = 0;
  for ( SubmodelTable::const_iterator si = my_smt->begin( ); si != my_smt->end( ); ++si, i++ )
   {
    key = my_smt->key( si );

    // cerr << "(debug) ModelProbability::write( ) key " << key << endl;

    bf = my_bf[i]; mp = my_mp[i];

    // cerr << "ModelProbability::write( ) key " << key << " bf " << bf << " mp " << mp << endl;

    if ( my_stdoutBF )
     cout << stdoutPrefix << " " << key << " " << my_bfChunkSuffix << " " << bf << endl;

    if ( my_stdoutMP )
     cout << stdoutPrefix << " " << key << " " << my_mpChunkSuffix << " " << mp << endl;

    if ( my_pghmriBF )
     my_mrh->setChunk( key + "." + my_bfChunkSuffix, 1, offset, PghMRIhandler::mriDouble, &bf );

    if ( my_pghmriMP )
     my_mrh->setChunk( key + "." + my_mpChunkSuffix, 1, offset, PghMRIhandler::mriDouble, &mp );
   }
 }

#endif /* __ModelProbability_H__ */
