/* ----------------------------------------------------------------------------
 *
 * Output Products
 *
 * required headers
 *   see include header file products
 *
 * notes
 *  This design is a "one product handler per submodel per product" approach,
 *  ideal for product quantities that may depend on other quantities in the
 *  same submodel, but not substantially on quantities in a different submodel.
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __products_H__
#define __products_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class OutputProduct
 {
  public:

   typedef varray<double> quantityType;

   virtual
   ~OutputProduct( ) = 0;

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

   inline
   const string&
   name( ) const;

   inline
   const string&
   chunkName( ) const;

   inline
   const string&
   chunkFile( ) const;

   inline
   const string&
   diagsName( ) const;

   inline
   const string&
   diagsFile( ) const;

   inline
   PghMRIhandler::mriArrayType
   defaultType( ) const;

   inline
   size_t
   numLeadDims( ) const;

   inline
   const vector<string>&
   productsNeeded( ) const;

   inline
   const TCreader*
   dataNavigator( ) const;

   inline
   const ParameterManager<fpType>&
   parameterManager( ) const;

   inline
   const MinQNB&
   minimizer( ) const;

   inline
   const dbProfile&
   drift( ) const;

   inline
   const acProfile&
   activation( ) const;

   inline
   const fpType*
   paramBuf( ) const;

   inline
   const TimeGrids&
   timeGrids( ) const;

   inline
   PghMRIhandler&
   mriHandler( ) const;

   inline
   bool
   isConfigured( ) const;

   inline
   bool
   hasProductsNeeded( ) const;

   inline
   bool
   isInitialized( ) const;

   inline
   bool
   stdout( ) const;                            // on/off status of stdout output

   inline
   bool
   pghmri( ) const;              // on/off status of Pittburgh MRI format output

   inline
   bool
   diagnostics( ) const;          // on/off status of product diagnostics output

   inline
   bool
   anyOutput( ) const;         // ( pghmri( ) || stdout( ) ) && isInitialized( )

   inline
   bool
   anyErrors( ) const;        // encountered any errors in product computations?

  // ------------------------------------- create all required product instances
  // ---------------------------------------------------------------------------

   bool
   createProductsNeeded( vector<OutputProduct*>& products, bool own = true );
                   //
                   // creates instances of  all products  needed by this product
                   // (if they do not exist  in products  already),  append  any
                   // pointers to products, renounce ownership if  own == false.
                   // If owner, the destructor will delete all product instances
                   // it owns. Output modes of created product instances are set
                   // as pghmri == stdout == false. Two paths for initialization
                   // for  created product instances  are envisioned:  all owned
                   // instances get configured and initialized in initialize( ),
                   // all  non-owned instances are  configured  and  initialized
                   // elsewhere, e. g., explicitly by the caller.

  // --------------------------------------------------------- configure product
  // ---------------------------------------------------------------------------

   void
   configure( const TCreader*                 dataNavigator,
              const ParameterManager<fpType>* parameterManager,
              const MinQNB*                   minimizer,
              const acProfile*                activation,
                    fpVector*                 paramBuf,
              const TimeGrids*                timeGrids );

   void
   setOutput( bool stdout = false, bool pghmri = true );
                       //
                       // stdout: write to stdout (for debug or few-voxel runs),
                       // pghmri: write using the Pittburgh MRI handler. turning
                       // both off is useful when  product quantity  needs to be
                       // computed (as part of another product) but not written.

  // -------------------------------------------------------- initialize product
  // ---------------------------------------------------------------------------

   virtual
   void
   initialize(       PghMRIhandler*              mriHandler,
               const string&                     chunkPrefix,
                     PghMRIhandler::mriArrayType chunkType    = -1,
               const string&                     leadDims     = "",
                     bool                        diagnostics  = false,
                     bool                        littleEndian = PghMRIhandler::endianness( ) );
                                      //
                                      // open mode  same as that of  mriHandler;
                                      // mriModify == mriWrite for this purpose;
                                      // mriRead dsallowed. appropriate defaults
                                      // for  chunk* and leadDims arguments  are
                                      // used if blank.
                                      //
                                      //     chunk name = chunk prefix + name( )
                                      //     chunk file = chunk name   +  ".dat"
                                      //
                                      // if !pghmri( ), then no  initializations
                                      // are performed. Conversely, initialize()
                                      // must be  invoked after  setting  pghmri
                                      // output on.
                                      //
                                      // for all owned product instances created
                                      // by createProductsNeeded( ), configure()
                                      // and  initialize( )  are  invoked   with
                                      // appropriate settings.  Output modes for
                                      // these are set to pghmri==stdout==false.

   virtual
   void
   subInitialize( const string& chunkPrefix );
               //
               // additional mechanism for any product-specific initializations;
               // invoked  by  initialize( ... )  after common  initializations.

  // -------------------------- product quantity update at current data location
  // ---------------------------------------------------------------------------

   inline
   const quantityType&
   operator()( bool tcIsConst = false ) const;
                                        //
                                        // check error status using anyErrors( )

   inline
   const TCreader::LocationType&
   whereAmI( ) const;
                //
                // current data location where product quantity was updated last

   virtual
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const = 0;
                               //
                               // product-specific computations:  store  product
                               // quantity in the first argument. return success.

  // ---------------------------------------------------- write product quantity
  // ---------------------------------------------------------------------------

   inline
   virtual
   void
   write( bool tcIsConst = false, const string& stdoutPrefix = "" ) const;
                 //
                 // if tcIsConst == true, no errors are reported for this voxel.

  // ---------------------------------------------------------- finalize product
  // ---------------------------------------------------------------------------

   virtual
   void
   finalize( );
                                  //
                                  // any product finalizations, but not close( )

   virtual
   void
   subFinalize( );
                     //
                     // additional mechanism for product-specific finalizations;
                     // invoked  by  finalize( )  after  common   finalizations.

  protected:

   OutputProduct( const OutputProduct& another );

   OutputProduct&
   operator=( const OutputProduct& rhs );

   OutputProduct( const string&                     name,
                        size_t                      numLeadDims         = 1,
                        PghMRIhandler::mriArrayType defaultType         = PghMRIhandler::mriDouble,
                  const vector<string>&             productsNeeded      = vector<string>( ),
                        bool                        disablePghMRIoutput = false );

   void
   setNeeds( const vector<string>& productsNeeded = vector<string>( ) );
                           //
                           // list of other products needed by that this product

   virtual
   uintVector
   leadSizes( ) const = 0;
                    //
                    // if  ( leadSizes( ).size( ) < my_numLeadDims ),   all  the
                    // remaining leading dimensions are assumed to have extent 1

   inline
   const vector<OutputProduct*>&
   needs( ) const;
                                              //
                                              // products needed by this product

   inline
   bool
   needsUpdate( ) const;
         //
         // true if cached-in product quantity is outdated w.r.t. data navigator

   inline
   TCreader::LocationType
   quantityLocation( const TCreader::LocationType& dataLocation ) const;
              //
              // first my_numLeadDims quantity indices are added to dataLocation

  private:

	   string                      my_name, my_chunkName, my_chunkFile, my_diagsName, my_diagsFile;
           size_t                      my_numLeadDims;
	   PghMRIhandler::mriArrayType my_defaultType;
           vector<string>              my_needs;
           vector<OutputProduct*>      my_products;
           vector<bool>                my_own;

   const   TCreader*                   my_dn;
   const   ParameterManager<fpType>*   my_pm;
   const   MinQNB*                     my_mn;
   const   acProfile*                  my_ac;
           fpVector*                   my_pb;
   const   TimeGrids*                  my_tg;

           PghMRIhandler*              my_mrh;

           bool                        my_configured, my_initialized;
           bool                        my_stdout, my_pghmri, my_disablePghMRI, my_diags;

	   ArrayIndexMap               my_aim, my_diagsAIM;

   mutable bool                        my_error;
   mutable quantityType                my_quantity;
   mutable TCreader::LocationType      my_dataLocation;
 };

struct ProductRegistry
 {
  static
  void
  expand( vector<string>& expanded,
          const vector<string>& productList = vector<string>( 1, "default.products" ) );
                     //
                     // Appends  an expanded list of individual product names to
                     // the first argument, given another list (second argument)
                     // that contains either individual product names  or  group
                     // names "default.products", "all.products", "misc.products"
                     // and  "diagnostics.products".

  static
  bool
  isKnown( const string& productName );
                                           //
                                           // true if productName is recognized,
                                           // same  as  create(  productName  );

  static
  bool
  create( const string& productName, OutputProduct*& p, bool testOnly = true );
                    //
                    // main  registry  of  inidividual  products.  If !testOnly,
                    // create instance of named product and return pointer in p.

  static
  bool
  add( const string& productName, vector<OutputProduct*>& products );
                               //
                               // create named product instance if known, append
                               // pointer to products, report success. No  other
                               // initializations performed.

  static
  bool
  del( const string& productName, vector<OutputProduct*>& products );
                                      //
                                      // finalize( )  and  delete  named product
                                      // instance from products, report success.

  inline
  static
  bool
  has( const string& productName, const vector<OutputProduct*>& products );
                                 //
                                 // if named product instance exists in products

  inline
  static
  OutputProduct*
  extract( const string& productName, const vector<OutputProduct*>& products );
                //
                // pointer to instance of named product if it exists in products
 };

class ProductSet : public OutputProduct
 {
  public:

   ProductSet( );

   ~ProductSet( );

   ProductSet( const ProductSet& another );
                                //
                                // ownership of the product instances created by
                                // this  product set is transferred to the copy.

   ProductSet&
   operator=( const ProductSet& rhs );
                                //
                                // ownership of the product instances created by
                                // this  product set is transferred to the copy.

   bool
   add( const string& productName );
                              //
                              // add named product instance to  this product set
                              // along with all products needed by this product;
                              // report success.  No  initializations performed.

   bool
   del( const string& productName );
                //
                // finalize( ) and delete named product instance; report success

   inline
   bool
   has( const string& productName ) const;
                            //
                            // if named product instance exists in this register

   inline
   OutputProduct*
   operator()( const string& productName ) const;
           //
           // pointer to instance of named product if it exists in this register

   inline
   const vector<OutputProduct*>&
   operator()( ) const;
                                     //
                                     // complete collection of product instances

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

   inline
   void
   initialize(       PghMRIhandler*              mriHandler,
               const string&                     chunkPrefix,
                     PghMRIhandler::mriArrayType chunkType    = -1,
               const string&                     leadDims     = "",
                     bool                        diagnostics  = false,
                     bool                        littleEndian = PghMRIhandler::endianness( ) );
                   //
                   // does nothing; initialize products in this set individually
                   // by  accessing  them  using   ProductSet::operator()( ... )

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;
                      //
                      // does nothing; compute an individual product in this set
                      // by  accessing it  using   ProductSet::operator()( ... )

   inline
   void
   write( bool tcIsConst = false, const string& stdoutPrefix = "" ) const;
                //
                // this will invoke write( ) for individual products in this set

   inline
   void
   finalize( );
                                  //
                                  // invoke finalize( ) for individual products.

   void
   setStdoutPrefix( const string& prefix = "" ) const;
                       //
                       // prefix for every write to stdout when stdout() is true

  protected:

   uintVector
   leadSizes( ) const { return uintVector( ); }

  private:

   mutable vector<OutputProduct*> my_products;
 };

class Iterations : public OutputProduct
 {
  public:
   Iterations( );

   Iterations( const Iterations& another );

   Iterations&
   operator=( const Iterations& rhs );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class Evaluations : public OutputProduct
 {
  public:
   Evaluations( );

   Evaluations( const Evaluations& another );

   Evaluations&
   operator=( const Evaluations& rhs );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class Restarts : public OutputProduct
 {
  public:
   Restarts( );

   Restarts( const Restarts& another );

   Restarts&
   operator=( const Restarts& rhs );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class Status : public OutputProduct
 {
  public:
   Status( );

   Status( const Status& another );

   Status&
   operator=( const Status& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class PrGradNorm : public OutputProduct
 {
  public:
   PrGradNorm( );

   PrGradNorm( const PrGradNorm& another );

   PrGradNorm&
   operator=( const PrGradNorm& rhs );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class Gradient : public OutputProduct
 {
  public:
   Gradient( );

   Gradient( const Gradient& another );

   Gradient&
   operator=( const Gradient& rhs );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class Estimates : public OutputProduct
 {
  public:
   Estimates( );

   Estimates( const Estimates& another );

   Estimates&
   operator=( const Estimates& rhs );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class Baseline : public OutputProduct
 {
  public:
   Baseline( );

   Baseline( const Baseline& another );

   Baseline&
   operator=( const Baseline& rhs );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class LogPostMax : public OutputProduct
 {
  public:
   LogPostMax( );

   LogPostMax( const LogPostMax& another );

   LogPostMax&
   operator=( const LogPostMax& rhs );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class DBprofile : public OutputProduct
 {
  public:
                                            // drift profile over the base grid,
                                            // including instants marked missing

   DBprofile( );

   DBprofile( const DBprofile& another );

   DBprofile&
   operator=( const DBprofile& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   dbProfile my_dbp;

   mutable quantityType my_quantity;
 };

class ACprofile : public OutputProduct
 {
  public:
                                     // activation  profile  over the base grid;
                                     // i. e., including instants marked missing

   ACprofile( );

   ACprofile( const ACprofile& another );

   ACprofile&
   operator=( const ACprofile& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   acProfile my_acp;

   mutable quantityType my_quantity;
 };

class Fits : public OutputProduct
 {
  public:
                                                // model fits over the base grid

   Fits( );

   Fits( const Fits& another );

   Fits&
   operator=( const Fits& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   ACprofile* my_acp;
   DBprofile* my_dbp;

   mutable quantityType my_quantity;
 };

class Residuals : public OutputProduct
 {
  public:

        // residuals are computed over the base grid consistent with  AR(n)  but
        // disregarding any MA(p) parameters (first n residuals are set to zero)

   Residuals( );

   Residuals( const Residuals& another );

   Residuals&
   operator=( const Residuals& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   int my_numAR, my_arOfs;

   Fits* my_fits;

   mutable quantityType my_quantity;
 };

class Hessian : public OutputProduct
 {
  public:

                // hessian matrix in a packed upper triangular columnwise format

   Hessian( );

   Hessian( const Hessian& another );

   Hessian&
   operator=( const Hessian& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   mutable quantityType my_quantity;
 };

class Covariances : public OutputProduct
 {
  public:

             // covariance matrix in a packed upper triangular columnwise format

   Covariances( double defaultOnError = -1 );
                         //
                         // When the covariance calculation encounters an error,
                         // all  the covariances  are are set to defaultOnError.

   Covariances( const Covariances& another );

   Covariances&
   operator=( const Covariances& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   double  my_default;
   Hessian* my_hessian;

   mutable FortranTypes::integer my_info1, my_info2;
   mutable quantityType my_quantity;
 };

class StdErrors : public OutputProduct
 {
  public:

   StdErrors( double defaultOnError = -1, double threshold = 1.0e-8 );
                        //
                        // When the covariance calculation encounters any error,
                        // all the standard errors are set to defaultOnError.
                        //
                        // any |variance| <= threshold will be taken to be zero.
                        // for any  variance < -threshold,  standard error := -1

   StdErrors( const StdErrors& another );

   StdErrors&
   operator=( const StdErrors& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   double my_default, my_threshold;
   Covariances* my_cov;

   mutable quantityType my_quantity;
 };

class BaselineStdErr : public OutputProduct
 {
  public:

   BaselineStdErr( double defaultOnError = -1, double threshold = 1.0e-8 );
                        //
                        // When the covariance calculation encounters any error,
                        // all the standard errors are set to defaultOnError.
                        //
                        // any |variance| <= threshold will be taken to be zero.
                        // for any  variance < -threshold,  standard error := -1

   BaselineStdErr( const BaselineStdErr& another );

   BaselineStdErr&
   operator=( const BaselineStdErr& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   double my_default, my_threshold;
   Covariances* my_cov;

   mutable quantityType my_quantity;
 };

class LogDeterminant : public OutputProduct
 {
  public:

   LogDeterminant( double defaultOnError = log( DBL_MIN ) );

   LogDeterminant( const LogDeterminant& another );

   LogDeterminant&
   operator=( const LogDeterminant& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   double my_default;
   Hessian* my_hessian;

   mutable FortranTypes::integer my_info;
   mutable quantityType my_quantity;
 };

class FitPositiveDefinite : public OutputProduct
 {
  public:

   FitPositiveDefinite( );

   FitPositiveDefinite( const FitPositiveDefinite& another );

   FitPositiveDefinite&
   operator=( const FitPositiveDefinite& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   LogDeterminant* my_logdet;
   mutable quantityType my_quantity;
 };

class LogIntegral : public OutputProduct
 {
  public:

   LogIntegral( double defaultOnError = -DBL_MAX );

   LogIntegral( const LogIntegral& another );

   LogIntegral&
   operator=( const LogIntegral& rhs );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   double my_default;
   LogDeterminant* my_logdet;
   LogPostMax* my_logpostmax;
   mutable quantityType my_quantity;

   static
   const double our_ln2pi;
 };

/* =============================================================================
class LogBayesFactor : public OutputProduct
 {
  public:

   LogBayesFactor( double defaultOnError = -DBL_MAX );

   LogBayesFactor( const LogBayesFactor& another );

   LogBayesFactor&
   operator=( const LogBayesFactor& rhs );

   void
   setNullLI( const LogIntegral* nullSubmodelLI );

   void
   subInitialize( const string& chunkPrefix );

   inline
   bool
   compute( quantityType& quantity, const bool tcIsConst = false ) const;

  protected:

   inline
   uintVector
   leadSizes( ) const;

  private:

   double my_default;
   LogIntegral* my_logint;
   const LogIntegral* my_logintNull;
   static const double our_lnDBLmax;
   mutable quantityType my_quantity;
 };
   ============================================================================= */

// --------------------------------------------------------------- OutputProduct
// -----------------------------------------------------------------------------

const string&
OutputProduct::name( ) const
 {
  return my_name;
 }

const string&
OutputProduct::chunkName( ) const
 {
  return my_chunkName;
 }

const string&
OutputProduct::chunkFile( ) const
 {
  return my_chunkFile;
 }

const string&
OutputProduct::diagsName( ) const
 {
  return my_diagsName;
 }

const string&
OutputProduct::diagsFile( ) const
 {
  return my_diagsFile;
 }

PghMRIhandler::mriArrayType
OutputProduct::defaultType( ) const
 {
  return my_defaultType;
 }

size_t
OutputProduct::numLeadDims( ) const
 {
  return my_numLeadDims;
 }

const vector<string>&
OutputProduct::productsNeeded( ) const
 {
  return my_needs;
 }

const TCreader*
OutputProduct::dataNavigator( ) const
 {
  return my_dn;
 }

const ParameterManager<fpType>&
OutputProduct::parameterManager( ) const
 {
  return *my_pm;
 }

const MinQNB&
OutputProduct::minimizer( ) const
 {
  return *my_mn;
 }

const dbProfile&
OutputProduct::drift( ) const
 {
  return my_ac->drift( );
 }

const acProfile&
OutputProduct::activation( ) const
 {
  return *my_ac;
 }

const fpType*
OutputProduct::paramBuf( ) const
 {
  return (*my_pb)( 0 );
 }

const TimeGrids&
OutputProduct::timeGrids( ) const
 {
  return *my_tg;
 }

PghMRIhandler&
OutputProduct::mriHandler( ) const
 {
  return *my_mrh;
 }

bool
OutputProduct::isConfigured( ) const
 {
  return my_configured;
 }

bool
OutputProduct::hasProductsNeeded( ) const
 {
  bool success = ( my_needs.size( ) == my_products.size( ) ) && ( my_needs.size( ) == my_own.size( ) );

  for ( size_t i = 0; success && ( i < my_needs.size( ) ); i++ )
   success = success && ProductRegistry::has( my_needs[i], my_products );

  // for ( size_t i = 0; ( i < my_needs.size( ) ) && success; i++ )
  //  success = success && my_products[i]->hasProductsNeeded( );

  return success;
 }

bool
OutputProduct::isInitialized( ) const
 {
  return my_initialized;
 }

bool
OutputProduct::stdout( ) const
 {
  return my_stdout;
 }

bool
OutputProduct::pghmri( ) const
 {
  return my_pghmri;
 }

bool
OutputProduct::diagnostics( ) const
 {
  return my_diags && ( my_pghmri || my_stdout );
 }

bool
OutputProduct::anyOutput( ) const
 {
  return my_initialized && ( my_stdout || my_pghmri );
 }

bool
OutputProduct::anyErrors( ) const
 {
  return my_error;
 }

// ---------------------------------- product quantity and current data location
// -----------------------------------------------------------------------------

const OutputProduct::quantityType&
OutputProduct::operator()( bool tcIsConst ) const
 {
  assert( my_initialized );

  // cerr << "(debug) OutputProduct::operator()( ) product " << my_name << " location " << my_dn->locationStr( ",", my_dn->dimsSansT( ) ) << " initialized " << my_initialized << " stdout " << my_stdout << " pghmri " << my_pghmri << " diags " << my_diags << " needsUpdate( ) " << needsUpdate( ) << endl;

  if ( needsUpdate( ) )
   {
    my_dataLocation.resize( my_dn->dimsSansT( ).size( ) );
    my_dataLocation = my_dn->locSansT( );
    my_error = !compute( my_quantity, tcIsConst );
   }

  return my_quantity;
 }

const TCreader::LocationType&
OutputProduct::whereAmI( ) const
 {
  return my_dataLocation;
 }

// ------------------------------------------------------ write product quantity
// -----------------------------------------------------------------------------

void
OutputProduct::write( bool tcIsConst, const string& stdoutPrefix ) const
 {
  assert( my_initialized );

  // cerr << "(debug) OutputProduct::write( ) product " << my_name << " location " << my_dn->locationStr( ",", my_dn->dimsSansT( ) ) << " initialized " << my_initialized << " stdout " << my_stdout << " pghmri " << my_pghmri << " diags " << my_diags << endl;

  operator()( tcIsConst );

  /*
  cerr << "(debug) OutputProduct::write( ) product " << my_name << " anyErrors( ) " << anyErrors( ) << " my_quantity.size( ) " << my_quantity.size( ) << endl;

  TCreader::LocationType ql = quantityLocation( my_dataLocation );
 
  cerr << "(debug) OutputProduct::write( ) my_dataLocation";
  for ( size_t i = 0; i < my_dataLocation.size( ); i++ ) cerr << " " << my_dataLocation[i];
  cerr << endl;

  cerr << "(debug) OutputProduct::write( ) ql";
  for ( size_t i = 0; i < ql.size( ); i++ ) cerr << " " << ql[i];
  cerr << endl;

  cerr << "(debug) OutputProduct::write( ) offsets " << my_aim.offset( ql ) << " " << my_diagsAIM.offset( my_dataLocation ) << endl;
  */

  if ( !anyOutput( ) ) return;

  if ( my_stdout )
   {
    if ( ( my_defaultType == PghMRIhandler::mriFloat ) || ( my_defaultType == PghMRIhandler::mriDouble ) )
     MiscFns::writeArray( my_quantity, cout, stdoutPrefix + " " + my_name, true, 21, 15, 4 );
    else if ( my_defaultType == PghMRIhandler::mriUchar )
     {
      varray<short> q( my_quantity.size( ) );
      for ( size_t i = 0; i < q.size( ); i++ ) q[i] = cpp_static_cast( short, my_quantity[i] );
      MiscFns::writeArray( q, cout, stdoutPrefix + " " + my_name, false, 0, 0, q.size( ) );
     }
    else
     MiscFns::writeArray( my_quantity, cout, stdoutPrefix + " " + my_name, false, 0, 0, my_quantity.size( ) );
   }

  if ( my_pghmri )
   my_mrh->setChunk( my_chunkName, my_quantity.size( ),
                     my_aim.offset( quantityLocation( my_dataLocation ) ),
                     PghMRIhandler::mriDouble, cpp_static_cast( void*, my_quantity( ) ) );

  if ( !diagnostics( ) ) return;

  if ( my_stdout )
   cout << stdoutPrefix << " " << name( ) << " status " << my_error << endl;

  if ( my_pghmri )
   {
    unsigned char s = my_error;
    my_mrh->setChunk( my_diagsName, 1, my_diagsAIM.offset( my_dataLocation ),
                      PghMRIhandler::mriUchar, cpp_static_cast( void*, &s ) );
   }
 }

// ------------------------------------------------------- protected functions
// ---------------------------------------------------------------------------

const vector<OutputProduct*>&
OutputProduct::needs( ) const
 {
  return my_products;
 }

bool
OutputProduct::needsUpdate( ) const
 {
  return ( !my_initialized || ( my_dataLocation != my_dn->locSansT( ) ) );
 }

TCreader::LocationType
OutputProduct::quantityLocation( const TCreader::LocationType& dataLocation ) const
 {
  TCreader::LocationType l( dataLocation.size( ) + my_numLeadDims );
  for ( size_t i = 0, j = my_numLeadDims; i < dataLocation.size( ); i++, j++ )
   l[j] = dataLocation[i];
  return l;
 }

// ------------------------------------------------------------ product registry
// -----------------------------------------------------------------------------

bool
ProductRegistry::has( const string& productName, const vector<OutputProduct*>& products )
 {
  // vector<OutputProduct*>::const_iterator
  // i = find_if( products.begin( ), products.end( ), nameIs<OutputProduct>( productName ) );
  // return ( i != products.end( ) );

  return
   (    find_if( products.begin( ), products.end( ), nameIs<OutputProduct>( productName ) )
     != products.end( ) );
 }

OutputProduct*
ProductRegistry::extract( const string& productName, const vector<OutputProduct*>& products )
 {
  vector<OutputProduct*>::const_iterator
  i = find_if( products.begin( ), products.end( ), nameIs<OutputProduct>( productName ) );
  return ( i == products.end( ) ) ? 0 : (*i);
 }

// ------------------------------------------------------------------ ProductSet
// -----------------------------------------------------------------------------

bool
ProductSet::has( const string& productName ) const
 {
  return ProductRegistry::has( productName, my_products );
 }

OutputProduct*
ProductSet::operator()( const string& productName ) const
 {
  return ProductRegistry::extract( productName, my_products );
 }

const vector<OutputProduct*>&
ProductSet::operator()( ) const
 {
  return my_products;
 }

void
ProductSet::initialize(       PghMRIhandler*              mriHandler,
                        const string&                     chunkPrefix,
                              PghMRIhandler::mriArrayType chunkType,
                        const string&                     leadDims,
                              bool                        diagnostics,
                              bool                        littleEndian )
 {
 }

bool
ProductSet::compute( quantityType& quantity, bool tcIsConst ) const
 {
  return true;
 }

void
ProductSet::write( bool tcIsConst, const string& stdoutPrefix ) const
 {
  for ( size_t i = 0; i < my_products.size( ); i++ )
   {
    // cerr << "(debug) ProductSet::write( ) " << my_products[i]->name( ) << " stdout " << my_products[i]->stdout( ) << " pghmri " << my_products[i]->pghmri( ) << endl;

    my_products[i]->write( tcIsConst, stdoutPrefix );
   }
 }

void
ProductSet::finalize( )
 {
  for ( size_t i = 0; i < my_products.size( ); i++ ) my_products[i]->finalize( );
 }

// ----------------------------------------------------- diagnostics: iterations
// -----------------------------------------------------------------------------

uintVector
Iterations::leadSizes( ) const
 {
  return uintVector( );
 }

bool
Iterations::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1 ); quantity[0] = minimizer( ).numIterations( );
  return true;
 }

// ---------------------------------------------------- diagnostics: evaluations
// -----------------------------------------------------------------------------

uintVector
Evaluations::leadSizes( ) const
 {
  return uintVector( );
 }

bool
Evaluations::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1 ); quantity[0] = minimizer( ).numEvals( );
  return true;
 }

// ------------------------------------------------------- diagnostics: restarts
// -----------------------------------------------------------------------------

uintVector
Restarts::leadSizes( ) const
 {
  return uintVector( );
 }

bool
Restarts::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1 ); quantity[0] = minimizer( ).numRestarts( );
  return true;
 }

// --------------------------------------------------------- diagnostics: status
// -----------------------------------------------------------------------------

uintVector
Status::leadSizes( ) const
 {
  return uintVector( );
 }

bool
Status::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1 ); quantity[0] = tcIsConst ? 17 : minimizer( ).statusCode( );
  return true;
 }

// ---------------------------------------- diagnostics: projected gradient norm
// -----------------------------------------------------------------------------

uintVector
PrGradNorm::leadSizes( ) const
 {
  return uintVector( );
 }

bool
PrGradNorm::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1 ); quantity[0] = minimizer( ).prGradNorm( paramBuf( ) );
  return true;
 }

// ------------------------------------------------------- diagnostics: gradient
// -----------------------------------------------------------------------------

uintVector
Gradient::leadSizes( ) const
 {
  uintVector ls( 1 );
  ls[0] = parameterManager( ).numParams( );
  return ls;
 }

bool
Gradient::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( minimizer( ).problemSize( ) );
  minimizer( ).objectiveFn( )->gradient( paramBuf( ), quantity( ) );
  return true;
 }

// ------------------------------------------------------- estimates: parameters
// -----------------------------------------------------------------------------

uintVector
Estimates::leadSizes( ) const
 {
  uintVector ls( 1 );
  ls[0] = parameterManager( ).numParams( );
  return ls;
 }

bool
Estimates::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( parameterManager( ).numParams( ) );
  for ( size_t i = 0; i < quantity.size( ); i++ ) quantity[i] = paramBuf( )[i];
  return true;
 }

// --------------------------------------------------------- estimates: baseline
// -----------------------------------------------------------------------------

uintVector
Baseline::leadSizes( ) const
 {
  return uintVector( );
 }

bool
Baseline::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1 );
  quantity[0] = drift( ).baseline( paramBuf( ) + parameterManager( ).blockOffset( ParamBlkNames::drift ) );
  return true;
 }

// ---------------------------------------------------- log posterior at maximum
// -----------------------------------------------------------------------------

uintVector
LogPostMax::leadSizes( ) const
 {
  return uintVector( );
 }

bool
LogPostMax::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1 );

  minimizer( ).objectiveFn( )->value( paramBuf( ), quantity( 0 ) );
  quantity[0] /= minimizer( ).objectiveFn( )->sign( );
  quantity[0] += minimizer( ).objectiveFn( )->additiveConst( );
          //
          // + log posterior at mode (including normalization-related constants)

  return true;
 }

// ---------------------------------------------------- base-grid profile: drift
// -----------------------------------------------------------------------------

uintVector
DBprofile::leadSizes( ) const
 {
  uintVector ls( 1 );
  ls[0] = timeGrids( ).baseGrid( ).size( );
  return ls;
 }

bool
DBprofile::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( my_dbp.timeGridSize( ) );
  quantity = my_dbp.profile( paramBuf( ) + parameterManager( ).blockOffset( ParamBlkNames::drift ) );
  return true;
 }

// ----------------------------------------------- base-grid profile: activation
// -----------------------------------------------------------------------------

uintVector
ACprofile::leadSizes( ) const
 {
  uintVector ls( 1 );
  ls[0] = timeGrids( ).baseGrid( ).size( );
  return ls;
 }

bool
ACprofile::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( my_acp.timeGridSize( ) );
  quantity = my_acp.profile( paramBuf( ) + parameterManager( ).blockOffset( ParamBlkNames::drift    ),
                                paramBuf( ) + parameterManager( ).blockOffset( ParamBlkNames::response ),
                                paramBuf( ) + parameterManager( ).blockOffset( ParamBlkNames::shape    ) );

  return true;
 }

// ----------------------------------------------------- base-grid profile: fits
// -----------------------------------------------------------------------------

uintVector
Fits::leadSizes( ) const
 {
  uintVector ls( 1 );
  ls[0] = timeGrids( ).baseGrid( ).size( );
  return ls;
 }

bool
Fits::compute( quantityType& quantity, bool tcIsConst ) const
 {
  const quantityType& a = my_acp->operator()( tcIsConst );
  quantity.resize( a.size( ) ); quantity = a; quantity += my_dbp->operator()( tcIsConst );
  return true;
 }

// ------------------------------------------------ base-grid profile: residuals
// -----------------------------------------------------------------------------

uintVector
Residuals::leadSizes( ) const
 {
  uintVector ls( 1 ); ls[0] = timeGrids( ).baseGrid( ).size( );
  return ls;
 }

bool
Residuals::compute( quantityType& quantity, bool tcIsConst ) const
 {
  const Fits::quantityType& f = my_fits->operator()( tcIsConst );

  quantity.resize( f.size( ) );
  quantity = dataNavigator( )->timeCourse( );
  quantity -= f;
  ARutils::getARvec( my_numAR, paramBuf( ) + my_arOfs, quantity );

  return true;
 }

// ---------------------------------------------------------- hessian at minimum
// -----------------------------------------------------------------------------

uintVector
Hessian::leadSizes( ) const
 {
  int n = parameterManager( ).numParams( );
  uintVector ls( 1 ); ls[0] = n * ( n + 1 ) / 2;
  return ls;
 }

bool
Hessian::compute( quantityType& quantity, bool tcIsConst ) const
 {
  matrix<fpType> h; minimizer( ).objectiveFn( )->hessian( paramBuf( ), h );
  // cerr << "(debug) Hessian::compute( ) h.numRows( ) " << h.numRows( ) << " h.numCols( ) " << h.numCols( ) << endl;
  BlasDrivers::pack( h, quantity, cppblas::UpperTriangle );
  // cerr << "(debug) Hessian::compute( ) quantity.size( ) " << quantity.size( ) << endl;
  return ( quantity.size( ) == leadSizes( )[0] );
 }

// ----------------------------------------------------------------- covariances
// -----------------------------------------------------------------------------

uintVector
Covariances::leadSizes( ) const
 {
  int n = parameterManager( ).numParams( );
  uintVector ls( 1 ); ls[0] = n * ( n + 1 ) / 2;
  return ls;
 }

bool
Covariances::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( leadSizes( )[0], my_default );

  /*
  size_t n = 0;
  cerr << "(debug) Covariances::compute( ) leadSizes( )[0] " << leadSizes( )[0] << endl;
  cerr << "(debug) Covariances::compute( ) quantity.size( ) " << quantity.size( )
       << " check " << BlasDrivers::checkPackedSize( quantity.size( ), n ) << " n " << n << endl;
  */

  const Hessian::quantityType& h = my_hessian->operator()( tcIsConst );
  // cerr << "(debug) Covariances::compute( ) my_hessian->anyErrors( ) " << my_hessian->anyErrors( ) << " h.size( ) " << h.size( ) << endl;
  if ( my_hessian->anyErrors( ) ) return false;

  my_info1 = my_info2 = 0;
  bool error = !LapackDrivers::RealSymmetric::invert( h, quantity, my_info1, my_info2, cppblas::UpperTriangle );

  // cerr << "(debug) Covariances::compute( ) info " << ( my_info1||my_info2 ) << " error " << error << " quantity.size( ) " << quantity.size( ) << endl;

  if ( error || my_info1 || my_info2 ) { quantity.resize( leadSizes( )[0], my_default ); return false; }

  return true;
 }

// ------------------------------------------------------------- standard errors
// -----------------------------------------------------------------------------

uintVector
StdErrors::leadSizes( ) const
 {
  uintVector ls( 1 ); ls[0] = parameterManager( ).numParams( );
  return ls;
 }

bool
StdErrors::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( parameterManager( ).numParams( ), my_default );

  const Covariances::quantityType& c = my_cov->operator()( tcIsConst );
  if ( my_cov->anyErrors( ) ) return false;

  BlasDrivers::unpack_diag( c, quantity, cppblas::UpperTriangle );

  for ( size_t i = 0; i < quantity.size( ); i++ )
   quantity[i] =   ( quantity[i] >= 0 ) ? sqrt( quantity[i] )
                 : ( ( fabs( quantity[i] ) <= my_threshold ) ? 0 : my_default );

  return true;
 }

// ----------------------------------------------------- baseline standard error
// -----------------------------------------------------------------------------

uintVector
BaselineStdErr::leadSizes( ) const
 {
  return uintVector( );
 }

bool
BaselineStdErr::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1, my_default );

  const Covariances::quantityType& c = my_cov->operator()( tcIsConst );
  if ( my_cov->anyErrors( ) ) return false;

  const fpVector& db = drift( ).Dbaseline( paramBuf( ) + parameterManager( ).blockOffset( ParamBlkNames::drift ) );

  quantityType a( db.size( ) );
  BlasDrivers::spmv( db.size( ), c( ), db( ), a( ), 1.0, 0.0, 1, 1, cppblas::UpperTriangle );
      //
      // ATTN it is  assumed here  that  the drift-drift block  in  the packed
      // ATTN covariance matrix is right at the beginning of c. In other words,
      // ATTN parameterManager( ).blockOffset( ParamBlkNames::drift ) == 0.

  fpType q = BlasDrivers::dot( db, a );

  if ( q >= 0 ) { quantity[0] = sqrt( q ); return true; }
  if ( fabs( q ) <= my_threshold ) { quantity[0] = 0; return true; }

  return false;
 }

// ------------------------------------------------------------- log determinant
// -----------------------------------------------------------------------------

uintVector
LogDeterminant::leadSizes( ) const
 {
  return uintVector( );
 }

bool
LogDeterminant::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1, my_default );

  const Hessian::quantityType& H = my_hessian->operator()( tcIsConst );
  if ( my_hessian->anyErrors( ) ) return false;

  varray<FortranTypes::dReal> h;
  bool error = !LapackDrivers::RealSymmetric::cholesky( H, h, my_info, cppblas::UpperTriangle );

  if ( my_info || error ) return false;

  varray<double> d; BlasDrivers::unpack_diag( h, d, cppblas::UpperTriangle );

  double lndet = 0;
  for ( size_t i = 0; i < d.size( ); i++ ) lndet += 2.0 * log( fabs( d[i] ) );
                                    //
                                    // determinant of hessian of - log posterior

  quantity[0] = -lndet;                                 // for *inverse* hessian

  return true;
 }

// -------------------------------------------- positive definiteness of the fit
// -----------------------------------------------------------------------------

uintVector
FitPositiveDefinite::leadSizes( ) const
 {
  return uintVector( );
 }

bool
FitPositiveDefinite::compute( quantityType& quantity, bool tcIsConst ) const
 {
  my_logdet->operator()( tcIsConst );
  quantity.resize( 1, !my_logdet->anyErrors( ) );
  return true;
 }

// ---------------------------------------------------------------- log integral
// -----------------------------------------------------------------------------

uintVector
LogIntegral::leadSizes( ) const
 {
  return uintVector( );
 }

bool
LogIntegral::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1, my_default );

  const quantityType& d = my_logdet->operator()( tcIsConst );
  if ( my_logdet->anyErrors( ) ) return false;

  const quantityType& p = my_logpostmax->operator()( tcIsConst );
  if ( my_logpostmax->anyErrors( ) ) return false;

  quantity[0] = p[0] + 0.5 * ( parameterManager( ).numParams( ) * our_ln2pi + d[0] );
             //
             // no number < -DBL_MAX can be represented as double; log(BF) value
             // will always be <= 0 with this choice (except for the null model).

  return true;
 }

/* =============================================================================
// ------------------------------------------------------------ log bayes factor
// -----------------------------------------------------------------------------

uintVector
LogBayesFactor::leadSizes( ) const
 {
  return uintVector( );
 }

bool
LogBayesFactor::compute( quantityType& quantity, bool tcIsConst ) const
 {
  quantity.resize( 1, my_default );

  if ( my_logintNull == my_logint ) { quantity[0] = 0; return true; }

  const LogIntegral::quantityType& liNull = my_logintNull->operator()( tcIsConst );
  if ( my_logintNull->anyErrors( ) ) return false;

  const LogIntegral::quantityType& li = my_logint->operator()( tcIsConst );
  if ( my_logint->anyErrors( ) ) return false;

  quantity[0] = min( our_lnDBLmax, li[0] - liNull[0] );

  return true;
 }
   ============================================================================= */

#endif /* __products_H__ */
