/* ----------------------------------------------------------------------------
 *
 * header settings.h
 *   implementation
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#include <brain.vmp/settings>

// ----------------------------------------------------------------- Acquisition
// -----------------------------------------------------------------------------

Acquisition::Acquisition( )
 : my_slices( 1 ), my_images( 1 ),
   my_shots( 1 ), my_tr( 1 ), my_iai( 1 ),
   my_Ti( 0 ), my_Tf( 1 ),
   my_sliceOrder( 1 ), my_sliceOffset( 1 )
 {
 }

Acquisition::Acquisition(       size_t      numSlices,
                                size_t      numImages,
                                fpType      tr,
                                fpType      iai,
                                size_t      numShots,
                          const uintVector& sliceOrder )
 {
  reinitialize( numSlices, numImages, tr, iai, numShots, sliceOrder );
 }

void
Acquisition::reinitialize(       size_t      slices,
                                 size_t      images,
                                 fpType      repeatTime,
                                 fpType      iaInterval,
                                 size_t      shots,
                           const uintVector& order )
 {
  /* numbers of slices, images and shots */

  my_slices = ( slices > 0 ) ? slices : ( ( order.size( ) > 0 ) ? order.size( ) : 1 );
  my_images = ( images > 0 ) ? images : 1;
  my_shots  = ( shots  > 0 ) ? shots  : 1;

  /* form a consistent set of value for iai, tr and my_shots */

  if ( ( repeatTime > 0 ) && ( iaInterval > 0 ) )
   {
    if ( my_shots * repeatTime >= iaInterval )
     {
      my_tr  = repeatTime;
      my_iai = my_shots * my_tr;
                                       //
                                       // iai inferred from my_shots, repeatTime
     }
    else
     {
      my_tr  = repeatTime;
      my_iai = iaInterval;
                           //
                           // treat iai and repeatTime as independent parameters
                           //     not sure if this is ever realized in real life
     }
   }
  else if ( ( repeatTime <= 0 ) && ( iaInterval <= 0 ) )
   {
    my_iai = 1;
    my_tr  = my_iai / my_shots;
   }
  else if ( repeatTime <= 0 )
   {
    my_iai = iaInterval;
    my_tr  = my_iai / my_shots;
   }
  else /* if ( iaInterval <= 0 ) */
   {
    my_tr  = repeatTime;
    my_iai = my_shots * my_tr;
   }

  /* initial and final acquisition times; assume initial == 0 */

  my_Tf  = my_Ti = 0;
  my_Tf += my_images * my_iai;

  /* slice order and acquisition offsets */

  bool allRepresented = false, noneRepeated = false;

  if ( order.size( ) >= my_slices )
   {
    allRepresented = true;
    noneRepeated   = true;

    bool found, repeated;

    for ( size_t i = 0; i < my_slices; i++ )
     {
      found    = false;
      repeated = false;

      for ( size_t i2 = 0; i2 < my_slices; i2++ )
       {
        repeated = found && ( order[i2] == i );
        found    = found || ( order[i2] == i );
       }

      allRepresented = allRepresented && found;
      noneRepeated   = noneRepeated   && !repeated;
     }
   }

  my_sliceOrder.resize( my_slices );
  my_sliceOffset.resize( my_slices );

  fpType dt = my_tr / my_slices;

  if ( allRepresented && noneRepeated )
   for ( size_t i = 0; i < my_slices; i++ )
    {
     my_sliceOrder[i]                 = order[i];
     my_sliceOffset[my_sliceOrder[i]] = i * dt;
    }
  else
   for ( size_t i = 0; i < my_slices; i++ )
    {
     my_sliceOrder[i]  = i;
     my_sliceOffset[i] = i * dt;
    }
 }

// ------------------------------------------------------------------- TimeGrids
// -----------------------------------------------------------------------------

TimeGrids::TimeGrids( ) { }
TimeGrids::TimeGrids( const Acquisition& ac, const MissingInfo& mi ) { reinitialize( ac, mi ); }

void
TimeGrids::reinitialize( const Acquisition& ac, const MissingInfo& mi )
 {
  assert( ac.numImages( ) == mi.numInstants( ) );
  assert( ac.numSlices( ) == mi.numSlices( ) );

  my_Ti = ac.initialTime( );
  my_Tf = ac.finalTime( );

  my_baseGrid.resize( ac.numImages( ) );
  for ( size_t i = 0; i < ac.numImages( ); i++ )
   my_baseGrid[i] = my_Ti + i * ac.iai( );

  vector<fpVector>::resize( ac.numSlices( ) );
  for ( size_t slice = 0; slice < ac.numSlices( ); slice++ )
   {
    fpVector fullGrid( my_baseGrid );
    fullGrid += ac.sliceOffset( )[slice];
    mi.compress( slice, fullGrid, vector<fpVector>::operator[]( slice ) );
   }
 }

// ---------------------------------------------------------------- BellRegistry
// -----------------------------------------------------------------------------

Bell<>     BellRegistry::un2;
Bell_ST3<> BellRegistry::st3;
Bell_ST4<> BellRegistry::st4;
Bell_V4<>  BellRegistry::v4;

BasicBell*
BellRegistry::name2bell( const string& name, const string& defaultName )
 {
  if ( name        == v4.name( )  ) return &v4;
  if ( name        == st3.name( ) ) return &st3;
  if ( name        == st4.name( ) ) return &st4;
  if ( name        == un2.name( ) ) return &un2;
  if ( defaultName == v4.name( )  ) return &v4;
  if ( defaultName == st3.name( ) ) return &st3;
  if ( defaultName == st4.name( ) ) return &st4;
  if ( defaultName == un2.name( ) ) return &un2;

  return &v4;
 }

// ---------------------------------------------------------------- TCisConstant
// -----------------------------------------------------------------------------

TCisConstant::TCisConstant( fpType threshold, size_t num2check )
 : my_threshold( threshold ), my_num2check( num2check )
 {
 }

// -------------------------------------------------------------------- Settings
// -----------------------------------------------------------------------------

Settings::Settings( const string& imageChunk, const string& missingChunk, char t, char z,
                    const string& submodelSpecs, const string& dataDescription,
                    const string& runDescription )
 : imageChunk( imageChunk ), missingChunk( missingChunk ), t( t ), z( z ),
   submodelSpecs( submodelSpecs ), dataDescription( dataDescription ),
   runDescription( runDescription )
 {
 }

// ------------------------------------------------------------- Settings::Drift
// -----------------------------------------------------------------------------

Settings::Drift::Drift( size_t basisOrder, fpType relWeight,
                        size_t smoothFn, size_t smoothMaxIt, fpType smoothTol )
 : basisOrder( basisOrder ),
   relWeight( relWeight ),
   smoothFn( smoothFn ),
   smoothMaxIt( smoothMaxIt ),
   smoothTol( smoothTol )
 {
 }

// ------------------------------------------------------------ Settings::Params
// -----------------------------------------------------------------------------

Settings::Params::Types::Types( const string& drift, const string& noise, const string& ar1,
                                const string& response, const string& shape )
 : drift( drift ), noise( noise ), ar1( ar1 ), response( response ), shape( shape )
 {
 }

Settings::Params::Bounds::Bounds( const BoundType& drift,    const BoundType& noise, const BoundType& ar1,
                                  const BoundType& response, const BoundType& shape )
 : drift( drift ), noise( noise ), ar1( ar1 ), response( response ), shape( shape )
 {
 }

Settings::Params::Scales::Scales( fpType drift, fpType noise, fpType ar1, fpType response, fpType shape )
 : drift( drift ), noise( noise ), ar1( ar1 ), response( response ), shape( shape )
 {
 }

// ------------------------------------------------------------- Settings::ARMA
// -----------------------------------------------------------------------------

Settings::ARMA::ARMA( size_t numAR, size_t numMA ) : numAR( numAR ), numMA( numMA ) { }

// --------------------------------------------------------- Settings::Minimizer
// -----------------------------------------------------------------------------

Settings::Minimizer::Minimizer( size_t maxCorrections, size_t maxIterations,
                                size_t maxEvaluations, fpType maxPrGradNorm,
                                fpType minRelReduceF,  size_t maxRestarts,
                                bool   checkHdet,      bool   stopAtPlusDefH,
                                fpType mrrfReduction,  int    verbosity )
 : maxCorrections( maxCorrections ), maxIterations( maxIterations ),
   maxEvaluations( maxEvaluations ), maxPrGradNorm( maxPrGradNorm ),
   minRelReduceF( minRelReduceF ),   maxRestarts( maxRestarts ),
   checkHdet( checkHdet ),           stopAtPlusDefH( stopAtPlusDefH ),
   mrrfReduction( mrrfReduction ),   verbosity( verbosity )
 {
 }

// ------------------------------------------------------- Settings::Initializer
// -----------------------------------------------------------------------------

Settings::Initializer::Initializer( size_t scheme, size_t maxIt, const Minimizer& minimizer )
 : scheme( scheme ), maxIt( maxIt ), minimizer( minimizer )
 {
 }

// -------------------------------------------------------- Settings::ObjFnNames
// -----------------------------------------------------------------------------

Settings::ObjFnNames::ObjFnNames( const string& noise, const string& response,
                                  const string& shape, const string& ll_prefix, const string& posterior )
 : noise( noise ), response( response ), shape( shape ), ll_prefix( ll_prefix ), posterior( posterior )
 {
 }

// ------------------------------------------------------------ Settings::Output
// -----------------------------------------------------------------------------

Settings::Output::Output( const string&                    name,
                                PghMRIhandler::mriOpenMode openMode,
                                bool                       stdout,
                                bool                       pghmri )
 : fileName( name ), openMode( openMode ), stdout( stdout ), pghmri( pghmri )
 {
 }
