/* ----------------------------------------------------------------------------
 *
 * header settings.h
 *   miscellaneous stuff
 *
 * required headers
 *   see include header file settings
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __settings_H__
#define __settings_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

struct Acquisition
 {
  Acquisition( );

  Acquisition(       size_t      numSlices,
                     size_t      numImages,
                     fpType      tr         = 1,
                     fpType      iai        = 1,
                     size_t      numShots   = 1,
               const uintVector& sliceOrder = uintVector( ) );

  void
  reinitialize(       size_t      numSlices,
                      size_t      numImages,
                      fpType      tr         = 1,
                      fpType      iai        = 1,
                      size_t      numShots   = 1,
                const uintVector& sliceOrder = uintVector( ) );

  inline
  size_t
  numSlices( ) const;

  inline
  size_t
  numImages( ) const;

  inline
  size_t
  numShots( ) const;

  inline
  fpType
  tr( ) const;

  inline
  fpType
  iai( ) const;

  inline
  fpType
  initialTime( ) const;

  inline
  fpType
  finalTime( ) const;

  inline
  const uintVector&
  sliceOrder( ) const;

  inline
  const fpVector&
  sliceOffset( ) const;

  private:

   size_t     my_slices, my_images, my_shots;
   fpType     my_tr, my_iai, my_Ti, my_Tf;         // assume initial time == 0
   uintVector my_sliceOrder;                     // slice acquisition sequence
   fpVector   my_sliceOffset;            // slicewise acquisition time offsets
 };

struct TimeGrids : public vector<fpVector>
 {
  TimeGrids( );
  TimeGrids( const Acquisition& ac, const MissingInfo& mi );

  void
  reinitialize( const Acquisition& ac, const MissingInfo& mi );

  inline
  const fpVector&
  baseGrid( ) const;

  inline
  fpType
  initialTime( ) const;

  inline
  fpType
  finalTime( ) const;

  private:

   fpVector my_baseGrid;
   fpType   my_Ti, my_Tf;
 };

struct BellRegistry
 {
  static Bell<>     un2;
  static Bell_ST3<> st3;
  static Bell_ST4<> st4;
  static Bell_V4<>  v4;

  static
  BasicBell*
  name2bell( const string& name, const string& defaultName = "v4" );
 };

struct TCisConstant                                 // constant timecourse check
 {
  TCisConstant( fpType threshold = 1.0e-10, size_t num2check = 10 );

  inline
  bool
  operator()( const fpVector& tc ) const;

  inline
  fpType
  threshold( ) const;

  inline
  size_t
  num2check( ) const;

  private:

   fpType my_threshold;
   size_t my_num2check;
 };

struct Settings
 {
  Settings( const string& imageChunk       = "images",
            const string& missingChunk     = "missing",
                  char    t                = 't',
                  char    z                = 'z',
            const string& submodelSpecs    = "+ name all end",
            const string& dataDescription  = "(none available)",
            const string& runDescription   = "(none available)" );

  string inputSpecsFile, imageData, missingSpecs;

  string imageChunk, missingChunk;
  char   t, z;

  string submodelSpecs;

  string dataDescription, runDescription;

  Acquisition  acquisition;
  TimeGrids    timegrids;
  TCisConstant tcIsConstant;

  struct Drift
   {
    Drift( size_t basisOrder  = 4,
           fpType relWeight   = 0.0,
           size_t smoothFn    = 0,
           size_t smoothMaxIt = 50,
           fpType smoothTol   = 1.0e+2 * DBL_EPSILON );

    int    basisOrder;
    fpType relWeight;
    int    smoothFn;
    int    smoothMaxIt;
    fpType smoothTol;

    fpType df, smoothClose, smoothPar;
    int    smoothIt, smoothInfo;
   }
  drift;

  struct Params
   {
               // ATTN need to be able to set types/bounds/scales for individual
               // ATTN shape parameters  for  more complicated  bell  functions.
               //
    struct Types
     {
      Types( const string& drift    = "none",
             const string& noise    = "lower",
             const string& ar1      = "both",
             const string& response = "lower",
             const string& shape    = "lower" );

      string drift, noise, ar1, response, shape;
     }
    types;

    struct Bounds
     {
      typedef pair<fpType,fpType> BoundType;

      Bounds( const BoundType& drift    = BoundType( 0,       0 ),
              const BoundType& noise    = BoundType( 1.0e-12, 0 ),
              const BoundType& ar1      = BoundType( -1,      1 ),
              const BoundType& response = BoundType( 0,       0 ),
              const BoundType& shape    = BoundType( 1.0e-8,  0 ) );

      BoundType drift, noise, ar1, response, shape;
     }
    bounds;

    struct Scales
     {
      Scales( fpType drift = 1, fpType noise = 1, fpType ar1 = 1, fpType response = 1, fpType shape = 1 );

      fpType drift, noise, ar1, response, shape;
     }
    scales;
   }
  params;

  struct ARMA
   {
    ARMA( size_t numAR = 0, size_t numMA = 0 );
    size_t numAR, numMA;
   }
  arma;

  struct Minimizer
   {
    Minimizer( size_t maxCorrections = 100,  size_t maxIterations  = 1000,
               size_t maxEvaluations = 2000, fpType maxPrGradNorm  = 1.0e-10,
               fpType minRelReduceF  = 0.0,  size_t maxRestarts    = 20,
               bool   checkHdet      = true, bool   stopAtPlusDefH = true,
               fpType mrrfReduction  = 0.8,  int    verbosity      = -1 );

    size_t maxCorrections, maxIterations, maxEvaluations;
    fpType maxPrGradNorm, minRelReduceF;
    size_t maxRestarts;
    bool   checkHdet, stopAtPlusDefH;
    fpType mrrfReduction;
    int    verbosity;
   }
  minimizer;

  struct Initializer
   {
    Initializer(       size_t     scheme    = 0,
                       size_t     maxIt     = 20,
                 const Minimizer& minimizer = Minimizer( 10, 25, 50, 1.0e-7, 0.0, 5, false, false, 0.8, -1 ) );

    size_t scheme, maxIt;
    Minimizer minimizer;
   }
  initializer;

  struct ObjFnNames
   {
    ObjFnNames( const string& noise      = "LP Noise",
                const string& response   = "LP Response: ",
                const string& shape      = "LP Shape: ",
                const string& ll_prefix  = "LL ",
                const string& posterior  = "- Log Posterior" );

    string noise, response, shape, ll_prefix, posterior;  // use default names for the rest
   }
  objFnNames;

  struct Output
   {
    Output( const string&                    name     = "vmp.master.output.mri",
                  PghMRIhandler::mriOpenMode openMode = PghMRIhandler::mriWrite,
                  bool                       stdout   = false,
                  bool                       pghmri   = true );

    string                     fileName;
    PghMRIhandler::mriOpenMode openMode;
    bool                       stdout, pghmri;
    bool                       bfRequested, mpRequested, liRequested, mpLittleE, bfLittleE;
    string                     mpType, bfType;

    vector<string>                      products;
    vector<PghMRIhandler::mriArrayType> types;
    vector<bool>                        littleE;
    vector<string>                      leadDims;
    vector<bool>                        diagnostics;

    fpVector                            priorProbabilities;
   }
  output;
 };

// ----------------------------------------------------------------- Acquisition
// -----------------------------------------------------------------------------

size_t
Acquisition::numSlices( ) const
 {
  return my_slices;
 }

size_t
Acquisition::numImages( ) const
 {
  return my_images;
 }

size_t
Acquisition::numShots( ) const
 {
  return my_shots;
 }

fpType
Acquisition::tr( ) const
 {
  return my_tr;
 }

fpType
Acquisition::iai( ) const
 {
  return my_iai;
 }

fpType
Acquisition::initialTime( ) const
 {
  return my_Ti;
 }

fpType
Acquisition::finalTime( ) const
 {
  return my_Tf;
 }

const uintVector&
Acquisition::sliceOrder( ) const
 {
  return my_sliceOrder;
 }

const fpVector&
Acquisition::sliceOffset( ) const
 {
  return my_sliceOffset;
 }

// ------------------------------------------------------------------- TimeGrids
// -----------------------------------------------------------------------------

const fpVector&
TimeGrids::baseGrid( ) const
 {
  return my_baseGrid;
 }

fpType
TimeGrids::initialTime( ) const
 {
  return my_Ti;
 }

fpType
TimeGrids::finalTime( ) const
 {
  return my_Tf;
 }

// ---------------------------------------------------------------- TCisConstant
// -----------------------------------------------------------------------------

bool
TCisConstant::operator()( const fpVector& tc ) const
 {
  if ( tc.size( ) == 0 ) return true;

  bool isConst = true;
  for ( size_t i = 0, j = tc.size( ) - 1; i < min( my_num2check, tc.size( ) ); i++, j-- )
   isConst = isConst && ( fabs( tc[0] - tc[i] ) <= my_threshold )
                     && ( fabs( tc[0] - tc[j] ) <= my_threshold );

  return isConst;
 }

fpType
TCisConstant::threshold( ) const
 {
  return my_threshold;
 }

size_t
TCisConstant::num2check( ) const
 {
  return my_num2check;
 }

#endif /* __settings_H__ */
