#include <cfloat>
#include <cstdio>
#include <cstdlib>

#include <algorithm>
#include <fstream>
#include <ios>
#include <iostream>
#include <iomanip>
#include <map>
#include <string>
#include <vector>

#include <cppconfig>

#include <basic/BasicTypes>
#include <basic/ParameterManager>

#include <b-splines/BSplineBasis>

#include <objective.functions/LPgamma>
#include <objective.functions/ObjFnManager>

#include <minimizer/MinQNB>

#include <pghmri/PghMRIhandler>
#include <missing.info/MissingInfo>
#include <timecourse.handlers/TCrwPghMRI>
#include <timecourse.handlers/DataNavigator>

#include <brain.model/acProfile>
#include <brain.model/dbProfile>
#include <brain.model/Bell>
#include <brain.model/ParamBlkNames>
#include <brain.model/SubmodelTable>
#include <brain.model/LLnormal>
#include <brain.model/LPbaseline>
#include <brain.model/LPdrift>

#include <brain.vmp/products>
#include <brain.vmp/settings>
#include <brain.vmp/initialize>
#include <brain.vmp/InSpec>
#include <brain.vmp/model.probability>

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

void
initNewSlice( Settings&  settings,
              size_t     newSlice,
              dbProfile& drift,
              LPdrift&   lpDrift,
              acProfile& activation )
 {
  drift.setTimeGrid( settings.timegrids.initialTime( ),
                     settings.timegrids.finalTime( ),
                     &( settings.timegrids[newSlice] ) );
  drift.initialize( );

  settings.drift.smoothPar
   = drift.smoothingPar( settings.drift.relWeight,   settings.drift.df,
                         settings.drift.smoothInfo,  settings.drift.smoothIt,
                         settings.drift.smoothClose, settings.drift.smoothFn,
                         settings.drift.smoothTol,   settings.drift.smoothMaxIt );

  assert( settings.drift.smoothInfo == 0 );

  lpDrift.setConstants( settings.drift.smoothPar, settings.drift.relWeight );

  activation.setTimeGrid( &( settings.timegrids[newSlice] ) );
 }

void
initObjFn( const Settings&                 settings,
           const ParameterManager<fpType>& parManager,
           const acProfile&                activation,
                 ObjFnManager&             objectiveFn,
                 LPbaseline&               lpBaseline,
                 LPdrift&                  lpDrift,
                 LPgamma&                  lpNoise,
                 vector<LPgamma>&          lpResponse,
                 map<string,LPgamma>&      lpShape,
                 LLnormal<false,0,0>&      llNormalNA,
                 LLnormal<true,0,0>&       llNormal,
                 LLnormal<false,1,0>&      llNormalNA_ar1,
                 LLnormal<true,1,0>&       llNormal_ar1 )
 {
  objectiveFn.setParamBufSize( parManager.numParams( ) );
  objectiveFn.delAll( );

  lpBaseline.setParamBufSize( parManager.numParams( ) );
  // lpBaseline.setParamOffsets( parManager );

  lpDrift.setParamBufSize( parManager.numParams( ) );
  // lpDrift.setParamOffsets( parManager );

  lpNoise.setParamBufSize( parManager.numParams( ) );
  // lpNoise.setParamOffset( parManager.blockOffset( ParamBlkNames::noise ) );

  objectiveFn.add( &( lpBaseline ) );
  objectiveFn.add( &( lpDrift    ) );
  objectiveFn.add( &( lpNoise    ) );

  for ( size_t a = 0; a < activation.numActive( ); a++ )
   {
    size_t c = activation.condition( a );

    lpResponse[c].setParamBufSize( parManager.numParams( ) );
    lpResponse[c].setParamOffset( parManager.blockOffset( ParamBlkNames::response ) + c );

    objectiveFn.add( &( lpResponse[c] ) );
   }

  for ( size_t a = 0; a < activation.numActive( ); a++ )
   {
    const acProfile::ConditionBellSet& cbs = activation.bellSet( activation.condition( a ) );

    lpShape[cbs.tag( )].setParamBufSize( parManager.numParams( ) );
    lpShape[cbs.tag( )].setParamOffset(   parManager.blockOffset( ParamBlkNames::shape )
                                        + cbs.shapeOffset( ) );

    objectiveFn.add( &( lpShape[cbs.tag( )] ) );
   }

  assert( ( settings.arma.numAR == 0 ) || ( settings.arma.numAR == 1 ) );

  if ( activation.numActive( ) == 0 )
   {
    if ( settings.arma.numAR == 0 )
     {
      llNormalNA.setParamBufSize( parManager.numParams( ) );
      llNormalNA.setParamOffset( parManager );

      objectiveFn.add( &( llNormalNA ) );
     }
    else if ( settings.arma.numAR == 1 )
     {
      llNormalNA_ar1.setParamBufSize( parManager.numParams( ) );
      llNormalNA_ar1.setParamOffset( parManager );

      objectiveFn.add( &( llNormalNA_ar1 ) );
     }
   }
  else
   {
    if ( settings.arma.numAR == 0 )
     {
      llNormal.setParamBufSize( parManager.numParams( ) );
      llNormal.setParamOffset( parManager );

      objectiveFn.add( &( llNormal ) );
     }
    else if ( settings.arma.numAR == 1 )
     {
      llNormal_ar1.setParamBufSize( parManager.numParams( ) );
      llNormal_ar1.setParamOffset( parManager );

      objectiveFn.add( &( llNormal_ar1 ) );
     }
   }
 }

void
initMinimizer( const Settings&                 settings,
               const ParameterManager<fpType>& parManager,
               const ObjFnManager&             objectiveFn,
                     MinQNB&                   minimizer )
 {
  minimizer.setSizes( parManager.numParams( ), settings.minimizer.maxCorrections );

  minimizer.setStoppingCriteria( settings.minimizer.minRelReduceF,
                                 settings.minimizer.maxPrGradNorm,
                                 settings.minimizer.maxEvaluations,
                                 settings.minimizer.maxIterations );

  minimizer.setRestartOpts( settings.minimizer.maxRestarts,
                            settings.minimizer.checkHdet,
                            settings.minimizer.stopAtPlusDefH,
                            settings.minimizer.mrrfReduction );

  minimizer.setVerbosity( settings.minimizer.verbosity );

  int len_d = parManager.blockLength( ParamBlkNames::drift    );
  int len_n = parManager.blockLength( ParamBlkNames::noise    );
  int len_a = parManager.blockLength( ParamBlkNames::ar       );
  int len_r = parManager.blockLength( ParamBlkNames::response );
  int len_s = parManager.blockLength( ParamBlkNames::shape    );

  int ofs_d = parManager.blockOffset( ParamBlkNames::drift    );
  int ofs_n = parManager.blockOffset( ParamBlkNames::noise    );

  minimizer.setBounds( ofs_d, len_d, settings.params.types.drift,
                       settings.params.bounds.drift.first,
                       settings.params.bounds.drift.second );
  minimizer.setScale( ofs_d, len_d, settings.params.scales.drift );

  minimizer.setBounds( ofs_n, len_n, settings.params.types.noise,
                       settings.params.bounds.noise.first,
                       settings.params.bounds.noise.second );
  minimizer.setScale( ofs_n, len_n, settings.params.scales.noise );

  if ( len_a > 0 )
   {
    int ofs_a = parManager.blockOffset( ParamBlkNames::ar );

    minimizer.setBounds( ofs_a, len_a, settings.params.types.ar1,
                         settings.params.bounds.ar1.first,
                         settings.params.bounds.ar1.second );
    minimizer.setScale( ofs_a, len_a, settings.params.scales.ar1 );
   }

  if ( len_r > 0 )
   {
    int ofs_r = parManager.blockOffset( ParamBlkNames::response );
    int ofs_s = parManager.blockOffset( ParamBlkNames::shape    );

    minimizer.setBounds( ofs_r, len_r, settings.params.types.response,
                         settings.params.bounds.response.first,
                         settings.params.bounds.response.second );
    minimizer.setScale( ofs_r, len_r, settings.params.scales.response );

    minimizer.setBounds( ofs_s, len_s, settings.params.types.shape,
                         settings.params.bounds.shape.first,
                         settings.params.bounds.shape.second );
    minimizer.setScale( ofs_s, len_s, settings.params.scales.shape );
   }

  minimizer.setObjectiveFn( &objectiveFn );
 }

void
initTCconstParams( const                 Settings&                 settings,
                   const                 ParameterManager<fpType>& parManager,
                   const                 acProfile&                activation,
                   const                 fpVector&                 timeCourse,
                /* in_principle_const */ map<string,LPgamma>&      lpShape,
                                         fpVector&                 parameters,
                                         fpType                    noisePrecisionDefault = 1.0e+16,
                                         bool                      priorMeanForShape     = true )
 {
  parameters.resize( parManager.numParams( ), timeCourse[0] );

  parameters[parManager.blockOffset( ParamBlkNames::noise )] = noisePrecisionDefault;
              //
              // arbitrary value of 1.0e+16 for noise precision is approximately
              // DBL_EPSILON^-1;  Here, noise precision is technically infinite.

  for ( int i = 0, j = parManager.blockOffset( ParamBlkNames::response );
            i < parManager.blockLength( ParamBlkNames::response ); i++, j++ )
   parameters[j] = 0.0;

  for ( size_t a = 0; a < activation.numActive( ); a++ )
   {
    const string& tag = activation.bellSet( activation.condition( a ) ).tag( );
    size_t n          = activation.bellSet( activation.condition( a ) ).numShapeParams( );

    for ( size_t i = 0,
                 j =   parManager.blockOffset( ParamBlkNames::shape )
                     + activation.bellSet( activation.condition( a ) ).shapeOffset( );
                 i < n; i++, j++ )
     parameters[j] =   priorMeanForShape
                     ? ( lpShape[tag].constA( i ) / lpShape[tag].constB( i ) )            // mean
                     : ( ( lpShape[tag].constA( i ) - 1.0 ) / lpShape[tag].constB( i ) ); // mode
   }
 }

void
initFirstVoxel( const Settings&                  settings,
                const string&                    submodelKey,
                const string&                    submodelDescription,
                      size_t                     smIndex,
                const DataNavigator<TCrwPghMRI>& dataNavigator,
                const ParameterManager<fpType>&  parManager,
                const MinQNB&                    minimizer,
                const acProfile&                 activation,
                      fpVector&                  parameters,
                      PghMRIhandler&             output,
                      vector<ProductSet>&        smProducts )
 {
  for ( size_t i = 0; i < settings.output.products.size( ); i++ )
   assert( smProducts[smIndex].add( settings.output.products[i] ) );

  size_t n; OutputProduct* p;
  string prefix = submodelKey + ".";

  /* ----------------------------------------------------------
  for ( size_t i = 0; i < smProducts[smIndex]( ).size( ); i++ )
   {
    p = smProducts[smIndex]( )[i];
    cerr << "(debug) submodel " << submodelKey << " product " << p->name( )
         << " configured " << p->isConfigured( )
         << " hasProductsNeeded " << p->hasProductsNeeded( )
         << " initialized " << p->isInitialized( )
         << " pghmri " << p->pghmri( )
         << " stdout " << p->stdout( ) << " p " << p << endl;
   }
  cerr << endl;
     ---------------------------------------------------------- */

  vector<string>::const_iterator i1;

  for ( size_t i = 0; i < smProducts[smIndex]( ).size( ); i++ )
   {
    p = smProducts[smIndex]( )[i];

    p->configure( &dataNavigator, &parManager, &minimizer,
                  &activation, &parameters, &settings.timegrids );

    i1 = find( settings.output.products.begin( ), settings.output.products.end( ), p->name( ) );

    /* ----------------------------------------------------------
       cerr << "(debug) submodel " << submodelKey << " product " << p->name( )
            << " configured " << p->isConfigured( )
            << " hasProductsNeeded " << p->hasProductsNeeded( )
            << " n " << n << endl;
       ---------------------------------------------------------- */

    if (    ( i1 == settings.output.products.end( ) )
         || ( ( p->name( ) == "log.integral" ) && !settings.output.liRequested ) )
     {
      p->setOutput( false, false );
      p->initialize( &output, prefix );
     }
    else
     {
      n = i1 - settings.output.products.begin( );
      p->setOutput( settings.output.stdout, settings.output.pghmri );
      p->initialize( &output, prefix, settings.output.types[n],
                     settings.output.leadDims[n], settings.output.diagnostics[n], settings.output.littleE[n] );
     }

    /*
    cerr << "(debug) submodel " << submodelKey << " product " << p->name( )
         << " configured " << p->isConfigured( )
         << " hasProductsNeeded " << p->hasProductsNeeded( )
         << " initialized " << p->isInitialized( )
         << " pghmri " << p->pghmri( )
         << " stdout " << p->stdout( )
         << " diagnostics " << p->diagnostics( )
         << " n " << n << endl;
    */
   }

  string pd = ParamBlkNames::paramDescr( parManager );
  string rd = ParamBlkNames::responseDescr( parManager, activation );
  string sd = ParamBlkNames::shapeDescr( parManager, activation );

  if ( pd.size( ) > 0 ) output.setString( submodelKey + ".paramblock.description", pd );
  if ( rd.size( ) > 0 ) output.setString( submodelKey + ".paramblock.description.response", rd );
  if ( sd.size( ) > 0 ) output.setString( submodelKey + ".paramblock.description.shape", sd );

  output.setString( submodelKey + ".submodel.description", submodelDescription );
 }

int
main ( int argc, char* argv[] )
 {
  if ( argc < 2 )
   {
    cerr << "usage: "
         << argv[0]
         << " input [output=" << Settings( ).output.fileName
         << "] [output_mode=" << PghMRIhandler::mode2name( Settings( ).output.openMode )
         << "]" << endl;
    exit( 1 );
   }

  // ------------------------------------------------ most of the objects needed
  // ---------------------------------------------------------------------------

  Settings                  settings;

  TCrwPghMRI                tcReader;
  MissingInfo               missingInfo;
  DataNavigator<TCrwPghMRI> dataNavigator;

  BSplineBasis              driftBasis;
  dbProfile                 drift;

  SubmodelTable             submodelTable;
  acProfile                 activation;

  ParameterManager<fpType>  parManager;

  LLnormal<false,0,0>       llNormalNA;
  LLnormal<true,0,0>        llNormal;
  LLnormal<false,1,0>       llNormalNA_ar1;
  LLnormal<true,1,0>        llNormal_ar1;

  LPbaseline                lpBaseline;
  LPdrift                   lpDrift;
  LPgamma                   lpNoise;
  vector<LPgamma>           lpResponse;
  map<string,LPgamma>       lpShape;          // condition tag <--> shape priors

  ObjFnManager              objectiveFn;

  fpVector                  parameters;

  MinQNB                    minimizer;
  vector<ProductSet>        smProducts;
  ModelProbability          modelProbability;


  InSpec                    inspec;
  PghMRIhandler             input, output;

  // ------------------------------ input processing and partial initializations
  // ---------------------------------------------------------------------------

  cout << "initializing" << endl;

  settings.inputSpecsFile = string( argv[1] );
  input.open( settings.inputSpecsFile, PghMRIhandler::mriRead );

  inspec.getSpecs_output( input, settings );

  if ( argc > 2 ) settings.output.fileName = string( argv[2] );
  if ( argc > 3 ) settings.output.openMode = PghMRIhandler::name2mode( string( argv[3] ),
                                                                       PghMRIhandler::mriWrite );
  
  assert( settings.output.openMode != PghMRIhandler::mriRead );
  assert( settings.output.openMode != PghMRIhandler::mriModifyData );

  output.open( settings.output.fileName, settings.output.openMode );

  if ( settings.output.openMode == PghMRIhandler::mriModify )
   {
         // this is a rather convolved way of overriding defaults  with settings
         // from output file in mriModify mode overriding with settings from the
         // input file;  anything better  will need  much redesigning of InSpec.

    TCrwPghMRI                tcr;
    MissingInfo               mi;
    DataNavigator<TCrwPghMRI> dn;
    BSplineBasis              db;
    dbProfile                 d;
    SubmodelTable             smt;
    acProfile                 a;
    ParameterManager<fpType>  pm;
    LPbaseline                lpB;
    LPdrift                   lpD;
    LPgamma                   lpN;
    vector<LPgamma>           lpR;
    map<string,LPgamma>       lpS;

    inspec.getSpecs_param( output, settings );
    inspec.getSpecs_minimizer( output, settings );
    inspec.getSpecs_initializer( output, settings );
    inspec.getSpecs_misc( output, settings );

    assert( ( settings.arma.numAR == 0 ) || ( settings.arma.numAR == 1 ) );

    inspec.initData( output, settings, tcr, mi, dn );
    inspec.initDrift( output, settings, db, d );
    inspec.initDesign( output, d, settings, a, smt,
                       settings.output.bfRequested || settings.output.mpRequested );

    inspec.getSpecs_products( output, settings, smt );

    pm.addBlock( ParamBlkNames::drift, db.basisSize( )     );
    pm.addBlock( ParamBlkNames::noise, 1                   );
    pm.addBlock( ParamBlkNames::ar,    settings.arma.numAR );

    inspec.initBDNpriors( output, d, a, pm, settings, lpB, lpD, lpN );
    inspec.initRSpriors( output, a, settings, lpR, lpS );

    inspec.writeSpecs_data( settings, dn, output );
    inspec.writeSpecs_drift( settings, db, output );
    inspec.writeSpecs_design( settings, a, smt, output );
    inspec.writeSpecs_BDNpriors( settings, lpB, lpD, lpN, output );
    inspec.writeSpecs_RSpriors( settings, a, lpR, lpS, output );
    inspec.writeSpecs_param( settings, output );
    inspec.writeSpecs_minimizer( settings, output );
    inspec.writeSpecs_initializer( settings, output );
    inspec.writeSpecs_misc( settings, output );
    inspec.writeSpecs_output( settings, output, smt );
   }

  inspec.getSpecs_param( input, settings );
  inspec.getSpecs_minimizer( input, settings );
  inspec.getSpecs_initializer( input, settings );
  inspec.getSpecs_misc( input, settings );

  assert( ( settings.arma.numAR == 0 ) || ( settings.arma.numAR == 1 ) );

  inspec.initData( input, settings, tcReader, missingInfo, dataNavigator );
  inspec.initDrift( input, settings, driftBasis, drift );
  inspec.initDesign( input, drift, settings, activation, submodelTable,
                     settings.output.bfRequested || settings.output.mpRequested );

  inspec.getSpecs_products( input, settings, submodelTable );

  parManager.addBlock( ParamBlkNames::drift, driftBasis.basisSize( ) );
  parManager.addBlock( ParamBlkNames::noise, 1                       );
  parManager.addBlock( ParamBlkNames::ar,    settings.arma.numAR     );
                        //
                        // rest of the blocks are submodel-dependent; add later.

  inspec.initBDNpriors( input, drift, activation, parManager, settings, lpBaseline, lpDrift, lpNoise );
  inspec.initRSpriors( input, activation, settings, lpResponse, lpShape );

  input.close( );

  objectiveFn.setSign( -1 );
  objectiveFn.setName( settings.objFnNames.posterior );

  if ( settings.arma.numAR == 0 )
   {
    llNormalNA.setDBprofile( &drift );
    llNormal.setDBprofile( &drift );
    llNormal.setACprofile( &activation );
   }
  else if ( settings.arma.numAR == 1 )
   {
    llNormalNA_ar1.setDBprofile( &drift );
    llNormal_ar1.setDBprofile( &drift );
    llNormal_ar1.setACprofile( &activation );
   }
                    //
                    // setParamBufSize( ), setParamOffset( ), setSignal( ) later

  // --------------------------------------------------------- initialize output
  // ---------------------------------------------------------------------------

  inspec.writeSpecs_data( settings, dataNavigator, output );
  inspec.writeSpecs_drift( settings, driftBasis, output );
  inspec.writeSpecs_design( settings, activation, submodelTable, output );
  inspec.writeSpecs_BDNpriors( settings, lpBaseline, lpDrift, lpNoise, output );
  inspec.writeSpecs_RSpriors( settings, activation, lpResponse, lpShape, output );
  inspec.writeSpecs_param( settings, output );
  inspec.writeSpecs_minimizer( settings, output );
  inspec.writeSpecs_initializer( settings, output );
  inspec.writeSpecs_misc( settings, output );
  inspec.writeSpecs_output( settings, output, submodelTable );

  // ---------------------------------- one-per-submodel output product handlers
  // ---------------------------------------------------------------------------

  smProducts.resize( submodelTable.size( ) );

  // ---------------------------- log bayes factor and model probability handler
  // ---------------------------------------------------------------------------

  modelProbability.setOutput( settings.output.pghmri && settings.output.mpRequested,
                              settings.output.stdout && settings.output.mpRequested,
                              settings.output.pghmri && settings.output.bfRequested,
                              settings.output.stdout && settings.output.bfRequested,
                              settings.output.mpRequested || settings.output.bfRequested );

  // ------------------------------------------------------------ voxelwise fits
  // ---------------------------------------------------------------------------

  cout << "computing voxelwise fits" << endl;

  size_t currentSlice = 0;
  size_t zdim = dataNavigator.sliceDim( );
  if ( zdim > dataNavigator.tdim( ) ) zdim -= 1;
                                        //
                                        // slice dimension index into v( ); v( )
                                        // is same as DataNavigator::locSansT( )

  size_t ldnt = dataNavigator.leadDimNT( );
  if ( ldnt > dataNavigator.tdim( ) ) ldnt -= 1;
                          //
                          // leading non-trivial non-t dimension index into v( )

  TCisConstant tcIsConstant;
  bool constTC;

  size_t nullIndex, smIndex;
  string submodelKey, voxelKey;

  bool firstVoxel = true, newSlice = true;

  for ( DataNavigator<TCrwPghMRI>::iterator v  = dataNavigator.begin( );
                                            v != dataNavigator.end( ); ++v )
   {
    voxelKey = dataNavigator.locationStr( ",", dataNavigator.dimsSansT( ) );

    // ---------------------------------------- constant time course? new slice?
    // -------------------------------------------------------------------------

    constTC  = tcIsConstant( (*v) );
    newSlice = firstVoxel || ( v( )[zdim] != currentSlice );

    if ( newSlice )
     {
      currentSlice = v( )[zdim];
      initNewSlice( settings, currentSlice, drift, lpDrift, activation );
     }

    // ------------------------------------------------- stdout status indicator
    // -------------------------------------------------------------------------

    if ( !settings.output.stdout )
     {
      if ( newSlice ) cout << endl << "slice " << currentSlice << endl;
      if ( v( )[ldnt] == dataNavigator.rangeLower( )[dataNavigator.leadDimNT( )] ) cout << endl;
      cout << ( constTC ? 'c' : '.' );
     }

    // -------------------------------------- likelihood: partial initialization
    // -------------------------------------------------------------------------

    if ( settings.arma.numAR == 0 )
     {
      llNormalNA.setSignal( &( *v ) );
      llNormal.setSignal( &( *v ) );
     }
    else if ( settings.arma.numAR == 1 )
     {
      llNormalNA_ar1.setSignal( &( *v ) );
      llNormal_ar1.setSignal( &( *v ) );
     }

    // ----------------------------------------------------- loop over submodels
    // -------------------------------------------------------------------------

    smIndex = 0;
    for ( SubmodelTable::const_iterator si  = submodelTable.begin( );
                                        si != submodelTable.end( ); ++si, smIndex++ )
     {
      submodelKey = submodelTable.key( si );

      // ------------------------------------------------- initialize activation
      // -----------------------------------------------------------------------

      activation.setSubmodel( &( *si ) ); activation.initialize( );

      // -------------------------------------------------- add parameter blocks
      // -----------------------------------------------------------------------

      parManager.delBlock( ParamBlkNames::response );
      parManager.delBlock( ParamBlkNames::shape );
      parManager.addBlock( ParamBlkNames::response, activation.numRespParams( ) );
      parManager.addBlock( ParamBlkNames::shape, activation.numShapeParams( ) );

      // ------------------------------------------ construct objective function
      // -----------------------------------------------------------------------

      initObjFn( settings, parManager, activation, objectiveFn,
                 lpBaseline, lpDrift, lpNoise, lpResponse, lpShape,
                 llNormalNA, llNormal, llNormalNA_ar1, llNormal_ar1 );

      // -------------------------------------------------- initialize minimizer
      // -----------------------------------------------------------------------

      initMinimizer( settings, parManager, objectiveFn, minimizer );

      // ------------------------------------------- minimize objective function
      // -----------------------------------------------------------------------

      if ( constTC )
       {
        initTCconstParams( settings, parManager, activation, (*v), lpShape, parameters );
        minimizer.reset( );
       }
      else
       {
        initialize( settings, (*v), parManager, activation, drift,
                    objectiveFn, llNormal, lpDrift, lpNoise,
                    lpResponse, lpShape, minimizer, parameters );
        minimizer.minimize( parameters( 0 ) );
       }

      // ------------------------------------------------------- output products
      // -----------------------------------------------------------------------

      if ( firstVoxel )
       {
        initFirstVoxel( settings, submodelKey,
                        submodelTable.description( si, '/' ), smIndex,
                        dataNavigator, parManager, minimizer, activation,
                        parameters, output, smProducts );
	if ( (*si).size( ) == 0 ) nullIndex = smIndex;
       }

      smProducts[smIndex].write( constTC, voxelKey + " " + submodelKey );
     }

    if ( firstVoxel )
     {
      modelProbability.initialize( &output, &dataNavigator, &submodelTable,
                                   smProducts, nullIndex, &settings.output.priorProbabilities );
      firstVoxel = false;
     }

    modelProbability.write( constTC, voxelKey );
   }

  for ( smIndex = 0; smIndex < smProducts.size( ); smIndex++ ) smProducts[smIndex].finalize( );
  output.setString( "baseline.coefficients", MiscFns::asString( drift.Dbaseline( 0 ), "%22.15e" ) );

  cout << endl << endl << "done" << endl;

  return 0;
 }
