#!/bin/tcsh -f

# ---------------------------------------------------------------------------
#
# csh script check.compliance.csh
#
# description
#   driver script  to assess the  ISO/ANSI standard compliance of a C++ compiler.
#   this is in no way a fool-proof, and the results need to be verified manually.
#
#   checks from the following four categories:
#
#   1. c++ languge features
#        needs user-supplied test files;
#        see csh script check.feature.csh
#   2. c++ standard headers
#        checks if <header> or <header.h> exist;
#        see csh script check.header.c++.csh
#   3. c++ standard headers for c standard library
#        checks if <cheader> or <header.h> exist;
#        see csh script check.header.c.csh
#   4. c++ standard classes
#        two possibilities for the headers X two possibilities for namespaces;
#        see csh script check.class.csh.
#        Not a fool-proof method; e. g., istream, ostream, streambuf, ios_base
#        do not support default constructors in the standard. This setup needs
#        to be modified to account for that (see cpp.check.ioclasses.cc).
# 
# usage
#   check.compliance.csh [-hFHC] [-l log_file] [-s stubs_directory] -c "compilation_command"
#
#   -h help/usage
#   -F disable features testing
#   -H disable headers testing
#   -C disable classes testing
#
#   -l log_file
#      compliance.log by default
#
#   -s stubs_directory
#      creates stub include files in the directory stubs_directory
#      for header files with nonstandard names
#
#   -c compilation command
#      use quotes whenever necessary; e. g., check.compliance.csh -c "g++ -g".
#      a.out must be the name of the executable produced
#
# notes
#   0. output in the form of cppconfig definitions
#   1. not a fool-proof method: all results need verification,
#      especially for:
#        .situations where a class is supported, but is declared/defined
#         under a different header; e. g., istream, ostream and streambuf
#         can be declared/defined under iostream.h (SGI MipsPRO 7.2.1)
#         or class ios may be declared under streambuf (gcc 2.95.2).
#        .c++ headers <limits>, <locale> and <string> which are
#         distinct from the corresponding c .h headers
#
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

# language features
#
 set features = (             keyword_explicit keyword_mutable )
 set features = ( ${features} keyword_namespace keyword_operators )
 set features = ( ${features} keyword_typename )
 set features = ( ${features} std_casts std_exceptions std_rtti )
 set features = ( ${features} template_defaults template_methods template_ttp )
 set features = ( ${features} template_partial )
 set features = ( ${features} type_bool type_wchar_t )

# standard c++ headers
#
 set cppheaders = ( algorithm bitset complex deque exception )
 set cppheaders = ( ${cppheaders} fstream functional )
 set cppheaders = ( ${cppheaders} iomanip ios iosfwd iostream istream iterator )
 set cppheaders = ( ${cppheaders} limits list locale map memory new numeric )
 set cppheaders = ( ${cppheaders} ostream queue )
 set cppheaders = ( ${cppheaders} set sstream stack stdexcept streambuf string )
 set cppheaders = ( ${cppheaders} typeinfo utility valarray vector )

# c++ headers for c standard library (without prefix "c")
#
 set cheaders = ( assert ctype errno float iso646 limits locale math )
 set cheaders = ( ${cheaders} setjmp signal stdarg stddef )
 set cheaders = ( ${cheaders} stdio stdlib string )
 set cheaders = ( ${cheaders} time wchar wctype )

# standard classes: containers
#
 set classes   = (              bitset )
 set classdecl = (              "bitset<32>" )
 set classhdrs = (              bitset )

 set classes   = ( ${classes}   deque )
 set classdecl = ( ${classdecl} "deque<int>" ) 
 set classhdrs = ( ${classhdrs} deque )

 set classes   = ( ${classes}   list )
 set classdecl = ( ${classdecl} "list<int>" ) 
 set classhdrs = ( ${classhdrs} list )

 set classes   = ( ${classes}   map )
 set classdecl = ( ${classdecl} "map<int,int>" ) 
 set classhdrs = ( ${classhdrs} map )

 set classes   = ( ${classes}   multimap )
 set classdecl = ( ${classdecl} "multimap<int,int>" ) 
 set classhdrs = ( ${classhdrs} map )

 set classes   = ( ${classes}   queue )
 set classdecl = ( ${classdecl} "queue<int>" ) 
 set classhdrs = ( ${classhdrs} queue )

 set classes   = ( ${classes}   priority_queue )
 set classdecl = ( ${classdecl} "priority_queue<int>" ) 
 set classhdrs = ( ${classhdrs} queue )

 set classes   = ( ${classes}   set )
 set classdecl = ( ${classdecl} "set<int>" ) 
 set classhdrs = ( ${classhdrs} set )

 set classes   = ( ${classes}   multiset )
 set classdecl = ( ${classdecl} "multiset<int>" ) 
 set classhdrs = ( ${classhdrs} set )

 set classes   = ( ${classes}   stack )
 set classdecl = ( ${classdecl} "stack<int>" ) 
 set classhdrs = ( ${classhdrs} stack )

 set classes   = ( ${classes}   vector )
 set classdecl = ( ${classdecl} "vector<float>" ) 
 set classhdrs = ( ${classhdrs} vector )

# standard classes: i/o
#
 set classes   = ( ${classes}   ios_base )
 set classdecl = ( ${classdecl} "ios_base" ) 
 set classhdrs = ( ${classhdrs} ios )

 set classes   = ( ${classes}   fstream )
 set classdecl = ( ${classdecl} "fstream" ) 
 set classhdrs = ( ${classhdrs} fstream )

 set classes   = ( ${classes}   istream )
 set classdecl = ( ${classdecl} "istream" ) 
 set classhdrs = ( ${classhdrs} istream )

 set classes   = ( ${classes}   ostream )
 set classdecl = ( ${classdecl} "ostream" ) 
 set classhdrs = ( ${classhdrs} ostream )

# standard classes: miscellaneous
#
 set classes   = ( ${classes}   complex )
 set classdecl = ( ${classdecl} "complex<float>" ) 
 set classhdrs = ( ${classhdrs} complex )

 set classes   = ( ${classes}   valarray )
 set classdecl = ( ${classdecl} "valarray<float>" ) 
 set classhdrs = ( ${classhdrs} valarray )

 set classes   = ( ${classes}   string )
 set classdecl = ( ${classdecl} "string" ) 
 set classhdrs = ( ${classhdrs} string )

 set classes   = ( ${classes}   stringstream )
 set classdecl = ( ${classdecl} "stringstream" ) 
 set classhdrs = ( ${classhdrs} sstream )

 set classes   = ( ${classes}   streambuf )
 set classdecl = ( ${classdecl} "streambuf" ) 
 set classhdrs = ( ${classhdrs} streambuf )

# command-line processing
#
 set help = 0
 if ( ${#argv} == 0 ) set help = 1

 set checkfeatures = 1
 set checkheaders  = 1
 set checkclasses  = 1

 set log           = compliance.log

 set createstubs   = 0
 set stubdir       = .

 set compile       = CC

 set n = 1
 while ( ${n} <= ${#argv} )
  set a = ${argv[${n}]}

  if ( `echo ${a} | awk '{print substr( $1, 1, 1 )}'` == "-" ) then
   if ( `echo ${a} | awk '{print index( $1, "h" )}'` != 0 ) set help = 1
   if ( `echo ${a} | awk '{print index( $1, "F" )}'` != 0 ) set checkfeatures = 0
   if ( `echo ${a} | awk '{print index( $1, "H" )}'` != 0 ) set checkheaders = 0
   if ( `echo ${a} | awk '{print index( $1, "C" )}'` != 0 ) set checkclasses = 0

   if ( `echo ${a} | awk '{print index( $1, "l" )}'` != 0 ) then
    @ n += 1
    set log = ${argv[n]}
   else if ( `echo ${a} | awk '{print index( $1, "s" )}'` != 0 ) then
    @ n += 1
    set createstubs = 1
    set stubdir    = ${argv[$n]}
   else if ( `echo ${a} | awk '{print index( $1, "c" )}'` != 0 ) then
    @ n += 1
    set compile = ( ${argv[$n]} )
   endif
  else
   echo "invalid argument(s): cannot proceed."
   exit 1
  endif

  @ n += 1
 end

# help/usage
#
 if ( ${help} == 1 ) then
  echo 'usage: '${0:t}' [-hFHC] [-l log_file] [-s stubs_directory] -c "compilation_command"'
  echo ''
  echo '-h help/usage'
  echo ''
  echo '-F disables features testing'
  echo '-H disables headers testing'
  echo '-C disables classes testing'
  echo ''
  echo '-l log_file'
  echo '   stdout by default'
  echo ''
  echo '-s stubs_directory'
  echo '   creates stub include files in the directory stubs_directory'
  echo '   for header files with nonstandard names'
  echo ''
  echo '-c compilation command'
  echo '   use quotes whenever necessary; e. g., '${0:t}' -c "g++ -g"'

  exit 0
 endif

# initial part of cppconfig.local
#
 if ( ( ${checkfeatures} == 1 ) && ( ${checkheaders} == 1 ) && ( ${checkclasses} == 1 ) ) then
  echo "#ifdef __cppconfig_local__"
  echo "#define __cppconfig_local__"
  echo ""
  echo "/* compilation command: ${compile} */"
 endif

 echo "${argv}" > ${log}

# check language features
#
 if ( ${checkfeatures} == 1 ) then
  set Prefix = "cpp.feature."

  foreach a ( ${features} )
   echo "\ncheck.feature.csh ${Prefix}${a}.cc ${compile}\n" >>& ${log}
   check.feature.csh ${Prefix}${a}.cc ${compile} >>& ${log}

   set returnstatus = ${status}

   if ( ${returnstatus} == 1 ) echo "#define cpp_missing_${a}"
   if ( ${returnstatus} == 2 ) echo "/* error: non-existent or zero-length file ${Prefix}${a}.cc */"
  end
 endif

# check c++ headers
#
 if ( ${checkheaders} == 1 ) then
  if ( ${checkfeatures} == 1 ) echo ""

  set missingheaders = 0

  foreach a ( ${cppheaders} )
   echo "\ncheck.header.c++.csh ${a} ${createstubs} "${stubdir}" ${compile}\n" >>& ${log}
   check.header.c++.csh ${a} ${createstubs} "${stubdir}" ${compile} >>& ${log}

   set returnstatus = ${status}

   if ( ${returnstatus} == 1 ) then
    echo ""
    echo "/* ${a}: unsupported? exists as ${a}.h? */"
    echo "/* #define cpp_missing_header_${a} */"
    echo ""
    if ( ${createstubs} == 1 ) echo "/* stub created: ${stubdir}/${a} */"
   endif

   if ( ${returnstatus} == 2 ) then
    echo "#define cpp_missing_header_${a}"
    set missingheaders = 1
   endif
  end

  echo ""

  foreach a ( ${cheaders} )
   echo "\ncheck.header.c.csh ${a} ${createstubs} "${stubdir}" ${compile}\n" >>& ${log}
   check.header.c.csh ${a} ${createstubs} "${stubdir}" ${compile} >>& ${log}

   set returnstatus = ${status}

   if ( ${returnstatus} == 1 ) then
    echo ""
    echo "/* c${a}: unsupported? exists as ${a}.h? */"
    echo "/* #define cpp_missing_header_c${a} */"
    echo ""
    if ( ${createstubs} == 1 ) echo "/* stub created: ${stubdir}/c${a} */"
   endif

   if ( ${returnstatus} == 2 ) then
    echo "#define cpp_missing_header_c${a}"
    set missingheaders = 1
   endif
  end

  if ( ${missingheaders} != 0 ) echo "\n#define cpp_missing_std_headers"
 endif

# check standard classes
#
 if ( ${checkclasses} == 1 ) then
  if ( ${checkheaders} == 1 ) echo ""

  set missingclasses = 0

  set n = 1
  while ( ${n} <= ${#classes} )
   echo "\ncheck.class.csh ${classes[${n}]} ${classhdrs[${n}]} "${classdecl[${n}]}" ${compile}\n" >>& ${log}
   check.class.csh ${classes[${n}]} ${classhdrs[${n}]} "${classdecl[${n}]}" ${compile} >>& ${log}

   if ( ${status} == 1 ) then
    echo "#define cpp_missing_class_${classes[${n}]}"
    set missingclasses = 1
   endif

   @ n += 1
  end

  if ( ${missingclasses} != 0 ) echo "\n#define cpp_missing_std_classes"
 endif

# final part of cppconfig.local
#
 if (    ( ${checkfeatures} == 1 ) \
      && ( ${checkheaders}  == 1 ) \
      && ( ${checkclasses}  == 1 ) ) then
  echo ""
  echo "#endif                                                 /* __cppconfig_local__ */"
 endif
