/* --------------------------------------------------------------------------
 *
 * namespace fcomplex
 *   complex number type that is compatible with corresponding fortran types
 *
 * required headers
 *   see include header file fcomplex
 *
 * notes
 *   this implementations is compatible with the fortran representation of
 *   a complex number only for single numbers. This interface would need
 *   substantial changes to make it compatible with fortran complex arrays.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __fcomplex_H__
#define __fcomplex_H__

cpp_using_namespace( std );

template< class T >
struct fcomplex
 {
  private:
   T my_z[2];

  public:
             // a representation of complex numbers compatible with  the fortran
             // complex type, a bare-basics version for interfacing with fortran.

  fcomplex( const T& real = T( ), const T& imag = T( ) );
  fcomplex( const T* z );               // z[0]: real part, z[1]: imaginary part

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

  template<class S>
  fcomplex( const fcomplex<S>& another );

  template<class S>
  fcomplex( const complex<S>& another );

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

  inline complex<T> asSTLcomplex( ) const;

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

  inline T& real( );
  inline T& imag( );
  inline T  real( ) const;
  inline T  imag( ) const;

  inline T& operator[]( bool imag );
  inline T* operator()( );

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

  inline fcomplex<T> conj( ) const;

  inline T norm( ) const;                                               // |z|^2
  inline T mag( )  const;                               //                   |z|
  inline T arg( )  const;                               // t: z = |z| exp( i t )

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

  template<class S>
  inline
  fcomplex<T>&
  operator=( const fcomplex<S>& rhs );

  template<class S>
  inline
  fcomplex<T>&
  operator=( const complex<S>& rhs );

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

  template<class S>
  inline
  fcomplex<T>&
  operator+=( const fcomplex<S>& rhs );

  template<class S>
  inline
  fcomplex<T>&
  operator-=( const fcomplex<S>& rhs );

  template<class S>
  inline
  fcomplex<T>&
  operator*=( const fcomplex<S>& rhs );

  template<class S>
  inline
  fcomplex<T>&
  operator/=( const fcomplex<S>& rhs );

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

  template<class S>
  inline
  fcomplex<T>&
  operator+=( const complex<S>& rhs );

  template<class S>
  inline
  fcomplex<T>&
  operator-=( const complex<S>& rhs );

  template<class S>
  inline
  fcomplex<T>&
  operator*=( const complex<S>& rhs );

  template<class S>
  inline
  fcomplex<T>&
  operator/=( const complex<S>& rhs );
 };

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class T >
fcomplex<T>::fcomplex( const T& real, const T& imag )
 {
  my_z[0] = real; my_z[1] = imag;
 }

template< class T >
fcomplex<T>::fcomplex( const T* z )
 {
  my_z[0] = z[0]; my_z[1] = z[1];
 }

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class T >
template< class S >
fcomplex<T>::fcomplex( const fcomplex<S>& another )
 {
  my_z[0] = static_cast<T>( another.my_z[0] );
  my_z[1] = static_cast<T>( another.my_z[1] );
 }

template< class T >
template< class S >
fcomplex<T>::fcomplex( const complex<S>& another )
 {
  my_z[0] = static_cast<T>( another.real( ) );
  my_z[1] = static_cast<T>( another.imag( ) );
 }

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class T >
complex<T>
fcomplex<T>::asSTLcomplex( ) const
 {
  return complex<T>( my_z[0], my_z[1] );
 }

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class T >
T&
fcomplex<T>::real( )
 {
  return my_z[0];
 }

template< class T >
T&
fcomplex<T>::imag( )
 {
  return my_z[1];
 }

template< class T >
T
fcomplex<T>::real( ) const
 {
  return my_z[0];
 }

template< class T >
T
fcomplex<T>::imag( ) const
 {
  return my_z[1];
 }

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class T >
T&
fcomplex<T>::operator[]( bool imag )
 {
  return my_z[imag];
 }

template< class T >
T*
fcomplex<T>::operator()( )
 {
  return my_z;
 }

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class T >
fcomplex<T>
fcomplex<T>::conj( ) const
 {
  return fcomplex( my_z[0], -my_z[1] );
 }

template< class T >
T
fcomplex<T>::norm( ) const
 {
  return ( my_z[0] * my_z[0] + my_z[1] * my_z[1] );
 }

template< class T >
T
fcomplex<T>::mag( ) const
 {
  return static_cast<T>( sqrt( static_cast<double>( norm( ) ) ) );
 }

template< class T >
T
fcomplex<T>::arg( ) const
 {
  return ( my_z[0] == 0 ) ? 0 : static_cast<T>( tanh( static_cast<double>( my_z[1] / my_z[0] ) ) );
 }

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator=( const fcomplex<S>& rhs )
 {
  if ( this == &rhs ) return *this;

  my_z[0] = static_cast<T>( another.my_z[0] );
  my_z[1] = static_cast<T>( another.my_z[1] );

  return *this;
 }

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator=( const complex<S>& rhs )
 {
  my_z[0] = static_cast<T>( rhs.real( ) );
  my_z[1] = static_cast<T>( rhs.imag( ) );

  return *this;
 }

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator+=( const fcomplex<S>& rhs )
 {
  my_z[0] += static_cast<T>( rhs.my_z[0] );
  my_z[1] += static_cast<T>( rhs.my_z[1] );

  return *this;
 }

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator-=( const fcomplex<S>& rhs )
 {
  my_z[0] -= static_cast<T>( another.my_z[0] );
  my_z[1] -= static_cast<T>( another.my_z[1] );

  return *this;
 }

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator*=( const fcomplex<S>& rhs )
 {
  T x = static_cast<T>( rhs.real( ) );
  T y = static_cast<T>( rhs.imag( ) );

  T t = x * my_z[0] - y * my_z[1];

  my_z[1] = y * my_z[0] + x * my_z[1];
  my_z[0] = t;

  return *this;
 }

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator/=( const fcomplex<S>& rhs )
 {
  ( *this ) *= rhs.conj( );
  ( *this ) /= rhs.norm( );

  return *this;
 }

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator+=( const complex<S>& rhs )
 {
  my_z[0] += static_cast<T>( rhs.real( ) );
  my_z[1] += static_cast<T>( rhs.imag( ) );

  return *this;
 }

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator-=( const complex<S>& rhs )
 {
  my_z[0] -= static_cast<T>( rhs.real( ) );
  my_z[1] -= static_cast<T>( rhs.imag( ) );

  return *this;
 }

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator*=( const complex<S>& rhs )
 {
  T x = static_cast<T>( rhs.real( ) );
  T y = static_cast<T>( rhs.imag( ) );

  T t = x * my_z[0] - y * my_z[1];

  my_z[1] = y * my_z[0] + x * my_z[1];
  my_z[0] = t;

  return *this;
 }

template< class T >
template< class S >
fcomplex<T>&
fcomplex<T>::operator/=( const complex<S>& rhs )
 {
  ( *this ) *= rhs.conj( );
  ( *this ) /= rhs.norm( );

  return *this;
 }

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------

template< class T >
inline
T
real( const fcomplex<T>& z )
 {
  return z.real( );
 }

template< class T >
inline
T
imag( const fcomplex<T>& z )
 {
  return z.imag( );
 }

template< class T >
inline
fcomplex<T>
conj( const fcomplex<T>& z )
 {
  return z.conj( );
 }

template< class T >
inline
T
norm( const fcomplex<T>& z )
 {
  return z.norm( );
 }

template< class T >
inline
T
abs( const fcomplex<T>& z )
 {
  return z.mag( );
 }

template< class T >
inline
T
arg( const fcomplex<T>& z )
 {
  return z.arg( );
 }

template< class T >
inline
complex<T>
asSTLcomplex( const fcomplex<T>& z )
 {
  return z.asSTLcomplex( );
 }

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------

template< class T, class S >
inline
fcomplex<T>
operator+( const fcomplex<T> z, const fcomplex<S>& a )
 {
  fcomplex<T> x( z ); x += a; return x;
 }

template< class T, class S >
inline
fcomplex<T>
operator-( const fcomplex<T> z, const fcomplex<S>& a )
 {
  fcomplex<T> x( z ); x -= a; return x;
 }

template< class T, class S >
inline
fcomplex<T>
operator*( const fcomplex<T> z, const fcomplex<S>& a )
 {
  fcomplex<T> x( z ); x *= a; return x;
 }

template< class T, class S >
inline
fcomplex<T>
operator/( const fcomplex<T> z, const fcomplex<S>& a )
 {
  fcomplex<T> x( z ); x /= a; return x;
 }

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------

template< class T, class S >
inline
fcomplex<T>
operator+( const fcomplex<T> z, const complex<S>& a )
 {
  fcomplex<T> x( z ); x += a; return x;
 }

template< class T, class S >
inline
fcomplex<T>
operator-( const fcomplex<T> z, const complex<S>& a )
 {
  fcomplex<T> x( z ); x -= a; return x;
 }

template< class T, class S >
inline
fcomplex<T>
operator*( const fcomplex<T> z, const complex<S>& a )
 {
  fcomplex<T> x( z ); x *= a; return x;
 }

template< class T, class S >
inline
fcomplex<T>
operator/( const fcomplex<T> z, const complex<S>& a )
 {
  fcomplex<T> x( z ); x /= a; return x;
 }

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------

template< class T, class S >
inline
complex<T>
operator+( const complex<T> z, const fcomplex<S>& a )
 {
  fcomplex<T> x( z ); x += a; return x.asSTLcomplex( );
 }

template< class T, class S >
inline
complex<T>
operator-( const complex<T> z, const fcomplex<S>& a )
 {
  fcomplex<T> x( z ); x -= a; return x.asSTLcomplex( );
 }

template< class T, class S >
inline
complex<T>
operator*( const complex<T> z, const fcomplex<S>& a )
 {
  fcomplex<T> x( z ); x *= a; return x.asSTLcomplex( );
 }

template< class T, class S >
inline
complex<T>
operator/( const complex<T> z, const fcomplex<S>& a )
 {
  fcomplex<T> x( z ); x /= a; return x.asSTLcomplex( );
 }

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------

/* -----------------------

// g++ doesn't seem to support basic_?stream<...> yet.

// template< class T, class Ch, class Tr >
// basic_istream<Ch,Tr>&
// operator>>( basic_istream<Ch,Tr>& is, fcomplex<T>& z );

// template< class T, class Ch, class Tr >
// basic_ostream<Ch,Tr>&
// operator<<( basic_ostream<Ch,Tr>& os, const fcomplex<T>& z );
   ----------------------- */

template< class T >
istream&
operator>>( istream& is, fcomplex<T>& z )
 {
  complex<T> x; is >> x; z = x; return is;
 }

template< class T >
ostream&
operator<<( ostream& os, const fcomplex<T>& z )
 {
  complex<T> x = z.asSTLcomplex( ); os << x; return os;
 }

#endif /* __fcomplex_H__ */
