/* --------------------------------------------------------------------------
 *
 * template class matrix
 *   simple, efficient wrapper class for matrices;
 *   compatible with fortran conventions of column-major
 *   order and a contiguous internal representation
 *
 * required headers
 *   see include header file matrix
 *
 * notes
 *   uses assert( ); use NDEBUG to disable.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __matrix_H__
#define __matrix_H__

cpp_using_namespace( std );

template<class T>
class matrix
 {
  public:
   // ---------------------------------------------------------- ctor/cctor/dtor
   // --------------------------------------------------------------------------

   matrix( );

   matrix( size_t mm, size_t nn, const T& value = T( ) );

   matrix(       size_t mm,
                 size_t nn,
           const T*     cArr,
                 size_t length,
           const T&     value = T( ) );
                        //
                        //  create an mm X nn matrix, copy first length elements
                        // from buffer cArr assumed to be in column-major order,
                        //          initialize the rest of the matrix with value

   matrix( size_t mm, size_t nn, T* cArr, bool wrapper = true );
                       //
                       //    this constructor allows the possibility of using an
                       //  existing matrix buffer in column-major format through
                       //                                         this interface
                       //
                       // If wrapper = true,  this class acts as a wrapper,  and
                       //    allows all operations except those where the buffer
                       //               would be destroyed (resize, assignment).
                       //          The last argument is irrelevant in this case.
                       //
                       //   If wrapper = false, a copy of the buffer is created,
                       //      and all operations including resize are permitted
                       //
                       //  In either case, size of the buffer cArr is assumed to
                       //                                            be mm * nn.

   matrix( const matrix<T>& another );

  ~matrix( );

   // --------------------------------------------------------------- assignment
   // --------------------------------------------------------------------------

   inline
   matrix<T>&
   operator=( const matrix<T>& rhs );

   inline
   matrix<T>&
   operator=( const T& value );

   inline
   matrix<T>&
   operator+=( const T& value );

   inline
   matrix<T>&
   operator-=( const T& value );

   inline
   matrix<T>&
   operator*=( const T& value );

   inline
   matrix<T>&
   operator/=( const T& value );

   // -------------------------------------------------------------------- sizes
   // --------------------------------------------------------------------------

   inline
   pair<size_t, size_t>
   size( ) const;

   inline
   size_t
   numRows( ) const;

   inline
   size_t
   numCols( ) const;

   inline
   bool
   isResizable( ) const;
                             //
                             //  matrix objects that only serve to wrap external
                             // column-major buffers into a convenient interface
                             //                                are not resizable

   // ------------------------------------------------- unchecked element access
   // --------------------------------------------------------------------------

   inline
   T&
   operator()( size_t rowIndex = 0, size_t colIndex = 0 );

   inline
   T
   operator()( size_t rowIndex = 0, size_t colIndex = 0 ) const;

   // --------------------------------------------------- checked element access
   // --------------------------------------------------------------------------

   class safeValue                             // proxy class for checked access
    {
     public:
      safeValue& operator=( const T& value )
       {
        if ( isValid( ) ) (*my_valptr) = value;
        return *this;
       };

      operator T( ) const
       {
        return value( );
       };

      bool isValid( ) const { return ( my_valptr != 0 ); }

      T value( ) const { return ( isValid( ) ? (*my_valptr) : T( ) ); }

     private:
      T* my_valptr;

      friend class matrix<T>;

      safeValue( matrix<T>& m, size_t rowIndex, size_t colIndex )
       : my_valptr(   (    ( rowIndex < m.numRows( ) )
                        && ( colIndex < m.numCols( ) ) )
                    ? &m( rowIndex, colIndex ) : 0 )
       {
       }
    };

   inline
   safeValue
   safeAccess( size_t rowIndex, size_t colIndex );

   inline
   const safeValue
   safeAccess( size_t rowIndex, size_t colIndex ) const;

   // ---------------------------------------------------- row and column access
   // --------------------------------------------------------------------------

   varray<T>
   column( size_t colIndex ) const;

   varray<T>
   row( size_t rowIndex ) const;

   void
   setCol( size_t colIndex, const varray<T>& column );

   void
   setRow( size_t rowIndex, const varray<T>& column );

   void
   setCol( size_t colIndex, const T* cArr, size_t length );

   void
   setRow( size_t rowIndex, const T* cArr, size_t length );

   void
   setCol( size_t colIndex, const T& value = T( ) );

   void
   setRow( size_t rowIndex, const T& value = T( ) );

   // ------------------------------------------------------------------- resize
   // --------------------------------------------------------------------------

   void
   resize( size_t numRows, size_t numCols,
           bool preserve = true, const T& value = T( ) );
                        //
                        //               resize matrix if isResizable( ) == true
                        //
                        // preserve = true:  preserve existing values in matrix;
                        //                  initialize any extra ones with value
                        // preserve = false:    do not preserve existing values;
                        //                             initialize all with value

   void
   resize( size_t numRows, size_t numCols, const T* cArr );
                        //
                        //              resize matrix is isResizable( ) == true;
                        // initialize using cArr  (assumed to be of size numRows
                        //                       * numCols, column-major format)

   // -------------------------------------------------------- internal pointers
   // --------------------------------------------------------------------------

         //        this function is intended to provide ease of interfacing with
         // fortran-based libraries such as lapack or blas.    Note: [][] access
         //               has order opposite to that of two-dimentional c-arrays

   inline
   const T*
   operator[]( size_t colIndex ) const;

   inline
   T*
   operator[]( size_t colIndex );

  private:
   bool   my_isResizable;
   T*     my_data;
   size_t my_bufSize;
   size_t my_numRows;
   size_t my_numCols;

   inline
   void
   destroy( );                                    // destroy data: reset to null

   inline
   size_t
   cmajor( size_t numRows, size_t column, size_t row = 0 ) const;
 };

// ----------------------------------------------------------- private functions
// -----------------------------------------------------------------------------

template<class T>
void
matrix<T>::destroy( )
 {
  delete [] my_data;
  my_numRows = my_numCols = my_bufSize = 0;
  my_data = 0;
 }

template<class T>
size_t
matrix<T>::cmajor( size_t numRows, size_t column, size_t row /* = 0 */ ) const
 {
  return ( numRows * column + row );
 }

// ------------------------------------------------------------- ctor/cctor/dtor
// -----------------------------------------------------------------------------

template<class T>
matrix<T>::matrix( )
 : my_isResizable( true ),
   my_data( 0 ),
   my_bufSize( 0 ),
   my_numRows( 0 ),
   my_numCols( 0 )
 {
 }

template<class T>
matrix<T>::matrix( size_t mm, size_t nn, const T& value /* = T( ) */ )
 : my_isResizable( true ),
   my_data( mm * nn ? new T[mm * nn] : 0 ),
   my_bufSize( my_data ? mm * nn : 0 ),
   my_numRows( my_data ? mm : 0 ),
   my_numCols( my_data ? nn : 0 )
 {
  (*this) = value;
 }

template<class T>
matrix<T>::matrix(       size_t mm,
                         size_t nn,
                   const T*     cArr,
                         size_t length,
                   const T&     value /* = T( )  */ )
 : my_isResizable( true ),
   my_data( mm * nn ? new T[mm * nn] : 0 ),
   my_bufSize( my_data ? mm * nn : 0 ),
   my_numRows( my_data ? mm : 0 ),
   my_numCols( my_data ? nn : 0 )
 {
  if ( cArr && my_data )
   memcpy( my_data, cArr, min( my_bufSize, length ) * sizeof( T ) );

  if ( length < my_bufSize )
   for( size_t i = length; i < my_bufSize; i++ ) my_data[i] = value;
 }

template<class T>
matrix<T>::matrix( size_t mm, size_t nn, T* cArr, bool wrapper /* = true */ )
 : my_isResizable( !wrapper || ( mm * nn == 0 ) || ( cArr == 0 ) ),
   my_data( my_isResizable ? ( mm * nn ? new T[mm * nn] : 0 ) : cArr ),
   my_bufSize( my_data ? mm * nn : 0 ),
   my_numRows( my_data ? mm : 0 ),
   my_numCols( my_data ? nn : 0 )
 {
  if ( my_isResizable && cArr && my_data )
   memcpy( my_data, cArr, my_bufSize * sizeof( T ) );
 }

template<class T>
matrix<T>::matrix( const matrix<T>& another )
 : my_isResizable( true ),
   my_data( another.my_bufSize ? new T[another.my_bufSize] : 0 ),
   my_bufSize( my_data ? another.my_bufSize : 0 ),
   my_numRows( my_data ? another.my_numRows : 0 ),
   my_numCols( my_data ? another.my_numCols : 0 )
 {
  if ( my_data && another.my_data )
   memcpy( my_data, another.my_data, my_bufSize * sizeof( T ) );
 }

template<class T>
matrix<T>::~matrix( )
 {
  if ( my_isResizable ) destroy( );
 }

// ------------------------------------------------------------------ assignment
// -----------------------------------------------------------------------------

template<class T>
matrix<T>&
matrix<T>::operator=( const matrix<T>& rhs )
 {
  if ( &rhs == this ) return *this;

                 // ATTN this needs more thought: (a) when this is not resizable
                 // ATTN             (b) when this and rhs share the same buffer

  if ( my_isResizable )
   {
    destroy( );

    my_data = rhs.my_bufSize ? new T[rhs.my_bufSize] : 0;

    my_bufSize = my_data ? rhs.my_bufSize : 0;
    my_numRows = my_data ? rhs.my_numRows : 0;
    my_numCols = my_data ? rhs.my_numCols : 0;

    if ( my_data && rhs.my_data )
     memcpy( my_data, rhs.my_data, my_bufSize * sizeof( T ) );
   }
  else
   {
    if ( my_data != rhs.my_data )
     {
      memcpy( my_data, rhs.my_data,
              min( my_bufSize, rhs.my_bufSize ) * sizeof( T ) );

      if ( my_bufSize > rhs.my_bufSize )
       for ( size_t i = rhs.my_bufSize; i < my_bufSize; i++ )
        my_data[i] = T( );
     }
   }

  return *this;
 }

template<class T>
matrix<T>&
matrix<T>::operator=( const T& value )
 {
  for ( size_t i = 0; i < my_bufSize; i++ ) my_data[i] = value;
  return *this;
 }

template<class T>
matrix<T>&
matrix<T>::operator+=( const T& value )
 {
  for ( size_t i = 0; i < my_bufSize; i++ ) my_data[i] += value;
  return *this;
 }

template<class T>
matrix<T>&
matrix<T>::operator-=( const T& value )
 {
  for ( size_t i = 0; i < my_bufSize; i++ ) my_data[i] -= value;
  return *this;
 }

template<class T>
matrix<T>&
matrix<T>::operator*=( const T& value )
 {
  for ( size_t i = 0; i < my_bufSize; i++ ) my_data[i] *= value;
  return *this;
 }

template<class T>
matrix<T>&
matrix<T>::operator/=( const T& value )
 {
  for ( size_t i = 0; i < my_bufSize; i++ ) my_data[i] /= value;
  return *this;
 }

// ----------------------------------------------------------------------- sizes
// -----------------------------------------------------------------------------

template<class T>
pair<size_t, size_t>
matrix<T>::size( ) const
 {
  return pair<size_t, size_t>( my_numRows, my_numCols );
 }

template<class T>
size_t
matrix<T>::numRows( ) const
 {
  return my_numRows;
 }

template<class T>
size_t
matrix<T>::numCols( ) const
 {
  return my_numCols;
 }

template<class T>
bool
matrix<T>::isResizable( ) const
 {
  return my_isResizable;
 }

// ---------------------------------------------------- unchecked element access
// -----------------------------------------------------------------------------

template<class T>
T&
matrix<T>::operator()( size_t rowIndex, size_t colIndex )
 {
  return my_data[cmajor( my_numRows, colIndex, rowIndex )];
 }

template<class T>
T
matrix<T>::operator()( size_t rowIndex, size_t colIndex ) const
 {
  return my_data[cmajor( my_numRows, colIndex, rowIndex )];
 }

// ------------------------------------------------------ checked element access
// -----------------------------------------------------------------------------

template<class T>
matrix<T>::safeValue
matrix<T>::safeAccess( size_t rowIndex, size_t colIndex )
 {
  return safeValue( *this, rowIndex, colIndex );
 }

template<class T>
const matrix<T>::safeValue
matrix<T>::safeAccess( size_t rowIndex, size_t colIndex ) const
 {
  return safeValue( *this, rowIndex, colIndex );
 }

// ------------------------------------------------------- row and column access
// -----------------------------------------------------------------------------

template<class T>
varray<T>
matrix<T>::column( size_t colIndex ) const
 {
  if ( colIndex >= my_numCols )
   return varray<T>( );
  else
   return varray<T>( my_data + cmajor( my_numRows, colIndex ), my_numRows );
 }

template<class T>
varray<T>
matrix<T>::row( size_t rowIndex ) const
 {
  if ( rowIndex >= my_numRows )
   return varray<T>( );
  else
   {
    varray<T> row( my_numCols );

    for ( size_t j = 0, i = rowIndex; j < my_numCols; j++, i += my_numRows )
     row[j] = my_data[i];

    return row;
   }
 }

template<class T>
void
matrix<T>::setCol( size_t colIndex, const varray<T>& column )
 {
  if ( colIndex < my_numCols )
   for ( size_t j = 0, i = cmajor( my_numRows, colIndex );
                j < min( column.size( ), my_numRows ); j++, i++ )
    my_data[i] = column[j];
 }

template<class T>
void
matrix<T>::setRow( size_t rowIndex, const varray<T>& row )
 {
  if ( rowIndex < my_numRows )
   for ( size_t i = 0, j = rowIndex;
                i < min( my_numCols, row.size( ) ); i++, j += my_numRows )
    my_data[j] = row[i];
 }

template<class T>
void
matrix<T>::setCol( size_t colIndex, const T* cArr, size_t length )
 {
  if ( ( colIndex < my_numCols ) && cArr )
   memcpy( my_data + cmajor( my_numRows, colIndex ), cArr,
           min( my_numRows, length ) * sizeof( T ) );
 }

template<class T>
void
matrix<T>::setRow( size_t rowIndex, const T* cArr, size_t length )
 {
  if ( ( rowIndex < my_numRows ) && cArr )
   for ( size_t j = 0, i = rowIndex;
                j < min( my_numCols, length ); j++, i += my_numRows )
    my_data[i] = cArr[j];
 }

template<class T>
void
matrix<T>::setCol( size_t colIndex, const T& value /* = T( ) */ )
 {
  if ( colIndex < my_numCols )
   for ( size_t j = 0, i = cmajor( my_numRows, colIndex );
                j < my_numRows; j++, i++ )
    my_data[i] = value;
 }

template<class T>
void
matrix<T>::setRow( size_t rowIndex, const T& value /* = T( ) */ )
 {
  if ( rowIndex < my_numRows )
   for ( size_t i = 0, j = rowIndex; i < my_numCols; i++, j += my_numRows )
    my_data[j] = value;
 }

// ---------------------------------------------------------------------- resize
// -----------------------------------------------------------------------------

template<class T>
void
matrix<T>::resize( size_t newrows, size_t newcols,
                   bool preserve /* = true */, const T& value /* = T( ) */ )
 {
  if ( !my_isResizable ) return;

  if ( newrows * newcols == 0 )
   {
    destroy();
    return;
   }

  if ( preserve )
   {
    if ( newrows * newcols <= my_bufSize )
     {
      if ( newrows == my_numRows )
       {
        if ( newcols > my_numCols )
         {
          for ( size_t l = my_numRows * my_numCols; l < newcols * newrows; l++ )
           my_data[l] = value;
         }
       }
      else if ( newrows > my_numRows )
       {
        for ( size_t i = newcols * my_numRows - 1,
                     inc = newcols * (newrows - my_numRows);
              i >= my_numRows - 1; i-- )
         {
          if ( !( ( i + inc + 1 ) % newrows ) )
           {
            inc -= newrows -  my_numRows;

            my_data[i + inc] = my_data[i];

            for ( size_t k = 1; k <= ( newrows - my_numRows ); k++ )
             my_data[i + inc + k] = value;
           }
          else my_data[i + inc] = my_data[i];
         }

        if ( newcols > my_numCols )
         {
          for ( size_t l = newrows * my_numCols; l < newcols * newrows; l++ )
           my_data[l] = value;
         }
       }
      else if ( newrows <  my_numRows )
       {
        for ( size_t i = newrows, inc = 0; i < newcols * newrows; i++ )
         {
          if ( !( ( i + inc + ( my_numRows - newrows ) ) % my_numRows ) )
           inc += my_numRows - newrows;

          my_data[i] = my_data[i+inc];
         }

        if ( newcols > my_numCols )
         {
          for ( size_t l = newrows * my_numCols; l < newcols * newrows; l++ )
           my_data[l] = value;
         }
       }
     }
    else
     {
      size_t newsize = newrows * newcols;

      T* tmp = new T[newsize];
      assert( tmp != 0 );
      T* old_ptr;
      T* new_ptr;

      for ( size_t column = 0; column < min(my_numCols, newcols); column++ )
       {
        old_ptr = my_data + column * my_numRows;
        new_ptr = tmp + column * newrows;

        memcpy( new_ptr, old_ptr, min( newrows, my_numRows ) * sizeof( T ) );

        if ( newrows > my_numRows )
         for(size_t r = my_numRows; r < newrows; r++) new_ptr[r] = value;
       }

      if ( newcols > my_numCols )
       for(size_t column = my_numCols; column < newcols; column++)
        {
         new_ptr = tmp + column * newrows;
         for ( size_t r = 0; r < newrows; r++ ) new_ptr[r] = value;
        }

       delete [] my_data;
       my_data = tmp;
       my_bufSize = newsize;
     }
   }
  else if( newrows * newcols > my_bufSize )   // !preserve but new buffer needed
   {
    my_bufSize = newrows * newcols;
    delete [] my_data;
    my_data = new T[my_bufSize];

    (*this) = value;
   }
  else                                    // !preserve but new buffer not needed
   (*this) = value;

                                            // this is common to all three cases
                                            //
   my_numRows = newrows;
   my_numCols = newcols;
 }

template<class T>
void
matrix<T>::resize( size_t newrows, size_t newcols, const T* cArr )
 {
  if ( my_isResizable && cArr )
   {
    resize( newrows, newcols, false );
    if ( my_data ) memcpy( my_data, cArr, newrows * newcols * sizeof( T ) );
   }
 }

// ----------------------------------------------------------- internal pointers
// -----------------------------------------------------------------------------

template<class T>
const T*
matrix<T>::operator[]( size_t colIndex ) const
 {
  return
   ( my_data == 0 ) ? 0 : my_data + cmajor( my_numRows, colIndex );
 }

template<class T>
T*
matrix<T>::operator[]( size_t colIndex )
 {
  return
   ( my_data == 0 ) ? 0 : my_data + cmajor( my_numRows, colIndex );
 }

#endif /* __matrix_H__ */
