/* --------------------------------------------------------------------------
 *
 * standard template class varray
 *   a subset of the standard valarray (draft standard November 1997)
 *
 * required headers
 *   see include header file varray
 *
 * notes
 *   this class is a bare-basics version of the standard valarray,
 *   with the following useful additions/changes:
 *    1 const T* operator()( size_t i ) const;
 *            T* operator()( size_t i );
 *    2 resize( ) takes the extra argument 'preserve'
 *    3 can act as a wrapper for a c-style array (not resizable in that case)
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __varray_H__
#define __varray_H__

cpp_using_namespace( std );

template<class T>
class varray
 {
  public:

  // ---------------------------------------------------------- ctors/cctor/dtor
  // ---------------------------------------------------------------------------

    varray( );

    explicit
    varray( size_t size );

    varray( const T& value, size_t size );

    varray( const T* cArray, size_t size );

    varray( T* cArray, size_t size, bool wrapper = false );

    varray( const varray<T>& another );

    ~varray( );

  // ---------------------------------------------------------------- assignment
  // ---------------------------------------------------------------------------

    inline
    varray<T>&
    operator=( const varray<T>& rhs );

    inline
    varray<T>&
    operator=( const T& rhs );

  // ------------------------------------------------------------ element access
  // ---------------------------------------------------------------------------

    inline
    T&
    operator[]( size_t index );

    inline
    T
    operator[]( size_t index ) const;

    inline
    const T*
    operator()( size_t index = 0 ) const;

    inline
    T*
    operator()( size_t index = 0 );

  // --------------------------------------------------------------- size/resize
  // ---------------------------------------------------------------------------

    inline
    size_t
    size( ) const;

    inline
    void
    resize( size_t newSize, const T& value = T( ), bool preserve = false );

  // ------------------------------------------------------- computed assignment
  // ---------------------------------------------------------------------------

    inline
    varray<T>&
    operator+=( const T& rhs );

    inline
    varray<T>&
    operator-=( const T& rhs );

    inline
    varray<T>&
    operator*=( const T& rhs );

    inline
    varray<T>&
    operator/=( const T& rhs );

    inline
    varray<T>&
    operator+=( const varray<T>& rhs );

    inline
    varray<T>&
    operator-=( const varray<T>& rhs );

    inline
    varray<T>&
    operator*=( const varray<T>& rhs );

    inline
    varray<T>&
    operator/=( const varray<T>& rhs );

  // ----------------------------------------------------------- unary operators
  // ---------------------------------------------------------------------------

    inline
    varray<T>
    operator+( ) const;

    inline
    varray<T>
    operator-( ) const;

  // ----------------------------------------------------------- private members
  // ---------------------------------------------------------------------------

  private:
    T*     my_data;
    size_t my_size;
    bool   my_resizable;

    inline
    void
    assignValue( const T& value = T( ) );

    inline
    void
    copyArray( const T* cArray, size_t cArraySize );

    inline
    void
    reset2null( );
 };

template<class T>
bool
operator==( const varray<T>& x, const varray<T>& y )
 {
  bool same = ( x.size( ) == y.size( ) );
  for ( size_t i = 0; same && ( i < x.size( ) ); i++ ) same = same && ( x[i] == y[i] );
  return same;
 }

template<class T>
bool
operator!=( const varray<T>& x, const varray<T>& y )
 {
  return !( x == y );
 }

// ------------------------------------------------------------ helper functions
// -----------------------------------------------------------------------------

template<class T>
inline
vector<T>
asSTLvector( const varray<T>& v )
 {
  vector<T> t( v.size( ) );
  for ( size_t i = 0; i < v.size( ); i++ ) t[i] = v[i];
  return t;
 }

template<class T>
inline
varray<T>
reverse( const varray<T>& v )
 {
  varray<T> t( v.size( ) );
  for ( size_t i = 0, j = v.size( ) - 1; i < v.size( ); i++, j-- ) t[i] = v[j];
  return t;
 }

// ----------------------------------------------------------- private functions
// -----------------------------------------------------------------------------

template<class T>
void
varray<T>::assignValue( const T& value )
 {
  for ( size_t i = 0; i < my_size; i++ ) my_data[i] = value;
 }

template<class T>
void
varray<T>::copyArray( const T* cArray, size_t cArraySize )
 {
  // for ( size_t i = 0; i < min( my_size, cArraySize ); i++ )
  //  my_data[i] = cArray[i];

  memcpy( my_data, cArray, min( my_size, cArraySize ) * sizeof( T ) );
 }

template<class T>
void
varray<T>::reset2null( )
 {
  my_size = 0;
  delete [] my_data;
  my_data = 0;
 }

// ------------------------------------------------------------ ctors/cctor/dtor
// -----------------------------------------------------------------------------

template<class T>
varray<T>::varray( size_t size )
 : my_data( size ? new T[size] : 0 ),
   my_size( my_data ? size : 0 ),
   my_resizable( true )
 {
  assignValue( );
 }

template<class T>
varray<T>::varray( )
 : my_data( 0 ), my_size( 0 ), my_resizable( true )
 {
 }

template<class T>
varray<T>::varray( const T& value, size_t size )
 : my_data( size ? new T[size] : 0 ),
   my_size( my_data ? size : 0 ),
   my_resizable( true )
 {
  assignValue( value );
 }

template<class T>
varray<T>::varray( const T* cArray, size_t size )
 : my_data( size ? new T[size] : 0 ),
   my_size( my_data ? size : 0 ),
   my_resizable( true )
 {
  if ( my_data && cArray ) copyArray( cArray, my_size );
 }

template<class T>
varray<T>::varray( T* cArray, size_t size, bool wrapper )
 : my_data( wrapper ? cArray : ( size ? new T[size] : 0 ) ),
   my_size( my_data ? size : 0 ),
   my_resizable( !wrapper )
 {
  if ( my_data && cArray && my_resizable ) copyArray( cArray, my_size );
 }

template<class T>
varray<T>::varray( const varray<T>& another )
 : my_data( another.my_size ? new T[another.my_size] : 0 ),
   my_size( my_data ? another.my_size : 0 ),
   my_resizable( true )
 {
  if ( my_data && another.my_data ) copyArray( another.my_data, another.my_size );
 }

template<class T>
varray<T>::~varray( )
 {
  reset2null( );
 }

// ------------------------------------------------------------------ assignment
// -----------------------------------------------------------------------------

template<class T>
varray<T>&
varray<T>::operator=( const varray<T>& rhs )
 {
  if ( ( &rhs != this ) && my_data && rhs.my_data )
   copyArray( rhs.my_data, rhs.my_size );

  return *this;
 }

template<class T>
varray<T>&
varray<T>::operator=( const T& rhs )
 {
  assignValue( rhs );

  return *this;
 }

// -------------------------------------------------------------- element access
// -----------------------------------------------------------------------------

template<class T>
T&
varray<T>::operator[]( size_t index )
 {
  return my_data[index];
 }

template<class T>
T
varray<T>::operator[]( size_t index ) const
 {
  return my_data[index];
 }

template<class T>
const T*
varray<T>::operator()( size_t index ) const
 {
  return ( my_data + index );
 }

template<class T>
T*
varray<T>::operator()( size_t index )
 {
  return ( my_data + index );
 }

// ----------------------------------------------------------------- size/resize
// -----------------------------------------------------------------------------

template<class T>
size_t
varray<T>::size( ) const
 {
  return my_size;
 }

template<class T>
void
varray<T>::resize( size_t newSize, const T& value, bool preserve )
 {
  if ( !my_resizable ) return;

  if ( newSize == 0 ) { reset2null( ); return; }

  T* tPtr = new T[newSize];

  if ( tPtr )
   {
    if ( preserve )
     {
      size_t k = min( newSize, my_size );
      for ( size_t i = 0; i < k; i++ ) tPtr[i] = my_data[i];
      for ( size_t i = k; i < newSize; i++ ) tPtr[i] = value;
     }
    else
     for ( size_t i = 0; i < newSize; i++ ) tPtr[i] = value;

    reset2null( );

    my_data = tPtr;
    my_size = newSize;
   }
 }

// --------------------------------------------------------- computed assignment
// -----------------------------------------------------------------------------

template<class T>
varray<T>&
varray<T>::operator+=( const T& rhs )
 {
  for ( size_t i = 0; i < my_size; i++ ) my_data[i] += rhs;

  return *this;
 }

template<class T>
varray<T>&
varray<T>::operator-=( const T& rhs )
 {
  for ( size_t i = 0; i < my_size; i++ ) my_data[i] -= rhs;

  return *this;
 }

template<class T>
varray<T>&
varray<T>::operator*=( const T& rhs )
 {
  for ( size_t i = 0; i < my_size; i++ ) my_data[i] *= rhs;

  return *this;
 }

template<class T>
varray<T>&
varray<T>::operator/=( const T& rhs )
 {
  for ( size_t i = 0; i < my_size; i++ ) my_data[i] /= rhs;

  return *this;
 }

template<class T>
varray<T>&
varray<T>::operator+=( const varray<T>& rhs )
 {
  for ( size_t i = 0; i < min( my_size, rhs.my_size ); i++ )
   my_data[i] += rhs.my_data[i];

  return *this;
 }

template<class T>
varray<T>&
varray<T>::operator-=( const varray<T>& rhs )
 {
  for ( size_t i = 0; i < min( my_size, rhs.my_size ); i++ )
   my_data[i] -= rhs.my_data[i];

  return *this;
 }

template<class T>
varray<T>&
varray<T>::operator*=( const varray<T>& rhs )
 {
  for ( size_t i = 0; i < min( my_size, rhs.my_size ); i++ )
   my_data[i] *= rhs.my_data[i];

  return *this;
 }

template<class T>
varray<T>&
varray<T>::operator/=( const varray<T>& rhs )
 {
  for ( size_t i = 0; i < min( my_size, rhs.my_size ); i++ )
   my_data[i] /= rhs.my_data[i];

  return *this;
 }

// ------------------------------------------------------------- unary operators
// -----------------------------------------------------------------------------

template<class T>
varray<T>
varray<T>::operator+( ) const
 {
  return varray<T>( *this );
 }

template<class T>
varray<T>
varray<T>::operator-( ) const
 {
  varray<T> tmp( my_size );

  for ( size_t i = 0; i < my_size; i++ ) tmp[i] = - my_data[i];

  return tmp;
 }

#endif /* __varray_H__ */
