/* --------------------------------------------------------------------------
 *
 * namespace cppblas
 *   C++ prototypes for fortran BLAS functions
 *
 * required headers
 *   see include header file cppblas
 *
 * notes
 *   1 function arguments have been changed into const/non-const references
 *     wherever appropriate.
 *   2 functions involving complex arguments (the c- and z- varieties) have
 *     not been included for want of a c++ representation of complex numbers
 *     that is compatible with the fortran representation, both in single
 *     numbers and arrays.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __cppblas_H__
#define __cppblas_H__

cpp_using_namespace( FortranTypes );

cpp_gnamespace_begin( cppblas )

  // -------------------------------- constants for blas level 2 and 3 functions
  // ---------------------------------------------------------------------------

  static const character NoTranspose        = 'N'; // values for arguments trans
  static const character Transpose          = 'T'; //          transa and transb
  static const character ConjugateTranspose = 'C';

  static const character UpperTriangle      = 'U'; //   values for argument uplo
  static const character LowerTriangle      = 'L';

  static const character NonUnitTriangular  = 'N'; //   values for argument diag
  static const character UnitTriangular     = 'U';

  static const character LeftSide           = 'L'; //   values for argument side
  static const character RightSide          = 'R';

  // ------------------------------------------------------- function prototypes
  // ---------------------------------------------------------------------------

  extern "C"
   {

   // ----------------------------------------------------- auxilliary functions
   // --------------------------------------------------------------------------

    logical
    lsame
    ( const character& ca, const character& cb );

    void
    xerbla
    ( const character* srname, const integer& info );

   // ------------------------------------------------- level 1: plane rotations
   // --------------------------------------------------------------------------

    void
    srotg
    ( sReal& a, sReal& b, sReal& c, sReal& s );

    void
    drotg
    ( dReal& a, dReal& b, dReal& c, dReal& s );

    void
    srotmg
    ( sReal& d1, sReal& d2, sReal& a, sReal& b, sReal* param );

    void
    drotmg
    ( dReal& d1, dReal& d2, dReal& a, dReal& b, dReal* param );

    void
    srot
    ( const integer& n,
      sReal* x, const integer& incx,
      sReal* y, const integer& incy,
      const sReal& c, const sReal& s );

    void
    drot
    ( const integer& n,
      dReal* x, const integer& incx,
      dReal* y, const integer& incy,
      const dReal& c, const dReal& s );

    void
    srotm
    ( const integer& n,
      sReal* x, const integer& incx,
      sReal* y, const integer& incy,
      const sReal* param );

    void
    drotm
    ( const integer& n,
      dReal* x, const integer& incx,
      dReal* y, const integer& incy,
      const dReal* param );

   // ---------------------------------------------------- level 1: swap x <-> y
   // --------------------------------------------------------------------------

    void
    sswap
    ( const integer& n,
      sReal* x, const integer& incx,
      sReal* y, const integer& incy );

    void
    dswap
    ( const integer& n,
      dReal* x, const integer& incx,
      dReal* y, const integer& incy );

    /*
    void
    cswap
    ( const integer& n,
      sComplex* x, const integer& incx,
      sComplex* y, const integer& incy );

    void
    zswap
    ( const integer& n,
      dComplex* x, const integer& incx,
      dComplex* y, const integer& incy );
    */

   // -------------------------------------------------- level 1: x <- alpha * x
   // --------------------------------------------------------------------------

    void
    sscal
    ( const integer& n, const sReal& alpha, sReal* x, const integer& incx );

    void
    dscal
    ( const integer& n, const dReal& alpha, dReal* x, const integer& incx );

    /*
    void
    cscal
    ( const integer& n,
      const sComplex& alpha,
      sComplex& x, const integer& incx );

    void
    zscal
    ( const integer& n,
      const dComplex& alpha,
      dComplex& x, const integer& incx );

    void
    csscal
    ( const integer& n,
      const sReal& alpha,
      sComplex& x, const integer& incx );

    void
    zdscal
    ( const integer& n,
      const dReal& alpha,
      dComplex& x, const integer& incx );
    */

   // ---------------------------------------------------------- level 1: y <- x
   // --------------------------------------------------------------------------

    void
    scopy
    ( const integer& n,
      const sReal* x, const integer& incx,
      sReal* y, const integer& incy );

    void
    dcopy
    ( const integer& n,
      const dReal* x, const integer& incx,
      dReal* y, const integer& incy );

    /*
    void
    ccopy
    ( const integer& n,
      const sComplex& x, const integer& incx,
      sComplex& y, const integer& incy );

    void
    zcopy
    ( const integer& n,
      const dComplex& x, const integer& incx,
      dComplex& y, const integer& incy );
    */

   // ---------------------------------------------- level 1: y <- alpha * x + y
   // --------------------------------------------------------------------------

    void
    saxpy
    ( const integer& n, const sReal& alpha,
      const sReal* x, const integer& incx,
      sReal* y, const integer& incy );

    void
    daxpy
    ( const integer& n, const dReal& alpha,
      const dReal* x, const integer& incx,
      dReal* y, const integer& incy );

    /*
    void
    caxpy
    ( const integer& n,
      const sComplex& alpha,
      const sComplex& x, const integer& incx,
      sComplex& y, const integer& incy );

    void
    zaxpy
    ( const integer& n,
      const dComplex& alpha,
      const dComplex& x, const integer& incx,
      dComplex& y, const integer& incy );
    */

   // -------------------------------------------------- level 1: dot <- x^T * y
   // --------------------------------------------------------------------------

    sReal
    sdot
    ( const integer& n,
      const sReal* x, const integer& incx,
      const sReal* y, const integer& incy );

    dReal
    ddot
    ( const integer& n,
      const dReal* x, const integer& incx,
      const dReal* y, const integer& incy );

    dReal
    dsdot
    ( const integer& n,
      const sReal* x, const integer& incx,
      const sReal* y, const integer& incy );

   // ------------------------------------------ level 1: dot <- alpha + x^T * y
   // --------------------------------------------------------------------------

    sReal
    sdsdot
    ( const integer& n, const sReal& alpha,
      const sReal* x, const integer& incx,
      const sReal* y, const integer& incy );

   // ------------------------------------------------- level 1: dotu <- x^T * y
   // --------------------------------------------------------------------------

    /*
    sComplex
    cdotu
    ( const integer& n,
      const sComplex& x, const integer& incx,
      const sComplex& y, const integer& incy );

    dComplex
    zdotu
    ( const integer& n,
      const dComplex& x, const integer& incx,
      const dComplex& y, const integer& incy );
    */

   // ------------------------------------------------- level 1: dotc <- x^H * y
   // --------------------------------------------------------------------------

    /*
    sComplex
    cdotc
    ( const integer& n,
      const sComplex& x, const integer& incx,
      const sComplex& y, const integer& incy );

    dComplex
    zdotc
    ( const integer& n,
      const dComplex& x, const integer& incx,
      const dComplex& y, const integer& incy );
    */

   // ------------------------------------------------- level 1: nrm2 <- ||x||_2
   // --------------------------------------------------------------------------

    sReal
    snrm2
    ( const integer& n, const sReal* x, const integer& incx );

    dReal
    dnrm2
    ( const integer& n, const dReal* x, const integer& incx );

    /*
    sReal
    scnrm2
    ( const integer& n, const sComplex& x, const integer& incx );

    dReal
    dznrm2
    ( const integer& n, const dComplex& x, const integer& incx );
    */

   // ------------------------------- level 1: asum <- ||re(x)||_1 + ||im(x)||_1
   // --------------------------------------------------------------------------

    sReal
    sasum
    ( const integer& n, const sReal* x, const integer& incx );

    dReal
    dasum
    ( const integer& n, const dReal* x, const integer& incx );

    /*
    sReal
    scasum
    ( const integer& n, const sComplex& x, const integer& incx );

    dReal
    dzasum
    ( const integer& n, const dComplex& x, const integer& incx );
    */

   // --------------------------------------- level 1: amax <- first k such that
   // ------- |re( x(k) )| + |im( x(k) )| = max_i( |re( x(i) )| + |im( x(i) )| )
   // --------------------------------------------------------------------------

    integer
    isamax
    ( const integer& n, const sReal* x, const integer& incx );

    integer
    idamax
    ( const integer& n, const dReal* x, const integer& incx );

    /*
    integer
    icamax
    ( const integer& n, const sComplex& x, const integer& incx );

    integer
    izamax
    ( const integer& n, const dComplex& x, const integer& incx );
    */

   // ----------------------------- level 2: y <- alpha * op( A ) * x + beta * y
   // ------------------------------------- A : m X n, op( A ) := A, A^T, or A^H
   // --------------------------------------------------------------------------

    void
    sgemv
    ( const character& trans, const integer& m, const integer& n,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      const sReal* x, const integer& incx,
      const sReal& beta,
      sReal* y, const integer& incy );

    void
    dgemv
    ( const character& trans, const integer& m, const integer& n,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      const dReal* x, const integer& incx,
      const dReal& beta,
      dReal* y, const integer& incy );

    /*
    void
    cgemv
    ( const character& trans,
      const integer& m, const integer& n,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& x, const integer& incx,
      const sComplex& beta,
      sComplex& y, const integer& incy );

    void
    zgemv
    ( const character& trans,
      const integer& m, const integer& n,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& x, const integer& incx,
      const dComplex& beta,
      dComplex& y, const integer& incy );
    */

   // ----------------------------- level 2: y <- alpha * op( A ) * x + beta * y
   // ------------------------------------- A : m X n, op( A ) := A, A^T, or A^H
   // --------------------------------------------------------------------------

    void
    sgbmv
    ( const character& trans,
      const integer& m, const integer& n,
      const integer& kl, const integer& ku,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      const sReal* x, const integer& incx,
      const sReal& beta,
      sReal* y, const integer& incy );

    void
    dgbmv
    ( const character& trans,
      const integer& m, const integer& n,
      const integer& kl, const integer& ku,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      const dReal* x, const integer& incx,
      const dReal& beta,
      dReal* y, const integer& incy );

    /*
    void
    cgbmv
    ( const character& trans,
      const integer& m, const integer& n,
      const integer& kl, const integer& ku,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& x, const integer& incx,
      const sComplex& beta,
      sComplex& y, const integer& incy );

    void
    zgbmv
    ( const character& trans,
      const integer& m, const integer& n,
      const integer& kl, const integer& ku,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& x, const integer& incx,
      const dComplex& beta,
      dComplex& y, const integer& incy );
    */

   // ----------------------------------- level 2: y <- alpha * A * x + beta * y
   // --------------------------------------------------------------------------

    /*
    void
    chemv
    ( const character& uplo,
      const integer& n,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& x, const integer& incx,
      const sComplex& beta,
      sComplex& y, const integer& incy );

    void
    zhemv
    ( const character& uplo,
      const integer& n,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& x, const integer& incx,
      const dComplex& beta,
      dComplex& y, const integer& incy );
    */

   // ----------------------------------- level 2: y <- alpha * A * x + beta * y
   // --------------------------------------------------------------------------

    /*
    void
    chbmv
    ( const character& uplo,
      const integer& n, const integer& k,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& x, const integer& incx,
      const sComplex& beta,
      sComplex& y, const integer& incy );

    void
    zhbmv
    ( const character& uplo,
      const integer& n, const integer& k,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& x, const integer& incx,
      const dComplex& beta,
      dComplex& y, const integer& incy );
    */

   // ----------------------------------- level 2: y <- alpha * A * x + beta * y
   // --------------------------------------------------------------------------

    /*
    void
    chpmv
    ( const character& uplo,
      const integer& n,
      const sComplex& alpha,
      const sComplex& ap,
      const sComplex& x, const integer& incx,
      const sComplex& beta,
      sComplex& y, const integer& incy );

    void
    zhpmv
    ( const character& uplo,
      const integer& n,
      const dComplex& alpha,
      const dComplex& ap,
      const dComplex& x, const integer& incx,
      const dComplex& beta,
      dComplex& y, const integer& incy );
    */

   // ----------------------------------- level 2: y <- alpha * A * x + beta * y
   // --------------------------------------------------------------------------

    void
    ssymv
    ( const character& uplo,
      const integer& n,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      const sReal* x, const integer& incx,
      const sReal& beta,
      sReal* y, const integer& incy );

    void
    dsymv
    ( const character& uplo,
      const integer& n,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      const dReal* x, const integer& incx,
      const dReal& beta,
      dReal* y, const integer& incy );

   // ----------------------------------- level 2: y <- alpha * A * x + beta * y
   // --------------------------------------------------------------------------

    void
    ssbmv
    ( const character& uplo,
      const integer& n, const integer& k,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      const sReal* x, const integer& incx,
      const sReal& beta,
      sReal* y, const integer& incy );

    void
    dsbmv
    ( const character& uplo,
      const integer& n, const integer& k,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      const dReal* x, const integer& incx,
      const dReal& beta,
      dReal* y, const integer& incy );

   // ----------------------------------- level 2: y <- alpha * A * x + beta * y
   // --------------------------------------------------------------------------

    void
    sspmv
    ( const character& uplo,
      const integer& n,
      const sReal& alpha,
      const sReal* ap,
      const sReal* x, const integer& incx,
      const sReal& beta,
      sReal* y, const integer& incy );

    void
    dspmv
    ( const character& uplo,
      const integer& n,
      const dReal& alpha,
      const dReal* ap,
      const dReal* x, const integer& incx,
      const dReal& beta,
      dReal* y, const integer& incy );

   // ------------------------------------------------ level 2: x <- op( A ) * x
   // ------------------------------------------------ op( A ) := A, A^T, or A^H
   // --------------------------------------------------------------------------

    void
    strmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const sReal* a, const integer& lda,
      sReal* x, const integer& incx );

    void
    dtrmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const dReal* a, const integer& lda,
      dReal* x, const integer& incx );

    /*
    void
    ctrmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const sComplex& a, const integer& lda,
      sComplex& x, const integer& incx );

    void
    ztrmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const dComplex& a, const integer& lda,
      dComplex& x, const integer& incx );
    */

   // ------------------------------------------------ level 2: x <- op( A ) * x
   // ------------------------------------------------ op( A ) := A, A^T, or A^H
   // --------------------------------------------------------------------------

    void
    stbmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const integer& k,
      const sReal* a, const integer& lda,
      sReal* x, const integer& incx );

    void
    dtbmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const integer& k,
      const dReal* a, const integer& lda,
      dReal* x, const integer& incx );

    /*
    void
    ctbmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const integer& k,
      const sComplex& a, const integer& lda,
      sComplex& x, const integer& incx );

    void
    ztbmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const integer& k,
      const dComplex& a, const integer& lda,
      dComplex& x, const integer& incx );
    */

   // ------------------------------------------------ level 2: x <- op( A ) * x
   // ------------------------------------------------ op( A ) := A, A^T, or A^H
   // --------------------------------------------------------------------------

    void
    stpmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const sReal* ap,
      sReal* x, const integer& incx );

    void
    dtpmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const dReal* ap,
      dReal* x, const integer& incx );

    /*
    void
    ctpmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const sComplex& ap,
      sComplex& x, const integer& incx );

    void
    ztpmv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const dComplex& ap,
      dComplex& x, const integer& incx );
    */

   // ------------------------------------------- level 2: solve x = op( A ) * b
   // ------------------------------------- op( A ) := A^(-1), A^(-T), or A^(-H)
   // --------------------------------------------------------------------------

    void
    strsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const sReal* a, const integer& lda,
      sReal* x, const integer& incx );

    void
    dtrsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const dReal* a, const integer& lda,
      dReal* x, const integer& incx );

    /*
    void
    ctrsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const sComplex& a, const integer& lda,
      sComplex& x, const integer& incx );

    void
    ztrsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const dComplex& a, const integer& lda,
      dComplex& x, const integer& incx );
    */

   // ------------------------------------------- level 2: solve x = op( A ) * b
   // ------------------------------------- op( A ) := A^(-1), A^(-T), or A^(-H)
   // --------------------------------------------------------------------------

    void
    stbsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const integer& k,
      const sReal* a, const integer& lda,
      sReal* x, const integer& incx );

    void
    dtbsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const integer& k,
      const dReal* a, const integer& lda,
      dReal* x, const integer& incx );

    /*
    void
    ctbsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const integer& k,
      const sComplex& a, const integer& lda,
      sComplex& x, const integer& incx );

    void
    ztbsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n, const integer& k,
      const dComplex& a, const integer& lda,
      dComplex& x, const integer& incx );
    */

   // ------------------------------------------- level 2: solve x = op( A ) * b
   // ------------------------------------- op( A ) := A^(-1), A^(-T), or A^(-H)
   // --------------------------------------------------------------------------

    void
    stpsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const sReal* ap,
      sReal* x, const integer& incx );

    void
    dtpsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const dReal* ap,
      dReal* x, const integer& incx );

    /*
    void
    ctpsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const sComplex& ap,
      sComplex& x, const integer& incx );

    void
    ztpsv
    ( const character& uplo, const character& trans, const character& diag,
      const integer& n,
      const dComplex& ap,
      dComplex& x, const integer& incx );
    */

   // ---------------------------------------- level 2: A <- alpha * x * y^T + A
   // ---------------------------------------------------------------- A : m X n
   // --------------------------------------------------------------------------

    void
    sger
    ( const integer& m, const integer& n,
      const sReal& alpha,
      const sReal* x, const integer& incx,
      const sReal* y, const integer& incy,
      sReal* a, const integer& lda );

    void
    dger
    ( const integer& m, const integer& n,
      const dReal& alpha,
      const dReal* x, const integer& incx,
      const dReal* y, const integer& incy,
      dReal* a, const integer& lda );

   // ---------------------------------------- level 2: A <- alpha * x * y^T + A
   // ---------------------------------------------------------------- A : m X n
   // --------------------------------------------------------------------------

    /*
    void
    cgeru
    ( const integer& m, const integer& n,
      const sComplex& alpha,
      const sComplex& x, const integer& incx,
      const sComplex& y, const integer& incy,
      sComplex& a, const integer& lda );

    void
    zgeru
    ( const integer& m, const integer& n,
      const dComplex& alpha,
      const dComplex& x, const integer& incx,
      const dComplex& y, const integer& incy,
      dComplex& a, const integer& lda );
    */

   // ---------------------------------------- level 2: A <- alpha * x * y^H + A
   // ---------------------------------------------------------------- A : m X n
   // --------------------------------------------------------------------------

    /*
    void
    cgerc
    ( const integer& m, const integer& n,
      const sComplex& alpha,
      const sComplex& x, const integer& incx,
      const sComplex& y, const integer& incy,
      sComplex& a, const integer& lda );

    void
    zgerc
    ( const integer& m, const integer& n,
      const dComplex& alpha,
      const dComplex& x, const integer& incx,
      const dComplex& y, const integer& incy,
      dComplex& a, const integer& lda );
    */

   // ---------------------------------------- level 2: A <- alpha * x * x^H + A
   // --------------------------------------------------------------------------

    /*
    void
    cher
    ( const character& uplo,
      const integer& n,
      const sComplex& alpha,
      const sComplex& x, const integer& incx,
      sComplex& a, const integer& lda );

    void
    zher
    ( const character& uplo,
      const integer& n,
      const dComplex& alpha,
      const dComplex& x, const integer& incx,
      dComplex& a, const integer& lda );
    */

   // ---------------------------------------- level 2: A <- alpha * x * x^H + A
   // --------------------------------------------------------------------------

    /*
    void
    chpr
    ( const character& uplo,
      const integer& n,
      const sComplex& alpha,
      const sComplex& x, const integer& incx,
      sComplex& ap );

    void
    zhpr
    ( const character& uplo,
      const integer& n,
      const dComplex& alpha,
      const dComplex& x, const integer& incx,
      dComplex& ap );
    */

   // -------------------- level 2: A <- alpha * x * y^H + y * (alpha * x)^H + A
   // --------------------------------------------------------------------------

    /*
    void
    cher2
    ( const character& uplo,
      const integer& n,
      const sComplex& alpha,
      const sComplex& x, const integer& incx,
      const sComplex& y, const integer& incy,
      sComplex& a, const integer& lda );

    void
    zher2
    ( const character& uplo,
      const integer& n,
      const dComplex& alpha,
      const dComplex& x, const integer& incx,
      const dComplex& y, const integer& incy,
      dComplex& a, const integer& lda );
    */

   // -------------------- level 2: A <- alpha * x * y^H + y * (alpha * x)^H + A
   // --------------------------------------------------------------------------

    /*
    void
    chpr2
    ( const character& uplo,
      const integer& n,
      const sComplex& alpha,
      const sComplex& x, const integer& incx,
      const sComplex& y, const integer& incy,
      sComplex& ap );

    void
    zhpr2
    ( const character& uplo,
      const integer& n,
      const dComplex& alpha,
      const dComplex& x, const integer& incx,
      const dComplex& y, const integer& incy,
      dComplex& ap );
    */

   // ---------------------------------------- level 2: A <- alpha * x * x^T + A
   // --------------------------------------------------------------------------

    void
    ssyr
    ( const character& uplo,
      const integer& n,
      const sReal& alpha,
      const sReal* x, const integer& incx,
      sReal* a, const integer& lda );

    void
    dsyr
    ( const character& uplo,
      const integer& n,
      const dReal& alpha,
      const dReal* x, const integer& incx,
      dReal* a, const integer& lda );

   // ---------------------------------------- level 2: A <- alpha * x * x^T + A
   // --------------------------------------------------------------------------

    void
    sspr
    ( const character& uplo,
      const integer& n,
      const sReal& alpha,
      const sReal* x, const integer& incx,
      sReal* ap );

    void
    dspr
    ( const character& uplo,
      const integer& n,
      const dReal& alpha,
      const dReal* x, const integer& incx,
      dReal* ap );

   // ---------------------- level 2: A <- alpha * x * y^T + alpha * y * x^T + A
   // --------------------------------------------------------------------------

    void
    ssyr2
    ( const character& uplo,
      const integer& n,
      const sReal& alpha,
      const sReal* x, const integer& incx,
      const sReal* y, const integer& incy,
      sReal* a, const integer& lda );

    void
    dsyr2
    ( const character& uplo,
      const integer& n,
      const dReal& alpha,
      const dReal& x, const integer& incx,
      const dReal& y, const integer& incy,
      dReal& a, const integer& lda );

   // ---------------------- level 2: A <- alpha * x * y^T + alpha * y * x^T + A
   // --------------------------------------------------------------------------

    void
    sspr2
    ( const character& uplo,
      const integer& n,
      const sReal& alpha,
      const sReal* x, const integer& incx,
      const sReal* y, const integer& incy,
      sReal* ap );

    void
    dspr2
    ( const character& uplo,
      const integer& n,
      const dReal& alpha,
      const dReal* x, const integer& incx,
      const dReal* y, const integer& incy,
      dReal* ap );


   // ----------------------- level 3: C <- alpha * op( A ) * op( B ) + beta * C
   // ------------------------------------- C : m X n, op( X ) := X, X^T, or X^H
   // --------------------------------------------------------------------------

    void
    sgemm
    ( const character& transa, const character& transb,
      const integer& m, const integer& n, const integer& k,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      const sReal* b, const integer& ldb,
      const sReal& beta,
      sReal* c, const integer& tdc );

    void
    dgemm
    ( const character& transa, const character& transb,
      const integer& m, const integer& n, const integer& k,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      const dReal* b, const integer& ldb,
      const dReal& beta,
      dReal* c, const integer& tdc );

    /*
    void
    cgemm
    ( const character& transa, const character& transb,
      const integer& m, const integer& n, const integer& k,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& b, const integer& ldb,
      const sComplex& beta,
      sComplex& c, const integer& tdc );

    void
    zgemm
    ( const character& transa, const character& transb,
      const integer& m, const integer& n, const integer& k,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& b, const integer& ldb,
      const dComplex& beta,
      dComplex& c, const integer& tdc );
    */

   // ----------------------------------- level 3: C <- alpha * A * B + beta * C
   // -------------------------------------------- C <- alpha * B * A + beta * C
   // ------------------------------------------------------ A = A^T, C is m X n
   // --------------------------------------------------------------------------

    void
    ssymm
    ( const character& side, const character& uplo,
      const integer& m, const integer& n,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      const sReal* b, const integer& ldb,
      const sReal& beta,
      sReal* c, const integer& tdc );

    void
    dsymm
    ( const character& side, const character& uplo,
      const integer& m, const integer& n,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      const dReal* b, const integer& ldb,
      const dReal& beta,
      dReal* c, const integer& tdc );

    /*
    void
    csymm
    ( const character& side, const character& uplo,
      const integer& m, const integer& n,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& b, const integer& ldb,
      const sComplex& beta,
      sComplex& c, const integer& tdc );

    void
    zsymm
    ( const character& side, const character& uplo,
      const integer& m, const integer& n,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& b, const integer& ldb,
      const dComplex& beta,
      dComplex& c, const integer& tdc );
    */

   // ----------------------------------- level 3: C <- alpha * A * B + beta * C
   // -------------------------------------------- C <- alpha * B * A + beta * C
   // ------------------------------------------------------ A = A^H, C is m X n
   // --------------------------------------------------------------------------

    /*
    void
    chemm
    ( const character& side, const character& uplo,
      const integer& m, const integer& n,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& b, const integer& ldb,
      const sComplex& beta,
      sComplex& c, const integer& tdc );

    void
    zhemm
    ( const character& side, const character& uplo,
      const integer& m, const integer& n,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& b, const integer& ldb,
      const dComplex& beta,
      dComplex& c, const integer& tdc );
    */

   // --------------------------------- level 3: C <- alpha * A * A^T + beta * C
   // ------------------------------------------ C <- alpha * A^T * A + beta * C
   // ---------------------------------------------------------------- C : n X n
   // --------------------------------------------------------------------------

    void
    ssyrk
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      const sReal& beta,
      sReal* c, const integer& tdc );

    void
    dsyrk
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      const dReal& beta,
      dReal* c, const integer& tdc );

    /*
    void
    csyrk
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& beta,
      sComplex& c, const integer& tdc );

    void
    zsyrk
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& beta,
      dComplex& c, const integer& tdc );
    */

   // --------------------------------- level 3: C <- alpha * A * A^H + beta * C
   // ------------------------------------------ C <- alpha * A^H * A + beta * C
   // ---------------------------------------------------------------- C : n X n
   // --------------------------------------------------------------------------

    /*
    void
    cherk
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& beta,
      sComplex& c, const integer& tdc );

    void
    zherk
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& beta,
      dComplex& c, const integer& tdc );
    */

   // ------------- level 3: C <- alpha * A * B^T + alpha^* * B * A^T + beta * C
   // ---------------------- C <- alpha * A^T * B + alpha^* * B^T * A + beta * C
   // ---------------------------------------------------------------- C : n X n
   // --------------------------------------------------------------------------

    void
    ssyr2k
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      const sReal* b, const integer& ldb,
      const sReal& beta,
      sReal* c, const integer& tdc );

    void
    dsyr2k
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      const dReal* b, const integer& ldb,
      const dReal& beta,
      dReal* c,  const integer& tdc );

    /*
    void
    csyr2k
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& b, const integer& ldb,
      const sComplex& beta,
      sComplex& c, const integer& tdc );

    void
    zsyr2k
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& b, const integer& ldb,
      const dComplex& beta,
      dComplex& c, const integer& tdc );
    */

   // ------------- level 3: C <- alpha * A * B^H + alpha^* * B * A^H + beta * C
   // ---------------------- C <- alpha * A^H * A + alpha^* * B^H * A + beta * C
   // ---------------------------------------------------------------- C : n X n
   // --------------------------------------------------------------------------

    /*
    void
    cher2k
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      const sComplex& b, const integer& ldb,
      const sComplex& beta,
      sComplex& c, const integer& tdc );

    void
    zher2k
    ( const character& uplo, const character& trans,
      const integer& n, const integer& k,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      const dComplex& b, const integer& ldb,
      const dComplex& beta,
      dComplex& c, const integer& tdc );
    */

   // ---------------------------------------- level 3: B <- alpha * op( A ) * B
   // ------------------------------------------------- B <- alpha * B * op( A )
   // -------------------------------------- B : m X n, op( A ) = A, A^T, or A^H
   // --------------------------------------------------------------------------

    void
    strmm
    ( const character& side, const character& uplo,
      const character& transa, const character& diag,
      const integer& m, const integer& n,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      sReal* b, const integer& ldb );

    void
    dtrmm
    ( const character& side, const character& uplo,
      const character& transa, const character& diag,
      const integer& m, const integer& n,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      dReal* b, const integer& ldb );

    /*
    void
    ctrmm
    ( const character& side, const character& uplo,
      const character& transa, const character& diag,
      const integer& m, const integer& n,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      sComplex& b, const integer& ldb );

    void
    ztrmm
    ( const character& side, const character& uplo,
      const character& transa, const character& diag,
      const integer& m, const integer& n,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      dComplex& b, const integer& ldb );
    */

   // ----------------------------------- level 3: B <- alpha * op( A^(-1) ) * B
   // -------------------------------------------- B <- alpha * B * op( A^(-1) )
   // -------------------------------------- B : m X n, op( A ) = A, A^T, or A^H
   // --------------------------------------------------------------------------

    void
    strsm
    ( const character& side, const character& uplo,
      const character& transa, const character& diag,
      const integer& m, const integer& n,
      const sReal& alpha,
      const sReal* a, const integer& lda,
      sReal* b, const integer& ldb );

    void
    dtrsm
    ( const character& side, const character& uplo,
      const character& transa, const character& diag,
      const integer& m, const integer& n,
      const dReal& alpha,
      const dReal* a, const integer& lda,
      dReal* b, const integer& ldb );

    /*
    void
    ctrsm
    ( const char side, const character& uplo,
      const character& transa, const character& diag,
      const integer& m, const integer& n,
      const sComplex& alpha,
      const sComplex& a, const integer& lda,
      sComplex& b, const integer& ldb );

    void
    ztrsm
    ( const character& side, const character& uplo,
      const character& transa, const character& diag,
      const integer& m, const integer& n,
      const dComplex& alpha,
      const dComplex& a, const integer& lda,
      dComplex& b, const integer& ldb );
    */
   } /* extern "C" {} */

cpp_gnamespace_end

#endif /* __cppblas_H__ */
