/* --------------------------------------------------------------------------
 *
 * namespace LapackDrivers
 *   miscellaneous black-box lapack drivers for common tasks
 *
 * required headers
 *   see include header file LapackDrivers
 *
 * notes
 *   1 these functions are intended to be convenient interfaces to blas and
 *     lapack functions, especially with respect to any type conversions and
 *     for use with matrix and varray types.
 *   2 contains function templates; may not be supported by all compilers
 *     (e. g., SUNWspro 5.0).
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __LapackDrivers_H__
#define __LapackDrivers_H__

cpp_namespace_begin( BlasDrivers )
 cpp_using_namespace( std );

#ifndef cpp_missing_template_methods

 // -------------------------------------------------- copy with type conversion
 // ----------------------------------------------------------------------------

 cpp_namespace_decl
 template< class T1, class T2 >
 inline
 void
 copy( size_t n, const T1* v1, T2* v2, size_t incv1 = 1, size_t incv2 = 1 );
               //
               // n  is the number of components of v1 to be  type-converted and
               // copied to v2. Size of v1 >= n * incv1, size of v2 >= n * incv2

 cpp_namespace_decl
 template< class T1, class T2 >
 inline
 void
 copy( const varray<T1>& v1, varray<T2>& v2, size_t incv1 = 1, size_t incv2 = 1, size_t v2minSize = 0 );

 cpp_namespace_decl
 template< class T1, class T2 >
 inline
 void
 copy( const matrix<T1>& m1, matrix<T2>& m2 );

 // -------------------------------------------------- vector-vector dot product
 // ----------------------------------------------------------------------------

 cpp_namespace_decl
 template< class T1, class T2 >
 inline
 T1
 dot( size_t n, const T1* v1, const T2* v2, size_t incv1 = 1, size_t incv2 = 1 );
                             //
                             // n  is  the number of terms  in  the dot product.
                             // Size of v1 >= n * incv1, size of v2 >= n * incv2

 cpp_namespace_decl
 template< class T1, class T2 >
 inline
 T1
 dot( const varray<T1>& v1, const varray<T2>& v2, size_t incv1 = 1, size_t incv2 = 1 );

 // ------------------------------------------------------------ vector 2-norm^2
 // ----------------------------------------------------------------------------

 cpp_namespace_decl
 template< class T >
 T
 nrm2( size_t n, const T* v, size_t incv = 1 );
          //
          // n: number of terms in the norm^2 computation; size of v >= n * incv

 cpp_namespace_decl
 template< class T >
 T
 nrm2( const varray<T>& v, size_t incv = 1 );

 // -------------------------------------- matrix-vector product: general matrix
 // ----------------------------------------------------------------------------

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 void
 gemv( size_t rows, size_t cols, const MT* a, const VT* x, VT* y,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1,
       FortranTypes::character trans = cppblas::NoTranspose );
                                       //
                                       // y <- alpha * trans( a ) * x + beta * y

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 void
 gemv( const matrix<MT>& a, const varray<VT>& x, varray<VT>& y,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1,
       FortranTypes::character trans = cppblas::NoTranspose );
                                       //
                                       // y <- alpha * trans( a ) * x + beta * y

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 varray<VT>
 gemv( size_t rows, size_t cols, const MT* a, const VT* x, const VT* y,
       FortranTypes::character trans = cppblas::NoTranspose,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1 );
                                            //
                                            // alpha * trans( a ) * x + beta * y

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 varray<VT>
 gemv( const matrix<MT>& a, const varray<VT>& x, const varray<VT>& y = varray<VT>( ),
       FortranTypes::character trans = cppblas::NoTranspose,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1 );
                                            //
                                            // alpha * trans( a ) * x + beta * y

 // ------------------------ real symmetric: packed and unpacked representations
 // ----------------------------------------------------------------------------

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 bool
 pack( const matrix<MT>& rsMatrix, varray<VT>& packed,
       FortranTypes::character uplo = cppblas::UpperTriangle );
                                                 //
                                                 // return value: report success

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 bool
 unpack( const varray<VT>& packed, matrix<MT>& rsMatrix,
         FortranTypes::character uplo = cppblas::UpperTriangle );
                                                 //
                                                 // return value: report success

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 bool
 unpack_diag( const varray<VT>& packed, varray<MT>& diagonal,
              FortranTypes::character uplo = cppblas::UpperTriangle );
                                                 //
                                                 // return value: report success

 cpp_namespace_decl
 inline
 bool
 checkPackedSize( size_t packedSize, size_t& rowSize );
                         //
                         // compute the row size corresponding to a packed size.
                         // return value  indicates validity of the packed size;
                         // if invalid, rowSize is set to 0.

 // ------------------------------- matrix-vector product: real symmetric matrix
 // ----------------------------------------------------------------------------

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 void
 symv( size_t rows, const MT* a, const VT* x, VT* y,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1,
       FortranTypes::character uplo = cppblas::UpperTriangle );
                                                //
                                                // y <- alpha * a * x + beta * y

 cpp_namespace_decl
 template< class MT, class VT >
 void
 symv( const matrix<MT>& a, const varray<VT>& x, varray<VT>& y,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1,
       FortranTypes::character uplo = cppblas::UpperTriangle );
                                                //
                                                // y <- alpha * a * x + beta * y

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 varray<VT>
 symv( size_t rows, const MT* a, const VT* x, const VT* y,
       FortranTypes::character uplo = cppblas::UpperTriangle,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1 );
                                                     //
                                                     // alpha * a * x + beta * y

 cpp_namespace_decl
 template< class MT, class VT >
 varray<VT>
 symv( const matrix<MT>& a, const varray<VT>& x, const varray<VT>& y = varray<VT>( ),
       FortranTypes::character uplo = cppblas::UpperTriangle,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1 );
                                                     //
                                                     // alpha * a * x + beta * y

 // ------------------------ matrix-vector product: real symmetric packed matrix
 // ----------------------------------------------------------------------------

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 void
 spmv( size_t rows, const MT* a, const VT* x, VT* y,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1,
       FortranTypes::character uplo = cppblas::UpperTriangle );
                                                //
                                                // y <- alpha * a * x + beta * y

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 void
 spmv( const varray<MT>& a, const varray<VT>& x, varray<VT>& y,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1,
       FortranTypes::character uplo = cppblas::UpperTriangle );
                                                //
                                                // y <- alpha * a * x + beta * y

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 varray<VT>
 spmv( size_t rows, const MT* a, const VT* x, const VT* y,
       FortranTypes::character uplo = cppblas::UpperTriangle,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1 );
                                                     //
                                                     // alpha * a * x + beta * y

 cpp_namespace_decl
 template< class MT, class VT >
 inline
 varray<VT>
 spmv( const varray<MT>& a, const varray<VT>& x, const varray<VT>& y = varray<VT>( ),
       FortranTypes::character uplo = cppblas::UpperTriangle,
       VT alpha = 1, VT beta = 0, size_t incx = 1, size_t incy = 1 );
                                                     //
                                                     // alpha * a * x + beta * y

#endif /* cpp_missing_template_methods */

cpp_namespace_end

cpp_namespace_begin( LapackDrivers )
 cpp_using_namespace( std );
 cpp_namespace_begin( RealSymmetric )

#ifndef cpp_missing_template_methods

  // ----------------------------------------------- real symmetric: eigenvalues
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  template< class MT, class VT >
  bool
  diagonalize( const matrix<MT>&             rsMatrix,
                     varray<VT>&             eigenvalues,
                     matrix<VT>&             eigenvectors,
                     FortranTypes::integer&  lapackInfo,
                     bool                    valuesOnly = true,
                     FortranTypes::character uplo       = cppblas::UpperTriangle );
                                 //
                                 // return value: report success; lapackInfo may
                                 // contain lapack error code reported by  dsyev

  cpp_namespace_decl
  template< class MT, class VT >
  bool
  diagonalize( const varray<MT>&             packedMatrix,
                     varray<VT>&             eigenvalues,
                     matrix<VT>&             eigenvectors,
                     FortranTypes::integer&  lapackInfo,
                     bool                    valuesOnly = true,
                     FortranTypes::character uplo       = cppblas::UpperTriangle );
                                 //
                                 // return value: report success; lapackInfo may
                                 // contain lapack error code reported by  dspev

  // --------------------------------------------------- real symmetric: inverse
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  template< class MT1, class MT2 >
  bool
  invert( const matrix<MT1>&            rsMatrix,
                matrix<MT2>&            rsInverse,
                FortranTypes::integer&  lapackInfo1,
                FortranTypes::integer&  lapackInfo2,
                FortranTypes::character uplo = cppblas::UpperTriangle );
                  //
                  //                               return value: reports success
                  //
                  // on error:  lapackInfo1, lapackInfo2 may contain error codes
                  // returned by lapack functions dsptrf and dsptri respectively.

  cpp_namespace_decl
  template< class MT1, class MT2 >
  bool
  invert( const varray<MT1>&            packedMatrix,
                varray<MT2>&            packedInverse,
                FortranTypes::integer&  lapackInfo1,
                FortranTypes::integer&  lapackInfo2,
                FortranTypes::character uplo = cppblas::UpperTriangle );
                  //
                  //                               return value: reports success
                  //
                  // on error:  lapackInfo1, lapackInfo2 may contain error codes
                  // returned by lapack functions dsptrf and dsptri respectively

  // ------------------------------------ real symmetric: cholesky decomposition
  // ---------------------------------------------------------------------------

  cpp_namespace_decl
  template< class MT1, class MT2 >
  bool
  cholesky( const matrix<MT1>&            rsMatrix,
                  varray<MT2>&            decomposition,
                  FortranTypes::integer&  lapackInfo,
                  FortranTypes::character uplo = cppblas::UpperTriangle );
                            //
                            // return value reports success. lapackInfo may contain
                            // any error code as reported by  dpotrf( ).   cholesky
                            // decomposition is returned in a packed format.

  cpp_namespace_decl
  template< class MT1, class MT2 >
  bool
  cholesky( const varray<MT1>&            packedMatrix,
                  varray<MT2>&            decomposition,
                  FortranTypes::integer&  lapackInfo,
                  FortranTypes::character uplo = cppblas::UpperTriangle );
                            //
                            // return value reports success. lapackInfo may contain
                            // any error code as reported by  dpotrf( ).   cholesky
                            // decomposition is returned in a packed format.

#endif /* cpp_missing_template_methods */

 cpp_namespace_end
cpp_namespace_end

#ifndef cpp_missing_template_methods

// --------------------------------------------------- copy with type conversion
// -----------------------------------------------------------------------------

template< class T1, class T2 >
void
BlasDrivers::copy( size_t n, const T1* v1, T2* v2, size_t incv1, size_t incv2 )
 {
  for ( size_t i = 0, j = 0, k = 0; i < n; j += incv1, k += incv2 )
   v2[k] = cpp_static_cast( T2, v1[j] );
 }

template< >
inline
void
BlasDrivers::copy<FortranTypes::dReal,FortranTypes::dReal>( size_t n, const FortranTypes::dReal* v1, FortranTypes::dReal* v2, size_t incv1, size_t incv2 )
 {
  cppblas::dcopy( n, v1, incv1, v2, incv2 );
 }

template< >
inline
void
BlasDrivers::copy<FortranTypes::sReal,FortranTypes::sReal>( size_t n, const FortranTypes::sReal* v1, FortranTypes::sReal* v2, size_t incv1, size_t incv2 )
 {
  cppblas::scopy( n, v1, incv1, v2, incv2 );
 }

template< class T1, class T2 >
void
BlasDrivers::copy( const varray<T1>& v1, varray<T2>& v2, size_t incv1, size_t incv2, size_t v2minSize )
 {
  size_t n = v1.size( ) / incv1 + ( ( v1.size( ) % incv1 ) > 0 );
  size_t m = max( v2minSize, ( n - 1 ) * incv2 + 1 );
  if ( v2.size( ) < m ) v2.resize( m, 0, true );
  copy( v1.size( ), v1( ), v2( ), incv1, incv2 );
 }

template< class T1, class T2 >
void
BlasDrivers::copy( const matrix<T1>& m1, matrix<T2>& m2 )
 {
  m2.resize( m1.numRows( ), m1.numCols( ) );
  copy( m1.numRows( ) * m1.numCols( ), m1[0], m2[0], 1, 1 );
 }

// template< class T1 >
// void
// LapackDrivers::copy<T1,T1>( const matrix<T1>& m1, matrix<T1>& m2 )
//  {
//   m2 = m1;
//  }

// -------------------------------------------------- vector-vector dot products
// -----------------------------------------------------------------------------

template< class T1, class T2 >
T1
BlasDrivers::dot( size_t n, const T1* v1, const T2* v2, size_t incv1, size_t incv2 )
 {
  T1 sum = T1( );
  for ( size_t i = 0, j = 0, k = 0; i < n; j += incv1, k += incv2 )
   sum += v1[j] * cpp_static_cast( T1, v2[k] );
  return sum;
 }

template< >
inline
FortranTypes::dReal
BlasDrivers::dot<FortranTypes::dReal,FortranTypes::dReal>( size_t n, const FortranTypes::dReal* v1, const FortranTypes::dReal* v2, size_t incv1, size_t incv2 )
 {
  return cppblas::ddot( n, v1, incv1, v2, incv2 );
 }

template< >
inline
FortranTypes::sReal
BlasDrivers::dot<FortranTypes::sReal,FortranTypes::sReal>( size_t n, const FortranTypes::sReal* v1, const FortranTypes::sReal* v2, size_t incv1, size_t incv2 )
 {
  return cppblas::sdot( n, v1, incv1, v2, incv2 );
 }

template< class T1, class T2 >
T1
BlasDrivers::dot( const varray<T1>& v1, const varray<T2>& v2, size_t incv1, size_t incv2 )
 {
  size_t n = v1.size( ) / incv1 + ( ( v1.size( ) % incv1 ) > 0 );
  size_t m = v2.size( ) / incv2 + ( ( v2.size( ) % incv2 ) > 0 );
  return dot( min( n, m ), v1( ), v2( ), incv1, incv2 );
 }

// ------------------------------------------------------------- vector 2-norm^2
// -----------------------------------------------------------------------------

template< class T >
T
BlasDrivers::nrm2( size_t n, const T* v, size_t incv )
 {
  T sum = T( );
  for ( size_t i = 0, j = 0; i < n; j += incv1 ) sum += v[j] * v[j];
  return cpp_static_cast( T, sqrt( cpp_static_cast( double, sum ) ) );
 }

template< >
inline
FortranTypes::dReal
BlasDrivers::nrm2<FortranTypes::dReal>( size_t n, const FortranTypes::dReal* v, size_t incv )
 {
  return cppblas::dnrm2( n, v, incv );
 }

template< >
inline
FortranTypes::sReal
BlasDrivers::nrm2<FortranTypes::sReal>( size_t n, const FortranTypes::sReal* v, size_t incv )
 {
  return cppblas::snrm2( n, v, incv );
 }

template< class T >
T
BlasDrivers::nrm2( const varray<T>& v, size_t incv )
 {
  return nrm2( v.size( ), v( ), incv );
 }

// --------------------------------------- matrix-vector product: general matrix
// -----------------------------------------------------------------------------

template< class MT, class VT >
void
BlasDrivers::gemv( size_t m, size_t n, const MT* A, const VT* X, VT* Y,
                     VT alpha, VT beta, size_t incX, size_t incY, FortranTypes::character trans )
 {
  size_t k = m * n; FortranTypes::dReal a[k]; copy( k, A, a );
  FortranTypes::dReal x[n]; copy( n, X, x, incX, 1 );
  FortranTypes::dReal y[m]; copy( m, Y, y, incY, 1 );
  cppblas::dgemv( trans, m, n, alpha, a, m, x, 1, beta, y, 1 );
  copy( m, y, Y, 1, incY );
 }

template< >
inline
void
BlasDrivers::gemv<FortranTypes::dReal,FortranTypes::dReal>( size_t m, size_t n,
                                                              const FortranTypes::dReal* a,
                                                              const FortranTypes::dReal* x,
                                                              FortranTypes::dReal* y,
                                                              FortranTypes::dReal alpha,
                                                              FortranTypes::dReal beta,
                                                              size_t incx, size_t incy,
                                                              FortranTypes::character trans )
 {
  cppblas::dgemv( trans, m, n, alpha, a, m, x, incx, beta, y, incy );
 }

template< >
inline
void
BlasDrivers::gemv<FortranTypes::sReal,FortranTypes::sReal>( size_t m, size_t n,
                                                              const FortranTypes::sReal* a,
                                                              const FortranTypes::sReal* x,
                                                              FortranTypes::sReal* y,
                                                              FortranTypes::sReal alpha,
                                                              FortranTypes::sReal beta,
                                                              size_t incx, size_t incy,
                                                              FortranTypes::character trans )
 {
  cppblas::sgemv( trans, m, n, alpha, a, m, x, incx, beta, y, incy );
 }

template< class MT, class VT >
void
BlasDrivers::gemv( const matrix<MT>& A, const varray<VT>& X, varray<VT>& Y,
                     VT alpha, VT beta, size_t incX, size_t incY, FortranTypes::character trans )
 {
  matrix<FortranTypes::dReal> a; copy( A, a );
  FortranTypes::dReal x[a.numCols( )], y[a.numRows( )];
  copy( a.numCols( ), X( ), x, incX, 1 );
  copy( a.numRows( ), Y( ), y, incY, 1 );
  cppblas::dgemv( trans, a.numRows( ), a.numCols( ), cpp_static_cast( FortranTypes::dReal, alpha ),
                  a[0], a.numRows( ), x, 1, cpp_static_cast( FortranTypes::dReal, beta ), y, 1 );
  copy( a.numCols( ), y, Y, 1, incY );
 }

template< >
inline
void
BlasDrivers::gemv<FortranTypes::dReal,FortranTypes::dReal>( const matrix<FortranTypes::dReal>& a,
                                                              const varray<FortranTypes::dReal>& x,
                                                                    varray<FortranTypes::dReal>& y,
                                                                    FortranTypes::dReal alpha,
                                                                    FortranTypes::dReal beta,
                                                                    size_t incx, size_t incy,
                                                                    FortranTypes::character trans )
 {
  const FortranTypes::dReal* xp = x( ); varray<FortranTypes::dReal> X;
  size_t n = ( a.numCols( ) - 1 ) * incx + 1, m = ( a.numRows( ) - 1 ) * incy + 1;
  if ( x.size( ) < n ) { X.resize( n ); X = x; xp = X( ); }
  if ( y.size( ) < m ) y.resize( m, 0, true );
  cppblas::dgemv( trans, a.numRows( ), a.numCols( ), alpha, a[0], a.numRows( ), xp, incx, beta, y( ), incy );
 }

template< >
inline
void
BlasDrivers::gemv<FortranTypes::sReal,FortranTypes::sReal>( const matrix<FortranTypes::sReal>& a,
                                                              const varray<FortranTypes::sReal>& x,
                                                                    varray<FortranTypes::sReal>& y,
                                                                    FortranTypes::sReal alpha,
                                                                    FortranTypes::sReal beta,
                                                                    size_t incx, size_t incy,
                                                                    FortranTypes::character trans )
 {
  const FortranTypes::sReal* xp = x( ); varray<FortranTypes::sReal> X;
  size_t n = ( a.numCols( ) - 1 ) * incx + 1, m = ( a.numRows( ) - 1 ) * incy + 1;
  if ( x.size( ) < n ) { X.resize( n ); X = x; xp = X( ); }
  if ( y.size( ) < m ) y.resize( m, 0, true );
  cppblas::sgemv( trans, a.numRows( ), a.numCols( ), alpha, a[0], a.numRows( ), xp, incx, beta, y( ), incy );
 }

template< class MT, class VT >
varray<VT>
BlasDrivers::gemv( size_t m, size_t n, const MT* A, const VT* X, const VT* Y,
                     FortranTypes::character trans, VT alpha, VT beta, size_t incX, size_t incY )
 {
  varray<VT> y( m ); copy( m, Y, y( ), incY, 1 );
  gemv( m, n, A, X, y( ), alpha, beta, incX, 1 );
  return y;
 }

template< class MT, class VT >
varray<VT>
BlasDrivers::gemv( const matrix<MT>& A, const varray<VT>& X, const varray<VT>& Y,
                     FortranTypes::character trans, VT alpha, VT beta, size_t incX, size_t incY )
 {
  varray<VT> y( Y );
  gemv( A, x, y, alpha, beta, incX, incY, trans );
  return y;
 }

// ------------------------- real symmetric: packed and unpacked representations
// -----------------------------------------------------------------------------

template< class MT, class VT >
bool
BlasDrivers::pack( const matrix<MT>& rsMatrix, varray<VT>& packed, FortranTypes::character uplo )
 {
  if ( rsMatrix.numRows( ) != rsMatrix.numCols( ) ) { packed.resize( 0 ); return false; }

  packed.resize( rsMatrix.numRows( ) * ( rsMatrix.numRows( ) + 1 ) / 2 );

  if ( packed.size( ) == 0 ) return true;

  size_t imin = 0, imax = rsMatrix.numRows( ) - 1;
  bool upper = ( uplo == cppblas::UpperTriangle );

  const MT* cj;

  size_t k = 0;
  for ( size_t j = 0; j < rsMatrix.numCols( ); j++ )
   {
    cj = rsMatrix[j]; if ( upper ) imax = j; else imin = j;

    for ( size_t i = imin; i <= imax; i++ )
     {
      packed[k] = cpp_static_cast( VT, cj[i] );
      k++;
     }
   }

  return true;
 }

template< class MT, class VT >
bool
BlasDrivers::unpack( const varray<VT>& packed, matrix<MT>& rsMatrix, FortranTypes::character uplo )
 {
  size_t n = 0;
  if ( !checkPackedSize( packed.size( ), n ) || ( n == 0 ) ) { rsMatrix.resize( 0, 0 ); return false; }

  size_t imin = 0, imax = n - 1;
  bool upper = ( uplo == cppblas::UpperTriangle );

  rsMatrix.resize( n, n );

  MT* cj;

  size_t k = 0;
  for ( size_t j = 0; j < n; j++ )
   {
    cj = rsMatrix[j]; if ( upper ) imax = j; else imin = j;

    for ( size_t i = imin; i <= imax; i++ )
     {
      rsMatrix( j, i ) = cj[i] = cpp_static_cast( MT, packed[k] );
      k++;
     }
   }

  return true;
 }

template< class MT, class VT >
bool
BlasDrivers::unpack_diag( const varray<VT>& packed, varray<MT>& diagonal, FortranTypes::character uplo )
 {
  size_t n = 0;
  if ( !checkPackedSize( packed.size( ), n ) || ( n == 0 ) ) { diagonal.resize( 0 ); return false; }

  diagonal.resize( n );

  if ( uplo == cppblas::UpperTriangle )
   for ( size_t i = 0, j = 2, m = 0; i < n; i++, m += j, j++ )
    diagonal[i] = cpp_static_cast( MT, packed[m] );
  else
   for ( size_t i = 0, j = n, m = 0; i < n; i++, m += j, j-- )
    diagonal[i] = cpp_static_cast( MT, packed[m] );

  return true;
 }

bool
BlasDrivers::checkPackedSize( size_t packedSize, size_t& n )
 {
  if ( ( packedSize == 0 ) || ( packedSize == 1 ) ) { n = packedSize; return true; }
  n = 0; size_t s = 0, k = 2 * packedSize;
  for ( ; s < k; n++ ) s = n * ( n + 1 ); n--;
  if ( s != k ) { n = 0; return false; }
  return true;
 }

// -------------------------------- matrix-vector product: real symmetric matrix
// -----------------------------------------------------------------------------

template< class MT, class VT >
void
BlasDrivers::symv( size_t m, const MT* A, const VT* X, VT* Y,
                     VT alpha, VT beta, size_t incX, size_t incY, FortranTypes::character uplo )
 {
  size_t k = m * m; FortranTypes::dReal a[k]; copy( k, A, a );
  FortranTypes::dReal x[m]; copy( m, X, x, incX, 1 );
  FortranTypes::dReal y[m]; copy( m, Y, y, incY, 1 );
  cppblas::dsymv( uplo, m, alpha, a, m, x, 1, beta, y, 1 );
  copy( m, y, Y, 1, incY );
 }

template< >
inline
void
BlasDrivers::symv<FortranTypes::dReal,FortranTypes::dReal>( size_t m,
                                                              const FortranTypes::dReal* a,
                                                              const FortranTypes::dReal* x,
                                                              FortranTypes::dReal* y,
                                                              FortranTypes::dReal alpha,
                                                              FortranTypes::dReal beta,
                                                              size_t incx, size_t incy,
                                                              FortranTypes::character uplo )
 {
  cppblas::dsymv( uplo, m, alpha, a, m, x, incx, beta, y, incy );
 }

template< >
inline
void
BlasDrivers::symv<FortranTypes::sReal,FortranTypes::sReal>( size_t m,
                                                              const FortranTypes::sReal* a,
                                                              const FortranTypes::sReal* x,
                                                              FortranTypes::sReal* y,
                                                              FortranTypes::sReal alpha,
                                                              FortranTypes::sReal beta,
                                                              size_t incx, size_t incy,
                                                              FortranTypes::character uplo )
 {
  cppblas::ssymv( uplo, m, alpha, a, m, x, incx, beta, y, incy );
 }

template< class MT, class VT >
void
BlasDrivers::symv( const matrix<MT>& A, const varray<VT>& x, varray<VT>& y,
                     VT alpha, VT beta, size_t incx, size_t incy, FortranTypes::character uplo )
 {
  if ( A.numRows( ) != A.numCols( ) ) { y.resize( 0 ); return; }
  varray<FortranTypes::dReal> a;
  if ( !pack( A, a, uplo ) ) { y.resize( 0 ); return; }
  spmv( a, x, y, alpha, beta, incx, incy, uplo );
 }

template< >
inline
void
BlasDrivers::symv<FortranTypes::dReal,FortranTypes::dReal>( const matrix<FortranTypes::dReal>& a,
                                                              const varray<FortranTypes::dReal>& x,
                                                                    varray<FortranTypes::dReal>& y,
                                                                    FortranTypes::dReal alpha,
                                                                    FortranTypes::dReal beta,
                                                                    size_t incx, size_t incy,
                                                                    FortranTypes::character uplo )
 {
  if ( a.numRows( ) != a.numCols( ) ) { y.resize( 0 ); return; }
  const FortranTypes::dReal* xp = x( ); varray<FortranTypes::dReal> X;
  size_t n = ( a.numCols( ) - 1 ) * incx + 1, m = ( a.numRows( ) - 1 ) * incy + 1;
  if ( x.size( ) < n ) { X.resize( n ); X = x; xp = X( ); }
  if ( y.size( ) < m ) y.resize( m, 0, true );
  cppblas::dsymv( uplo, a.numRows( ), alpha, a[0], a.numRows( ), xp, incx, beta, y( ), incy );
 }

template< >
inline
void
BlasDrivers::symv<FortranTypes::sReal,FortranTypes::sReal>( const matrix<FortranTypes::sReal>& a,
                                                              const varray<FortranTypes::sReal>& x,
                                                                    varray<FortranTypes::sReal>& y,
                                                                    FortranTypes::sReal alpha,
                                                                    FortranTypes::sReal beta,
                                                                    size_t incx, size_t incy,
                                                                    FortranTypes::character uplo )
 {
  if ( a.numRows( ) != a.numCols( ) ) { y.resize( 0 ); return; }
  const FortranTypes::sReal* xp = x( ); varray<FortranTypes::sReal> X;
  size_t n = ( a.numCols( ) - 1 ) * incx + 1, m = ( a.numRows( ) - 1 ) * incy + 1;
  if ( x.size( ) < n ) { X.resize( n ); X = x; xp = X( ); }
  if ( y.size( ) < m ) y.resize( m, 0, true );
  cppblas::ssymv( uplo, a.numRows( ), alpha, a[0], a.numRows( ), xp, incx, beta, y( ), incy );
 }

template< class MT, class VT >
varray<VT>
BlasDrivers::symv( size_t m, const MT* A, const VT* X, const VT* Y,
                     FortranTypes::character uplo, VT alpha, VT beta, size_t incX, size_t incY )
 {
  varray<VT> y( m ); copy( m, Y, y( ), incY, 1 );
  symv( m, A, X, y( ), alpha, beta, incX, 1 );
  return y;
 }

template< class MT, class VT >
varray<VT>
BlasDrivers::symv( const matrix<MT>& A, const varray<VT>& x, const varray<VT>& y,
                     FortranTypes::character uplo, VT alpha, VT beta, size_t incx, size_t incy )
 {
  if ( A.numRows( ) != A.numCols( ) ) return varray<VT>( );
  varray<FortranTypes::dReal> a;
  if ( !pack( A, a, uplo ) ) return varray<VT>( );
  return spmv( a, x, y, uplo, alpha, beta, incx, incy );
 }

// ------------------------- matrix-vector product: real symmetric packed matrix
// -----------------------------------------------------------------------------

template< class MT, class VT >
void
BlasDrivers::spmv( size_t m, const MT* A, const VT* X, VT* Y,
                   VT alpha, VT beta, size_t incX, size_t incY, FortranTypes::character uplo )
 {
  size_t k = m * ( m + 1 ) / 2; FortranTypes::dReal a[k]; copy( k, A, a );
  FortranTypes::dReal x[m]; copy( m, X, x, incX, 1 );
  FortranTypes::dReal y[m]; copy( m, Y, y, incY, 1 );
  cppblas::dspmv( uplo, m, alpha, a, x, 1, beta, y, 1 );
  copy( m, y, Y, 1, incY );
 }

template< >
inline
void
BlasDrivers::spmv<FortranTypes::dReal,FortranTypes::dReal>( size_t m,
                                                            const FortranTypes::dReal* a,
                                                            const FortranTypes::dReal* x,
                                                            FortranTypes::dReal* y,
                                                            FortranTypes::dReal alpha,
                                                            FortranTypes::dReal beta,
                                                            size_t incx, size_t incy,
                                                            FortranTypes::character uplo )
 {
  cppblas::dspmv( uplo, m, alpha, a, x, incx, beta, y, incy );
 }

template< >
inline
void
BlasDrivers::spmv<FortranTypes::sReal,FortranTypes::sReal>( size_t m,
                                                            const FortranTypes::sReal* a,
                                                            const FortranTypes::sReal* x,
                                                            FortranTypes::sReal* y,
                                                            FortranTypes::sReal alpha,
                                                            FortranTypes::sReal beta,
                                                            size_t incx, size_t incy,
                                                            FortranTypes::character uplo )
 {
  cppblas::sspmv( uplo, m, alpha, a, x, incx, beta, y, incy );
 }

template< class MT, class VT >
void
BlasDrivers::spmv( const varray<MT>& A, const varray<VT>& X, varray<VT>& Y,
                     VT alpha, VT beta, size_t incX, size_t incY, FortranTypes::character uplo )
 {
  size_t n = 0;
  if ( !checkPackedSize( A.size( ), n ) || ( n == 0 ) ) { Y.resize( 0 ); return; }

  varray<FortranTypes::dReal> a, x, y;
  copy( A, a ); copy( X, x, incX, 1, n ); copy( Y, y, incY, 1, n );

  cppblas::dspmv( uplo, n, cpp_static_cast( FortranTypes::dReal, alpha ),
                  a( ), x( ), 1, cpp_static_cast( FortranTypes::dReal, beta ), y( ), 1 );

  copy( y, Y );
 }

template< >
inline
void
BlasDrivers::spmv<FortranTypes::dReal,FortranTypes::dReal>( const varray<FortranTypes::dReal>& a,
                                                            const varray<FortranTypes::dReal>& x,
                                                                  varray<FortranTypes::dReal>& y,
                                                                  FortranTypes::dReal alpha,
                                                                  FortranTypes::dReal beta,
                                                                  size_t incx, size_t incy,
                                                                  FortranTypes::character uplo )
 {
  size_t n = 0;
  if ( !checkPackedSize( a.size( ), n ) || ( n == 0 ) ) { y.resize( 0 ); return; }

  const FortranTypes::dReal* xp = x( ); varray<FortranTypes::dReal> X;
  size_t m = ( n - 1 ) * incx + 1;
  if ( x.size( ) < n ) { X.resize( n ); X = x; xp = X( ); }
  m = ( n - 1 ) * incy + 1;
  if ( y.size( ) < m ) y.resize( m, 0, true );

  cppblas::dspmv( uplo, n, alpha, a( ), xp, incx, beta, y( ), incy );
 }

template< >
inline
void
BlasDrivers::spmv<FortranTypes::sReal,FortranTypes::sReal>( const varray<FortranTypes::sReal>& a,
                                                            const varray<FortranTypes::sReal>& x,
                                                                  varray<FortranTypes::sReal>& y,
                                                                  FortranTypes::sReal alpha,
                                                                  FortranTypes::sReal beta,
                                                                  size_t incx, size_t incy,
                                                                  FortranTypes::character uplo )
 {
  size_t n = 0;
  if ( !checkPackedSize( a.size( ), n ) || ( n == 0 ) ) { y.resize( 0 ); return; }

  const FortranTypes::sReal* xp = x( ); varray<FortranTypes::sReal> X;
  size_t m = ( n - 1 ) * incx + 1;
  if ( x.size( ) < n ) { X.resize( n ); X = x; xp = X( ); }
  m = ( n - 1 ) * incy + 1;
  if ( y.size( ) < m ) y.resize( m, 0, true );

  cppblas::sspmv( uplo, n, alpha, a( ), xp, incx, beta, y( ), incy );
 }

template< class MT, class VT >
varray<VT>
BlasDrivers::spmv( size_t m, const MT* A, const VT* X, const VT* Y,
                     FortranTypes::character uplo, VT alpha, VT beta, size_t incX, size_t incY )
 {
  varray<VT> y( m ); copy( m, Y, y( ), incY, 1 );
  spmv( m, A, X, y( ), alpha, beta, incX, 1 );
  return y;
 }

template< class MT, class VT >
varray<VT>
BlasDrivers::spmv( const varray<MT>& A, const varray<VT>& X, const varray<VT>& Y,
                     FortranTypes::character uplo, VT alpha, VT beta, size_t incX, size_t incY )
 {
  varray<VT> y( Y );
  spmv( A, X, y, alpha, beta, incX, incY, uplo );
  return y;
 }

// ------------------------------------------------- real symmetric: eigenvalues
// -----------------------------------------------------------------------------

template< class MT, class VT >
bool
LapackDrivers::RealSymmetric::diagonalize( const matrix<MT>&             rsMatrix,
                                                 varray<VT>&             eigenvalues,
                                                 matrix<VT>&             eigenvectors,
                                                 FortranTypes::integer&  lapackInfo,
                                                 bool                    valuesOnly,
                                                 FortranTypes::character uplo )
 {
  if ( rsMatrix.numRows( ) != rsMatrix.numCols( ) )
   {
    eigenvalues.resize( 0 ); eigenvectors.resize( 0, 0 ); lapackInfo = 0; return false;
   }

  varray<FortranTypes::dReal> a; BlasDrivers::pack( rsMatrix, a );
  return diagonalize( a, eigenvalues, eigenvectors, lapackInfo, valuesOnly, uplo );
 }

template< class MT, class VT >
bool
LapackDrivers::RealSymmetric::diagonalize( const varray<MT>&             packedMatrix,
                                                 varray<VT>&             eigenvalues,
                                                 matrix<VT>&             eigenvectors,
                                                 FortranTypes::integer&  lapackInfo,
                                                 bool                    valuesOnly,
                                                 FortranTypes::character uplo )
 {
  eigenvalues.resize( 0 ); eigenvectors.resize( 0, 0 ); lapackInfo = 0;

  size_t n = 0;
  if ( !BlasDrivers::checkPackedSize( packedMatrix.size( ), n ) || ( n == 0 ) ) return false;

  varray<FortranTypes::dReal> ap;
  copy( packedMatrix, ap );

  varray<FortranTypes::dReal> ev( n );

  FortranTypes::dReal work[3 * n - 1];

  if ( valuesOnly )
   {
    FortranTypes::dReal* z = 0;
    cpplapack::dspev( cpplapack::jobz_0, uplo, n, ap( ), ev( ), z, n, work, lapackInfo );
    if ( lapackInfo ) return false;
    copy( ev, eigenvalues );
    return true;
   }
  else
   {
    FortranTypes::dReal z[n*n];
    cpplapack::dspev( cpplapack::jobz_1, uplo, n, ap( ), ev( ), z, n, work, lapackInfo );
    if ( lapackInfo ) return false;
    copy( ev, eigenvalues );
    copy( matrix<FortranTypes::dReal>( n, n, z, false ), eigenvectors );
    return true;
   }
 }

template< >
inline
bool
LapackDrivers::RealSymmetric::diagonalize<FortranTypes::dReal,FortranTypes::dReal>( const varray<FortranTypes::dReal>& packedMatrix,
                                                 varray<FortranTypes::dReal>& eigenvalues,
                                                 matrix<FortranTypes::dReal>& eigenvectors,
                                                 FortranTypes::integer& lapackInfo,
                                                 bool valuesOnly, FortranTypes::character uplo )
 {
  eigenvalues.resize( 0 ); eigenvectors.resize( 0, 0 ); lapackInfo = 0;

  size_t n = 0;
  if ( !BlasDrivers::checkPackedSize( packedMatrix.size( ), n ) || ( n == 0 ) ) return false;

  varray<FortranTypes::dReal> ap( packedMatrix( ), packedMatrix.size( ) );

  eigenvalues.resize( n );
  FortranTypes::dReal work[3 * n - 1];

  if ( valuesOnly )
   {
    cpplapack::dspev( cpplapack::jobz_0, uplo, n, ap( ), eigenvalues( ), 0, n, work, lapackInfo );
    if ( lapackInfo ) { eigenvalues.resize( 0 ); return false; }
    return true;
   }
  else
   {
    eigenvectors.resize( n, n );
    cpplapack::dspev( cpplapack::jobz_1, uplo, n, ap( ), eigenvalues( ), eigenvectors[0], n, work, lapackInfo );
    if ( lapackInfo ) { eigenvalues.resize( 0 ); eigenvectors.resize( 0, 0 ); return false; }
    return true;
   }
 }

template< >
inline
bool
LapackDrivers::RealSymmetric::diagonalize<FortranTypes::sReal,FortranTypes::sReal>( const varray<FortranTypes::sReal>& packedMatrix,
                                                 varray<FortranTypes::sReal>& eigenvalues,
                                                 matrix<FortranTypes::sReal>& eigenvectors,
                                                 FortranTypes::integer& lapackInfo,
                                                 bool valuesOnly, FortranTypes::character uplo )
 {
  eigenvalues.resize( 0 ); eigenvectors.resize( 0, 0 ); lapackInfo = 0;

  size_t n = 0;
  if ( !BlasDrivers::checkPackedSize( packedMatrix.size( ), n ) || ( n == 0 ) ) return false;

  varray<FortranTypes::sReal> ap( packedMatrix( ), packedMatrix.size( ) );

  eigenvalues.resize( n );
  FortranTypes::sReal work[3 * n - 1];

  if ( valuesOnly )
   {
    cpplapack::sspev( cpplapack::jobz_0, uplo, n, ap( ), eigenvalues( ), 0, n, work, lapackInfo );
    if ( lapackInfo ) { eigenvalues.resize( 0 ); return false; }
    return true;
   }
  else
   {
    eigenvectors.resize( n, n );
    cpplapack::sspev( cpplapack::jobz_1, uplo, n, ap( ), eigenvalues( ), eigenvectors[0], n, work, lapackInfo );
    if ( lapackInfo ) { eigenvalues.resize( 0 ); eigenvectors.resize( 0, 0 ); return false; }
    return true;
   }
 }

// ----------------------------------------------------- real symmetric: inverse
// -----------------------------------------------------------------------------

template< class MT1, class MT2 >
bool
LapackDrivers::RealSymmetric::invert( const matrix<MT1>&            rsMatrix,
                                            matrix<MT2>&            rsInverse,
                                            FortranTypes::integer&  lapackInfo1,
                                            FortranTypes::integer&  lapackInfo2,
                                            FortranTypes::character uplo )
 {
  rsInverse.resize( 0, 0 ); lapackInfo1 = lapackInfo2 = 0;
  if ( rsMatrix.numRows( ) != rsMatrix.numCols( ) ) return false;
  varray<FortranTypes::dReal> a, ai; BlasDrivers::pack( rsMatrix, a, uplo );
  if ( invert( a, ai, lapackInfo1, lapackInfo2, uplo ) ) { BlasDrivers::unpack( ai, rsInverse, uplo ); return true; }
  return false;
 }

template< class MT1, class MT2 >
bool
LapackDrivers::RealSymmetric::invert( const varray<MT1>&            packedMatrix,
                                            varray<MT2>&            packedInverse,
                                            FortranTypes::integer&  lapackInfo1,
                                            FortranTypes::integer&  lapackInfo2,
                                            FortranTypes::character uplo )
 {
  packedInverse.resize( 0 ); lapackInfo1 = lapackInfo2 = 0;
  size_t n = 0;
  if ( !BlasDrivers::checkPackedSize( packedMatrix.size( ), n ) || ( n == 0 ) ) return false;
  varray<FortranTypes::dReal> a, ai; copy( packedMatrix, a );
  if ( invert( a, ai, lapackInfo1, lapackInfo2, uplo ) ) { BlasDrivers::unpack( ai, rsInverse, uplo ); return true; }
  return false;
 }

template< >
inline
bool
LapackDrivers::RealSymmetric::invert<FortranTypes::dReal,FortranTypes::dReal>( const varray<FortranTypes::dReal>& packedMatrix,
                                                                 varray<FortranTypes::dReal>& packedInverse,
                                                                 FortranTypes::integer& lapackInfo1,
                                                                 FortranTypes::integer& lapackInfo2,
                                                                 FortranTypes::character uplo )
 {
  packedInverse.resize( 0 ); lapackInfo1 = lapackInfo2 = 0;
  size_t n = 0;

  /*
  cerr << "(debug) LapackDrivers::invert<double,double>( ) packedMatrix.size( ) " << packedMatrix.size( )
       << " check " << BlasDrivers::checkPackedSize( packedMatrix.size( ), n ) << " n " << n << endl;
  */

  if ( !BlasDrivers::checkPackedSize( packedMatrix.size( ), n ) || ( n == 0 ) ) return false;
  packedInverse.resize( packedMatrix.size( ) );
  packedInverse = packedMatrix;
  FortranTypes::integer ipiv[n];
  cpplapack::dsptrf( uplo, n, packedInverse( ), ipiv, lapackInfo1 );
  // cerr << "(debug) LapackDrivers::invert<double,double>( ) lapackInfo1 " << lapackInfo1 << endl;
  if ( lapackInfo1 ) { packedInverse.resize( 0 ); return false; }
  FortranTypes::dReal work[n];
  cpplapack::dsptri( uplo, n, packedInverse( ), ipiv, work, lapackInfo2 );
  // cerr << "(debug) LapackDrivers::invert<double,double>( ) lapackInfo2 " << lapackInfo2 << endl;
  return ( lapackInfo2 ? false : true );
 }

template< >
inline
bool
LapackDrivers::RealSymmetric::invert<FortranTypes::sReal,FortranTypes::sReal>( const varray<FortranTypes::sReal>&  packedMatrix,
                                                                 varray<FortranTypes::sReal>& packedInverse,
                                                                 FortranTypes::integer& lapackInfo1,
                                                                 FortranTypes::integer& lapackInfo2,
                                                                 FortranTypes::character uplo )
 {
  packedInverse.resize( 0 ); lapackInfo1 = lapackInfo2 = 0;
  size_t n = 0;
  if ( !BlasDrivers::checkPackedSize( packedMatrix.size( ), n ) || ( n == 0 ) ) return false;
  packedInverse.resize( packedMatrix.size( ) );
  packedInverse = packedMatrix;
  FortranTypes::integer ipiv[n];
  cpplapack::ssptrf( uplo, n, packedInverse( ), ipiv, lapackInfo1 );
  if ( lapackInfo1 ) { packedInverse.resize( 0 ); return false; }
  FortranTypes::sReal work[n];
  cpplapack::ssptri( uplo, n, packedInverse( ), ipiv, work, lapackInfo2 );
  return ( lapackInfo2 ? false : true );
 }

// -------------------------------------- real symmetric: cholesky decomposition
// -----------------------------------------------------------------------------

template< class MT1, class MT2 >
bool
LapackDrivers::RealSymmetric::cholesky( const matrix<MT1>&            rsMatrix,
                                              varray<MT2>&            decomposition,
                                              FortranTypes::integer&  lapackInfo,
                                              FortranTypes::character uplo )
 {
  decomposition.resize( 0 ); lapackInfo = 0;
  if ( rsMatrix.numRows( ) != rsMatrix.numCols( ) ) return false;
  FortranTypes::integer info = 0;
  matrix<FortranTypes::dReal> a; copy( rsMatrix, a ); varray<FortranTypes::dReal> d;
  if ( cholesky( a, d, lapackInfo, uplo ) ) { copy( d, decomposition ); return true; }
  return false;
 }

template< class MT1, class MT2 >
bool
LapackDrivers::RealSymmetric::cholesky( const varray<MT1>&            packedMatrix,
                                              varray<MT2>&            decomposition,
                                              FortranTypes::integer&  lapackInfo,
                                              FortranTypes::character uplo )
 {
  decomposition.resize( 0 ); lapackInfo = 0;
  matrix<FortranTypes::dReal> a;
  if ( !BlasDrivers::unpack( packedMatrix, a ) ) return false;
  varray<FortranTypes::dReal> d;
  if ( cholesky( a, d, lapackInfo, uplo ) ) { BlasDrivers::copy( d, decomposition ); return true; }
  return false;
 }

template< >
inline
bool
LapackDrivers::RealSymmetric::cholesky<FortranTypes::dReal,FortranTypes::dReal>( const matrix<FortranTypes::dReal>& rsMatrix,
                                                                 varray<FortranTypes::dReal>& decomposition,
                                                                 FortranTypes::integer& lapackInfo,
                                                                 FortranTypes::character uplo )
 {
  decomposition.resize( 0 ); lapackInfo = 0;
  if ( rsMatrix.numRows( ) != rsMatrix.numCols( ) ) return false;
  matrix<FortranTypes::dReal> c( rsMatrix );
  cpplapack::dpotrf( uplo, c.numRows( ), c[0], c.numRows( ), lapackInfo );
  if ( lapackInfo ) return false;
  BlasDrivers::pack( c, decomposition, uplo );
  return true;
 }

template< >
inline
bool
LapackDrivers::RealSymmetric::cholesky<FortranTypes::sReal,FortranTypes::sReal>( const matrix<FortranTypes::sReal>& rsMatrix,
                                                                 varray<FortranTypes::sReal>& decomposition,
                                                                 FortranTypes::integer& lapackInfo,
                                                                 FortranTypes::character uplo )
 {
  decomposition.resize( 0 ); lapackInfo = 0;
  if ( rsMatrix.numRows( ) != rsMatrix.numCols( ) ) return false;
  matrix<FortranTypes::sReal> c( rsMatrix );
  cpplapack::spotrf( uplo, c.numRows( ), c[0], c.numRows( ), lapackInfo );
  if ( lapackInfo ) return false;
  BlasDrivers::pack( c, decomposition, uplo );
  return true;
 }

#endif /* cpp_missing_template_methods */

#endif /* __LapackDrivers_H__ */
