/* --------------------------------------------------------------------------
 *
 * namespace pava
 *   Pooled Adjascent Violators Algorithm
 *   (Robertson, Wright, Dykstra 1988)
 *
 * required headers
 *   see include header file pava
 *
 * notes
 *   thanks to Chris Genovese for his original R code.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------- */

#ifndef __pava_H__
#define __pava_H__

cpp_using_namespace( std );

cpp_namespace_begin( pava )

  /* ----------------------------------------------------------------------
     template parameter tonicity is pava::isotonic or pava::antitonic;
     C is any container that supports indexed access.
     FP is *preferably* a floating-point type.

     If w.size( ) == 0, weights are assumed to be unit,
     else w.size( ) must match g.size( ).
     ---------------------------------------------------------------------- */

  /* ----------------------------------------------------------------------
     PAV algorithm for antitonic or isotonic regression with bound

     minimize
       sum_{i=from:to} w_i (g_i - f_i)^2
     subject to the constraint
       f_from <= ... <= f_to <= bound (isotonic)
       f_from >= ... >= f_to >= bound (antitonic)

     appropriately applied simple bound maintains optimality.
     ---------------------------------------------------------------------- */

  template<class tonicity, template<typename W> class C, class T>
  void
  pav_b( const C<T>&  g,
         const C<T>&  w,
               C<T>&  fit,
         const T&     bound,
               size_t from = 0,
               size_t to   = -1 );

  template<class tonicity, template<typename W> class C, class T>
  C<T>
  pav_b( const C<T>&  g,
         const C<T>&  w,
         const T&     bound,
               size_t from = 0,
               size_t to   = -1 );

  /* ----------------------------------------------------------------------
     PAV algorithm for antitonic or isotonic regression without bound

     minimize
       sum_{i=from:to} w_i (g_i - f_i)^2
     subject to the constraint
       f_from <= ... <= f_to (isotonic)
       f_from >= ... >= f_to (antitonic)
     ---------------------------------------------------------------------- */

  template<class tonicity, template<typename W> class C, class T>
  void
  pav( const C<T>&  g,
       const C<T>&  w,
             C<T>&  fit,
             size_t from = 0,
             size_t to   = -1 );

  template<class tonicity, template<typename W> class C, class T>
  C<T>
  pav( const C<T>&  g,
       const C<T>&  w,
             size_t from = 0,
             size_t to   = -1 );

  /* ----------------------------------------------------------------------
           sum_{i=from:to} w_i (y_i - fit_i)^2
     or
     sqrt( sum_{i=from:to} w_i (y_i - fit_i)^2 )
     ---------------------------------------------------------------------- */

  template<template<typename W> class C, class FP>
  FP
  ssr( const C<FP>& y,
       const C<FP>& w,
       const C<FP>& fit,
             size_t from       = 0,
             size_t to         = -1,
             bool   returnSqrt = true );

  /* ----------------------------------------------------------------------
     For any container t that supports indexed access,
     this returns the reverse sequence.
     ---------------------------------------------------------------------- */

  template<class T>
  inline
  T
  reverse( const T& t );

  /* ----------------------------------------------------------------------
     For any container t that supports indexed access, this
     returns the specified subrange of the input sequence.
     ---------------------------------------------------------------------- */

  template<class T>
  inline
  T
  subrange( const T& t, size_t from = 0, size_t to = -1 );

  /* ----------------------------------------------------------------------
     Tonicity classes.
     ---------------------------------------------------------------------- */

  struct antitonic
   {
    template<class T>
    static
    size_t
    firstViolator( const vector<T>& f )
     {
      size_t fv = 1;
      while ( ( fv < f.size( ) ) && ( f[fv] <= f[fv-1] ) ) fv++;
      return fv;
     }

    template<class T>
    static
    vector<T>
    bounded( const vector<T>& t, const T& lowerBound )
     {
      vector<T> lt( t );
      putBound( lt, lowerBound );
      return lt;
     }

    template<class T>
    static
    void
    putBound( vector<T>& t, const T& lowerBound, size_t from = 0, size_t to = -1 )
     {
      if ( t.size( ) == 0 ) return;

      to   = min( to, t.size( ) - 1 );
      from = min( to, from          );

      for ( size_t i = from; i <= to; i++ ) t[i] = max( lowerBound, t[i] );
     }
   };

  struct isotonic
   {
    template<class T>
    static
    size_t
    firstViolator( const vector<T>& f )
     {
      size_t fv = 1;
      while ( ( fv < f.size( ) ) && ( f[fv] >= f[fv-1] ) ) fv++;
      return fv;
     }

    template<class T>
    static
    vector<T>
    bounded( const vector<T>& t, const T& upperBound )
     {
      vector<T> lt( t );
      putBound( lt, upperBound );
      return lt;
     }

    template<class T>
    static
    void
    putBound( vector<T>& t, const T& upperBound, size_t from = 0, size_t to = -1 )
     {
      if ( t.size( ) == 0 ) return;

      to   = min( to, t.size( ) - 1 );
      from = min( to, from          );

      for ( size_t i = from; i <= to; i++ ) t[i] = min( upperBound, t[i] );
     }
   };

cpp_namespace_end

template<class tonicity, template<typename W> class C, class T>
void
pava::pav( const C<T>&  g,
           const C<T>&  w,
                 C<T>&  fit,
                 size_t from = 0,
                 size_t to   = -1 )
 {
  /* ----------------------------------------------------------------------
     PAV algorithm for antitonic or isotonic regression without bound

     minimize
       sum_{i=from:to} w_i (g_i - f_i)^2
     subject to the constraint
       f_from <= ... <= f_to (isotonic)
       f_from >= ... >= f_to (antitonic)

     template parameter tonicity is pava::isotonic or pava::antitonic.
     ---------------------------------------------------------------------- */

  if ( g.size( ) == 0 ) return;

  to   = min( to, g.size( ) - 1 );
  from = min( to, from          );

  size_t m  = to - from + 1;

  vector<T> f( m    );
  vector<T> W( m, 1 );

  for ( size_t i = 0, j = from; i < m; i++, j++ ) f[i] = g[j];

  if ( w.size( ) > 0 )
   for ( size_t i = 0, j = from; i < m; i++, j++ ) W[i] = w[j];

  vector< pair<size_t,size_t> > index( m );
  for ( size_t i = 0; i < m; i++ ) index[i].first = index[i].second = i;

  // first i such that f[i] > f[i-1] (antitonic),
  // or                f[i] < f[i-1] (isotonic),
  // or                f.size( ) (no violators).
  //
  size_t fv = tonicity::firstViolator( f );

  T a = T( );

  while ( fv < f.size( ) )
   {
    a = W[fv] + W[fv-1];

    index[fv-1].second = index[fv].second;
    index.erase( index.begin( ) + fv );

    f[fv-1] = ( f[fv] * W[fv] + f[fv-1] * W[fv-1] ) / a;
    f.erase( f.begin( ) + fv );

    W[fv-1] = a;
    W.erase( W.begin( ) + fv );

    // first i such that f[i] > f[i-1] (antitonic),
    // or                f[i] < f[i-1] (isotonic),
    // or                f.size( ) (no violators).
    //
    fv = tonicity::firstViolator( f );
   }

  if ( fit.size( ) < g.size( ) ) fit.resize( g.size( ) );

  for ( size_t i = 0; i < f.size( ); i++ )
   {
    a = f[i];
    for ( size_t j = index[i].first, k = j + from; j <= index[i].second; j++, k++ )
     fit[k] = a;
   }
 }

template<class tonicity, template<typename W> class C, class T>
void
pava::pav_b( const C<T>&  g,
             const C<T>&  w,
                   C<T>&  fit,
             const T&     bound,
                   size_t from = 0,
                   size_t to   = -1 )
 {
  /* ----------------------------------------------------------------------
     PAV algorithm for antitonic or isotonic regression without bound

     minimize
       sum_{i=from:to} w_i (g_i - f_i)^2
     subject to the constraint
       f_from <= ... <= f_to <= bound (isotonic)
       f_from >= ... >= f_to >= bound (antitonic)

     appropriately applied simple bound maintains optimality.
     template parameter tonicity is pava::isotonic or pava::antitonic.
     ---------------------------------------------------------------------- */

  pav<tonicity>( g, w, fit, from, to );
  typename tonicity::putBound( fit, bound, from, to );
 }

template<class tonicity, template<typename W> class C, class T>
C<T>
pava::pav( const C<T>&  g,
           const C<T>&  w,
                 size_t from = 0,
                 size_t to   = -1 )
 {
  C<T> fit( g.size( ) );
  pav<tonicity>( g, w, fit, from, to );
  return fit;
 }

template<class tonicity, template<typename W> class C, class T>
C<T>
pava::pav_b( const C<T>&  g,
             const C<T>&  w,
             const T&     bound,
                   size_t from = 0,
                   size_t to   = -1 )
 {
  C<T> fit( g.size( ) );
  pav_b<tonicity>( g, w, fit, bound, from, to );
  return fit;
 }

template<template<typename W> class C, class FP>
FP
pava::ssr( const C<FP>& y,
           const C<FP>& w,
           const C<FP>& fit,
                 size_t from       = 0,
                 size_t to         = -1,
                 bool   returnSqrt = true )
 {
  if ( ( y.size( ) == 0 ) || ( fit.size( ) != y.size( ) ) ) return -1;

  to   = min( to, y.size( ) - 1 );
  from = min( to, from          );

  FP s = FP( ), t = FP( );
  if ( w.size( ) == 0 )
   for ( size_t i = from; i <= to; i++ ) { t = y[i] - fit[i]; s += t * t; }
  else
   for ( size_t i = from; i <= to; i++ ) { t = y[i] - fit[i]; s += w[i] * t * t; }

  return ( returnSqrt ? sqrt( s ) : s );
 }

template<class T>
T
pava::reverse( const T& t )
 {
  /* ----------------------------------------------------------------------
     For any container t that supports indexed access,
     this returns the reverse sequence.
     ---------------------------------------------------------------------- */

  T rt( t.size( ) );
  for ( size_t i = 0, j = t.size( ) - 1; i < t.size( ); i++, j-- ) rt[j] = t[i];
  return rt;
 }

template<class T>
T
pava::subrange( const T& t, size_t from = 0, size_t to = -1 )
 {
  /* ----------------------------------------------------------------------
     For any container t that supports indexed access, this
     returns the specified subrange of the input sequence.
     ---------------------------------------------------------------------- */

  if ( t.size( ) == 0 ) return T( );

  to   = min( to, t.size( ) - 1 );
  from = min( to, from          );

  size_t n = to - from + 1;

  if ( n == t.size( ) ) return t;

  T s( n );
  for ( size_t i = 0, j = from; i < n; i++, j++ ) s[i] = t[j];

  return s;
 }

// template<class tonicity, class T>
// vector<T>
// pava::pav( const vector<T>& g, const vector<T>& w )
//  {
//   /* ----------------------------------------------------------------------
//      PAV algorithm for antitonic or isotonic regression without bound
// 
//      minimize
//        sum_i w_i (g_i - f_i)^2
//      subject to the constraint
//        f_1 <= f_2 <= ... <= f_m (isotonic)
//        f_1 >= f_2 >= ... >= f_m (antitonic)
// 
//      template parameter tonicity is pava::isotonic or pava::antitonic.
//      ---------------------------------------------------------------------- */
// 
//   vector<T> f( g );
//   vector<T> W( w );
// 
//   if ( W.size( ) == 0 ) W.resize( f.size( ), (T) 1 );
// 
//   size_t m = f.size( );
// 
//   vector< pair<size_t,size_t> > index( m );
//   for ( size_t i = 0; i < m; i++ ) index[i].first = index[i].second = i;
// 
//   // first i such that f[i] > f[i-1] (antitonic),
//   // or                f[i] < f[i-1] (isotonic),
//   // or                f.size( ) (no violators).
//   //
//   size_t fv = tonicity::firstViolator( f );
// 
//   T a = T( );
// 
//   while ( fv < f.size( ) )
//    {
//     a = W[fv] + W[fv-1];
// 
//     index[fv-1].second = index[fv].second;
//     index.erase( index.begin( ) + fv );
// 
//     f[fv-1] = ( f[fv] * W[fv] + f[fv-1] * W[fv-1] ) / a;
//     f.erase( f.begin( ) + fv );
// 
//     W[fv-1] = a;
//     W.erase( W.begin( ) + fv );
// 
//     // first i such that f[i] > f[i-1] (antitonic),
//     // or                f[i] < f[i-1] (isotonic),
//     // or                f.size( ) (no violators).
//     //
//     fv = tonicity::firstViolator( f );
//    }
// 
//   vector<T> fstar( m );
//   for ( size_t i = 0; i < f.size( ); i++ )
//    {
//     a = f[i];
//     for ( size_t j = index[i].first; j <= index[i].second; j++ ) fstar[j] = a;
//    }
// 
//   return fstar;
//  }
// 
// template<class tonicity, class T>
// vector<T>
// pava::pav_b( const vector<T>& g, const vector<T>& w, const T& bound )
//  {
//   /* ----------------------------------------------------------------------
//      PAV algorithm for antitonic or isotonic regression with bound
// 
//      minimize
//        sum_i w_i (g_i - f_i)^2
//      subject to the constraint
//        f_1 <= f_2 <= ... <= f_m <= bound (isotonic)
//        f_1 >= f_2 >= ... >= f_m >= bound (antitonic)
// 
//      appropriately applied simple bound maintains optimality.
//      template parameter tonicity is pava::isotonic or pava::antitonic.
//      ---------------------------------------------------------------------- */
// 
//   return typename tonicity::bounded( pav<tonicity>( g, w ), bound );
//  }
// 
// template<class T>
// T
// pava::reverse( const T& t )
//  {
//   /* ----------------------------------------------------------------------
//      For any container t that supports iterators and reverse iterators,
//      this returns the reverse sequence.
//      ---------------------------------------------------------------------- */
// 
//   T rt( t.size( ) );
//   typename T::const_reverse_iterator p = t.rbegin( );
//   typename T::iterator               q = rt.begin( );
//   while ( p != t.rend( ) ) { (*q) = (*p); ++p; ++q; }
//   return rt;
//  }

#endif /* __pava_H__ */
