/* --------------------------------------------------------------------------
 *
 * namespace wls
 *   weighted least squares
 *
 * required headers
 *   see include header file wls
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __wls_H__
#define __wls_H__

cpp_namespace_begin( wls )

 cpp_namespace_begin( bases )

  template< template<typename W> class C, class FP >
  struct polynomial_basis
   {
    polynomial_basis( const C<FP>& x, size_t degree = 2 )
     : my_x( x ), my_degree( min( x.size( ), degree ) )
     {
     }

    size_t
    numRows( size_t from = 0, size_t to = -1 ) const
     {
      to   = min( to, my_x.size( ) - 1 );
      from = min( to, from             );

      return ( to - from + 1 );
     }

    size_t
    numCols( size_t from = 0, size_t to = -1 ) const
     {
      return ( my_degree + 1 );
     }

    matrix<FP>
    operator()( size_t from = 0, size_t to = -1 ) const
     {
      to   = min( to, my_x.size( ) - 1 );
      from = min( to, from             );

      size_t n = to - from + 1;
      size_t m = my_degree + 1;

      matrix<FP> bm( n, m, 1.0 );

      varray<FP> column( n );
      for ( size_t j = 0, k = from; j < n; j++, k++ ) column[j] = my_x[k];

      for ( size_t i = 1; i < m; i++ )
       {
        bm.setCol( i, column );
        for ( size_t j = 0, k = from; j < n; j++, k++ ) column[j] *= my_x[k];
       }

      return bm;
     }

    private:
     C<FP>  my_x;
     size_t my_degree;
   };

 cpp_namespace_end;

 template< template<typename W> class C, class FP,
           template< template<typename W> class C, class FP > class Basis >
 int
 wlsfit( const C<FP>&       data,
         const C<FP>&       weights,
         const Basis<C,FP>& basis,
               C<FP>&       coefficients,
               C<FP>&       fit,
               C<FP>&       weighted_residuals,
               size_t       from =  0,
               size_t       to   = -1 )
  {
   return wlsfit( data, weights, basis( from, to ), coefficients,
                  fit, weighted_residuals, from, to );
  }

 template< template<typename W> class C, class FP >
 int
 wlsfit( const C<FP>&      data,
         const C<FP>&      weights,
         const matrix<FP>& basis_matrix,
               C<FP>&      coefficients,
               C<FP>&      fit,
               C<FP>&      weighted_residuals,
               size_t      from =  0,
               size_t      to   = -1 )
  {
   if ( data.size( ) == 0 ) return 1;

   to   = min( to, data.size( ) - 1 );
   from = min( to, from             );

   size_t nn = basis_matrix.numRows( );
   size_t mm = basis_matrix.numCols( );

   if ( nn != ( to - from + 1 ) ) return 2;

   FortranTypes::integer n = nn;
   FortranTypes::integer m = mm;

   matrix<FortranTypes::dReal> a( nn, mm );

   cpp_scope_begin
    const FortranTypes::dReal* const cA = basis_matrix[0];
          FortranTypes::dReal*       ca = a[0];
    size_t k = nn * mm;
    for ( size_t i = 0; i < k; i++ )
     ca[i] = cpp_static_cast( FortranTypes::dReal, cA[i] );
   cpp_scope_end

   matrix<FortranTypes::dReal> w( nn, nn );
   for ( size_t j = 0; j < nn; j++ ) w( j, j ) = 1.0;

   if ( weights.size( ) > 0 )
    for ( size_t j = 0, i = from; j < nn; j++, i++ )
     w( j, j ) /= cpp_static_cast( FortranTypes::dReal, weights[i] );

   varray<FortranTypes::dReal> d( nn );
   for ( size_t j = 0, i = from; j < nn; j++, i++ )
    d[j] = cpp_static_cast( FortranTypes::dReal, data[i] );

   varray<FortranTypes::dReal> x( mm ), y( nn );

   FortranTypes::integer lwork = -1;
   FortranTypes::integer info  =  0;

   cpp_scope_begin
    dReal work = 0.0;
    cpplapack::dggglm( n, m, n, a[0], n, w[0], n, d( ), x( ), y( ), &work, lwork, info );
    lwork = cpp_static_cast( FortranTypes::integer, work );
   cpp_scope_end

   dReal work[lwork];
   cpplapack::dggglm( n, m, n, a[0], n, w[0], n, d( ), x( ), y( ), work, lwork, info );

   if ( info == 0 )
    {
     coefficients.resize( mm );
     for ( size_t i = 0; i < mm; i++ )
      coefficients[i] = cpp_static_cast( FP, x[i] );

     if ( weighted_residuals.size( ) != data.size( ) )
      weighted_residuals.resize( data.size( ) );
     for ( size_t j = 0, i = from; j < nn; j++, i++ )
      weighted_residuals[i] = cpp_static_cast( FP, y[j] );

     if ( fit.size( ) != data.size( ) ) fit.resize( data.size( ) );
     if ( weights.size( ) > 0 )
      for ( size_t i = from; i <= to; i++ )
       fit[i] = data[i] - weighted_residuals[i] / weights[i];
     else
      for ( size_t i = from; i <= to; i++ )
       fit[i] = data[i] - weighted_residuals[i];
    }

   return info;
  }

cpp_namespace_end;

#endif /* __wls_H__ */
