/* --------------------------------------------------------------------------
 *
 * class MinQNB
 *   implementation
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#include <minimizer/MinQNB>

// ---------------------------------------------------------------- static stuff
// -----------------------------------------------------------------------------

enum
statusIndices
 {
  /*  0 */ sm_err_n = 0,
  /*  1 */ sm_err_m,
  /*  2 */ sm_err_factr,
  /*  3 */ sm_err_nbd,
  /*  4 */ sm_err_solution,
  /*  5 */ sm_err_stpLTmin,
  /*  6 */ sm_err_stpGTmax,
  /*  7 */ sm_err_initialG,
  /*  8 */ sm_err_ftol,
  /*  9 */ sm_err_gtol,
  /* 10 */ sm_err_xtol,
  /* 11 */ sm_err_stpmin,
  /* 12 */ sm_err_stpmax,

  /* 13 */ sm_warn_rounding,
  /* 14 */ sm_warn_xtol,
  /* 15 */ sm_warn_stpEQmax,
  /* 16 */ sm_warn_stpEQmin,

  /* 17 */ sm_converge,
  /* 18 */ sm_converge_prG,
  /* 19 */ sm_converge_f,

  /* 20 */ sm_abnormal,

  /* 21 */ sm_stop_evals,
  /* 22 */ sm_stop_iters,
  /* 23 */ sm_stop_prG,

  /* 24 */ sm_msgCount
 };

static
vector<string>
msgList( )
 {
  vector<string> msgs( static_cast<size_t>( sm_msgCount ) );

  msgs[sm_err_n]         = "ERROR: N .LE. 0                                             ";
  msgs[sm_err_m]         = "ERROR: M .LE. 0                                             ";
  msgs[sm_err_factr]     = "ERROR: FACTR .LT. 0                                         ";
  msgs[sm_err_nbd]       = "ERROR: INVALID NBD                                          ";
  msgs[sm_err_solution]  = "ERROR: NO FEASIBLE SOLUTION                                 ";
  msgs[sm_err_stpLTmin]  = "ERROR: STP .LT. STPMIN                                      ";
  msgs[sm_err_stpGTmax]  = "ERROR: STP .GT. STPMAX                                      ";
  msgs[sm_err_initialG]  = "ERROR: INITIAL G .GE. ZERO                                  ";
  msgs[sm_err_ftol]      = "ERROR: FTOL .LT. ZERO                                       ";
  msgs[sm_err_gtol]      = "ERROR: GTOL .LT. ZERO                                       ";
  msgs[sm_err_xtol]      = "ERROR: XTOL .LT. ZERO                                       ";
  msgs[sm_err_stpmin]    = "ERROR: STPMIN .LT. ZERO                                     ";
  msgs[sm_err_stpmax]    = "ERROR: STPMAX .LT. STPMIN                                   ";

  msgs[sm_warn_rounding] = "WARNING: ROUNDING ERRORS PREVENT PROGRESS                   ";
  msgs[sm_warn_xtol]     = "WARNING: XTOL TEST SATISFIED                                ";
  msgs[sm_warn_stpEQmax] = "WARNING: STP = STPMAX                                       ";
  msgs[sm_warn_stpEQmin] = "WARNING: STP = STPMIN                                       ";

  msgs[sm_converge]      = "CONVERGENCE                                                 ";
  msgs[sm_converge_prG]  = "CONVERGENCE: NORM OF PROJECTED GRADIENT <= PGTOL            ";
  msgs[sm_converge_f]    = "CONVERGENCE: REL_REDUCTION_OF_F <= FACTR*EPSMCH             ";

  msgs[sm_abnormal]      = "ABNORMAL_TERMINATION_IN_LNSRCH                              ";

  msgs[sm_stop_evals]    = "STOP: MAXIMUM_EVALUATIONS_REACHED                           ";
  msgs[sm_stop_iters]    = "STOP: MAXIMUM_ITERATIONS_REACHED                            ";
  msgs[sm_stop_prG]      = "STOP: ||PROJECTED_GRADIENT||_IS_SUFFICIENTLY_SMALL          ";

  return msgs;
 };

const vector<string>
MinQNB::statusMsgs( msgList( ) );

static
void
minimizeWithoutRestart( dReal* const paramBuf,
                        dReal& f, dReal& prGradNorm,
                        integer& numIts, integer& numEvals, string& status,
                        const BasicObjFn* const objectiveFn,
                        integer problemSize, integer maxCorrections,
                        dReal minRelReduceF, dReal maxPrGradNorm, dReal maxRelPrGradNorm,
                        integer maxIterations, integer maxEvaluations,
                        integer verbosity,
                        const integer* const boundType,
                        const dReal* const lowerBound,
                        const dReal* const upperBound,
                        const dReal* const scale,
                        bool doRescale,
                        dReal* dWork, integer* iSave, dReal* dSave, dReal epsilon )
 {
  if ( doRescale )
   for ( integer i = 0; i < problemSize; i++ ) paramBuf[i] *= scale[i];

  dReal   grad[problemSize];
  integer iWork[3 * problemSize];
  logical lSave[4];
  char    task[61];
  char    cSave[61];

  /* ------------------------------------------------------- begin testing/debug

  memset( iSave, (char) 0, 44 * sizeof( integer ) );
  memset( dSave, (char) 0, 29 * sizeof( dReal ) );
  memset( dWork, (char) 0, (   (  2 * maxCorrections + 4 ) * problemSize
                             + ( 11 * maxCorrections + 8 ) * maxCorrections ) * sizeof( dReal ) );

     --------------------------------------------------------- end testing/debug */

  task[60]  = '\0';
  cSave[60] = '\0';
                 //
                 //  fortran strings aren't null-terminated. The function setulb
                 // needs character buffers of size 60. This last null is padded
                 //  at 61st position for the string to be well-defined in c/c++

  strcpy( task,
          "START                                                       " );
          //
          //     fortran doesn't like null-terminated strings. The 55 blanks are
          //   padded for this purpose. The function setulb required a character
          // buffer of size 60. a null at 61st position won't cause any problems

  for ( ; ; )
   {
    cpplbfgs::setulb( problemSize, maxCorrections,
                      paramBuf[0], lowerBound[0], upperBound[0], boundType[0],
                      f, grad[0],
                      minRelReduceF, maxPrGradNorm,
                      dWork[0], iWork[0],
                      task[0], verbosity, cSave[0],
                      lSave[0], iSave[0], dSave[0], epsilon );

    if ( task[0] == 'F' && task[1] == 'G' )                             // "FG*"
     {
      if ( doRescale )
       for ( int i = 0; i < problemSize; i++ ) paramBuf[i] /= scale[i];

      objectiveFn->value( paramBuf, &f );
      objectiveFn->gradient( paramBuf, grad );

      if ( doRescale )
       for ( int i = 0; i < problemSize; i++ )
        {
         paramBuf[i] *= scale[i];
         grad[i]     /= scale[i];
        }

      if ( iSave[33] >= maxEvaluations )
       strcpy( task, msgList( )[sm_stop_evals].c_str( ) );
     }
    else if ( ( task[0] == 'N' ) && ( task[1] == 'E' ) )             // "NEW_X*"
     {
      if ( iSave[29] >= maxIterations )
       strcpy( task, msgList( )[sm_stop_iters].c_str( ) );
      else if ( dSave[12] <= maxRelPrGradNorm * ( 1.0 + fabs( dSave[1] ) ) )
       strcpy( task, msgList( )[sm_stop_prG].c_str( ) );
     }
    else                         // "STOP*", "ERROR*", "WARN*", "CONV*", "ABNO*"
     break;
   }

  numIts     = iSave[29];
  numEvals   = iSave[33];
  f          = dSave[1];
  prGradNorm = dSave[12];

  status     = task;

                                // check bounds in case of ABNORMAL_TERMINATION*
                                //
  if ( ( task[0] == 'A' ) && ( task[1] == 'B' ) )
   for ( int i = 0; i < problemSize; i++ )
    if ( boundType[i] > 0 )
     {
      if ( ( boundType[i] <= 2 ) && ( paramBuf[i] < lowerBound[i] ) )
       paramBuf[i] = lowerBound[i];
      else if ( ( boundType[i] >= 2 ) && ( paramBuf[i] > upperBound[i] ) )
       paramBuf[i] = upperBound[i];
     }

  if ( doRescale )
   for ( int i = 0; i < problemSize; i++ ) paramBuf[i] /= scale[i];
 }

// -------------------------------------------------- incremental initialization
// -----------------------------------------------------------------------------

MinQNB::MinQNB( )
 : my_problemSize( ),
   my_maxCorrections( ),
   my_minRelReduceF( 1.0e+07 ),
   my_maxPrGradNorm( 1.0e-10 ),
   my_maxEvaluations( 2000 ),
   my_maxIterations( 1000 ),
   my_maxRelPrGradNorm( 0.0 ),
   my_maxRestarts( 10 ),
   my_checkHdet( false ),
   my_stopAtPlusDefH( false ),
   my_mrrfReduction( 0.1 ),
   my_verbosity( -1 ),
   my_lowerBound( ),
   my_upperBound( ),
   my_boundType( ),
   my_scale( ),
   my_objectiveFn( ),
   my_numIterations( 0 ),
   my_numEvaluations( 0 ),
   my_numRestarts( 0 ),
   my_status( ),
   my_s( ),
   my_y( ),
   my_sTy( ),
   my_invTheta( ),
   my_choleskyH( ),
   my_lapackInfo( -25 )
 {
 }

void
MinQNB::setSizes( integer problemSize, integer maxCorrections )
 {
  my_problemSize    = problemSize;
  my_maxCorrections = maxCorrections;

  my_lowerBound.resize( problemSize, -DBL_MAX );
  my_upperBound.resize( problemSize, +DBL_MAX );
  my_boundType.resize( problemSize, 0 );

  my_scale.resize( problemSize, 1.0 );
 }

void
MinQNB::setStoppingCriteria( dReal   minRelReduceF    /* = 1.0e+07 */,
                             dReal   maxPrGradNorm    /* = 1.0e-10 */,
                             integer maxEvaluations   /* =    2000 */,
                             integer maxIterations    /* =    1000 */,
                             dReal   maxRelPrGradNorm /* =     0.0 */  )
 {
  my_minRelReduceF    = minRelReduceF;
  my_maxPrGradNorm    = maxPrGradNorm;
  my_maxEvaluations   = maxEvaluations;
  my_maxIterations    = maxIterations;
  my_maxRelPrGradNorm = maxRelPrGradNorm;
 }

void
MinQNB::setVerbosity( integer verbosity /* = -1 */ )
 {
  my_verbosity = verbosity;
 }

void
MinQNB::setBounds(       integer index,
                   const string& boundType  /* = "none" */,
                         dReal   lowerBound /* = -DBL_MAX */,
                         dReal   upperBound /* = +DBL_MAX */ )
 {
  my_lowerBound[index] = lowerBound;
  my_upperBound[index] = upperBound;

  if ( boundType == our_boundTypes[1] )
   my_boundType[index] = 1;
  else if ( boundType == our_boundTypes[2] )
   my_boundType[index] = 2;
  else if ( boundType == our_boundTypes[3] )
   my_boundType[index] = 3;
  else
   my_boundType[index] = 0;
                                                     //
                                                     // 0 <= index < problemSize
 }

void
MinQNB::setBounds(       integer indexMin,
                         integer indexRange,
                   const string& boundType  /* = "none" */,
                         dReal   lowerBound /* = -DBL_MAX */,
                         dReal   upperBound /* = +DBL_MAX */ )
 {
  for ( int i = indexMin; i < indexMin + indexRange; i++ )
   setBounds( i, boundType, lowerBound, upperBound );
 }

void
MinQNB::setScale( integer index, dReal scale /* = 1.0 */ )
 {
  my_scale[index] = ( scale > 0.0 ) ? scale : 1.0;
 }

void
MinQNB::setScale( integer indexMin, integer indexRange, dReal scale /* = 1.0 */ )
 {
  for ( int i = indexMin; i < indexMin + indexRange; i++ ) setScale( i, scale );
 }

void
MinQNB::setObjectiveFn( const BasicObjFn* objectiveFn )
 {
  my_objectiveFn = objectiveFn;
 }

void
MinQNB::setRestartOpts( integer maxRestarts,
                        bool    checkHdet,
                        bool    stopAtPlusDefH,
                        dReal   mrrfReduction )
 {
  my_maxRestarts    = maxRestarts;
  my_checkHdet      = checkHdet;
  my_stopAtPlusDefH = checkHdet ? stopAtPlusDefH : false;

  my_mrrfReduction = 0.5;
  if ( ( mrrfReduction >= 0.0 ) && ( mrrfReduction < 1.0 ) )
   my_mrrfReduction = mrrfReduction;
 }

// ------------------------------------------------------------ copy constructor
// -----------------------------------------------------------------------------

MinQNB::MinQNB( const MinQNB& another )
 : my_problemSize( another.my_problemSize ),
   my_maxCorrections( another.my_maxCorrections ),
   my_minRelReduceF( another.my_minRelReduceF ),
   my_maxPrGradNorm( another.my_maxPrGradNorm ),
   my_maxEvaluations( another.my_maxEvaluations ),
   my_maxIterations( another.my_maxIterations ),
   my_maxRelPrGradNorm( another.my_maxRelPrGradNorm ),
   my_maxRestarts( another.my_maxRestarts ),
   my_checkHdet( another.my_checkHdet ),
   my_stopAtPlusDefH( another.my_stopAtPlusDefH ),
   my_mrrfReduction( another.my_mrrfReduction ),
   my_verbosity( another.my_verbosity ),
   my_lowerBound( another.my_lowerBound ),
   my_upperBound( another.my_upperBound ),
   my_boundType( another.my_boundType ),
   my_scale( another.my_scale ),
   my_objectiveFn( another.my_objectiveFn ),
   my_numIterations( another.my_numIterations ),
   my_numEvaluations( another.my_numEvaluations ),
   my_numRestarts( another.my_numRestarts ),
   my_status( another.my_status ),
   my_s( another.my_s ),
   my_y( another.my_y ),
   my_sTy( another.my_sTy ),
   my_invTheta( another.my_invTheta ),
   my_choleskyH( another.my_choleskyH ),
   my_lapackInfo( another.my_lapackInfo )
 {
 }

// ------------------------------------------------------------------ assignment
// -----------------------------------------------------------------------------

MinQNB&
MinQNB::operator=( const MinQNB& rhs )
 {
  my_problemSize    = rhs.my_problemSize;
  my_maxCorrections = rhs.my_maxCorrections;

  my_minRelReduceF    = rhs.my_minRelReduceF;
  my_maxPrGradNorm    = rhs.my_maxPrGradNorm;
  my_maxEvaluations   = rhs.my_maxEvaluations;
  my_maxIterations    = rhs.my_maxIterations;
  my_maxRelPrGradNorm = rhs.my_maxRelPrGradNorm;

  my_maxRestarts    = rhs.my_maxRestarts;
  my_checkHdet      = rhs.my_checkHdet;
  my_stopAtPlusDefH = rhs.my_stopAtPlusDefH;
  my_mrrfReduction  = rhs.my_mrrfReduction;

  my_verbosity = rhs.my_verbosity;

  my_lowerBound.resize( rhs.my_lowerBound.size( ) );
  my_upperBound.resize( rhs.my_upperBound.size( ) );
  my_boundType.resize( rhs.my_boundType.size( ) );

  my_scale.resize( rhs.my_scale.size( ) );

  my_lowerBound = rhs.my_lowerBound;
  my_upperBound = rhs.my_upperBound;
  my_boundType  = rhs.my_boundType;

  my_scale = rhs.my_scale;

  my_objectiveFn = rhs.my_objectiveFn;

  my_numIterations  = rhs.my_numIterations;
  my_numEvaluations = rhs.my_numEvaluations;
  my_numRestarts    = rhs.my_numRestarts;
  my_status         = rhs.my_status;

  my_s         = rhs.my_s;
  my_y         = rhs.my_y;
  my_sTy       = rhs.my_sTy;
  my_invTheta  = rhs.my_invTheta;

  my_choleskyH  = rhs.my_choleskyH;
  my_lapackInfo = rhs.my_lapackInfo;

  return *this;
 }

// -------------------------------------------------------------------- minimize
// -----------------------------------------------------------------------------

void
MinQNB::minimize( dReal* const paramBuf, dReal epsilon )
 {
  // cerr << "begin minimize( )" << endl;
  // cerr.precision( 15 );

  if ( my_maxIterations <= 0 )
   {
    my_status = msgList( )[sm_stop_iters];
    return;
   }
  if ( my_maxEvaluations <= 0 )
   {
    my_status = msgList( )[sm_stop_evals];
    return;
   }

  dReal lowerBound[my_problemSize];
  dReal upperBound[my_problemSize];

  for ( integer i = 0; i < my_problemSize; i++ )
   {
    lowerBound[i] = my_lowerBound[i] * my_scale[i];
    upperBound[i] = my_upperBound[i] * my_scale[i];
   }

  bool doRescale = false;
  for ( integer i = 0; i < my_problemSize; i++ )
   doRescale = doRescale || ( my_scale[i] != 1.0 );

  my_numIterations  = 0;
  my_numEvaluations = 0;
  my_numRestarts    = 0;

  dReal params[my_problemSize];

  memcpy( static_cast<void*>( params ), static_cast<const void*>( paramBuf ),
          my_problemSize * sizeof( dReal ) );
       //
       // ATTN params maintains a copy of paramBuf[i], i = 0,...,my_problemSize.
       // In case a minimization cycle (restart) is unsuccessful in reducing the
       // objective  function  value,  these  original values  are restored into
       // paramBuf.  Another reason for working with the user-supplied buffer is
       // that paramBuf[i], i < 0, i >= my_problemSize  may be  relevant for the
       // objective function computations depending on how these are been coded.

  dReal f = 0.0, prGradNorm = 0.0;
  integer numIts = 0, numEvals = 0;

  dReal fOld;
  my_objectiveFn->value( paramBuf, &fOld );

  bool plusDef = false, plusDefOld = false;
  dReal mrrfReduction = 1.0;

  if ( my_checkHdet )
   {
    my_objectiveFn->hessian( paramBuf, my_choleskyH );
    cpplapack::dpotrf( cppblas::UpperTriangle, my_choleskyH.numRows( ),
                       my_choleskyH[0], my_choleskyH.numRows( ), my_lapackInfo );
    plusDef = ( my_lapackInfo == 0 );
   }

  dReal   dWork[   (  2 * my_maxCorrections + 4 ) * my_problemSize
                 + ( 11 * my_maxCorrections + 8 ) * my_maxCorrections ];
  integer iSave[44];
  dReal   dSave[29];

  bool restart    = true;
  bool acceptable = false;

  while ( restart )
   {
    minimizeWithoutRestart( paramBuf, f, prGradNorm, numIts, numEvals, my_status,
                            my_objectiveFn, my_problemSize, my_maxCorrections,
                            mrrfReduction * my_minRelReduceF, my_maxPrGradNorm,
                            my_maxRelPrGradNorm, my_maxIterations - my_numIterations,
                            my_maxEvaluations - my_numEvaluations, my_verbosity,
                            my_boundType( 0 ), lowerBound, upperBound, my_scale( 0 ),
                            doRescale, dWork, iSave, dSave, epsilon );

    acceptable = ( f < fOld );

                         // a restart is performed if all these conditions hold:
                         //
                         //                      (a) enough restarts are allowed
                         //       (b) task is "CONVERGENCE: REL*" or "ABNORMAL*"
                         //  (c) for second restart onwards: the current restart
                         //                has helped in reducing function value
                         //
    restart =    acceptable
              && ( my_numRestarts < my_maxRestarts )
              && (    ( my_status.substr( 0, 16 ) == "CONVERGENCE: REL" )
                   || ( my_status.substr( 0, 8  ) == "ABNORMAL"         ) );

                         //   if the user has requested checking the hessian for
                         //                      positive definiteness, do this:
                         //
                         //           (a) reduce convergence criterion using the
                         //                       user-supplied reduction factor
                         //    (b) if ( f == fOld ) accept the new parameters if
                         //     the old Hessian was not +def, but the new one is
                         //
    if ( my_checkHdet )
     {
      my_objectiveFn->hessian( paramBuf, my_choleskyH );
      cpplapack::dpotrf( cppblas::UpperTriangle, my_choleskyH.numRows( ),
                         my_choleskyH[0], my_choleskyH.numRows( ), my_lapackInfo );

      plusDefOld = plusDef;
      plusDef    = ( my_lapackInfo == 0 );

      if ( plusDef )
       {
        if ( my_stopAtPlusDefH ) restart = false;
       }
      else
       mrrfReduction *= my_mrrfReduction;

      acceptable = acceptable || ( ( f == fOld ) && plusDef && !plusDefOld );

      if ( !acceptable ) plusDef = plusDefOld;
     }

                       // ATTN         maintain a copy the current best; else if
                       // no progress, restore the best available into paramBuf.
                       //
    if ( acceptable )
     memcpy( static_cast<void*>( params ), static_cast<const void*>( paramBuf ),
             my_problemSize * sizeof( dReal ) );
    else
     memcpy( static_cast<void*>( paramBuf ), static_cast<const void*>( params ),
              my_problemSize * sizeof( dReal ) );

    // cerr << "my_maxRestarts " << my_maxRestarts << " my_numRestarts " << my_numRestarts << " numIts " << numIts << " numEvals " << numEvals << " fOld " << fOld << " f " << f << " plusDefOld " << plusDefOld << " plusDef " << plusDef << " acceptable " << acceptable << " restart " << restart << " status " << my_status << endl;

    my_numIterations  += numIts;
    my_numEvaluations += numEvals;

    my_numRestarts++; fOld = f;
   }

  my_numRestarts--;

  if ( !plusDef ) my_choleskyH.resize( 0, 0 );

        // extract S, Y and R matrices from within dWork using offsets in iSave.
        // theta is dSave[0]. These are required to compute the BFGS approximate
        //                                                      inverse Hessian.
        //
        //  S, Y and R are most probably not available if the termination status
        // begins with ERROR. Not sure what happens for WARNING. The rest should
        //    be OK. # of iterations should be decremented in case of ABNORMAL_*
        //
        //      ATTN Assume that the L-BFGS-B codes have been modified to update
        //                        these even when convergence has been achieved.

  integer nBFGS = iSave[29] - iSave[25];
  if ( my_status[0] == 'A' ) nBFGS -= 1;
  nBFGS = max( 0, min( my_maxCorrections, nBFGS ) );
                              //
                              // min( ..., current iteration - skipped updates )
                              //  current iteration - 1 in case of ABNORMAL_TERM

  // cerr << "iterations        " << iSave[29]         << endl;
  // cerr << "skipped updates   " << iSave[25]         << endl;
  // cerr << "status            " << my_status         << endl;
  // cerr << "my_maxCorrections " << my_maxCorrections << endl;
  // cerr << "nBFGS             " << nBFGS             << endl;

  if ( ( my_status[0] != 'E' ) && ( nBFGS > 0 ) )
   {
    my_invTheta = 1.0 / dSave[0];

                                                // extract BFGS matrices S and Y
                                                //
    my_s.resize( my_problemSize, nBFGS, dWork + iSave[3] - 1 );
    my_y.resize( my_problemSize, nBFGS, dWork + iSave[4] - 1 );

                              // actual S^T Y is nBFGS X nBFGS; physical size in
                              // dWork is my_maxCorrections X my_maxCorrections.
                              //         Extract only the relevant portion using
                              // R as a wrapper for the physical S^T Y in dWork.
                              //
    matrix<dReal> R( my_maxCorrections, nBFGS, dWork + iSave[5] - 1, false );

    my_sTy.resize( nBFGS, nBFGS, false, 0.0 );

    dReal* r;
    dReal* sTy;
    for ( size_t j = 0; j < my_sTy.numCols( ); j++ )
     {
      r   = R[j];                                        // j-th column pointers
      sTy = my_sTy[j];
      for ( size_t i = 0; i < my_sTy.numRows( ); i++ ) sTy[i] = r[i];
     }
   }
  else
   {
    my_invTheta = 0.0;
    my_s.resize( 0, 0 );
    my_y.resize( 0, 0 );
    my_sTy.resize( 0, 0 );
   }

  // cerr << "end minimize( )" << endl;
 }

// ---------------------------------------------------------- projected gradient
// -----------------------------------------------------------------------------

dReal
MinQNB::prGradNorm( const dReal* const paramBuf ) const
 {
  dReal grad[my_problemSize];
  my_objectiveFn->gradient( paramBuf, grad );
  return prGradNorm( paramBuf, grad );
 }

dReal
MinQNB::prGradNorm( const dReal* const paramBuf, fpVector& gradient ) const
 {
  gradient.resize( my_problemSize );
  my_objectiveFn->gradient( paramBuf, gradient( 0 ) );
  return prGradNorm( paramBuf, gradient( 0 ) );
 }

dReal
MinQNB::prGradNorm( const dReal* const paramBuf, const dReal* const gradient ) const
 {
  // cpplbfgs::projgr( my_problemSize,
  //                   my_lowerBound[0], my_upperBound[0], my_boundType[0],
  //                   paramBuf[0], grad[0], norm );
  //
  // below is an exact translation of projgr( ... ). For some mysterious reasons,
  // when projgr( ) is called as above, it gets wrong values in my_lowerBound,
  // my_upperBound, my_boundType. Calls to setulb in minimize( ) don't suffer
  // from this, somehow.

  dReal gi;
  dReal norm = 0.0;
  for ( int i = 0; i < my_problemSize; i++ )
   {
    gi = gradient[i];

    if ( my_boundType[i] != 0 )
     {
      if ( gi < 0.0 )
       {
        if ( my_boundType[i] >= 2 ) gi = max( paramBuf[i] - my_upperBound[i], gi );
       }
      else
       {
        if ( my_boundType[i] <= 2 ) gi = min( paramBuf[i] - my_lowerBound[i], gi );
       }
     }

    norm = max( norm, fabs( gi ) );
   }

  return norm;
 }

// ---------------------- diagnostics: BFGS approximation to the inverse Hessian
// -----------------------------------------------------------------------------

void
MinQNB::getInvHessian( matrix<dReal>& invHessian ) const
 {
     //        compute BFGS approximation to inverse hessian using BFGS matrices
     //
     // H = S R^-T D R^-1 S^T + theta^-1 ( I - S R^-T Y^T ) ( I - S R^-T Y^T )^T
     //
     //     this is expression 3.7 in  R. H. Byrd, P. Lu, J. Nocedal and C. Zhu,
     //         "A Limited Memory Algorithm for Bound Constrained Optimization",
     //             Northwestern University Department of Electrical Engineering
     //              Technical Report NAM-08; see MinQNB.h for more information.
     //
     // S, Y and R are not available if the termination status begins with ERROR
     //  Not sure how to handle WARNING. Other termination statuses should be OK

  if ( my_status[0] != 'E' )
   {
                                        // copy BFGS matrices S, Y and R = S^T Y
                                        //
    matrix<dReal> S( my_s );
    matrix<dReal> Y( my_y );
    matrix<dReal> R( my_sTy );

    /* -------------------------------------------------------
    cerr << "S.numRows( ) " << S.numRows( ) << " S.numCols( ) " << S.numCols( ) << endl;
    for ( size_t i = 0; i < S.numRows( ); i++ )
     {
      for ( size_t j = 0, k = 0; j < S.numCols( ); j++, k++ )
       {
        if ( k % 5 == 0 ) cerr << endl;
        cerr << setw( 23 ) << S( i, j );
       }
      cerr << endl;
     }

    cerr << "Y.numRows( ) " << Y.numRows( ) << " Y.numCols( ) " << Y.numCols( ) << endl;
    for ( size_t i = 0; i < Y.numRows( ); i++ )
     {
      for ( size_t j = 0, k = 0; j < Y.numCols( ); j++, k++ )
       {
        if ( k % 5 == 0 ) cerr << endl;
        cerr << setw( 23 ) << Y( i, j );
       }
      cerr << endl;
     }

    cerr << "R.numRows( ) " << nBFGS << " R.numCols( ) " << nBFGS << endl;
    for ( size_t i = 0; i < nBFGS; i++ )
     {
      for ( size_t j = 0, k = 0; j < nBFGS; j++, k++ )
       {
        if ( k % 5 == 0 ) cerr << endl;
        cerr << setw( 23 ) << R( i, j );
       }
      cerr << endl;
     }
     ------------------------------------------------------- */

    fpType sum1 = 0.0, sum2 = 0.0;

                                                      // copy diagonal part of R
                                                      //
    varray<dReal> D( R.numRows( ) );
    for ( size_t i = 0; i < R.numRows( ); i++ ) D[i] = R( i, i );

                        // replace S by S R^{-T}; required R is the UT of my_sTy
                        //
    cppblas::dtrsm( cppblas::RightSide, cppblas::UpperTriangle,
                    cppblas::Transpose, cppblas::NonUnitTriangular,
                    S.numRows( ), S.numCols( ), 1.0,
                    R[0], R.numRows( ), S[0], S.numRows( ) );

                        // set R = ( I - S Y^T ); original R is no longer needed
                        //
    R.resize( my_problemSize, my_problemSize, false, 0.0 );

    for ( size_t i = 0; i < R.numRows( ); i++ ) R( i, i ) = 1.0;

    for ( size_t i = 0; i < R.numRows( ); i++ )
     for ( size_t j = 0; j < R.numCols( ); j++ )
      {
       sum1 = 0.0;
       for ( size_t k = 0; k < S.numCols( ); k++ ) sum1 += S( i, k ) * Y( j, k );

       R( i, j ) -= sum1;
      }

  /* -------------------------------------------------------
    cerr << "S.numRows( ) " << S.numRows( ) << " S.numCols( ) " << S.numCols( ) << endl;
    for ( size_t i = 0; i < S.numRows( ); i++ )
     {
      for ( size_t j = 0, k = 0; j < S.numCols( ); j++, k++ )
       {
        if ( k % 5 == 0 ) cerr << endl;
        cerr << setw( 23 ) << S( i, j );
       }
      cerr << endl;
     }

    cerr << "R.numRows( ) " << R.numRows( ) << " R.numCols( ) " << R.numCols( ) << endl;
    for ( size_t i = 0; i < R.numRows( ); i++ )
     {
      for ( size_t j = 0, k = 0; j < R.numCols( ); j++, k++ )
       {
        if ( k % 5 == 0 ) cerr << endl;
        cerr << setw( 23 ) << R( i, j );
       }
      cerr << endl;
     }
     ------------------------------------------------------- */

                                              //    H = S D S^T + theta^-1 R R^T
                                              // correct for any scaling factors
                                              //
    invHessian.resize( my_problemSize, my_problemSize, false, 0.0 );

    for ( size_t i = 0; i < invHessian.numRows( ); i++ )
     {
      sum1 = 0.0;
      for ( size_t k = 0; k < D.size( ); k++ )
       sum1 += S( i, k ) * D[k] * S( i, k );

      sum2 = 0.0;
      for ( size_t k = 0; k < R.numRows( ); k++ )
       sum2 += R( i, k ) * R( i, k );

      invHessian( i, i )
       = ( sum1 + my_invTheta * sum2 ) / ( my_scale[i] * my_scale[i] );

      for ( size_t j = 0; j < i; j++ )
       {
        sum1 = 0.0;
        for ( size_t k = 0; k < D.size( ); k++ )
         sum1 += S( i, k ) * D[k] * S( j, k );

        sum2 = 0.0;
        for ( size_t k = 0; k < R.numRows( ); k++ )
         sum2 += R( i, k ) * R( j, k );

        invHessian( i, j )
         = invHessian( j, i )
         = ( sum1 + my_invTheta * sum2 ) / ( my_scale[i] * my_scale[j] );
       }
     }

  /* -------------------------------------------------------
      cerr << "H.numRows( ) " << invHessian.numRows( ) << " H.numCols( ) " << invHessian.numCols( ) << endl;
      for ( size_t i = 0; i < invHessian.numRows( ); i++ )
       {
        for ( size_t j = 0, k = 0; j < invHessian.numCols( ); j++, k++ )
         {
          if ( k % 5 == 0 ) cerr << endl;
          cerr << setw( 23 ) << invHessian( i, j );
         }
        cerr << endl;
       }
     ------------------------------------------------------- */

    // cerr << "end H-computation" << endl;
   }
  else
   invHessian.resize( 0, 0 );
 }

const string
MinQNB::our_boundTypes[] = { "none", "lower", "both", "upper" };
