/* --------------------------------------------------------------------------
 *
 * class MinQNB
 *   interface for the BFGS/quasi-Newton minimizer L-BFGS-B
 *   ftp://ece.nwu.edu/pub/lbfgs/lbfgs_bcm
 *
 * required headers
 *   see include header file MinQNB
 *
 * notes
 *   this implementation assumes a precompiled library/object file
 *   of the original fortran source; this allows certain flexibility
 *   in using references in place of pointers in the prototype for
 *   function setulb. This implementation may not work in its present
 *   form with c or c++ translations of lbfgs fortran source.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __MinQNB_H__
#define __MinQNB_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class MinQNB
 {
  public:

   // ----------------------------------------------- incremental initialization
   // --------------------------------------------------------------------------

   MinQNB( );

   void
   setSizes( integer problemSize, integer maxCorrections );

   void
   setStoppingCriteria( dReal   minRelReduceF    = 1.0e+07,
                        dReal   maxPrGradNorm    = 1.0e-10,
                        integer maxEvaluations   =    2000,
                        integer maxIterations    =    1000,
                        dReal   maxRelPrGradNorm =     0.0  );
                       //
                       // minRelReduceF*epsilon is used (in minimize( ))  as the
                       //      stopping criterion on relative function reduction
                       //
                       //     maxPrGradNorm is used as the stopping criterion on
                       //            the infinity norm of the projected gradient
                       //
                       //  maxRelPrGradNorm is used in the convergence criterion
                       // ||projected gradient|| <= maxRelPrGradNorm * (1 + |f|)
                       //                  where ||...|| is infinity norm of ...

   void
   setRestartOpts( integer maxRestarts    = 0,
                   bool    checkHdet      = false,
                   bool    stopAtPlusDefH = false,
                   dReal   mrrfReduction  = 0.5 );
                      //
                      //  the minimization process is restarted from the current
                      //  point when status( ) is either "CONVERGENCE: REL*"  or
                      //                                 "ABNORMAL_TERMINATION*"
                      //
                      //        basic stopping criteria for the restart loop are
                      //                 (a) allowed # of restarts (maxRestarts)
                      //         (b) minimizer status (stop if other than above)
                      //       (c) function value (stop if no further reduction)
                      //
                      //    in addition, if checkHdet == stopAtPlusDefH == true,
                      //         stop as soon as the hessian becomes + definite.
                      //      Further, so long as the hessian is not + definite,
                      //         reduce minRelReduceF by 0 <= mrrfReduction < 1.
                      //     If checkHdet == true, the Cholesky decomposition of
                      //  hessian (upper triangle) is available via choleskyH( )

   void
   setVerbosity( integer verbosity = -1 );
              //
              // verbosity             frequency and type of intermediate output
              //
              // < 0                                      no output is generated
              // = 0                   print only one line at the last iteration
              // 1 ... 98   print also f and |proj g| every verbosity iterations
              // 99            print details of every iteration except n-vectors
              // 100            print also the changes of active set and final x
              // > 100        print details of every iteration including x and g
              //
              //                When verbosity > 0, the file iterate.dat will be
              //                              created to summarize the iteration

   void
   setBounds(       integer index,
              const string& boundType  = "none",
                    dReal   lowerBound = -DBL_MAX,
                    dReal   upperBound = +DBL_MAX );
                                //
                                // bound types: "none", "lower", "both", "upper"
                                //                   0 <= index < problemSize( )

   void
   setBounds(       integer indexMin,
                    integer indexRange,
              const string& boundType  = "none",
                    dReal   lowerBound = -DBL_MAX,
                    dReal   upperBound = +DBL_MAX );
                                //
                                //                     set bounds over the range
                                //         indexMin .. indexMin + indexRange - 1
                                //
                                //                0 <= indexMin < problemSize( )
                                //  0 <= indexMin + indexRange <= problemSize( )

   void
   setScale( integer index, dReal scale = 1.0 );

   void
   setScale( integer indexMin, integer indexRange, dReal scale = 1.0 );
                                //
                                //    measurement unit for given parameter range

   void
   setObjectiveFn( const BasicObjFn* objectiveFn );

   // --------------------------------------------------------- copy constructor
   // --------------------------------------------------------------------------

   MinQNB( const MinQNB& another );

   // --------------------------------------------------------------- assignment
   // --------------------------------------------------------------------------

   MinQNB&
   operator=( const MinQNB& rhs );

   // ------------------------------------------------------------ configuration
   // --------------------------------------------------------------------------

   inline
   integer
   problemSize( ) const;

   inline
   integer
   maxCorrections( ) const;

   inline
   dReal
   lowerBound( integer index ) const;

   inline
   dReal
   upperBound( integer index ) const;

   inline
   const string&
   boundType( integer index ) const;

   inline
   dReal
   scale( integer index ) const;

   inline
   dReal
   minRelReduceF( ) const;

   inline
   dReal
   maxPrGradNorm( ) const;

   inline
   integer
   maxIterations( ) const;

   inline
   integer
   maxEvaluations( ) const;

   inline
   dReal
   maxRelPrGradNorm( ) const;

   inline
   integer
   maxRestarts( ) const;

   inline
   bool
   fReductionCheck( ) const;

   inline
   bool
   detHcheck( ) const;

   inline
   dReal
   mrrfReduction( ) const;

   inline
   integer
   verbosity( ) const;

   inline
   const BasicObjFn*
   objectiveFn( ) const;

   // ----------------------------------------------------------------- minimize
   // --------------------------------------------------------------------------

   void
   minimize( dReal* const paramBuf, dReal epsilon = DBL_EPSILON );
      //
      // inverse Hessian computed is its BFGS approximation at the maximum found

   // -------------------------------------------------------------- diagnostics
   // --------------------------------------------------------------------------

   static
   const vector<string>
   statusMsgs;
                                //
                                // complete list of minimization status messages

   inline
   const string&
   status( ) const;
                   //
                   // current minimization status as reported by setulb/minimize

   inline
   size_t
   statusCode( ) const;
                                            //
                                            // index of status( ) in statusMsgs,
                                            // statusMsgs.size( ) if unavailable

   inline
   integer
   numIterations( ) const;
                                //
                                // total number of iterations (for all restarts)

   inline
   integer
   numEvals( ) const;
             //
             // total number of function/gradient evaluations (for all restarts)

   inline
   integer
   numRestarts( ) const;
                                                 //
                                                 // number of minimizer restarts

   inline
   void
   reset( );
           //
           // appropriately resets diagnostics; e. g., number of iterations to 0

   // ------------------------------------------ diagnostics: projected gradient
   // --------------------------------------------------------------------------

   dReal
   prGradNorm( const dReal* const paramBuf ) const;
          //
          // infinity norm of projected gradient; gradient calculated internally

   dReal
   prGradNorm( const dReal* const paramBuf, varray<dReal>& gradient ) const;
            //
            // gradient and infinity norm of projected gradient given parameters

   dReal
   prGradNorm( const dReal* const paramBuf, const dReal* const gradient ) const;
           //
           // infinity norm of projected gradient given parameters and gradients

   // -------------------------------- diagnostics: exact Hessian + definiteness
   // --------------------------------------------------------------------------

   inline
   const matrix<dReal>&
   choleskyH( ) const;
                           //
                           // Cholesky decomposition of hessian (upper triangle)
                           //    when checkHdet == true; see setHdetCheck( ... )
                           //
                           //  returns a 0x0 matrix if hessian is not + definite

   inline
   integer
   lapackInfo( ) const;
                           //
                           //              corresponding lapack info from dpotrf

   // ------------------- diagnostics: BFGS approximation to the inverse Hessian
   // --------------------------------------------------------------------------

   void
   getInvHessian( matrix<dReal>& invHessian ) const;
                         //
                         // returns a 0 X 0 matrix if this could not be computed
                         //                  (e. g., when status( ) == "ERROR*")

  private:
   integer           my_problemSize;                                   //      n
   integer           my_maxCorrections;                                //      m

   dReal             my_minRelReduceF;                                 //  factr
   dReal             my_maxPrGradNorm;                                 //  pgtol

   integer           my_maxEvaluations;
   integer           my_maxIterations;
   dReal             my_maxRelPrGradNorm;

   integer           my_maxRestarts;
   bool              my_checkHdet, my_stopAtPlusDefH;
   dReal             my_mrrfReduction;

   integer           my_verbosity;                                     // iprint

   varray<dReal>      my_lowerBound;                                   //      l
   varray<dReal>      my_upperBound;                                   //      u
   varray<integer>    my_boundType;                                    //    nbd

   varray<dReal>      my_scale;

   const BasicObjFn* my_objectiveFn;

   integer           my_numIterations;
   integer           my_numEvaluations;
   integer           my_numRestarts;

   string            my_status;

   matrix<dReal>     my_s, my_y, my_sTy;                // BFGS compact matrices
   dReal             my_invTheta;

   matrix<dReal>     my_choleskyH;
   integer           my_lapackInfo;

   static
   const string
   our_boundTypes[];
 };

// --------------------------------------------------------------- configuration
// -----------------------------------------------------------------------------

integer
MinQNB::problemSize( ) const
 {
  return my_problemSize;
 }

integer
MinQNB::maxCorrections( ) const
 {
  return my_maxCorrections;
 }

dReal
MinQNB::lowerBound( integer index ) const
 {
  return my_lowerBound[index];
                                                     //
                                                     // 0 <= index < problemSize
 }

dReal
MinQNB::upperBound( integer index ) const
 {
  return my_upperBound[index];
                                                     //
                                                     // 0 <= index < problemSize
 }

const string&
MinQNB::boundType( integer index ) const
 {
  return our_boundTypes[my_boundType[index]];        // 0 <= index < problemSize
 }

dReal
MinQNB::scale( integer index ) const
 {
  return my_scale[index];
                                                     //
                                                     // 0 <= index < problemSize
 }

dReal
MinQNB::minRelReduceF( ) const
 {
  return my_minRelReduceF;
 }

dReal
MinQNB::maxPrGradNorm( ) const
 {
  return my_maxPrGradNorm;
 }

integer
MinQNB::maxEvaluations( ) const
 {
  return my_maxEvaluations;
 }

integer
MinQNB::maxIterations( ) const
 {
  return my_maxIterations;
 }

dReal
MinQNB::maxRelPrGradNorm( ) const
 {
  return my_maxRelPrGradNorm;
 }

integer
MinQNB::maxRestarts( ) const
 {
  return my_maxRestarts;
 }

bool
MinQNB::detHcheck( ) const
 {
  return my_checkHdet;
 }

dReal
MinQNB::mrrfReduction( ) const
 {
  return my_mrrfReduction;
 }

integer
MinQNB::verbosity( ) const
 {
  return my_verbosity;
 }

const BasicObjFn*
MinQNB::objectiveFn( ) const
 {
  return my_objectiveFn;
 }

// ----------------------------------------------------------------- diagnostics
// -----------------------------------------------------------------------------

const string&
MinQNB::status( ) const
 {
  return my_status;
 }

size_t
MinQNB::statusCode( ) const
 {
  return
   distance( statusMsgs.begin( ),
             find( statusMsgs.begin( ), statusMsgs.end( ), my_status ) );
 }

integer
MinQNB::numIterations( ) const
 {
  return my_numIterations;
 }

integer
MinQNB::numEvals( ) const
 {
  return my_numEvaluations;
 }

integer
MinQNB::numRestarts( ) const
 {
  return my_numRestarts;
 }

void
MinQNB::reset( )
 {
  my_numIterations  = 0;
  my_numEvaluations = 0;
  my_numRestarts    = 0;
  my_status         = "";
  my_invTheta       = 0.0;

  my_s.resize( 0, 0 );
  my_y.resize( 0, 0 );
  my_sTy.resize( 0, 0 );

  my_choleskyH.resize( 0, 0 );
 }

// ------------------------------------ diagnostics: exact Hessian +definiteness
// -----------------------------------------------------------------------------

const matrix<dReal>&
MinQNB::choleskyH( ) const
 {
  return my_choleskyH;
 }

integer
MinQNB::lapackInfo( ) const
 {
  return my_lapackInfo;
 }

#endif /* __MinQNB_H__ */
