/* --------------------------------------------------------------------------
 *
 * class MissingInfo
 *   interface to FIASCO missing slices
 *   information in Pittsburgh MRI format
 *
 * required headers
 *   see include header file MissingInfo
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __MissingInfo_H__
#define __MissingInfo_H__

class MissingInfo
 {
  public:
   typedef varray<bool> FlagArray;

  // ---------------------------------------------------------- construct/reinit
  // ---------------------------------------------------------------------------

   MissingInfo( const MissingInfo& another );

   MissingInfo&
   operator=( const MissingInfo& rhs );

   MissingInfo( size_t numInstants = 0, size_t numSlices   = 0,
                size_t zLowerBound = 0, size_t tLowerBound = 0 );
                                  //
                                  // "instants" here = "images" in fiasco jargon

   void
   reinitialize( size_t numInstants = 0, size_t numSlices   = 0,
                 size_t zLowerBound = 0, size_t tLowerBound = 0 );

   MissingInfo( const string& specification, const string& separators = " \t" );
               //
               //                       construct using the specification string 
               //
               //            separators are characters used to delimit tokens in
               //              specification; ':' is not allowed as a separator.
               //
               //                          specification is a string of the form
               //   "op args op args ... end", where ops and args are as follows
               //
               //                                     op                    args
               //                                     --                    ----
               //                                     sizes            (#t) (#z)
               //                                     zlower            (zlower)
               //                                     tlower            (tlower)
               //
               // "sizes #t #z" specifies size of the missing information table.
               // "zlower z" means that slice indices begin with z instead of 0.
               // "tlower t" means that time  indices begin with t instead of 0.
               //      These together are equivalent to one of the constructors.
               //  If present, these specifications take the highest precedance.
               //
               //                                     op                    args
               //                                     --                    ----
               //                                     file           (fname).mri
               //                                     tname              (tname)
               //                                     zname              (zname)
               //                                     data       (datachunkname)
               //                                     missing (missingchunkname)
               //
               //  If used without  "sizes n zlower z tlower t",   missing  info
               //  is read from specified .mri file with specified dimension and
               //  chunk names. Defaults are as follows:
               //
               //                           data chunk name               images
               //                           missing chunk name           missing
               //                           t-dimension (instants) name        t
               //                           z-dimension (slices) name          z
               //
               //  If used with "sizes n zlower z tlower t", missing information
               //  is *updated* from the .mri file only if  t- and z- sizes  and
               //  lower bounds match.
               //
               //                                     op                    args
               //                                     --                    ----
               //                                     +          [n]:[m] [i]:[j]
               //                                     -          [n]:[m] [i]:[j]
               //
               //                           [] denotes optional range arguments.
               //                      + implies missing, - implies non-missing.
               //        specification [n]:[m] for instants, [i]:[j] for slices.
               //               out-of-range or unspecified range specifications
               //                               are set to appropriate defaults.

   string
   reinitialize( const string& specification,
                       bool    updateOnly = true,
                 const string& separators = " \t" );
                      //
                      // if updateOnly == true: update missing information using
                      // specification.   Any "sizes ... zlower ..." is ignored.
                      // for an  "fname ..." specification,  information is read
                      // from the specified .mri file only if appropriate  chunk
                      //     extents match with numInstants( ) and numSlices( ).
                      //
                      // if updateOnly == false:  sizes  and  zLowerBound may be
                      // changed using a "sizes ... zlower ..." or a "fname ..."
                      // specification. If both are specified, "sizes..zlower.."
                      //                       gets precedance over "fname ...".
                      //
                      //                       returns error message; "" if none

  // -------------------------------------------------------- information access
  // ---------------------------------------------------------------------------

   inline
   size_t
   numInstants( ) const;

   inline
   size_t
   numSlices( ) const;

   inline
   size_t
   tLower( ) const;

   inline
   size_t
   tUpper( ) const;

   inline
   size_t
   zLower( ) const;

   inline
   size_t
   zUpper( ) const;

                                            // zLower( ) <= instant <= zUpper( )
                                            // tLower( ) <= instant <= tUpper( )

   inline
   size_t
   numMissingSlices( size_t instant ) const;

   inline
   size_t
   numMissingInstants( size_t slice ) const;


   inline
   bool
   operator()( size_t instant, size_t slice ) const;

   inline
   bool&
   operator()( size_t instant, size_t slice );

   inline
   FlagArray
   missingSlices( size_t instant ) const;

   inline
   FlagArray
   missingInstants( size_t slice ) const;

  // -------------------------------------------------- set/unset missing status
  // ---------------------------------------------------------------------------

   inline
   void
   setStatus( size_t instantMin, size_t instantMax,
              size_t sliceMin,   size_t sliceMax, bool status );
                   //
                   // mark given range of slices/instants as missing/non-missing

  // --------------------------------------------------- time course compression
  // ---------------------------------------------------------------------------

   inline
   void
   compress(       size_t    slice,
             const fpVector& timeCourse,
                   fpVector& compressed,
                   bool      matchIndices = true ) const;
                 //
                 //   output argument compressed contains only those values from
                 // timeCourse that are *not* marked missing for the given slice
                 //
                 // if matchIndices == true,  then timeCourse indices  0 through
                 // numInstants( ) - 1  are considered  identical  to  tLower( )
                 // through tUpper( ). In this case, timeCourse.size( )  must be
                 // >= numInstants( ).
                 //
                 // If matchIndices == false, then compression affects only  the
                 // tLower( ) through tUpper( ) portion of timeCourse.  In  this
                 // case, timeCourse.size( ) must be >= tUpper( ) + 1.

  // -------------------------------------------------------------------- output
  // ---------------------------------------------------------------------------

   string
   write2mri( const string&                    mriFile,
              const string&                    datFile          = "",
                    PghMRIhandler::mriOpenMode openMode
                                                                = PghMRIhandler::mriModify,
              const string&                    missingChunkName = "missing",
                    char                       tDimName         = 't',
                    char                       zDimName         = 'z' ) const;
                                            //
                                            // returns error message; "" if none
                                            //
                                            //        openMode cannot be mriRead

   void
   printInfo(       size_t   instantMin,
                    size_t   instantMax,
                    size_t   sliceMin,
                    size_t   sliceMax,
              const string&  hTitle  = "z",
              const string&  vTitle  = "t",
              const char*    symbols = ".X",
                    ostream& os      = cout ) const;
                      //
                      //  Print out missing slices information in a tabular form
                      //
                      // symbol[true] for missing, symbol[false] for non-missing

   void
   printInfo( ) const;

  private:
   matrix<bool> my_missing;                           // numInstants X numSlices
   size_t       my_zLower, my_tLower;

   string
   updateWith( const string& mriFile,
                     bool    updateOnly   = true,
               const string& dataChunk    = "images",
               const string& missingChunk = "missing",
                     char    tDimName     = 't',
                     char    zDimName     = 'z' );
 };

// ------------------------------------------------------------- construct/reinit
// ------------------------------------------------------------------------------

void
MissingInfo::setStatus( size_t instantMin, size_t instantMax,
                        size_t sliceMin,   size_t sliceMax, bool status )
 {
  // cerr << "# setStatus( " << instantMin << ", " << instantMax << ", " << sliceMin << ", " << sliceMax << ", " << status << " )" << endl;

  size_t sMin = max( my_zLower, min( sliceMin, zUpper( ) ) );
  size_t sMax = max( my_zLower, min( sliceMax, zUpper( ) ) );

  size_t tMin = max( my_tLower, min( instantMin, tUpper( ) ) );
  size_t tMax = max( my_tLower, min( instantMax, tUpper( ) ) );

  // cerr << "# sMin " << sMin << " sMax " << sMax << endl;
  // cerr << "# tMin " << tMin << " tMax " << tMax << endl;

  bool* slicePtr;
  for ( size_t s = sMin, z = s - my_zLower; s <= sMax; s++, z++ )
   {
    slicePtr = my_missing[z];
    for ( size_t t = tMin, i = t - my_tLower; t <= tMax; t++, i++ ) slicePtr[i] = status;
   }

  // cerr << "# setStatus( )" << endl;
 }

// ---------------------------------------------------------- information access
// -----------------------------------------------------------------------------

size_t
MissingInfo::numInstants( ) const
 {
  return my_missing.numRows( );
 }

size_t
MissingInfo::numSlices( ) const
 {
  return my_missing.numCols( );
 }

size_t
MissingInfo::zLower( ) const
 {
  return my_zLower;
 }

size_t
MissingInfo::zUpper( ) const
 {
  size_t n = my_zLower + numSlices( );
  return ( n == 0 ) ? 0 : ( n - 1 );
 }

size_t
MissingInfo::tLower( ) const
 {
  return my_tLower;
 }

size_t
MissingInfo::tUpper( ) const
 {
  size_t n = my_tLower + numInstants( );
  return ( n == 0 ) ? 0 : ( n - 1 );
 }

size_t
MissingInfo::numMissingSlices( size_t instant ) const
 {
  size_t n = 0; instant -= my_tLower;
  for ( size_t s = 0; s < numSlices( ); s++ )
   n += my_missing( instant, s );

  return n;
 }

size_t
MissingInfo::numMissingInstants( size_t slice ) const
 {
  const bool* m = my_missing[slice - my_zLower];

  size_t n = 0;
  for ( size_t i = 0; i < numInstants( ); i++ ) n += m[i];

  return n;
 }

bool&
MissingInfo::operator()( size_t instant, size_t slice )
 {
  return my_missing( instant - my_tLower, slice - my_zLower );
 }

bool
MissingInfo::operator()( size_t instant, size_t slice ) const
 {
  return my_missing( instant - my_tLower, slice - my_zLower );
 }

MissingInfo::FlagArray
MissingInfo::missingSlices( size_t instant ) const
 {
  return my_missing.row( instant - my_tLower );
 }

MissingInfo::FlagArray
MissingInfo::missingInstants( size_t slice ) const
 {
  return my_missing.column( slice - my_zLower );
 }

// ----------------------------------------------------- time course compression
// -----------------------------------------------------------------------------

void
MissingInfo::compress(       size_t    slice,
                       const fpVector& timeCourse,
                             fpVector& compressed,
                             bool      matchIndices ) const
 {
  const bool* m = my_missing[slice - my_zLower];

  if ( matchIndices )
   {
    if ( timeCourse.size( ) < numInstants( ) )
     {
      compressed.resize( 0 );
      return;
     }

    compressed.resize( timeCourse.size( ) - numMissingInstants( slice ) );

    size_t k = 0;

    for ( size_t i = 0; i < numInstants( ); i++ )
     if ( !m[i] ) { compressed[k] = timeCourse[i]; k++; }

    for ( size_t i = numInstants( ); i < timeCourse.size( ); i++, k++ )
     compressed[k] = timeCourse[i];

    return;
   }
  else
   {
    if ( timeCourse.size( ) < tUpper( ) + 1 )
     {
      compressed.resize( 0 );
      return;
     }

    compressed.resize( timeCourse.size( ) - numMissingInstants( slice ) );

    size_t k = 0;

    for ( size_t i = tLower( ), j = 0; i <= tUpper( ); i++, j++ )
     if ( !m[j] ) { compressed[k] = timeCourse[i]; k++; }

    for ( size_t i = tUpper( ) + 1; i < timeCourse.size( ); i++, k++ )
     compressed[k] = timeCourse[i];

    return;
   }
 }

#endif /* __MissingInfo_H__ */
