/* --------------------------------------------------------------------------
 *
 * class BasicObjFn
 *   objective function interface
 *
 * required headers
 *   see include header file BasicObjFn
 *
 * notes
 *   Suggested convention: treat the parameters outside the parameter buffer,
 *   i. e., those with offsets < 0 or >= paramBufSize( ), as constants in the
 *   objective  function  for  minimization  purposes.  Derivatives  are  not
 *   calculated  for  these  parameters.  Use of these parameters in function
 *   evaluation depends on specific functional forms.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __BacisObjFn_H__
#define __BacisObjFn_H__

cpp_using_namespace( BasicTypes );

class BasicObjFn
 {
  public:
   virtual
   ~BasicObjFn( ) = 0;

   void
   setParamBufSize( int paramBufSize );
                                          //
                                          // total number of function parameters

   void
   setSign( fpType sign = +1.0 );
                               //
                               //        overall sign of the objective function;
                               // used only in the with-initialization functions

   void
   setName( const string& name = "some objective function" );
                               //
                               //               name for this objective function

   void
   reinitialize(       int     paramBufSize,
                       fpType  sign = 1.0,
                 const string& name = "some objective function" );

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   int
   paramBufSize( ) const;

   inline
   fpType
   sign( ) const;

   inline
   const string&
   name( ) const;

   virtual
   fpType
   additiveConst( ) const = 0;
                                //
                                // any overall additive constant in the function
                                //  not accounted for by value( ) or value_ni( )
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   // ----------------------- with initialization of output arguments f, Df, DDf
   // --------------------------------------------------------------------------

   inline
   void
   value( const fpType* const paramBuf, fpType* f ) const;

   inline
   void
   gradient( const fpType* const paramBuf, fpType* Df ) const;
                                               //
                                               // Df: buffer of appropriate size

   inline
   void
   hessian( const fpType* const paramBuf, fpMatrix& DDf ) const;
                        //
                        // DDf is resized appropriately if it is resizable; else
                        // it is assumed to be paramBufSize( ) X paramBufSize( )

   // -------------------- without initialization of output arguments f, Df, DDf
   // --------------------------------------------------------------------------

   virtual
   void
   value_ni( const fpType* const paramBuf, fpType* f ) const = 0;

   virtual
   void
   gradient_ni( const fpType* const paramBuf, fpType* Df ) const = 0;
                                           //
                                           // Df: buffer of size paramBufSize( )

   virtual
   void
   hessian_ni( const fpType* const paramBuf, fpMatrix& DDf ) const = 0;
                        //
                        // DDf: matrix of size paramBufSize( ) X paramBufSize( )

  protected:
   BasicObjFn( );

   BasicObjFn( const BasicObjFn& another );

   BasicObjFn&
   operator=( const BasicObjFn& rhs );

   BasicObjFn(       int     paramBufSize,
                     fpType  sign = 1.0,
               const string& name = "some objective function" );

  private:
   int    my_paramBufSize;
   fpType my_sign;
   string my_name;
 };

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

inline
int
BasicObjFn::paramBufSize( ) const
 {
  return my_paramBufSize;
 }

inline
fpType
BasicObjFn::sign( ) const
 {
  return my_sign;
 }

inline
const string&
BasicObjFn::name( ) const
 {
  return my_name;
 }

// --------------------------------------------- objective function computations
// -------------------------- with initialization of output arguments f, Df, DDf
// -----------------------------------------------------------------------------

void
BasicObjFn::value( const fpType* const paramBuf, fpType* f ) const
 {
  // cerr << "BasicObjFn::value( ) *f before update " << (*f) << " f before update " << f << endl;

  (*f) = 0.0;
  value_ni( paramBuf, f );
  (*f) *= my_sign;

  // fpType ff = 0.0;
  // cerr << "BasicObjFn::value( ) ff before update " << ff << " &ff " << &ff << endl;
  // value_ni( paramBuf, &ff );
  // cerr << "BasicObjFn::value( ) ff after update " << ff << endl;
  // (*f) = my_sign * ff;

  // cerr << "BasicObjFn::value( ) *f after update " << (*f) << " f after update " << f << endl;
 }

void
BasicObjFn::gradient( const fpType* const paramBuf, fpType* Df ) const
 {
  for ( int i = 0; i < my_paramBufSize; i++ ) Df[i] = 0.0;

  gradient_ni( paramBuf, Df );

  if ( my_sign != 1.0 )
   for ( int i = 0; i < my_paramBufSize; i++ ) Df[i] *= my_sign;
 }

void
BasicObjFn::hessian( const fpType* const paramBuf, fpMatrix& DDf ) const
 {
  if ( DDf.isResizable( ) )
   DDf.resize( my_paramBufSize, my_paramBufSize, false, 0.0 );
  else
   DDf = 0.0;

  hessian_ni( paramBuf, DDf );

  if ( my_sign != 1.0 ) DDf *= my_sign;
 }

#endif /* __BacisObjFn_H__ */
