/* --------------------------------------------------------------------------
 *
 * class LPflatCS
 *   log ( flat prior with compact support ) - constants
 *   for a set of independent parameters
 *
 * required headers
 *   see include header file LPflatCS
 *
 * notes
 *   follows convention for out-of-range parameter offsets; see BasicObjFn.h
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __LPflatCS_H__
#define __LPflatCS_H__

cpp_using_namespace( BasicTypes );

class LPflatCS : public BasicObjFn
 {
  public:
   // ---------------------------- construct/reinitialize/incremental initialize
   // --------------------------------------------------------------------------

   LPflatCS( );

   LPflatCS( const LPflatCS& another );

   LPflatCS&
   operator=( const LPflatCS& rhs );

   LPflatCS(       int       paramOffset,
                   int       numParams,
             const fpVector& lowerBound,
             const fpVector& upperBound,
             const fpVector& riseFactor,
             const fpVector& fallFactor,
                   int       paramBufSize,
                   fpType    sign = 1.0,
             const string&   name ="log flat prior/compact support" );

                       //                                            assumptions
                       //
                       //     paramOffset is the offset in a parameter buffer of
                       //                         size paramBufSize >= numParams
                       //
                       //   paramOffset >= 0, numParams >= 0, paramBufSize >= 0,
                       //                paramOffset + numParams <= paramBufSize
                       //
                       //                        lowerBound.size( ) >= numParams
                       //                        upperBound.size( ) >= numParams
                       //                        riseFactor.size( ) >= numParams
                       //                        fallFactor.size( ) >= numParams

   void
   setParamOffset( int paramOffset );

   void
   setNumParams( int numParams );

   void
   setConstants( int    paramIndex,
                 fpType lowerBound, fpType upperBound,
                 fpType riseFactor, fpType fallFactor );

   void
   reinitialize(       int       paramOffset,
                       int       numParams,
                 const fpVector& lowerBound,
                 const fpVector& upperBound,
                 const fpVector& riseFactor,
                 const fpVector& fallFactor,
                       int       paramBufSize,
                       fpType    sign = 1.0,
                 const string&   name ="log flat prior/compact support" );

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   int
   numParams( ) const;

   inline
   int
   paramOffset( ) const;

   inline
   fpType
   lowerBound( int paramIndex ) const;

   inline
   fpType
   upperBound( int paramIndex ) const;

   inline
   fpType
   riseFactor( int paramIndex ) const;

   inline
   fpType
   fallFactor( int paramIndex ) const;

   inline
   fpType
   lowerSupport( int paramIndex ) const;

   inline
   fpType
   upperSupport( int paramIndex ) const;

   inline
   fpType
   additiveConst( int paramIndex ) const;
                                //
                                // additive constants arising from normalization
                                //   that are not included in value computation.
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   inline
   fpType
   additiveConst( ) const;
                                //
                                //     overall additive constant in the function
                                //  not accounted for by value( ) or value_ni( )
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   // ------------------------------------------------------- prior computations
   // ----------------- without initialization of output arguments lp, Dlp, DDlp
   // --------------------------------------------------------------------------

   inline
   void
   value_ni( const fpType* const paramBuf, fpType* lp ) const;

   inline
   void
   gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const;

   inline
   void
   hessian_ni( const fpType* const parameters, fpMatrix& DDlp ) const;

  private:
   int      my_paramOffset;
   int      my_numParams;

   fpVector my_lowerBound, my_upperBound, my_riseLength, my_fallLength;
   fpVector my_lowerSupport;

   static
   SplineRampC2 our_ramp;
 };

// ------------------------------------------------------- current configuration
// -----------------------------------------------------------------------------

int
LPflatCS::paramOffset( ) const
 {
  return my_paramOffset;
 }

int
LPflatCS::numParams( ) const
 {
  return my_numParams;
 }

fpType
LPflatCS::lowerBound( int paramIndex ) const
 {
  return my_lowerBound[paramIndex];
 }

fpType
LPflatCS::upperBound( int paramIndex ) const
 {
  return my_upperBound[paramIndex];
 }

fpType
LPflatCS::riseFactor( int paramIndex ) const
 {
  return
     my_riseLength[paramIndex]
   / ( my_upperBound[paramIndex] - my_lowerBound[paramIndex] );
 }

fpType
LPflatCS::fallFactor( int paramIndex ) const
 {
  return
     my_fallLength[paramIndex]
   / ( my_upperBound[paramIndex] - my_lowerBound[paramIndex] );
 }

fpType
LPflatCS::lowerSupport( int paramIndex ) const
 {
  return my_lowerSupport[paramIndex];
 }

fpType
LPflatCS::upperSupport( int paramIndex ) const
 {
  return
   my_upperBound[paramIndex] + my_fallLength[paramIndex];
 }

fpType
LPflatCS::additiveConst( int paramIndex ) const
 {
  return
   log(   our_ramp.integral( )
        * ( my_riseLength[paramIndex] + my_fallLength[paramIndex] )
        + ( my_upperBound[paramIndex] - my_lowerBound[paramIndex] ) );
 }

fpType
LPflatCS::additiveConst( ) const
 {
  fpType c = 0.0;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   c += additiveConst( i );

  return c;
 }

// ---------------------------------------------------------- prior computations
// -------------------- without initialization of output arguments lp, Dlp, DDlp
// -----------------------------------------------------------------------------

void
LPflatCS::value_ni( const fpType* const paramBuf, fpType* lp ) const
 {
  fpType p;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   {
    p = paramBuf[j];

    if ( p < my_lowerBound[i] )
     (*lp) += log( our_ramp( ( p - my_lowerSupport[i] ) / my_riseLength[i] ) );
    else if ( p > my_upperBound[i] )
     (*lp) += log( 1.0 - our_ramp( ( p - my_upperBound[i] ) / my_fallLength[i] ) );
   }
 }

void
LPflatCS::gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const
 {
  fpType p, s, t;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   {
    p = paramBuf[j];

    if ( p < my_lowerBound[i] )
     {
      s = 1.0 / my_riseLength[i];
      t = s * ( p - my_lowerSupport[i] );

      Dlp[j] += s * our_ramp.Dt( t, 1 ) / our_ramp( t );
     }
    else if ( p > my_upperBound[i] )
     {
      s = 1.0 / my_fallLength[i];
      t = s * ( p - my_upperBound[i] );

      Dlp[j] -= s * our_ramp.Dt( t, 1 ) / ( 1.0 - our_ramp( t ) );
     }
   }
 }

void
LPflatCS::hessian_ni( const fpType* const paramBuf, fpMatrix& DDlp ) const
 {
  fpType p, s, t, r0, r1;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   {
    p = paramBuf[j];

    if ( p < my_lowerBound[i] )
     {
      s  = 1.0 / my_riseLength[i];
      t  = s * ( p - my_lowerSupport[i] );
      s *= s;

      r0  = 1.0 / our_ramp( t );
      r1  = our_ramp.Dt( t, 1 );
      r1 *= r1;

      DDlp( j, j ) += s * r0 * ( our_ramp.Dt( t, 2 ) - r1 * r0 );
     }
    else if ( p > my_upperBound[i] )
     {
      s  = 1.0 / my_fallLength[i];
      t  = s * ( p - my_upperBound[i] );
      s *= s;

      r0  = 1.0 / ( 1.0 - our_ramp( t ) );
      r1  = our_ramp.Dt( t, 1 );
      r1 *= r1;

      DDlp( j, j ) -= s * r0 * ( our_ramp.Dt( t, 2 ) + r1 * r0 );
     }
   }
 }

#endif /* __LPflatCS_H__ */
