/* --------------------------------------------------------------------------
 *
 * class LPgamma
 *   log gamma prior - constants == sum_i ( a - 1 ) log x_i - b x_i
 *   for a set of independent parameters x_i
 *
 * required headers
 *   see include header file LPgamma
 *
 * notes
 *   0. follows convention for out-of-range parameter offsets; see BasicObjFn.h
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __LPgamma_H__
#define __LPgamma_H__

cpp_using_namespace( BasicTypes );

class LPgamma : public BasicObjFn
 {
  public:
   // ---------------------------- construct/reinitialize/incremental initialize
   // --------------------------------------------------------------------------

   LPgamma( );

   LPgamma( const LPgamma& another );

   LPgamma&
   operator=( const LPgamma& rhs );

   LPgamma(       int       paramOffset,
                  int       numParams,
            const fpVector& constA,
            const fpVector& constB,
                  int       paramBufSize,
                  fpType    sign = 1.0,
            const string&   name = "log gamma prior" );

                       //                                            assumptions
                       //
                       //     paramOffset is the offset in a parameter buffer of
                       //                         size paramBufSize >= numParams
                       //
                       //   paramOffset >= 0, numParams >= 0, paramBufSize >= 0,
                       //                paramOffset + numParams <= paramBufSize
                       //
                       //                            constA.size( ) >= numParams
                       //                            constB.size( ) >= numParams

   void
   setParamOffset( int paramOffset );

   void
   setNumParams( int numParams );

   void
   setConstants( int paramIndex, fpType constA, fpType constB );

   void
   reinitialize(       int       paramOffset,
                       int       numParams,
                 const fpVector& constA,
                 const fpVector& constB,
                       int       paramBufSize,
                       fpType    sign = 1.0,
                 const string&   name = "log gamma prior" );

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   int
   paramOffset( ) const;

   inline
   int
   numParams( ) const;

   inline
   fpType
   constA( int paramIndex ) const;

   inline
   fpType
   constB( int paramIndex ) const;

   inline
   fpType
   additiveConst( int paramIndex ) const;
                                //
                                // additive constants arising from normalization
                                //   that are not included in value computation.
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   inline
   fpType
   additiveConst( ) const;
                                //
                                //     overall additive constant in the function
                                //  not accounted for by value( ) or value_ni( )
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   // ------------------------------------------------------- prior computations
   // ----------------- without initialization of output arguments lp, Dlp, DDlp
   // --------------------------------------------------------------------------

   inline
   void
   value_ni( const fpType* const paramBuf, fpType* lp ) const;

   inline
   void
   gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const;

   inline
   void
   hessian_ni( const fpType* const parameters, fpMatrix& DDlp ) const;

  private:
   int      my_paramOffset;
   int      my_numParams;
   fpVector my_constA, my_constB;
 };

// ------------------------------------------------------- current configuration
// -----------------------------------------------------------------------------

int
LPgamma::paramOffset( ) const
 {
  return my_paramOffset;
 }

int
LPgamma::numParams( ) const
 {
  return my_numParams;
 }

fpType
LPgamma::constA( int paramIndex ) const
 {
  return my_constA[paramIndex] + 1.0;
 }

fpType
LPgamma::constB( int paramIndex ) const
 {
  return my_constB[paramIndex];
 }

fpType
LPgamma::additiveConst( int paramIndex ) const
 {
  fpType a = constA( paramIndex );

  return
   a * log( constB( paramIndex ) ) - lgamma( a );
 }

fpType
LPgamma::additiveConst( ) const
 {
  fpType c = 0.0;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   c += additiveConst( i );

  return c;
 }

// ---------------------------------------------------------- prior computations
// -------------------- without initialization of output arguments lp, Dlp, DDlp
// -----------------------------------------------------------------------------

void
LPgamma::value_ni( const fpType* const paramBuf, fpType* lp ) const
 {
  fpType p, a, f;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   {
    p = paramBuf[j];
    a = my_constA[i];
    f = ( a > 0.0 ) ? a * log( p ) : 0.0;

    (*lp) += ( f - my_constB[i] * p );
   }
 }

void
LPgamma::gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const
 {
  fpType a, f;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   {
    a = my_constA[i];
    f = ( a > 0.0 ) ? a / paramBuf[j] : 0.0;

    Dlp[j] += ( f - my_constB[i] );
   }
 }

void
LPgamma::hessian_ni( const fpType* const paramBuf, fpMatrix& DDlp ) const
 {
  fpType p, a, f;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   {
    p = 1.0 / paramBuf[j];
    a = my_constA[i];
    f = ( a > 0.0 ) ? a * p * p : 0.0;

    DDlp( j, j ) -= f;
   }
 }

#endif /* __LPgamma_H__ */
