/* --------------------------------------------------------------------------
 *
 * class LPnormal
 *   log normal prior - constants == 0.5 * sum_i ( x_i - mu_i )^2 / sigma_i^2
 *   for a set of independent parameters x_i
 *
 * required headers
 *   see include header file LPnormal
 *
 * notes
 *   follows convention for out-of-range parameter offsets; see BasicObjFn.h
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __LPnormal_H__
#define __LPnormal_H__

cpp_using_namespace( BasicTypes );

class LPnormal : public BasicObjFn
 {
  public:
   // ---------------------------- construct/reinitialize/incremental initialize
   // --------------------------------------------------------------------------

   LPnormal( );

   LPnormal( const LPnormal& another );

   LPnormal&
   operator=( const LPnormal& rhs );

   LPnormal(       int       paramOffset,
                   int       numParams,
             const fpVector& mean,
             const fpVector& stdev,
                   int       paramBufSize,
                   fpType    sign = 1.0,
             const string&   name = "log normal prior" );

                       //                                            assumptions
                       //
                       //     paramOffset is the offset in a parameter buffer of
                       //                         size paramBufSize >= numParams
                       //
                       //   paramOffset >= 0, numParams >= 0, paramBufSize >= 0,
                       //                paramOffset + numParams <= paramBufSize
                       //
                       //                              mean.size( ) >= numParams
                       //                             stdev.size( ) >= numParams

   void
   setParamOffset( int paramOffset );

   void
   setNumParams( int numParams );

   void
   setConstants( int paramIndex, fpType mean, fpType stdev );

   void
   reinitialize(       int       paramOffset,
                       int       numParams,
                 const fpVector& mean,
                 const fpVector& stdev,
                       int       paramBufSize,
                       fpType    sign = 1.0,
                 const string&   name = "log normal prior" );
                       //
                       //              assumption: mean.size( ) == stdev.size( )
                       //             == # of parameters described by this prior
                       //
                       //     paramOffset is the offset in a parameter buffer of
                       //                         size paramBufSize >= numParams

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   int
   paramOffset( ) const;

   inline
   int
   numParams( ) const;

   inline
   fpType
   mean( int paramIndex ) const;

   inline
   fpType
   stdev( int paramIndex ) const;

   inline
   fpType
   additiveConst( int paramIndex ) const;
                                //
                                // additive constants arising from normalization
                                //   that are not included in value computation.
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   inline
   fpType
   additiveConst( ) const;
                                //
                                //     overall additive constant in the function
                                //  not accounted for by value( ) or value_ni( )
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   // ------------------------------------------------------- prior computations
   // ----------------- without initialization of output arguments lp, Dlp, DDlp
   // --------------------------------------------------------------------------

   inline
   void
   value_ni( const fpType* const paramBuf, fpType* lp ) const;

   inline
   void
   gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const;

   inline
   void
   hessian_ni( const fpType* const parameters, fpMatrix& DDlp ) const;

  private:
   int      my_paramOffset;
   int      my_numParams;
   fpVector my_mean, my_precision;
 };

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

int
LPnormal::paramOffset( ) const
 {
  return my_paramOffset;
 }

int
LPnormal::numParams( ) const
 {
  return my_numParams;
 }

fpType
LPnormal::mean( int paramIndex ) const
 {
  return my_mean[paramIndex];
 }

fpType
LPnormal::stdev( int paramIndex ) const
 {
  return 1.0 / my_precision[paramIndex];
 }

fpType
LPnormal::additiveConst( int paramIndex ) const
 {
  return 0.5 * ( log( my_precision[paramIndex] ) - log( 2.0 * M_PI ) );
 }

fpType
LPnormal::additiveConst( ) const
 {
  fpType c = 0.0;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   c += additiveConst( i );

  return c;
 }

// ---------------------------------------------------------- prior computations
// -------------------- without initialization of output arguments lp, Dlp, DDlp
// -----------------------------------------------------------------------------

void
LPnormal::value_ni( const fpType* const paramBuf, fpType* lp ) const
 {
  fpType p;

  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   {
    p = ( paramBuf[j] - my_mean[i] ) * my_precision[i];
    (*lp) -= 0.5 * p * p;
   }
 }

void
LPnormal::gradient_ni( const fpType* const paramBuf, fpType* Dlp ) const
 {
  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   Dlp[j] -= ( paramBuf[j] - my_mean[i] ) * my_precision[i] * my_precision[i];
 }

void
LPnormal::hessian_ni( const fpType* const paramBuf, fpMatrix& DDlp ) const
 {
  int ofsMax = min( my_paramOffset + my_numParams, paramBufSize( ) );
  for ( int j = max( 0, my_paramOffset ),
            i = j - my_paramOffset;
            j < ofsMax; j++, i++ )
   DDlp( j, j ) -= my_precision[i] * my_precision[i];
 }

#endif /* __LPnormal_H__ */
