/* --------------------------------------------------------------------------
 *
 * class ObjFnManager
 *   interface for handling objective functions
 *   with a number of additive terms
 *
 * required headers
 *   see include header file BasicObjFn
 *
 * --------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __ObjFnManager_H__
#define __ObjFnManager_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class ObjFnManager : public BasicObjFn
 {
  public:
   typedef list<const BasicObjFn*>   ObjFnList;
   typedef ObjFnList::const_iterator const_iterator;

   // --------------------------------------------------- construct/reinitialize
   // --------------------------------------------------------------------------

   ObjFnManager( );

   ObjFnManager( const ObjFnManager& another );

   ObjFnManager&
   operator=( const ObjFnManager& rhs );

   ObjFnManager(       int     paramBufSize,
                       fpType  sign = 1.0,
                 const string& name = "objective function/additive terms" );

   void
   reinitialize(       int     paramBufSize,
                       fpType  sign = 1.0,
                 const string& name = "objective function/additive terms" );

                                      //                              assumption
                                      //
                                      // paramBufSize for all objective function
                                      //  terms match with paramBufSize set here

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   fpType
   additiveConst( ) const;
                                //
                                // any overall additive constant in the function
                                //  not accounted for by value( ) or value_ni( )
                                //
                                // sign (as set by setSign( ) or constructor) is
                                //             not included in additiveConst( ).

   // ------------------------------------------------- objective function terms
   // --------------------------------------------------------------------------

   inline
   void
   add( const BasicObjFn* term );
                              //
                              // add a named objective function term to the list
                              //                                unless term == 0
                              //                    or it has already been added

   inline
   void
   del( const BasicObjFn* term );
                              //
                              //    delete all occurrances of term from the list

   inline
   void
   delAll( );
                              //
                              //             clear objective function terms list

   inline
   void
   delIfNameIs( const string& name );
                              //
                              //   delete all occurrances of the named objective
                              //                                   function term

   inline
   void
   delIfNameHas( const string& name );
                              //
                              //  delete all objective function terms whose name
                              //                                   contains name

   inline
   bool
   has( const string& name ) const;
                              //
                              //    true if named objective function term exists

   inline
   bool
   has( const BasicObjFn* term ) const;
                              //
                              //    true if given objective function term exists

                              //              for more elaborate list operations
                              //
   inline
   const ObjFnList&
   termList( ) const;

   /* --------------------- subsumed by predicate/predicates
   class nameIs : public unary_function< const BasicObjFn*, bool >
    {
     public:
      explicit
      nameIs( const string& name )
       : my_name( name )
       {
       }

      bool
      operator()( const BasicObjFn* term ) const
       {
        return ( term ? ( term->name( ) == my_name ) : false );
       }

     private:
      const string& my_name;
    };

   class nameHas : public unary_function< const BasicObjFn*, bool >
    {
     public:
      explicit
      nameHas( const string& name )
       : my_name( name )
       {
       }

      bool
      operator()( const BasicObjFn* term ) const
       {
        return ( term ? ( term->name( ).find( my_name ) != string::npos ) : false );
       }

     private:
      const string& my_name;
    };
   --------------------- subsumed by predicate/predicates */

   // ------------------------------------------ objective function computations
   // -------------------- without initialization of output arguments f, Df, DDf
   // --------------------------------------------------------------------------

   inline
   void
   value_ni( const fpType* const paramBuf, fpType* f ) const;

   inline
   void
   gradient_ni( const fpType* const paramBuf, fpType* Df ) const;

   inline
   void
   hessian_ni( const fpType* const paramBuf, fpMatrix& DDlp ) const;

  private:
   ObjFnList my_termList;
 };

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

fpType
ObjFnManager::additiveConst( ) const
 {
  fpType c = 0.0;
  for ( const_iterator ti  = my_termList.begin( );
                       ti != my_termList.end( ); ++ti )
   c += (*ti)->additiveConst( );
  return c;
 }

// ---------------------------------------------------- objective function terms
// -----------------------------------------------------------------------------

void
ObjFnManager::add( const BasicObjFn* term )
 {
  if (    ( term != 0 )
       && (    find( my_termList.begin( ), my_termList.end( ), term )
            == my_termList.end( ) ) )
   my_termList.push_back( term );
 }

void
ObjFnManager::del( const BasicObjFn* term )
 {
  my_termList.remove( term );
 }

void
ObjFnManager::delAll( )
 {
  my_termList.clear( );
 }

void
ObjFnManager::delIfNameIs( const string& name )
 {
  my_termList.remove_if( nameIs<BasicObjFn>( name ) );
 }

void
ObjFnManager::delIfNameHas( const string& name )
 {
  my_termList.remove_if( nameHas<BasicObjFn>( name ) );
 }

bool
ObjFnManager::has( const string& name ) const
 {
  return
   find_if( my_termList.begin( ), my_termList.end( ), nameIs<BasicObjFn>( name ) )
   != my_termList.end( );
 }

bool
ObjFnManager::has( const BasicObjFn* term ) const
 {
  return
   find( my_termList.begin( ), my_termList.end( ), term ) != my_termList.end( );
 }

const ObjFnManager::ObjFnList&
ObjFnManager::termList( ) const
 {
  return my_termList;
 }

// --------------------------------------------- objective function computations
// ----------------------- without initialization of output arguments f, Df, DDf
// -----------------------------------------------------------------------------

void
ObjFnManager::value_ni( const fpType* const paramBuf, fpType* f ) const
 {
  for ( const_iterator ti  = my_termList.begin( );
                       ti != my_termList.end( ); ++ti )
   (*ti)->value_ni( paramBuf, f );
 }

void
ObjFnManager::gradient_ni( const fpType* const paramBuf, fpType* Df ) const
 {
  for ( const_iterator ti  = my_termList.begin( );
                       ti != my_termList.end( ); ++ti )
   (*ti)->gradient_ni( paramBuf, Df );
 }

void
ObjFnManager::hessian_ni( const fpType* const paramBuf, fpMatrix& DDf ) const
 {
  for ( const_iterator ti  = my_termList.begin( );
                       ti != my_termList.end( ); ++ti )
   (*ti)->hessian_ni( paramBuf, DDf );
 }

#endif /* __ObjFnManager_H__ */
