/* --------------------------------------------------------------------------
 *
 * class PghMRIhandler
 *   C++ interface to some of the libmri functions; see details below.
 *   For details about libmri, see mri.h, mri-pgh.doc, mri-c.doc.
 *
 * required headers
 *   see include header file PghMRIhandler
 *
 * notes
 *   0. errorMode is global: if there are several coexisting PghMRIhandler
 *      instances, setting errorMode for one would affect all.
 *   1. It would be nice if the original C library supported another version
 *      of mri_get_chunk which read the requested data into a user-supplied
 *      buffer instead of allocating one on its own. The present C++ interface
 *      is intended to be something slightly better than 'extern "C" { ... }';
 *      it does not attempt to improve upon the original libmri in such
 *      respects.
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __PghMRIhandler_H__
#define __PghMRIhandler_H__

cpp_using_namespace( std );

class PghMRIhandler
 {
  public:

             // This is a wrapper class for libmri functions in mri.h. low-level
             // typedefs, #defines, struct definitions, etc., that do not appear
             // in any of the user functions have been omitted here.

   // ---------------------------------------------------- library configuration
   // --------------------------------------------------------------------------

   static const int mriMaxDims            = MRI_MAX_DIMS;
   static const int mriMaxKeyLength       = MRI_MAX_KEY_LENGTH;
   static const int mriMaxValueLength     = MRI_MAX_VALUE_LENGTH;
   static const int mriMaxChunks          = MRI_MAX_CHUNKS;
   static const int mriMaxFilenameLength  = MRI_MAX_FILENAME_LENGTH;
   static const int mriMaxOpenFiles       = MRI_MAX_OPEN_FILES;
   static const int mriSafeBufferCount    = MRI_SAFE_BUFFER_COUNT;
   static const int mriMaxBufferCount     = MRI_MAX_BUFFER_COUNT;
   static const int mriAlignmentThreshold = MRI_ALIGNMENT_THRESHOLD;
   static const int mriAlignmentBoundary  = MRI_ALIGNMENT_BOUNDARY;

   // ------------------------------------------------ type conversion constants
   // --------------------------------------------------------------------------

   typedef MRI_ArrayType mriArrayType;

   static const mriArrayType mriRaw;
   static const mriArrayType mriUchar;
   static const mriArrayType mriShort;
   static const mriArrayType mriInt;
   static const mriArrayType mriFloat;
   static const mriArrayType mriDouble;
   static const mriArrayType mriLong;
   static const mriArrayType mriLongLong;

   static const mriArrayType mriScalar;
   static const mriArrayType mriComplex;
   static const mriArrayType mriVector;

   static const mriArrayType mriComplexShort;
   static const mriArrayType mriComplexInt;
   static const mriArrayType mriComplexFloat;
   static const mriArrayType mriComplexDouble;
   static const mriArrayType mriComplexLong;
   static const mriArrayType mriComplexLongLong;

   static const mriArrayType mriVectorShort;
   static const mriArrayType mriVectorInt;
   static const mriArrayType mriVectorFloat;
   static const mriArrayType mriVectorDouble;
   static const mriArrayType mriVectorLong;
   static const mriArrayType mriVectorLongLong;

   // ---------------------------------------------------------- file open modes
   // --------------------------------------------------------------------------

   typedef MRI_OpenMode mriOpenMode;

   static const mriOpenMode mriRead;
   static const mriOpenMode mriWrite;
   static const mriOpenMode mriModify;
   static const mriOpenMode mriModifyData;

   // ----------------------------------------------------------- error handling
   // --------------------------------------------------------------------------

   typedef MRI_Error_Handling mriErrorMode;

   static const mriErrorMode mriAbortOnError;
   static const mriErrorMode mriReportErrors;
   static const mriErrorMode mriIgnoreErrors;

   // -------------------------------------------------------------- mri dataset
   // --------------------------------------------------------------------------

   typedef MRI_Dataset mriDataset;

   // ------------------------------------- cctor, dtor; explicit open and close
   // --------------------------------------------------------------------------

   PghMRIhandler( const string&      fileName,
                        mriOpenMode  openMode  = mriRead,
                        mriErrorMode errorMode = mriAbortOnError );
                       //
                       // if openMode == mriWrite, the basic mri key-value pairs
                       // "!format = pgh" and "!version = 1.0" are automatically
                       //                                      added to the file

   ~PghMRIhandler( );

                           // copy and default constructors are provided for the
                           //  sole purpose of creating arrays of PghMRIhandlers

   PghMRIhandler( );

   PghMRIhandler( const PghMRIhandler& another );

                                                 // copy with openMode = mriRead

   PghMRIhandler&
   operator=( const PghMRIhandler& rhs );

   void
   open( const string&      fileName,
               mriOpenMode  openMode  = mriRead,
               mriErrorMode errorMode = mriAbortOnError );
                       //
                       // if openMode == mriWrite, the basic mri key-value pairs
                       // "!format = pgh" and "!version = 1.0" are automatically
                       //                                      added to the file

   void
   close( );

   // ------------------------------------------------------------- copy dataset
   // --------------------------------------------------------------------------

   inline
   void
   copyTo( const string& destination ) const;

   // --------------------------------------------------------------- attributes
   // --------------------------------------------------------------------------

   inline
   static
   int
   endianness( );
                                   //
                                   // 0 if big, 1 if little, -1 if indeterminate

   inline
   bool
   isOpen( ) const;

   inline
   const string&
   fileName( ) const;

   inline
   mriOpenMode
   openMode( ) const;

   inline
   mriErrorMode
   errorMode( ) const;

   // ----------------------------------------------------------- error handling
   // --------------------------------------------------------------------------

   static
   void
   setErrorMode( mriErrorMode mode );
                                       //
                                       // unfortunately, setErrorMode( ) affects
                                       // all coexisting PghMRIhandler instances

   inline
   bool
   anyError( ) const;
                             //
                             // inoperative when errorMode( ) == mriAbortOnError

   inline
   string
   error( ) const;
                             //
                             // inoperative when errorMode( ) == mriAbortOnError

   // ---------------------------------------------------------- keys and values
   // --------------------------------------------------------------------------

   inline
   int
   getInt( const string& key ) const;

   inline
   double
   getFloat( const string& key ) const;

   inline
   string
   getString( const string& key ) const;

   inline
   void
   setInt( const string& key, int value );

   inline
   void
   setFloat( const string& key, float value );

   inline
   void
   setString( const string& key, const string& value );

   inline
   bool
   hasKey( const string& key ) const;

   inline
   void
   remove( const string& key );

   inline
   void
   iterateOverKeys( ) const;

   inline
   string
   nextKey( ) const;

   // ------------------------------------------------------- chunks and buffers
   // --------------------------------------------------------------------------

   inline
   void
   createChunk( const string& key );

   inline
   void
   updateChunk( const string& key );

   inline
   void*
   readChunk( const string& key, int size, int offset, mriArrayType type, void* buffer ) const;
                            //
                            // copies requested chunk in buffer; returns buffer.
                            // This is different than the current libmri version
                            // of mri_read_chunk( ) which merely uses buffer for
                            //                                temporary storage.

   inline
   void*
   getChunk( const string& key, int size, int offset, mriArrayType type ) const;

   inline
   void
   setChunk( const string& key, int size, int offset, mriArrayType type, void* bufferPtr );

   inline
   void
   retainBuffer( void* bufferPtr ) const;

   inline
   void
   discardBuffer( void* bufferPtr ) const;

   // ------------------------------------------------------------------- images
   // --------------------------------------------------------------------------

   inline
   void*
   getImage( const int time, const int slice, mriArrayType type ) const;

   inline
   void
   setImage( const int time, const int slice, mriArrayType type, void* image );

   // --------------------------------------- miscellaneous non-libmri utilities
   // --------------------------------------------------------------------------

   static
   inline
   string
   type2name( mriArrayType mriType, string defaultTypeName = "" );
                    //
                    // defaultTypeName is returned for invalid values of mriType

   static
   inline
   mriArrayType
   name2type( const string& mriTypeName, mriArrayType defaultType = -1 );
                    //
                    // defaultType is returned for invalid values of mriTypeName
                    // valid values are  "uint8",  "int16",  "int32", "float32",
                    // "float64".

   static
   inline
   mriArrayType
   type2type( mriArrayType mriType, mriArrayType defaultType = mriFloat );
                           //
                           // return defaultType for an invalid value of mriType

   static
   inline
   string
   mode2name( mriOpenMode mode, string defaultModeName = "" );
                         //
                         // defaultModeName is returned for values of mode other
                         // than  mriRead,  mriWrite,  mriModify, mriModifyData.

   static
   inline
   mriOpenMode
   name2mode( const string& mriModeName, mriOpenMode defaultMode = -1 );
                   //
                   // defaultMode is returned for invalid values of mriModeName.
                   // valid mode names are "mriRead",  "mriWrite",  "mriModify",
                   // "mriModifyData".

   static
   inline
   size_t
   sizeOf( const string& mriTypeName );

   static
   inline
   size_t
   sizeOf( mriArrayType type );

   inline
   long
   availableChunkOffset( ) const;
                           //
                           //      next available chunk offset calculated as the
                           // largest existing chunk offset + size of that chunk

   static
   inline
   string
   availableDimNames( const string& toExclude = "" );
                             //
                             // available dimension names (a subset of printable
                             // ascii characters)  excluding those in toExclude.

   inline
   void
   createChunk( const string& name,
                const string& dimensions,
                const string& type,
                const string& datafile,
                const int*    extents,
                const int*    indexLowerBounds = 0,
                      bool    littleEndian     = endianness( ),
                      long    offset           = 0 );
                                          //
                                          // everything needed to create a chunk

  inline
  static
  bool
  matchChunk( const PghMRIhandler& mriHandler,
              const string& chunkName, const string& chunkType,
	      const string& dimensions, const int* sizes, const int* lower );
                               //
                               // true if chunk exists in mriHandler and all its
                               // essential attributes  match the specifications

  private:
   string      my_fileName;
   mriOpenMode my_openMode;
   mriDataset* my_dataset;

   mutable char*        my_error;
   static  mriErrorMode our_errorMode;

   static  char         our_fileNotOpenMsg[];

   inline
   void
   noOpenFileErrorHandler( ) const;
 };

// ---------------------------------------------------------------- copy dataset
// -----------------------------------------------------------------------------

void
PghMRIhandler::copyTo( const string& destination ) const
 {
  if ( my_dataset )
   mri_close_dataset( mri_copy_dataset( destination.c_str( ), my_dataset ) );
  else
   noOpenFileErrorHandler( );
 }

// ------------------------------------------------------------------ attributes
// -----------------------------------------------------------------------------

int
PghMRIhandler::endianness( )
 {
     // this is a slightly compressed version of InitBIO() from fiasco libbio.c.
     // calling  InitBIO()  might have been an easier solution (libbio is linked
     // along with libmri anyways) but for the difficulties in  including bio.h.
     // see ~fiasco/Fiasco5.1/src/util/libbio.c::InitBIO() for details.

  if ( !(    ( sizeof( char )      <  sizeof( int ) )
          && ( sizeof( short )     >= 2 )
          && ( sizeof( int )       >= 4 )
          && ( sizeof( long long ) >= 8 )
          && ( sizeof( float )     == 4 )
          && ( sizeof( double )    == 8 )
          && ( ( sizeof( long ) == 4 ) || ( sizeof( long ) == 8 ) ) ) ) return -1;
                   //
                   // compiler uses nonstandard lengths for some variable types?

  union { long l; unsigned char c[sizeof( long )]; } t;
  t.l = 1;

  int e = -1;
  if ( t.c[0] == 1 )
   e = 1;
  else if ( t.c[sizeof( long ) - 1] == 1 )
   e = 0;

  return e;
 }

bool
PghMRIhandler::isOpen( ) const
 {
  return my_dataset;
 }

const string&
PghMRIhandler::fileName( ) const
 {
  return my_fileName;
 }

PghMRIhandler::mriOpenMode
PghMRIhandler::openMode( ) const
 {
  return my_openMode;
 }

PghMRIhandler::mriErrorMode
PghMRIhandler::errorMode( ) const
 {
  return our_errorMode;
 }

// -------------------------------------------------------------- error handling
// -----------------------------------------------------------------------------

bool
PghMRIhandler::anyError( ) const
 {
  return ( my_error != 0 );
 }

string
PghMRIhandler::error( ) const
 {
  return ( my_error == 0 ) ? string( ) : string( my_error );
 }

// ------------------------------------------------------------- keys and values
// -----------------------------------------------------------------------------

int
PghMRIhandler::getInt( const string& key ) const
 {
  if ( my_dataset )
   {
    int value = mri_get_int( my_dataset, key.c_str( ) );
    my_error = mri_error;
    return value;
   }
  else
   {
    noOpenFileErrorHandler( );
    return 0;
   }
 }

double
PghMRIhandler::getFloat( const string& key ) const
 {
  if ( my_dataset )
   {
    double value = mri_get_float( my_dataset, key.c_str( ) );
    my_error = mri_error;
    return value;
   }
  else
   {
    noOpenFileErrorHandler( );
    return 0.0;
   }
 }

string
PghMRIhandler::getString( const string& key ) const
 {
  if ( my_dataset )
   {
    char* value = mri_get_string( my_dataset, key.c_str( ) );
    my_error = mri_error;
    return ( value == 0 ) ? string( ) : string( value );
   }
  else
   {
    noOpenFileErrorHandler( );
    return string( );
   }
 }

void
PghMRIhandler::setInt( const string& key, int value )
 {
  if ( my_dataset )
   {
    mri_set_int( my_dataset, key.c_str( ), value );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

void
PghMRIhandler::setFloat( const string& key, float value )
 {
  if ( my_dataset )
   {
    mri_set_float( my_dataset, key.c_str( ), value );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

void
PghMRIhandler::setString( const string& key, const string& value )
 {
  if ( my_dataset )
   {
    mri_set_string( my_dataset, key.c_str( ), value.c_str( ) );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

bool
PghMRIhandler::hasKey( const string& key ) const
 {
  if ( my_dataset )
   return mri_has( my_dataset, key.c_str( ) );
  else
   {
    noOpenFileErrorHandler( );
    return false;
   }
 }

void
PghMRIhandler::remove( const string& key )
 {
  if ( my_dataset )
   {
    mri_remove( my_dataset, key.c_str( ) );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

void
PghMRIhandler::iterateOverKeys( ) const
 {
  if ( my_dataset )
   {
    mri_iterate_over_keys( my_dataset );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

string
PghMRIhandler::nextKey( ) const
 {
  if ( my_dataset )
   {
    char* key = mri_next_key( my_dataset );
    my_error = mri_error;
    return ( key == 0 ) ? string( ) : string( key );
   }
  else
   {
    noOpenFileErrorHandler( );
    return string( );
   }
 }

// ---------------------------------------------------------- chunks and buffers
// -----------------------------------------------------------------------------

void
PghMRIhandler::createChunk( const string& key )
 {
  if ( my_dataset )
   {
    mri_create_chunk( my_dataset, key.c_str( ) );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

void
PghMRIhandler::updateChunk( const string& key )
 {
  if ( my_dataset )
   {
    mri_update_chunk( my_dataset, key.c_str( ) );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

void*
PghMRIhandler::readChunk( const string& key, int size, int offset, mriArrayType type,
                          void* buffer ) const
 {
  if ( my_dataset )
   {
    // void* chunk = mri_read_chunk( my_dataset, key.c_str( ), size, offset, type, buffer );
    // my_error = mri_error;
             //
             // mri_read_chunk( ) does something funny with the buffer provided.
             // instead of returning the results into this buffer,  it only uses
             // it for temporary storage; the real results are returned in a new
             // buffer it creates. Seems like Joel misunderstood my requirements
             // for read_chunk( ). The following assertion always fails when the
             // data being read is of a different type than type.
             //
             // assert ( chunk == buffer );
             // return chunk;

    void* chunk = mri_get_chunk( my_dataset, key.c_str( ), size, offset, type );
    my_error = mri_error;
    memcpy( buffer, chunk, sizeOf( type ) * static_cast<size_t>( size ) );
                   //
                   // substitute code for readChunk based  on  mri_get_chunk( ).
                   // check revisions of libmri for whether this has been fixed.

    return buffer;
   }
  else
   {
    noOpenFileErrorHandler( );
    return 0;
   }
 }

void*
PghMRIhandler::getChunk( const string& key, int size, int offset, mriArrayType type ) const
 {
  if ( my_dataset )
   {
    void* chunk = mri_get_chunk( my_dataset, key.c_str( ), size, offset, type );
    my_error = mri_error;
    return chunk;
   }
  else
   {
    noOpenFileErrorHandler( );
    return 0;
   }
 }

void
PghMRIhandler::setChunk( const string& key, int size, int offset, mriArrayType type, void* bufferPtr )
 {
  if ( my_dataset )
   {
    mri_set_chunk( my_dataset, key.c_str( ), size, offset, type, bufferPtr );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

void
PghMRIhandler::retainBuffer( void* bufferPtr ) const
 {
  if ( my_dataset )
   {
    mri_retain_buffer( my_dataset, bufferPtr );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

void
PghMRIhandler::discardBuffer( void* bufferPtr ) const
 {
  if ( my_dataset )
   {
    mri_discard_buffer( my_dataset, bufferPtr );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

// ---------------------------------------------------------------------- images
// -----------------------------------------------------------------------------

void*
PghMRIhandler::getImage( const int time, const int slice, mriArrayType type ) const
 {
  if ( my_dataset )
   {
    void* image = mri_get_image( my_dataset, time, slice, type );
    my_error = mri_error;
    return image;
   }
  else
   {
    noOpenFileErrorHandler( );
    return 0;
   }
 }

void
PghMRIhandler::setImage( const int time, const int slice, mriArrayType type, void* image )
 {
  if ( my_dataset )
   {
    mri_set_image( my_dataset, time, slice, type, image );
    my_error = mri_error;
   }
  else
   noOpenFileErrorHandler( );
 }

// ------------------------------------------ miscellaneous non-libmri utilities
// -----------------------------------------------------------------------------

string
PghMRIhandler::type2name( mriArrayType type, string defaultTypeName )
 {
  if ( type == mriUchar )
   return "uint8";
  if ( type == mriShort )
   return "int16";
  if ( type == mriInt )
   return "int32";
  if ( type == mriFloat )
   return "float32";
  if ( type == mriDouble )
   return "float64";
  else
   return defaultTypeName;
 }

PghMRIhandler::mriArrayType
PghMRIhandler::name2type( const string& mriTypeName, mriArrayType defaultType )
 {
  if ( mriTypeName == "uint8" )
   return mriUchar;
  else if ( mriTypeName == "int16" )
   return mriShort;
  else if ( mriTypeName == "int32" )
   return mriInt;
  else if ( mriTypeName == "float32" )
   return mriFloat;
  else if ( mriTypeName == "float64" )
   return mriDouble;
  else
   return defaultType;
 }

PghMRIhandler::mriArrayType
PghMRIhandler::type2type( mriArrayType type, mriArrayType defaultType )
 {
  return (    ( type == mriUchar ) || ( type == mriShort )
           || ( type == mriInt   ) || ( type == mriFloat ) || ( type == mriDouble ) ) ? type : defaultType;
 }

string
PghMRIhandler::mode2name( mriOpenMode mode, string defaultModeName )
 {
  if ( mode == mriRead )
   return "mriRead";
  if ( mode == mriWrite )
   return "mriWrite";
  if ( mode == mriModify )
   return "mriModify";
  if ( mode == mriModifyData )
   return "mriModifyData";
  else
   return defaultModeName;
 }

PghMRIhandler::mriOpenMode
PghMRIhandler::name2mode( const string& mriModeName, mriOpenMode defaultMode )
 {
  if ( mriModeName == "mriRead" )
   return mriRead;
  if ( mriModeName == "mriWrite" )
   return mriWrite;
  if ( mriModeName == "mriModify" )
   return mriModify;
  if ( mriModeName == "mriModifyData" )
   return mriModifyData;
  else
   return defaultMode;
 }

size_t
PghMRIhandler::sizeOf( const string& mriType )
 {
  if ( mriType == "uint8" )
   return sizeof( unsigned char );
  else if ( mriType == "int16" )
   return sizeof( short );
  else if ( mriType == "int32" )
   return sizeof( int );
  else if ( mriType == "float32" )
   return sizeof( float );
  else if ( mriType == "float64" )
   return sizeof( double );
  else
   return 0;
 }

size_t
PghMRIhandler::sizeOf( mriArrayType type )
 {
  if ( type == mriUchar )
   return sizeof( unsigned char );
  else if ( type == mriShort )
   return sizeof( short );
  else if ( type == mriInt )
   return sizeof( int );
  else if ( type == mriFloat )
   return sizeof( float );
  else if ( type == mriDouble )
   return sizeof( double );
  else
   return 0;
 }

long
PghMRIhandler::availableChunkOffset( ) const
 {
  long offset = 0;
  bool ofsInitialized = false;

  string chunkName = "";

  iterateOverKeys( );
  string key = nextKey( );

  long ofs;

  while ( key != "" )
   {
    if ( getString( key ) == "[chunk]" )
     {
      if ( !ofsInitialized )
       {
        chunkName = key;
        offset    = getInt( chunkName + ".offset" );

        ofsInitialized = true;
       }

      ofs = getInt( chunkName + ".offset" );

      if ( ofs > offset )
       {
        chunkName = key;
        offset    = ofs;
       }
     }

    key = nextKey( );
   }

  size_t size = 0;
  if ( hasKey( chunkName + ".size" ) )
   size = getInt( chunkName + ".size" );

  return offset + size;
 }

string
PghMRIhandler::availableDimNames( const string& toExclude )
 {
  // ATTN any value allowed for char seems to be allowed by libmri as a dimension name.

  string dims = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789=+-*/<>()[]{}:;,.!?#$%&|_^~@'`\"\\";

  for ( size_t i = 0; i < toExclude.size( ); i++ )
   dims.erase( dims.find( toExclude[i] ), 1 );

  return dims;
 }

void
PghMRIhandler::createChunk( const string& name,
                            const string& dimensions,
                            const string& type,
                            const string& datafile,
                            const int*    extents,
                            const int*    indexLowerBounds,
                                  bool    littleEndian,
                                  long    offset )
 {
  setString( name, "[chunk]" );
  setString( name + ".datatype", type );
  setString( name + ".dimensions", dimensions );
  setInt( name + ".offset", offset );

  if ( datafile != "" ) setString( name + ".file", datafile );
  if ( littleEndian   ) setInt( name + ".little_endian", 1 );

  size_t size = 1;
  for ( size_t d = 0; d < dimensions.size( ); d++ )
   {
    setInt( name + ".extent." + dimensions[d], extents[d] );
    size *= extents[d];
   }

  setInt( name + ".size", sizeOf( type ) * size );

  if ( indexLowerBounds )
   for ( size_t d = 0; d < dimensions.size( ); d++ )
    if ( indexLowerBounds[d] ) setInt( name + ".min." + dimensions[d], indexLowerBounds[d] );

  createChunk( name );
 }

bool
PghMRIhandler::matchChunk( const PghMRIhandler& mriHandler,
                           const string& chunkName, const string& chunkType,
			   const string& dimensions, const int* sizes, const int* lower )
 {
  if ( !mriHandler.hasKey( chunkName ) ) return false;
  if ( mriHandler.getString( chunkName ) != "[chunk]" ) return false;

  string dims = mriHandler.getString( chunkName + ".dimensions" );
  if ( dims != dimensions ) return false;

  if ( mriHandler.getString( chunkName + ".datatype" ) != chunkType ) return false;

  string keyE = chunkName + ".extent.?";
  string keyM = chunkName + ".min.?";
  for ( size_t i = 0; i < dimensions.size( ); i++ )
   {
    keyE[keyE.size( ) - 1] = dimensions[i];
    if ( mriHandler.getInt( keyE ) != sizes[i] ) return false;

    keyM[keyM.size( ) - 1] = dimensions[i];
    if ( mriHandler.getInt( keyM ) != lower[i] ) return false;
   }

  return true;
 }

// --------------------------------------------------------------------- private
// -----------------------------------------------------------------------------

void
PghMRIhandler::noOpenFileErrorHandler( ) const
 {
  if ( our_errorMode == mriAbortOnError )
   {
    cerr << our_fileNotOpenMsg << endl;
    abort( );
   }
  else if ( our_errorMode == mriReportErrors )
   cerr << our_fileNotOpenMsg << endl;
  else /* if ( our_errorMode == mriIgnoreErrors ) */
   my_error = our_fileNotOpenMsg;
 }

#endif /* __PghMRIhandler_H__ */
