#!/bin/tcsh -f

# ---------------------------------------------------------------------------
#
# try -help for details
#
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

# defaults
#
 set myname = ${0:t}

# temporary files prefix
#
 mkdir -p /tmp/${user}
 set tmpdir = /tmp/${user}/
 set tmp    = ${tmpdir}`echo ${myname} | sed 's/\.csh$//'`.`date '+%m%d%T' | sed 's/://g'`

# verbosity
#
 set verbmax   = 1
 set verbosity = ${verbmax}

# misc
#
 set comparison = absolute
 set threshold  = 1.0e-6
 set files      = ()

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${myname}
    echo ' numeric comparison of chunk1 in file1.mri with chunk2 in file2.mri'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo ' [-t|-threshold numeric_threshold]'
    echo ' [-a|-absolute|-r|-relative]'
    echo ''
    echo ' file1[.mri]:chunk1 file2[.mri]:chunk2'
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination; chunks identical to within threshold'
    echo ' 1: abnormal termination, except'
    echo ' 2: unspecified/non-existent input file(s)/chunk(s)'
    echo ' 3: unknown comparison'
    echo ' 4: chunks not identical to within threshold'
    echo ' 5-: failed utility/script'
    echo ''
    echo 'Required utilities/scripts'
    echo ' errmsg.csh'
    echo ' mri_rpn_math mri_subsample mri_printfield'
    echo ' mri.rename.chunk.csh mri.create.chunk.csh mri.print.chunk.csh mri.sizeof.csh'
    echo ''
    echo 'Notes'
    echo ' 1 mri_rpn_math does  single-precision  calculations  (as of today).'
    echo ' 2 "absolute" and "relative" refer to the numerical comparison type:'
    echo '   the two chunks are considered identical  if  c(a, b) <= threshold'
    echo '   for all a in chunk 1 and corresponding b in chunk 2, where c() is'
    echo '   defined as'
    echo '    absolute: c(a,b) := |a-b|'
    echo '    relative: c(a,b) := 0              if a == b == 0'
    echo '                     := |a-b|/max(|a|,|b|) otherwise.'
#   echo ''
#   echo 'Defaults'
#   echo " threshold: ${threshold}, comparison: ${comparison}"
#   echo ''
#   echo " verbosity: ${verbosity} (maximum ${verbmax})"
#   echo " temporary files prefix: ${tmp}"
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

   case "-t":
   case "-threshold":
    set threshold = "${argv[2]}"; shift; shift; breaksw;

   case "-r":
   case "-relative":
    set comparison = relative; shift; breaksw;

   case "-a":
   case "-absolute":
    set comparison = absolute; shift; breaksw;

   default:
    set files = ( ${files} ${argv[1]} ); shift; breaksw;
  endsw
 end

# error checks
#
 if ( ${#files} < 2 ) then
  errmsg.csh ${verbosity} ${myname} 2 "unspecified/non-existent input file(s)/chunk(s)" 1 ""
  exit ${status}
 endif

 set files = ( `echo ${files[1]} ${files[2]} | sed -e 's/\.mri:/:/g' -e 's/:/ /g'` )

 if ( ${#files} < 4 ) then
  errmsg.csh ${verbosity} ${myname} 2 "unspecified/non-existent input file(s)/chunk(s)" 1 ""
  exit ${status}
 else if ( !( -f ${files[1]}.mri ) || !( -f ${files[3]}.mri ) ) then
  errmsg.csh ${verbosity} ${myname} 2 "unspecified/non-existent input file(s)/chunk(s)" 1 ""
  exit ${status}
 endif

 if ( !( ( "${comparison}" == "absolute" ) || ( "${comparison}" == "relative" ) ) ) then
  errmsg.csh ${verbosity} ${myname} 3 "unknown comparison" 1 ""
  exit ${status}
 endif

# file and chunk names
#
 set file1  = ${files[1]}
 set chunk1 = ${files[2]}
 set file2  = ${files[3]}
 set chunk2 = ${files[4]}

# script for mri_rpn_math
#
 if ( "${comparison}" == "absolute" ) then
  set rpnscript = '$1,$2,-,abs'
 else if ( "${comparison}" == "relative" ) then
  set rpnscript = '$1,$2,-,abs,$1,abs,$2,abs,max,1,swap,dup,if_keep,/'
 endif

# create single-chunk header for the two chunks
#
# this is necessary to ensure that mri_rpn_math works correctly on
# .mri files containing multiple chunks with explicitly listed files
# (not just extensions).
#
# disregard chunk.min.? keys for the present purpose
#
 set size   = `mri_printfield -input ${file1} -field ${chunk1}.size`
 set type   = `mri_printfield -input ${file1} -field ${chunk1}.datatype`
 @   extent = ${size} / `mri.sizeof.csh ${type}`

 set l_endian = `mri_printfield -input ${file1} -field ${chunk1}.little_endian`
 if ( ${status} ) set l_endian = 0

 set ofs    = `mri_printfield -input ${file1} -field ${chunk1}.offset`

 set file   = ( `mri.datfile.csh -infile ${file1} -chunk ${chunk1}` )

 errmsg.csh ${verbosity} ${myname} 6 "failed command" ${status} "mri.datfile.csh -infile ${file1} -chunk ${chunk1}"
 if ( ${status} ) exit ${status}

 mri.create.chunk.csh -verbosity ${verbosity} -headeronly -little_endian ${l_endian} -offset ${ofs} -file ${file[1]} -mri ${tmp}.1 -chunk images -type ${type} -dimensions t -extents ${extent}

 errmsg.csh ${verbosity} ${myname} 7 "failed command" ${status} "mri.create.chunk.csh -verbosity ${verbosity} -headeronly -little_endian ${l_endian} -offset ${ofs} -file ${file[1]} -mri ${tmp}.1 -chunk images -type ${type} -dimensions t -extents ${extent}"
 if ( ${status} ) exit ${status}

 set size   = `mri_printfield -input ${file2} -field ${chunk2}.size`
 set type   = `mri_printfield -input ${file2} -field ${chunk2}.datatype`
 @   extent = ${size} / `mri.sizeof.csh ${type}`

 set l_endian = `mri_printfield -input ${file2} -field ${chunk2}.little_endian`
 if ( ${status} ) set l_endian = 0

 set ofs    = `mri_printfield -input ${file2} -field ${chunk2}.offset`

 set file   = ( `mri.datfile.csh -infile ${file2} -chunk ${chunk2}` )

 errmsg.csh ${verbosity} ${myname} 8 "failed command" ${status} "mri.datfile.csh -infile ${file2} -chunk ${chunk2}"
 if ( ${status} ) exit ${status}

 mri.create.chunk.csh -verbosity ${verbosity} -headeronly -little_endian ${l_endian} -offset ${ofs} -file ${file[1]} -mri ${tmp}.2 -chunk images -type ${type} -dimensions t -extents ${extent}

 errmsg.csh ${verbosity} ${myname} 9 "failed command" ${status} "mri.create.chunk.csh -verbosity ${verbosity} -headeronly -little_endian ${l_endian} -offset ${ofs} -file ${file[1]} -mri ${tmp}.2 -chunk images -type ${type} -dimensions t -extents ${extent}"
 if ( ${status} ) exit ${status}

# compute differences
#
 mri_rpn_math -o ${tmp} "${rpnscript}" -c images ${tmp}.1 ${tmp}.2 >& ${tmp}.err

 errmsg.csh ${verbosity} ${myname} 10 "failed command" ${status} "mri_rpn_math -o ${tmp} '${rpnscript}' -c images ${tmp}.1 ${tmp}.2 >& ${tmp}.err"
 if ( ${status} ) exit ${status}

# compute sum of all comparisons
#
 mri_subsample -d t -e 1 -max ${tmp} ${tmp}.3 >& ${tmp}.err

 errmsg.csh ${verbosity} ${myname} 11 "failed command" ${status} "mri_subsample -d t -e 1 -max ${tmp} ${tmp}.3 >& ${tmp}.err"
 if ( ${status} ) exit ${status}

 set absdiffmax = `mri.print.chunk.csh -t 0 -C '<=' -i ${tmp}.3 -c images | awk '{print $2}'`

# report success/failure
#
 set s = `echo ${absdiffmax} ${threshold} | awk '{s=0;if($1>$2)s=4;print s}'`

 if ( ${s} && ( ${verbosity} > 0 ) ) then
  echo "comparison: ${comparison}, threshold: ${threshold}, maximum mismatch: ${absdiffmax}"
 endif

# cleanup
#
 rm -f ${tmp}* >& /dev/null

 exit ${s}
