#!/bin/tcsh -f

# ---------------------------------------------------------------------------
#
# try -help for details
#
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

# defaults
#
 set myname = ${0:t}

# temporary files prefix
#
 mkdir -p /tmp/${user}
 set tmpdir = /tmp/${user}/
 set tmp    = ${tmpdir}`echo ${myname} | sed 's/\.csh$//'`.`date '+%m%d%T' | sed 's/://g'`

# verbosity
#
 set verbmax   = 1
 set verbosity = ${verbmax}

# misc
#
 set comparison = absolute
 set threshold  = 1.0e-6
 set files = ()
 set subset = common

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${myname}
    echo ' comparison of text and chunkwise contents of two .mri files.'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo ' [-t|-threshold numeric_threshold]'
    echo ' [-a|-absolute|-r|-relative]'
    echo ''
    echo ' [-n|-notext]'
    echo ' [-T|-terminate]'
    echo ''
    echo ' [-s|-subset all|common]'
    echo ''
    echo ' file1[.mri] file2[.mri]'
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination; text contents match, all chunks match to within threshold'
    echo ' 1: abnormal termination, except'
    echo ' 2: unspecified/non-existent input file(s)'
    echo ' 3: non-identical text contents'
    echo ' 4: not all chunks in the first file are present in the second (-subset all)'
    echo ' 5: no common chunks (-subset common)'
    echo ' 6: non-identical chunks contents for at least one chunk (to within threshold)'
    echo ' 7-: failed utility/script'
    echo ''
    echo 'Required utilities/scripts'
    echo ' errmsg.csh'
    echo ' mri.cmp.txt.csh mri.cmp.chunk.csh mri.list.chunk.names.csh'
    echo ''
    echo 'Notes'
    echo ' 1 -notext skips the text comparison, -terminate terminates if text contents differ.'
    echo ' 2 see the notes section of mri.cmp.chunk.csh -help for details on comparisons.'
    echo ' 3 all chunk names in the first file should exist in the second.'
#   echo ''
#   echo 'Defaults'
#   echo " threshold: ${threshold}, comparison: ${comparison}"
#   echo ''
#   echo " verbosity: ${verbosity} (maximum ${verbmax})"
#   echo " temporary files prefix: ${tmp}"
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

   case "-t":
   case "-threshold":
    set threshold = "${argv[2]}"; shift; shift; breaksw;

   case "-r":
   case "-relative":
    set comparison = relative; shift; breaksw;

   case "-a":
   case "-absolute":
    set comparison = absolute; shift; breaksw;

   case "-n":
   case "-notext":
    set notext; shift; breaksw;

   case "-T":
   case "-terminate":
    set terminate; shift; breaksw;

   case "-s":
   case "-subset":
    set subset = "${argv[2]}"; shift; shift; breaksw;

   default:
    set files = ( ${files} ${argv[1]} ); shift; breaksw;
  endsw
 end

# error checks
#
 if ( ${#files} < 2 ) then
  errmsg.csh ${verbosity} ${myname} 2 "unspecified/non-existent input file(s)" 1 ""
  exit ${status}
 endif

 set files = ( `echo ${files[1]} ${files[2]} | sed -e 's/\.mri / /g' -e 's/\.mri$//'` )

 if ( !( -f ${files[1]}.mri ) || !( -f ${files[2]}.mri ) ) then
  errmsg.csh ${verbosity} ${myname} 2 "unspecified/non-existent input file(s)" 1 ""
  exit ${status}
 endif

# text comparison
#
 if ( !( ${?notext} ) ) then
  mri.cmp.txt.csh -verbosity ${verbosity} ${files[1]} ${files[2]}

  set s = ${status}
  if ( ${s} == 3 ) then
   if ( ${?terminate} ) exit 3
  else if ( ${s} != 0 ) then
   errmsg.csh ${verbosity} ${myname} 6 "failed command" ${s} "mri.cmp.txt.csh -verbosity ${verbosity} ${files[1]} ${files[2]}"
   exit ${status}
  endif
 endif

# subset of chunks to compare
#
 mri.list.chunk.names.csh ${files[1]} > ${tmp}.0
 mri.list.chunk.names.csh ${files[2]} > ${tmp}.1

 sed -e 's/^/^/' -e 's/$/$/' ${tmp}.0 > ${tmp}.2
 grep -f ${tmp}.2 ${tmp}.1 > ${tmp}.3

 if ( ${subset} == all ) then
  cmp -s ${tmp}.0 ${tmp}.3
  if ( ${status} ) then
   if ( ${verbosity} > 0 ) echo 'not all chunks in the first file are present in the second (-subset all)'
   exit 4
  endif
 endif

 if ( ( ${subset} == common ) && ( `wc -l ${tmp}.3 | awk '{print $1}'` == 0 ) ) then
  if ( ${verbosity} > 0 ) echo 'no common chunks (-subset common)'
  exit 5
 endif

# chunkwise comparison
#
 set exitstatus = 0

 foreach c ( `cat ${tmp}.3` )
  if ( ${verbosity} > 0 ) echo -n "chunk ${c}: "

  mri.cmp.chunk.csh -verbosity ${verbosity} -threshold ${threshold} -${comparison} ${files[1]}:${c} ${files[2]}:${c}

  set s = ${status}
  if ( ${s} == 0 ) then
   echo "match"
  else if ( ${s} == 4 ) then
   set exitstatus = 6
  else
   errmsg.csh ${verbosity} ${myname} 7 "failed command" ${s} "mri.cmp.chunk.csh -verbosity ${verbosity} -threshold ${threshold} -${comparison} ${files[1]}:${c} ${files[2]}:${c}"
   exit ${status}
  endif
 end

# cleanup
#
 rm -f ${tmp}* >& /dev/null

 exit ${exitstatus}
