#!/bin/tcsh -f

# ---------------------------------------------------------------------------
#
# try -help for details
#
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

# defaults
#
 set myname = ${0:t}

# temporary files prefix
#
 mkdir -p /tmp/${user}
 set tmpdir = /tmp/${user}/
 set tmp    = ${tmpdir}`echo ${myname} | sed 's/\.csh$//'`.`date '+%m%d%T' | sed 's/://g'`

# verbosity
#
 set verbmax   = 1
 set verbosity = ${verbmax}

# misc
#
 set mode      = 'single_record' # modes: timecourse, single_record
 set chunkinfo = ''              # preprocessed chunk info (mrichunkinfo format), if available
 set indspecs  = ''
 set infile    = ''
 set chunk     = ''

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${myname}
    echo ' Extract a single record from data in .mri format.'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo ' [-tc|-timecourse]'
    echo ' -i|-infile infile[.mri] -c|-chunk chunk'
    echo ' -o|-outfile outfile[.mri]'
    echo ' [indspecs]'
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination'
    echo ' 1: abnormal termination, except'
    echo ' 2: unspecified input file/chunk'
    echo ' 3: unspecified output file'
    echo ' 5: nonexistent data file'
    echo ' 4,6-: failed utility/script'
    echo ''
    echo 'Required utilities/scripts'
    echo ' errmsg.csh'
    echo ' mri.chunk.info.csh mri.offset.csh mri.create.chunk.csh'
    echo ' mri_subset'
    echo ''
    echo 'Notes'
    echo ' 1 Refer to proc.indspecs.csh for index specification format.'
    echo ' 2 A  time course  is defined as  the entire  range of values'
    echo '   corresponding  to the  first  nontrivial  fastest  varying'
    echo '   dimension   given  the  coordinate  indices  for  the rest'
    echo '   of the dimensions.  In this mode  (i. e.,  if  -timecourse'
    echo '   is  specified),  any  index  value  corresponding  to  the'
    echo '   fastest-varying   data   dimension   is  ignored.  If  all'
    echo '   dimensions have size = 1, fastest == last.'
#   echo ''
#   echo 'Defaults'
#   echo " verbosity: ${verbosity} (maximum ${verbmax})"
#   echo " temporary files prefix: ${tmp}"
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

   case "-tc":
   case "-timecourse":
    set mode = timecourse; shift; breaksw;

   case "-i":
   case "-infile":
    set infile = `echo ${argv[2]} | sed 's/\.mri$//'`;
    shift; shift; breaksw;

   case "-o":
   case "-outfile":
    set outfile = `echo ${argv[2]} | sed 's/\.mri$//'`;
    shift; shift; breaksw;

   case "-c":
   case "-chunk":
    set chunk = ${argv[2]}; shift; shift; breaksw;

   case "-ci":
   case "-chunkinfo":
    set chunkinfo = ${argv[2]}; shift; shift; breaksw;

   default:
    set indspecs = ${argv[1]}; shift; breaksw;
  endsw
 end

# error checks
#
 if ( ( ( ${infile} == '' ) || ( ${chunk} == '' ) ) && ( ${chunkinfo} == '' ) ) then
  errmsg.csh ${verbosity} ${myname} 2 "unspecified input file/chunk" 1 ""
  exit ${status}
 else if ( !( ${?outfile} ) ) then
  errmsg.csh ${verbosity} ${myname} 3 "unspecified output file" 1 ""
  exit ${status}
 endif

# chunk attributes
#
 if ( "${chunkinfo}" == '' ) then
  mri.chunk.info.csh -verbosity ${verbosity} -infile ${infile} -chunk ${chunk} > ${tmp}

  errmsg.csh ${verbosity} ${myname} 4 "failed command" ${status} "mri.chunk.info.csh -verbosity ${verbosity} -infile ${infile} -chunk ${chunk} > ${tmp}"
  if ( ${status} ) exit ${status}

  set chunkinfo = ${tmp}
 endif

 set file = ( `grep '^physical_file ' ${chunkinfo} | sed 's/physical_file//'` )

 if ( "${file[2]}" != "exists" ) then
  errmsg.csh ${verbosity} ${myname} 5 "nonexistent data file ${file[1]}" 1 ""
  if ( ${status} ) exit ${status}
 endif

 set file     = ${file[1]}
 set chunk    =   `grep '^chunk '         ${chunkinfo} | sed 's/chunk//'`
 set type     =   `grep '^datatype '      ${chunkinfo} | sed 's/datatype//'`
 set l_endian =   `grep '^little_endian ' ${chunkinfo} | sed 's/little_endian//'`
 set dims     = ( `grep '^dimensions '    ${chunkinfo} | sed 's/dimensions//'` )
 set sizes    = ( `grep '^extents '       ${chunkinfo} | sed 's/extents//'`    )
 set lower    = ( `grep '^lower '         ${chunkinfo} | sed 's/lower//'`      )
#set upper    = ( `grep '^upper '         ${chunkinfo} | sed 's/upper//'`      )
#set index    = ( ${lower} )

 switch ( ${mode} )
  case timecourse:
   set fd = 1
   while ( ( ${sizes[${fd}]} == 1 ) && ( ${fd} < ${#sizes} ) )
    @ fd += 1
   end
   #
   # fastest-varying non-trivial dimension (last if all sizes == 1)

   set fastest  = ${dims[${fd}]}
   set size     = ${sizes[${fd}]}
   set indspecs = "${indspecs},${dims[${fd}]}=${lower[${fd}]}"

   set description = "timecourse"

   breaksw

  default:
   set fastest = f
   set size    = 1
   #
   # assign an arbitrary name to an imaginary leading dimension with size 1

   set description = "single_record"

   breaksw
 endsw

# byte offset for timecourse/single_record
# (mrioffset will substitute appropriate defaults for unspecified indices)
#
 set offset = ( `mri.offset.csh -verbosity ${verbosity} -chunkinfo "${chunkinfo}" -infile "${infile}" -chunk "${chunk}" ${indspecs}` )

 errmsg.csh ${verbosity} ${myname} 6 "failed command" ${status} "mri.offset.csh -verbosity ${verbosity} -chunkinfo "'"'${chunkinfo}'"'" -infile "'"'${infile}'"'" -chunk "'"'${chunk}'"'" ${indspecs}"
 if ( ${status} ) exit ${status}

 set description = "${description} ${infile}.mri ${chunk} ${indspecs} byte_offset=${offset[1]} rec_offset=${offset[2]}"

# generate a mri file describing only this one timecourse/record
#
 mri.create.chunk.csh -verbosity ${verbosity} -headeronly -little_endian ${l_endian} -offset ${offset[1]} -file ${file} -mri ${tmp} -chunk ${chunk} -type ${type} -dimensions ${fastest} -extents ${size} -description "${description}"

 errmsg.csh ${verbosity} ${myname} 7 "failed command" ${status} "mri.create.chunk.csh -verbosity ${verbosity} -headeronly -little_endian ${l_endian} -offset ${offset[1]} -file ${file} -mri ${tmp} -chunk ${chunk} -type ${type} -dimensions ${fastest} -extents ${size} -description "'"'${description}'"'
 if ( ${status} ) exit ${status}

# extract time course using mri_subset
#
 mri_subset -d ${fastest} -s 0 -l ${size} ${tmp}.mri ${outfile}.mri >& ${tmp}.err

 errmsg.csh ${verbosity} ${myname} 8 "failed command" ${status} "mri_subset -d ${fastest} -s 0 -l ${size} ${tmp}.mri ${outfile}.mri >& ${tmp}.err"
 if ( ${status} ) exit ${status}

# clean up
#
 rm -f ${tmp}* >& /dev/null; exit 0
