#!/bin/tcsh -f

# ---------------------------------------------------------------------------
#
# try -help for details
#
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

# defaults
#
 set myname = ${0:t}

# temporary files prefix
#
 mkdir -p /tmp/${user}
 set tmpdir = /tmp/${user}/
 set tmp    = ${tmpdir}`echo ${myname} | sed 's/\.csh$//'`.`date '+%m%d%T' | sed 's/://g'`

# verbosity
#
 set verbmax   = 1
 set verbosity = 0

# preprocessed chunk information (mrichunkinfo format) if available
#
 set chunkinfo = ''

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${myname}
    echo ' Extract  indices  corresponding to specified dimensions from an'
    echo ' index specification; see below.  Optionally, if an mri file and'
    echo ' a chunk is specified, then  (1) substitute appropriate defaults'
    echo ' (2) check for out-of-range indices.'
    echo ''
    echo 'Index specification format'
    echo ' a=value,...,z=value'
    echo ' a-z are 1-character dimension names  (converted to lower case).'
    echo ' Multiple  "a=value"  specifications ok; the last one supercedes'
    echo ' all the earlier ones.  Special values "min", "max"  are allowed'
    echo ' for indices. Chunk indices are considered 0-based.  Leading and'
    echo ' trailing commas are ignored (so are multiple ones).'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo ' form 1: [-d|-detailed] -i|-infile file[.mri] -c|-chunk chunk [indspecs]'
    echo ' form 2: indspecs'
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination'
    echo ' 1: abnormal termination, except'
    echo ' 2: no index and/or file+chunk specification'
    echo ' 3: illegal index specification'
    echo ' 4: failed utility/script'
    echo ' 5: out-of-range index'
    echo ''
    echo 'Required utilities/scripts'
    echo ' errmsg.csh'
    echo ' mri.chunk.info.csh'
    echo ''
    echo 'Notes'
    echo ' 1 Not smart enough  to detect  syntax problems  in  index'
    echo '   specification.  This may lead to abnormal terminations.'
    echo ' 2 output format when no mri file and chunk are specified'
    echo '    dimension index'
    echo '    ...'
    echo ' 3 output format when mri file and chunk are specified'
    echo '    dimension index size lower_bound upper_bound'
    echo '    ...'
#   echo ''
#   echo 'Defaults'
#   echo " verbosity: ${verbosity} (maximum ${verbmax})"
#   echo " temporary files prefix: ${tmp}"
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

   case "-d":
   case "-detailed":
    set detailed; shift; breaksw;

   case "-i":
   case "-infile":
    set infile = `echo ${argv[2]} | sed 's/\.mri$//'`;
    shift; shift; breaksw;

   case "-c":
   case "-chunk":
    set chunk = ${argv[2]}; shift; shift; breaksw;

   case "-ci":
   case "-chunkinfo":
    set chunkinfo = ${argv[2]}; shift; shift; breaksw;

   default:
    set indspecs = ${argv[1]}; shift; breaksw;
  endsw
 end

# error checks
#
 if ( ( ${?infile} && ${?chunk} ) || ( "${chunkinfo}" != '' ) ) set bothspecified

 if ( !( ${?indspecs} || ${?bothspecified} ) ) then
  errmsg.csh ${verbosity} ${myname} 2 "no index and/or file+chunk specification" 1 ""
  exit ${status}
 endif

 if ( ${?indspecs} ) then
  set indspecs = ( `echo ${indspecs} | awk '{print tolower($0)}' | sed -e 's/,/\\n/g' -e 's/=+/=/g'` )

  echo ${indspecs} | grep -v '^$' | grep -q -v -e '[a-z]=[0-9]' -e '[a-z]=-[0-9]' -e '[a-z]=+[0-9]' -e '[a-z]=min' -e '[a-z]=max'

  if ( ${status} == 0 ) then
   errmsg.csh ${verbosity} ${myname} 3 "illegal index specification" 1 ""
   exit ${status}
  endif

 # extract dimensions and indices from index specification
 #
  set dimensions = ( `echo ${indspecs} | grep -v '^$' | awk -F= '{print $1}' | sort | uniq` )

  set indices = ()
  foreach d ( ${dimensions} )
   set indices = ( ${indices} `echo ${indspecs} | grep -v '^$' | awk -F= '{print $1" "$2}' | grep "^${d}" | tail -1 | awk '{print $2}'` )
  end

 # if an mri file and a chunk hasn't been specified, stop here
 #
  if ( !( ${?bothspecified} ) ) then
   echo ${dimensions} ${indices} | awk '{k='${#dimensions}'; for( i = 1; i <= k; i++ ) print $i" "$(i+k)}'
   exit 0
  endif
 endif

# dimensionwise sizes, lower and upper bounds on indices
#
 if ( "${chunkinfo}" == '' ) then
  mri.chunk.info.csh -verbosity ${verbosity} -infile ${infile} -chunk ${chunk} > ${tmp}

  errmsg.csh ${verbosity} ${myname} 4 "failed command" ${status} "mri.chunk.info.csh -verbosity ${verbosity} -infile ${infile} -chunk ${chunk} > ${tmp}"
  if ( ${status} ) exit ${status}

  set chunkinfo = ${tmp}
 endif

 set dims  = ( `grep '^dimensions ' ${chunkinfo} | sed 's/dimensions//'` )
 set sizes = ( `grep '^extents '    ${chunkinfo} | sed 's/extents//'`    )
 set lower = ( `grep '^lower '      ${chunkinfo} | sed 's/lower//'`      )
 set upper = ( `grep '^upper '      ${chunkinfo} | sed 's/upper//'`      )
 set index = ( ${lower} )

# replace "min", "max" by lower and upper bounds;
# optional strict range enforcement
#
 if ( ${?indspecs} ) then
  while ( ${#dimensions} != 0 )
   set i = 1; set notfound;
   while ( ( ${i} <= ${#dims} ) && ${?notfound} )
    set d = ${dims[${i}]}
    if ( ${dimensions[1]} == ${d} ) then
     unset notfound

     switch ( ${indices[1]} )
      case "min":
       breaksw

      case "max":
       set index[${i}] = ${upper[${i}]}
       breaksw

      default:
       if ( ( ${indices[1]} < ${lower[${i}]} ) || ( ${indices[1]} > ${upper[${i}]} ) ) then
        errmsg.csh ${verbosity} ${myname} 5 "out-of-range ${d}=${indices[1]} (min=${lower[${i}]}, max=${upper[${i}]})" 1 ""
        exit ${status}
       else
        set index[${i}] = ${indices[1]}
       endif
       breaksw
     endsw
    else
     set notfound
    endif

    @ i += 1
   end

   shift dimensions
   shift indices
  end
 endif

# print processed indices
#
 if ( ${?detailed} ) then
  cat ${chunkinfo}
  echo "\nindices ${index}"
 else
  set len = `echo ${sizes} ${lower} ${upper} | awk '{l=0;for(i=1;i<=NF;i++){s=length($i);if(l<s)l=s};print l+1}'`

  set fmt = "%c%${len}d%${len}d%${len}d%${len}d\n"

  echo ${dims} ${index} ${sizes} ${lower} ${upper} | awk '{k='${#dims}'; for( i = 1; i <= k; i++ ) printf "'${fmt}'", $i, $(i+k), $(i+2*k), $(i+3*k), $(i+4*k)}'
 endif

# cleanup
#
 rm -f ${tmp}* >& /dev/null; exit 0
