#!/bin/tcsh -f

# ---------------------------------------------------------------------------
#
# try -help for details
#
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

# defaults
#
 set myname = ${0:t}

# temporary files prefix
#
 mkdir -p /tmp/${user}
 set tmpdir = /tmp/${user}/
 set tmp    = ${tmpdir}`echo ${myname} | sed 's/\.csh$//'`.`date '+%m%d%T' | sed 's/://g'`

# verbosity
#
 set verbmax   = 1
 set verbosity = ${verbmax}

# vmp output prefix, submodels to exclude
#
 set prefix   = './'
 set excluded = ''

# probabilities file/chunk names
#
 set mpchunk = model.probabilities

# zero
#
 set zero = 1.0e-8

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${0}
    echo ' Compute submodel-averaged estimates from vmp output.'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo ' [-p|-prefix input_prefix]'
    echo ' [-o|-outfile output_file_name]'
    echo ' [-e|-exclude submodel1,submodel2,...]'
    echo ' [-z|-zero numerical_zero]'
    echo ' quantity'
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination'
    echo ' 1: abnormal termination, except'
    echo ' 2: unspecified quantity to average'
    echo ' 3: nothing to average over (all submodels excluded)'
    echo ' 4-: failed utility/script'
    echo ''
    echo 'Required utilities/scripts'
    echo ' errmsg.csh'
    echo ' mri.rename.chunk.csh extract.mp.csh'
    echo ' mri_printfield mri_rpn_math'
    echo ''
    echo 'Notes'
    echo ' 1 "-exclude submodel1,submodel2,..."  excludes  all the'
    echo '   specified submodels from averaging  and  renormalizes'
    echo '   probabilities for included submodels using'
    echo '   ( 1 - p(submodel1) - p(submodel2) - ... )^-1.'
    echo ' 2 "-zero z" specifies the numerical zero threshold used'
    echo '   in deciding whether |1-p(submodel1)-p(submodel2)-...|'
    echo '   should be treated as zero.'
    echo ' 3 Individual (submodel-specific)  files for quantity to'
    echo '   be submodel-averaged are assumed to have names of the'
    echo '   form (input_prefix)(submodel).(quantity).mri.'
#   echo ''
#   echo 'Defaults'
#   echo " prefix: '${prefix}'"
#   echo ''
#   echo " verbosity: ${verbosity} (maximum ${verbmax})"
#   echo " temporary files prefix: ${tmp}"
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

   case "-o":
   case "-outfile":
    set outfile = `echo ${argv[2]} | sed 's/\.mri$//'`;
    shift; shift; breaksw;

   case "-p":
   case "-prefix":
    set prefix = ${argv[2]}; shift; shift; breaksw;

   case "-z":
   case "-zero":
    set zero = ${argv[2]}; shift; shift; breaksw;

   case "-e":
   case "-exclude":
    set excluded = ( `echo ${argv[2]} | sed 's/,/ /g'` );
    shift; shift; breaksw;

   default:
    set quantity = ${argv[1]}; shift; breaksw;
  endsw
 end

# error checks
#
 if ( !( ${?quantity} ) ) then
  errmsg.csh ${verbosity} ${myname} 2 "unspecified quantity" 1 ""
  exit ${status}
 endif

 if ( !( ${?outfile} ) ) set outfile = average.${quantity}

# extract model probabilities
#
 extract.mp.csh -verbosity ${verbosity} -prefix ${prefix} -outprefix ${tmp}. -outsuffix '' -outchunk ${quantity} >& ${tmp}.mp

 errmsg.csh ${verbosity} ${myname} 4 "failed command" ${status} "extract.mp.csh -verbosity ${verbosity} -prefix ${prefix} -outprefix ${tmp}. -outsuffix '' -outchunk ${quantity} >& ${tmp}.mp"
 if ( ${status} ) exit ${status}

 # echo excluded \"$excluded\"

# excluded submodels list: throw away those not present in the actual analysis
#
 set xlist  = ""
 set togrep = ""
 set xfiles = ""
 foreach s ( ${excluded} )
  set ss = ( `grep "^submodel: ${s}" ${tmp}.mp | awk '{print $2" "$4}'` )
  if ( "${ss[1]}" == "${s}" ) then
   set xlist  = ( ${xlist} ${s} )
   set togrep = ( ${togrep} "-e '^submodel: ${s}'" )
   set xfiles = ( ${xfiles} ${ss[2]} )
  endif
 end
 set excluded = ( ${xlist} )

 # echo excluded \"$excluded\"
 # echo xfiles \"$xfiles\"
 # echo togrep \"$togrep\"

# submodel and filenames for model probabilities
#
 if ( ${#excluded} == 0 ) then
  set submodels = ( `awk '{print $2}' ${tmp}.mp` )
  set mfiles    = ( `awk '{print $4}' ${tmp}.mp` )
 else
  alias mygrep "grep -v ${togrep}"
  set submodels = ( `mygrep ${tmp}.mp | awk '{print $2}'` )
  set mfiles    = ( `mygrep ${tmp}.mp | awk '{print $4}'` )
  unalias mygrep
 endif

 if ( ${#submodels} == 0 ) then
  errmsg.csh ${verbosity} ${myname} 3 "nothing to average over: all submodels excluded" 1 ""
  exit ${status}
 endif

 # echo submodels \"$submodels\"
 # echo mfiles \"$mfiles\"

# submodelwise filenames for quantity estimates
#
 set qfiles = ''
 foreach s ( ${submodels} )
  set qfiles = ( ${qfiles} ${prefix}${s}.${quantity} )
 end

 # echo qfiles \"$qfiles\"

# compute renormalization factors
#
 if ( ${#excluded} > 0 ) then
  mri_rpn_math -o ${tmp}.x -c ${quantity} '$1' ${xfiles[1]}

  set i = 2
  while ( ${i} <= ${#excluded} )
   mri_rpn_math -o ${tmp} -c ${quantity} '$1,$2,+' ${tmp}.x ${xfiles[${i}]} >& ${tmp}.err

   errmsg.csh ${verbosity} ${myname} 5 "failed command" ${status} "mri_rpn_math -o ${tmp} -c ${quantity} '"'$1,$2,+'"' ${tmp}.x ${xfiles[${i}]} >& ${tmp}.err"
   if ( ${status} ) exit ${status}

   mv ${tmp}.dat ${tmp}.x.dat

   @ i += 1
  end

  mri_rpn_math -o ${tmp}.n -c ${quantity} '1.0,1.0,$1,-,/,0.0,'${zero}',1.0,$1,-,abs,<=,if_keep' ${tmp}.x >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 6 "failed command" ${status} "mri_rpn_math -o ${tmp}.n -c ${quantity} '"'1.0,1.0,$1,-,/,0.0,'${zero}',1.0,$1,-,abs,<=,if_keep'"' ${tmp}.x >& ${tmp}.err"
  if ( ${status} ) exit ${status}
 else
  mri_rpn_math -o ${tmp}.n -c ${quantity} '1.0' ${mfiles[1]} >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 7 "failed command" ${status} "mri_rpn_math -o ${tmp}.n -c ${quantity} '1.0' ${mfiles[1]} >& ${tmp}.err"
  if ( ${status} ) exit ${status}
 endif

# submodel-averaged quantity estimate
#  := Sum{submodel != excluded} p(submodel) (1-p(excluded))^-1 quantity(submodel)
#
 mri_rpn_math -o ${outfile} -c ${quantity} '$1,$2,$3,*,*' ${qfiles[1]} ${mfiles[1]} ${tmp}.n >& ${tmp}.err

 errmsg.csh ${verbosity} ${myname} 8 "failed command" ${status} "mri_rpn_math -o ${outfile} -c ${quantity} '$1,$2,$3,*,*' ${qfiles[1]} ${mfiles[1]} ${tmp}.n >& ${tmp}.err"
 if ( ${status} ) exit ${status}

 set i = 2
 while ( ${i} <= ${#submodels} )
  mri_rpn_math -o ${tmp}.p -c ${quantity} '$1,$2,$3,*,*' ${qfiles[${i}]} ${mfiles[${i}]} ${tmp}.n >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 9 "failed command" ${status} "mri_rpn_math -o ${tmp}.p -c ${quantity} '$1,$2,$3,*,*' ${qfiles[${i}]} ${mfiles[${i}]} ${tmp}.n >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  mri_rpn_math -o ${tmp} -c ${quantity} '$1,$2,+' ${outfile} ${tmp}.p >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 10 "failed command" ${status} "mri_rpn_math -o ${tmp} -c ${quantity} '"'$1,$2,+'"' ${outfile} ${tmp}.p >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  mv ${tmp}.dat ${outfile}.dat

  @ i += 1
 end

# cleanup
#
 rm -f ${tmp}* >& /dev/null; exit 0
