#!/bin/tcsh -f

# ---------------------------------------------------------------------------
#
# try -help for details
#
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

# defaults
#
 set myname = ${0:t}

# temporary files prefix
#
 mkdir -p /tmp/${user}
 set tmpdir = /tmp/${user}/
 set tmp    = ${tmpdir}`echo ${myname} | sed 's/\.csh$//'`.`date '+%m%d%T' | sed 's/://g'`

# verbosity
#
 set verbmax   = 1
 set verbosity = ${verbmax}

# vmp output prefix, submodels to process
#
 set prefix     = './'
 set submodels  = ''

# misc
#
 set param      = baseline
 set plotlimits = 'auto:auto'
 set zero       = 1.0e-8 # numerical zero for relative error calculation

 set plot_estimate
 set plot_stderr
 set plot_relerr

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${myname}
    echo ' Make baseline/stderr images from vmp output.'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo ' [-ne|-noestimate] [-ns|-nostderr] [-nr|-norelerr]'
    echo ' [-pl|-plotlimits [lower][:[upper]]]'
    echo ' [-p|-prefix any_path/prefix]'
    echo ' submodel1 submodel2 ...'
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination'
    echo ' 1: abnormal termination, except'
    echo ' 2-: failed utility/script'
    echo ''
    echo 'Required utilities/scripts'
    echo ' errmsg.csh'
    echo ' mri.to.img.csh mri.rename.chunk.csh proc.rangeopts.csh'
    echo ' mri_printfield mri_rpn_math mri_to_img'
#   echo ''
#   echo 'Defaults'
#   echo " prefix: '${prefix}'"
#   echo ''
#   echo " verbosity: ${verbosity} (maximum ${verbmax})"
#   echo " temporary files prefix: ${tmp}"
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

   case "-p":
   case "-prefix":
    set prefix = `echo ${argv[2]} | sed 's/\.mri$//'`;
    shift; shift; breaksw;

   case "-pl":
   case "-plotlimits":
    set plotlimits = ${argv[2]}; shift; shift; breaksw;

   case "-ne":
   case "-noestimate":
    unset plot_estimate; shift; breaksw;

   case "-ns":
   case "-nostderr":
    unset plot_stderr; shift; breaksw;

   case "-nr":
   case "-norelerr":
    unset plot_relerr; shift; breaksw;

   default:
    set submodels = ( ${submodels} ${argv[1]} ); shift; breaksw;
  endsw
 end

# error checks
#
 if ( !( ${?plot_estimate} || ${?plot_stderr} || ${?plot_relerr} ) || ( "${submodels}" == "" ) ) then
# errmsg.csh ${verbosity} ${myname} 0 "nothing to plot" 1 ""
# exit ${status}
  exit 0
 endif

# convert plot range options into a form that can be passed on to mri_to_img
#
 set plotlimits = ( `proc.rangeopts.csh -verbosity ${verbosity} -defaults "auto:auto" ${plotlimits}` )

 errmsg.csh ${verbosity} ${myname} 2 "failed command" ${status} "proc.rangeopts.csh -verbosity ${verbosity} -defaults "auto:auto" ${plotlimits}"
 if ( ${status} ) exit ${status}

 if ( "${plotlimits[1]}" == "auto" ) then
  set plotlimits[1] = ""
 else
  set plotlimits[1] = "-n ${plotlimits[1]}"
 endif

 if ( "${plotlimits[2]}" == "auto" ) then
  set plotlimits[2] = ""
 else
  set plotlimits[2] = "-x ${plotlimits[2]}"
 endif

# loop over submodels
#
 foreach submodel ( ${submodels} )

  # image titles
  #
  set slices = `mri_printfield -input ${prefix}${submodel}.${param} -field ${param}.extent.z`

  errmsg.csh ${verbosity} ${myname} 3 "failed command" ${status} "mri_printfield -input ${prefix}${submodel}.${param} -field ${param}.extent.z"
  if ( ${status} ) exit ${status}

  setenv F_PS_TITLE1 `mri_printfield -input ${prefix}${submodel}.${param} -field imagedata.description`
  setenv F_PS_TITLE2 "submodel: ${submodel}, slices: ${slices}"

  # estimate
  #
  if ( !( ${?plot_estimate} ) ) goto label_skipE

  setenv F_PS_TITLE3 "${param}"

  mri.to.img.csh -infile ${prefix}${submodel}.${param} -chunk ${param} -fiasco "-mosaic -t -c -p -r -g ${plotlimits}"

  errmsg.csh ${verbosity} ${myname} 4 "failed command" ${status} "mri.to.img.csh -infile ${prefix}${submodel}.${param} -chunk ${param} -fiasco '-mosaic -t -c -p -r -g ${plotlimits}'"
  if ( ${status} ) exit ${status}

  label_skipE:

  # standard error
  #
  if ( !( ${?plot_stderr} ) ) goto label_skipS

  setenv F_PS_TITLE3 "${param} standard error"

  mri.to.img.csh -infile ${prefix}${submodel}.${param}.stderr -chunk ${param}.stderr -fiasco "-mosaic -t -c -p -r -g ${plotlimits}" -outfile ${submodel}.${param}.se.ps

  errmsg.csh ${verbosity} ${myname} 5 "failed command" ${status} "mri.to.img.csh -infile ${prefix}${submodel}.${param}.stderr -chunk ${param}.stderr -fiasco '-mosaic -t -c -p -r -g ${plotlimits}' -outfile ${submodel}.${param}.se.ps"
  if ( ${status} ) exit ${status}

  label_skipS:

  # relative error
  #
  if ( !( ${?plot_relerr} ) ) goto label_skipR

  # rename estimate/stderr chunks to "images" for relative error calculation
  #
  mri.rename.chunk.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.${param} -chunk ${param} -outfile ${tmp}.e -to images

  errmsg.csh ${verbosity} ${myname} 6 "failed command" ${status} "mri.rename.chunk.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.${param} -chunk ${param} -outfile ${tmp}.e -to images"
  if ( ${status} ) exit ${status}

  mri.rename.chunk.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.${param}.stderr -chunk ${param}.stderr -outfile ${tmp}.s -to images

  errmsg.csh ${verbosity} ${myname} 7 "failed command" ${status} "mri.rename.chunk.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.${param}.stderr -chunk ${param}.stderr -outfile ${tmp}.s -to images"
  if ( ${status} ) exit ${status}

  # compute relative error := ( abs(mean) > zero ) ? stderr/abs(mean) : stderr
  #
  mri_rpn_math -o ${tmp} -c images '$1,$2,abs,/,$1,$2,abs,'${zero}',>,if_keep' ${tmp}.s ${tmp}.e >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 8 "failed command" ${status} "mri_rpn_math -o ${tmp} -c images '"'$1,$2,abs,/,$1,$2,abs,'${zero}',>,if_keep'"' ${tmp}.s ${tmp}.e >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  # plot relative error
  #
  setenv F_PS_TITLE3 "${param} relative error"

  # mri_to_img -mosaic -title -date_created -date_printed -reverse -gray_scale ${plotlimits} ${tmp} ${submodel}.${param}.re.ps

  mri_to_img -o ${submodel}.${param}.re.ps -mosaic -t -c -p -r -g ${plotlimits} ${tmp} >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 9 "failed command" ${status} "mri_to_img -o ${submodel}.${param}.re.ps -mosaic -t -c -p -r -g ${plotlimits} ${tmp} >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  label_skipR:
 end

# cleanup
#
 rm -f ${tmp}* >& /dev/null; exit 0
