#!/bin/tcsh -f

# ---------------------------------------------------------------------------
#
# try -help for details
#
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

# defaults
#
 set myname = ${0:t}

# temporary files prefix
#
 mkdir -p /tmp/${user}
 set tmpdir = /tmp/${user}/
 set tmp    = ${tmpdir}`echo ${myname} | sed 's/\.csh$//'`.`date '+%m%d%T' | sed 's/://g'`

# verbosity
#
 set verbmax   = 1
 set verbosity = ${verbmax}

# vmp output prefix, submodels to process
#
 set prefix    = './'
 set submodels = ''

# misc
#
 set param      = noiselevel
 set plotlimits = 'auto:auto'

 set plot_estimate
 set plot_stderr
 set plot_relerr

# numerical zero for relative error calculation
#
 set zero = 1.0e-8

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${myname}
    echo ' Make noise level/stderr images from vmp output.'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo ' [-ne|-noestimate] [-ns|-nostderr] [-nr|-norelerr]'
    echo ' [-pl|-plotlimits [lower][:[upper]]]'
    echo ' [-p|-prefix any_path/prefix]'
    echo ' submodel1 submodel2 ...'
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination'
    echo ' 1: abnormal termination, except'
    echo ' 2-: failed utility/script'
    echo ''
    echo 'Required utilities/scripts'
    echo ' errmsg.csh'
    echo ' mri.rename.chunk.csh param.block.info.csh proc.rangeopts.csh'
    echo ' mri_printfield mri_rpn_math mri_to_img'
#   echo ''
#   echo 'Defaults'
#   echo " prefix: '${prefix}'"
#   echo ''
#   echo " verbosity: ${verbosity} (maximum ${verbmax})"
#   echo " temporary files prefix: ${tmp}"
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

   case "-p":
   case "-prefix":
    set prefix = `echo ${argv[2]} | sed 's/\.mri$//'`;
    shift; shift; breaksw;

   case "-pl":
   case "-plotlimits":
    set plotlimits = ${argv[2]}; shift; shift; breaksw;

   case "-ne":
   case "-noestimate":
    unset plot_estimate; shift; breaksw;

   case "-ns":
   case "-nostderr":
    unset plot_stderr; shift; breaksw;

   case "-nr":
   case "-norelerr":
    unset plot_relerr; shift; breaksw;

   default:
    set submodels = ( ${submodels} ${argv[1]} ); shift; breaksw;
  endsw
 end

# error checks
#
 if ( !( ${?plot_estimate} || ${?plot_stderr} || ${?plot_relerr} ) || ( "${submodels}" == "" ) ) then
# errmsg.csh ${verbosity} ${myname} 0 "nothing to plot" 1 ""
# exit ${status}
  exit 0
 endif

# convert plot range options into a form that can be passed on to mri_to_img
#
 set plotlimits = ( `proc.rangeopts.csh -verbosity ${verbosity} -defaults "auto:auto" ${plotlimits}` )

 errmsg.csh ${verbosity} ${myname} 2 "failed command" ${status} "proc.rangeopts.csh -verbosity ${verbosity} -defaults "auto:auto" ${plotlimits}"
 if ( ${status} ) exit ${status}

 if ( "${plotlimits[1]}" != "auto" ) then
  set plotlimits[1] = "-n ${plotlimits[1]}"
 else
  set plotlimits[1] = ""
 endif

 if ( "${plotlimits[2]}" != "auto" ) then
  set plotlimits[2] = "-x ${plotlimits[2]}"
 else
  set plotlimits[2] = ""
 endif

# loop over submodels
#
 foreach submodel ( ${submodels} )

  # parameter names and offsets
  #
  # if there are more than one parameters in the "noise" block,
  # assume that the first one represents noise precision
  # and needs to be plotted as noiselevel.
  #
  param.block.info.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.estimates -block noise -outfile ${tmp}

  errmsg.csh ${verbosity} ${myname} 3 "failed command" ${status} "param.block.info.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.estimates -block noise -outfile ${tmp}"
  if ( ${status} ) exit ${status}

  set poffset = `head -1 ${tmp} | awk '{print $2}'`
  set pname   = ${param}

  # image titles
  #
  set slices = `mri_printfield -input ${prefix}${submodel}.estimates -field estimates.extent.z`

  errmsg.csh ${verbosity} ${myname} 4 "failed command" ${status} "mri_printfield -input ${prefix}${submodel}.estimates -field estimates.extent.z"
  if ( ${status} ) exit ${status}

  setenv F_PS_TITLE1 `mri_printfield -input ${prefix}${submodel}.estimates -field imagedata.description`
  setenv F_PS_TITLE2 "submodel: ${submodel}, slices: ${slices}"

  # noise level estimate
  #
  mri.rename.chunk.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.estimates -chunk estimates -outfile ${tmp} -to images

  errmsg.csh ${verbosity} ${myname} 5 "failed command" ${status} "mri.rename.chunk.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.estimates -chunk estimates -outfile ${tmp} -to images"
  if ( ${status} ) exit ${status}

  # extract noise precision estimate
  #
  mri_subset -d w -s ${poffset} -l 1 ${tmp} ${tmp}.e >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 6 "failed command" ${status} "mri_subset -d w -s ${poffset} -l 1 ${tmp} ${tmp}.e >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  # compute noise level := 1.0 / sqrt( noise precision )
  #
  mri_rpn_math -o ${tmp}.nl -c images '1.0,$1,sqrt,/' ${tmp}.e >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 7 "failed command" ${status} "mri_rpn_math -o ${tmp}.nl -c images '"'1.0,$1,sqrt,/'"' ${tmp}.e >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  # make noise level image
  #
  if ( !( ${?plot_estimate} ) ) goto label_SE

  setenv F_PS_TITLE3 "${pname}"

  mri_to_img -o ${submodel}.${pname}.ps -mosaic -t -c -p -r -g ${plotlimits} ${tmp}.nl >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 8 "failed command" ${status} "mri_to_img -o ${submodel}.${pname}.ps -mosaic -t -c -p -r -g ${plotlimits} ${tmp}.nl >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  label_SE:

  # noise level standard error
  #
  mri.rename.chunk.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.standard.errors -chunk standard.errors -outfile ${tmp} -to images

  errmsg.csh ${verbosity} ${myname} 9 "failed command" ${status} "mri.rename.chunk.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.standard.errors -chunk standard.errors -outfile ${tmp} -to images"
  if ( ${status} ) exit ${status}

  # extract standard error
  #
  mri_subset -d w -s ${poffset} -l 1 ${tmp} ${tmp}.s >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 10 "failed command" ${status} "mri_subset -d w -s ${poffset} -l 1 ${tmp} ${tmp}.s >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  # compute noise level standard error
  # := abs( 0.5 * ( noise precision standard error ) * ( noise level )^3 )
  #
  mri_rpn_math -o ${tmp}.nlse -c images '0.5,$1,*,$1,*,$1,*,$2,*' ${tmp}.nl ${tmp}.s >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 11 "failed command" ${status} "mri_rpn_math -o ${tmp}.nlse -c images '"'0.5,$1,*,$1,*,$1,*,$2,*'"' ${tmp}.nl ${tmp}.s >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  # make noise level standard error image
  #
  if ( !( ${?plot_stderr} ) ) goto label_RE

  setenv F_PS_TITLE3 "${pname} standard error"

  mri_to_img -o ${submodel}.${pname}.se.ps -mosaic -t -c -p -r -g ${plotlimits} ${tmp}.nlse >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 12 "failed command" ${status} "mri_to_img -o ${submodel}.${pname}.se.ps -mosaic -t -c -p -r -g ${plotlimits} ${tmp}.nlse >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  label_RE:

  if ( !( ${?plot_relerr} ) ) goto label_skipRE

  # compute relative error := ( abs(mean) > zero ) ? stderr/abs(mean) : stderr
  #
  mri_rpn_math -o ${tmp}.nlre -c images '$1,$2,abs,/,$1,$2,abs,'${zero}',>,if_keep' ${tmp}.nlse ${tmp}.nl >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 13 "failed command" ${status} "mri_rpn_math -o ${tmp}.nlre -c images '"'$1,$2,abs,/,$1,$2,abs,'${zero}',>,if_keep'"' ${tmp}.nlse ${tmp}.nl >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  # make noise level relative error image
  #
  setenv F_PS_TITLE3 "${pname} relative error"
 
  mri_to_img -o ${submodel}.${pname}.re.ps -mosaic -t -c -p -r -g ${plotlimits} ${tmp}.nlre >& ${tmp}.err
 
  errmsg.csh ${verbosity} ${myname} 14 "failed command" ${status} "mri_to_img -o ${submodel}.${pname}.re.ps -mosaic -t -c -p -r -g ${plotlimits} ${tmp}.nlre >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  label_skipRE:
 end

# cleanup
#
 rm -f ${tmp}* >& /dev/null; exit 0
