#!/bin/tcsh -f

# ---------------------------------------------------------------------------
#
# try -help for details
#
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

# defaults
#
 set myname = ${0:t}

# temporary files prefix
#
 mkdir -p /tmp/${user}
 set tmpdir = /tmp/${user}/
 set tmp    = ${tmpdir}`echo ${myname} | sed 's/\.csh$//'`.`date '+%m%d%T' | sed 's/://g'`

# verbosity
#
 set verbmax   = 1
 set verbosity = ${verbmax}

# misc
#
 set indspecs  = ''
 set prefix    = './'
 set submodels = ()
 set pretty

# standard file/chunk names
#
 set f_modelp    = model.probabilities
 set f_pdef      = log.determinant.lapackinfo

 set f_lnbf      = log.bayes.factors
 set f_lndet     = log.determinant
 set f_lnpm      = log.posterior.max
 set f_li        = log.integral

 set f_estimates = estimates

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${myname}
    echo ' Produce,  for a voxel,  a  submodelwise summary  of  model'
    echo ' probabilities and related diagnostics:  model probability,'
    echo ' fit positive definiteness, log Bayes factor, log integral,'
    echo ' log posterior value at maximum, log determinant, number of'
    echo ' parameters.'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo ' [-n|-nopretty]'
    echo ' [-o|-out output_file]'
    echo ' [-p|-prefix any_directory_prefix]'
    echo ' [-s|-submodels s1,s2,s3,...]'
    echo ' [indspecs]'
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination'
    echo ' 1: abnormal termination, except'
    echo ' 2: invalid voxel specification'
    echo ' 3: submodel information not available'
    echo ''
    echo 'Required utilities/scripts'
    echo ' errmsg.csh'
    echo ' mri.print.record.csh param.block.info.csh proc.indspecs.csh pretty.print.csh'
    echo ' mri_printfield'
    echo ''
    echo 'Notes'
    echo ' 1 Refer to proc.indspecs.csh for index specification format.'
#   echo ''
#   echo 'Defaults'
#   echo " verbosity: ${verbosity} (maximum ${verbmax})"
#   echo " temporary files prefix: ${tmp}"
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

   case "-s":
   case "-submodels":
    set submodels = ( `echo ${argv[2]} | sed 's/,/ /g'` ); shift; shift; breaksw;

   case "-p":
   case "-prefix":
    set prefix = ${argv[2]}; shift; shift; breaksw;

   case "-o":
   case "-out":
    set out = ${argv[2]}; shift; shift; breaksw;

   case "-n":
   case "-nopretty":
    unset pretty; shift; breaksw;

   default:
    set indspecs = ${argv[1]}; shift; breaksw;
  endsw
 end

# voxel
#
 proc.indspecs.csh -verbosity ${verbosity} -infile ${prefix}${f_modelp} -chunk ${f_modelp} ${indspecs} > ${tmp}

 errmsg.csh ${verbosity} ${myname} 2 "invalid voxel" ${status} ""
 if ( ${status} ) exit ${status}

 set indspecs = ( `grep '^[xyz]' ${tmp} | sort | awk '{print $1"="$2}'` )
 set indspecs = `echo ${indspecs} | sed 's/ /,/g'`

# submodels
#
 if ( "${submodels}" == "" ) then
  set fitsubmodels = ( `mri_printfield -input ${prefix}${f_modelp} -field submodel.list` )

  errmsg.csh ${verbosity} ${myname} 3 "submodel information not available; failed command:" ${status} "mri_printfield -input ${prefix}${f_modelp} -field submodel.list"
  if ( ${status} ) exit ${status}

  set sm        = ( ${submodels} )
  set submodels = ''
  foreach s ( ${sm} )
   if ( `echo '' | awk '{print index("'"${fitsubmodels}"'","'${s}'")}'` != 0 ) set submodels = ( ${submodels} ${s} )
  end

  if ( "${submodels}" == "" ) set submodels = ( ${fitsubmodels} )
 endif

# headers
#
 echo -n '' > ${tmp}
 if ( ${?pretty} ) echo 'Submodel ModelProbability PlusDef LnBF LnIntegral LnPostMax LnDet NumParams\n' > ${tmp}

# model probabilities
#
 if ( -f ${prefix}${f_modelp}.mri ) then
  set mp = ( `mri.print.record.csh -verbosity ${verbosity} -timecourse -infile ${prefix}${f_modelp} -chunk ${f_modelp} ${indspecs}` )

  if ( ${status} ) set mp = ( `echo ${#submodels} | awk '{for(i=1;i<=$1;i++)printf " (NA)"}'` )

  set mp = ( ${mp} '(NA)' )
 endif

# log bayes factors w.r.t null submodel
#
 if ( -f ${prefix}${f_lnbf}.mri ) then
  set lnbf = ( `mri.print.record.csh -verbosity ${verbosity} -timecourse -infile ${prefix}${f_lnbf} -chunk ${f_lnbf} ${indspecs}` )

  if ( ${status} ) set lnbf = ( `echo ${#submodels} | awk '{for(i=1;i<=$1;i++)printf " (NA)"}'` )

  set lnbf = ( ${lnbf} '(NA)' )
 endif

# submodelwise model probabilities and related diagnostics
#
 foreach submodel ( ${submodels} )
  if ( ${submodel} == "average" ) goto label_skip_submodel

  set imp = `mri_printfield -nofail -input ${prefix}${f_modelp} -field submodel.index.${submodel}`

  if ( ${status} ) then
   set imp  = ${#mp}
  else
   @   imp += 1
  endif

  set ilnbf = `mri_printfield -nofail -input ${prefix}${f_lnbf} -field submodel.index.${submodel}`

  if ( ${status} ) then
   set ilnbf  = ${#lnbf}
  else
   @   ilnbf += 1
  endif

  # positive definiteness
  #
  set pdef = '(NA)'
  if ( -f ${prefix}${submodel}.${f_pdef}.mri ) then
   set pdef = `mri.print.record.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.${f_pdef} -chunk ${f_pdef} ${indspecs}`

   if ( !( ${status} ) ) then
    if ( ${pdef} == 0 ) then
     set pdef = yes
    else
     set pdef = no
    endif
   endif
  endif

  # number of parameters
  #
  set npar = '(NA)'
  if ( -f ${prefix}${submodel}.estimates.mri ) then
   param.block.info.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.estimates >& ${tmp}.bi

   if ( !( ${status} ) ) set npar = `awk '{s+=$3}END{print s}' ${tmp}.bi`
  endif

  # log determinant
  #
  set lndet = '(NA)'
  if ( -f ${prefix}${submodel}.${f_lndet}.mri ) then
   set lndet = `mri.print.record.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.${f_lndet} -chunk ${f_lndet} ${indspecs}`

   if ( ${status} ) set lndet = '(NA)'
  endif

  # log posterior value at maximum
  #
  set lnpm = '(NA)'
  if ( -f ${prefix}${submodel}.${f_lnpm}.mri ) then
   set lnpm = `mri.print.record.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.${f_lnpm} -chunk ${f_lnpm} ${indspecs}`

   if ( ${status} ) set lnpm = '(NA)'
  endif

  # log integral
  #
  set li = '(NA)'
  if ( -f ${prefix}${submodel}.${f_li}.mri ) then
   set li = `mri.print.record.csh -verbosity ${verbosity} -infile ${prefix}${submodel}.${f_li} -chunk ${f_li} ${indspecs}`

   if ( ${status} ) set li = '(NA)'
  endif

  echo ${submodel} ${mp[${imp}]} ${pdef} ${lnbf[${ilnbf}]} ${li} ${lnpm} ${lndet} ${npar} >> ${tmp}

  label_skip_submodel:
 end

# pretty format
#
 if ( ${?pretty} ) then
  pretty.print.csh ${tmp} > ${tmp}.1
  mv ${tmp}.1 ${tmp}
 endif

# screendump if no output file is specified
#
 if ( ${?out} ) then
  echo -n '' > ${out}
  if ( ${?pretty} ) echo "Voxel ${indspecs}\n" > ${out}
  cat ${tmp} >> ${out}
 else
  if ( ${?pretty} ) echo "Voxel ${indspecs}\n"
  cat ${tmp}
 endif

# cleanup
#
 rm -f ${tmp}* >& /dev/null
