# Spectral density analysis functions from
# the book by Sam Efromovich; http://math.unm.edu/~efrom/book1/

ab.vec <- function( f = NA )
 {
  n   <- length( f ) + 1.
  f   <- c( f, 0. )
  vec <- c( 0., 0. )

  if ( all( f > 0. ) )
   {
    vec <- c( vec, 1., n )
   }
  else
   {
    seq.pos <- ( 1.:n )[f > 0.]
    seq.neg <- ( 1.:n )[f <= 0.]
    seq.neg <- seq.neg[seq.neg > 1.]
    a <- 1.
    while ( length( seq.pos ) * length( seq.neg ) > 0. )
     {
      if ( f[a + 1.] > 0. )
       {
        b <- min( seq.neg ) - 1.
       }
      else
       {
        a <- min( seq.pos )
        seq.neg <- seq.neg[seq.neg > a]
        b <- min( seq.neg ) - 1.
       }
      vec <- c( vec, a, b )
      seq.pos <- seq.pos[seq.pos > b]
      seq.neg <- seq.neg[seq.neg > b]
      a <- b
     }
   }

  if ( vec[length( vec )] == n )
   {
    vec[length( vec )] <- n - 1.
   }

  vec
 }

rem.bump1 <- function( f = NA, AREASQ = NA, coef = 1. )
 {
  n   <- length( f )
  vec <- ab.vec( f )

  if ( length( vec ) > 2. )
   {
    vec <- vec[ - c( 1., 2. )]
    k   <- length( vec ) / 2.
    for ( s in 1.:k )
     {
      if ( sum( ( f[vec[2. * s - 1.]:vec[2. * s]] )^2. ) / n <= coef * AREASQ )
       {
        f[vec[2. * s - 1.]:vec[2. * s]] <- 0.
       }
     }
   }

  f
 }

negden <- function( f = NA, delta = 0.01, FLAGBUMP = 1., cB = 2. )
 {
  flag <- 0.
  f1   <- f
  k    <- length( f )
  AREA <- ( k/( k - 1. ) ) * mean( f ) - ( f[1.] + f[k] )/( 2. * ( k - 1. ) )

  if ( all( f >= 0. ) )
   {
    flag <- 1.
   }

  if ( all( f <= 2. * delta ) | ( AREA <= 2. * delta ) )
   {
    flag <- 2.
   }

   while ( flag == 0. )
    {
     f <- f - delta
     f[f < 0.] <- 0.
     int <- ( k/( k - 1. ) ) * mean( f ) - ( f[1.] + f[k] )/( 2. * ( k - 1. ) )
     if ( int <= AREA )
      {
       if ( int > ( 10. * delta ) )
        {
         f <- f * ( AREA/int )
        }
       flag <- 1.
      }
    }

   if ( FLAGBUMP == 1. )
    {
     AREASQ <- mean( ( f - f1 )^2. )
     f      <- rem.bump1( f = f, AREASQ = AREASQ, coef = cB )
    }

   if ( flag == 1. )
    {
     if ( mean( f ) > ( 10. * delta ) )
      {
       f <- f * ( AREA/mean( f ) )
      }
    }

   f[f < 0.] <- 0.

   f
 }

Updiag <- function( J = 2. )
 {
  m <- matrix( 1., nrow = J, ncol = J )
  m[row( m ) > col( m )] <- 0.
  m
 }

estspden <- function( X   = NA, TT  = NA,  FLAGMIS = 0., knots = 100.,
                      cJ0 = 4., cJ1 = 0.5, cJM     = 6., cT    = 4., cB = 2., FLAGNEG = 1. )
 {
  n <- length( X )
  if ( FLAGMIS == 1. )
   {
    n <- sum( TT )
   }

  JMAX   <- ceiling( cJ0 + cJ1 * logb( n + 3. ) )
  theta1 <- acf( X, lag.max = cJM * JMAX + 1., type = "covariance", plot = F )$acf[, 1., 1.]

  if ( FLAGMIS == 1. )
   {
    cc     <- acf( TT, cJM * JMAX + 1., type = "covariance", plot = F )$acf[, 1., 1.] + mean( TT )^2.
    theta1 <- ( theta1 + mean( X )^2. )/cc
   }

  d <- 2. * sum( theta1[1.:JMAX]^2. ) - theta1[1.]^2.
  thetasq1 <- theta1^2. - d/n
  thetasq1[thetasq1 < 0.] <- 0.
  theta <- theta1[1.:JMAX]
  thetasq <- thetasq1[1.:JMAX]
  error <- matrix( d/n - thetasq, nrow = 1. ) %*% Updiag( JMAX )
  J <- order( error )[1.]
  theta <- theta[1.:J]
  thetasq <- thetasq[1.:J]
  arg <- outer( seq( 0., pi, len = knots ), ( 0.:( cJM * JMAX ) ) )
  bas <- cos( arg )/pi
  theta <- ( theta * thetasq )/( thetasq + ( d/n ) )
  theta[1.] <- theta[1.]/2.

  if ( cJM <= 1. )
   {
    if ( JMAX > J )
     {
      theta <- c( theta, rep( 0., JMAX - J + 1. ) )
     }
   }

  if ( cJM > 1. )
   {
    rest.theta <- theta1[( J + 1. ):( cJM * JMAX + 1. )]
    rest.theta[rest.theta^2. < ( cT * d * logb( n + 3. ) )/n] <- 0.
    theta <- c( theta, rest.theta )
   }

  f <- bas %*% theta

  if ( FLAGNEG == 1. )
   {
    f <- negden( f, FLAGBUMP = 1., cB = cB )
   }
  else
   {
    f[f < 0.] <- 0.
   }

  f
 }
