#!/bin/csh -f

# defaults
#
 set tmp = tmp.f2b

 set verbmax   = 1
 set verbosity = ${verbmax}

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${0}
    echo ' Experimental script for generating a template brain input file'
    echo ' using  any  of the  fiasco scripts  {prefix}.local.csh,  where'
    echo ' {prefix}  stands for  epi,  epi_brain,  ts,  ts_brain, spiral,'
    echo ' spiral_brain.'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo 'Defaults'
    echo " verbosity: ${verbosity} (maximum ${verbmax})"
    echo ' image datachunk name: images'
    echo ' process summary using latex; cleanup all the mess.'
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination'
    echo ' 1: abnormal termination, except'
    echo ' 2: unspecified/invalid fiasco script'
    echo ' 3-: failed script/utility'
    echo ''
    echo 'Temporary files prefix'
    echo " ${tmp}"
    echo ''
    echo 'Notes'
    echo ' 1 Output needs to be hand-tuned before use;  e. g.,  there is'
    echo '   no easy way of generating keys  prior.probability.s through'
    echo '   this script, where  s  is a binary key encoding a  submodel'
    echo '   (for  two  conditions: s  =  00, 01, 10, 11).  This  script'
    echo '   introduces  these   key-value pairs assuming two conditions'
    echo '   and  arbitrary numbers for the  prior probabilities.  Also,'
    echo '   e. g., paths to files need to be hand-checked.
    echo ' 2 See Brain/misc/input.specification for details of the brain'
    echo '   input specification format.
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

#  case "-s":
#  case "-submodel":
#   set submodel = ${argv[2]}; shift; shift; breaksw;

   default:
#   set indspecs = ${argv[1]}; shift; breaksw;
  endsw
 end

# error checks
#
 if ( !( ${?master} ) ) then
  if ( ${verbosity} > 0 ) echo ${0}':' unspecified/invalid BRAIN output file
  exit 3
 else if ( !( -f ${master} ) ) then
  if ( ${verbosity} > 0 ) echo ${0}':' unspecified/invalid BRAIN output file
  exit 3
 else if ( !( ${?submodel} ) ) then
  if ( ${verbosity} > 0 ) echo ${0}':' unspecified submodel
  exit 4
 endif

# .local.csh filename
#
 set localcsh = epi.local.csh
 if ( ${#argv} > 0 ) set localcsh = $1

# source localcsh
#
 set F_SPLIT_FILE = ${localcsh}.design
 source ${localcsh}

# image data
#
 set imagedata = ${F_DIR}/data/outlier-vtxyz.mri

# acquisition: standard fiasco scripts do not have any explicit
# variables for number of shots, tr, te, voxel dimensions, etc.
#
# Here we look for environment variables NSHOTS, TR, SLICEORDER,
# in case they are defined, else construct reasonable defaults.
# 
# If NSHOTS is not defined, this looks for the clues "reader"
# and "twice" on the same line in the script: "reader sees twice as many..."
# which indicates NSHOTS = 2. Else it is set to 1.
#
 if ( "`printenv NSHOTS`" == "" ) then
  if ( "`grep 'twice' ${localcsh} | grep 'reader'`" != "" ) then
   set NSHOTS = 2
  else
   set NSHOTS = 1
  endif
 endif

# acquisition: tr
#
 if ( "`printenv TR`" == "" ) set TR = `echo ${F_IAI} ${NSHOTS} | awk '{printf "%25.15e", $1 / $2}'`

# acquisition: slice order
#
# default is 0 2 4 ... 1 3 5 ...
#
 set sliceorder = ( `printenv SLICEORDER` )
 if ( "${sliceorder}" == "" ) then
  set sliceorder = "0"
  @ s = 2
  while ( ${s} < ${F_NSLICE} )
   set sliceorder = "$sliceorder ${s}"
   @ s += 2
  end
  @ s = 1
  while ( ${s} < ${F_NSLICE} )
   set sliceorder = "$sliceorder ${s}"
   @ s += 2
  end
 endif

# write everything except experiment-design related stuff
#
cat << @EOF
!format = pgh
!version = 1.0

master.output = brain.out.mri

imagedata.description = "${F_DESCRIPTION}"
thisrun.description = "analysis with default fiasco2brain.csh settings"

imagedata = ${imagedata}
missing.specs = "file ${imagedata} end"

voxelRange.x = ":"
voxelRange.y = ":"
voxelRange.z = ":"

acquisition.numShots   = ${NSHOTS}
acquisition.repeatTime = ${TR}
acquisition.iai        = ${F_IAI}
acquisition.sliceOrder = "${sliceorder}"

driftBasis.order    = 4
driftBasis.numKnots = 26

submodels = "+ name all end"

prior.baseline.scale = 1.0e-6
prior.baseline.shift = 2.0e+3

prior.drift.norm.penalty         = 0.0
prior.drift.smoothing.df         = 5.0
prior.drift.smoothing.choice     = 0
prior.drift.smoothing.iterations = 50
prior.drift.smoothing.tolerance  = 1.0e-13

prior.noise.a = 1.8
prior.noise.b = 2.0e+2

params.bounds.drift    = "none"
params.bounds.noise    = "lower 1.0e-12"
params.bounds.response = "lower 0.0"
params.bounds.shape    = "lower 1.0e-08"

params.scale.drift    = 1.0e+0
params.scale.noise    = 1.0e+1
params.scale.response = 1.0e+0
params.scale.shape    = 1.0e-1

init.scheme        =  0
init.maxIterations = 10

init.minimizer.maxCorrections = 5
init.minimizer.maxIterations  = 25
init.minimizer.maxEvaluations = 50
init.minimizer.maxPrGradNorm  = 1.0e-7
init.minimizer.minRelReduceF  = 0.0
init.minimizer.restartOpts    = "10 off false 0.8"
init.minimizer.verbosity      = -1

minimizer.maxCorrections = 20
minimizer.maxIterations  = 5000
minimizer.maxEvaluations = 10000
minimizer.maxPrGradNorm  = 1.0e-10
minimizer.minRelReduceF  = 1.0e-5
minimizer.restartOpts    = "50 on true 0.8"
minimizer.verbosity      = -1

output.prefix = ${F_DIR}/brain/
output.products = "estimates standard.errors covariances baseline baseline.stderr fit residuals log.bayes.factors model.probabilities averages"
output.type.log.integral = float64

prior.probability.00 = 1.0
prior.probability.01 = 0.0025
prior.probability.10 = 0.0025
@EOF

# experiment design
#
 if ( `grep -c bynumimages ${F_SPLIT_FILE}` == 1 ) then
  grep -v -e 'bynumimages' -e 'Condition' ${F_SPLIT_FILE} > ${tmp}.${F_SPLIT_FILE}
 
  set conditions = ( `cat ${tmp}.${F_SPLIT_FILE} | awk '{print $1}' | sort | uniq` )
 
  set tags = "T0"
  set bells = "UN"
  @ t = 1
  while ( ${t} < ${#conditions} )
   set tags = ( ${tags} T${t} )
   set bells = ( ${bells} UN )
   @ t += 1
  end
 
  echo ''
  echo design.conditions       = \"${conditions}\"
  echo design.conditions.tags  = \"${tags}\"
  echo design.conditions.bells = \"${bells}\"
 
  set fullclist   = ( `cat ${tmp}.${F_SPLIT_FILE} | awk '{print $1}'` )
  set fulllengths = ( `cat ${tmp}.${F_SPLIT_FILE} | awk '{print $2}'` )
 
  set clist = ""
  set lengths = ""
  @ e = 1
  while ( ${e} <= ${#fulllengths} )
   if ( $fulllengths[$e] != 0 ) then
    set clist = ( ${clist} $fullclist[$e] )
    set el = `echo $fulllengths[$e] ${F_IAI} | awk '{print $1 * $2}'`
    set lengths = ( ${lengths} $el )
   endif
   @ e += 1
  end
 
  set keyprefix = "design.epochs"
 
  set prespconsta = ""
  set prespconstb = ""
 
  foreach c ( ${conditions} )
   set prespconsta = ( ${prespconsta} 1.0 )
   set prespconstb = ( ${prespconstb} 50.0 )
 
   set epochbegin = ""
   set epochlength = ""
 
   @ e = 1
   @ b = 0
   while ( ${e} <= ${#lengths} )
    if ( "$clist[$e]" == "${c}" ) then
     set epochbegin = ( ${epochbegin} ${b} )
     set epochlength = ( ${epochlength} $lengths[$e] )
    endif
    @ b += $lengths[$e]
    @ e += 1
   end
 
   echo ''
   echo ${keyprefix}.begin.${c} = \"${epochbegin}\"
   echo ${keyprefix}.length.${c} = \"${epochlength}\"
  end
 
  echo ''
  echo prior.response.constA = \"${prespconsta}\"
  echo prior.response.constB = \"${prespconstb}\"
 
  set uniqtags = ( `echo $tags | sort | uniq` )
 
  foreach t ( $uniqtags )
   echo ''
   echo prior.shape.constA.${t} = \"4.00 4.00\"
   echo prior.shape.constB.${t} = \"1.00 0.43\"
  end
 endif
 
 /bin/rm ${F_SPLIT_FILE} ${tmp}.${F_SPLIT_FILE}
