#!/bin/tcsh -f

# defaults
#
 set myname = `basename ${0}` # name

# temporary files prefix
#
#if ( ${?TMPPREF} ) then
# set tmp = ${TMPPREF}
#else
  mkdir -p /tmp/`whoami`
  set tmpdir = /tmp/`whoami`/
  set tmp    = ${tmpdir}`echo ${myname} | sed 's/\.csh$//'`.`date '+%m%d%T' | sed 's/://g'`
#endif

 # echo tmp \"$tmp\"

# verbosity
#
 set verbmax   = 1
 set verbosity = ${verbmax}

# brain output prefix, submodels to exclude
#
 set prefix   = './'
 set excluded = ''

# probabilities file/chunk names
#
 set mpchunk = model.probabilities

# zero
#
 set zero = 1.0e-8

# command line args
#
 if ( ${#argv} == 0 ) set argv = -help

 while ( ${#argv} > 0 )
  switch ( ${argv[1]} )
   case "-h":
   case "-help":
    echo ${0}
    echo ' Compute submodel-averaged estimates and standard errors'
    echo ' from BRAIN output.'
    echo ''
    echo 'Command line'
    echo ' [-h|-help]'
    echo ' [-q|-quiet] [-v|-verbosity level] [-V|-verbose]'
    echo ''
    echo ' [-p|-prefix input_prefix]'
    echo ' [-oe|-oute output_file_for_estimates]'
    echo ' [-os|-outs output_file_for_stderrors]'
    echo ' [-e|-exclude submodel1,submodel2,...]'
    echo ' [-z|-zero numerical_zero]'
    echo ' quantity'
#   echo ''
#   echo 'Defaults'
#   echo " verbosity: ${verbosity} (maximum ${verbmax})"
#   echo " prefix: '${prefix}'"
    echo ''
    echo 'Termination status codes'
    echo ' 0: normal termination'
    echo ' 1: abnormal termination, except'
    echo ' 2: unspecified quantity to average'
    echo ' 3: nothing to average over (all submodels excluded)'
    echo ' 4-: failed utility/script'
    echo ''
    echo 'Required utilities/scripts'
    echo ' errmsg.csh'
    echo ' mrirenchunk.csh extract.mp.csh'
    echo ' mri_rpn_math'
#   echo ''
#   echo 'Temporary files prefix'
#   echo " ${tmp}"
    echo ''
    echo 'Notes'
    echo ' 1 "-exclude submodel1,submodel2,..."  excludes  all the'
    echo '   specified submodels from averaging  and  renormalizes'
    echo '   probabilities for included submodels using'
    echo '   ( 1 - p(submodel1) - p(submodel2) - ... )^-1.'
    echo ' 2 "-zero z" specifies the numerical zero threshold used'
    echo '   in deciding whether |1-p(submodel1)-p(submodel2)-...|'
    echo '   should be treated as zero.'
    echo ' 3 Submodel-specific file names for quantity and standard'
    echo '   error are assumed to be of the form'
    echo '    (input_prefix)(submodel).(quantity).mri'
    echo '   and'
    echo '    (input_prefix)(submodel).(quantity).stderr.mri.'
    echo '   Corresponding chunk names are assumed to be'
    echo '   (quantity) and (quantity).stderr respectively.'
    exit 0;

   case "-q":
   case "-quiet":
    set verbosity = 0; shift; breaksw;

   case "-v":
   case "-verbosity":
    set verbosity = ${argv[2]}; shift; shift; breaksw;

   case "-V":
   case "-verbose":
    set verbosity = ${verbmax}; shift; breaksw;

   case "-oe":
   case "-oute":
    set oute = `echo ${argv[2]} | sed 's/\.mri$//'`;
    shift; shift; breaksw;

   case "-os":
   case "-outs":
    set outs = `echo ${argv[2]} | sed 's/\.mri$//'`;
    shift; shift; breaksw;

   case "-p":
   case "-prefix":
    set prefix = ${argv[2]}; shift; shift; breaksw;

   case "-z":
   case "-zero":
    set zero = ${argv[2]}; shift; shift; breaksw;

   case "-e":
   case "-exclude":
    set excluded = ( `echo ${argv[2]} | sed 's/,/ /g'` );
    shift; shift; breaksw;

   default:
    set quantity = ${argv[1]}; shift; breaksw;
  endsw
 end

# error checks
#
 if ( !( ${?quantity} ) ) then
  errmsg.csh ${verbosity} ${myname} 2 "unspecified quantity" 1 ""
  exit ${status}
 endif

 if ( !( ${?oute} ) ) set oute = average.${quantity}
 if ( !( ${?outs} ) ) set outs = average.${quantity}.stderr

# extract model probabilities
#
 extract.mp.csh -verbosity ${verbosity} -prefix ${prefix} -outprefix ${tmp}. -outsuffix '' -outchunk ${quantity} >& ${tmp}.mp

 errmsg.csh ${verbosity} ${myname} 4 "failed command" ${status} "extract.mp.csh -verbosity ${verbosity} -prefix ${prefix} -outprefix ${tmp}. -outsuffix '' -outchunk ${quantity} >& ${tmp}.mp"
 if ( ${status} ) exit ${status}

 # echo excluded \"$excluded\"

# excluded submodels list: throw away those not present in the actual analysis
#
 set xlist  = ""
 set togrep = ""
 set xfiles = ""
 foreach s ( ${excluded} )
  set ss = ( `grep "^submodel: ${s}" ${tmp}.mp | awk '{print $2" "$4}'` )
  if ( "${ss[1]}" == "${s}" ) then
   set xlist  = ( ${xlist} ${s} )
   set togrep = ( ${togrep} "-e '^submodel: ${s}'" )
   set xfiles = ( ${xfiles} ${ss[2]} )
  endif
 end
 set excluded = ( ${xlist} )

 # echo excluded \"$excluded\"
 # echo xfiles \"$xfiles\"
 # echo togrep \"$togrep\"

# submodel and filenames for model probabilities
#
 if ( ${#excluded} == 0 ) then
  set submodels = ( `awk '{print $2}' ${tmp}.mp` )
  set mfiles    = ( `awk '{print $4}' ${tmp}.mp` )
 else
  alias mygrep "grep -v ${togrep}"
  set submodels = ( `mygrep ${tmp}.mp | awk '{print $2}'` )
  set mfiles    = ( `mygrep ${tmp}.mp | awk '{print $4}'` )
  unalias mygrep
 endif

 if ( ${#submodels} == 0 ) then
  errmsg.csh ${verbosity} ${myname} 3 "nothing to average over: all submodels excluded" 1 ""
  exit ${status}
 endif

 # echo submodels \"$submodels\"
 # echo mfiles \"$mfiles\"

# submodelwise filenames for quantity estimates;
# rename standard errors chunks
#
 set efiles = ''
 set sfiles = ''
 set i = 1
 foreach s ( ${submodels} )
  set efiles = ( ${efiles} ${prefix}${s}.${quantity} )
  set sfiles = ( ${sfiles} ${tmp}.${s}.se            )

  mrirenchunk.csh -verbosity ${verbosity} -infile ${prefix}${s}.${quantity}.stderr -outfile ${sfiles[${i}]} -chunk ${quantity}.stderr -to ${quantity}

  errmsg.csh ${verbosity} ${myname} 5 "failed command" ${status} "mrirenchunk.csh -verbosity ${verbosity} -infile ${prefix}${s}.${quantity}.stderr -outfile ${sfiles[${i}]} -chunk ${quantity}.stderr -to ${quantity}"
  if ( ${status} ) exit ${status}

  @ i += 1
 end

 # echo efiles \"$efiles\"
 # echo sfiles \"$sfiles\"

# compute renormalization factors
#
 if ( ${#excluded} > 0 ) then
  mri_rpn_math -o ${tmp}.x -c ${quantity} '$1' ${xfiles[1]}

  set i = 2
  while ( ${i} <= ${#excluded} )
   mri_rpn_math -o ${tmp} -c ${quantity} '$1,$2,+' ${tmp}.x ${xfiles[${i}]} >& ${tmp}.err

   errmsg.csh ${verbosity} ${myname} 6 "failed command" ${status} "mri_rpn_math -o ${tmp} -c ${quantity} '"'$1,$2,+'"' ${tmp}.x ${xfiles[${i}]} >& ${tmp}.err"
   if ( ${status} ) exit ${status}

   mv ${tmp}.dat ${tmp}.x.dat

   @ i += 1
  end

  mri_rpn_math -o ${tmp}.n -c ${quantity} '1.0,1.0,$1,-,/,0.0,'${zero}',1.0,$1,-,abs,<=,if_keep' ${tmp}.x >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 7 "failed command" ${status} "mri_rpn_math -o ${tmp}.n -c ${quantity} '"'1.0,1.0,$1,-,/,0.0,'${zero}',1.0,$1,-,abs,<=,if_keep'"' ${tmp}.x >& ${tmp}.err"
  if ( ${status} ) exit ${status}
 else
  mri_rpn_math -o ${tmp}.n -c ${quantity} '1.0' ${mfiles[1]} >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 8 "failed command" ${status} "mri_rpn_math -o ${tmp}.n -c ${quantity} '1.0' ${mfiles[1]} >& ${tmp}.err"
  if ( ${status} ) exit ${status}
 endif

# submodel-averaged estimate
#  := Sum{submodel != excluded} p(submodel) (1-p(excluded))^-1 estimate(submodel)
#
# submodel-averaged variance
#  := Sum{submodel != excluded} p(submodel) (1-p(excluded))^-1 [ var(submodel) + estimate(submodel)^2 ]
#     - average_estimate^2
#
 mri_rpn_math -o ${oute} -c ${quantity} '$1,$2,$3,*,*' ${efiles[1]} ${mfiles[1]} ${tmp}.n >& ${tmp}.err

 errmsg.csh ${verbosity} ${myname} 9 "failed command" ${status} "mri_rpn_math -o ${oute} -c ${quantity} '$1,$2,$3,*,*' ${efiles[1]} ${mfiles[1]} ${tmp}.n >& ${tmp}.err"
 if ( ${status} ) exit ${status}

 mri_rpn_math -o ${tmp}.av -c ${quantity} '$1,dup,*,$2,dup,*,+,$3,$4,*,*' ${efiles[1]} ${sfiles[1]} ${mfiles[1]} ${tmp}.n >& ${tmp}.err

 errmsg.csh ${verbosity} ${myname} 10 "failed command" ${status} "mri_rpn_math -o ${tmp}.av -c ${quantity} '$1,dup,*,$2,dup,*,+,$3,$4,*,*' ${efiles[1]} ${sfiles[1]} ${mfiles[1]} ${tmp}.n >& ${tmp}.err"
 if ( ${status} ) exit ${status}

 set i = 2
 while ( ${i} <= ${#submodels} )
  # estimate
  #
  mri_rpn_math -o ${tmp}.p -c ${quantity} '$1,$2,$3,*,*' ${efiles[${i}]} ${mfiles[${i}]} ${tmp}.n >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 11 "failed command" ${status} "mri_rpn_math -o ${tmp}.p -c ${quantity} '$1,$2,$3,*,*' ${efiles[${i}]} ${mfiles[${i}]} ${tmp}.n >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  mri_rpn_math -o ${tmp} -c ${quantity} '$1,$2,+' ${oute} ${tmp}.p >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 12 "failed command" ${status} "mri_rpn_math -o ${tmp} -c ${quantity} '"'$1,$2,+'"' ${oute} ${tmp}.p >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  mv -f ${tmp}.dat ${oute}.dat

  # variance
  #
  mri_rpn_math -o ${tmp}.v -c ${quantity} '$1,dup,*,$2,dup,*,+,$3,$4,*,*' ${efiles[${i}]} ${sfiles[${i}]} ${mfiles[${i}]} ${tmp}.n >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 13 "failed command" ${status} "mri_rpn_math -o ${tmp}.v -c ${quantity} '$1,dup,*,$2,dup,*,+,$3,$4,*,*' ${efiles[${i}]} ${sfiles[${i}]} ${mfiles[${i}]} ${tmp}.n >& ${tmp}.err"
 if ( ${status} ) exit ${status}

  mri_rpn_math -o ${tmp} -c ${quantity} '$1,$2,+' ${tmp}.v ${tmp}.p >& ${tmp}.err

  errmsg.csh ${verbosity} ${myname} 14 "failed command" ${status} "mri_rpn_math -o ${tmp} -c ${quantity} '"'$1,$2,+'"' ${tmp}.v ${tmp}.p >& ${tmp}.err"
  if ( ${status} ) exit ${status}

  mv -f ${tmp}.dat ${tmp}.v.dat

  @ i += 1
 end

# model-averaged standard errors: second term;
# if any variance turns out to be < zero, set stderr to 2*|estimate|
#
 mri_rpn_math -o ${outs} -c ${quantity} '2.0,$2,abs,*,$1,$2,dup,*,-,dup,sqrt,swap,0.0,<=,if_keep' ${tmp}.v ${oute} >& ${tmp}.err

 errmsg.csh ${verbosity} ${myname} 15 "failed command" ${status} "mri_rpn_math -o ${outs} -c ${quantity} '2.0,$2,abs,*,$1,$2,dup,*,-,dup,sqrt,swap,0.0,<=,if_keep' ${tmp}.v ${oute} >& ${tmp}.err"
 if ( ${status} ) exit ${status}

# rename standard errors chunk
#
 strings ${outs}.mri | grep -v 'history' | sed 's/^'${quantity}'/'${quantity}'\.stderr/g' > ${tmp}.mri
 mv -f ${tmp}.mri ${outs}.mri

# cleanup
#
 rm -f ${tmp}* >& /dev/null; exit 0
