/* ----------------------------------------------------------------------------
 *
 * class DataNavigator
 *   A TCreader that combines another TCreader with a MissingInfo object
 *
 * required headers
 *   see include header file DataNavigator
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __DataNavigator_H__
#define __DataNavigator_H__

cpp_using_namespace( std );

template< class TCR >
class DataNavigator : public TCreader
 {
  public:

            // DataNavigator, a TCreader itself, is designed to be a wrapper for
            // another TCreader and a MissingInfo object. Time courses retrieved
            // by  DataNavigator  are filtered out  for instants that are marked
            // "missing".  In addition,  it provides a convenient iterator-based
            // interface to forward navigation. Template parameter TCR must be a
            // TChandler with the same read capabilities as TCreader or TCrw.

   // ----------------------------------------------------------------- iterator
   // --------------------------------------------------------------------------

   // DataNavigator  allows navigation through time course data  in  two ways: a
   // stand-alone mode common to all time course readers,  and  via an iterator.
   // Iterators allow forward iteration only.  The two  navigation modes  can be
   // intermixed.  Only one  active iterator instance  is permitted at any  time
   // (auto_ptr semantics):  copying an  active iterator makes it inactive,  the
   // copy is then the only active instance. In the same spirit, begin() returns
   // inactive  instances  until  the  destruction  of  the active instance.  An
   // inactive iterator is the same as end( ). Operator -> allows access to  the
   // underlying DataNavigator functions for full control over navigation.

   class iterator
    {
     public:

      // ---------------------------------------- construct/destruct/copy/assign
      // -----------------------------------------------------------------------

      iterator( );

      ~iterator( );

      iterator( iterator& another );

      iterator&
      operator=( iterator& rhs );

      // ----------------------------------------------------------- comparisons
      // -----------------------------------------------------------------------

      inline
      operator bool( ) const;

      inline
      bool
      operator==( const iterator& another ) const;

      inline
      bool
      operator!=( const iterator& another ) const;

      // ---------------------------------------------- increment/deref/location
      // -----------------------------------------------------------------------

      inline
      iterator&
      operator++( );
                                                     //
                                                     // DataNavigator::toNext( )

      inline
      const varray<double>&
      operator*( ) const;
                                                 //
                                                 // DataNavigator::timeCourse( )

      inline
      DataNavigator*
      operator->( ) const;

      inline
      LocationType
      operator()( ) const;
                                                   //
                                                   // DataNavigator::locSansT( )

     private:
      DataNavigator<TCR>* my_dn;
      explicit iterator( DataNavigator<TCR>* dn );
      friend class DataNavigator<TCR>;
    };

   // --------------------------------------------------- construct/reinitialize
   // --------------------------------------------------------------------------

   DataNavigator( );

   DataNavigator( const DataNavigator& another );

   DataNavigator&
   operator=( const DataNavigator& rhs );

   DataNavigator( const TCR*         tcr,
                  const MissingInfo* mi,
                  const string&      sliceDimName = "z" );

                // DataNavigator  acts as a wrapper for the  tcr and mi  objects
                // (must remain in existence throught the life of DataNavigator)

   void
   reinitialize( const TCR*         tcr,
                 const MissingInfo* mi,
                 const string&      sliceDimName = "z" );

   // --------------------------------------- voxel navigation through iterators
   // --------------------------------------------------------------------------

   inline
   iterator&
   begin( ) const;

   inline
   iterator
   end( ) const;

   // ---------------------------------------------- raw and edited time courses
   // --------------------------------------------------------------------------

   inline
   const varray<double>&
   timeCourse( ) const;
                                   //
                                   // time course *sans* instants marked missing

   inline
   const varray<double>&
   tcEntire( ) const;
                                    //
                                    // this is a hook to TCreader::timeCourse( )

   // -------------------------------------------- references to wrapped objects
   // --------------------------------------------------------------------------

   inline
   const MissingInfo&
   mi( ) const;

   // -------------------------------------------------- "slice" dimension index
   // --------------------------------------------------------------------------

   inline
   size_t
   sliceDim( ) const;

   // -------------------------------------------------------- seek and retrieve
   // --------------------------------------------------------------------------

   inline
   bool
   seek( ) const;

   inline
   bool
   retrieve( varray<double>& timeCourse ) const;

   inline
   bool
   retrieveFloat( varray<float>& timeCourse ) const;

   inline
   bool
   retrieveInt( varray<int>& timeCourse ) const;

   inline
   bool
   retrieveShort( varray<short>& timeCourse ) const;

   inline
   bool
   retrieveUchar( varray<unsigned char>& timeCourse ) const;

  // ---------------------------------------- data type of the stored timecourse
  // ---------------------------------------------------------------------------

   inline
   string
   dataType( ) const;

  // ------------------------------------------------------- explicit file close
  // ---------------------------------------------------------------------------

   inline
   void
   close( ) const;

  // ------------------------------------------------------------- redefinitions
  // ---------------------------------------------------------------------------

                                       // this is  to ensure that TCR is tightly
                                       // coupled with its DataNavigator wrapper

   inline
   bool
   toPrev( ) const;

   inline
   bool
   toNext( ) const;

   inline
   bool
   toLast( ) const;

   inline
   bool
   toFirst( ) const;

   inline
   long
   offset( ) const;

   inline
   long
   offsetB( ) const;

  private:
   const TCR*         my_tcr;
   const MissingInfo* my_mi;
         size_t       my_zdim;

   mutable iterator  my_begin;
   mutable varray<double> my_timeCourse;

   bool isConsistent( ) const;

   friend iterator::~iterator( );
   friend iterator& iterator::operator++( );
   friend iterator& iterator::operator=( iterator& rhs );
 };

// ------------------------------------------------------ construct/reinitialize
// -----------------------------------------------------------------------------

template< class TCR >
DataNavigator<TCR>::DataNavigator( )
 : TCreader( ),
   my_tcr( ),
   my_mi( ),
   my_zdim( ),
   my_begin( this ),
   my_timeCourse( )
 {
 }

template< class TCR >
DataNavigator<TCR>::DataNavigator( const DataNavigator& another )
 : TCreader( another ),
   my_tcr( another.my_tcr ),
   my_mi( another.my_mi ),
   my_zdim( another.my_zdim ),
   my_begin( this ),
   my_timeCourse( another.my_timeCourse )
 {
 }

template< class TCR >
DataNavigator<TCR>&
DataNavigator<TCR>::operator=( const DataNavigator& rhs )
 {
  if ( this != &rhs )
   {
    TCreader::operator=( rhs );

    my_tcr  = rhs.my_tcr;
    my_mi   = rhs.my_mi;
    my_zdim = rhs.my_zdim;
    
    my_timeCourse.resize( rhs.my_timeCourse.size( ) );
    my_timeCourse = rhs.my_timeCourse;
                                           //
                                           // ATTN this may reset rhs.anyMove( )

    // ATTN my_begin will be set/reset by iterator instances
   }

  return *this;
 }

template< class TCR >
DataNavigator<TCR>::DataNavigator( const TCR*         tcr,
                                   const MissingInfo* mi,
                                   const string&      sliceDimName )
 : TCreader( ),
   my_tcr( ),
   my_mi( ),
   my_zdim( ),
   my_begin( this ),
   my_timeCourse( )
 {
  reinitialize( tcr, mi, sliceDimName );
 }

template< class TCR >
void
DataNavigator<TCR>::reinitialize( const TCR*         tcr,
                                  const MissingInfo* mi,
                                  const string&      sliceDimName )
 {
  // cerr << "(debug) DataNavigator::reinitialize( ) begin" << endl;
  assert( tcr != 0 ); assert( mi  != 0 );

  TCreader::reinitialize( tcr->sizes( ), tcr->tdim( ), tcr->order( ),
                          tcr->indexLower( ), tcr->names( ), false );

  for ( size_t d = 0; d < numDims( ); d++ )
   setRange( d, tcr->rangeLower( )[d], tcr->rangeUpper( )[d] );

  my_tcr  = tcr;
  my_mi   = mi;
  my_zdim = dimension( sliceDimName );

  my_timeCourse.resize( 0 );

  assert( isConsistent( ) );
  // cerr << "(debug) DataNavigator::reinitialize( ) end" << endl;
 }

// ------------------------------------------ voxel navigation through iterators
// -----------------------------------------------------------------------------

template< class TCR >
DataNavigator<TCR>::iterator&
DataNavigator<TCR>::begin( ) const
 {
  if ( my_begin ) toFirst( );

  return my_begin;
 }

template< class TCR >
DataNavigator<TCR>::iterator
DataNavigator<TCR>::end( ) const
 {
  iterator i;
  return i;
 }

// ------------------------------------------------- raw and edited time courses
// -----------------------------------------------------------------------------

template< class TCR >
const varray<double>&
DataNavigator<TCR>::timeCourse( ) const
 {
  if ( anyMove( ) )
   my_mi->compress( whereAmI( )[my_zdim], TCreader::timeCourse( ), my_timeCourse );

  return my_timeCourse;
 }

template< class TCR >
const varray<double>&
DataNavigator<TCR>::tcEntire( ) const
 {
  return TCreader::timeCourse( );
 }

// ----------------------------------------------- references to wrapped objects
// -----------------------------------------------------------------------------

template< class TCR >
const MissingInfo&
DataNavigator<TCR>::mi( ) const
 {
  return *my_mi;
 }

// ----------------------------------------------------- "slice" dimension index
// -----------------------------------------------------------------------------

template< class TCR >
size_t
DataNavigator<TCR>::sliceDim( ) const
 {
  return my_zdim;
 }

// ------------------------------------ iterator: construct/destruct/copy/assign
// -----------------------------------------------------------------------------

template< class TCR >
DataNavigator<TCR>::iterator::iterator( )
 : my_dn( 0 )
 {
 }

template< class TCR >
DataNavigator<TCR>::iterator::iterator( DataNavigator* dn )
 : my_dn( dn )
 {
 }

template< class TCR >
DataNavigator<TCR>::iterator::~iterator( )
 {
  if ( my_dn != 0 ) my_dn->my_begin.my_dn = my_dn;
  my_dn = 0;
 }

template< class TCR >
DataNavigator<TCR>::iterator::iterator( iterator& another )
 : my_dn( another.my_dn )
 {
  another.my_dn = 0;
 }

template< class TCR >
DataNavigator<TCR>::iterator&
DataNavigator<TCR>::iterator::operator=( iterator& rhs )
 {
  if ( this == &rhs ) return *this;

  if ( my_dn != 0 ) my_dn->my_begin.my_dn = my_dn;
             //
             // takes care of two iterators pointing to different DataNavigators

  my_dn = rhs.my_dn;
  rhs.my_dn = 0;

  return *this;
 }

// ------------------------------------------------------- iterator: comparisons
// -----------------------------------------------------------------------------

template< class TCR >
DataNavigator<TCR>::iterator::operator bool( ) const
 {
  return ( my_dn != 0 );
 }

template< class TCR >
bool
DataNavigator<TCR>::iterator::operator==( const iterator& another ) const
 {
  return ( my_dn == another.my_dn );
 }

template< class TCR >
bool
DataNavigator<TCR>::iterator::operator!=( const iterator& another ) const
 {
  return ( my_dn != another.my_dn );
 }

// ------------------------------------------ iterator: increment/deref/location
// -----------------------------------------------------------------------------

template< class TCR >
DataNavigator<TCR>::iterator&
DataNavigator<TCR>::iterator::operator++( )
 {
  if ( !my_dn->toNext( ) ) { my_dn->my_begin.my_dn = my_dn; my_dn = 0; }
  return *this;
 }

template< class TCR >
const varray<double>&
DataNavigator<TCR>::iterator::operator*( ) const
 {
  return my_dn->timeCourse( );
 }

template< class TCR >
DataNavigator<TCR>*
DataNavigator<TCR>::iterator::operator->( ) const
 {
  return my_dn;
 }

template< class TCR >
DataNavigator<TCR>::LocationType
DataNavigator<TCR>::iterator::operator()( ) const
 {
  return my_dn->locSansT( );
 }

// ----------------------------------------------------------- seek and retrieve
// -----------------------------------------------------------------------------

template< class TCR >
bool
DataNavigator<TCR>::seek( ) const
 {
  my_tcr->moveTo( whereAmI( ) );
  return my_tcr->seek( );
 }

template< class TCR >
bool
DataNavigator<TCR>::retrieve( varray<double>& timeCourse ) const
 {
  setMove( false ); return my_tcr->retrieve( timeCourse );
 }

template< class TCR >
bool
DataNavigator<TCR>::retrieveFloat( varray<float>& timeCourse ) const
 {
  setMove( false ); return my_tcr->retrieveFloat( timeCourse );
 }

template< class TCR >
bool
DataNavigator<TCR>::retrieveInt( varray<int>& timeCourse ) const
 {
  setMove( false ); return my_tcr->retrieveInt( timeCourse );
 }

template< class TCR >
bool
DataNavigator<TCR>::retrieveShort( varray<short>& timeCourse ) const
 {
  setMove( false ); return my_tcr->retrieveShort( timeCourse );
 }

template< class TCR >
bool
DataNavigator<TCR>::retrieveUchar( varray<unsigned char>& timeCourse ) const
 {
  setMove( false ); return my_tcr->retrieveUchar( timeCourse );
 }

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

template< class TCR >
void
DataNavigator<TCR>::close( ) const
 {
  my_tcr->close( );
 }

template< class TCR >
string
DataNavigator<TCR>::dataType( ) const
 {
  return my_tcr->dataType( );
 }

// ------------------------------------------------ ArrayNavigator redefinitions
// -----------------------------------------------------------------------------

template< class TCR >
bool
DataNavigator<TCR>::toPrev( ) const
 {
  bool m = TCreader::toPrev( );
  my_tcr->moveTo( whereAmI( ) );
  my_tcr->setMove( anyMove( ) );
  return m;
 }

template< class TCR >
bool
DataNavigator<TCR>::toNext( ) const
 {
  bool m = TCreader::toNext( );
  my_tcr->moveTo( whereAmI( ) );
  my_tcr->setMove( anyMove( ) );
  return m;
 }

template< class TCR >
bool
DataNavigator<TCR>::toFirst( ) const
 {
  bool m = TCreader::toFirst( );
  my_tcr->moveTo( whereAmI( ) );
  my_tcr->setMove( anyMove( ) );
  return m;
 }

template< class TCR >
bool
DataNavigator<TCR>::toLast( ) const
 {
  bool m = TCreader::toLast( );
  my_tcr->moveTo( whereAmI( ) );
  my_tcr->setMove( anyMove( ) );
  return m;
 }

template< class TCR >
long
DataNavigator<TCR>::offset( ) const
 {
  long ofs = TCreader::offset( );
  my_tcr->moveTo( whereAmI( ) );
  my_tcr->offset( );
  my_tcr->setMove( anyMove( ) );
  return ofs;
 }

template< class TCR >
long
DataNavigator<TCR>::offsetB( ) const
 {
  long ofs = TCreader::offsetB( );
  my_tcr->moveTo( whereAmI( ) );
  my_tcr->offsetB( );
  my_tcr->setMove( anyMove( ) );
  return ofs;
 }

// --------------------------------------------------------------------- private
// -----------------------------------------------------------------------------

template< class TCR >
bool
DataNavigator<TCR>::isConsistent( ) const
 {
  return
   (
        ( my_tcr                 && my_mi                 )
     && ( my_zdim                != numDims( )            )
     && ( sizes( )[tdim( )]      == my_mi->numInstants( ) )
     && ( sizes( )[my_zdim]      == my_mi->numSlices( )   )
     && ( indexLower( )[tdim( )] == my_mi->tLower( )      )
     && ( indexLower( )[my_zdim] == my_mi->zLower( )      )
   );
 }

#endif /* __DataNavigator_H__ */
