/* ----------------------------------------------------------------------------
 *
 * class TChandler
 *   base class for readers and writers for  specific time-course data formats.
 *   A time-course is defined to be a scalar function of one discrete  variable
 *   (time).  Time-course data  is visualized as  a multi-dimensional array  of
 *   constant-length time courses.
 *
 * required headers
 *   see include header file TChandler
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __TChandler_H__
#define __TChandler_H__

cpp_using_namespace( std );
cpp_using_namespace( BasicTypes );

class TChandler : public ArrayNavigator
 {
  public:

      // A time course is a vector of values. The data is visualized as a multi-
      // dimensional array of same-length time courses. For given  set of  array
      // indices, time course handler provides  a base structure  for retrieving
      // or storing the corresponding time course (typically from or into a file
      // in some format) or portion thereof. A time course handler also provides
      // basic navigational capabilities.  Coordinate indices  specifying a time
      // course need not necessarily be 0-based. Further, data navigation can be
      // restricted to a subrange/slice/rectangular  region of the data. Setting
      // range for the time dimension leads to accessing (reading/writing)  that
      // portion  of  a time course. Individual time course readers  and writers
      // derived from this base may demand that the  t-dimension  be the leading
      // non-trivial  (i. e.,  the  fastest varying)  dimension.  However,  that
      // assumption is not part of this base class.

   virtual
   ~TChandler( ) = 0;

  // ----------------------------------------------------------- dimension names
  // ---------------------------------------------------------------------------

   inline
   const vector<string>&
   names( ) const;
                                             //
                                             // names associated with dimensions

   inline
   bool
   isValid( const string& name ) const;
                                  //
                                  // true if name is associated with a dimension

   inline
   size_t
   dimension( const string& name ) const;
               //
               // dimension associated with name; numDims( ) if !isValid( name )

   inline
   string
   nameStr( const string&     separator  = ",",
            const uintVector& dimsSubset = uintVector( ) ) const;
           //
           // names( )[dimsSubset]  merged into  a single string,  separated  by
           // separator. dimsSubset == uintVector( ) defaults to all dimensions.

   inline
   string
   locationStr( const string&     separator  = ",",
                const uintVector& dimsSubset = uintVector( ) ) const;
                      //
                      // current location  in  the  format  "name=location,...".
                      // dimsSubset == uintVector( ) defaults to all dimensions.

  // ----------------------------------------------------------- the t-dimension
  // ---------------------------------------------------------------------------

   inline
   size_t
   tdim( ) const;
                          //
                          // position of the t-dimension in the dimensions list.
                          // E. g.,  names()[tdim()] is the name associated with
                          // the t-dimension;  dimension(names(tdim()))==tdim().

   inline
   bool
   isLeadDimT( ) const;
           //
           // is t-dimension the leading nontrivial (fastest varying) dimension?

   inline
   size_t
   leadDimNT( ) const;
                       //
                       // leading nontrivial non-t dimension; numDims( ) if none

  // ------------------------------------------- attributes sans the t-dimension
  // ---------------------------------------------------------------------------

   inline
   const uintVector&
   dimsSansT( ) const;
                                  //
                                  // dimension indices excluding the t-dimension

   inline
   LocationType
   locSansT( ) const;
                //
                // coordinates of the current location excluding the t-dimension

   inline
   uintVector
   sizesSansT( ) const;
                                //
                                // dimensional extents excluding the t-dimension

   inline
   uintVector
   lowerSansT( ) const;
                     //
                     // dimensional index lower bounds excluding the t-dimension

  // ---------------------------------------------- ArrayNavigator redefinitions
  // ---------------------------------------------------------------------------

                                     // this is to make sure that the t-index is
                                     // always  locked at rangeLower( )[my_tdim]

   inline
   bool
   toPrev( ) const;

   inline
   bool
   toNext( ) const;

   inline
   bool
   toLast( ) const;

   inline
   long
   offset( ) const;

   inline
   long
   offsetB( ) const;                        // same as ArrayNavigator::offset( )

  // ---------------------------------------------------------- file positioning
  // ---------------------------------------------------------------------------

   virtual
   bool
   seek( ) const = 0;
                          //
                          // position file to access time course at whereAmI( ).
                          // must set the t-coordinate to rangeLower( )[tdim( )]
                          // explicitly. report success.

  // --------------------------------------- data type of stored timecourse data
  // ---------------------------------------------------------------------------

   virtual
   string
   dataType( ) const = 0;

  // ------------------------------------------------------- explicit file close
  // ---------------------------------------------------------------------------

   virtual
   void
   close( ) const = 0;

  protected:

  // ---------------------------------------------------------- construct/reinit
  // ---------------------------------------------------------------------------

   TChandler( );

   TChandler( const TChandler& another );
   TChandler& operator=( const TChandler& rhs );

                   // sizes.size( ) == number of data dimensions including time.
                   // default name for dimension  tDimension  is  "t".  If  name
                   // for any other dimension is unavailable, or is identical to
                   // any  existing  name,  a  suitable default  is substituted.
                   // deleteTrivial == true deletes dimensions with  extent == 1
                   // (except for the designated time dimension).

   TChandler( const uintVector&     sizes,
                    size_t          tDimension    = 0,
              const uintVector&     indexOrder    = LocationType( ),
              const LocationType&   lowerBounds   = LocationType( ),
              const vector<string>& names         = vector<string>( ),
                    bool            deleteTrivial = true );

   void
   reinitialize( const uintVector&     sizes,
                       size_t          tDimension    = 0,
                 const uintVector&     indexOrder    = LocationType( ),
                 const LocationType&   lowerBounds   = LocationType( ),
                 const vector<string>& names         = vector<string>( ),
                       bool            deleteTrivial = true );

   void
   setNames( const vector<string>& names, bool excludeT = false );

   void
   setName( size_t dimension, const string& name, bool excludeT = false );
                                        //
                                        // assume my_names.size( ) == numDims( )

  private:

   size_t         my_tdim;
   vector<string> my_names;
   uintVector     my_dimsSansT;
   bool           my_isLeadDimT;
   size_t         my_leadDimNT;
 };

// ------------------------------------------------------------- dimension names
// -----------------------------------------------------------------------------

const vector<string>&
TChandler::names( ) const
 {
  return my_names;
 }

size_t
TChandler::dimension( const string& name ) const
 {
  size_t d = 0;
  while ( ( d < numDims( ) ) && ( my_names[d] != name ) ) d++;
  return d;
 }

bool
TChandler::isValid( const string& name ) const
 {
  return ( dimension( name ) < numDims( ) );
 }

string
TChandler::nameStr( const string& separator, const uintVector& dimsSubset ) const
 {
  if ( dimsSubset.size( ) == 0 )
   {
    string ds;
    for ( size_t i = 0; i < my_names.size( ); i++ )
     {
      if ( i > 0 ) ds.append( separator );
      ds.append( my_names[i] );
     }

    return ds;
   }

  size_t i = 0;
  for ( ; ( i < dimsSubset.size( ) ) && ( dimsSubset[i] >= numDims( ) ); i++ )

  if ( i == dimsSubset.size( ) ) return string( );

  string ds( my_names[dimsSubset[i]] );
  for ( i += 1; i < dimsSubset.size( ); i++ )
   if ( dimsSubset[i] < numDims( ) )
    {
     ds.append( separator );
     ds.append( my_names[dimsSubset[i]] );
    }

  return ds;
 }

string
TChandler::locationStr( const string& separator, const uintVector& dimsSubset ) const
 {
  char s[32];
  string ds;

  if ( dimsSubset.size( ) == 0 )
   {
    for ( size_t i = 0; i < my_names.size( ); i++ )
     {
      if ( i > 0 ) ds.append( separator );
      sprintf( s, "%d", whereAmI( )[i] );
      ds.append( my_names[i] ); ds.append( "=" ); ds.append( string( s ) );
     }

    return ds;
   }

  size_t i = 0;
  for ( ; ( i < dimsSubset.size( ) ) && ( dimsSubset[i] >= numDims( ) ); i++ )

  if ( i == dimsSubset.size( ) ) return string( );

  ds = my_names[dimsSubset[i]]; ds.append( "=" );
  sprintf( s, "%d", whereAmI( )[dimsSubset[i]] );
  ds.append( string( s ) );
  for ( i += 1; i < dimsSubset.size( ); i++ )
   if ( dimsSubset[i] < numDims( ) )
    {
     ds.append( separator );
     ds.append( my_names[dimsSubset[i]] ); ds.append( "=" );
     sprintf( s, "%d", whereAmI( )[dimsSubset[i]] );
     ds.append( string( s ) );
    }

  return ds;
 }

// ------------------------------------------------------------- the t-dimension
// -----------------------------------------------------------------------------

size_t
TChandler::tdim( ) const
 {
  return my_tdim;
 }

bool
TChandler::isLeadDimT( ) const
 {
  return my_isLeadDimT;
 }

size_t
TChandler::leadDimNT( ) const
 {
  return my_leadDimNT;
 }

// --------------------------------------- current location sans the t-dimension
// -----------------------------------------------------------------------------

const uintVector&
TChandler::dimsSansT( ) const
 {
  return my_dimsSansT;
 }

ArrayNavigator::LocationType
TChandler::locSansT( ) const
 {
  LocationType l( my_dimsSansT.size( ) );
  for ( size_t d = 0; d < l.size( ); d++ ) l[d] = whereAmI( )[my_dimsSansT[d]];
  // for ( size_t d = 0, i = 0; d < numDims( ); d++ )
  //  if ( d != my_tdim ) { l[i] = whereAmI( )[d]; i++; }
  return l;
 }

uintVector
TChandler::sizesSansT( ) const
 {
  uintVector s( my_dimsSansT.size( ) );
  for ( size_t d = 0; d < s.size( ); d++ ) s[d] = sizes( )[my_dimsSansT[d]];
  return s;
 }

uintVector
TChandler::lowerSansT( ) const
 {
  uintVector l( my_dimsSansT.size( ) );
  for ( size_t d = 0; d < l.size( ); d++ ) l[d] = indexLower( )[my_dimsSansT[d]];
  return l;
 }

// ------------------------------------------------ ArrayNavigator redefinitions
// -----------------------------------------------------------------------------

long
TChandler::offset( ) const
 {
              // ArrayNavigator::anyMove( ) indicates a change in location since
              // the  last  call  to  ArrayNavigator::offset( ).  Since,  e. g.,
              // TCreader::timeCourse() uses ArrayNavigator::anyMove() to decide
              // whether to update TCreader::my_timeCourse,  the internal  state
              // of the underlying ArrayNavigator  needs  to  be  restored  when
              // offset( ) is invoked.  It is important for  seek( )  and  other
              // low-level  access functions to invoke ArrayNavigator::offset( )
              // instead of this offset( ): this function provides access to the
              // ArrayNavigator::offset( )  if invoked  directly by a user,  and
              // restores the move status of the underlying ArrayNavigator.

  bool move = ArrayNavigator::anyMove( );
  long ofs  = ArrayNavigator::offset( );
  setMove( move ); return ofs;
 }

long
TChandler::offsetB( ) const
 {
  return ArrayNavigator::offset( );
 }

bool
TChandler::toPrev( ) const
 {
  // this is to make sure that t-index is at rangeLower( )[my_tdim]

  moveTo( my_tdim, rangeLower( )[my_tdim] );
  bool b = ArrayNavigator::toPrev( );
  moveTo( my_tdim, rangeLower( )[my_tdim] );
  return b;
 }

bool
TChandler::toNext( ) const
 {
  // this is to make sure that t-index is at rangeLower( )[my_tdim]

  moveTo( my_tdim, rangeUpper( )[my_tdim] );
  return ArrayNavigator::toNext( );
 }

bool
TChandler::toLast( ) const
 {
  // this is to make sure that t-index is at rangeLower( )[my_tdim]

  bool b = ArrayNavigator::toLast( );
  moveTo( my_tdim, rangeLower( )[my_tdim] );
  return b;
 }

#endif /* __TChandler_H__ */
