/* ----------------------------------------------------------------------------
 *
 * class TCrwPghMRI
 *   reader/writer for time-course data in Pittsburgh MRI format
 *
 * required headers
 *   see include header file TCrwPghMRI
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __TCrwPghMRI_H__
#define __TCrwPghMRI_H__

cpp_using_namespace( std );

class TCrwPghMRI : public TCrw
 {
  public:

                // It is preferable to have the data chunk ordered externally so
                // that the time dimension is the  leading nontrivial dimension.
                // Although this reader  is capable of  handling the other case,
                // the retrieve operation may be slow in that case.

   TCrwPghMRI( const TCrwPghMRI& another );

   TCrwPghMRI&
   operator=( const TCrwPghMRI& rhs );

   TCrwPghMRI( );

   TCrwPghMRI( const string&                     fileName,
               const string&                     chunkName     = "images",
                     char                        tDimName      = 't',
                     bool                        deleteTrivial = true,
                     PghMRIhandler::mriErrorMode errorMode     = PghMRIhandler::mriAbortOnError );
                     //
                     // open read-only. deleteTrivial == true deletes dimensions
                     // with extent=1, except for the designated time dimension.
                     //
                     // Standard  Pittsburgh MRI  chunk descriptors  of the form
                     //     chunk, chunk.file, chunk.datatype, chunk.dimensions,
                     //     chunk.extent.?, chunk.offset,    chunk.little_endian
                     // are required for read and modify.
                     //
                     // In addition,  lower bounds on indices  may be  specified
                     // through descriptors of the form chunk.min.?.

   TCrwPghMRI( const string&                     fileName,
                     char                        tDimName      = 't',
               const string&                     chunkName     = "images",
                     bool                        deleteTrivial = true,
                     PghMRIhandler::mriErrorMode errorMode     = PghMRIhandler::mriAbortOnError );
                     //
                     // open for  data modification  only. Chunk characteristics
                     // such as extents cannot be changed.  Notice change in the
                     // order of the second and third arguments compared to  the
                     // read-only constructor.

   TCrwPghMRI( const string&                     fileName,
               const string&                     dimensions,
               const uintVector&                 sizes,
                     char                        tDimName      = 't',
               const string&                     chunkName     = "images",
               const LocationType&               chunkLower    = LocationType( ),
                     PghMRIhandler::mriArrayType chunkType     = PghMRIhandler::mriDouble,
               const string&                     chunkFile     = ".dat",
                     bool                        littleEndian  = PghMRIhandler::endianness( ),
                     PghMRIhandler::mriErrorMode errorMode     = PghMRIhandler::mriAbortOnError );
                     //
                     // create  a new file or overwrite an existing one with the
                     // same name completely.  PghMRIhandler::mriDouble  is used
                     // as default  for unsupported values of  chunkType.  Note:
                     // although a TChandler allows strings for dimension names,
                     // the  Pittsburgh MRI format allows only single characters
                     // for dimesnsion names.

   TCrwPghMRI( const string&                     fileName,
               const TChandler*                  tch,
               const string&                     chunkName    = "images",
                     PghMRIhandler::mriArrayType chunkType    = PghMRIhandler::mriDouble,
               const string&                     chunkFile    = ".dat",
                     bool                        littleEndian = PghMRIhandler::endianness( ),
                     PghMRIhandler::mriErrorMode errorMode    = PghMRIhandler::mriAbortOnError );
               //
               // write: dimensions, sizes and chunkLower are obtained from tch.
               // tch dimension names  must be single character to be compatible
               // with the Pittsburgh MRI format.

  // ----------------------------- open: read-only, modify-data-only, write-only
  // ---------------------------------------------------------------------------

   void
   readOnly( const string&                     fileName,
             const string&                     chunkName     = "images",
                   char                        tDimName      = 't',
                   bool                        deleteTrivial = true,
                   PghMRIhandler::mriErrorMode errorMode     = PghMRIhandler::mriAbortOnError );

   void
   modifyData( const string&                     fileName,
                     char                        tDimName      = 't',
               const string&                     chunkName     = "images",
                     bool                        deleteTrivial = true,
                     PghMRIhandler::mriErrorMode errorMode     = PghMRIhandler::mriAbortOnError );

   void
   write( const string&                     fileName,
          const string&                     dimensions,
          const uintVector&                 sizes,
                char                        tDimName,
          const string&                     chunkName,
          const LocationType&               chunkLower,
                PghMRIhandler::mriArrayType chunkType,
          const string&                     chunkFile,
                bool                        littleEndian,
                PghMRIhandler::mriErrorMode errorMode );

   void
   write( const string&                     fileName,
          const TChandler*                  tch,
          const string&                     chunkName    = "images",
                PghMRIhandler::mriArrayType chunkType    = PghMRIhandler::mriDouble,
          const string&                     chunkFile    = ".dat",
                bool                        littleEndian = PghMRIhandler::endianness( ),
                PghMRIhandler::mriErrorMode errorMode    = PghMRIhandler::mriAbortOnError );

  // ------------------------------------------ Pittsburgh MRI handler and chunk
  // ---------------------------------------------------------------------------

   inline
   PghMRIhandler&
   mriHandler( ) const;

   inline
   const string&
   chunkName( ) const;

  // -------------------------------------- retrieve current-location timecourse
  // ---------------------------------------------------------------------------

                      // all the retrieve functions return false  and do nothing
                      // if mriHandler( ).openMode( ) == PghMRIhandler::mriWrite

   inline
   bool
   retrieve( varray<double>& timeCourse ) const;

   inline
   bool
   retrieveFloat( varray<float>& timeCourse ) const;

   inline
   bool
   retrieveInt( varray<int>& timeCourse ) const;

   inline
   bool
   retrieveShort( varray<short>& timeCourse ) const;

   inline
   bool
   retrieveUchar( varray<unsigned char>& timeCourse ) const;

  // ---------------------------------- store timecourse at the current location
  // ---------------------------------------------------------------------------

                       // all the store functions  return false  and  do nothing
                       // if mriHandler( ).openMode( ) == PghMRIhandler::mriRead

   inline
   bool
   store( const varray<double>& timeCourse ) const;

   inline
   bool
   storeFloat( const varray<float>& timeCourse ) const;

   inline
   bool
   storeInt( const varray<int>& timeCourse ) const;

   inline
   bool
   storeShort( const varray<short>& timeCourse ) const;

   inline
   bool
   storeUchar( const varray<unsigned char>& timeCourse ) const;

  // ---------------------------------------------------------- file positioning
  // ---------------------------------------------------------------------------

   inline
   bool
   seek( ) const;

  // --------------------------------------- data type of stored timecourse data
  // ---------------------------------------------------------------------------

   inline
   string
   dataType( ) const;

  // ------------------------------------------------------- explicit file close
  // ---------------------------------------------------------------------------

   inline
   void
   close( ) const;

  private:
   string my_chunkName;
   long   my_tskip;

   mutable PghMRIhandler my_mriHandler;

   void openRM( char tDimName, bool deleteTrivial );
 };

// -------------------------------------------- Pittsburgh MRI handler and chunk
// -----------------------------------------------------------------------------

PghMRIhandler&
TCrwPghMRI::mriHandler( ) const
 {
  return my_mriHandler;
 }

const string&
TCrwPghMRI::chunkName( ) const
 {
  return my_chunkName;
 }

// ---------------------------------------- retrieve current-location timecourse
// -----------------------------------------------------------------------------

bool
TCrwPghMRI::retrieve( varray<double>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriWrite ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    double* tc
     = (double*) my_mriHandler.readChunk( my_chunkName, timeCourse.size( ),
                                          ofs, PghMRIhandler::mriDouble,
                                          timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    double* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (double*) my_mriHandler.readChunk( my_chunkName, 1, ofs,
                                              PghMRIhandler::mriDouble,
                                              timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

bool
TCrwPghMRI::retrieveFloat( varray<float>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriWrite ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    float* tc
     = (float*) my_mriHandler.readChunk( my_chunkName, timeCourse.size( ),
                                         ofs, PghMRIhandler::mriFloat,
                                         timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    float* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (float*) my_mriHandler.readChunk( my_chunkName, 1,
                                             ofs, PghMRIhandler::mriFloat,
                                             timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

bool
TCrwPghMRI::retrieveInt( varray<int>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriWrite ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    int* tc
     = (int*) my_mriHandler.readChunk( my_chunkName, timeCourse.size( ),
                                       ofs, PghMRIhandler::mriInt,
                                       timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    int* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (int*) my_mriHandler.readChunk( my_chunkName, 1,
                                           ofs, PghMRIhandler::mriInt,
                                           timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

bool
TCrwPghMRI::retrieveShort( varray<short>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriWrite ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    short* tc
     = (short*) my_mriHandler.readChunk( my_chunkName, timeCourse.size( ),
                                         ofs, PghMRIhandler::mriShort,
                                         timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    short* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (short*) my_mriHandler.readChunk( my_chunkName, 1,
                                             ofs, PghMRIhandler::mriShort,
                                             timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

bool
TCrwPghMRI::retrieveUchar( varray<unsigned char>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriWrite ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    unsigned char* tc
     = (unsigned char*) my_mriHandler.readChunk( my_chunkName, timeCourse.size( ),
                                                 ofs, PghMRIhandler::mriUchar,
                                                 timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    unsigned char* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (unsigned char*) my_mriHandler.readChunk( my_chunkName, 1,
                                                     ofs, PghMRIhandler::mriUchar,
                                                     timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

// ------------------------------------ store timecourse at the current location
// -----------------------------------------------------------------------------

bool
TCrwPghMRI::store( const varray<double>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriRead ) return false;

  varray<double> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   my_mriHandler.setChunk( my_chunkName, n, ofs, PghMRIhandler::mriDouble, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler.setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriDouble, tc( i ) );

  setMove( true );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

bool
TCrwPghMRI::storeFloat( const varray<float>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriRead ) return false;

  varray<float> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   my_mriHandler.setChunk( my_chunkName, n, ofs, PghMRIhandler::mriFloat, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler.setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriFloat, tc( i ) );

  setMove( true );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

bool
TCrwPghMRI::storeInt( const varray<int>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriRead ) return false;

  varray<int> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   my_mriHandler.setChunk( my_chunkName, n, ofs, PghMRIhandler::mriInt, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler.setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriInt, tc( i ) );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

bool
TCrwPghMRI::storeShort( const varray<short>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriRead ) return false;

  varray<short> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   my_mriHandler.setChunk( my_chunkName, n, ofs, PghMRIhandler::mriShort, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler.setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriShort, tc( i ) );

  setMove( true );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

bool
TCrwPghMRI::storeUchar( const varray<unsigned char>& timeCourse ) const
 {
  if ( my_mriHandler.openMode( ) == PghMRIhandler::mriRead ) return false;

  varray<unsigned char> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   my_mriHandler.setChunk( my_chunkName, n, ofs, PghMRIhandler::mriUchar, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler.setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriUchar, tc( i ) );

  setMove( true );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

// --------------------------------------------------------- explicit file close
// -----------------------------------------------------------------------------

void
TCrwPghMRI::close( ) const
 {
  my_mriHandler.close( );
 }

// ----------------------------------------- data type of stored timecourse data
// -----------------------------------------------------------------------------

string
TCrwPghMRI::dataType( ) const
 {
  return my_mriHandler.getString( my_chunkName + ".datatype" );
 }

// ------------------------------------------------------------ file positioning
// -----------------------------------------------------------------------------

bool
TCrwPghMRI::seek( ) const
 {
  moveTo( tdim( ), rangeLower( )[tdim( )] );
  TChandler::offset( );
  return true;
 }

#endif /* __TCrwPghMRI_H__ */
