/* ----------------------------------------------------------------------------
 *
 * class TCrwPghMRI_W
 *   reader/writer for time-course data in Pittsburgh MRI format;
 *   acts as a wrapper for an external PghMRIhandler object.
 *
 * required headers
 *   see include header file TCrwPghMRI_W
 *
 * ----------------------------------------------------------------------------
 * $Id$
 * ---------------------------------------------------------------------------- */

#ifndef __TCrwPghMRI_W_H__
#define __TCrwPghMRI_W_H__

cpp_using_namespace( std );

class TCrwPghMRI_W : public TCrw
 {
  public:

                // It is preferable to have the data chunk ordered externally so
                // that the time dimension is the  leading nontrivial dimension.
                // Although this reader  is capable of  handling the other case,
                // the retrieve operation may be slow in that case.

  // ------------------- construct/open: read-only, modify-data-only, write-only
  // ---------------------------------------------------------------------------

   TCrwPghMRI_W( const TCrwPghMRI_W& another );

   TCrwPghMRI_W&
   operator=( const TCrwPghMRI_W& rhs );

   TCrwPghMRI_W( );

   TCrwPghMRI_W( PghMRIhandler* mriHandler );

   void
   setMRIhandler( PghMRIhandler* mriHandler );

                     // mriHandler is assumed to be in  mriRead or mriModifyData
                     // modes.  deleteTrivial == true  deletes  dimensions  with
                     // with extent=1, except for the designated time dimension.
                     //
                     // Standard  Pittsburgh MRI  chunk descriptors  of the form
                     //     chunk, chunk.file, chunk.datatype, chunk.dimensions,
                     //     chunk.extent.?, chunk.offset,    chunk.little_endian
                     // are required for read and modify.
                     //
                     // In addition,  lower bounds on indices  may be  specified
                     // through descriptors of the form chunk.min.?.

   TCrwPghMRI_W(       PghMRIhandler* mriHandler,
                 const string&        chunkName     = "images",
                       char           tDimName      = 't',
                       bool           deleteTrivial = true );

   void
   setChunk( const string& chunkName     = "images",
                   char    tDimName      = 't',
                   bool    deleteTrivial = true );

                  // create a new chunk. mriHandler is assumed to be in mriWrite
                  // or  mriModify  modes.  PghMRIhandler::mriDouble  is used as
                  // default for unsupported values of chunkType. Note: although
                  // TChandler allows strings for dimension names,the Pittsburgh 
                  // MRI format allows only single characters for dimesnsions.

   TCrwPghMRI_W(       PghMRIhandler*              mriHandler,
                 const string&                     dimensions,
                 const uintVector&                 sizes,
                       char                        tDimName      = 't',
                 const string&                     chunkName     = "images",
                 const LocationType&               chunkLower    = LocationType( ),
                       PghMRIhandler::mriArrayType chunkType     = PghMRIhandler::mriDouble,
                 const string&                     chunkFile     = ".dat",
                       bool                        littleEndian  = PghMRIhandler::endianness( ) );

   void
   setChunk( const string&                     dimensions,
             const uintVector&                 sizes,
                   char                        tDimName,
             const string&                     chunkName,
             const LocationType&               chunkLower,
                   PghMRIhandler::mriArrayType chunkType,
             const string&                     chunkFile,
                   bool                        littleEndian );

    // create new chunk: dimensions, sizes and chunkLower are obtained from tch.
    // tch dimension names must be single character  to be compatible  with  the
    // Pittsburgh MRI format.

   TCrwPghMRI_W(       PghMRIhandler*              mriHandler,
                 const TChandler*                  tch,
                 const string&                     chunkName    = "images",
                       PghMRIhandler::mriArrayType chunkType    = PghMRIhandler::mriDouble,
                 const string&                     chunkFile    = ".dat",
                       bool                        littleEndian = PghMRIhandler::endianness( ) );

   void
   setChunk( const TChandler*                  tch,
             const string&                     chunkName    = "images",
                   PghMRIhandler::mriArrayType chunkType    = PghMRIhandler::mriDouble,
             const string&                     chunkFile    = ".dat",
                   bool                        littleEndian = PghMRIhandler::endianness( ) );

  // ------------------------------------------ Pittsburgh MRI handler and chunk
  // ---------------------------------------------------------------------------

   inline
   PghMRIhandler&
   mriHandler( ) const;

   inline
   const string&
   chunkName( ) const;

  // -------------------------------------- retrieve current-location timecourse
  // ---------------------------------------------------------------------------

                      // all the retrieve functions return false  and do nothing
                      // if mriHandler( ).openMode( ) == PghMRIhandler::mriWrite

   inline
   bool
   retrieve( varray<double>& timeCourse ) const;

   inline
   bool
   retrieveFloat( varray<float>& timeCourse ) const;

   inline
   bool
   retrieveInt( varray<int>& timeCourse ) const;

   inline
   bool
   retrieveShort( varray<short>& timeCourse ) const;

   inline
   bool
   retrieveUchar( varray<unsigned char>& timeCourse ) const;

  // ---------------------------------- store timecourse at the current location
  // ---------------------------------------------------------------------------

                       // all the store functions  return false  and  do nothing
                       // if mriHandler( ).openMode( ) == PghMRIhandler::mriRead

   inline
   bool
   store( const varray<double>& timeCourse ) const;

   inline
   bool
   storeFloat( const varray<float>& timeCourse ) const;

   inline
   bool
   storeInt( const varray<int>& timeCourse ) const;

   inline
   bool
   storeShort( const varray<short>& timeCourse ) const;

   inline
   bool
   storeUchar( const varray<unsigned char>& timeCourse ) const;

  // ---------------------------------------------------------- file positioning
  // ---------------------------------------------------------------------------

   inline
   bool
   seek( ) const;

  // --------------------------------------- data type of stored timecourse data
  // ---------------------------------------------------------------------------

   inline
   string
   dataType( ) const;

  // ------------------------------------------------------- explicit file close
  // ---------------------------------------------------------------------------

   inline
   void
   close( ) const;

  private:
   string         my_chunkName;
   long           my_tskip;
   PghMRIhandler* my_mriHandler;

   static PghMRIhandler my_mri;
                                  //
                                  // default for mriHandler( ) if !my_mriHandler

   void openRM( char tDimName, bool deleteTrivial );
 };

// -------------------------------------------- Pittsburgh MRI handler and chunk
// -----------------------------------------------------------------------------

PghMRIhandler&
TCrwPghMRI_W::mriHandler( ) const
 {
  return my_mriHandler ? ( *my_mriHandler ) : my_mri;
 }

const string&
TCrwPghMRI_W::chunkName( ) const
 {
  return my_chunkName;
 }

// ---------------------------------------- retrieve current-location timecourse
// -----------------------------------------------------------------------------

bool
TCrwPghMRI_W::retrieve( varray<double>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriWrite ) ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    double* tc
     = (double*) my_mriHandler->readChunk( my_chunkName, timeCourse.size( ),
                                          ofs, PghMRIhandler::mriDouble,
                                          timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    double* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (double*) my_mriHandler->readChunk( my_chunkName, 1, ofs,
                                              PghMRIhandler::mriDouble,
                                              timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

bool
TCrwPghMRI_W::retrieveFloat( varray<float>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriWrite ) ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    float* tc
     = (float*) my_mriHandler->readChunk( my_chunkName, timeCourse.size( ),
                                         ofs, PghMRIhandler::mriFloat,
                                         timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    float* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (float*) my_mriHandler->readChunk( my_chunkName, 1,
                                             ofs, PghMRIhandler::mriFloat,
                                             timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

bool
TCrwPghMRI_W::retrieveInt( varray<int>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriWrite ) ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    int* tc
     = (int*) my_mriHandler->readChunk( my_chunkName, timeCourse.size( ),
                                       ofs, PghMRIhandler::mriInt,
                                       timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    int* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (int*) my_mriHandler->readChunk( my_chunkName, 1,
                                           ofs, PghMRIhandler::mriInt,
                                           timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

bool
TCrwPghMRI_W::retrieveShort( varray<short>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriWrite ) ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    short* tc
     = (short*) my_mriHandler->readChunk( my_chunkName, timeCourse.size( ),
                                         ofs, PghMRIhandler::mriShort,
                                         timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    short* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (short*) my_mriHandler->readChunk( my_chunkName, 1,
                                             ofs, PghMRIhandler::mriShort,
                                             timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

bool
TCrwPghMRI_W::retrieveUchar( varray<unsigned char>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriWrite ) ) return false;

  seek( ); long ofs = ArrayNavigator::offset( );

  if ( isLeadDimT( ) )
   {
    unsigned char* tc
     = (unsigned char*) my_mriHandler->readChunk( my_chunkName, timeCourse.size( ),
                                                 ofs, PghMRIhandler::mriUchar,
                                                 timeCourse( 0 ) );

    return ( tc != 0 );
   }
  else
   {
    unsigned char* tc;

    bool anyErrors = false;
    for ( size_t i = 0; i < timeCourse.size( ); i++, ofs += my_tskip )
     {
      tc = (unsigned char*) my_mriHandler->readChunk( my_chunkName, 1,
                                                     ofs, PghMRIhandler::mriUchar,
                                                     timeCourse( i ) );
      anyErrors = anyErrors || ( tc == 0 );
     }

    return anyErrors;
   }
 }

// ------------------------------------ store timecourse at the current location
// -----------------------------------------------------------------------------

bool
TCrwPghMRI_W::store( const varray<double>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriRead ) ) return false;

  varray<double> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  // cerr << "(debug) TCrwPghMRI_W::store( ) location " << locationStr( ",", dimsSansT( ) ) << " offset " << ofs << endl;

  if ( isLeadDimT( ) )
   my_mriHandler->setChunk( my_chunkName, n, ofs, PghMRIhandler::mriDouble, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler->setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriDouble, tc( i ) );

  setMove( true );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

bool
TCrwPghMRI_W::storeFloat( const varray<float>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriRead ) ) return false;

  varray<float> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  // cerr << "(debug) TCrwPghMRI_W::storeFloat( ) location " << locationStr( ",", dimsSansT( ) ) << " offset " << ofs << endl;

  if ( isLeadDimT( ) )
   my_mriHandler->setChunk( my_chunkName, n, ofs, PghMRIhandler::mriFloat, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler->setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriFloat, tc( i ) );

  setMove( true );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

bool
TCrwPghMRI_W::storeInt( const varray<int>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriRead ) ) return false;

  varray<int> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  // cerr << "(debug) TCrwPghMRI_W::storeInt( ) location " << locationStr( ",", dimsSansT( ) ) << " offset " << ofs << endl;

  if ( isLeadDimT( ) )
   my_mriHandler->setChunk( my_chunkName, n, ofs, PghMRIhandler::mriInt, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler->setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriInt, tc( i ) );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

bool
TCrwPghMRI_W::storeShort( const varray<short>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriRead ) ) return false;

  varray<short> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  // cerr << "(debug) TCrwPghMRI_W::storeShort( ) location " << locationStr( ",", dimsSansT( ) ) << " offset " << ofs << endl;

  if ( isLeadDimT( ) )
   my_mriHandler->setChunk( my_chunkName, n, ofs, PghMRIhandler::mriShort, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler->setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriShort, tc( i ) );

  setMove( true );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

bool
TCrwPghMRI_W::storeUchar( const varray<unsigned char>& timeCourse ) const
 {
  if ( !my_mriHandler || ( my_mriHandler->openMode( ) == PghMRIhandler::mriRead ) ) return false;

  varray<unsigned char> tc( timeCourse );
  size_t n = min( rangeUpper( )[tdim( )] - rangeLower( )[tdim( )] + 1, tc.size( ) );

  seek( ); long ofs = ArrayNavigator::offset( );

  // cerr << "(debug) TCrwPghMRI_W::storeUchar( ) location " << locationStr( ",", dimsSansT( ) ) << " offset " << ofs << endl;

  if ( isLeadDimT( ) )
   my_mriHandler->setChunk( my_chunkName, n, ofs, PghMRIhandler::mriUchar, tc( 0 ) );
  else
   for ( size_t i = 0; i < n; i++, ofs += my_tskip )
    my_mriHandler->setChunk( my_chunkName, 1, ofs, PghMRIhandler::mriUchar, tc( i ) );

  setMove( true );

  return true;

  // unfortunately, there is very little control over error reporting in libmri.
  // setChunk( ), e. g., does not report success/failure. setErrorMode( ) can be
  // used  to customize  error handling to some extent,  but that affects  *all*
  // existing instances of PghMRIhandler.
 }

// --------------------------------------------------------- explicit file close
// -----------------------------------------------------------------------------

void
TCrwPghMRI_W::close( ) const
 {
  if ( my_mriHandler ) my_mriHandler->close( );
 }

// ----------------------------------------- data type of stored timecourse data
// -----------------------------------------------------------------------------

string
TCrwPghMRI_W::dataType( ) const
 {
  return my_mriHandler ? my_mriHandler->getString( my_chunkName + ".datatype" ) : "";
 }

// ------------------------------------------------------------ file positioning
// -----------------------------------------------------------------------------

bool
TCrwPghMRI_W::seek( ) const
 {
  moveTo( tdim( ), rangeLower( )[tdim( )] );
  TChandler::offset( );
  return true;
 }

#endif /* __TCrwPghMRI_W_H__ */
