#!/bin/csh -f

# temporary files
#
 set myname = `basename ${0}`

 mkdir -p /tmp/`whoami`
 set tmp = /tmp/`whoami`/`echo ${myname} | sed 's/\.csh$//'`

 set verbmax   = 1
 set verbosity = ${verbmax}

# command line
#
 if ( ${#argv} < 6 ) then
  echo "${myname}"
  echo " make frequencywise spatial images for sqrt(power spectrum)."
  echo ""
  echo "usage"
  echo " `basename ${0}` infile dim2fft numchop iMin iMax outfile smopts"
  exit 0
 endif

 set infile  = $1   # that's the data to be ffted
 set dim2fft = $2   # e. g., t
 set numchop = $3   # number of initial points to chop off along dim2fft
 set iMin    = $4   # frequency number: print images from
 set iMax    = $5   # frequency number: print images to
 set outfile = $6   # this is the output postscript file containing frequencywise images
 set smopts  = "$7" # mri_smooth smoother options, e. g., "-smoother_type none"

# allowed frequency indices for frequencywise images:
# 1, ..., max div 2 - 1; 0 and pi (i. e., 0.5) excluded
#
 @ max      = `printmrifield -i ${infile} -f images.extent.${dim2fft}` - ${numchop}
 @ maxBy2   = ${max} / 2
 @ maxBy2p1 = ${maxBy2} + 1
 @ maxBy2m1 = ${maxBy2} - 1

# bounds on iMin, iMax
#
 if ( ${iMin} <= 0         ) @ iMin = 1
 if ( ${iMin} >= ${maxBy2} ) @ iMin = ${maxBy2} - 1

 if ( ${iMax} <= 0         ) @ iMax = 1
 if ( ${iMax} >= ${maxBy2} ) @ iMax = ${maxBy2} - 1

# chop off initial points
#
 mri_subset -d ${dim2fft} -o ${numchop} -r ${max} ${infile} ${tmp}

 errmsg.csh ${verbosity} ${myname} 2 "failed command" ${status} "mri_subset -d ${dim2fft} -o ${numchop} -r ${max} ${infile} ${tmp}"
 if ( ${status} ) exit ${status}

# fft
#
 mri_fft -d ${dim2fft} -f -m ${tmp} ${tmp}.fft.0

 errmsg.csh ${verbosity} ${myname} 3 "failed command" ${status} "mri_fft -d ${dim2fft} -f -m ${tmp} ${tmp}.fft.0"
 if ( ${status} ) exit ${status}

# subset positive half of the frequencies; exclude 0 and pi (i. e., 0.5)
#
 mri_subset -d ${dim2fft} -o ${maxBy2p1} -r ${maxBy2m1} ${tmp}.fft.0 ${tmp}.fft.1

 errmsg.csh ${verbosity} ${myname} 4 "failed command" ${status} "mri_subset -d ${dim2fft} -o ${maxBy2p1} -r ${maxBy2m1} ${tmp}.fft.0 ${tmp}.fft.1"
 if ( ${status} ) exit ${status}

# smooth
#
 mri_smooth -d ${dim2fft} ${smopts} ${tmp}.fft.1 ${tmp}.fft.2

 errmsg.csh ${verbosity} ${myname} 5 "failed command" ${status} "mri_smooth -d ${dim2fft} ${smopts} ${tmp}.fft.1 ${tmp}.fft.2"
 if ( ${status} ) exit ${status}

# make frequencywise images; stack them together in a single ps file
# by exploiting the structure of ps files produced by mri_to_img
#
 @ i      = ${iMin}
 @ offset = ${i} - 1
 while ( ${i} <= ${iMax} )
  mri_subset -d ${dim2fft} -o ${offset} -r 1 ${tmp}.fft.2 ${tmp}

  errmsg.csh ${verbosity} ${myname} 6 "failed command" ${status} "mri_subset -d ${dim2fft} -o ${offset} -r 1 ${tmp}.fft.2 ${tmp}"
  if ( ${status} ) exit ${status}

  mri_to_img -o ${tmp}.ps -all -r -g -mosaic ${tmp}

  errmsg.csh ${verbosity} ${myname} 7 "failed command" ${status} "mri_to_img -o ${tmp}.ps -all -r -g -mosaic ${tmp}"
  if ( ${status} ) exit ${status}

  if ( ${i} == ${iMin} ) then
   @ tpos = `grep -n '%%Title: ' ${tmp}.ps | awk -F: '{print $1}'` - 1

   head -${tpos} ${tmp}.ps     > ${outfile}

   echo "%%Title: ${outfile}" >> ${outfile}

   @ pages  = ${iMax} - ${iMin}
   @ pages += 1

   echo "%%Pages: ${pages}"   >> ${outfile}

   @ lines = `wc -l ${tmp}.ps | awk '{print $1}'` - ${tpos} + 1

   tail -${lines} ${tmp}.ps | sed 's/Page: 1 1 /Page: '${i}' '${i}'/' >> ${outfile}
  else
   @ lines = `wc -l ${tmp}.ps | awk '{print $1}'` - `grep -n EndProlog ${tmp}.ps | awk -F: '{print $1}'`

   tail -${lines} ${tmp}.ps | sed 's/Page: 1 1 /Page: '${i}' '${i}'/' >> ${outfile}
  endif

  @ i      += 1
  @ offset += 1
 end

# cleanup
#
 rm -f ${tmp}* >& /dev/null; exit 0
